#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>

#include "bcm_host.h"

#include "GLES/gl.h"
#include "EGL/egl.h"
#include "EGL/eglext.h"

extern "C"
{
	typedef struct 
	{
		uint32_t	screenOriginX;
		uint32_t	screenOriginY;
		uint32_t 	screenWidth;
		uint32_t 	screenHeight;

		uint32_t 	desktopWidth;
		uint32_t 	desktopHeight;


		EGLDisplay 	display;
		EGLSurface 	surface;
		EGLContext 	context;
	} DISP_STATE_T;
	DISP_STATE_T gGLstate; // global

	static int width = 640, height=480;

	void NativeEGLInitParams(int set_width, int set_height)
	{
		width=set_width; height=set_height;
	}

	void NativeEGLInfo(int* pwidth, int* pheight)
	{
		*pwidth = gGLstate.desktopWidth;
		*pheight= gGLstate.desktopHeight;
	}

	int NativeEGLInitOpenGL()
	{
		bcm_host_init();
		memset( &gGLstate, 0, sizeof( gGLstate ) );

		int32_t success = 0;
		EGLBoolean result;
		EGLint num_config;
		static EGL_DISPMANX_WINDOW_T nativewindow;
		DISPMANX_ELEMENT_HANDLE_T dispman_element;
		DISPMANX_DISPLAY_HANDLE_T dispman_display;
		DISPMANX_UPDATE_HANDLE_T dispman_update;

		// BIZARRE!
		// Need 255 here.  Providing 0 will produce fully transparent
		// 127 will obviously be half way
		VC_DISPMANX_ALPHA_T alpha = { DISPMANX_FLAGS_ALPHA_FIXED_ALL_PIXELS, 255, 0 };

		VC_RECT_T dst_rect;
		VC_RECT_T src_rect;

 
#if 0
	// consts
	#define COLOURDEPTH_RED_SIZE  		5
	#define COLOURDEPTH_GREEN_SIZE 		6
	#define COLOURDEPTH_BLUE_SIZE 		5
	#define COLOURDEPTH_DEPTH_SIZE		16
	static const EGLint attribute_list[] =
		{
//		EGL_CONTEXT_CLIENT_VERSION,	1,

		EGL_BUFFER_SIZE,			0,
/*
		EGL_RED_SIZE,				0,
		EGL_GREEN_SIZE,				0,
		EGL_BLUE_SIZE,				0,
		EGL_DEPTH_SIZE,				0,
*/
		EGL_LUMINANCE_SIZE,			0,
		EGL_ALPHA_SIZE,				0,
		EGL_CONFIG_CAVEAT,			EGL_DONT_CARE,
		EGL_CONFIG_ID,				EGL_DONT_CARE,
		EGL_LEVEL,					0,
		EGL_NATIVE_RENDERABLE,		EGL_DONT_CARE,
		EGL_NATIVE_VISUAL_TYPE,		EGL_DONT_CARE,
		EGL_SAMPLE_BUFFERS,			0,
		EGL_SAMPLES,				0,
		EGL_STENCIL_SIZE,			0,
		EGL_TRANSPARENT_TYPE,		EGL_NONE,
		EGL_TRANSPARENT_RED_VALUE,	EGL_DONT_CARE,
		EGL_TRANSPARENT_GREEN_VALUE,EGL_DONT_CARE,
		EGL_TRANSPARENT_BLUE_VALUE,	EGL_DONT_CARE,

/*
*/
		EGL_RED_SIZE,		  	    COLOURDEPTH_RED_SIZE,
		EGL_GREEN_SIZE,    		    COLOURDEPTH_GREEN_SIZE,
		EGL_BLUE_SIZE,     		    COLOURDEPTH_BLUE_SIZE,
		EGL_DEPTH_SIZE,				COLOURDEPTH_DEPTH_SIZE,

		EGL_SURFACE_TYPE,         EGL_WINDOW_BIT,
		//EGL_SURFACE_TYPE,       EGL_PBUFFER_BIT,
		EGL_RENDERABLE_TYPE,      EGL_OPENGL_ES_BIT,// 0
//		EGL_BIND_TO_TEXTURE_RGBA, EGL_TRUE,
		EGL_NONE
		};
#else
		static const EGLint attribute_list[] = 
		{
			//EGL_RENDERABLE_TYPE, EGL_OPENGL_ES_BIT,
			//EGL_CONFORMANT,      EGL_OPENGL_ES_BIT,

			EGL_RED_SIZE,   8,
			EGL_GREEN_SIZE, 8,
			EGL_BLUE_SIZE,  8,
			EGL_ALPHA_SIZE, 8,
			EGL_DEPTH_SIZE, 24,
/*
*/

			//EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
			EGL_SURFACE_TYPE, EGL_PBUFFER_BIT,

			EGL_NONE
		};
#endif

		EGLConfig config;

		gGLstate.display = eglGetDisplay(EGL_DEFAULT_DISPLAY);
		assert(gGLstate.display!=EGL_NO_DISPLAY);

		result = eglInitialize(gGLstate.display, NULL, NULL);
		assert(EGL_FALSE != result);

		result = eglChooseConfig(gGLstate.display, attribute_list, &config, 1, &num_config);
		assert(EGL_FALSE != result);

		result = eglBindAPI(EGL_OPENGL_ES_API);
		assert(EGL_FALSE != result);

		gGLstate.context = eglCreateContext(gGLstate.display, config, EGL_NO_CONTEXT, NULL);
		assert(gGLstate.context!=EGL_NO_CONTEXT);

		success = graphics_get_display_size(0, &gGLstate.desktopWidth, &gGLstate.desktopHeight);
		assert( success >= 0 );

		// Pi will not make window with border - always go fullscreen
		if(1) // width < 0 || width > gGLstate.desktopWidth || height < 0 || height > gGLstate.desktopHeight) 
		{
			printf("pi goes fullscreen %dx%d\n", gGLstate.desktopWidth, gGLstate.desktopHeight);
			// Fullscreen
			gGLstate.screenOriginX = 0;
			gGLstate.screenOriginY = 0;
			gGLstate.screenWidth  =gGLstate.desktopWidth ;
			gGLstate.screenHeight =gGLstate.desktopHeight;
		}
		else
		{
			// Centered Window
			printf("pi goes windowed %dx%d\n", width, height);
			gGLstate.screenOriginX = (gGLstate.desktopWidth  - width)  / 2;
			gGLstate.screenOriginY = (gGLstate.desktopWidth  - height) / 2;
			gGLstate.screenWidth  = width;
			gGLstate.screenHeight = height;

			// dummy @ 0,0
			gGLstate.screenOriginX = 0;
			gGLstate.screenOriginY = 0;
			gGLstate.screenWidth  =width;
			gGLstate.screenHeight =height;
		}
		dst_rect.x = gGLstate.screenOriginX;
		dst_rect.y = gGLstate.screenOriginY;
		dst_rect.width = gGLstate.screenWidth;
		dst_rect.height = gGLstate.screenHeight;
      
		src_rect.x = gGLstate.screenOriginX;
		src_rect.y = gGLstate.screenOriginY;
		src_rect.width = gGLstate.screenWidth << 16;
		src_rect.height = gGLstate.screenHeight << 16;        

		dispman_display = vc_dispmanx_display_open( 0);
		dispman_update = vc_dispmanx_update_start( 0 );
         
		dispman_element = vc_dispmanx_element_add ( dispman_update, dispman_display,
			0, &dst_rect, 0, &src_rect, DISPMANX_PROTECTION_NONE, &alpha, 0, DISPMANX_NO_ROTATE);
      
		nativewindow.element = dispman_element;
		nativewindow.width = gGLstate.screenWidth;
		nativewindow.height = gGLstate.screenHeight;
		vc_dispmanx_update_submit_sync( dispman_update );
      
		gGLstate.surface = eglCreateWindowSurface( gGLstate.display, config, &nativewindow, NULL );
		assert(gGLstate.surface != EGL_NO_SURFACE);

		result = eglMakeCurrent(gGLstate.display, gGLstate.surface, gGLstate.surface, gGLstate.context);
		assert(EGL_FALSE != result);

		glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
		glClear( GL_COLOR_BUFFER_BIT );
		
		glEnable( GL_DEPTH_TEST );
		glClear( GL_DEPTH_BUFFER_BIT );
		glDisable( GL_DEPTH_TEST );
		
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrthof(0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f);

		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_FALSE);


	}

	void NativeEGLTerminateOpenGL()
	{
		glClear( GL_COLOR_BUFFER_BIT );
		eglSwapBuffers(gGLstate.display, gGLstate.surface);

		eglMakeCurrent( gGLstate.display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT );
		eglDestroySurface( gGLstate.display, gGLstate.surface );
		eglDestroyContext( gGLstate.display, gGLstate.context );
		eglTerminate( gGLstate.display );
	}
 
 
	int NativeEGLSwapBuffers()
	{
		eglSwapBuffers(gGLstate.display, gGLstate.surface);
	}



void glb_pi_test()
{

static GLfloat triangle[] = 
{
	-0.5f, -0.5f, 0.0f,
	0.0f, 0.5f, 0.0f,
	0.5f, -0.5f, 0.0f
};

static GLfloat colors[] =
{
	0.0f, 1.0f, 0.0f, 1.0f,
	1.0f, 0.0f, 0.0f, 1.0f,
	0.0f, 0.0f, 1.0f, 1.0f
};
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrthof(0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f);
	glClearColor( 1.0f, 0.0f, 0.0f, 1.0f );
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, triangle);
	glColorPointer(4, GL_FLOAT, 0, colors);

	glShadeModel(GL_SMOOTH);
	glEnable(GL_DITHER);

	glDrawArrays(GL_TRIANGLE_STRIP, 0, 3);

	NativeEGLSwapBuffers();
}



}// extern "C"
