// - HEADERS ---------------------------------------------------
#if defined(TINYGL) || defined(TRENKI3D)
#include "stdafx.h"

#include "obitmap.h"
#include <math.h>

#ifdef TINYGL
	#include <GL/zgl.h>
#endif

#include "openglimage.h"

// - GLOBALS ---------------------------------------------------

// - FUNCTIONS -------------------------------------------------

// GP2X - Blitter
#if defined(GP2X)
	extern unsigned int GetUpperRealAddress(void* pVirtual);
	extern void gp2x_blit(unsigned int hardware_src, int x_from, int y_from, int src_stride_bytes, int x, int y, int w, int h);
	extern void gp2x_blit_cookie(unsigned int hardware_src, int x_from, int y_from, int src_stride_bytes, int x, int y, int w, int h, unsigned short cookie);
#endif


#ifdef _DEBUG
void dbg_gp2x_blit(unsigned char* src, int x_from, int y_from, int src_stride_bytes, int x, int y, int w, int h)
{
	src+=x_from;
	src+=(y_from * src_stride_bytes);
	for (int i=0; i<h; ++i)
	{
		VALID_SCREEN(Screen.GetPixelBuffer(x,y+i));
		VALID_SCREEN(Screen.GetPixelBuffer(x+w-1, y+i));
		memcpy(Screen.GetPixelBuffer(x,y+i), (unsigned short*)src, w*2);
		src+=src_stride_bytes;
	}
}
#endif



// - MEMBER FUNCTIONS ------------------------------------------
// --------------------------------------------------------------- //
// C'tor
// --------------------------------------------------------------- //
oBitmap::oBitmap()
{
	m_CookieInside=false;
	m_Height=0; m_Width=0; m_Buffer=NULL; m_colorkey = PackColor( RGB(255,0,128) ); m_CookieInLine=NULL;
	#ifdef TINYGL
	m_glTextureId=0;
	#endif
}

// --------------------------------------------------------------- //
// D'tor
// --------------------------------------------------------------- //
oBitmap::~oBitmap()
{
	Delete();
}

// --------------------------------------------------------------- //
// Create XButton image
// --------------------------------------------------------------- //
void oBitmap::XButton()
{
OPENGL_IMAGE img;

	img.resize(14,14);
	long j;
	for (j=0; j<img.m_sizex * img.m_sizey; j++)
	{
		img.m_txdata[j*4  ] = 0;
		img.m_txdata[j*4+1] = 0;
		img.m_txdata[j*4+2] = 255;
		img.m_txdata[j*4+3] = 255;
	}
	long n = img.m_sizey; if (img.m_sizex<n) n=img.m_sizey;
	for (j=4; j<n-4; j++)
	{
		img.m_txdata[(j*n+j)*4  ] = 255;
		img.m_txdata[(j*n+j)*4+1] = 255;
		img.m_txdata[(j*n+j)*4+2] = 255;
		img.m_txdata[(j*n+j)*4+3] = 255;

		if (j<n-1 && j>3)
		{
			img.m_txdata[(j*n+n-j+1)*4  ] = 255;
			img.m_txdata[(j*n+n-j+1)*4+1] = 255;
			img.m_txdata[(j*n+n-j+1)*4+2] = 255;
			img.m_txdata[(j*n+n-j+1)*4+3] = 255;
		}
		img.m_txdata[(j*n+n-j)*4  ] = 255;
		img.m_txdata[(j*n+n-j)*4+1] = 255;
		img.m_txdata[(j*n+n-j)*4+2] = 255;
		img.m_txdata[(j*n+n-j)*4+3] = 255;
 	}
	CreateBy(img, (COLORREF)-1);
}


// --------------------------------------------------------------- //
// Create from file
// --------------------------------------------------------------- //
void oBitmap::Create(const char* filename, COLORREF cookie, RECT *argRectangle)
{
	Delete();
	OPENGL_IMAGE img;
	
	if(img.LoadImageBMP(filename, cookie) || img.LoadImagePNG(filename, cookie))
	{
		CreateBy(img, cookie, argRectangle);
	}
}


// --------------------------------------------------------------- //
// CreateBy
// cookie is the internal stored color for the transparent color.
// Transparency is picked by the aplha channel, though.
// --------------------------------------------------------------- //
bool oBitmap::CreateBy(OPENGL_IMAGE& img, COLORREF cookie, RECT *argRectangle)
{
	m_colorkey=PackColor(cookie);
	Delete();
	DeleteGLTexture();
	unsigned long Offset1;
	unsigned long Offset2;
	register unsigned int x, y;
	unsigned char  r,g,b;

	RECT	tmprect;
	// - CREATE THE BITMAP -

	// If no source rectangle was specified:
	if(!argRectangle)
	{
		argRectangle = &tmprect;
		// Specify the whole bitmap:
		argRectangle->left   = 0;
		argRectangle->top    = 0;
		argRectangle->right  = img.m_sizex-1;
		argRectangle->bottom = img.m_sizey-1;
	}
	else // Clip if bmp is too small
	{
		if (argRectangle->bottom>=img.m_sizey) argRectangle->bottom=img.m_sizey-1;
		if (argRectangle->right >=img.m_sizex) argRectangle->right =img.m_sizex-1;
	}
unsigned int sx, ex, sy, ey;

	m_Width  = (int)(argRectangle->right-argRectangle->left) + 1;
	m_Height = (int)(argRectangle->bottom-argRectangle->top) + 1;

	sx=(unsigned int)argRectangle->left;
	ex=(unsigned int)argRectangle->right;
	sy=(unsigned int)argRectangle->top;
	ey=(unsigned int)argRectangle->bottom;

	m_Buffer = (unsigned short*)gl_malloc(m_Width*m_Height*sizeof(unsigned short));
	m_CookieInLine = (unsigned char*)gl_malloc(m_Height*sizeof(unsigned char));
	if(!m_Buffer || !m_CookieInLine)
	{
		g_debug("Out of memory\n");
		Delete();
		return false;
	}

	m_CookieInside=false;
	// - LOAD THE BITMAP -
	Offset2 = 0;
	unsigned short col;
	for(y = sy; y <= ey; y++)
	{
		unsigned char HasCookie=0;
			Offset1    = (sx + (          y  )*img.m_sizex) *4;
		for(x = sx; x <= ex; x++)
		{
			// Safety check
			if (Offset1>=(unsigned)(img.m_sizex*img.m_sizey*4))
				Offset1=0;
			r = img[ Offset1++ ];
			g = img[ Offset1++ ];
			b = img[ Offset1++ ];
			// check alpha channel
			if (img[Offset1++] < 5)
			{
				m_CookieInside=true;
				HasCookie=1;
				col=m_colorkey;
			}
			else
			{
				col = PackColor(r,g,b);
			}
			m_Buffer[Offset2++] = col;
		}
		m_CookieInLine[y-sy] = HasCookie;
	}
	
	if(!m_CookieInside)
		m_colorkey=(unsigned short)-1;
	
//	if(!Screen.Landscape) Rotate90Degs();
	return true;
}

// --------------------------------------------------------------- //
// Delete the attached glTexture
// --------------------------------------------------------------- //
void oBitmap::DeleteGLTexture()
{
	#ifndef TRENKI3D
	if(m_glTextureId)
		glDeleteTextures(1, &m_glTextureId);
	m_glTextureId = 0;
	#endif
}

#ifdef TINYGL
// --------------------------------------------------------------- //
// Make a glTexture from this bitmap.
// returns texture index.
// stays in TinyGL even when oBitmap gets
// deleted
// --------------------------------------------------------------- //
unsigned int oBitmap::MakeGLTexture()
{
	DeleteGLTexture();
	if(!m_Width || !m_Height) return 0;

	glGenTextures(1, &m_glTextureId);
	glBindTexture(GL_TEXTURE_2D, m_glTextureId);

	// Doing glTexImage2D ourself here - much faster!
	GLImage *im;
	GLContext *c=gl_get_context();
	im=&c->current_texture->images[0];
	im->xsize=256;
	im->ysize=256;
	im->pixmap = gl_zalloc(256*256*sizeof(short));
	*(unsigned short*)im->pixmap = (unsigned short)0xffff;

	// scale the image to a 256x256 texture for tinygl
	short* pDest = (short*)im->pixmap;
	for(int y=0; y<256; ++y)
	{
	int row = (y*m_Height)>>8;
	for(int x=0; x<256; ++x)
		{
			*pDest = m_Buffer[(x*m_Width)/256 + (m_Height-row-1) * m_Width];
			++pDest;
		}
	}
	return m_glTextureId;
}
#endif


// --------------------------------------------------------------- //
// Rotate by 90 degs (Landscape -> Portrait)
// --------------------------------------------------------------- //
void oBitmap::Rotate90Degs()
{
if(!good()) return;

int x,y;
	unsigned short* pNew = new unsigned short[m_Width*m_Height];
	if(!pNew)
	{
		g_debug("Out of memory");
		Delete();
		return;
	}
	memset(pNew, 0, m_Width*m_Height*2);
	for(x=0; x<m_Width; ++x)
	{
		for(y=0; y<m_Height; ++y)
		{
			pNew[y+x*m_Height] = m_Buffer[x+y*m_Width];
		}
	}
	delete m_Buffer;
	m_Buffer=pNew;
	x = m_Width; m_Width=m_Height; m_Height=x;
}


// --------------------------------------------------------------- //
// Delete
// --------------------------------------------------------------- //
void oBitmap::Delete(void)
{
	m_Width=0; m_Height=0;
	if(m_Buffer)
		gl_free(m_Buffer);
	m_Buffer = NULL;

	if(m_CookieInLine)
		gl_free(m_CookieInLine);
	m_CookieInLine=NULL;
	m_CookieInside=false;
//	if(m_glTextureId) glDeleteTextures(1, &m_glTextureId); No glContext at d'tor
	#ifndef TRENKI3D
	m_glTextureId=0;
	#endif
}

// --------------------------------------------------------------- //
// Copy Operator
// --------------------------------------------------------------- //
oBitmap& oBitmap::operator=(const oBitmap& b)
{
	Delete();
	DeleteGLTexture();
	int ncopy =  m_Width*m_Height*sizeof(unsigned short); // number of bytes
	m_Width  = b.m_Width;
	m_Height = b.m_Height;
	m_Buffer = (unsigned short*)gl_malloc(ncopy);
	if(!m_Buffer)
	{
		g_debug("Out of memory\n");
		Delete();
		return *this;
	}

	if((ncopy & 0x00000003) == 0) // bits 1 and 2 not set - 4 byte aligned
	{
		ncopy /= 4;
		QCopy4(m_Buffer, b.m_Buffer, ncopy);
	}
	else
	{
		ncopy /= 2;
		QCopy2(m_Buffer, b.m_Buffer, ncopy);
	}
	if (b.m_CookieInLine)
	{
		m_CookieInLine = (unsigned char*)gl_malloc(m_Height);
		memcpy(m_CookieInLine, b.m_CookieInLine, m_Height);
	}
	m_CookieInside = b.m_CookieInside;	
return *this;
}

// --------------------------------------------------------------- //
// Copy to Screen (backgrnd bitmap)
// --------------------------------------------------------------- //
void oBitmap::MemCpyToScreen()
{
	if (!m_Buffer) return;
#ifdef GP2X
	gp2x_blit(GetUpperRealAddress(m_Buffer), 0,0, m_Width, 0,0, min(320, m_Width), min(240, m_Height));
	return;
#endif

	if(	m_Width==Screen.getWidth() && m_Height==Screen.getHeight())
	{
		QCopy4((unsigned short*)Screen.getColorBuffer(), m_Buffer, Screen.ScreenRange/2);
	}
	else
	{
		Draw(0,0);
	}
}
// --------------------------------------------------------------- //
// Copy from Screen (Use as bitmap)
// --------------------------------------------------------------- //
void oBitmap::MemCpyFromScreen()
{
	Delete();
	m_Width=Screen.getWidth();
	m_Height=Screen.getHeight();
	m_Buffer = (unsigned short*)gl_malloc(m_Width*m_Height*sizeof(unsigned short));
	m_CookieInLine = (unsigned char*)gl_malloc(m_Height);
	if(!m_Buffer || !m_CookieInLine)
	{
		g_debug("Out of memory\n");
		Delete();
		return;
	}

	memset(m_CookieInLine, 0, m_Height);
	QCopy4(m_Buffer, (unsigned short*)Screen.getColorBuffer(), Screen.ScreenRange/2);
}


static float gf_qsine(float x)
{
//return sin(x/57.296);
	const float pi = 3.14158f;
    const float B = 4/pi;
    const float C = -4/(pi*pi);
    
    while(x>360.0f)  x-=360.0f;
    while(x<0.0f)    x+=360.0f;
    if (x>180.0f){x = 180.0f-x;}
    
    x/=57.296f;
	float abs_x = x>0 ? x : -x;
    float y = B * x + C * x * abs_x;

	float abs_y = y>0 ? y : -y;
	const float P = 0.225f; y = P*(y*abs_y-y)+y;
	// extra2 (even more accurate)
	// const float Q = 0.775f; y = Q * y + P * y * abs_y;
return y;
}


// --------------------------------------------------------------- //
// Create a rotated image
// --------------------------------------------------------------- //
void oBitmap::CreateRotatation(float degrees, oBitmap& out)
{
	// 1.42* max dimesion - enough space for 45
int newsize = (142*max(m_Width, m_Height))/100;

	if (newsize > out.m_Width || newsize > out.m_Height || newsize==0)
	{
		out.Delete();
		if(!newsize) return;
		out.m_Buffer = (unsigned short*)gl_malloc(newsize*newsize*sizeof(unsigned short));
		out.m_CookieInLine = new unsigned char[newsize];
		out.m_CookieInLine = (unsigned char*)gl_malloc(newsize);
	}
	
	if (!out.m_Buffer || !out.m_CookieInLine){return;} // uh-oh! Out of memory
	out.m_Width=newsize;
	out.m_Height=newsize;
	out.m_colorkey = m_colorkey;

#define BITS 16
	int sinq = (int)(gf_qsine(degrees)   *(float)(1<<BITS));
	int cosq = (int)(gf_qsine(degrees+90)*(float)(1<<BITS));
	#define qsin(a) ((sinq*(a))>>BITS) // sin(-degrees)*a
	#define qcos(a) ((cosq*(a))>>BITS) // cos(-degrees)*a

	register int x2, y2;
	register unsigned short col;
	for(register int x=0; x<newsize; ++x)
	{
		for(register int y=0; y<newsize; ++y)
		{
			x2 = qcos(x-newsize/2)-qsin(y-newsize/2) + m_Width/2;
			y2 = qsin(x-newsize/2)+qcos(y-newsize/2) + m_Height/2;
			if (x2>-1 && y2>-1 && x2<m_Width && y2<m_Height)
			{
				col = m_Buffer[m_Width*y2+x2]; // GetPixel(x2,y2)
			}
			else
			{
				col = m_colorkey;
			}
			out.m_Buffer[out.m_Width*y+x] = col;
			if(col == m_colorkey)
			{
				out.m_CookieInLine[y] = 1;
				out.m_CookieInside=true;
			}
		}
	}
}


// --------------------------------------------------------------- //
// DrawClipped
// ClipX/CLipY = upper left pixel at position argX/argY on screen
// ClipWidth = lower right pixel (not "width") of rect to draw
// --------------------------------------------------------------- //
void oBitmap::DrawClip(int argX, int argY, int ClipX, int ClipY, int ClipWidth, int ClipHeight)
{
	if(!m_Buffer)	return;

	argX-=ClipX;
	argY-=ClipY;

	if (ClipHeight>m_Height) ClipHeight=m_Height;
	if (ClipWidth>m_Width)   ClipWidth=m_Width;


	// If the bitmap overlaps the left edge:
	if(argX+ClipX < 0)
		ClipX = -argX;
	// If the bitmap overlaps the rigth edge:
	if(argX + ClipWidth > Screen.getViewportWidth())
		ClipWidth = Screen.getViewportWidth() - argX;

	// If the bitmap overlaps the top edge:
	if(argY+ClipY < 0)
		ClipY = -argY;
	// If the bitmap overlaps the bottom edge:
	if(argY + ClipHeight > Screen.getViewportHeight())
		ClipHeight = Screen.getViewportHeight() - argY;


	// offset the viewport
	argX+=Screen.getViewportX();
	argY+=Screen.getViewportY();
	// Real drawing position after clipping
	argX+=ClipX;
	argY+=ClipY;

	// - DRAW -
	int linelen = ClipWidth-ClipX; // num shorts to draw/copy per line

	// If the bitmap is not on screen:
	if(linelen<=0 || (ClipHeight-ClipY)<=0)
		return;

	int scanline = Screen.getWidth();

	#ifndef GP2X
		register int x, y=ClipY;
		register unsigned short* pScreen = Screen.getColorBuffer() + argX + argY*scanline;
		register const unsigned short* pImage  = m_Buffer + ClipX + ClipY*m_Width;
	#endif
	if(!Screen.alpha_blending)
	{
		if(m_CookieInside && Screen.alpha_testing) // Cookie Cut in block
		{
			bool linelen_is_4_bytes = (linelen & 0x00000001) ? false:true; // can copy 4 bytes at once?
			int linelen_half = linelen/2;
			#ifdef GP2X
				gp2x_blit_cookie(GetUpperRealAddress(m_Buffer), ClipX, ClipY, m_Width<<1, argX, argY, linelen, ClipHeight-ClipY, m_colorkey);
				return;
			#else
				do
				{
					if(m_CookieInLine[y])
					{
						x=0;
						do
						{
							if(*pImage == m_colorkey) continue;
							VALID_SCREEN(pScreen);
							*(pScreen) = *pImage;
						}while(++pScreen, ++pImage, ++x<linelen);
						pScreen-=linelen;
						pImage -=linelen;
					}
					else
					{
						if(linelen_is_4_bytes)
							QCopy4(pScreen, pImage, linelen_half);
						else
							QCopy2(pScreen, pImage, linelen);
					}
					// one line down
					pScreen+=scanline;
					pImage+=m_Width;
				}while(++y<ClipHeight);
			#endif
		}
		else // Completely solid block - muhahaha
		{
			#ifdef _DEBUG
				dbg_gp2x_blit((unsigned char*)m_Buffer,  ClipX, ClipY, m_Width<<1, argX, argY, linelen, ClipHeight-ClipY);
				return;
			#else
				#ifdef GP2X
					gp2x_blit(GetUpperRealAddress(m_Buffer), ClipX, ClipY, m_Width<<1, argX, argY, linelen, ClipHeight-ClipY);
					return;
				#else
					if((linelen & 0x00000001) == 0)
					{
						int linelen_half = linelen/2;
						do
						{
							VALID_SCREEN(pScreen);
							VALID_SCREEN(pScreen+linelen_half*2-1);
							VALID_IMG(pImage);
							VALID_IMG(pImage + linelen_half*2-1);
							QCopy4(pScreen, pImage, linelen_half); // 4 bytes at once
							// one line down
							pScreen+=scanline;
							pImage+=m_Width;

							argY+=scanline;
						}while(++y<ClipHeight);
					}
					else
					{
						do
						{
							VALID_SCREEN(pScreen);
							VALID_SCREEN(pScreen+linelen-1);
							VALID_IMG(pImage);
							VALID_IMG(pImage + linelen-1);
							QCopy2(pScreen, pImage, linelen); // 2 bytes at once
							// one line down
							pScreen+=scanline;
							pImage+=m_Width;

							argY+=scanline;
						}while(++y<ClipHeight);
					}

				#endif
			#endif
		}
	}
	else // Must have alpha blending
	{
		#ifdef GP2X
		register unsigned short* pScreen = Screen.getColorBuffer() + argX + argY*scanline;
		register unsigned short* pImage  = m_Buffer + ClipX + ClipY*m_Width;
		register unsigned int x;
		register int y=ClipY;
		#endif
		do
		{
			x=0;
			do
			{
				if(m_CookieInLine[y] && m_CookieInside && *pImage == m_colorkey && Screen.alpha_testing) continue;
				VALID_SCREEN(pScreen);
				VALID_IMG(pImage);
				Screen.m_BlendFunc(pScreen, *pImage);
			}while(++pScreen, ++pImage, ++x<linelen);
			pScreen-=linelen;
			pImage -=linelen;

			// one line down
			pScreen+=scanline;
			pImage+=m_Width;
		}while(++y<ClipHeight);
	}
}

// --------------------------------------------------------------- //
// ZOOM
// --------------------------------------------------------------- //
void oBitmap::Zoom(int argX, int argY, int endX, int endY)
{
int ClipX, ClipWidth, ClipY, ClipHeight;
	if(!m_Buffer) return;

	if (argX>endX) SwapMe(argX, endX);
	if (argY>endY) SwapMe(argY, endY);

	// - CLIP -
	ClipX = 0, ClipWidth  = (endX-argX);
	ClipY = 0, ClipHeight = (endY-argY);

	if(!ClipHeight || !ClipHeight) return;

	// for scale factor - save before clipping
	register int zHeight= ClipHeight;
	register int zWidth = ClipWidth;

	// If the bitmap overlaps the left edge:
	if(argX < 0)
		ClipX = -argX;
	// If the bitmap overlaps the rigth edge:
	if(endX > Screen.getViewportWidth())
		ClipWidth = Screen.getViewportWidth() - argX;

	// If the bitmap overlaps the top edge:
	if(argY < 0)
		ClipY = -argY;
	// If the bitmap overlaps the bottom edge:
	if(endY > Screen.getViewportHeight())
		ClipHeight = Screen.getViewportHeight() - argY;

	// offset viewport
	argX+=Screen.getViewportX();
	argY+=Screen.getViewportY();
	endX+=Screen.getViewportX();
	endY+=Screen.getViewportY();
	
	// offset clip rect
	argX+=ClipX;
	argY+=ClipY;

	// - DRAW -
	int linelen  = ClipWidth-ClipX;
	if(linelen<=0 || (ClipHeight-ClipY)<=0)
		return;

	int scanline = Screen.getWidth();
	register unsigned long Offset2;
	register int x, y, YHeight, XWidth;
	// YHeight = Y*Height, XWidth=X*Width -> zWidth * Position in src image

	register unsigned short* pScreen=Screen.getColorBuffer()+argX + argY*scanline;

	register int y_val, XWidth_reset = ClipX*m_Width;
	int fracx = (1<<16)/zWidth;
	int fracy = (1<<16)/zHeight;
	YHeight = ClipY*m_Height;
	y=ClipY;
	scanline-=linelen; // number of bytes to move from end of line to next line
	
	if(Screen.alpha_blending)
	{
		do
		{
			y_val = m_Width* ((YHeight*fracy) >> 16);
			XWidth = XWidth_reset; x=0;
			do
			{
				Offset2 = y_val + ( (XWidth * fracx) >>16);
				// cookie
				if(m_Buffer[Offset2] == m_colorkey) continue;
				// blendfunc
				VALID_SCREEN(pScreen);
				Screen.m_BlendFunc(pScreen, m_Buffer[Offset2]);
			}while(++pScreen, XWidth+=m_Width, ++x<linelen);
			pScreen+=scanline;
		}while(++y, YHeight+=m_Height, y<ClipHeight);
	}
	else // Solid block, pixel by pixel
	{
		if(m_CookieInside && Screen.alpha_testing)
		{
			do
			{
				y_val = m_Width* ((YHeight*fracy) >> 16);
				XWidth = XWidth_reset; x=0;
				do
				{
					Offset2 = y_val + ( (XWidth * fracx) >>16);
					// cookie
					if(m_Buffer[Offset2] == m_colorkey) continue;
					VALID_SCREEN(pScreen);
					*pScreen = m_Buffer[Offset2];
				}while(++pScreen, XWidth+=m_Width, ++x<linelen);
				pScreen+=scanline;
			}while(++y, YHeight+=m_Height, y<ClipHeight);
		}
		else // fully solid
		{
			do
			{
				y_val = m_Width* ((YHeight*fracy) >> 16);
				XWidth = XWidth_reset; x=0;
				do
				{
					Offset2 = y_val + ( (XWidth * fracx) >>16);
					VALID_SCREEN(pScreen);
					*pScreen = m_Buffer[Offset2];
				}while(++pScreen, XWidth+=m_Width, ++x<linelen);
				pScreen+=scanline;
			}while(++y, YHeight+=m_Height, y<ClipHeight);
		}
	}
}

// --------------------------------------------------------------- //
// Grab From Screen Buffer - Enabled cookie cut here!
// --------------------------------------------------------------- //
void oBitmap::GrabFrom(int argX, int argY, int endX, int endY, COLORREF cookie)
{
bool didclip=false;
int ClipX, ClipWidth, ClipY, ClipHeight;
	Delete();
	DeleteGLTexture();
	#ifndef TRENKI3D
	glFlush();
	#endif
	m_Height = abs(argY - endY);
	m_Width  = abs(argX - endX);

	// - CLIP -
	ClipX = 0, ClipWidth  = m_Width;
	ClipY = 0, ClipHeight = m_Height;

	// If the bitmap is not on screen:
	if((argX + m_Width) < 0 || argX >= Screen.getViewportWidth()
	|| (argY + m_Height)< 0 || argY >= Screen.getViewportHeight()) return;

	// If the bitmap overlaps the left edge:
	if(argX < 0)
	{
		ClipX = -argX; didclip=true;
	}
	// If the bitmap overlaps the rigth edge:
	else
	if(argX + m_Width > Screen.getViewportWidth())
	{
		ClipWidth = Screen.getViewportWidth() - argX; didclip=true;
	}

	// If the bitmap overlaps the top edge:
	if(argY < 0)
	{
		ClipY = -argY; didclip=true;
	}
	// If the bitmap overlaps the bottom edge:
	else 
	if(argY + m_Height > Screen.getViewportHeight())
	{
		ClipHeight = Screen.getViewportHeight() - argY; didclip=true;
	}

	// offset viewport rect
	argX+=Screen.getViewportX();
	argY+=Screen.getViewportY();

//	unsigned long linelen = ClipWidth-ClipX; // num shorts
	unsigned long scanline = Screen.getWidth();
	m_Buffer = (unsigned short*)gl_malloc(m_Width*m_Height*sizeof(unsigned short)); //memset(Buffer, 0, Width*Height*2);
	m_colorkey = PackColor(cookie);
	if(didclip)
	{
		int all = m_Width*m_Height; unsigned short* pC = m_Buffer;
		for(int i=0; i<all; ++i) *pC++ = m_colorkey;
	}
	m_CookieInLine = (unsigned char*)gl_malloc(m_Height*sizeof(unsigned char));
	m_CookieInside=didclip;
	memset(m_CookieInLine, didclip ? 1:0, m_Height);
	if(!m_Buffer || !m_CookieInLine)
	{
		g_debug("Out of memory\n");
		Delete();
		return;
	}
	register unsigned short* pScreen = (unsigned short*)Screen.getColorBuffer() + (argY+ClipY)*scanline + ClipX+argX;
	register unsigned short* pBuf = m_Buffer+ClipX+ClipY*m_Width;
	ClipHeight-=ClipY;
	ClipWidth -=ClipX;
	register int x, y;
	for(y = 0; y < ClipHeight; y++)
	{
		for(x = 0; x < ClipWidth; x++)
		{
			VALID_SCREEN(pScreen);
			if(*pScreen == m_colorkey)
			{
				m_CookieInLine[y]=1;
				m_CookieInside=true;
			}
			*pBuf++ = *pScreen++;
		}
		pBuf-=ClipWidth;
		pBuf+=m_Width;
		pScreen-=ClipWidth;
		pScreen+=scanline;
	}

	if(!m_CookieInside)
		m_colorkey=(unsigned short)-1;
}


// ------------------------------------------------------------- //
// SAVE BMP FILE
// ------------------------------------------------------------- //
bool oBitmap::SaveBMPFile(const char* szFilename)
{
BITMAPFILEHEADER fh;
BITMAPINFOHEADER bi;
BYTE *txdata; 
int bWidth;

	//Make sure its not a NULL filenam
	if(szFilename == NULL)
		return false;

	//Attempt to create the output
	GFFileSystem file_out;
	if (!file_out.OpenOut(szFilename))
	{
		return false;
	}
	// Setup output buffer stuff, since Windows has paging and we're in flat mode, I just made
	// it as big as the bitmap 


	// Get 4 byte aligned width
	#define BITS_PER_PIXEL 24
	#define DWORD_24BIT_WIDTHBYTES_ALIGN( width )  ((( (width)* (BITS_PER_PIXEL/8)) + 3) & ~3 )

	bWidth=DWORD_24BIT_WIDTHBYTES_ALIGN(m_Width);

	txdata = new BYTE [bWidth* m_Height * BITS_PER_PIXEL/8];  //width*height*24-bit

	//Write the file header
	((char *)&(fh . bfType))[0] = 'B';
	((char *)&(fh . bfType))[1] = 'M';
	fh . bfSize = 9270; // (long)(sizeof (BITMAPINFOHEADER)+sizeof (BITMAPFILEHEADER)+bWidth*Height*BITS_PER_PIXEL/8); //Size in BYTES
	fh . bfReserved1 = 0;
	fh . bfReserved2 = 0;
	fh . bfOffBits = 54; // sizeof (BITMAPINFOHEADER)+sizeof (BITMAPFILEHEADER);
	bi . biSize = 40; // sizeof (BITMAPINFOHEADER);
	bi . biWidth = m_Width;
	bi . biHeight =m_Height;
	bi . biPlanes = 1;
	bi . biBitCount = BITS_PER_PIXEL;
	bi . biCompression = BI_RGB;
	bi . biSizeImage = 0;
	bi . biXPelsPerMeter = 10000;
	bi . biYPelsPerMeter = 10000;
	bi . biClrUsed = 0;
	bi . biClrImportant = 0;

	// fh.bfType has 4 bytes on #*~$% implementation. pack(1) not accepted!?
	file_out.Write((void*) &fh.bfType,  2);
	file_out.Write((void*) &fh.bfSize,  12);
	file_out.Write((void*) &bi,         40);

	long postx;
	register short pix;
	for (long x=0; x<m_Width; x++)
	for (long y=0; y<m_Height; y++)
	{
		postx = y*bWidth + x*(BITS_PER_PIXEL/8);
		pix = m_Buffer[(m_Height-y-1)*m_Width + x];
		txdata[postx  ] = (BYTE)GETB(pix);
		txdata[postx+1] = (BYTE)GETG(pix);
		txdata[postx+2] = (BYTE)GETR(pix);
	}
	file_out.Write(txdata, m_Width*m_Height*3);
	file_out.Close();
	if (txdata) delete [] txdata;
return true;
}


// --------------------------------------------------------------- //
// Colliion of 2 oBitmaps
// --------------------------------------------------------------- //
// Rect Collision?
#define qmax_(a, b) ((a)>(b) ? (a) : (b))
#define qmin_(a, b) ((a)<(b) ? (a) : (b))
#define __RectRectIntersection(x1, y1, xe1, ye1, x2, y2, xe2, ye2) \
							(qmax_(qmin_(x1, xe1), qmin_(x2, xe2)) <= qmin_(qmax_(x1, xe1), qmax_(x2, xe2))) \
						 && (qmax_(qmin_(y1, ye1), qmin_(y2, ye2)) <= qmin_(qmax_(y1, ye1), qmax_(y2, ye2)))
// For already normalized rects
#define __NormRectRectIntersection(x1, y1, xe1, ye1, x2, y2, xe2, ye2) \
							   ((qmax_(x1, x2) <= qmin_(xe1, xe2)) \
							&& (qmax_(y1, y2) <= qmin_(ye1, ye2)))

// --------------------------------------------------------------- //
bool oBitmap::Collision(oBitmap& b1, int x1, int y1, oBitmap& b2, int x2, int y2)
{
	if (!__NormRectRectIntersection(x1, y1, x1+b1.m_Width, y1+b1.m_Height, x2, y2, x2+b2.m_Width, y2+b2.m_Height)) return false;
	
int xs, ys, xe, ye;
	xs=qmax_(x1, x2);						ys=qmax_(y1, y2);
	xe=qmin_(x1+b1.m_Width, x2+b2.m_Width);	ye=qmin_(y1+b1.m_Height, y2+b2.m_Height);

	return _PixelCollision(b1,x1,y1, b2,x2,y2, xs,xe,ys,ye);
}

// --------------------------------------------------------------- //
bool oBitmap::CollisionClip(	oBitmap& b1, int x1, int y1, int cx1, int cy1, int cw1, int ch1,
								oBitmap& b2, int x2, int y2, int cx2, int cy2, int cw2, int ch2)
{
	if (!__NormRectRectIntersection(x1, y1, x1+cw1, y1+ch1,
									x2, y2, x2+cw2, y2+ch2)) return false;

int xs,ys,xe,ye;
	xs=qmax_(x1,     x2);		ys=qmax_(y1,     y2);
	xe=qmin_(x1+cw1, x2+cw2);	ye=qmin_(y1+ch1, y2+ch2);
	return _PixelCollision(b1,x1-cx1,y1-cy1, b2,x2-cx2,y2-cy2, xs,xe,ys,ye);
}


// --------------------------------------------------------------- //
bool oBitmap::_PixelCollision(	oBitmap& b1, int x1, int y1, oBitmap& b2, int x2, int y2,
								int xs, int xe, int ys, int ye)
{
	// xs/xe and ys/ye are real pixel coords on the "screen"
	// Find Overlapping Rect
	register int x, y;

	if(b1.m_CookieInside)
	{
		// even X
			// even Y
			for (x=xs; x<xe; x+=2)
			for (y=ys; y<ye; y+=2)
			{
				// pBits[x][y] = pBits[x + y*w]
				if (b1.m_Buffer[(x-x1)+(y-y1)*b1.m_Width]==b1.m_colorkey)
					continue; // No collision possible with this bit
				if (b2.m_Buffer[(x-x2)+(y-y2)*b2.m_Width]!=b2.m_colorkey) // Got Collision
					return true;
			}
			// odd Y
			for (x=xs; x<xe; x+=2)
			for (y=ys+1; y<ye; y+=2)
			{
				// pBits[x][y] = pBits[x + y*w]
				if (b1.m_Buffer[(x-x1)+(y-y1)*b1.m_Width]==b1.m_colorkey)
					continue; // No collision possible with this bit
				if (b2.m_Buffer[(x-x2)+(y-y2)*b2.m_Width]!=b2.m_colorkey) // Got Collision
					return true;
			}

		// odd X
			// even Y
			for (x=xs+1; x<xe; x+=2)
			for (y=ys; y<ye; y+=2)
			{
				// pBits[x][y] = pBits[x + y*w]
				if (b1.m_Buffer[(x-x1)+(y-y1)*b1.m_Width]==b1.m_colorkey)
					continue; // No collision possible with this bit
				if (b2.m_Buffer[(x-x2)+(y-y2)*b2.m_Width]!=b2.m_colorkey) // Got Collision
					return true;
			}
			// odd Y
			for (x=xs+1; x<xe; x+=2)
			for (y=ys+1; y<ye; y+=2)
			{
				// pBits[x][y] = pBits[x + y*w]
				if (b1.m_Buffer[(x-x1)+(y-y1)*b1.m_Width]==b1.m_colorkey)
					continue; // No collision possible with this bit
				if (b2.m_Buffer[(x-x2)+(y-y2)*b2.m_Width]!=b2.m_colorkey) // Got Collision
					return true;
			}
	}
	else
	{
		if(!b2.m_CookieInside) return true; // rect-rect suffices

		// even X
			// even Y
			for (x=xs; x<xe; x+=2)
			for (y=ys; y<ye; y+=2)
			{
				// pBits[x][y] = pBits[x + y*w]
				if (b2.m_Buffer[(x-x2)+(y-y2)*b2.m_Width]!=b2.m_colorkey) // Got Collision
					return true;
			}
			// odd Y
			for (x=xs; x<xe; x+=2)
			for (y=ys+1; y<ye; y+=2)
			{
				// pBits[x][y] = pBits[x + y*w]
				if (b2.m_Buffer[(x-x2)+(y-y2)*b2.m_Width]!=b2.m_colorkey) // Got Collision
					return true;
			}
		// odd X
			// even Y
			for (x=xs+1; x<xe; x+=2)
			for (y=ys; y<ye; y+=2)
			{
				// pBits[x][y] = pBits[x + y*w]
				if (b2.m_Buffer[(x-x2)+(y-y2)*b2.m_Width]!=b2.m_colorkey) // Got Collision
					return true;
			}
			// odd Y
			for (x=xs+1; x<xe; x+=2)
			for (y=ys+1; y<ye; y+=2)
			{
				// pBits[x][y] = pBits[x + y*w]
				if (b2.m_Buffer[(x-x2)+(y-y2)*b2.m_Width]!=b2.m_colorkey) // Got Collision
					return true;
			}
	}
return false;
}


#endif // #if defined(TINYGL) || defined(TRENKI3D)













