// - HEADERS ---------------------------------------------------
#if defined(TINYGL) || defined(TRENKI3D)
#include "stdafx.h"
#include "oscreen.h"


#ifdef GP2X
	#include <sys/mman.h>
	#include <fcntl.h>
	#include "mmsp2_regs.h"
#endif

#if defined(GPHWIZ)
	#include <sys/mman.h>
	#include <fcntl.h>
	#include <sys/ioctl.h>
	int g_dev_fb0 = 0;
	int g_dev_mem = 0;
	volatile unsigned short* g_memregs16 = NULL;
	volatile unsigned long * g_memregs32 = NULL;
#endif


#ifdef _WIN32_WCE
	#include "ceFramework.h" // GXBeginDraw and stuff
#endif

#ifdef TINYGL
	#include "gfglinit.h"
	#include <GL/zgl.h>
	#include <GL/gl.h>
#endif


#ifdef THREADED_UPDATE
DWORD WINAPI FlipThread(LPVOID pParam);
#endif

#define _abs(a) ((a)<0 ? -(a) : a)



// - GLOBALS ---------------------------------------------------
oScreen Screen;

// -------------------------------------------------------- //
// BlendPixel - Set
// -------------------------------------------------------- //
static void BlendPixelSet(unsigned short* pDest, unsigned short colsrc)
{*pDest = colsrc;}



#define RED		((0x00ff >> 3)<<SHIFTRED)
#define GREEN	((0x00ff >> LEFTGREEN)<<5)
#define BLUE	((0x00ff >> 3))

// -------------------------------------------------------- //
// BlendPixel - Add
// -------------------------------------------------------- //
inline unsigned int _blend_mix(unsigned int a, unsigned int b, unsigned int lim)
	{a+=b; return a>lim?lim:a&lim;}

static void BlendPixelAdd(unsigned short* pDest, unsigned short colsrc)
{
// alphablending = 1 .. 255
register unsigned int alpha = (unsigned int)Screen.alpha_blending;
#define   ONE(a) ((unsigned int)(a))
#define  AONE(a) (((unsigned int)(a)*alpha) >>8)
	*pDest = (unsigned short)
			(
			(_blend_mix(ONE(*pDest & RED  ), AONE(colsrc & RED  ), RED   ))
			|(_blend_mix(ONE(*pDest & GREEN), AONE(colsrc & GREEN), GREEN ))
			|(_blend_mix(ONE(*pDest & BLUE ), AONE(colsrc & BLUE ), BLUE  ))
			);
#undef ONE
#undef AONE
}

// -------------------------------------------------------- //
// BlendPixel - Sub
// -------------------------------------------------------- //
static void BlendPixelSub(unsigned short* pDest, unsigned short colsrc)
{
// alphablending = -1 .. -255
register unsigned int one_minus_alpha = 256+Screen.alpha_blending,
					alpha = -Screen.alpha_blending;
#define   ONE(a) ( ((a)*alpha) >>8)
#define  AONE(a) ( ((a)*one_minus_alpha) >>8)

	*pDest =(unsigned short) (
			((AONE(*pDest & RED  ) + ONE(colsrc & RED  ) ) & RED   )
			|((AONE(*pDest & GREEN) + ONE(colsrc & GREEN) ) & GREEN )
			|((AONE(*pDest & BLUE ) + ONE(colsrc & BLUE ) ) & BLUE  )  );

#undef ONE
#undef AONE
}



#ifdef GPHWIZ
int gWizTearing = 1;
extern "C" void WizEnableTearing(int bEnable)
{
	// Wiz device (open in 240x320 mode to avoid tearing)
	#define FBIO_MAGIC                      'D'
	#define FBIO_LCD_CHANGE_CONTROL         _IOW(FBIO_MAGIC, 90, unsigned int[2])
	#define LCD_DIRECTION_ON_CMD            5       /* 320x240 */
	#define LCD_DIRECTION_OFF_CMD           6       /* 240x320 */
	unsigned int send[2];
	if(bEnable)
	{
		gWizTearing=1;
		send[0] = LCD_DIRECTION_ON_CMD;
	}
	else
	{
		gWizTearing=0;
		send[0] = LCD_DIRECTION_OFF_CMD;
	}
	ioctl(g_dev_fb0, FBIO_LCD_CHANGE_CONTROL, &send);
}
#endif




// -------------------------------------------------------- //
// Rotate a 320x240 WORD buffer by 90 degrees - windows only
// -------------------------------------------------------- //
void oScreen::Rotate90Degs(unsigned short* Buffer)
{
register int x,y, x319240, y320;
	// ignore for non VGA screens
	if(Screen.ScreenRange != 320*240)
	{
		unsigned short* rotB = new unsigned short[Screen.ScreenRange];
		x=0; x319240 = (Screen.cBufferWidth-1)*Screen.cBufferHeight;
		do
		{
			y=x319240; y320=0;
			do
			{
				// rotBuffer[(319-x)*240+y] = Buffer[y*320 + x];
				rotB[y] = Buffer[y320 + x]; ++y; y320+=Screen.cBufferWidth;
				rotB[y] = Buffer[y320 + x]; ++y; y320+=Screen.cBufferWidth;
				rotB[y] = Buffer[y320 + x]; ++y; y320+=Screen.cBufferWidth;
				rotB[y] = Buffer[y320 + x]; ++y; y320+=Screen.cBufferWidth;
			} while(y<Screen.cBufferHeight+x319240);
		} while(x319240-=Screen.cBufferHeight, ++x<Screen.cBufferWidth);
		QCopy4(Buffer, rotB, Screen.ScreenRange*sizeof(unsigned short)/4);
		delete []rotB;
		return;
	}
	
	// Rotate by 90 degs:
static unsigned short rotBuffer[320*240];
	// x319240 = (319-x)*240
	x=0; x319240 = 319*240;
	do
	{
		y=x319240; y320=0;
		do
		{
			// rotBuffer[(319-x)*240+y] = Buffer[y*320 + x];
			rotBuffer[y] = Buffer[y320 + x]; ++y; y320+=320;
			rotBuffer[y] = Buffer[y320 + x]; ++y; y320+=320;
			rotBuffer[y] = Buffer[y320 + x]; ++y; y320+=320;
			rotBuffer[y] = Buffer[y320 + x]; ++y; y320+=320;
		} while(y<240+x319240);
	} while(x319240-=240, ++x<320);
	QCopy4(Buffer, rotBuffer, 320*240*sizeof(unsigned short)/4);
}


// -------------------------------------------------------- //
// If you have to use this, you've got a problem
// Smartphone (176x220)
// VGA        (640x480)
// -------------------------------------------------------- //

/*
Screen W,H,BPP = 480, 640, 16
Pitch X,Y = 2, 960
*/
inline void UpdateDisplayNon320x240(unsigned short* Buffer, bool Landscape)
{
#ifdef _WIN32_WCE
unsigned short* dst;
int movex, movey;
	if(Screen.pRawFrameBuffer)
		dst = Screen.pRawFrameBuffer;
	else
		dst = (unsigned short*)GXBeginDraw();

	movex = Screen.xpitch >> 1;
	movey = Screen.ypitch >> 1;
	unsigned short* pDstLine = dst, *pDstPix, *pSrcPix, *pSrcLine=Buffer;
	int inc_src_x, inc_src_y;
	if(Landscape)
	{
		inc_src_x = Screen.getWidth();
		inc_src_y = -1;
		pSrcLine += (Screen.getWidth()-1);
		for (register unsigned int y=0; y<Screen.getWidth(); y++)
		{
			pDstPix=pDstLine;
			pSrcPix=pSrcLine;
			for (register unsigned int x=0; x<Screen.getHeight(); x++)
			{
				// *pDstPix = Buffer[x*Screen.width + (Screen.width - y - 1)];
				*pDstPix = *pSrcPix;
				pDstPix += movex;
				pSrcPix += inc_src_x;
			}
			pDstLine   += movey;
			pSrcLine--; //   += inc_src_y;
		}
	}
	else
	{
		inc_src_x = 1;
		inc_src_y = Screen.getWidth();
		for (register unsigned int y=0; y<Screen.getHeight(); y++)
		{
			pDstPix=pDstLine;
			pSrcPix=pSrcLine;
			for (register unsigned int x=0; x<Screen.getWidth(); x++)
			{
				// *pDstPix = Buffer[x + y*Screen.width]; //Screen.GetpDstPix(x,y);
				*pDstPix = *pSrcPix;
				pDstPix += movex;
				pSrcPix ++; // += inc_src_x;
			}
			pDstLine    += movey;
			pSrcLine    += inc_src_y;
		}
	}
	if(!Screen.pRawFrameBuffer)
		GXEndDraw();
#endif
}


// POCKETPC or WIN32, native
#if defined(WIN32) && !defined(WANT_SDL)
	// -------------------------------------------------------- //
	// POCKETPC - FlipScreen
	// -------------------------------------------------------- //
	inline void UpdateDisplay(unsigned short* Buffer, bool Landscape)
	{
	#define REDMASK (31<<11)
	#define GREENMASK (63<<5)
	#define BLUEMASK (31)

	#ifdef INTERLACED
	static long	LastInterlaced=0;
	#endif
	
	#if defined(WIN32) && !defined(WANT_SDL)
		if(!Screen.pRawFrameBuffer)
		{
			if(Landscape) oScreen::Rotate90Degs(Buffer);
			HDC hdc = GetDC(Screen.Window);
			HDC Context; Context = CreateCompatibleDC(hdc);
			SelectObject(Context, Screen.Buffer2);
			BitBlt(hdc, 0, 0, ::GetSystemMetrics(SM_CXSCREEN), ::GetSystemMetrics(SM_CYSCREEN), Context, 0, 0, SRCCOPY);
			DeleteDC(Context);
			ReleaseDC(Screen.Window, hdc);
			return;
		}
	#endif


	#ifdef _WIN32_WCE
		// is this a Smartphone or something?
		if(Screen.getWidth()+Screen.getHeight() != 320+240)
		{
			UpdateDisplayNon320x240(Buffer, Landscape);
			return;
		}

		if(Landscape)
		{
			if (Screen.BPP == 16)
			{
			unsigned short* src = Buffer;
			unsigned short* dst;
			if(Screen.pRawFrameBuffer)
				dst = Screen.pRawFrameBuffer;
			else
				dst = (unsigned short*)GXBeginDraw();

				// Let's do this really clean (and slow:)
				register unsigned short x, y, *ldst;
				
				// 2, 480 - ACER n35			
				short xptch = (Screen.xpitch >> 1), yptch = (Screen.ypitch >> 1);
				register unsigned short s1, s2, s3, s4, s5, s6, s7, s8;
				ldst=dst + 319 * yptch;
				y=0;
				do // for (y = 0; y < 240; ++y)
				{
					// ldst=dst + 319 * yptch + y*xptch;
					x=0;
					do //for (x = 0; x < 320/8; x++)
					{
						// *(ldst) = *src++; // one pixel right on src
						// ldst -= yptch;    // add a pixel to the right on dest
						s1 = *src++;
						s2 = *src++;
						s3 = *src++;
						s4 = *src++;
						s5 = *src++;
						s6 = *src++;
						s7 = *src++;
						s8 = *src++;
						*ldst=s1; ldst-=yptch;
						*ldst=s2; ldst-=yptch;
						*ldst=s3; ldst-=yptch;
						*ldst=s4; ldst-=yptch;
						*ldst=s5; ldst-=yptch;
						*ldst=s6; ldst-=yptch;
						*ldst=s7; ldst-=yptch;
						*ldst=s8; ldst-=yptch;
					} while(++x<40);
					ldst += xptch;
					ldst += (yptch*320); // undo the ldst-=yptch
				} while(++y<240);
			}
			if(!Screen.pRawFrameBuffer)
				GXEndDraw();
		}
		else
		{
			#ifdef INTERLACED
			LastInterlaced = 1-LastInterlaced;
			#endif
			if (Screen.ipaq3800)
			{
				unsigned short* dst = (unsigned short*)0xac0755a0;
				unsigned short* src = Buffer;
				register unsigned short s1, s2, s3, s4, s5, s6, s7, s8;
				for ( int y = 0; y < 320; y++ )
				{
					unsigned short* ldst = dst;
					#ifdef INTERLACED
					dst+=LastInterlaced; src+=240;
					#endif
					register unsigned short x=0;
					do //for ( int x = 0; x < 240/8; x++ )
					{
						s1 = *src++;
						s2 = *src++;
						s3 = *src++;
						s4 = *src++;
						s5 = *src++;
						s6 = *src++;
						s7 = *src++;
						s8 = *src++;
						*ldst=s1; ldst-=320;
						*ldst=s2; ldst-=320;
						*ldst=s3; ldst-=320;
						*ldst=s4; ldst-=320;
						*ldst=s5; ldst-=320;
						*ldst=s6; ldst-=320;
						*ldst=s7; ldst-=320;
						*ldst=s8; ldst-=320;
						// *(ldst) = *src++; ldst -= 320;
					} while (++x< 30);
					dst++;
					#ifdef INTERLACED
					dst++; y++; src+=240;
					#endif
				}
			}
			else
			{
				if (Screen.BPP == 16)
				{
					unsigned short* src = Buffer;
					unsigned short* dst;
					if(Screen.pRawFrameBuffer)
						dst = Screen.pRawFrameBuffer;
					else
						dst = (unsigned short*)GXBeginDraw();
					if (Screen.xpitch == 2)
					{
						#ifdef INTERLACED
						if (LastInterlaced) {dst+=256; src+=240;}
						#endif
						if (Screen.ypitch == 480)
						{
							// This is a really nice display (casio's / DELL AXIM)
							// memcpy(dst, src, 240 * 320 * 2 );
							QCopy4(dst, src, 240*160);
						}
						else
						{
							for (int y = 0; y < 320; y++)
							{
								QCopy4(dst, src, 120);
								// memcpy( dst, src, 480 );
								dst += 256;
								src += 240;
								#ifdef INTERLACED
								dst+=256; src += 240; y++;
								#endif
							}
						}
					}
					else
					{
						/*
							* HP IPAQ 36xx series          (xPitch = 640, yPitch =  -2) (Thomas Rhode)
							* HP IPAQ 38xx series          (xPitch =-640, yPitch =   2)
							* Cassiopeia E-115             (xPitch =   2, yPitch = 512)
							* Toshiba e740,Axim X3,Acer nXX(xPitch =   2, yPitch = 480)
						*/
						// Let's do this really clean (and slow:)
						register unsigned int x, y;
						register unsigned short *ldst;
						int xptch = (Screen.xpitch >> 1), yptch = (Screen.ypitch >> 1);
						register unsigned short s1, s2, s3, s4, s5, s6, s7, s8;
					#ifdef INTERLACED
						if(LastInterlaced) {dst+=yptch; src+=240;}
						for (y = 0; y < 160; ++y)
					#else
						for (y = 0; y < 320; ++y)
					#endif
						{
							ldst = dst; // Get current line address
							/*
							for (x = 0; x < 240; x++ )
							{
								*(ldst) = *src++; // one pixel rigth on src
								ldst += xptch;    // add a pixel to the right on dest
							}
							*/
							x=0;
							do
							{
								s1 = *src++;
								s2 = *src++;
								s3 = *src++;
								s4 = *src++;
								s5 = *src++;
								s6 = *src++;
								s7 = *src++;
								s8 = *src++;
								*ldst=s1; ldst+=xptch;
								*ldst=s2; ldst+=xptch;
								*ldst=s3; ldst+=xptch;
								*ldst=s4; ldst+=xptch;
								*ldst=s5; ldst+=xptch;
								*ldst=s6; ldst+=xptch;
								*ldst=s7; ldst+=xptch;
								*ldst=s8; ldst+=xptch;
							} while(++x<30);
							dst += yptch; // add a line to dst buffer
							#ifdef INTERLACED
							dst += yptch; src+=240;
							#endif
						}
					}
					if(!Screen.pRawFrameBuffer)
						GXEndDraw();
				}
				else if (Screen.BPP == 4)
				{
					// Assume this is a monochrome iPaq
					unsigned short* src = Buffer;
					unsigned short* dst;
					if(Screen.pRawFrameBuffer)
						dst = Screen.pRawFrameBuffer;
					else
						dst = (unsigned short*)GXBeginDraw();
					src += 319 * 240; // start at lower left corner
					for ( int y = 0; y < 240; y++ )
					{
						unsigned char* ldst = (unsigned char*)dst;
						unsigned short* lsrc = src;
						for ( int x = 0; x < 160; x++ )
						{
							unsigned short c1 = *(lsrc - 240);
							unsigned short c2 = *lsrc;
							c1 = ((c1 & REDMASK) >> 11) + ((c1 & GREENMASK) >> 5) + (c1 & BLUEMASK);
							c2 = ((c2 & REDMASK) >> 11) + ((c2 & GREENMASK) >> 5) + (c2 & BLUEMASK);
							*(ldst) = (c1 >> 3) + ((c2 >> 3) << 4);
							ldst++;
							lsrc -= 480;
						}
						dst += 160;
						src++;
					}
					if(!Screen.pRawFrameBuffer)
						GXEndDraw();
				}
			} // ipaq3800
		}// landscape
	#endif //_WIN32_WCE
	}

#elif defined(GP2X)
	// GP2X special routines
	// ---------------------

	unsigned long devMem=0;
	volatile unsigned long *gp2x_memreg32=NULL;
	volatile unsigned short *gp2x_memreg16=NULL;
	volatile unsigned long* blitter32=NULL;
	int   vsync_polarity=0;

	extern unsigned int GetUpperRealAddress(void*);

	// Overclocking:
	void CLOCKSPEED(int MHZ)
	{
		#define SYS_CLK_FREQ 7372800
		unsigned v;
		unsigned mdiv,pdiv=3,scale=0;
		MHZ*=1000000;
		mdiv=(MHZ*pdiv)/SYS_CLK_FREQ;
		//printf ("Old value = %04X\r",MEM_REG[0x924>>1]," ");
		//printf ("APLL = %04X\r",MEM_REG[0x91A>>1]," ");
		mdiv=((mdiv-8)<<8) & 0xff00;
		pdiv=((pdiv-2)<<2) & 0xfc;
		scale&=3;
		gp2x_memreg16[0x910>>1]= mdiv | pdiv | scale;
	}


	const unsigned long SCREEN0_HW = 0x3101000; // this is the one to restore afterwards
	const unsigned long SCREEN1_HW = 0x3381000; // this is the back buffer when ending

	// swap these toghether with pRawFrameBuffer and cBuffer
	unsigned long FRONTBUFFER_HW = SCREEN0_HW;
	unsigned long BACKBUFFER_HW = SCREEN1_HW;

	#define UNSATURATE 	asm volatile ("nop"); asm volatile ("nop"); asm volatile ("nop"); asm volatile ("nop");


	// dummy blit to force MMSP2's blitter to flush it's cache
	void gp2x_dummy_blit(void)
	{
	/*
		//Set the source address
		blitter32[0x0010 >> 2] = FRONTBUFFER_HW; // MESGSRCADDR

		// Blitter seems to have a 16 byte buffer, so force flushing by
		// drawing 8 16bit pixels onto themselves.
		blitter32[0x0000 >> 2] = (1 << 5) | (1 << 6); // MESGDSTCTRL = MESG_DSTBPP_16|MESG_DSTENB
		blitter32[0x0004 >> 2] = FRONTBUFFER_HW; // MESGDSTADDR = front screen buffer
		blitter32[0x000C>>2] = (1<<8)|(1<<5); // MESGSRCCTRL=MESG_INVIDEO | *MESG_SRCBPP_16*
		// blitter32[0x0020>>2] = (1<<5)|(1<<4); // MESGPATCTRL = 0 | *MESG_PATENB | MESG_PATBPP_1*
		blitter32[0x002C>>2] = (1 << 16) | (1<<4); // MESGSIZE = (1<<MESGHEIGHT)|MESGPATBPP_1 (number of bytes to copy)
		blitter32[0x0030>>2] = (1<<8) | (1<<9) | (0xaa); // MESGCTRL = 

		// Make sure blitter is ready
		while (blitter32[0x0034 >> 2] & 1)
		{
			UNSATURATE
		}
		asm volatile ("" ::: "memory");
		blitter32[0x0034 >> 2] = 0x0001;
	*/
		// SDL method
		do {} while (blitter32[MESGSTATUS] & MESG_BUSY);
		blitter32[MESGDSTCTRL] = MESG_DSTENB | MESG_DSTBPP_16;
		blitter32[MESGDSTADDR] = 0x3101000;
		blitter32[MESGDSTSTRIDE] = 4;
		blitter32[MESGSRCCTRL] = MESG_SRCBPP_16 | MESG_INVIDEO;
		blitter32[MESGPATCTRL] = MESG_PATENB | MESG_PATBPP_1;
		blitter32[MESGFORCOLOR] = ~0;
		blitter32[MESGBACKCOLOR] = ~0;
		blitter32[MESGSIZE] = (1 << MESG_HEIGHT) | 32;
		blitter32[MESGCTRL] = MESG_FFCLR | MESG_XDIR_POS | MESG_YDIR_POS | 0xAA;
		asm volatile ("":::"memory");
		blitter32[MESGSTATUS] = MESG_BUSY;
		do {
			UNSATURATE
		} while (blitter32[MESGSTATUS] & MESG_BUSY);
	}

	inline void gp2x_vblank()
	{
		// gp2x-wiki:
		while(  gp2x_memreg16[GPIOB_PINLVL] & GPIOB_VSYNC ){UNSATURATE}
		while(!(gp2x_memreg16[GPIOB_PINLVL] & GPIOB_VSYNC)){UNSATURATE}
	/*
		if (vsync_polarity)
			while ( (gp2x_memreg16[GPIOB_PINLVL] & GPIOB_VSYNC)) {UNSATURATE}
		else
			while (!(gp2x_memreg16[GPIOB_PINLVL] & GPIOB_VSYNC)) {UNSATURATE}
	*/
	}


	// Flip the framebuffer pointer
	void gp2x_video_flip()
	{
		// swap pointers
		SwapMe(BACKBUFFER_HW, FRONTBUFFER_HW);
		SwapMe(Screen.cBuffer, Screen.pRawFrameBuffer);


		// VBlank
		// gp2x_dummy_blit(); - this seems to cause trouble

		// wait for the blitter to finish, if not, yet (maybe?)
		do {} while (blitter32[MESGSTATUS] & MESG_BUSY); // do _NOT_ nop; here!
		gp2x_vblank();

		// Wait to be on even field (non-interlaced always returns 0)
		//  do {} while (data->io[SC_STATUS] & SC_DISP_FIELD);

		// set still RGB image pointer
		// odd part
		gp2x_memreg16[MLC_STL_EADRL]=(unsigned short)(FRONTBUFFER_HW & 0xFFFF);
		// even part
		gp2x_memreg16[MLC_STL_EADRH]=(unsigned short)(FRONTBUFFER_HW >> 16);

		// this only for interlace?
		// odd part
		gp2x_memreg16[MLC_STL_OADRL]=(unsigned short)(FRONTBUFFER_HW & 0xFFFF);
		// even part
		gp2x_memreg16[MLC_STL_OADRH]=(unsigned short)(FRONTBUFFER_HW >> 16);

		// Wait for vblank to end (to prevent 2 close page flips in one frame)
		// while (!(data->io[GPIOB_PINLVL] & GPIOB_VSYNC));
	}


	// restore the framebuffer pointer for next apps
	void gp2x_video_restore()
	{
		if(FRONTBUFFER_HW != SCREEN0_HW)
			gp2x_video_flip();
	}

	// fill solid rectangle -workes, yahoo!
	// seems wo mess up screen y scaling!?
	void gp2x_fillrect(int x, int y, int w, int h, unsigned short col)
	{
		blitter32[0x4 >> 2] = (BACKBUFFER_HW + (y*640) + (x<<1) ) & ~3; // dest ptr
		blitter32[0x0] = 1<<5  | ((x & 0x00000001) << 4); 				// Destination is 16 bpp
		blitter32[0x8 >> 2] = 640;										// Destination stride size in bytes
		blitter32[0xC >> 2] = 0;  // source controlled by CPU

	//	blitter32[0x20 >> 2] = 1 << 5 | 2 << 3;			// Perform ROP with a 1bpp pattern
		blitter32[0x20 >> 2] = (1<<5) | (1<<4);			// Perform ROP with a 1bpp pattern
		blitter32[0x24 >> 2] = col;						// Setup the foreground and background colors for the pattern to be the same
		blitter32[0x28 >> 2] = col;						// so we don't actually have to upload a pattern image.
		blitter32[0x2c >> 2] = (h << 16) | (w << 0);	// Height and width to blit
	//	blitter32[0x30 >> 2] = (1<<8) | (1<<9 ) | 0xf0;	// Fill from top left to bottom right
		blitter32[0x30 >> 2] = (1 << 10) | (1 << 9) | (1 << 8) | 0xf0;	// Fill from top left to bottom right

		// Wait for blitter to be free, start it, and then wait for completion.
		// Throw in some nop's so we don't saturate the address bus with polling requests.
		while (blitter32[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
		asm volatile ("" ::: "memory");
		blitter32[0x34 >> 2] = 1;
		gp2x_dummy_blit();
	}

	void gp2x_clear_back_buffer(void)
	{
		blitter32[0x4 >> 2] = BACKBUFFER_HW;
		blitter32[0x0] = 1<<5; 											// Destination is 16 bpp
		blitter32[0x8 >> 2] = 640;										// Destination stride size in bytes
		blitter32[0x20 >> 2] = 1 << 5  | 2 << 3;					// Perform ROP with a 1bpp pattern
		blitter32[0x24 >> 2] = 0;								// Setup the foreground and background colors for the pattern to be the same
		blitter32[0x28 >> 2] = 0;								// so we don't actually have to upload a pattern image.
		blitter32[0x2c >> 2] = (240 << 16) | (320 << 0);				// Height and width to blit
		blitter32[0x30 >> 2] = (1<<8) | (1<<9 ) | 0xf0;			// Fill from top left to bottom right

		// Wait for blitter to be free, start it, and then wait for completion.
		// Throw in some nop's so we don't saturate the address bus with polling requests.
		while (blitter32[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
		asm volatile ("" ::: "memory");
		blitter32[0x34 >> 2] = 1;
		gp2x_dummy_blit();
	}



	void gp2x_blit(unsigned int hardware_src, int x_from, int y_from, int src_stride_bytes, int x, int y, int w, int h)
	{
		blitter32[0x4 >> 2] = (BACKBUFFER_HW + (y*640) + (x<<1) ) & ~3; // dest ptr
		blitter32[0x0] = 1<<5  | ((x & 0x00000001) << 4) ; 			// Destination is 16 bpp
		blitter32[0x8 >> 2] = 640;									// Destination stride size in bytes

		//Set the source address
		blitter32[0x0010 >> 2] = (hardware_src +(y_from*src_stride_bytes)+(x_from<<1) )&~3;
		//Set the pitch of source in bytes
		blitter32[0x0014 >> 2] = src_stride_bytes;
		//Do nothing with pattern
		blitter32[0x0020 >> 2] = 0;

		// Set a 16bit source, enable source and say the source is not controlled by CPU(?)
		blitter32[0x000C >> 2] = (1 << 8) | (1 << 7) | (1 << 5) | ((x_from & 0x00000001) << 4);

		// Clear the source input FIFO, positive X,Y. And do a copy ROP.
		blitter32[0x0030 >> 2] = (1 << 10) | (1 << 9) | (1 << 8) | 0xCC;

		blitter32[0x002c >> 2] = (h << 16) | (w << 0);				// Height and width to blit

		// Wait for blitter to be free, start it, and then wait for completion.
		// Throw in some nop's so we don't saturate the address bus with polling requests.
		while (blitter32[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
		asm volatile ("" ::: "memory");
		blitter32[0x34 >> 2] = 1;
		gp2x_dummy_blit();
	}


	void gp2x_blit_cookie(unsigned int hardware_src, int x_from, int y_from, int src_stride_bytes, int x, int y, int w, int h, unsigned short cookie)
	{
		blitter32[0x4 >> 2] = (BACKBUFFER_HW + (y*640) + (x<<1) ) & ~3; // dest ptr
		blitter32[0x0] = 1<<5  | ((x & 0x00000001) << 4) ; 			// Destination is 16 bpp
		blitter32[0x8 >> 2] = 640;									// Destination stride size in bytes

		//Set the source address
		blitter32[0x0010 >> 2] = (hardware_src +(y_from*src_stride_bytes)+(x_from<<1) )&~3;
		//Set the pitch of source in bytes
		blitter32[0x0014 >> 2] = src_stride_bytes;
		//Do nothing with pattern
		blitter32[0x0020 >> 2] = 0;

		// Set a 16bit source, enable source and say the source is not controlled by CPU(?)
		blitter32[0x000C >> 2] = (1 << 8) | (1 << 7) | (1 << 5) | ((x_from & 0x00000001) << 4);

		// Clear the source input FIFO, positive X,Y. And do a copy ROP.  cookie        enable cookie
		blitter32[0x0030 >> 2] = (1 << 10) | (1 << 9) | (1 << 8) | 0xCC | (cookie<<16)|(1<<11);

		blitter32[0x2c >> 2] = (h << 16) | (w << 0);				// Height and width to blit

		// Wait for blitter to be free, start it, and then wait for completion.
		// Throw in some nop's so we don't saturate the address bus with polling requests.
		while (blitter32[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
		asm volatile ("" ::: "memory");
		blitter32[0x34 >> 2] = 1;
		gp2x_dummy_blit();
	}

	// ---------------------------------
	// GP2X - Flipscreen
	// ---------------------------------
	inline void UpdateDisplay(unsigned short* Buffer, bool Landscape)
	{
		gp2x_video_flip();
	}


#elif defined(GPHWIZ)

	#define TIMER_BASE3 0x1980
	#define TIMER_REG(x) g_memregs32[(TIMER_BASE3 + x) >> 2]
	int gTimerIsInit=0;
	void wiz_timer_init(void)
	{
			TIMER_REG(0x44) = 0x922;
			TIMER_REG(0x40) = 0x0c;
			TIMER_REG(0x08) = 0x6b;
			gTimerIsInit=1;
	}

	extern "C" unsigned int wiz_timer_get_ticks_us(void)
	{
		static unsigned int last_time=0;
		static unsigned int accumulator=0;
		TIMER_REG(0x08) = 0x4b;  /* run timer, latch value */
		unsigned int val = TIMER_REG(0);

		if(val < last_time) // wraps
		{
			accumulator += last_time;
		}
		last_time=val;
		return val + accumulator;
	}

	void wiz_timer_cleanup(void)
	{
			if(!gTimerIsInit) return;
			TIMER_REG(0x40) = 0x0c;
			TIMER_REG(0x08) = 0x23;
			TIMER_REG(0x00) = 0;
			TIMER_REG(0x40) = 0;
			TIMER_REG(0x44) = 0;
			gTimerIsInit=0;
	}


	void wiz_wait_vsync_poll()
	{
		typedef enum
		{
			DPC_HTOTAL     = 0x307C,
			DPC_VTOTAL     = 0x3084,
			DPC_VSWIDTH    = 0x3086,
			DPC_VASTART    = 0x3088,
			DPC_VAEND      = 0x308A,
			DPC_CONTROL0   = 0x308C
		} wiz_dplc_regs_enum;
		
		// >>1 due to 16 bit pointer (instead of 32 bit)
		while((g_memregs16[DPC_CONTROL0 >> 1] & (1 << 10)) == 0);
		g_memregs16[DPC_CONTROL0 >> 1] |= (1 << 10);
	}
	
	extern "C" void do_rotated_blit (void *dst, void *src);

	// ---------------------------------
	// WIZ - Flipscreen
	// ---------------------------------
	inline void UpdateDisplay(unsigned short* Buffer, bool Landscape)
	{
		register const unsigned short* pSrc;
		register unsigned short* pDst;
		pSrc=Buffer; pDst = Screen.pRawFrameBuffer;


		if(gWizTearing)
		{
			QCopy4(pDst, pSrc, 320*240/2);
		}
		else
		{
			// ASM code - really fast
			// wiz_wait_vsync_poll(); <-- causes a bad tearing in right corner
			do_rotated_blit ((void *)pDst, (void *)pSrc); return;
			
			// WIZ: 240x320 display -> 320x240 buffer
			// Display is:
			// y<-+-----+
			//   O|     |O
			//    +-----+
			//          |
			//          v x
			// (pSrc++, pDst+=240) - reversed loop - faster reads than writes
			register unsigned short* pLine = pDst + 319*240;
			for(register int y=0; y<240; ++y)
			{
				pDst = pLine++;

				#define D_ONE {*pDst = *pSrc++; pDst-=240;}
				#define D_EIGHT D_ONE D_ONE D_ONE D_ONE    D_ONE D_ONE D_ONE D_ONE
				#define D_EIGHTY D_EIGHT D_EIGHT D_EIGHT D_EIGHT D_EIGHT   D_EIGHT D_EIGHT D_EIGHT D_EIGHT D_EIGHT

				// one line
				// the loop is supposed to be faster?? http://www.gp32x.com/board/index.php?/topic/50876-copy-320x240-buffer-to-240x320-screen/
				// for(register int x=0; x<320; ++x){D_ONE;}
				D_EIGHTY D_EIGHTY D_EIGHTY D_EIGHTY // full 320 pixels
			}
		}
	}
	

#elif defined(WANT_SDL)
	#include <SDL/SDL.h>
	// ---------------------------------
	// SDL - Flipscreen
	// ---------------------------------
	inline void UpdateDisplay(unsigned short* Buffer, bool Landscape)
	{
		SDL_Surface* pSurf = SDL_GetVideoSurface();
		if(!pSurf)
		{
			printf("SDL_GetVideoSurface == NULL\n");
			int* pNULL=NULL;
			*pNULL = 0;
		}
		switch(pSurf->format->BitsPerPixel)
		{
			case 16:
			{
				const unsigned short* pSrc;
				unsigned short* pDst;
				SDL_LockSurface(pSurf);
				pSrc=Buffer; pDst = (unsigned short*) pSurf->pixels; // + row_ptr*(pSurf->h - 1);
				QCopy4(pDst, pSrc, (pSurf->format->BytesPerPixel * pSurf->w * pSurf->h) >> 2);

				SDL_UnlockSurface(pSurf); // Used to be SDL_FreeSurface - but that's wrong, no??
				// SDL_Flip(pSurf);
				SDL_UpdateRect(pSurf, 0, 0, 0, 0);
			}
			break;
			case 24: // RGB
			{
				const unsigned short* pSrc;
				unsigned char* pDst;
				SDL_LockSurface(pSurf);
				pSrc = Buffer;
				pDst = (unsigned char*)pSurf->pixels; // + row_ptr*(pSurf->h - 1);

				int count = pSurf->w * pSurf->h;
				unsigned char rgb;
				while(--count > 0)
				{
				if(*pSrc != 0)
					rgb = GETR( *pSrc ); *pDst++ = rgb;
					rgb = GETG( *pSrc ); *pDst++ = rgb;
					rgb = GETB( *pSrc ); *pDst++ = rgb;
					++pSrc;
				}
				SDL_UnlockSurface(pSurf);
				SDL_UpdateRect(pSurf, 0, 0, 0, 0);
			}
			break;
			case 32: // Emscripten - RGB + a=15!
			{
				const unsigned short* pSrc;
				unsigned char* pDst;
				SDL_LockSurface(pSurf);
				pSrc = Buffer;
				pDst = (unsigned char*)pSurf->pixels; // + row_ptr*(pSurf->h - 1);

				unsigned int* pDst32 = (unsigned int*)pSurf->pixels;
				#if 1
					#if defined(INTERLACED)
						static int iInterlace = 0;
						iInterlace = 1-iInterlace;
						register int count = (int)(pSurf->w * pSurf->h); // NOT unsigned!
						count >>= 1; // half the pixels per run.
						if(iInterlace)
						{
							pDst32+=count; pSrc+=count;
						}
						count >>=2; // loop unroll 4 lines
						while(count--)
						{
							*pDst32++ = UnPackColor(*pSrc++) | 0xff000000;
							*pDst32++ = UnPackColor(*pSrc++) | 0xff000000;
							*pDst32++ = UnPackColor(*pSrc++) | 0xff000000;
							*pDst32++ = UnPackColor(*pSrc++) | 0xff000000;
						}
					#else
						register int count = (int)(pSurf->w * pSurf->h); // NOT unsigned!
						while(count--)
						{
							*pDst32++ = UnPackColor(*pSrc++) | 0xff000000;
						}
					#endif


				#else
					int count = pSurf->w * pSurf->h;
					unsigned char rgb;
					while(--count > 0)
					{
						rgb = GETR( *pSrc ); *pDst++ = rgb;
						rgb = GETG( *pSrc ); *pDst++ = rgb;
						rgb = GETB( *pSrc ); *pDst++ = rgb;
						*pDst++ = 0xff; // A
						++pSrc;
					}
				#endif
				SDL_UnlockSurface(pSurf);
				SDL_UpdateRect(pSurf, 0, 0, 0, 0);
			}
			break;
		}
	}
#endif


// ---------------------------------
// PutRect - fill a rectangle
// dx,dy - number of pixels to write
// ---------------------------------
void oScreen::PutRect(int x, int y, int dx, int dy, unsigned short col)
{
	if(!db_width || !db_height) return;

	// negative width/height
	if(dx<0){x+=dx; dx=-dx;}
	if(dy<0){y+=dy; dy=-dy;}

	// clipping
	if(x<0) {dx+=x; x=0;}
	if(y<0) {dy+=y; y=0;}
	x+=viewport_x;
	y+=viewport_y;
	if(x+dx > viewport_x+viewport_w ) {dx=(viewport_x+viewport_w) -x;}
	if(y+dy > viewport_y+viewport_h)  {dy=(viewport_y+viewport_h) -y;}

	// rect x,y out of bounds
	if(dx<=0 || dy<=0) return;

	#if defined(GP2X)
		// Quick solid putrect with Blitter on GP2X hardware
		if(this->alpha_blending==0)
		{
			gp2x_fillrect(x,y,dx,dy,col);
			return;
		}
	#endif

	// can speed this up now
	unsigned short* pScreen = getColorBuffer() + x + y*db_width;

	int num_q = dx>>2, num_r = (dx & 0x00000003); // number of 4 pixels vs. remaining pixels -> loop unrolling
	int shiftline = db_width - dx;

	// draw solid, go fast! (Even on SDL)
	if(alpha_blending == 0)
	{
		unsigned long colcol = col | ((unsigned long)col << 16);
		// unsigned long* pScLong;
		for(y=0; y<dy; ++y)
		{
			register int rx;
			// pScLong = (unsigned long*)pScreen;
			for(rx=0; rx<num_q; ++rx)
			{
				/* crashes on PocketPC. GP2X will use blitter anyway

				DRAWRECT 1,20,18,1,0 // crashed on 2x2 ints drawrect

				VALID_SCREEN(pScLong);
				VALID_SCREEN(((unsigned short*)pScLong)+1);
				VALID_SCREEN(pScLong+1);
				VALID_SCREEN(((unsigned short*)pScLong+1)+1);
				*pScLong++ = colcol;
				*pScLong++ = colcol;
				pScreen+=4;
				*/
				
				VALID_SCREEN(pScreen); *pScreen++ = col;
				VALID_SCREEN(pScreen); *pScreen++ = col;
				VALID_SCREEN(pScreen); *pScreen++ = col;
				VALID_SCREEN(pScreen); *pScreen++ = col;
			}

			if(num_r)
			{
				rx = num_r;
				VALID_SCREEN(pScreen); *pScreen++ = col;
				if(--rx)
				{
					VALID_SCREEN(pScreen); *pScreen++ = col;
					if(--rx)
					{
						VALID_SCREEN(pScreen); *pScreen++ = col;
					}
				}
			}
			pScreen+=shiftline;
		}
	}
	else
	{
		for(y=0; y<dy; y++)
		{
			register int rx;
			for(rx=0; rx<num_q; ++rx)
			{
				// PutPixel(x, y, col);
				VALID_SCREEN(pScreen); m_BlendFunc(pScreen++, col);
				VALID_SCREEN(pScreen); m_BlendFunc(pScreen++, col);
				VALID_SCREEN(pScreen); m_BlendFunc(pScreen++, col);
				VALID_SCREEN(pScreen); m_BlendFunc(pScreen++, col);
			}
			if(num_r)
			{
				VALID_SCREEN(pScreen); m_BlendFunc(pScreen++, col);
				rx = num_r;
				if(--rx)
				{
					VALID_SCREEN(pScreen); m_BlendFunc(pScreen++, col);
					if(--rx)
					{
						VALID_SCREEN(pScreen); m_BlendFunc(pScreen++, col);
					}
				}
			}
			pScreen+=shiftline;
		}
	}
}




// - MEMBER FUNCTIONS ------------------------------------------


// --------------------------------------------------------------- //
// 
// --------------------------------------------------------------- //
oScreen::oScreen()
{
	m_clear_color=0;
	#ifdef THREADED_UPDATE
	BufferLock=false;
	ThreadTempBuffer=NULL;
	#endif
	Window=NULL; alpha_blending=0; cBuffer=NULL; Landscape=true;ScreenRange=0;
	m_BlendFunc = BlendPixelSet;
	alpha_testing=true;
}

// --------------------------------------------------------------- //
// 
// --------------------------------------------------------------- //
#ifdef GP2X
struct OLDSTATE
{
	unsigned short fastio, gfx_state;
};
static OLDSTATE g_oldState;

#endif

oScreen::~oScreen()
{
extern void DestroyMemPool();
	Close();

	#ifdef GP2X
		printf("Cleaning up...");
		// restore registers
		gp2x_memreg16[0x0904 >> 1]=g_oldState.fastio;
		gp2x_memreg16[0x090a >> 1]=g_oldState.gfx_state;
		// wait for blitter before quit
		while (blitter32[0x0034 >> 2] & 1)
		{
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
		}
		CLOCKSPEED(200);
	#endif
	
	DestroyMemPool();
}




// --------------------------------------------------------------- //
// 
// --------------------------------------------------------------- //
void  oScreen::Close()
{
	g_debug("setdrawbuf\n");
	SetDrawBuffer();
#ifdef THREADED_UPDATE
	if (ThreadID)
	{
		ExitThread(ThreadID);
	}
	gl_free(ThreadTempBuffer);
	ThreadTempBuffer=NULL;
#endif

#if defined(WIN32) && !defined(WANT_SDL)
	//	DeleteDC(Context);
	// if (Window && Buffer1 && ::IsWindow(Window)) ReleaseDC(Window, Buffer1);
	#ifdef _WIN32_WCE
	if(pRawFrameBuffer && GXCloseDisplay)
	{
		g_debug("gxclosedsp\n");
		GXCloseDisplay();
	}
	#endif
#endif

	if(cBuffer)
	{
		#ifdef TINYGL
		glFinishOnBuffer();
		#endif

		#if !defined(WANT_SDL) && defined(WIN32)
			if(Buffer2)
			{
				g_debug("delete Buffer2\n");
				DeleteObject(Buffer2);
				cBuffer=NULL; // that was a pointer to the pixels of Buffer2, gained from CreateDIBSection()
			}
			Buffer2=NULL;
		#endif

		#if !defined(POCKETPC) || defined(_WIN32_WCE)
			if(cBuffer)
			{
				g_debug("free cbuffer\n");
				gl_free(cBuffer);
			}
		#endif
	}
	
	#ifdef TRENKI3D
	if(zBuffer)
	{
		g_debug("free zbuffer\n");
		gl_free(zBuffer);
	}
	#endif

	#if defined(GPHWIZ)
		WizEnableTearing( 1 );
		close(g_dev_fb0);
		g_dev_fb0 = 0;
		wiz_timer_cleanup();
		close(g_dev_mem);
		g_dev_mem=0;
		g_memregs16 = NULL;
		g_memregs32 = NULL;
	#endif

	
	// Set to NULL
	zBuffer = NULL;
	cBuffer=NULL;
	pRawFrameBuffer=NULL;
	g_debug("close screen done\n");
}


void oScreen::Alphamode(short mode)
{
	alpha_blending=max(-255, min(255, mode));
	if (mode==0)
		m_BlendFunc = BlendPixelSet;
	else if (mode>0)
		m_BlendFunc = BlendPixelAdd;
	else
		m_BlendFunc = BlendPixelSub;
}

#ifdef _WIN32_WCE
void oScreen::createpalette( void )
{
	// Code by L Bailey, modified by Phantom
	LPLOGPALETTE logPalette = NULL;
	WORD palEntries = 256;
	if (BPP == 8) 
	{
		// logPalette = (LOGPALETTE *) LocalAlloc (LMEM_FIXED, sizeof(LOGPALETTE) + sizeof(PALETTEENTRY) * palEntries);	

		logPalette = (LOGPALETTE*) new BYTE[sizeof(LOGPALETTE) + sizeof(PALETTEENTRY) * palEntries];
		logPalette->palVersion = (WORD) 0x0300;
		logPalette->palNumEntries = palEntries;
		// Create a 2-3-2 palette
		for (int i = 0; i < palEntries; i++) 
		{
			int r = (i & (3 << 5)) >> 5;
			int g = (i & (7 << 2)) >> 2;
			int b = i & 3;
			logPalette->palPalEntry[i].peRed = r << 6;
			logPalette->palPalEntry[i].peGreen = g << 5;
			logPalette->palPalEntry[i].peBlue = b << 6;
			logPalette->palPalEntry[i].peFlags = 0;
		}
		hPalettize = CreatePalette(logPalette);
		// LocalFree ((HLOCAL) logPalette);
		delete [] (BYTE*)logPalette;
	}
	// Activate the palette
	HDC paletteHDC = GetDC(0);
	SelectPalette(paletteHDC, hPalettize, false);
	RealizePalette(paletteHDC);
	ReleaseDC(0, paletteHDC);
}
#endif


void oScreen::GetScreenMetrics(int* swidth, int* sheight)
{
	if (Landscape)
	{
		*swidth=db_height; *sheight=db_width;
	}
	else
	{
		*swidth=db_width; *sheight=db_height;
	}
}


// --------------------------------------------------------------- //
// 
// --------------------------------------------------------------- //
void oScreen::Create(HWND hwnd, int swidth, int sheight, bool landscape)
{
extern void InitMemPool();
	g_debug("mempool\n");
	InitMemPool();
	Close();
	Window = hwnd;
	Landscape=landscape;
	cBufferWidth =swidth;
	cBufferHeight=sheight;
	db_width =swidth;
	db_height=sheight;
	ScreenRange=db_width*db_height;
	Viewport(0,0,0,0);
	pRawFrameBuffer=NULL;
	g_debug("get screen ptr\n");

// Hardcode this in
#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
	swidth=320; sheight=240;
#endif


#if defined(POCKETPC)
	#ifdef _WIN32_WCE // True PocketPC
		// METHOD 1 - GETRAWFRAMEBUFFER
		#define GETRAWFRAMEBUFFER   0x00020001

		#define FORMAT_565 1
		#define FORMAT_555 2
		#define FORMAT_OTHER 3

		typedef struct _RawFrameBufferInfo
		{
		WORD wFormat;
		WORD wBPP;
		VOID *pFramePointer;
			int  cxStride;
			int  cyStride;
		int  cxPixels;
		int  cyPixels;
		} RawFrameBufferInfo;
		RawFrameBufferInfo rfbi;
		memset(&rfbi, 0, sizeof(rfbi));
		HDC hdc = GetDC(NULL);
		ExtEscape(hdc, GETRAWFRAMEBUFFER, 0, NULL, 
		sizeof(RawFrameBufferInfo), (char *) &rfbi);
		ReleaseDC(NULL, hdc);
		if(rfbi.pFramePointer)
		{
			xpitch = rfbi.cxStride;
			ypitch = rfbi.cyStride;
			db_width  = rfbi.cxPixels;
			db_height = rfbi.cyPixels;
			BPP    = rfbi.wBPP;
			if((getWidth() == swidth && getHeight()==sheight) || 
			   (getHeight()== swidth && getWidth() ==sheight) )
			{
				g_debug("Using GETRAWFRAMEBUFFER\n");
				pRawFrameBuffer = (unsigned short*)rfbi.pFramePointer;
			}
		}
		if(!pRawFrameBuffer) // Using GAPI
		{
			g_debug("Trying GAPI\n");
			if(GXOpenDisplay && GXOpenDisplay(hwnd, GX_FULLSCREEN) != 0)
			{
				g_debug("Using GAPI\n");
				// Get display properties
				GXDisplayProperties dp = GXGetDisplayProperties();
				BPP = (short)dp.cBPP;
				if (BPP == 8) createpalette();
				xpitch = dp.cbxPitch; // 640
				ypitch = dp.cbyPitch; // -2
				db_width  = dp.cxWidth;
				db_height = dp.cyHeight;
			}
			else
			{
				g_debug("No GAPI - fallback to GDI.\n");
			}
		}
		
		// landscape device!! Turn it so you assume it's portrait again
		// Minion:Pitch X,Y = 2, 640
		if(pRawFrameBuffer && db_width > db_height)
		{
			g_debug("landscape device\n");
			int swp;
			swp = db_width; db_width=db_height; db_height=swp;    // w=240,h=320
			swp = xpitch; xpitch=-ypitch; ypitch=swp; // px=-640, py=2
			pRawFrameBuffer -= ((db_width-1) * (xpitch>>1) );
		}

		// force display orientation of this program
		if(pRawFrameBuffer && swidth>sheight && db_width<db_height)
		{
			int swp = db_width; db_width=db_height; db_height=swp;
		}
	#endif // !WIN_CE
	
	#if defined(WIN32) && !defined(WANT_SDL)
		// PC/WinCE emulation with Bitmap
		if(!pRawFrameBuffer)
		{
			HDC hdc = GetDC(hwnd);
			// GETGR/G/B -> Convert 16Bit short to 255 char component
			struct oBMPInfo
			{
				BITMAPINFOHEADER bmiHeader;
				unsigned long  bmiColors[3]; // For 16 bit BI_BITFIELDS compression
			};
			oBMPInfo BitmapInfo;
			
			BitmapInfo.bmiHeader.biSize          = sizeof(BITMAPINFOHEADER);
			if(landscape)
			{
				BitmapInfo.bmiHeader.biWidth         =  db_height;
				BitmapInfo.bmiHeader.biHeight        = -db_width;
			}
			else
			{
				BitmapInfo.bmiHeader.biWidth         =  db_width;
				BitmapInfo.bmiHeader.biHeight        = -db_height;
			}
			BitmapInfo.bmiHeader.biPlanes        = 1;
			BitmapInfo.bmiHeader.biBitCount      = 16;
			BitmapInfo.bmiHeader.biCompression   = BI_BITFIELDS; //BI_RGB;
			BitmapInfo.bmiHeader.biSizeImage     = 0;
			BitmapInfo.bmiHeader.biXPelsPerMeter = 0;
			BitmapInfo.bmiHeader.biYPelsPerMeter = 0;
			BitmapInfo.bmiHeader.biClrUsed       = 0;
			BitmapInfo.bmiHeader.biClrImportant  = 0;
			((DWORD*) BitmapInfo.bmiColors)[0] = (31 << SHIFTRED); /* red mask */
			((DWORD*) BitmapInfo.bmiColors)[1] = (ANDGREEN << 5); /* green mask */
			((DWORD*) BitmapInfo.bmiColors)[2] = (31 << 0); /* blue mask */

			Buffer2 = CreateDIBSection(hdc, (BITMAPINFO*)&BitmapInfo, DIB_RGB_COLORS, (VOID **)&Screen.cBuffer, NULL, NULL);
			::ReleaseDC(hwnd, hdc);

			pRawFrameBuffer=NULL; // must lock. See UpdateDisplay function
			BPP=16;
			xpitch=1; ypitch=db_width;
		}
	#endif

	g_debug("Screen W,H,bpp: %d %d %d\n", db_width, db_height, BPP);
	g_debug("Pitch X,Y: %d %d\n", xpitch, ypitch);
	if(!cBuffer) // might be bitmap buffer already
		cBuffer = (unsigned short*) gl_malloc(db_width*db_height*sizeof(unsigned short));
	
	#if defined(_WIN32_WCE)
		//Detect ipaq 3800 series - they might have a bug
		WCHAR oemstr[100];
		SystemParametersInfo(SPI_GETOEMINFO, sizeof( oemstr ), oemstr, 0 ); 
		ipaq3800 = false;
		if ((oemstr[12] == 'H') && (oemstr[13] == '3') && (oemstr[14] == '8')) ipaq3800 = true;
	#endif

	
#elif defined(GP2X)
	// Blitter access
	extern volatile int Uppermemfd; // memory.cpp (TinyGL)

	// Uppermemfd = open("/dev/mem", O_RDWR);
	gp2x_memreg32=(volatile unsigned long *)mmap(0, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, Uppermemfd, 0xc0000000);
	gp2x_memreg16=(volatile unsigned short *)gp2x_memreg32;

	// Get vsync polarity (GPH changes this every now and then)
	vsync_polarity = gp2x_memreg16[DPC_V_SYNC] & DPC_VSFLDPOL;

	blitter32 = (unsigned long*)mmap(0, 0x100, PROT_READ|PROT_WRITE, MAP_SHARED, Uppermemfd, 0xE0020000);
	g_debug("Got blitter at: %p\n", blitter32);
	
	// store old state
	g_oldState.fastio   = gp2x_memreg16[0x0904 >> 1];
	g_oldState.gfx_state= gp2x_memreg16[0x090a >> 1];
	
	// enable all video and graphic devices
	gp2x_memreg16[0x090a >> 1] = 0xFFFF;
	// enable fastio
	gp2x_memreg16[0x0904 >> 1] |= (1 << 10);
	pRawFrameBuffer=(unsigned short*)mmap(0, 320*240*2, PROT_READ|PROT_WRITE, MAP_SHARED, Uppermemfd, SCREEN0_HW); // 1st frame buffer
	// we _do_ need volatile here
	static volatile unsigned short* pBackBuffer = (volatile unsigned short*)mmap(0, 320*240*2, PROT_READ|PROT_WRITE, MAP_SHARED, Uppermemfd, SCREEN1_HW); // 2nd frame buffer
	cBuffer=(unsigned short*) pBackBuffer;

	CLOCKSPEED(240);

	// set still RGB image pointer, gp2xmenu might have swapped it
	FRONTBUFFER_HW = SCREEN0_HW;
	BACKBUFFER_HW = SCREEN1_HW;
	gp2x_memreg16[0x290E>>1]=gp2x_memreg16[0x2912>>1]=(unsigned short)(SCREEN0_HW & 0xFFFF);
	gp2x_memreg16[0x2910>>1]=gp2x_memreg16[0x2914>>1]=(unsigned short)(SCREEN0_HW >> 16);

	// Do dummy blit w/o waiting
	blitter32[0x0000 >> 2] = (1 << 5) | (1 << 6);
	blitter32[0x0004 >> 2] = SCREEN0_HW; // front screen buffer
	blitter32[0x000C>>2] = 1<<8;
	blitter32[0x0020>>2] = 0;
	blitter32[0x002C>>2] = (1 << 16) | 8; // number of bytes to copy
	blitter32[0x0030>>2] = 1<< 8 | 1<< 9 | 0xaa;
	asm volatile ("" ::: "memory");
	blitter32[0x0034 >> 2] = 0x0001;

	// Now fait for dummy blit to complete
	while (blitter32[0x0034 >> 2] & 1)
	{
		asm volatile ("nop");
		asm volatile ("nop");
		asm volatile ("nop");
		asm volatile ("nop");
	}
	// and clear
	gp2x_clear_back_buffer();

	// Wiz
#elif defined(GPHWIZ)

	g_debug("wiz: %d\n", __LINE__);
	g_dev_mem = open("/dev/mem", O_RDWR);
	g_memregs32 = (volatile unsigned long*)mmap(0, 0x20000, PROT_READ|PROT_WRITE, MAP_SHARED, g_dev_mem, 0xC0000000);
	g_memregs16 = (volatile unsigned short*)g_memregs32;
	g_dev_fb0 = open("/dev/fb0", O_RDWR);
	g_debug("wiz: %d\n", __LINE__);
	WizEnableTearing( 0 ); // start in nice mode
	g_debug("wiz: %d\n", __LINE__);
	wiz_timer_init();

	// Open the raw framebuffer to draw on that (quicker than SDL)
	g_debug("wiz: %d\n", __LINE__);
	pRawFrameBuffer=(unsigned short *)mmap(0, 320*240*sizeof(short), PROT_WRITE, MAP_SHARED, g_dev_fb0, 0);
	g_debug("wiz: %d\n", __LINE__);
	cBuffer = (unsigned short*) gl_malloc(getWidth()*getHeight()*sizeof(unsigned short));
	g_debug("wiz: %d\n", __LINE__);

#else // SDL
	pRawFrameBuffer=NULL; // must lock. See UpdateDisplay function
	cBuffer = (unsigned short*) gl_malloc(cBufferWidth * cBufferHeight*sizeof(unsigned short));
#endif


#ifdef THREADED_UPDATE
	ThreadTempBuffer = (unsigned short*) gl_malloc(getWidth()*getHeight());
	BufferLock=false;
	ThreadID=0;
	CreateThread(NULL, 0, FlipThread, 0, 0, &ThreadID);
#endif

	#ifdef TINYGL
	glInitOnBuffer(getWidth(), getHeight(), cBuffer);
	glFlush();
	#endif
	
	#ifdef TRENKI3D
	zBuffer = (unsigned short*) gl_malloc(cBufferWidth * cBufferHeight*sizeof(unsigned short));
	tglInit(cBuffer, zBuffer, cBufferWidth);
	SetZBufferUsed();
	ClearZBuffer();
	#endif

	SetDrawBuffer();

	g_debug("Drawbuffer is %p\n", cBuffer);
}


void oScreen::Clear() // Clear the color buffer only (2D mode)
{
//	SetDrawBuffer();
#if defined(GP2X) && !defined(WANT_SDL)
	if(m_clear_color == 0)
	{
		gp2x_clear_back_buffer();
	}
	else
	{
		unsigned short nbPixels = (cBufferWidth * cBufferHeight);
		unsigned short* pixels   = (unsigned short*)cBuffer;
		while (nbPixels--)
			*pixels++ = m_clear_color;
	
	}
#else
	if(m_clear_color == 0)
	{
		long nbPixels = (cBufferWidth * cBufferHeight) >> 1; // not unsigned!
		long* pixels   = (long*)cBuffer;
		while (nbPixels--)
			*pixels++ = 0;
	}
	else
	{
		int    nbPixels = (cBufferWidth * cBufferHeight); // not unsigned!
		unsigned short* pixels   = (unsigned short*)cBuffer;
		while (nbPixels--)
			*pixels++ = m_clear_color;
	
	}
#endif
}

#ifdef TRENKI3D
void oScreen::ClearZBuffer()
{
	if(m_z_buffer_used)
	{
		long nbPixels = (cBufferWidth * cBufferHeight) >> 1;
		long* pixels   = (long*)zBuffer;
		while (nbPixels--)
			*pixels++ = 0x7f7f7f7f;
		m_z_buffer_used=0;
	}
}
#endif



void oScreen::DrawLine(int sx, int sy, int ex, int ey, unsigned short col)
{
	int	fbXincr, fbYincr, dPr, dPru, P;
	int	dX, dY;
	dX = _abs(ex-sx);
	dY = _abs(ey-sy);
	if(dX==0 && dY==0) {PutPixel(sx, sy, col); return;}
	fbXincr=(sx==ex) ? 0 : ((sx < ex) ? 1 : -1);
	fbYincr=(sy==ey) ? 0 : ((sy < ey) ? 1 : -1);

	if (dY < dX)
	{
		dPr = dY<<1;
		dPru = dPr-(dX<<1);
		P = dPr-dX;
	XLOOP:
		PutPixel(sx, sy, col);
		if (P > 0) goto RightAndUp;
		sx+=fbXincr;
		P+=dPr;
		dX--;
		if (dX>=0) goto XLOOP;
		return;
	RightAndUp:
		sx+=fbXincr;
		sy+=fbYincr;
		P+=dPru;
		dX--;
		if (dX>=0) goto XLOOP;
		return;
	}
	else
	{
		dPr = dX<<1;
		dPru = dPr-(dY<<1);
		P = dPr-dY;

	YLOOP:
		PutPixel(sx, sy, col);
		if (P > 0) goto RightAndUp2;
		sy+=fbYincr;
		P+=dPr;
		dY--;
		if (dY>=0) goto YLOOP;
		return;
	RightAndUp2:
		sx+=fbXincr;
		sy+=fbYincr;
		P+=dPru;
		dY--;
		if (dY>=0) goto YLOOP;
		return;
	}
}

#ifdef THREADED_UPDATE
unsigned short* threadbuffer=NULL;
#endif


void oScreen::SetDrawBuffer(unsigned short* pBuffer, int width, int height)
{
	if(!pBuffer || width<=0 || height<=0)
	{
		this->pDrawBuffer = this->cBuffer;
		this->db_width = this->cBufferWidth;
		this->db_height= this->cBufferHeight;
	}
	else
	{
		this->pDrawBuffer = pBuffer;
		this->db_width    = width;
		this->db_height   = height;
		Viewport(0,0,db_width, db_height);
	}
}


void oScreen::Show(void)
{
	SetDrawBuffer(); // restore cBuffer for enverything you do
#ifdef THREADED_UPDATE
unsigned short *swapBuffer;
	LockBuffer();
	if(threadbuffer)
		QCopy4(threadbuffer, Screen.ThreadTempBuffer, Screen.ScreenRange/2);
	UnlockBuffer();
#else
	UpdateDisplay(cBuffer, Landscape);
#endif

	#ifdef TRENKI3D
	tglSetBuffers(cBuffer, zBuffer, Screen.cBufferWidth);
	#endif

	Alphamode(0);
}


#ifdef THREADED_UPDATE

DWORD WINAPI FlipThread(LPVOID pParam)
{
	threadbuffer = new unsigned short[Screen.ScreenRange];
	while(true)
	{
		UpdateDisplay(threadbuffer, Screen.Landscape);	
		Sleep(4);
	}
	delete [] threadbuffer;
	threadbuffer=NULL;
return 0;
}

#endif // THREADED_UPDATE


#endif // #if defined(TINYGL) || defined(TRENKI3D)


