#if defined(TINYGL) || defined(TRENKI3D)

#ifndef GRAPHICS_H
#define GRAPHICS_H

// #define THREADED_UPDATE
// #define INTERLACED // fast but uh - that's ugly...


// - HEADERS ---------------------------------------------------
#include "gffilesystem.h"
#include "16bitcolor.h"

class oBitmap;

#ifdef _DEBUG
	#define VALID_SCREEN(ptr) {if((ptr)<Screen.getColorBuffer() || (ptr)>=Screen.getColorBuffer()+Screen.getWidth()*Screen.getHeight()) *(int*)0=42;}
	#define VALID_IMG(ptr)    {if((ptr)<m_Buffer || (ptr)>=m_Buffer+(m_Width*m_Height)) *(int*)0=42;}
#else
	#define VALID_SCREEN(ptr) {}
	#define VALID_IMG(ptr)    {}
#endif



// - OBJECTS ---------------------------------------------------
class oScreen
{
public:
	oScreen();
	~oScreen();
	void Create(HWND hwnd, int swidth, int sheight, bool landscape);
	void Close();

	inline unsigned short* GetPixelBuffer(int x, int y);
	#ifdef TRENKI3D
	inline	void	SetZBufferUsed(int used=1)			{m_z_buffer_used = used;}
			void	ClearZBuffer();
	#endif
	inline void PutPixel(int x, int y, unsigned short col); // Handle alpha blending
	inline unsigned short GetPixel(int x, int y); // Handle alpha blending
	void PutRect(int x, int y, int dx, int dy, unsigned short col);
	void GetScreenMetrics(int* width, int* height); // Take care of Landscaping
	void DrawLine(int sx, int sy, int ex, int ey, unsigned short col);
	void Alphamode(short mode); // -255 -> +255
	void AlphaTesting(bool enable)						{alpha_testing = enable;}
	static void Rotate90Degs(unsigned short* Buffer);
	
	#ifdef _WIN32_WCE
	bool ipaq3800;
	HPALETTE hPalettize;
	void createpalette( void );
	#endif

	inline int getHeight() {return db_height;}
	inline int getWidth()  {return db_width;}
	inline int getViewportHeight()	{return viewport_h;}
	inline int getViewportWidth()	{return viewport_w;}
	inline int getViewportX()		{return viewport_x;}
	inline int getViewportY()		{return viewport_y;}
	inline unsigned short* getColorBuffer();

	void (*m_BlendFunc)(unsigned short* pdest, unsigned short src);
	void SetClearColor(COLORREF clearcol) {m_clear_color = PackColor(clearcol); }
	void Clear();
	void  Show(void);

	void Viewport(int x, int y, int w, int h)
	{
		if (w<=0 || h<=0)
		{
			x=0; y=0;
			w = db_width;
			h = db_height;
		}
		else
		{
			if (x<0) x=0;
			if (y<0) y=0;
		}
		viewport_x = x;
		viewport_y = y;
		viewport_w = w;
		viewport_h = h;
		if (viewport_x+viewport_w > db_width) viewport_w = db_width-viewport_x;
		if (viewport_y+viewport_h > db_height)viewport_h = db_height-viewport_y;
	}

	// DATA
public:
	#if !defined(WANT_SDL) && defined(WIN32)
		// HDC     Buffer1; // The visible buffer. - get only temporarily
		HBITMAP Buffer2; // The 16 bit (X1R5G5B5 format) non-visible buffer where drawing occurs.
	#endif


#ifdef THREADED_UPDATE
	unsigned short  *ThreadTempBuffer;
	bool			BufferLock;
	DWORD			ThreadID;
	inline void LockBuffer()  {while( BufferLock){Sleep(1);} BufferLock=true;}
	inline void UnlockBuffer(){BufferLock=false;}
#endif

	void SetDrawBuffer(unsigned short* pBuffer=NULL, int width=0, int height=0); // Set a frame buffer object, or default to cBuffer


	int db_width, db_height;			// size of the current drawbuffer (offscreen - NOT screen!! cBufferWidth, cBufferHeight)
	int viewport_x, viewport_y, viewport_w, viewport_h; // current viewport rect

	#ifdef TRENKI3D
	unsigned short* zBuffer;			// zBuffer bits
	int				m_z_buffer_used;	// did you write to it?
	#endif
	HWND	Window;
	RECT	Rectangle;
	bool	Landscape; // Rotated by 90 degrees for a 320x240 view
	int		alpha_blending; // -255 -> +255. 0=off
	bool	alpha_testing;  // true/false if cookie taken care of
	unsigned long ScreenRange;	// width*height (number of pixels)
	int xpitch, ypitch, BPP;	// pocketpc bytes to draw x,y direction

	unsigned short*	pRawFrameBuffer; // if (NULL) use GAPI, pointer to real screen (SDL: NULL->must lock, GP2X: swapped with cBuffer, Wiz: /dev/fb0)
	unsigned short* cBuffer;	// back buffer memory
	int cBufferWidth, cBufferHeight; // real dimensions for the color buffer
	
	unsigned short* pDrawBuffer; // pointer to where to draw -> might be an "frame buffer object"
	unsigned short m_clear_color;
	
};

inline unsigned short* oScreen::getColorBuffer()
{
	return pDrawBuffer;
}

unsigned short* oScreen::GetPixelBuffer(int x, int y)
{
	x+=viewport_x;
	y+=viewport_y;
	if(x<0 || x>=db_width || y<0 || y>=db_height)
	{
		return NULL;
	}
	return (pDrawBuffer)+(x + y*db_width);
}


unsigned short oScreen::GetPixel(int x, int y)
{
static unsigned short* pBuffer;
	pBuffer = GetPixelBuffer(x, y);
	if(pBuffer) return *pBuffer;
return 0;
}


void oScreen::PutPixel(int x, int y, unsigned short col)
{
	unsigned short* pBuffer = GetPixelBuffer(x, y);
	if(pBuffer)
		m_BlendFunc(pBuffer, col);
}



// - EXTERNALS -------------------------------------------------
extern oScreen Screen;



// --------------------------------------------------------------- //
// Faster than memcpy
// --------------------------------------------------------------- //
#ifndef QCopy4
#define QCopy4(dst, src, ndwords) _QCopy4((unsigned long*)(dst), (const unsigned long*)(src), (unsigned int)ndwords)
#define QCopy2(dst, src, nwords) _QCopy2((unsigned short*)(dst), (const unsigned short*)(src), (unsigned int)nwords)
#endif

inline void _QCopy4(register unsigned long* a, register const unsigned long* b, register unsigned int ndwords) // must be 4 byte aligned
{
	for(; ndwords>0; ndwords--) *a++=*b++; // Makes 4 bytes in one!
}
inline void _QCopy2(register unsigned short* a, register const unsigned short* b, register unsigned int nshorts) // must be 2 byte aligned
{
	for(; nshorts>0; nshorts--) *a++=*b++; // Makes 2 bytes in one!
}

template<class T> inline void SwapMe(T& a, T& b)
{
register T c;
	c=a;
	a=b;
	b=c;
}



#endif

#endif // #if defined(TINYGL) || defined(TRENKI3D)




