//////////////////////////////////////////////////////////////////////////////
//
//  Sound extension for PocketFrog ver 0.93
//    Written by Thomas 'gersen' Maniero
//
//////////////////////////////////////////////////////////////////////////////


#ifndef _SOUND_H
#define _SOUND_H


#if !defined(_WIN32_WCE) && defined(WIN32_LEAN_AND_MEAN)
#include <mmsystem.h>
#pragma comment (lib, "winmm.lib")
#endif
#include <windows.h>


#include <mmreg.h>

#include "SoundData.h"
#include "SoundChannel.h"



#define MOD_MAX 2

#define SFX_MAX 8

#define MOD_MIXBUF_11025 298   //496
#define MOD_MIXBUF_22050 588   //984
#define MOD_MIXBUF_44100 1176  //1968

#define MOD_PPC_WHDR 5



class Sound {

private:
    HWAVEOUT m_hwaveout;
    WAVEFORMATEX m_waveformatx;
    WAVEHDR m_wavehdr[MOD_PPC_WHDR];

    HANDLE m_thread;
    DWORD m_threadId;
    CRITICAL_SECTION m_critSecRtp;
    HANDLE m_killEvent, m_exitEvent;

    bool m_open;

    unsigned long m_freqmix;

    // canali per MOD
    ModChannel m_mc[MOD_MAX];

    // canali per SFX
    SFXChannel m_sc[SFX_MAX];

    unsigned short *m_samplefreq;

    // buffer mix
    unsigned char *m_bufmix;
    // lunghezza di un frame (senza il doppio per lo stereo)
    unsigned long m_bufmixframe;
    // totale lunghezza bufmix
    unsigned long m_bufmixtotal;
    // puntatore al buffer precedente (per il double/triple... buffer)
    unsigned char *m_pbufmix;
    // indica se output in stereo o mono, uso unsigned long per facilitarmi i <<...
    unsigned long m_stereo;

    unsigned long m_globalVolume;   // global volume


private:
    static DWORD WINAPI update(LPVOID lpParameter);
    void updateModSFX();


public:
    Sound();
    ~Sound();

    bool open(unsigned long freqmix, bool stereo);
    bool close();

    void suspend();
    void resume();

    Channel *getChannel(unsigned long chl);

    unsigned long numChannelMod();
    unsigned long numChannelSFX();

    unsigned long playingChannelMod();
    unsigned long playingChannelSFX();


    unsigned long playSFX(SoundFX *sndfx);
    bool isPlayingSFX(unsigned long chl);
    void stopSFX(unsigned long chl);
    void pauseSFX();
    void resumeSFX();

    unsigned long playMod(Module *mod);
    void stopMod(unsigned long chl);
    void pauseMod();
    void resumeMod();

    void setVolume(unsigned long v);
    unsigned long getVolume();
};




#endif
