//////////////////////////////////////////////////////////////////////////////
//
//  Sound extension for PocketFrg ver 0.94
//    Written by Thomas 'gersen' Maniero
//
//////////////////////////////////////////////////////////////////////////////

#ifndef _SOUNDCHANNEL_H
#define _SOUNDCHANNEL_H

#include "soundData.h"

class Channel {

    friend class Sound;

protected:
    unsigned long m_chlvol, m_gblvol;
    unsigned long m_volume;
    bool m_loop;

    bool m_pause, m_stop;


protected:
    Channel();
    void play();

public:
    virtual void setVolume(unsigned long v);
    unsigned long getVolume();

    void setLoop(bool l);
    bool getLoop();

    void stop();
    void pause();
    void resume();

    bool playing();
};


class ModChannel : public Channel {

    friend class Sound;

private:
    // dati attuale modulo
    ModHeader *m_modheader;
    ModPattern *m_modpattern;
    // attuale pattern e note e numero nella sequenza pattern
    unsigned long m_pattern, m_note, m_seq, m_restart;
    // dati frequenze e tempo
    unsigned long m_bytextick, m_bytecnt;
    unsigned long m_tickxdivision, m_tickcnt;
    // dati per note e comandi canali
    unsigned long m_notefreq[MOD_MAXCHANNEL];
    unsigned long m_sample[MOD_MAXCHANNEL];
    unsigned long m_freqfactor[MOD_MAXCHANNEL];
    unsigned long m_clcvol[MOD_MAXCHANNEL];      // mi serve per i calcoli del vol slide
    unsigned long m_glbvol[MOD_MAXCHANNEL];      // gvolume contiene il vol modificato dal vol del m_mod
    unsigned long m_command[MOD_MAXCHANNEL];
    unsigned long m_opx[MOD_MAXCHANNEL];
    unsigned long m_opy[MOD_MAXCHANNEL];
    unsigned long m_opxy[MOD_MAXCHANNEL];
    unsigned long m_position[MOD_MAXCHANNEL];
    unsigned long m_pan[MOD_MAXCHANNEL];
    // dati per sample
    unsigned long m_slength[MOD_MAXSAMPLE];
    unsigned long m_sloopstart[MOD_MAXSAMPLE];
    unsigned long m_slooplength[MOD_MAXSAMPLE];
    unsigned long m_svol[MOD_MAXSAMPLE];
    unsigned long m_sfinetune[MOD_MAXSAMPLE];
    unsigned long m_sdata[MOD_MAXSAMPLE];
    // dati per effetti
    long m_voldelta[MOD_MAXCHANNEL];
    unsigned long m_slideupdown[MOD_MAXCHANNEL];
    unsigned long m_slidetonote[MOD_MAXCHANNEL];
    unsigned long m_slidetonotedelta[MOD_MAXCHANNEL];
    unsigned long m_arpeggioidx[MOD_MAXCHANNEL];
    unsigned long m_arpeggiofq[MOD_MAXCHANNEL];   // freqfactor note
    unsigned long m_arpeggiofqx[MOD_MAXCHANNEL];  // freqfactor note+x
    unsigned long m_arpeggiofqy[MOD_MAXCHANNEL];  // freqfactor note+y
    unsigned long m_vibratoidx[MOD_MAXCHANNEL];   // indice alla tabella
    unsigned long m_vibratospeed[MOD_MAXCHANNEL]; // adder all'indice
    unsigned long m_vibratodepth[MOD_MAXCHANNEL]; // profondit curva, da moltiplicare al valore tabella
    unsigned long m_delaynote[MOD_MAXCHANNEL];
    unsigned long m_delaynotefreq[MOD_MAXCHANNEL];
    unsigned long m_tickcutsample[MOD_MAXCHANNEL]; // per effetto 0xEC
    unsigned long m_jumpseq;
    unsigned long m_jumpnote;
    unsigned long m_delaypattern;


private:
    ModChannel();

public:
    virtual void setVolume(unsigned long v);
};


class SFXChannel : public Channel {

    friend class Sound;

private:
    unsigned char *m_sample;
    unsigned long m_freqfactor;
    unsigned long m_position;
    unsigned long m_length;


private:
    SFXChannel();
};



#endif
