//////////////////////////////////////////////////////////////////////////////
//
//  Sound extension for PocketFrog ver 0.94
//    Written by Thomas 'gersen' Maniero
//
//////////////////////////////////////////////////////////////////////////////
#include "Stdafx.h"

#include "soundData.h"
#include <stdio.h>
#include "GFFileSystem.h"

Module::Module() {

    m_mh = NULL;
    m_volume = 64;
    m_loop = false;

    m_delete = false;
}


Module::~Module() {

    if ((m_delete) && (m_mh)) {
        delete [] m_mh;
        m_mh = NULL;
        m_delete = false;
    }
}


void Module::setVolume(unsigned long v) {

    if (v > 64)
        v = 64;

    m_volume = v;
}


unsigned long Module::getVolume() {

    return m_volume;
}


void Module::setLoop(bool l) {

    m_loop = l;
}


bool Module::getLoop() {

    return m_loop;
}


bool Module::load(const char *modname)
{   unsigned long modsize;
    unsigned char *mod;

//    if (PocketPC::GetFullPathName(modname, pathname) == 0)
//        return false;
GFFileSystem f;
    if (!f.OpenIn(modname))
        return false;

	modsize = f.GetSize();

    if (modsize == 0) {
        return false;
    }

    mod = new unsigned char[modsize];
    if (mod == 0) {
        return false;
    }

    f.Read(mod, modsize);

    m_mh = mod;

    m_delete = true;

    return true;
}

bool Module::load(HINSTANCE /*hInstance*/, DWORD /*dwResourceID*/)
{
/*
    HRSRC hr; 
    HGLOBAL hg; 

    // trovo risorsa
    hr = FindResource(hInstance, MAKEINTRESOURCE(dwResourceID), TEXT("PTMOD")); 

    if (hr != NULL) { 

        // carico risorsa
        hg = LoadResource(hInstance, hr);

        if (hg != NULL) { 
            // lock resource
            m_mh = (unsigned char *)LockResource(hg); 
            if (m_mh != NULL) {
                m_delete = false;
                return true;
            }

        }

    }
*/

    return false;
}




SoundFX::SoundFX() {

    m_sample = NULL;
    m_volume = 64;
    m_loop = false;

    m_delete = false;
}


SoundFX::~SoundFX() {

    if ((m_delete) && (m_sample)) {
        delete [] m_sample;
        m_sample = NULL;
        m_delete = false;
    }
}



void SoundFX::setVolume(unsigned long v) {

    if (v > 64)
        v = 64;

    m_volume = v;
}


unsigned long SoundFX::getVolume() {

    return m_volume;
}


void SoundFX::setLoop(bool l) {

    m_loop = l;
}


bool SoundFX::getLoop() {

    return m_loop;
}


bool SoundFX::load(const char *wavname)
{
GFFileSystem f;

    if (!f.OpenIn(wavname))
        return false;

    if (!f.SeekTo(12, GFFileSystem::seek_start))
        return false;

    WaveChunk wck;
    WaveHeader whdr;

    // read chunk fmt   
    if ((f.Read(&wck, sizeof(WaveChunk)) != 1) || (wck.ckID != WAVE_CHUNK_FMT)) {
        return false;
    }

    // read header
    f.Read(&whdr, sizeof(WaveHeader));
//    if ((whdr.nBitsPerSample != 8) || (whdr.nChannels != 1)) return false;

    // adjust position...
    if (wck.nChunkSize > sizeof(WaveHeader))
        f.SeekTo(wck.nChunkSize - sizeof(WaveHeader), GFFileSystem::seek_current);

    // find data chunk
    f.Read(&wck, sizeof(WaveChunk));
    while (wck.ckID != WAVE_CHUNK_DATA) {
        f.SeekTo(wck.nChunkSize, GFFileSystem::seek_current);
        if (!f.Read(&wck, sizeof(WaveChunk))) {
            return false;
        }
    }
    m_freq = whdr.nSamplesPerSec;

    // Is ist the format we need? Quickly read it!
    if(whdr.nChannels==1 && whdr.nBitsPerSample==8)
    {
		m_length=wck.nChunkSize - 8;
		m_sample = new unsigned char[m_length];
		if(!m_sample) return false;
		f.Read(m_sample, m_length);
    }
    else // Uh-oh. Have to convert it!
    {
		// Read original data
		switch(whdr.nBitsPerSample)
		{
			case 16: m_length=(wck.nChunkSize - 8)/2; break;
			case  8: m_length=(wck.nChunkSize - 8);   break;
			default:
				return false;
		}
		m_sample = new unsigned char[m_length];
		if (!m_sample) return false;
		
		unsigned char* buffer = new unsigned char[m_length*whdr.nChannels*whdr.nBitsPerSample/8];
		f.Read(buffer, m_length*whdr.nChannels*whdr.nBitsPerSample/8);
		char* pCurCh  = (char*)buffer;
		short* pCurSh = (short*) buffer;		

		register int sample;
		if(whdr.nBitsPerSample==16)
		{
			for (register unsigned long n = 0; n<m_length; ++n)
			{
				int stereo=0;sample=0;
				do
				{
					sample += (*pCurSh>>8)+127; ++pCurSh;
				}while(++stereo < whdr.nChannels);
				sample /= whdr.nChannels;
				m_sample[n] = (unsigned char) sample;
			}
		}
		else
		{
			for (register unsigned long n = 0; n<m_length; ++n)
			{
				int stereo=0;sample=0;
				do
				{
					sample+= (*pCurCh++);
				}while(++stereo < whdr.nChannels);
				sample /= whdr.nChannels;
				m_sample[n] = (unsigned char) sample;
			}
		}
		delete[] buffer;
    }
    m_delete = true;
return true;
}

bool SoundFX::load(HINSTANCE /*hInstance*/, DWORD /*dwResourceID*/) {
/*
    HRSRC hr; 
    HGLOBAL hg; 

    // trovo risorsa
    hr = FindResource(hInstance, MAKEINTRESOURCE(dwResourceID), TEXT("WAVE")); 

    if (hr != NULL) { 

        // carico risorsa
        hg = LoadResource(hInstance, hr);

        if (hg != NULL) { 

            WaveChunk wck;
            WaveHeader whdr;
            unsigned char *pwav;

            // lock resource
            pwav = (unsigned char *)((unsigned long)LockResource(hg) + 12);
            if (pwav == NULL) 
                return false;

            // first chunk, must be fmt
            memcpy(&wck, pwav, sizeof(WaveChunk));
            if (wck.ckID != WAVE_CHUNK_FMT)
                return false;

            pwav += sizeof(WaveChunk);

            memcpy(&whdr, pwav, sizeof(WaveHeader));
            if ((whdr.nBitsPerSample != 8) || (whdr.nChannels != 1))
                return false;

            // next chunk
            pwav += wck.nChunkSize;

            // trovo il chunk data
            memcpy(&wck, pwav, sizeof(WaveChunk));
            while (wck.ckID != WAVE_CHUNK_DATA) {
                pwav += sizeof(WaveChunk) + wck.nChunkSize;
                memcpy(&wck, pwav, sizeof(WaveChunk));
            }


            m_freq = whdr.nSamplesPerSec;
            m_length = wck.nChunkSize - 8;

            m_sample = (unsigned char *)((unsigned long)pwav + sizeof(WaveChunk));

            m_delete = false;

            return true;
        }

    }
*/
return false;
}

