//////////////////////////////////////////////////////////////////////////////
//
//  Sound extension for PocketFrg ver 0.94
//    Written by Thomas 'gersen' Maniero
//
//////////////////////////////////////////////////////////////////////////////

#ifndef _SOUNDDATA_H
#define _SOUNDDATA_H


#include <windows.h>


#define MOD_MAXCHANNEL 4
#define MOD_MAXSAMPLE 32



// ATTENZIONE tutti i word vanno capovolti... l'amiga  una macchina big-endian


// struttura per i mod sample... 
// sono 15 o 31. se 31 dopo i primi 15 NON abbiamo la stringa di 4 byte "M.K."
struct struct_ModSample {
    char name[22];       // nome del sample
    unsigned short length;        // lunghezza in word (va quindi moltiplicato per 2) (* 2)
    unsigned char finetune;       // nota... lower 4 bits (-7..8) gestisco come unsigned...
    unsigned char volume;         // 0..64 (64? dovrebbe essere 63) -- vol = 20*log10(Vol/64)
    unsigned short loopstart;     // offset in word inizio del sample loop (* 2)
    unsigned short looplength;    // lunghezza in word del sample loop (loop se > 1) (* 2)
};
typedef struct struct_ModSample ModSample;


union union_ModNote {
    struct _AA{
        unsigned notefreq : 12;
        unsigned : 4;
        unsigned y : 4;
        unsigned x : 4;    
        unsigned command : 4;
        unsigned sample : 4;    // 0 = NULL, 1..31 valid
    };
    struct _BB{
        unsigned : 16;
        unsigned xy : 8;
        unsigned : 8;
    };
    unsigned long bits;
};
typedef union union_ModNote ModNote;


struct struct_ModPattern {
    ModNote notes[256];
    //ModNote notes[64][4];
};
typedef struct struct_ModPattern ModPattern;


// module header
struct struct_ModHeader {
    char name[20];           // nome del modulo
    ModSample samples[31]; // sample del modulo
    unsigned char numSequence;        // numero sequenze (?) (1..128)
    unsigned char restart;            // non pi usato (?)
    unsigned char sequences[128];     // sequenza (numSequence) suono pattern (ogni byte compreso 0..63)
    char mkSign[4];          // "M.K." signature

    // pattern data...
    // ModPattern *patterns; // array di pattern, contiene 1024 note divise per 64

    // sample data...

};
typedef struct struct_ModHeader ModHeader;



struct struct_WaveChunk {
    unsigned long ckID;            // chunk type s8[4] = u32
    unsigned long nChunkSize;      // lunghezza di questo chunk (non comprende ckID e nChunkSize (8 byte))
};
typedef struct struct_WaveChunk WaveChunk;


#define WAVE_CHUNK_FMT  0x20746D66
#define WAVE_CHUNK_DATA 0x61746164


struct struct_WaveHeader {
    unsigned short wFormatTag;      // formato del wave
    unsigned short nChannels;       // numero canali
    unsigned long nSamplesPerSec;  // sample rate (11025 - 22050 - 44100...)
    unsigned long nAvgBytesPerSec;
    unsigned short nBlockAlign;     
    unsigned short nBitsPerSample;  // bit campionamento... (8 - 16)
};
typedef struct struct_WaveHeader WaveHeader;




class Module {

    friend class Sound;

private:
    unsigned char *m_mh;
    unsigned long m_volume;
    bool m_loop;

    bool m_delete;

public:
    Module();
    ~Module();

    void setVolume(unsigned long v);
    unsigned long getVolume();

    void setLoop(bool l);
    bool getLoop();

    bool load(const char *modname);
    bool load(HINSTANCE hInstance, DWORD dwResourceID);
};


// struttura per gli effetti sonori
class SoundFX {   

    friend class Sound;

private:
    unsigned char *m_sample;        // puntatore ai dati

    unsigned long m_freq;          // frequenza campionamento
    unsigned long m_length;        // lunghezza in byte del suono
    unsigned long m_left;          // 1 - canale sinistro, 0 - canale destro
    unsigned long m_volume;        // volume suono
    bool m_loop;

    bool m_delete;

public:
    SoundFX();
    ~SoundFX();

    void setVolume(unsigned long v);
    unsigned long getVolume();

    void setLoop(bool l);
    bool getLoop();
    bool IsLoaded() {return m_sample ? true:false;}

    bool load(const char *wavname);
    bool load(HINSTANCE hInstance, DWORD dwResourceID);
};



#endif
