#include "oSound.h"

oSound gSound;

#define SIMPLE_SOUND // use PlaySound function...



#ifndef SIMPLE_SOUND
void CALLBACK __WaveOutProc(HWAVEOUT hwo, UINT uMsg, DWORD /*dwInstance*/, DWORD dwParam1, DWORD /*dwParam2*/)
{

	switch(uMsg)
	{
	case WOM_OPEN: // B
		break;
	case WOM_DONE: // C
		{
			WAVEHDR *whdr = (WAVEHDR*)dwParam1;
			if(!whdr || !hwo) return;
			if(whdr->dwUser) // loop
				waveOutWrite(hwo, whdr, sizeof(WAVEHDR));
			else
			{
				waveOutUnprepareHeader(hwo, whdr, sizeof(WAVEHDR));
				//waveOutPause(hwo);
				//waveOutReset(hwo);
				waveOutClose(hwo);
				delete whdr;
			}
		}
		break;
	case WOM_CLOSE: // D
		break;
	}
}
#endif

bool oWaveFile::Load(const char* szFileName)
{
	Clear();
GFFileSystem f; // Structure to load the file into memory

    // Open and read the wave file
    if(!f.OpenIn(szFileName)) return false;

    // Determine the file size
	file_size = f.GetSize();

    // Request enough memory to store the entire file
    lpfile = new char [file_size];
    // 'Copy' the file to memory
	f.Read(lpfile, file_size);
	f.Close();

	// Fill WAVEFORMATEX with the data from the file
    wfex.wFormatTag         = *((WORD* )(lpfile + OFFSET_FORMATTAG     ));
    wfex.nChannels          = *((WORD* )(lpfile + OFFSET_CHANNELS      ));
    wfex.nSamplesPerSec     = *((DWORD*)(lpfile + OFFSET_SAMPLESPERSEC ));
    wfex.nAvgBytesPerSec    = *((DWORD*)(lpfile + OFFSET_AVGBYTESPERSEC));
    wfex.nBlockAlign        = *((WORD* )(lpfile + OFFSET_BLOCKALIGN    ));
    wfex.wBitsPerSample     = *((WORD* )(lpfile + OFFSET_BITSPERSAMPLE ));
return true;
}

oWaveFile::~oWaveFile()
{
	Clear();
}

void oWaveFile::Clear()
{
#ifndef SIMPLE_SOUND
	if (hwo)
	{
		waveOutClose(hwo);
	}
#endif
	hwo=NULL;
	if (lpfile) delete[]lpfile; lpfile=NULL;
	memset(&wfex,0, sizeof(wfex)); file_size=0;
}


WAVEHDR* oWaveFile::Play(bool loop)
{
#ifdef SIMPLE_SOUND
	if(file_size)
		PlaySound(lpfile, NULL, SND_MEMORY | SND_ASYNC);
	return NULL;
#else
	WAVEHDR *whdr = NULL;
	whdr = new WAVEHDR();
	if(!whdr) return NULL;
	whdr->lpData = lpfile+HEADER_SIZE;
	whdr->dwBufferLength = file_size-HEADER_SIZE;
	whdr->dwUser = (DWORD)loop;

	// Find a waveOut device and open it
	for(UINT devid = 0; devid < waveOutGetNumDevs(); devid++)
	{
		if(devid == waveOutGetNumDevs())
			// Error, no free devices found
			return NULL;
		if(waveOutOpen(&hwo, WAVE_MAPPER, &wfex, (DWORD)__WaveOutProc, 0, CALLBACK_FUNCTION) == MMSYSERR_NOERROR)
			// Usable device found, stop searching
			break;
	}

	if(waveOutPrepareHeader(hwo, whdr, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
		return NULL;
	if(waveOutWrite(hwo, whdr, sizeof(WAVEHDR)) != MMSYSERR_NOERROR)
		return NULL;

	// Return the pointer to the WAVEHDR we used here
	return whdr;
#endif
}

void oWaveFile::Stop(WAVEHDR *whdr) // Only for looping sounds - the others alread might have been deleted!
{
	if(whdr)
		(whdr)->dwUser = (DWORD)false;
}
