#ifndef OSOUND_H_INC
#define OSOUND_H_INC

#ifndef _WIN32_WCE
#pragma comment(lib, "winmm.lib")
#endif

/* MultipleWaves.cpp
 * Multiple wave files simultaneously playback
 * By Fabricio Kury, April 12, 2003
 * mailto:justt@mail.ru
 * http://www.cellosoft.com/
 */

#include <windows.h>
#include "GFQueue.h"
#include "GFFileSystem.h"

#define OFFSET_FORMATTAG			20
#define OFFSET_CHANNELS				22
#define OFFSET_SAMPLESPERSEC		24
#define OFFSET_AVGBYTESPERSEC		28
#define OFFSET_BLOCKALIGN			32
#define OFFSET_BITSPERSAMPLE		34
#define OFFSET_WAVEDATA				44
#define HEADER_SIZE	                OFFSET_WAVEDATA

class oWaveFile
{
public:
	oWaveFile()	{lpfile=NULL; hwo=NULL; memset(&wfex,0, sizeof(wfex)); file_size=0;}
	~oWaveFile();
	bool Load(const char* szFileName);
	WAVEHDR* Play(bool loop=false);
	void Stop(WAVEHDR *whdr);
	void Clear();
private:
	char *lpfile;
	int file_size;
	WAVEFORMATEX	wfex;
	HWAVEOUT		hwo;
};


class oSound
{
public:
	~oSound()	{Hush();}
	bool Load(long n, const char* fn)
	{
		oWaveFile wf;
		WAVEHDR*  wh=NULL;
		if (n<0) return false;
		while (n>=(long)wavs.Size())
		{
			wavs.Push(wf);
			playing.Push(wh);
		}
		return wavs.GetAt(n)->Load(fn);
	}
	void Play(long n, bool loop=false)
	{
		if (n>=0 && n<(long)wavs.Size())
		{
			if (wavs.GetAt(n) && playing.GetAt(n) && *playing.GetAt(n))
			{
				wavs.GetAt(n)->Stop(*playing.GetAt(n));
				*playing.GetAt(n) = NULL;
			}
			if (wavs.GetAt(n))
			{
				*playing.GetAt(n) = wavs.GetAt(n)->Play(loop);
				if(!loop) *playing.GetAt(n) = NULL; // Don't have to stop non-looping sounds
			}
		}
	}
	void Hush()
	{
		for (long n=0; n<(long)wavs.Size(); n++)
		{
			if (wavs.GetAt(n) && playing.GetAt(n) && *playing.GetAt(n))
			{
				wavs.GetAt(n)->Stop(*playing.GetAt(n));
				*playing.GetAt(n) = NULL;
			}
		}
	}
private:
	Queue<oWaveFile>	wavs;
	Queue<WAVEHDR*>		playing;
};

extern oSound gSound;

#endif // OSOUND_H_INC

