#ifndef __POLLUXREGS_H__
#define __POLLUXREGS_H__

extern volatile uint32_t *memregs32;
extern volatile uint16_t *memregs16;
extern volatile uint8_t  *memregs8;
extern volatile uint16_t *uppermem;
#define BIT(number) (1<<(number))

/* CPU control */

#define PLLSETREG0		(memregs32[0xF004>>2])
#define PWRMODE			(memregs32[0xF07C>>2])


/* MLC base registers */

/* MLC TOP CONTROL REGISTER, pg 256 */
typedef struct {
	bool 	 FIELDENB:1;	/* bit 0 */
	bool	 MLCENB:1;
	bool	 :1;
	bool	 DITTYFLAG:1;
	uint32_t :4;
	uint32_t PRIORITY:2;
	bool	 PIXELBUFFER_SLD:1;
	bool	 PIXELBUFFER_PWD:1;
	uint32_t :2;
	uint32_t :18;
} st_MLCCONTROLT;
#define	REG_MLCCONTROLT	((volatile st_MLCCONTROLT*)&memregs32[lc_abstorel(0xC0004000,32)])

/* MLC SCREEN SIZE REGISTER, pg 256 */
typedef struct {
	uint32_t SCREENWIDTH:12;
	uint32_t :4;
	uint32_t SCREENHEIGHT:12;
	uint32_t :4;
} st_MLCSCREENSIZE;
#define	REG_MLCSCREENSIZE	((volatile st_MLCSCREENSIZE*)&memregs32[lc_abstorel(0xC0004004,32)])

/* MLC BACKGROUND COLOR REGISTER, pg 256*/
typedef struct {
	uint32_t DEFAULTCOLOR:24;
	uint32_t :8;
} st_MLCBGCOLOR;
#define	REG_MLCBGCOLOR	((volatile st_MLCBGCOLOR*)&memregs32[lc_abstorel(0xC0004008,32)])

/***************************************************************************************************************/

/* RGB layer 0 registers */

/* MLC RGB LAYER 0 LEFT RIGHT REGISTER, pg 256 */
typedef struct {
	uint32_t RIGHT:12;
	uint32_t :4;
	uint32_t LEFT:12;
	uint32_t :4;
} st_MLCLEFTRIGHT0;
#define	REG_MLCLEFTRIGHT0	((volatile st_MLCLEFTRIGHT0*)&memregs32[lc_abstorel(0xC000400C,32)])

/* MLC RGB LAYER 0 TOP BOTTOM REGISTER, pg 256 */
typedef struct {
	uint32_t BOTTOM:12;
	uint32_t :4;
	uint32_t TOP:12;
	uint32_t :4;
} st_MLCTOPBOTTOM0;
#define	REG_MLCTOPBOTTOM0	((volatile st_MLCTOPBOTTOM0*)&memregs32[lc_abstorel(0xC0004010,32)])

/* MLC RGB LAYER 0 INVALID AREA 0 LEFT RIGHT REGISTER, pg 257 */
typedef struct {
	uint32_t RIGHT:11;
	uint32_t :5;
	uint32_t LEFT:11;
	bool	 :1;
	bool	 INVALIDENB0:1;
	uint32_t :3;
} st_MLCLEFTRIGHT0_0;
#define REG_MLCLEFTRIGHT0_0	((volatile st_MLCLEFTRIGHT0_0*)&memregs32[lc_abstorel(0xC0004014,32)])

/* MLC RGB LAYER 0 INVALID AREA 0 BOTTOM TOP REGISTER, pg 257 */
typedef struct {
	uint32_t BOTTOM:12;
	uint32_t :4;
	uint32_t TOP:12;
	uint32_t :4;
} st_MLCTOPBOTTOM0_0;
#define	REG_MLCTOPBOTTOM0_0	((volatile st_MLCTOPBOTTOM0_0*)&memregs32[lc_abstorel(0xC0004018,32)])

/* MLC RGB LAYER 0 INVALID AREA 1 LEFT RIGHT REGISTER, pg 257 */
typedef struct {
	uint32_t RIGHT:11;
	uint32_t :5;
	uint32_t LEFT:11;
	bool	 :1;
	bool	 INVALIDENB1:1;
	uint32_t :3;
} st_MLCLEFTRIGHT0_1;
#define REG_MLCLEFTRIGHT0_1	((volatile st_MLCLEFTRIGHT0_1*)&memregs32[lc_abstorel(0xC000401C,32)])

/* MLC RGB LAYER 0 INVALID AREA 1 BOTTOM TOP REGISTER, pg 257 */
typedef struct {
	uint32_t BOTTOM:12;
	uint32_t :4;
	uint32_t TOP:12;
	uint32_t :4;
} st_MLCTOPBOTTOM0_1;
#define	REG_MLCTOPBOTTOM0_1	((volatile st_MLCTOPBOTTOM0_1*)&memregs32[lc_abstorel(0xC0004020,32)])

/* MLC RGB LAYER 0 CONTROL REGISTER, pg 258 */
typedef struct {
	bool	 TPENB:1; /* transparency */
	bool	 INVENB:1; /* colour inversion */
	bool	 BLENDENB:1; /* alpha */
	bool	 :1;
	bool	 DIRTYFLAG:1; /* use changes */
	bool	 LAYERENB:1; /* enable layer */
	uint32_t :2;
	bool	 :1;
	uint32_t :3;
	uint32_t LOCKSIZE:2; /* amount of data to be read at once */
	bool	 PALETTESLD:1; /* palette table sleep mode */
	bool	 PALETTEPWD:1; /* palette table power on/off */
	uint32_t FORMAT:16; /* colour format */
} st_MLCCONTROL0;
#define	REG_MLCCONTROL0	((volatile st_MLCCONTROL0*)&memregs32[lc_abstorel(0xC0004024,32)])

/* MLC RGB LAYER 0 HORIZONTAL STRIDE REGISTER, pg 258 */
typedef struct {
	uint32_t HSTRIDE:31;
	bool	 :1;
} st_MLCHSTRIDE0;
#define	REG_MLCHSTRIDE0	((volatile st_MLCHSTRIDE0*)&memregs32[lc_abstorel(0xC0004028,32)])

/* MLC RGB LAYER 0 VERTICAL STRIDE REGISTER, pg 258 */
typedef struct {
	uint32_t VSTRIDE:32;
} st_MLCVSTRIDE0;
#define	REG_MLCVSTRIDE0	((volatile st_MLCVSTRIDE0*)&memregs32[lc_abstorel(0xC000402C,32)])

/* MLC RGB LAYER 0 TRANSPARENCY COLOR REGISTER, pg 258 */
typedef struct {
	uint32_t TPCOLOR:24; /* transparency colour */
	uint32_t :4;
	uint32_t ALPHA:4; /* alpha value */
} st_MLCTPCOLOR0;
#define	REG_MLCTPCOLOR0	((volatile st_MLCTPCOLOR0*)&memregs32[lc_abstorel(0xC0004030,32)])

/* MLC RGB LAYER 0 INVERSION COLOR REGISTER, pg 258 */
typedef struct {
	uint32_t INVCOLOR:24; /* inversion colour */
	uint32_t :8;
} st_MLCINVCOLOR0;
#define	REG_MLCINVCOLOR0	((volatile st_MLCINVCOLOR0*)&memregs32[lc_abstorel(0xC0004034,32)])

/* MLC RGB LAYER 0 BASE ADDERSS REGISTER, pg 259 */
typedef struct {
	uint32_t ADDRESS:32;
} st_MLCADDRESS0;
#define	REG_MLCADDRESS0	((volatile st_MLCADDRESS0*)&memregs32[lc_abstorel(0xC0004038,32)])

/* MLC RGB LAYER 0 PALETTE ADDRESS DATA, pg 260 */
typedef struct {
	uint32_t PALETTEDATA:16;
	uint32_t :8;
	uint32_t PALETTEADDR:8;
} st_MLCPALETTE0;
#define	REG_MLCPALETTE0	((volatile st_MLCPALETTE0*)&memregs32[lc_abstorel(0xC000403C,32)])

/***************************************************************************************************************/

/* RGB layer 1 registers */

#define	REG_MLCLEFTRIGHT1	((volatile st_MLCLEFTRIGHT0*)&memregs32[lc_abstorel(0xC0004040,32)])
#define	REG_MLCTOPBOTTOM1	((volatile st_MLCTOPBOTTOM0*)&memregs32[lc_abstorel(0xC0004044,32)])
#define REG_MLCLEFTRIGHT1_0	((volatile st_MLCLEFTRIGHT0_0*)&memregs32[lc_abstorel(0xC0004048,32)])
#define	REG_MLCTOPBOTTOM1_0	((volatile st_MLCTOPBOTTOM0_0*)&memregs32[lc_abstorel(0xC000404C,32)])
#define REG_MLCLEFTRIGHT1_1	((volatile st_MLCLEFTRIGHT0_1*)&memregs32[lc_abstorel(0xC0004050,32)])
#define	REG_MLCTOPBOTTOM1_1	((volatile st_MLCTOPBOTTOM0_1*)&memregs32[lc_abstorel(0xC0004054,32)])
#define	REG_MLCCONTROL1		((volatile st_MLCCONTROL0*)&memregs32[lc_abstorel(0xC0004058,32)])
#define	REG_MLCHSTRIDE1		((volatile st_MLCHSTRIDE0*)&memregs32[lc_abstorel(0xC000405C,32)])
#define	REG_MLCVSTRIDE1		((volatile st_MLCVSTRIDE0*)&memregs32[lc_abstorel(0xC0004060,32)])
#define	REG_MLCTPCOLOR1		((volatile st_MLCTPCOLOR0*)&memregs32[lc_abstorel(0xC0004064,32)])
#define	REG_MLCINVCOLOR1	((volatile st_MLCINVCOLOR0*)&memregs32[lc_abstorel(0xC0004068,32)])
#define	REG_MLCADDRESS1		((volatile st_MLCADDRESS0*)&memregs32[lc_abstorel(0xC000406C,32)])
#define	REG_MLCPALETTE1		((volatile st_MLCPALETTE0*)&memregs32[lc_abstorel(0xC0004070,32)])

/* MLC */
#define	MLCCONTROLT		(memregs32[0x4000>>2])
#define	MLCSCREENSIZE		(memregs32[0x4004>>2])
#define	MLCBGCOLOR		(memregs32[0x4008>>2])

#define	MLCLEFTRIGHT0		(memregs32[0x400C>>2])
#define	MLCTOPBOTTOM0		(memregs32[0x4010>>2])
#define MLCLEFTRIGHT0_0		(memregs32[0x4014>>2])
#define	MLCTOPBOTTOM0_0		(memregs32[0x4018>>2])
#define MLCLEFTRIGHT0_1		(memregs32[0x401C>>2])
#define	MLCTOPBOTTOM0_1		(memregs32[0x4020>>2])
#define	MLCCONTROL0		(memregs32[0x4024>>2])
#define	MLCHSTRIDE0		(memregs32[0x4028>>2])
#define	MLCVSTRIDE0		(memregs32[0x402C>>2])
#define	MLCTPCOLOR0		(memregs32[0x4030>>2])
#define	MLCINVCOLOR0		(memregs32[0x4034>>2])
#define	MLCADDRESS0		(memregs32[0x4038>>2])
#define	MLCPALETTE0		(memregs32[0x403C>>2])

#define	MLCLEFTRIGHT1		(memregs32[0x4040>>2])
#define	MLCTOPBOTTOM1		(memregs32[0x4044>>2])
#define MLCLEFTRIGHT1_0		(memregs32[0x4048>>2])
#define	MLCTOPBOTTOM1_0		(memregs32[0x404C>>2])
#define MLCLEFTRIGHT1_1		(memregs32[0x4050>>2])
#define	MLCTOPBOTTOM1_1		(memregs32[0x4054>>2])
#define	MLCCONTROL1		(memregs32[0x4058>>2])
#define	MLCHSTRIDE1		(memregs32[0x405C>>2])
#define	MLCVSTRIDE1		(memregs32[0x4060>>2])
#define	MLCTPCOLOR1		(memregs32[0x4064>>2])
#define	MLCINVCOLOR1		(memregs32[0x4068>>2])
#define	MLCADDRESS1		(memregs32[0x406C>>2])
#define	MLCPALETTE1		(memregs32[0x4070>>2])

/* Graphics modes */
enum {
	RGB565		= 0x4432,
	BGR565		= 0xC432,
	XRGB1555	= 0x4342,
	XBGR1555	= 0xC342,
	XRGB4444	= 0x4211,
	XBGR4444	= 0xC211,
	XRGB8332	= 0x4120,
	XBGR8332	= 0xC120,
	ARGB1555	= 0x3342,
	ABGR1555	= 0xB342,
	ARGB4444	= 0x2211,
	ABGR4444	= 0xA211,
	ARGB8332	= 0x1120,
	ABGR8332	= 0x9120,
	RGB888		= 0x4653,
	BGR888		= 0xC653,
	XRGB8888	= 0x4653,
	XBGR8888	= 0xC653,
	ARGB8888	= 0x0653,
	ABGR8888	= 0x8653,
	PTRGB565	= 0x443A
};

/* Display controller */
#define DPCHTOTAL		(memregs16[0x307C>>1])
#define DPCHAEND		(memregs16[0x3082>>1])
#define DPCVTOTAL		(memregs16[0x3084>>1])
#define DPCVASTART		(memregs16[0x3088>>1])
#define DPCVAEND		(memregs16[0x308A>>1])
#define DPCCTRL0		(memregs16[0x308C>>1])
#define DPCCTRL1		(memregs16[0x308E>>1])
#define DITHER_NONE	0
#define DITHER_4BIT	1
#define DITHER_5BIT	2
#define DITHER_6BIT	3

/* GPIO */
#define GPIOAPAD		(memregs16[0xA018>>1])
#define GPIOBPAD		(memregs16[0xA058>>1])
#define GPIOCPAD		(memregs16[0xA098>>1])

#endif
