REM -------------------------------------------------------------------------------------------
REM Must set these environment variables before calling this sccript:

REM PLATFORM     = folder within GLBasic\Compiler\platform = extension of libs

REM GCC_PREFIX   = prefix of g++.exe (path must be in "%COMPILER%\bin" )

REM GCC_ARGS     = args and more includes and defines

REM SPECIAL_CPP  = optional: gcc input files for this platform (relative pathes using %COMPILER%)

REM SPECIAL_O    = optional: gcc linker object of SPECIAL_CPP files (only filenames, space separated)

REM GLB_CORE     = optional: WINCE    - Core for WindowsCE (first GLBasic port - a true mess :/)

REM GLB_GFX_CORE = optional: TRENKI3D - Software rasterizer library
REM                          OPENGL   - OpenGL (usually version 1.1 with shaders by glGetProcAddress)
REM                          OPENGLES - OpenGL|ES mode (version 1.1, no shaders)

REM LIB_POSTFIX  = optional: postfix characters for final lib name (console, trenki...)

REM -------------------------------------------------------------------------------------------


IF "%PLATFORM%"=="" (
CLS
ECHO Must call this with special parameters.
PAUSE
GOTO :EOF
)


SET COMPILER=..\%PLATFORM%
SET PATH=%COMPILER%\bin;%COMPILER%;%SystemRoot%\System32

REM Some compilers (Android) fail on long pathes here.
SET TMPDIR=C:\Temp
IF NOT EXIST "%TMPDIR%" (
	MKDIR "%TMPDIR%"
)

REM -------------------------------------------------------------------------------------------
REM -------------------------------------------------------------------------------------------
ECHO ..libGLBasic Platform:[%PLATFORM%] Core:[%GLB_CORE%] GFX-Core:[%GLB_GFX_CORE%] Lib:[%LIB_POSTFIX%]
REM CD to "GLBasic\Compiler\platform\Source"
CD /D "%~dp0"

SET GCC_CMD=%GCC_PREFIX%g++ -pipe -c -O3 -funroll-loops -w %GCC_ARGS% -DNDEBUG -I"%COMPILER%/include" -I./ -I./glbasic -I../Include -isystem./SDL-1.2.11/include -I./trenki3D -I./WinCE 

REM GLBasic engine components for all platforms
SET CPP_FILES=glbasic/glb-full.cpp glbasic/FBuffers.cpp glbasic/glb3D.cpp glbasic/OpenGLRainbows.cpp glbasic/Rainbows3D.cpp glbasic/3DMath.cpp glbasic/MeshObject.cpp glbasic/GFCompression.cpp glbasic/pathfinding/astarplanner.cpp glbasic/pathfinding/node.cpp glbasic/pathfinding/nodehashtable.cpp glbasic/pathfinding/nodehashtable-iterator.cpp glbasic/pathfinding/planner.cpp glbasic/network/hthread.cpp glbasic/network/htmutex.cpp glbasic/network/htcondition.cpp glbasic/gf_sockets.cpp glbasic/blowfish.cpp
SET OBJECTS=glb-full.o fbuffers.o astarplanner.o node.o nodehashtable.o nodehashtable-iterator.o planner.o Rainbows3D.o glb3D.o OpenGLRainbows.o 3DMath.o GFCompression.o MeshObject.o hthread.o htmutex.o htcondition.o gf_sockets.o blowfish.o

REM Special components by "%GLB_CORE%"
if "%GLB_CORE%" == "WINCE" (
REM *** WINDOWS-CE CORE (glb_ppc core, init, sound and input is totally different) ***
	SET CPP_FILES=%CPP_FILES% WinCE\glb_ppc.cpp WinCE\ceframework.cpp WinCE\sound.cpp WinCE\soundchannel.cpp WinCE\sounddata.cpp WinCE\okeyboard.cpp WinCE\omouse.cpp WinCE\obmp.cpp
	SET OBJECTS=%OBJECTS% glb_ppc.o ceframework.o sound.o soundchannel.o sounddata.o okeyboard.o omouse.o obmp.o
) else (
REM *** NORMAL CORE (glb_pc core: initialization, sound and input) ***
	SET CPP_FILES=%CPP_FILES% glbasic/glb_pc.cpp glbasic/dxguid.cpp glbasic/DXInput.cpp glbasic/DXSound.cpp
	SET OBJECTS=%OBJECTS% glb_pc.o dxguid.o DXInput.o DXSound.o
)

REM Software rasterizer "Trenki3D" code
if "%GLB_GFX_CORE%" == "TRENKI3D" (
	SET CPP_FILES=%CPP_FILES% WinCE\obitmap.cpp WinCE\ofont.cpp WinCE\oscreen.cpp trenki3D\trenki3d.cpp trenki3D\renderer\geometry_processor.cpp trenki3D\fixedpoint\fixed_func.cpp
	SET OBJECTS=%OBJECTS% obitmap.o ofont.o oscreen.o trenki3d.o geometry_processor.o fixed_func.o
)

REM some CPP files especially for this platfrom?
if NOT "%SPECIAL_CPP%" == "" (
	SET CPP_FILES=%CPP_FILES% %SPECIAL_CPP%
	SET OBJECTS=%OBJECTS% %SPECIAL_O%
)

REM *** Compile the GLBasic core engine components ***
%GCC_CMD% %CPP_FILES%

REM append: "2>&1|more" to have the output in pages.

if "%LIB_EXTENSION%" == "" (
	SET LIB_EXTENSION=a
)

REM and build a library from the object files
SET LIBPATH=%COMPILER%\lib\libGLBasic%PLATFORM%%LIB_POSTFIX%.%LIB_EXTENSION%
DEL "%LIBPATH%" /Q >nul 2>&1
REM Android needs S -> no warning "internal ranlib" -> manual ranlib req.
"%GCC_PREFIX%ar" crlS "%LIBPATH%" %OBJECTS%

"%GCC_PREFIX%ranlib" "%LIBPATH%"
DEL %OBJECTS% /Q > NUL

