#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <setjmp.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include <sys/wait.h>
#include <sys/poll.h>
#include <signal.h>
#include <errno.h>

#include "te9_tf9_hybrid_driver.h"
#include "te9_tf9_main.h"
#include "te9_tf9_regs.h"

int accel_fd;
struct sigaction sigact, oldact;

char str[64];
int exitIrqChecker = 0;

/*
//-------------------------------------------------------------------------------------------------
// Test Function Prototypes.... 
//-------------------------------------------------------------------------------------------------
void AccelTest_Init(void);
void AccelTest_deinit(void);
void AccelTest_Reset(void);
void AccelTest_Dump_Registers(void);
void AccelTest_XYZ_LPF_cnt_g(void);
void AccelTest_XYZ_HPF_cnt_g(void);
void AccelTest_Alg_MotionDetection(void);
void AccelTest_Alg_Rotation(void);
void AccelTest_Alg_DirectionalTap(void);
void integrated_test(void); 			//added by lars-20100308
void directional_tap(void);
void screen_rotation(void);
void motion_detect(void);
void shake_detect(void);
*/

void * func_accel_test[][2]=
{	
	(void *)AccelTest_Init, 				"ACCEL Init",
	(void *)AccelTest_deinit,				"ACCEL deinit",
	(void *)AccelTest_Reset,				"ACCEL Reset",
	(void *)AccelTest_Dump_Registers,		"Display/Dump Register Value",
	(void *)AccelTest_XYZ_LPF_cnt_g,		"Read XYZ Count/g (LPF)",
	(void *)AccelTest_XYZ_HPF_cnt_g,		"Read XYZ Count/g (HPF) - Not available, sorry~! ^^",
	(void *)AccelTest_Alg_MotionDetection,	"Algorithm : Motion Detection",
	(void *)AccelTest_Alg_Rotation,			"Algorithm : Screen Rotation",
	(void *)AccelTest_Alg_DirectionalTap,		"Algorithm : Directional-Tap \
		                                        \n   ----- LARS added below functions -----",
	(void *)integrated_test,					"Integrated Test(Rotation, Tap, XYZ)",
	(void *)directional_tap,					"Directional Tap",
	(void *)screen_rotation,					"Screen Rotation",
	(void *)motion_detect,						"Motion Detection",
	(void *)shake_detect,						"Shake Detection(Use Motion Detection)",
	0,0
};


/* Signal Handler... */
static void irqChecker_sigHandler(int signo)
{
	switch(signo)
	{
	case SIGIO:
//		KxPrint("-I- SIGIO \n");
		KIONIX_ACCEL_service_interrupt();
		break;
	case SIGINT: case SIGQUIT:
		KxPrint("-I- SIGINT or SIGQUIT \n");
		exitIrqChecker = 1;
		break;
	}
	return;
}

/* Device File Initialization... */
int init_accel_device_file(void)
{
	accel_fd = open("/dev/accel", O_RDWR);		//blocking mode open... 
	if (accel_fd == -1){
		printf("-E- Error opening accel adapter\n");
		return -1;
	}
	
	return 0;
}

/* Sig Action Function Initialization... */
void init_sigact(void)
{
	int oflag;

	sigact.sa_handler = irqChecker_sigHandler;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = SA_INTERRUPT;
	if(sigaction(SIGIO, &sigact, &oldact) < 0)
	{
		KxPrint("-W- can't receive SIGIO event\n");
		if(accel_fd) close(accel_fd);
	}
	fcntl(accel_fd, F_SETOWN, getpid());
	oflag = fcntl(accel_fd, F_GETFL);
	fcntl(accel_fd, F_SETFL, oflag | FASYNC);
}

int main(void)
{
	int i, sel, retval;
	pid_t pid;
	unsigned long data;
	
	if(init_accel_device_file())
		return -1;

	init_sigact();

	while(1)
	{
		i=0; sel=-1;
		KxPrint("+--------------[ Accelerometer Test ]-----------------+\n");
		KxPrint("|  Kionix Driver Version   : "KIONIX_DRV_VERSION"\n");
		KxPrint("|  \n");

		while(1)
		{   //display menu
			KxPrint("| %2d:%s\n",i+1, func_accel_test[i][1]);
			i++;
			if((int)(func_accel_test[i][0])==0)	break;
		}
		KxPrint("| %2d:Previous menu\n", i+1);
		KxPrint("+-------------------------------------------------+\n");
		KxPrint(" Select the number to test : ");
		scanf("%d", &sel);
		fflush(stdin);
		sel--;
		KxPrint("+-------------------------------------------------+\n\n\n");
		if(sel == i){
			return;
		}if(sel>=0 && (sel<((sizeof(func_accel_test)-1)/8)) ){
			( (void (*)(void)) (func_accel_test[sel][0]) )();
			KxPrint("\n\nDone.\n");
		}else{
			KxPrint("Wrong number seleted.. Try again!!\n\n\n");
		}
	}
	AccelTest_deinit();

	if(accel_fd) close(accel_fd);

	exit(0);
}

void motion_detect(void)		// 단말기의 움직이는 정도...? 
{
	int device_type;

	AccelTest_Init();		 
	device_type = KIONIX_ACCEL_get_device_type();
	if(device_type != DEVICE_TYPE_KIONIX_KXTF9)	{
		KxPrint("Wrong device type, val: %d\n", device_type);
	}

	KIONIX_ACCEL_deinit();

	KXTF9_set_G_range(2);
	KXTF9_set_resolution(12);
	KIONIX_ACCEL_enable_wake_up_function();
	KIONIX_ACCEL_set_odr_motion (50);
	KIONIX_ACCEL_wuf_timer(5);
	KIONIX_ACCEL_wuf_thresh(5);

	KIONIX_ACCEL_int_activeh();
	KIONIX_ACCEL_int_latch();
	KIONIX_ACCEL_enable_interrupt();
	KIONIX_ACCEL_enable_outputs();		/* enable acceleration data output */

	//KxPrint("\n\nPress any key to exit....\n\n");
	//getchar();
	return;
}

void shake_detect(void)
{
	int device_type;

	AccelTest_Init();		 
	device_type = KIONIX_ACCEL_get_device_type();
	if(device_type != DEVICE_TYPE_KIONIX_KXTF9)	{
		KxPrint("Wrong device type, val: %d\n", device_type);
	}

	KIONIX_ACCEL_deinit();

	KXTF9_set_G_range(2);
	KXTF9_set_resolution(12);
	KIONIX_ACCEL_enable_wake_up_function();
	KIONIX_ACCEL_set_odr_motion (50);
	KIONIX_ACCEL_wuf_timer(20);
	KIONIX_ACCEL_wuf_thresh(20);

	KIONIX_ACCEL_int_activeh();
	KIONIX_ACCEL_int_latch();
	KIONIX_ACCEL_enable_interrupt();
	KIONIX_ACCEL_enable_outputs();		/* enable acceleration data output */

	//KxPrint("\n\nPress any key to exit....\n\n");
	//getchar();
	return;
}

void directional_tap(void)
{
	int device_type;

	AccelTest_Init();		 
	device_type = KIONIX_ACCEL_get_device_type();
	if(device_type != DEVICE_TYPE_KIONIX_KXTF9)	{
		KxPrint("Wrong device type, val: %d\n", device_type);
	}
	KIONIX_ACCEL_deinit();
	KXTF9_set_G_range(2);	
	KXTF9_set_resolution(12);
	KXTF9_set_odr_tap(400);
	KXTF9_enable_tap_detection();
	KXTF9_tap_mask_TLE();
	KXTF9_tap_mask_TRI();
	KXTF9_tap_mask_TDO();
	KXTF9_tap_mask_TUP();
	KXTF9_tap_mask_TFD();
	KXTF9_tap_mask_TFU();
	KIONIX_ACCEL_int_activeh();
	KIONIX_ACCEL_int_latch();
	KIONIX_ACCEL_enable_interrupt();
	KIONIX_ACCEL_enable_outputs();		/* enable acceleration data output */
	return;
}

void screen_rotation(void)
{
	int device_type;

	AccelTest_Init();		 
	
	device_type = KIONIX_ACCEL_get_device_type();
	if(device_type != DEVICE_TYPE_KIONIX_KXTF9)	{
		KxPrint("Wrong device type, val: %d\n", device_type);
	}

	KIONIX_ACCEL_deinit();

	KXTF9_set_G_range(2);
	KXTF9_set_resolution(12);
	KXTF9_set_odr_tilt(12);
	KIONIX_ACCEL_enable_tilt_function();
	KIONIX_ACCEL_position_mask_le();
	KIONIX_ACCEL_position_mask_ri();
	KIONIX_ACCEL_position_mask_do();
	KIONIX_ACCEL_position_mask_up();
	KIONIX_ACCEL_position_mask_fd();
	KIONIX_ACCEL_position_mask_fu();
	KIONIX_ACCEL_tilt_timer(5);
	KIONIX_ACCEL_int_activeh();
	KIONIX_ACCEL_int_latch();
	KIONIX_ACCEL_enable_interrupt();
	KIONIX_ACCEL_enable_outputs();		/* enable acceleration data output */

	return;
}


void integrated_test(void)
{	
	int device_type;
	static ACCEL_XYZ accel_val;
	
	//register initializtion and identify device type...
	AccelTest_Init();		 
	//check device type...
	device_type = KIONIX_ACCEL_get_device_type();
	if(device_type != DEVICE_TYPE_KIONIX_KXTF9)	{
		KxPrint("Wrong device type, val: %d\n", device_type);
	}

	//set registers...
	KXTF9_set_G_range(2);	//2-g range
	KXTF9_set_resolution(12); 	// for using tap function
	KIONIX_ACCEL_int_activeh();
	KIONIX_ACCEL_int_latch();

	//directional - tap
	KXTF9_set_odr_tap(400); // 400 HZ
	KXTF9_tap_unmask_TFU();
	KXTF9_tap_unmask_TFD();
	KXTF9_tap_unmask_TUP();
	KXTF9_tap_unmask_TDO();
	//KXTF9_tap_mask_TFU();		// Face Up/Down Tap Detection...
	//KXTF9_tap_mask_TFD();
	//KXTF9_tap_mask_TUP();		// Up/Down Tap Detection... 
	//KXTF9_tap_mask_TDO();
	KXTF9_tap_mask_TRI();		// 좌우(X+/-) 방향에대한 Tap Detection...
	KXTF9_tap_mask_TLE();
	KXTF9_enable_tap_detection();

	//Screen Rotation
	KXTF9_set_odr_tilt(12);		
	KIONIX_ACCEL_tilt_timer(5);		//ratation sensitivity... 
	KIONIX_ACCEL_position_unmask_fu();  //Face up/down 에 대해서 disable..
	KIONIX_ACCEL_position_unmask_fd();
	//KIONIX_ACCEL_position_mask_fu();  //Face up/down 에 대해서 disable..
	//KIONIX_ACCEL_position_mask_fd();
	KIONIX_ACCEL_position_mask_up();
	KIONIX_ACCEL_position_mask_do();
	KIONIX_ACCEL_position_mask_ri();
	KIONIX_ACCEL_position_mask_le();
	KIONIX_ACCEL_enable_tilt_function();

	//interrupt enable and operating start
	KIONIX_ACCEL_enable_interrupt();
	KIONIX_ACCEL_enable_outputs();		//enable acceleration data output
	

	while(1) {
		usleep(200000);		//200 milliseconds
		KIONIX_ACCEL_read_LPF_g(&accel_val.x, &accel_val.y, &accel_val.z);
		KxPrint("X-Y-Z = %d mg, %d mg, %d mg\n", accel_val.x, accel_val.y, accel_val.z);
	}

}

void AccelTest_Dump_Registers(void)
{
	int i=0, reg_len;
	char value;
	struct __kx__9_regs *pReg;

	switch(KIONIX_ACCEL_get_device_type())
	{
	case 0 : // KXTE9
		pReg = kxte9_regs; reg_len = sizeof(kxte9_regs);
		break;
	case 1 : // KXTF9
		pReg = kxtf9_regs; reg_len = sizeof(kxtf9_regs);
		break;
	case 2 : // KXSD9
		KxPrint("\nKXSD9 Registers \n");
		return;
	default : KxPrint("-E- Unknown device type \n");
		return;
	}
	KxPrint("REG_NAME | REG_ADDR = RESET VALUE | CURRENT VALUE \n");

	for(i=0; i<reg_len/sizeof(struct __kx__9_regs) ; i++)
	{
		KIONIX_ACCEL_read_bytes((int)(pReg[i]).reg_addr, &value, 1);
		KxPrint("%-20s | 0x%02x = 0x%02x | 0x%02x \n",
			pReg[i].regname, pReg[i].reg_addr, pReg[i].rstval, value);
	}
}


void AccelTest_Init(void)
{
	KIONIX_ACCEL_init();
	// set sensor to sleep mode
	KIONIX_ACCEL_sleep();
}

void AccelTest_deinit(void)
{
	KIONIX_ACCEL_deinit();
}

void AccelTest_Reset(void)
{
	KIONIX_ACCEL_reset();
	KxMSleep (10); // wait for 10us
}

int getyn(void)
{
	int ch=-1;

	while(ch!='y' && ch!= 'n' ) // && !isalpha(ch) )
	{
		ch = tolower(getchar());
	}
	fflush(stdin);
	return ch;
}

void AccelTest_XYZ_LPF_cnt_g(void)
{
	int inch, disp_mode;
	int i=0, res = 0;
	static ACCEL_XYZ accel_val;

	KIONIX_ACCEL_deinit();

	switch(KIONIX_ACCEL_get_device_type())
	{
	case 0: // KXTE9,  2g, 6bit fixed
		KxPrint("Support G range [2/?] (2g) \n");
		KxPrint("Support A/D resolution bit [6/?] (6-bit) \n"); 
		break;
	case 1: // KXTF9
		KxPrint("Support G range [2/4/8/?] (2g) ");  while(inch!=2 && inch!=4 && inch!=8) scanf("%d", &inch);
		res |= KXTF9_set_G_range( inch ); inch=-1;

		KxPrint("Support A/D resolution bit [8/12/?] (12-bit) ");  while(inch!=8 && inch!=12) scanf("%d", &inch);
		res |= KXTF9_set_resolution( inch);inch=-1;

		KxPrint("Support Output Data Rate for low-pass filtered acceleration output\n");
		KxPrint("[6.25/12.5/25/50/100/200/400/?] (25Hz) ");
		while(inch!=6&&inch!=12&&inch!=25&&inch!=50&&inch!=100&&inch!=200&&inch!=400) scanf("%d", &inch);
		res |= KXTF9_set_lpf_odr( inch);inch=-1;

		break;
	case 2: // KXSD9
		break;
	default : KxPrint("-E- Unknown device type \n");
		return;
	}

	res |= KIONIX_ACCEL_enable_outputs();		/* enable acceleration data output */

	KxPrint("Support display type Counts or G [c/g/?] (c) "); while(inch!='c' && inch!='g') inch = tolower(getchar());
	if(inch == 'c') disp_mode = 1;
	else disp_mode = 2;

	inch=50;
	KxPrint("Support display loop count ? (%d) ", inch); scanf("%d", &inch);

	for(i=0; i<inch; i++)
	{
		if(disp_mode == 1) {
			KIONIX_ACCEL_read_LPF_cnt(&accel_val.x, &accel_val.y, &accel_val.z);
			KxPrint("X-Y-Z = %d , %d , %d \n", accel_val.x, accel_val.y, accel_val.z);
		}
		else if(disp_mode == 2) {
			KIONIX_ACCEL_read_LPF_g(&accel_val.x, &accel_val.y, &accel_val.z);
			KxPrint("X-Y-Z = %d mg, %d mg, %d mg\n", accel_val.x, accel_val.y, accel_val.z);
		}
		KxMSleep(100);
	}
}


void AccelTest_XYZ_HPF_cnt_g(void)
{
	KxPrint("\n-I- Sorry!!! Not available. \n");
	return ;
}


void AccelTest_Alg_MotionDetection(void)
{
	int device_type=-1, res=-1, inch;
	char ctrl_reg3, value;

	device_type = KIONIX_ACCEL_get_device_type();
	if ( device_type != 1 && device_type != 0 ) {
		KxPrint("-E- Invalid device type (%d) \n", device_type);
		return;
	}

	KIONIX_ACCEL_deinit();

	switch(device_type)
	{
	case 0 : // KXTE9,
		KxPrint("Support G range [2/?] (2g) \n");
		KxPrint("Support A/D resolution bit [6/?] (6-bit) \n"); 

		res |= KIONIX_ACCEL_enable_wake_up_function();
		res |= KIONIX_ACCEL_enable_back_to_sleep();

		KxPrint("Support Output Data Rate when in the {in}active mode\n");
		KxPrint("[1/3/10/40/? ] (40Hz) "); while(inch!=1&&inch!=3&&inch!=10&&inch!=40) scanf("%d", &inch);
		res |= KIONIX_ACCEL_set_odr_motion ( inch);inch=-1;

		KIONIX_ACCEL_read_bytes(KXTE9_I2C_B2S_TIMER, &value, 1);
		KxPrint("Support B2S_TIMER [0~255/? ] (%d) ", value);	scanf("%d", &inch);
		KXTE9_b2s_timer(inch);
		KIONIX_ACCEL_read_bytes(KXTE9_I2C_B2S_THRESH, &value, 1);
		KxPrint("Support B2S_THRESH [0~255/? ] (%d) ", value); scanf("%d", &inch);
		KXTE9_b2s_thresh(inch);inch=-1;

		break;
	case 1 : // KXTF9
		KxPrint("Support G range [2/4/8/?] (2g) ");  while(inch!=2 && inch!=4 && inch!=8) scanf("%d", &inch);
		res |= KXTF9_set_G_range( inch );inch=-1;

		KxPrint("Support A/D resolution bit [8/12/?] (12-bit) ");  while(inch!=8 && inch!=12) scanf("%d", &inch);
		res |= KXTF9_set_resolution( inch);inch=-1;

		res |= KIONIX_ACCEL_enable_wake_up_function();

		KxPrint("Support Output Data Rate for the general motion detection function\n");
		KxPrint("[25/50/100/200/? ] (50Hz) "); while(inch!=25&&inch!=50&&inch!=100&&inch!=200) scanf("%d", &inch);
		res |= KIONIX_ACCEL_set_odr_motion ( inch );inch=-1;

		break;
	case 2 : // KXSD9
		break;
	}

	KIONIX_ACCEL_read_bytes(KIONIX_ACCEL_I2C_WUF_TIMER, &value, 1);
	KxPrint("Support WUF_TIMER [0~255/? ] (%d)  ", value);		scanf("%d", &inch);
	KIONIX_ACCEL_wuf_timer(inch);

	KIONIX_ACCEL_read_bytes(KIONIX_ACCEL_I2C_WUF_THRESH, &value, 1);
	KxPrint("Support WUF_THRESH [0~255/? ] (%d) ", value);	scanf("%d", &inch);
	KIONIX_ACCEL_wuf_thresh(inch);

	res |= KIONIX_ACCEL_int_activeh();
	res |= KIONIX_ACCEL_int_latch();
	KIONIX_ACCEL_enable_interrupt();
	res |= KIONIX_ACCEL_enable_outputs();		/* enable acceleration data output */

	KxPrint("\n\nPress any key to exit....\n\n");
	getchar();

	return;
}

void AccelTest_Alg_Rotation(void)
{
	int device_type=-1, res=-1, inch;
	char ctrl_reg3, value;

	device_type = KIONIX_ACCEL_get_device_type();
	if ( device_type != 1 && device_type != 0 ) {
		KxPrint("-E- Invalid device type (%d) \n", device_type);
		return;
	}

	KIONIX_ACCEL_deinit();

	switch(device_type)
	{
	case 0: // KXTE9,  2g, 6bit fixed
		KxPrint("Support G range [2/?] (2g) \n");
		KxPrint("Support A/D resolution bit [6/?] (6-bit) \n"); 

		res |= KIONIX_ACCEL_enable_tilt_function();
		break;
	case 1: // KXTF9
		KxPrint("Support G range [2/4/8/?] (2g) ");  while(inch!=2 && inch!=4 && inch!=8) scanf("%d", &inch);
		res |= KXTF9_set_G_range( inch );inch=-1;

		KxPrint("Support A/D resolution bit [8/12/?] (12-bit) ");  while(inch!=8 && inch!=12) scanf("%d", &inch);
		res |= KXTF9_set_resolution( inch);inch=-1;

		KxPrint("Support Output Data Rate for Tilt Position function \n");
		KxPrint("[1.6/6.3/12.5/50/? ] (12.5Hz) "); while(inch!=1&&inch!=6&&inch!=12&&inch!=50) scanf("%d", &inch);
		res |= KXTF9_set_odr_tilt( inch );inch=-1;

		res |= KIONIX_ACCEL_enable_tilt_function();

		KxPrint("Support tilt position axis mask for Left state [Y/n/? ] (y) ");		inch = getyn();
		if(inch=='n') res |= KIONIX_ACCEL_position_unmask_le();	else res|= KIONIX_ACCEL_position_mask_le();

		KxPrint("Support tilt position axis mask for Right state [Y/n/? ] (y) ");		inch = getyn();
		if(inch=='n') res |= KIONIX_ACCEL_position_unmask_ri();	else res|= KIONIX_ACCEL_position_mask_ri();

		KxPrint("Support tilt position axis mask for Down state [Y/n/? ] (y) ");		inch = getyn();
		if(inch=='n') res |= KIONIX_ACCEL_position_unmask_do();	else res|= KIONIX_ACCEL_position_mask_do();

		KxPrint("Support tilt position axis mask for Up state [Y/n/? ] (y) ");		inch = getyn();
		if(inch=='n') res |= KIONIX_ACCEL_position_unmask_up();	else res|= KIONIX_ACCEL_position_mask_up();

		KxPrint("Support tilt position axis mask for Face-Down state [Y/n/? ] (y) ");		inch = getyn();
		if(inch=='n') res |= KIONIX_ACCEL_position_unmask_fd();	else res|= KIONIX_ACCEL_position_mask_fd();

		KxPrint("Support tilt position axis mask for Face-Up state [Y/n/? ] (y) ");		inch = getyn();
		if(inch=='n') res |= KIONIX_ACCEL_position_unmask_fu();	else res|= KIONIX_ACCEL_position_mask_fu();

		break;
	default : KxPrint("-E- Unknown device type \n");
		return;
	}

	KIONIX_ACCEL_read_bytes(KIONIX_ACCEL_I2C_TILT_TIMER, &value, 1);
	KxPrint("Support Tilt State Timer [0~255/? ] (%d) ", value); scanf("%d", &inch);
	res |= KIONIX_ACCEL_tilt_timer(inch);

	res |= KIONIX_ACCEL_int_activeh();
	res |= KIONIX_ACCEL_int_latch();
	KIONIX_ACCEL_enable_interrupt();
	res |= KIONIX_ACCEL_enable_outputs();		/* enable acceleration data output */

	KxPrint("\n\nPress any key to exit....\n\n");
	getchar();

	return;
}

void AccelTest_Alg_DirectionalTap(void)
{
	int res, inch;
	char ctrl_reg3, value;

	if(KIONIX_ACCEL_get_device_type() != 1) {
		KxPrint("-E- Only supported for KXTF9 !!! \n");
		return;
	}

	KIONIX_ACCEL_deinit();

	KxPrint("Support G range [2/4/8/?] (2g) ");  while(inch!=2 && inch!=4 && inch!=8) scanf("%d", &inch);
	res |= KXTF9_set_G_range( inch );inch=-1;

	KxPrint("Support A/D resolution bit [8/12/?] (12-bit) ");  while(inch!=8 && inch!=12) scanf("%d", &inch);
	res |= KXTF9_set_resolution( inch);inch=-1;

	KxPrint("Support Output Data Rate for the directional-tap function\n");
	KxPrint("50/100/200/400/? ] (400Hz) "); while(inch!=50&&inch!=100&&inch!=200&&inch!=400) scanf("%d", &inch);
	res |= KXTF9_set_odr_tap( inch );inch=-1;

	res |= KXTF9_enable_tap_detection();

	KxPrint("Support tap detection axis mask for - X Negative (X-)  [Y/n/? ] (y) ");		inch = getyn();
	if(inch=='n') res |= KXTF9_tap_unmask_TLE();	else res |= KXTF9_tap_mask_TLE();
	
	KxPrint("Support tap detection axis mask for - X Positive (X+)  [Y/n/? ] (y) "); 		inch = getyn();
	if(inch=='n') res |= KXTF9_tap_unmask_TRI();	else res |= KXTF9_tap_mask_TRI();
	
	KxPrint("Support tap detection axis mask for - Y Negative (Y-)  [Y/n/? ] (y) "); 		inch = getyn();
	if(inch=='n') res |= KXTF9_tap_unmask_TDO();	else res |= KXTF9_tap_mask_TDO();
	
	KxPrint("Support tap detection axis mask for - Y Positive (Y+)  [Y/n/? ] (y) "); 		inch = getyn();
	if(inch=='n') res |= KXTF9_tap_unmask_TUP();	else res |= KXTF9_tap_mask_TUP();
	
	KxPrint("Support tap detection axis mask for - Z Negative (Z-)  [Y/n/? ] (y) "); 		inch = getyn();
	if(inch=='n') res |= KXTF9_tap_unmask_TFD();	else res |= KXTF9_tap_mask_TFD();
	
	KxPrint("Support tap detection axis mask for - Z Positive (Z+)  [Y/n/? ] (y) "); 		inch = getyn();
	if(inch=='n') res |= KXTF9_tap_unmask_TFU();	else res |= KXTF9_tap_mask_TFU();

#if 0
	KIONIX_ACCEL_read_bytes(KXTF9_I2C_TDT_TIMER, &value, 1);
	KxPrint("Support TDT_TIMER [0~255/? ] (%d) ", value );  scanf("%d", &inch);
	res |= KXTF9_tdt_timer(inch);

	KIONIX_ACCEL_read_bytes(KXTF9_I2C_TDT_H_THRESH, &value, 1);
	KxPrint("Support TDT_H_THRESH [0~255/? ] (%d) ", value );  scanf("%d", &inch);
	res |= KXTF9_tdt_h_thresh(inch);

	KIONIX_ACCEL_read_bytes(KXTF9_I2C_TDT_L_THRESH, &value, 1);
	KxPrint("Support TDT_L_THRESH [0~255/? ] (%d) ", value );  scanf("%d", &inch);
	res |= KXTF9_tdt_l_thresh(inch);

	KIONIX_ACCEL_read_bytes(KXTF9_I2C_TDT_TAP_TIMER, &value, 1);
	KxPrint("Support TDT_TAP_TIMER [0~255/? ] (%d) ", value );  scanf("%d", &inch);
	res |= KXTF9_tdt_tap_timer(inch);

	KIONIX_ACCEL_read_bytes(KXTF9_I2C_TDT_TOTAL_TIMER, &value, 1);
	KxPrint("Support TDT_TOTAL_TIMER [0~255/? ] (%d) ", value );  scanf("%d", &inch);
	res |= KXTF9_tdt_total_timer(inch);

	KIONIX_ACCEL_read_bytes(KXTF9_I2C_TDT_LATENCY_TIMER, &value, 1);
	KxPrint("Support TDT_LATENCY_TIMER [0~255/? ] (%d) ", value );  scanf("%d", &inch);
	res |= KXTF9_tdt_latency_timer(inch);

	KIONIX_ACCEL_read_bytes(KXTF9_I2C_TDT_WINDOW_TIMER, &value, 1);
	KxPrint("Support TDT_WINDOW_TIMER [0~255/? ] (%d) ", value );  scanf("%d", &inch);
	res |= KXTF9_tdt_window_timer(inch);
#endif

	res |= KIONIX_ACCEL_int_activeh();
	res |= KIONIX_ACCEL_int_latch();
	KIONIX_ACCEL_enable_interrupt();
	res |= KIONIX_ACCEL_enable_outputs();		/* enable acceleration data output */

	KxPrint("\n\nPress any key to exit....\n\n");
	getchar();

	return;
}



