#include <setjmp.h>
#include <sys/sem.h>
#include <sys/poll.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/ipc.h>

#include <sys/wait.h>
#include <signal.h>
#include <errno.h>

#include "3G_sensor/te9_tf9_hybrid_driver.h"
#include "3G_sensor/te9_tf9_main.h"
#include "3G_sensor/te9_tf9_regs.h"

static char str[64];
static int accel_fd, exitIrqChecker = 0;
struct sigaction sigact, oldact;
static ACCEL_XYZ accel_val;

/* Signal Handler... */
static void irqChecker_sigHandler(int signo) {
        switch(signo) {
                case SIGIO:
                        // KxPrint("-I- SIGIO\n");
                        KIONIX_ACCEL_service_interrupt();
                        break;

                case SIGINT:
                case SIGQUIT:
                        KxPrint("-I- SIGINT or SIGQUIT\n");
                        exitIrqChecker = 1;
                        break;
        }

        return;
}

/* Device File Initialization... */
int init_accel_device_file(void) {
        accel_fd = open("/dev/accel", O_RDWR); // blocking mode open... 
        if (accel_fd == -1){
                printf("-E- Error opening accel adapter\n");
                return -1;
        }

        return 0;
}

/* Sig Action Function Initialization... */
void init_sigact(void) {
    int oflag;

    sigact.sa_handler = irqChecker_sigHandler;
    sigemptyset(&sigact.sa_mask);
    sigact.sa_flags = SA_INTERRUPT;

    if(sigaction(SIGIO, &sigact, &oldact) < 0) {
        KxPrint("-W- can't receive SIGIO event\n");
        if(accel_fd)
            close(accel_fd);
    }

    fcntl(accel_fd, F_SETOWN, getpid());
    oflag = fcntl(accel_fd, F_GETFL);
    fcntl(accel_fd, F_SETFL, oflag | FASYNC);
}

void AccelTest_Init(void)
{
    int device_type;

    KIONIX_ACCEL_init();

    device_type = KIONIX_ACCEL_get_device_type();
    if(device_type != DEVICE_TYPE_KIONIX_KXTF9)     {
            printf("Wrong device type, val: %d\n", device_type);
    }

    KXTF9_set_G_range(2);
    KXTF9_set_resolution(12);
    KIONIX_ACCEL_int_activeh();
    KIONIX_ACCEL_int_latch();

    KXTF9_set_odr_tap(400);

    KXTF9_tap_unmask_TFU();
    KXTF9_tap_unmask_TFD();
    KXTF9_tap_unmask_TUP();
    KXTF9_tap_unmask_TDO();

    KXTF9_tap_mask_TRI();
    KXTF9_tap_mask_TLE();

    KXTF9_enable_tap_detection();
    KXTF9_set_odr_tilt(12);         
    KIONIX_ACCEL_tilt_timer(5);

    KIONIX_ACCEL_position_unmask_fu();
    KIONIX_ACCEL_position_unmask_fd();

    KIONIX_ACCEL_position_mask_up();
    KIONIX_ACCEL_position_mask_do();
    KIONIX_ACCEL_position_mask_ri();
    KIONIX_ACCEL_position_mask_le();

    KIONIX_ACCEL_enable_tilt_function();

    KIONIX_ACCEL_enable_interrupt();
    KIONIX_ACCEL_enable_outputs();
}



extern "C" void caanoo_sensor_init()
{
	init_accel_device_file();
	init_sigact();
	AccelTest_Init();
}

extern "C" void caanoo_sensor_deinit(void)
{
    KIONIX_ACCEL_deinit();

    if(accel_fd)
            close(accel_fd);
	accel_fd = 0;
}

extern "C" void caanoo_sensor_poll(int* ax, int* ay, int* az)
{
    KIONIX_ACCEL_read_LPF_g(&accel_val.x, &accel_val.y, &accel_val.z);
    // printf("X-Y-Z = %d mg, %d mg, %d mg\n", accel_val.x, accel_val.y, accel_val.z);

	*ax = accel_val.x * 33;
	*ay = accel_val.y * 33;
	*az = accel_val.z * 33;
/*
    if (accel_val.y < -80) {} // Up
    if (accel_val.x < -120) {} // Left
    if (accel_val.y > 80) {} // Down
    if (accel_val.x > 120) {} // Right
*/
}

