

#include <stdio.h>
#include <stdlib.h>
extern "C"
{
#define FAST_FLOAT  float
#define HAVE_UNSIGNED_CHAR

#include "jpeglib.h"
#include "jerror.h"
}
#include "stdafx.h"
#include "gffilesystem.h"


METHODDEF(void) my_jpeg_init_source(j_decompress_ptr cinfo)
{

}

METHODDEF(boolean) my_jpeg_fill_input_buffer(j_decompress_ptr cinfo)
{
	ERREXIT(cinfo, 0);
	return FALSE;
}

METHODDEF(void) my_jpeg_skip_input_data(j_decompress_ptr cinfo, long num_bytes)
{
	if(num_bytes > (long) (cinfo)->src->bytes_in_buffer) {
		ERREXIT(cinfo, 0);
	}
	(cinfo)->src->next_input_byte += num_bytes;
	(cinfo)->src->bytes_in_buffer -= num_bytes;
}

METHODDEF(void) my_jpeg_term_source(j_decompress_ptr cinfo)
{

}

// This is based on the FILE source manager in libjpeg

const static JOCTET EOI_BUFFER[1] = { JPEG_EOI };
struct my_source_mgr {
	struct jpeg_source_mgr pub;
	const JOCTET *data;
	size_t       len;
};

static void my_init_source(j_decompress_ptr cinfo) {}

static boolean my_fill_input_buffer(j_decompress_ptr cinfo) {
	my_source_mgr* src = (my_source_mgr*)cinfo->src;
	// No more data.  Probably an incomplete image;  just output EOI.
	src->pub.next_input_byte = EOI_BUFFER;
	src->pub.bytes_in_buffer = 1;
	return TRUE;
}
static void my_skip_input_data(j_decompress_ptr cinfo, long num_bytes) {
	my_source_mgr* src = (my_source_mgr*)cinfo->src;
	if ((long)src->pub.bytes_in_buffer < num_bytes) {
		// Skipping over all of remaining data;  output EOI.
		src->pub.next_input_byte = EOI_BUFFER;
		src->pub.bytes_in_buffer = 1;
	} else {
		// Skipping over only some of the remaining data.
		src->pub.next_input_byte += num_bytes;
		src->pub.bytes_in_buffer -= num_bytes;
	}
}
static void my_term_source(j_decompress_ptr cinfo) {}

static void my_set_source_mgr(j_decompress_ptr cinfo, const char* data, size_t len) {
	my_source_mgr* src;
	if (cinfo->src == 0) { // if this is first time;  allocate memory
		cinfo->src = (struct jpeg_source_mgr *)(*cinfo->mem->alloc_small)
			((j_common_ptr) cinfo, JPOOL_PERMANENT, sizeof(my_source_mgr));
	}
	src = (my_source_mgr*) cinfo->src;
	src->pub.init_source = my_init_source;
	src->pub.fill_input_buffer = my_fill_input_buffer;
	src->pub.skip_input_data = my_skip_input_data;
	src->pub.resync_to_restart = jpeg_resync_to_restart; // default
	src->pub.term_source = my_term_source;
	// fill the buffers
	src->data = (const JOCTET *)data;
	src->len = len;
	src->pub.bytes_in_buffer = len;
	src->pub.next_input_byte = src->data;
}

// ---------------------------------------------------- //
// gf_jpeg_read
// ---------------------------------------------------- //
int gf_jpeg_read(const char* data, int len, int* pWidth, int* pHeight, int* pChannels, unsigned char** pRGB)
{
	struct jpeg_decompress_struct info; 
	struct jpeg_error_mgr jerr;

	// Setup decompression structure
	info.err = jpeg_std_error(&jerr); 
	jpeg_create_decompress(&info); 
	my_set_source_mgr(&info, data,len); // here's the magic

	// read info from header.
	int r = jpeg_read_header(&info, TRUE);

	// now decrompress from the info structure

	jpeg_start_decompress(&info);

	int channels;      //the channels of the image 3 = RGA 4 = RGBA
	channels = info.num_components;
	*pWidth  = info.output_width;
	*pHeight = info.output_height;
	*pChannels = channels;

	unsigned long size;     //length of the file
	size = info.output_width * info.output_height * /* 3 */ channels;

	// read turn the uncompressed data into something ogl can read
	*pRGB = new unsigned char[size];

	BYTE* p1 = *pRGB;
	BYTE** p2 = &p1;
	int numlines = 0;

	while(info.output_scanline < info.output_height)
	{
		numlines = jpeg_read_scanlines(&info, p2, 1);
		*p2 += numlines * 3 * info.output_width;
	}

	jpeg_finish_decompress(&info);   //finish decompressing this file

	return 1;
}




// ---------------------------------------------------- //
// gf_write_jpeg
// ---------------------------------------------------- //
int gf_write_jpeg(const char *file_name, const unsigned char* pixels32, int width, int height)
{
	int quality = 80; // 0 .. 100

	bool ret=false;

	struct jpeg_compress_struct cinfo; // the JPEG OBJECT
	struct jpeg_error_mgr jerr; // error handler struct
	unsigned char *row_pointer[1]; // pointer to JSAMPLE row[s]
	unsigned char *flip=0;
	FILE* shot;
	int row_stride; // width of row in image buffer

	shot = fopen(file_name, "wb");
	if(shot) { // jpeg file
		// initialization
		cinfo.err = jpeg_std_error(&jerr); // error handler
		jpeg_create_compress(&cinfo); // compression object
		jpeg_stdio_dest(&cinfo, shot); // tie stdio object to JPEG object
		row_stride = width * 3;

		flip = (unsigned char *)malloc(sizeof(unsigned char)*width*height*3);

		if (flip!=NULL) {
			// give some specifications about the image to save to libjpeg
			cinfo.image_width = width;
			cinfo.image_height = height;
			cinfo.input_components = 3; // 3 for R, G, B
			cinfo.in_color_space = JCS_RGB; // type of image

			jpeg_set_defaults(&cinfo);
			jpeg_set_quality(&cinfo, quality, TRUE);
			jpeg_start_compress(&cinfo, TRUE);

			// OpenGL writes from bottom to top.
			// libjpeg goes from top to bottom.
			// flip lines.
			for (int y=0;y<height;++y)
			{
				for (int x=0;x<width;++x)
				{
					flip[(y*width+x)*3]   = pixels32[(y*width+x)*4+2]; // 4 for RGBA
					flip[(y*width+x)*3+1] = pixels32[(y*width+x)*4+1];
					flip[(y*width+x)*3+2] = pixels32[(y*width+x)*4+0];
				}
			}

			// write the lines
			while (cinfo.next_scanline < cinfo.image_height) {
				row_pointer[0] = &flip[cinfo.next_scanline * row_stride];
				jpeg_write_scanlines(&cinfo, row_pointer, 1);
			}

			ret=true;
			// finish up and free resources
			jpeg_finish_compress(&cinfo);
			jpeg_destroy_compress(&cinfo);
		}
		fclose(shot);
	}

	if(flip!=0)
		free(flip);

	return ret;
}

