
#define PNG_NO_STDIO


#include "png.h"
#include <stdlib.h>

#include "stdafx.h"
#include "gffilesystem.h"


static png_structp png_ptr=0;
static png_infop   info_ptr=0;

// FILE *png_fp=0;
static GFFileSystem* png_fp=NULL;


static int g_PNG_error = 0;
static void gf_PNG_default_error(png_structp png_ptr_, png_const_charp error_message)
{
	g_PNG_error=1;
}
static void gf_PNG_default_warning(png_structp png_ptr_, png_const_charp error_message)
{
}



/* The replacement fread wrapper which we will pass to
 * png_set_read_fn(). */
void gf_PNG_fread(png_structp png_ptr_, png_bytep data, png_size_t length)
{
    if ( (length <= 0)  || png_get_io_ptr(png_ptr_) == NULL)
    {
    
    }
    else
    {
		((GFFileSystem*)png_get_io_ptr(png_ptr_))->Read(data, (int)length);
	}
}


/* The replacement fread wrapper which we will pass to
 * png_set_read_fn(). */
void gf_PNG_fwrite(png_structp png_ptr_, png_bytep data, png_size_t length)
{
    if ( (length <= 0)  || png_get_io_ptr(png_ptr_) == NULL)
    {
    
    }
    else
    {
		((GFFileSystem*)png_get_io_ptr(png_ptr_))->Write(data, (int)length);
	}
}
void gf_PNG_fflush(png_structp png_ptr_)
{
}



void gf_png_close();


int gf_png_open(const char* szFileName, int* width, int* height, int* rowbytes, int* channels)
{
int ok=1;
g_PNG_error=0;
	for(;;)
	{
		gf_png_close();
		png_fp = new GFFileSystem();
		/* open the file */
		if(!png_fp->OpenIn(szFileName))
		{
			ok=0;
			break;
		}
	
		// read the file header yourself - just to be very safe
		char header[4];
		png_fp->Read(header, 4);
		png_fp->SeekTo(0);
		// 89 50 4E 47
		if(/*header[0]!=0x89 || */ header[1]!=0x50 || header[2]!=0x4E || header[3]!=0x47)
		{
			png_fp->Close();
			delete png_fp; png_fp = NULL;
			return 0;
		}

		/* Create and initialize the png_struct with the desired error handler
		functions.  If you want to use the default stderr and longjump method,
		you can supply NULL for the last three parameters.  We also check that
		the header file is compatible with the library version.
		*/
		png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
		if (!png_ptr) {ok=0; break;}
		png_set_error_fn(png_ptr, 0, &gf_PNG_default_error, gf_PNG_default_warning);

		info_ptr = png_create_info_struct(png_ptr);
		if (!info_ptr) {ok=0; break;}

		png_set_read_fn(png_ptr, (void *)png_fp, gf_PNG_fread);

		/* set error handling if you are using the setjmp/longjmp method */
		// if (setjmp(png_ptr->jmpbuf)) {ok=0; break;}

		/* set up the input control if you are using standard C streams */
//		png_init_io(png_ptr, png_fp);

		/* read the file information */
		png_read_info(png_ptr, info_ptr);
		if(g_PNG_error) {ok=0; break;}

		/* set up the transformations you want.  Note that these are all optional.  Only call them if you want them */

		/* expand paletted colors into true RGB triplets */
		if (png_get_color_type(png_ptr, info_ptr) == PNG_COLOR_TYPE_PALETTE)
			png_set_expand(png_ptr);

		/* expand grayscale images to the full 8 bits */
		if (png_get_color_type(png_ptr, info_ptr) == PNG_COLOR_TYPE_GRAY && png_get_bit_depth(png_ptr, info_ptr) < 8)
			png_set_expand(png_ptr);

		/* expand paletted or RGB images with transparency to full alpha channels
		* so the data will be available as RGBA quartets */
		if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
			png_set_expand(png_ptr);
		if(g_PNG_error) {ok=0; break;}

/*
		// tell libpng to handle the gamma conversion for you
		double screen_gamma = 2.0;
		if (info_ptr->valid & PNG_INFO_gAMA)
			png_set_gamma(png_ptr, screen_gamma, info_ptr->gamma);
		else
			png_set_gamma(png_ptr, screen_gamma, 0.45);
*/

		/* PNG can have files with 16 bits per channel.  If you only can handle
		8 bits per channel, this will strip the pixels down to 8 bit. */
		if (png_get_bit_depth(png_ptr, info_ptr) == 16)
			png_set_strip_16(png_ptr);

		/* pack multiple pixels with bit depths of 1, 2, and 4 into bytes */
		if (png_get_bit_depth(png_ptr, info_ptr)<8)
			png_set_packing(png_ptr);

		/* optional call to gamma correct and add the background to the palette and update info structure. */
		png_read_update_info(png_ptr, info_ptr);
		if(g_PNG_error) {ok=0; break;}

		int nchannels = png_get_channels(png_ptr, info_ptr);
		if (nchannels!=1 && nchannels!=3 && nchannels!=4) {
			// printf("File %s has %d channels.\nTry using 1, 3 or 4 channels.\n",filename, info_ptr->channels);
			ok = 0;
			break;
		}

		if (png_get_bit_depth(png_ptr, info_ptr)!=8)
		{
			printf("File %s has bit depth %d.\n", szFileName, (int)info_ptr->bit_depth);
			printf("Only bit depth of 8 is supported.\n");
			ok = 0;
			break;
		}
		if (   png_get_color_type(png_ptr, info_ptr)!=PNG_COLOR_TYPE_RGB
			&& png_get_color_type(png_ptr, info_ptr)!=PNG_COLOR_TYPE_GRAY
			&& png_get_color_type(png_ptr, info_ptr)!=PNG_COLOR_TYPE_RGB_ALPHA)
		{
			printf("File %s has unsupported color type.\n", szFileName);
			printf("Types GRAY, RGB and RGB_ALPHA are supported.\n");
			ok = 0;
			break;
		}
		if(g_PNG_error) {ok=0; break;}

		// the easiest way to read the image
		if (!png_get_image_height(png_ptr, info_ptr) )
		{
			ok = 0;
			break;
		}
		
		*rowbytes=(int)png_get_rowbytes(png_ptr, info_ptr);
		*width =png_get_image_width(png_ptr, info_ptr);
		*height=png_get_image_height(png_ptr, info_ptr);
		*channels = nchannels;

		break; // don't loop
	}
	if(g_PNG_error) {ok=0;}


	/* that's it */
	return ok;
}

void gf_png_read(unsigned char* buffer, unsigned char** row_pointers)
{
unsigned int row;
	for (row = 0; row < info_ptr->height; row++)
		row_pointers[row] = &buffer[row*info_ptr->rowbytes];

	png_read_image(png_ptr, row_pointers);

	/* read the rest of the file, getting any additional chunks in info_ptr */
//	if (setjmp(png_ptr->jmpbuf))
	if(g_PNG_error)
		png_read_end(png_ptr, info_ptr);
}

void gf_png_close()
{
	if(png_ptr)
	{
		/* clean up after the read, and free any memory allocated */
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		png_ptr = NULL;
		info_ptr = NULL;
	}
	/* close the file */
	if (png_fp)
	{
		((GFFileSystem*)png_fp)->Close();
		delete png_fp; png_fp = NULL;
	}
}


/* write a png file */
int gf_write_png(const char *file_name, const unsigned char* pixels32, int width, int height)
{
//   png_structp png_ptr; <- global objects
//   png_infop info_ptr;
//   png_colorp palette;
   png_uint_32 k;
   png_bytep* row_pointers;
   png_color_16 white;
   white.red = 0xff;
    white.green = 0xff;
    white.blue = 0xff;

   /* open the file */
GFFileSystem* fp;
	fp = new GFFileSystem();
	
	if(!fp->OpenOut(file_name))
	{
		delete fp; fp=NULL;
		return 0;
	}


   /* Create and initialize the png_struct with the desired error handler
    * functions.  If you want to use the default stderr and longjump method,
    * you can supply NULL for the last three parameters.  We also check that
    * the library version is compatible with the one used at compile time,
    * in case we are using dynamically linked libraries.  REQUIRED.
    */
   png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0,0,0); //png_voidp user_error_ptr, user_error_fn, user_warning_fn);

   if (png_ptr == NULL)
   {
		delete fp; fp=NULL;
      return 0;
   }

   /* Allocate/initialize the image information data.  REQUIRED */
   info_ptr = png_create_info_struct(png_ptr);
   if (info_ptr == NULL)
   {
	  delete fp; fp=NULL;
	  png_destroy_write_struct(&png_ptr, NULL);
      return 0;
   }

   /* Set error handling.  REQUIRED if you aren't supplying your own
    * error handling functions in the png_create_write_struct() call.
    */
   // if (setjmp(png_jmpbuf(png_ptr)))
   if(g_PNG_error)
   {
      /* If we get here, we had a problem reading the file */
      delete fp; fp=NULL;
      png_destroy_write_struct(&png_ptr, &info_ptr);
      return 0;
   }

   /* set up the output control if you are using standard C streams */
//   png_init_io(png_ptr, fp);
    png_set_write_fn(png_ptr, (void *)fp, gf_PNG_fwrite, gf_PNG_fflush);


   /* This is the hard way */

   /* Set the image information here.  Width and height are up to 2^31,
    * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
    * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
    * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
    * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
    * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
    * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
    */
   png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_RGB_ALPHA,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);


#if 0
   /* set the palette if there is one.  REQUIRED for indexed-color images */
   palette = (png_colorp)png_malloc(png_ptr, PNG_MAX_PALETTE_LENGTH
             * sizeof (png_color));
   /* ... set palette colors ... */
   png_set_PLTE(png_ptr, info_ptr, palette, PNG_MAX_PALETTE_LENGTH);
   /* You must not free palette here, because png_set_PLTE only makes a link to
      the palette that you malloced.  Wait until you are about to destroy
      the png structure. */
#endif


/*
   // optional significant bit chunk 
   // if we are dealing with a grayscale image then 
   sig_bit.gray = true_bit_depth;
   // otherwise, if we are dealing with a color image then 
   sig_bit.red = true_red_bit_depth;
   sig_bit.green = true_green_bit_depth;
   sig_bit.blue = true_blue_bit_depth;
   // if the image has an alpha channel then 
   sig_bit.alpha = true_alpha_bit_depth;
   png_set_sBIT(png_ptr, info_ptr, sig_bit);
*/

   /* Optional gamma chunk is strongly suggested if you have any guess
    * as to the correct gamma of the image.
    */
//   png_set_gAMA(png_ptr, info_ptr, gamma);

   /* Optionally write comments into the image */
/*
   struct png_text text_ptr[3];
   text_ptr[0].key = "Title";
   text_ptr[0].text = "Export Image";
   text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
   text_ptr[1].key = "Author";
   text_ptr[1].text = "GLBasic";
   text_ptr[1].compression = PNG_TEXT_COMPRESSION_NONE;
   text_ptr[2].key = "Description";
   text_ptr[2].text = "it's an image";
   text_ptr[2].compression = PNG_TEXT_COMPRESSION_zTXt;
#ifdef PNG_iTXt_SUPPORTED
   text_ptr[0].lang = NULL;
   text_ptr[1].lang = NULL;
   text_ptr[2].lang = NULL;
#endif
   png_set_text(png_ptr, info_ptr, text_ptr, 3);
*/


   /* other optional chunks like cHRM, bKGD, tRNS, tIME, oFFs, pHYs, */
   /* note that if sRGB is present the gAMA and cHRM chunks must be ignored
    * on read and must be written in accordance with the sRGB profile */

   /* Write the file header information.  REQUIRED */
   png_write_info(png_ptr, info_ptr);

   /* If you want, you can write the info in two steps, in case you need to
    * write your private chunk ahead of PLTE:
    *
    *   png_write_info_before_PLTE(write_ptr, write_info_ptr);
    *   write_my_chunk();
    *   png_write_info(png_ptr, info_ptr);
    *
    * However, given the level of known- and unknown-chunk support in 1.1.0
    * and up, this should no longer be necessary.
    */

   /* Once we write out the header, the compression type on the text
    * chunks gets changed to PNG_TEXT_COMPRESSION_NONE_WR or
    * PNG_TEXT_COMPRESSION_zTXt_WR, so it doesn't get written out again
    * at the end.
    */

   /* set up the transformations you want.  Note that these are
    * all optional.  Only call them if you want them.
    */

   /* invert monochrome pixels */
//   png_set_invert_mono(png_ptr);

   /* Shift the pixels up to a legal bit depth and fill in
    * as appropriate to correctly scale the image.
    */
//   png_set_shift(png_ptr, &sig_bit);

   /* pack pixels into bytes */
   png_set_packing(png_ptr);

   /* swap location of alpha bytes from ARGB to RGBA */
//   png_set_swap_alpha(png_ptr);

   /* Get rid of filler (OR ALPHA) bytes, pack XRGB/RGBX/ARGB/RGBA into
    * RGB (4 channels -> 3 channels). The second parameter is not used.
    */
//   png_set_filler(png_ptr, 0, PNG_FILLER_BEFORE);


// Background color for transparent pixels
	png_set_bKGD(png_ptr, info_ptr, &white);

   /* flip BGR pixels to RGB */
   // png_set_bgr(png_ptr);


   /* swap bytes of 16-bit files to most significant byte first */
//   png_set_swap(png_ptr);

   /* swap bits of 1, 2, 4 bit packed pixel formats */
//   png_set_packswap(png_ptr);


   /* The easiest way to write the image (you may have a different memory
    * layout, however, so choose what fits your needs best).  You need to
    * use the first method if you aren't handling interlacing yourself.
    */
   row_pointers = (png_bytep*)malloc(height * sizeof(void*));
   for (k = 0; k <(png_uint_32)height; k++)
     row_pointers[k] = (png_byte*)pixels32 + k*width*4  /*bytes_per_pixel*/;

   /* One of the following output methods is REQUIRED */
   png_write_image(png_ptr, row_pointers);
   free(row_pointers);


   /* You can write optional chunks like tEXt, zTXt, and tIME at the end
    * as well.  Shouldn't be necessary in 1.1.0 and up as all the public
    * chunks are supported and you can use png_set_unknown_chunks() to
    * register unknown chunks into the info structure to be written out.
    */

   /* It is REQUIRED to call this to finish writing the rest of the file */
   png_write_end(png_ptr, info_ptr);



   /* If you png_malloced a palette, free it here (don't free info_ptr->palette,
      as recommended in versions 1.0.5m and earlier of this example; if
      libpng mallocs info_ptr->palette, libpng will free it).  If you
      allocated it with malloc() instead of png_malloc(), use free() instead
      of png_free(). */
 //  png_free(png_ptr, palette);
 //  palette=NULL;

   /* Similarly, if you png_malloced any data that you passed in with
      png_set_something(), such as a hist or trans array, free it here,
      when you can be sure that libpng is through with it. */
/*
   png_free(png_ptr, trans);
   trans=NULL;
*/
   /* clean up after the write, and free any memory allocated */
   png_destroy_write_struct(&png_ptr, &info_ptr);

   /* close the file */
   delete fp; fp=NULL;

   /* that's it */
   return 1;
}




#include <stdio.h>
#include <stdarg.h>

#ifndef png_snprintf
extern "C" int png_snprintf( char * p_szBuffer, unsigned int p_nCount, const char * p_cszFormat, ... )
{
    int retval = 0;
    va_list va_args;
    va_start( va_args, p_cszFormat );
    retval = sprintf( p_szBuffer, p_cszFormat, va_args );
    va_end(va_args);
    return retval;
}
#endif
#ifndef png_snprintf2
extern "C" int png_snprintf2( char * p_szBuffer, unsigned int p_nCount, const char * p_cszFormat, ... )
{
    int retval = 0;
    va_list va_args;
    va_start( va_args, p_cszFormat );
    retval = sprintf( p_szBuffer, p_cszFormat, va_args );
    va_end(va_args);
    return retval;
}
#endif
#ifndef png_snprintf6
extern "C" int png_snprintf6( char * p_szBuffer, unsigned int p_nCount, const char * p_cszFormat, ... )
{
    int retval = 0;
    va_list va_args;
    va_start( va_args, p_cszFormat );
    retval = sprintf( p_szBuffer, p_cszFormat, va_args );
    va_end(va_args);
    return retval;
}
#endif
