//***********************************************************************//
//																		 //
//		- "Talk to me like I'm a 3 year old!" Programming Lessons -		 //
//                                                                       //
//		$Author:		DigiBen			DigiBen@GameTutorials.com		 //
//																		 //
//		$Program:		SpherePolygonCollision							 //
//																		 //
//		$Description:	This demonstrates sphere and polygon collision.	 //
//																		 //
//		$Date:			1/23/02											 //
//																		 //
//***********************************************************************//


#include "stdafx.h"
#include <math.h>
#include "3dmath.h"
#include "vectormath.h"

#ifdef HAVE_OPENGL
	#include <GL/gl.h>
#endif


#ifndef PI
#define PI 3.1415926535897932f
#endif

#define BEHIND		0
#define INTERSECTS	1
#define FRONT		2
#define FABS(a) ((a)<0 ? -(a) : (a))


/////////////////////////////////////// NORMAL \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
//          ID INDX
	// Get 2 vectors from the polygon (2 sides), Remember the order!
	// float[3] vVector1 = vPoly[2] - vPoly[0];
	// float[3] vVector2 = vPoly[1] - vPoly[0];
inline void TriangleNormal(float* Vn, float Poly[][3])
{
float V1[3], V2[3];
	V1[0]=Poly[2][0]-Poly[0][0];
	V1[1]=Poly[2][1]-Poly[0][1];
	V1[2]=Poly[2][2]-Poly[0][2];
	V2[0]=Poly[1][0]-Poly[0][0];
	V2[1]=Poly[1][1]-Poly[0][1];
	V2[2]=Poly[1][2]-Poly[0][2];
	gCross(V1, V2, Vn);
	gUnify(Vn);
}


/////////////////////////////////// PLANE DISTANCE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
// Use the plane equation to find the distance (Ax + By + Cz + D = 0)  We want to find D.
// So, we come up with D = -(Ax + By + Cz)
// Basically, the negated dot product of the normal of the plane and the point.
#define PlaneDistance(Normal, Point) (- ((Normal[0] * Point[0]) + (Normal[1] * Point[1]) + (Normal[2] * Point[2])))


/////////////////////////////////// INTERSECTED PLANE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////	This checks to see if a line intersects a plane
inline bool IntersectedPlane(float vPoly[3][3], float vLine[3][3], float vNormal[3], float &originDistance)
{
float distance1, distance2;
	 TriangleNormal(vNormal, vPoly);							// We need to get the normal of our plane to go any further

	// Let's find the distance our plane is from the origin.  We can find this value
	// from the normal to the plane (polygon) and any point that lies on that plane (Any vertex)
	originDistance = PlaneDistance(vNormal, vPoly[0]);

	// Get the distance from point1 from the plane using: Ax + By + Cz + D = (The distance from the plane)
	distance1 = ((vNormal[0] * vLine[0][0])  +					// Ax +
		         (vNormal[1] * vLine[0][1])  +					// Bx +
				 (vNormal[2] * vLine[0][2])) + originDistance;	// Cz + D
	
	// Get the distance from point2 from the plane using Ax + By + Cz + D = (The distance from the plane)
	distance2 = ((vNormal[0] * vLine[1][0])  +					// Ax +
		         (vNormal[1] * vLine[1][1])  +					// Bx +
				 (vNormal[2] * vLine[1][2])) + originDistance;	// Cz + D

	// Now that we have 2 distances from the plane, if we times them together we either
	// get a positive or negative number.  If it's a negative number, that means we collided!
	// This is because the 2 points must be on either side of the plane (IE. -1 * 1 = -1).

	if(distance1 * distance2 >= 0)			// Check to see if both point's distances are both negative or both positive
	   return false;						// Return false if each point has the same sign.  -1 and 1 would mean each point is on either side of the plane.  -1 -2 or 3 4 wouldn't...
					
return true;							// The line intersected the plane, Return TRUE
}


/////////////////////////////////// ANGLE BETWEEN VECTORS \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////	This checks to see if a point is inside the ranges of a polygon
inline float AngleBetweenVectors(float Vector1[3], float Vector2[3])
{
float z = (gGetLength(Vector1) * gGetLength(Vector2));
	if (z<=1E-8)
		return 0;
return acosf( gDot(Vector1, Vector2)/z);
}


/////////////////////////////////// INTERSECTION POINT \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////	This returns the intersection point of the line that intersects the plane
void IntersectionPoint(float vPoint[3], float vNormal[3], float vLine[3][3], float distance)
{
float vLineDir[3];
float Numerator = 0.0, Denominator = 0.0, dist = 0.0;
	vLineDir[0] = vLine[1][0] - vLine[0][0];		// Get the Vector of the line
	vLineDir[1] = vLine[1][1] - vLine[0][1];		// Get the Vector of the line
	vLineDir[2] = vLine[1][2] - vLine[0][2];		// Get the Vector of the line
	gUnify(vLineDir);				// Normalize the lines vector
	// distance from one of our points to the plane.
	Numerator = - (vNormal[0] * vLine[0][0] +		// Use the plane equation with the normal and the line
				   vNormal[1] * vLine[0][1] +
				   vNormal[2] * vLine[0][2] + distance);
	// 3) If we take the dot product between our line vector and the normal of the polygon,
	Denominator = gDot(vNormal, vLineDir);			// Get the dot product of the line's vector and the normal of the plane

	if( Denominator == 0.0)							// Check so we don't divide by zero
	{
		vPoint[0]=vLine[0][0];						// Return an arbitrary point on the line
		vPoint[1]=vLine[0][1];
		vPoint[2]=vLine[0][2];
		return;
	}
	dist = Numerator / Denominator;				// Divide to get the multiplying (percentage) factor
	
	// Now, like we said above, we times the dist by the vector, then add our arbitrary point.
	vPoint[0] = (float)(vLine[0][0] + (vLineDir[0] * dist));
	vPoint[1] = (float)(vLine[0][1] + (vLineDir[1] * dist));
	vPoint[2] = (float)(vLine[0][2] + (vLineDir[2] * dist));
}


/////////////////////////////////// INSIDE POLYGON \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////
/////	This checks to see if a point is inside the ranges of a polygon
/////
/////////////////////////////////// INSIDE POLYGON \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*

bool InsidePolygon(float vIntersection[3], float Poly[3][3], long verticeCount)
{
const float MATCH_FACTOR = 0.99f;
float Angle = 0.0f;
float vA[3], vB[3];
int		n;

	for (int i = 0; i < verticeCount; i++)		// Go in a circle to each vertex and get the angle between
	{	
		vA[0] = Poly[i][0] - vIntersection[0];			// Subtract the intersection point from the current vertex
		vA[1] = Poly[i][1] - vIntersection[1];			// Subtract the intersection point from the current vertex
		vA[2] = Poly[i][2] - vIntersection[2];			// Subtract the intersection point from the current vertex
		
		n=(i + 1) % verticeCount;								// Subtract the point from the next vertex
		vB[0] = Poly[n][0]-vIntersection[0];
		vB[1] = Poly[n][1]-vIntersection[1];
		vB[2] = Poly[n][2]-vIntersection[2];

		Angle += AngleBetweenVectors(vA, vB);	// Find the angle between the 2 vectors and add them all up as we go along
	}

	if(Angle >= (MATCH_FACTOR * (2.0f * PI_F)) )	// If the angle is greater than 2 PI, (360 degrees)
		return true;							// The point is inside of the polygon
return false;								// If you get here, it obviously wasn't inside the polygon, so Return FALSE
}


/////////////////////////////////// INTERSECTED POLYGON \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////	This checks if a line is intersecting a polygon
bool IntersectedPolygon(float vPoly[3][3], float vLine[3][3], int verticeCount)
{
float vNormal[3];
float originDistance = 0;

	// First, make sure our line intersects the plane
									 // Reference   // Reference
	if(!IntersectedPlane(vPoly, vLine,   vNormal,   originDistance))
		return false;

	// Now that we have our normal and distance passed back from IntersectedPlane(), 
	// we can use it to calculate the intersection point.  
	float vIntersection[3];
		IntersectionPoint(vIntersection, vNormal, vLine, originDistance);

	// Now that we have the intersection point, we need to test if it's inside the polygon.
	if(InsidePolygon(vIntersection, vPoly, verticeCount))
		return true;							// We collided!	  Return success

return false;								// There was no collision, so return false
}

////////////////////////////// CLOSEST POINT ON LINE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////	This returns the point on the line vA_vB that is closest to the point vPoint
inline static void ClosestPointOnLine(float vRes[3], float vA[3], float vB[3], float vPoint[3])
{
    static float t, d;
	// Create the vector from end point vA to our point vPoint.
	static float vVector1[3];
	vVector1[0] = vPoint[0] - vA[0];
	vVector1[1] = vPoint[1] - vA[1];
	vVector1[2] = vPoint[2] - vA[2];

	// Create a normalized direction vector from end point vA to end point vB
    static float vVector2[3];
	vVector2[0] = vA[0] - vB[0];
	vVector2[1] = vA[1] - vB[1];
	vVector2[2] = vA[2] - vB[2];
	gUnify(vVector2);

	// Use the distance formula to find the distance of the line segment (or magnitude)
	d = Distance(vA, vB);

	// Point A and B are the same
	if (d<1E-8)
	{
		vRes[0]=vA[0]; vRes[1]=vA[1]; vRes[2]=vA[2];
		return;
	}

	// Using the dot product, we project the vVector1 onto the vector vVector2.
	// This essentially gives us the distance from our projected vector from vA.
	t = gDot(vVector2, vVector1)/d;

	// If our projected distance from vA, "t", is less than or equal to 0, it must
	// be closest to the end point vA.  We want to return this end point.
    if (t <= 0)
	{
		vRes[0]=vA[0]; vRes[1]=vA[1]; vRes[2]=vA[2];
		return;
	}

	// If our projected distance from vA, "t", is greater than or equal to the magnitude
	// or distance of the line segment, it must be closest to the end point vB.  So, return vB.
    if (t*t >= d) 
	{
		vRes[0]=vB[0]; vRes[1]=vB[1]; vRes[2]=vB[2];
		return;
	}

	// To find the closest point on the line segment, we just add vVector3 to the original
	// end point vA.  
	vRes[0] = vA[0] + vVector2[0] * t;
	vRes[1] = vA[1] + vVector2[1] * t;
	vRes[2] = vA[2] + vVector2[2] * t;
}


///////////////////////////////// CLASSIFY SPHERE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////	This tells if a sphere is BEHIND, in FRONT, or INTERSECTS a plane, also it's distance
int ClassifySphere(float vCenter[3], float vNormal[3], float vPoint[3], float radius, float &distance)
{
float d = PlaneDistance(vNormal, vPoint);
	distance = (vNormal[0] * vCenter[0] + vNormal[1] * vCenter[1] + vNormal[2] * vCenter[2] + d);

	if(FABS(distance) < radius)
		return INTERSECTS;
	else if(distance > radius)
		return FRONT;
return BEHIND;
}


///////////////////////////////// EDGE SPHERE COLLSIION \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////	This returns true if the sphere is intersecting any of the edges of the polygon
bool EdgeSphereCollision(float vCenter[3], float vPolygon[3][3], int vertexCount, float radiusSQ)
{
	float vPoint[3];
	for(int i = 0; i < vertexCount; i++)
	{
		ClosestPointOnLine(vPoint, vPolygon[i], vPolygon[(i + 1) % vertexCount], vCenter);
		float distance = Distance(vPoint, vCenter);
		if(distance < radiusSQ)
			return true;
	}
return false;
}


////////////////////////////// SPHERE POLYGON COLLISION \\\\\\\\\\\\\\\\\\\\\\\\\\\\\*
/////	This returns true if our sphere collides with the polygon passed in
bool SpherePolygonCollision(float vPolygon[3][3], float* vCenter, int vertexCount, float radius)
{
float rvPolygon[3][3], rvCenter[3];
rvPolygon[0][0]=(float)vPolygon[0][0];
rvPolygon[1][0]=(float)vPolygon[1][0];
rvPolygon[2][0]=(float)vPolygon[2][0];
rvPolygon[0][1]=(float)vPolygon[0][1];
rvPolygon[1][1]=(float)vPolygon[1][1];
rvPolygon[2][1]=(float)vPolygon[2][1];
rvPolygon[0][2]=(float)vPolygon[0][2];
rvPolygon[1][2]=(float)vPolygon[1][2];
rvPolygon[2][2]=(float)vPolygon[2][2];
rvCenter[0]=(float)vCenter[0];
rvCenter[1]=(float)vCenter[1];
rvCenter[2]=(float)vCenter[2];

// Let's use our Normal() function to return us the normal to this polygon
static float vNormal[3];
	TriangleNormal(vNormal, rvPolygon);
// This will store the distance our sphere is from the plane
static float distance;
static int classification;
static float vOffset[3], vPosition[3];
	distance = 0.0f;
	classification = ClassifySphere(rvCenter, vNormal, rvPolygon[0], radius, distance);
	if(classification == INTERSECTS && distance<=radius) 
	{
		vOffset[0] = vNormal[0] * distance;
		vOffset[1] = vNormal[1] * distance;
		vOffset[2] = vNormal[2] * distance;
		vPosition[0] = vCenter[0] - vOffset[0];
		vPosition[1] = vCenter[1] - vOffset[1];
		vPosition[2] = vCenter[2] - vOffset[2];

		if(InsidePolygon(vPosition, rvPolygon, vertexCount))
			return true;	// We collided!
		else
		{
			if(EdgeSphereCollision(rvCenter, rvPolygon, vertexCount, radius))
			{
				return true;	// We collided! "And you doubted me..." - Sphere
			}
		}
	}
// If we get here, there is obviously no collision happening up in this crib
return false;
}




/*
Cuboid + Triangle intersection

The theory is difficult to explain in details, but roughly it goes like this. If there is a plane where
the box is totally on one side of the plane while the triangle is totally on the other side of the plane,
then the plane separates the two objects and the objects do not intersect. If on the contrary, you cannot find
such a plane, then the objects intersect.

Of course, there is potentially an infinite number of planes that can separate a box and a triangle, but you
can reduce the number of planes to test to 13. It's proven somewhere, but it;s actually hard to prove.

The planes are also refered as 'separation axis', where the normal of the plane is the separation axis.

ok, so given a box
O,
A0, A1, A2,
a0, a1, a1

where O is the centre,
A0, A1, A2 are the direction of the faces (basically, A0=(1, 0, 0), A1=(0, 1, 0), A2=(0, 0, 1) for an axis aligned box),
and a0, a1, a2 are the extents of the box along the respective axes A0, A1, A2
(for a Min-Max type of box,a0=(MaxX-MinX)/2, a1=(MaxY-MinY)/2, ...).

and a triangle
V0, V1, V2
N (normal of the triangle, N=(V2-V1) x (V1-V0)).

the 13 separation axes are
N
A0
A1
A2
A0 x (V1-V0)
A0 x (V2-V1)
A0 x (V0-V2)
A1 x (V1-V0)
A1 x (V2-V1)
A1 x (V0-V2)
A2 x (V1-V0)
A2 x (V2-V1)
A2 x (V0-V2)

As you may see, it's a combination of plane normals, and edge dot products. That also relates
to collision detection where the main collisions you get from colliding two polygonal models are
either point-plane collision or edge-ege collisions. All other collisions are special cases of these
(point-edge, edge-plane, ....).

these vectors represent the normal of a potential separating axis. To decide if this axis is separating
the box and triangle, you 'project' the two objects along the axes and see if the intervals overlap,

S is the separation axis we are testing (one of the 13 options above).
'' is a vector dot product.
'x' is a vector cross product.
'|...|' is absolute value.

minb = OS - |A0S| * a0 + |A1S| * a1 + |A2S| * a2
maxb = OS + |A0S| * a0 + |A1S| * a1 + |A2S| * a2

mint = min(min(V0S, V1S), V2S)
maxt = max(max(V0S, V1S), V2S)

if (minb > maxt || maxb < mint), then the axis S is indeed separating the objects, and the objects do not intersect.

As soon as you found an axis as being a true separation axis, you can quit the test, safe in the
knowledge that the objects are spearated by a plane and therefore cannot intersect.

if the 13 tests all fail, then the two objects intersect.

so, no divisions, just dot products and cross products, so fast, and no worries of divisions by 0.

// GF:
Since the a0,a1,a2 values could be set to 1.0 and the Ax values might be scaled before,
I leave the ax values off here for more speed.
*/

bool IntersectCuboidAndTriangle(
													const float* Org, const float* A0, const float* A1, const float* A2,
													const float* V0, const float* V1, const float* V2)
{
	float A0S_A1S_A2S, tmp[3], tmp2[3], V0S, V1S, V2S, OrgS;
	
	const float* S;
	
	#define DOCHECK													\
	A0S_A1S_A2S=FABS(gDot(A0,S))+FABS(gDot(A1,S))+FABS(gDot(A2,S)); \
	V0S = gDot(V0, S); V1S = gDot(V1, S); V2S = gDot(V2, S);		\
	OrgS = gDot(Org, S);											\
	if(OrgS - A0S_A1S_A2S > max(max(V0S, V1S),V2S)) return false;	\
	if(OrgS + A0S_A1S_A2S < min(min(V0S, V1S),V2S)) return false;

	S=A0; DOCHECK;
	S=A1; DOCHECK;
	S=A2; DOCHECK;
	S=tmp2;
	gCross(V1, V2, tmp2); DOCHECK;
	gDiffVertex(tmp, V1, V0); gCross(A0, tmp, tmp2); DOCHECK;
	gDiffVertex(tmp, V2, V1); gCross(A0, tmp, tmp2); DOCHECK;
	gDiffVertex(tmp, V0, V2); gCross(A0, tmp, tmp2); DOCHECK;
	gDiffVertex(tmp, V1, V0); gCross(A1, tmp, tmp2); DOCHECK;
	gDiffVertex(tmp, V2, V1); gCross(A1, tmp, tmp2); DOCHECK;
	gDiffVertex(tmp, V0, V2); gCross(A1, tmp, tmp2); DOCHECK;
	gDiffVertex(tmp, V1, V0); gCross(A2, tmp, tmp2); DOCHECK;
	gDiffVertex(tmp, V2, V1); gCross(A2, tmp, tmp2); DOCHECK;
	gDiffVertex(tmp, V0, V2); gCross(A2, tmp, tmp2); DOCHECK;
return true;
}



static void invertMatrix(const float *m, float *out )
{
/* NB. OpenGL Matrices are COLUMN major. */
#define MAT(m,r,c) (m)[(c)*4+(r)]

/* Here's some shorthand converting standard (row,column) to index. */
#define m11 MAT(m,0,0)
#define m12 MAT(m,0,1)
#define m13 MAT(m,0,2)
#define m14 MAT(m,0,3)
#define m21 MAT(m,1,0)
#define m22 MAT(m,1,1)
#define m23 MAT(m,1,2)
#define m24 MAT(m,1,3)
#define m31 MAT(m,2,0)
#define m32 MAT(m,2,1)
#define m33 MAT(m,2,2)
#define m34 MAT(m,2,3)
#define m41 MAT(m,3,0)
#define m42 MAT(m,3,1)
#define m43 MAT(m,3,2)
#define m44 MAT(m,3,3)

   float det;
   float d12, d13, d23, d24, d34, d41;
   float tmp[16]; /* Allow out == in. */

   /* Inverse = adjoint / det. (See linear algebra texts.)*/

   /* pre-compute 2x2 dets for last two rows when computing */
   /* cofactors of first two rows. */
   d12 = (m31*m42-m41*m32);
   d13 = (m31*m43-m41*m33);
   d23 = (m32*m43-m42*m33);
   d24 = (m32*m44-m42*m34);
   d34 = (m33*m44-m43*m34);
   d41 = (m34*m41-m44*m31);

   tmp[0] =  (m22 * d34 - m23 * d24 + m24 * d23);
   tmp[1] = -(m21 * d34 + m23 * d41 + m24 * d13);
   tmp[2] =  (m21 * d24 + m22 * d41 + m24 * d12);
   tmp[3] = -(m21 * d23 - m22 * d13 + m23 * d12);

   /* Compute determinant as early as possible using these cofactors. */
   det = m11 * tmp[0] + m12 * tmp[1] + m13 * tmp[2] + m14 * tmp[3];

   /* Run singularity test. */
   if (det == 0.0)
   {
	   gIdentity(out);
   }
   else
   {
      float invDet = 1.0f / det;
      /* Compute rest of inverse. */
      tmp[0] *= invDet;
      tmp[1] *= invDet;
      tmp[2] *= invDet;
      tmp[3] *= invDet;

      tmp[4] = -(m12 * d34 - m13 * d24 + m14 * d23) * invDet;
      tmp[5] =  (m11 * d34 + m13 * d41 + m14 * d13) * invDet;
      tmp[6] = -(m11 * d24 + m12 * d41 + m14 * d12) * invDet;
      tmp[7] =  (m11 * d23 - m12 * d13 + m13 * d12) * invDet;

      /* Pre-compute 2x2 dets for first two rows when computing */
      /* cofactors of last two rows. */
      d12 = m11*m22-m21*m12;
      d13 = m11*m23-m21*m13;
      d23 = m12*m23-m22*m13;
      d24 = m12*m24-m22*m14;
      d34 = m13*m24-m23*m14;
      d41 = m14*m21-m24*m11;

      tmp[8] =  (m42 * d34 - m43 * d24 + m44 * d23) * invDet;
      tmp[9] = -(m41 * d34 + m43 * d41 + m44 * d13) * invDet;
      tmp[10] =  (m41 * d24 + m42 * d41 + m44 * d12) * invDet;
      tmp[11] = -(m41 * d23 - m42 * d13 + m43 * d12) * invDet;
      tmp[12] = -(m32 * d34 - m33 * d24 + m34 * d23) * invDet;
      tmp[13] =  (m31 * d34 + m33 * d41 + m34 * d13) * invDet;
      tmp[14] = -(m31 * d24 + m32 * d41 + m34 * d12) * invDet;
      tmp[15] =  (m31 * d23 - m32 * d13 + m33 * d12) * invDet;

      memcpy(out, tmp, 16*sizeof(float));
   }

#undef m11
#undef m12
#undef m13
#undef m14
#undef m21
#undef m22
#undef m23
#undef m24
#undef m31
#undef m32
#undef m33
#undef m34
#undef m41
#undef m42
#undef m43
#undef m44
#undef MAT
}

//////////////////////////////////////////////////////////////////////////////
static void GFgluMultMatricesf(const float a[16], const float b[16], float r[16])
{
    for (int i = 0; i < 4; i++) {
	    r[i*4+0] = 	a[i*4+0]*b[0*4+0] +	a[i*4+1]*b[1*4+0] +	a[i*4+2]*b[2*4+0] +	a[i*4+3]*b[3*4+0];
	    r[i*4+1] = 	a[i*4+0]*b[0*4+1] +	a[i*4+1]*b[1*4+1] +	a[i*4+2]*b[2*4+1] +	a[i*4+3]*b[3*4+1];
	    r[i*4+2] = 	a[i*4+0]*b[0*4+2] +	a[i*4+1]*b[1*4+2] +	a[i*4+2]*b[2*4+2] +	a[i*4+3]*b[3*4+2];
	    r[i*4+3] = 	a[i*4+0]*b[0*4+3] +	a[i*4+1]*b[1*4+3] +	a[i*4+2]*b[2*4+3] +	a[i*4+3]*b[3*4+3];
    }
}


//////////////////////////////////////////////////////////////////////////////
bool GFglu_Project(float objx, float objy, float objz,
					float* winx, float* winy, float* winz, const int viewport[4], const float proj[16], const float model[16])
{
#ifndef GLB_CONSOLE
	// static int viewport[4];
	static float /*proj[16], model[16],*/ in[4], out[4];

	// first load the matrices for the project call.
	// glGetIntegerv( GL_VIEWPORT,         viewport );
	// glGetFloatv  ( GL_PROJECTION_MATRIX, proj );
	// glGetFloatv  ( GL_MODELVIEW_MATRIX,  model);
	in[0]=objx; in[1]=objy; in[2]=objz; in[3]=1.0;

/*
#define dmat(mat) g_debug("[%s]=\n%5.1f %5.1f %5.1f %5.1f\n%5.1f %5.1f %5.1f %5.1f\n%5.1f %5.1f %5.1f %5.1f\n%5.1f %5.1f %5.1f %5.1f\n\n", #mat, \
	mat[0], mat[1], mat[2], mat[3], mat[4], mat[5], mat[6], mat[7], mat[8], mat[9], mat[10], mat[11], mat[12], mat[13], mat[14], mat[15]);
#define dvec(mat) g_debug("[%s]=\n%5.1f %5.1f %5.1f %5.1f\n\n", #mat, \
	mat[0], mat[1], mat[2], mat[3] );
	dvec(in);
	dmat(model);
*/
	transform_point(out,model,in);
//	dvec(out);

//	dmat(proj);
	transform_point(in,proj,out);
//	dvec(in);

	if (in[3]==0.0F) return false;

	// Normalize between -1 and 1 :
	float one_over_in3 = 1.0f / in[3];
	in[0]*=one_over_in3;in[1]*=one_over_in3;in[2]*=one_over_in3;

//	printf("Viewp: %d %d %d %d\n", viewport[0], viewport[1],viewport[2], viewport[3]);

	// In screen coordinates :
	*winx = viewport[0]+(1.0f+in[0])*viewport[2]/2.0f;
	*winy = viewport[3] - (viewport[1]+(1.0f+in[1])*viewport[3] / 2.0f); // GLBasic origin = top, (+) goes down


/*
int viewport_real[4];
glGetIntegerv( GL_VIEWPORT,         viewport_real );
	*winx = viewport_real[0]+(1+in[0])*viewport_real[2]/2;
	*winy = viewport_real[3] - (viewport_real[1]+(1+in[1])*viewport_real[3]/2); // GLBasic origin = top, (+) goes down
*/

/* en coordonnees ecran */
// *winx = viewport[0] + (1 + in[0]) * viewport[2] / 2;
// *winy = viewport[1] + (1 + in[1]) * viewport[3] / 2;

	// In z [-1,1] -> [0,1] :
	*winz = (1.0f+in[2]) / 2.0f;
	return true;
#else
	return false;
#endif
}

//////////////////////////////////////////////////////////////////////////////
bool GFglu_UnProject(float winx, float winy, float winz, float *objx, float *objy, float *objz, const int viewport[4], const float proj[16], const float model[16])
{
#ifndef GLB_CONSOLE
//		static int viewport[4];
	static float /*proj[16], model[16],*/ finalMatrix[16], in[4], out[4];

	// first load the matrices for the project call.
	// glGetIntegerv( GL_VIEWPORT,         viewport );
	// glGetFloatv  ( GL_PROJECTION_MATRIX, (GLfloat*)proj );
	// glGetFloatv  ( GL_MODELVIEW_MATRIX,  (GLfloat*)model);

	GFgluMultMatricesf(model, proj, finalMatrix);
	//  gInvertMatrix4x4(finalMatrix);
	invertMatrix(finalMatrix, finalMatrix); // might be non-symetrical!
	in[0]=winx;
	in[1]=winy;
	in[2]=winz;
	in[3]=1.0;

	/* Map x and y from window coordinates */
	in[0] = (in[0] - viewport[0]) / viewport[2];
	in[1] = (in[1] - viewport[1]) / viewport[3]; // flip y in GLBasic

	/* Map to range -1 to 1 */
	in[0] = in[0] * 2 - 1;
	in[1] = in[1] * 2 - 1;
	in[2] = in[2] * 2 - 1;

	transform_point(out, finalMatrix, in);
	if (out[3] == 0.0) return false;
	out[0] /= out[3];
	out[1] /= out[3];
	out[2] /= out[3];
	*objx = out[0];
	*objy = out[1];
	*objz = out[2];
	return true;
#else
	return false;
#endif
}

// intersect_RayTriangle(): intersect a ray with a 3D triangle
//    Input:  a ray R, and a triangle T
//    Output: *I = intersection point (when it exists)
//    Return: 0: No intersection, value = factor of distance R0->R1
float IntersectTriangleAndRay(const float tri[3][3], const float R0[3], const float R1[3])
{
    float    u[3], v[3], n[3], I[3];             // triangle vectors
    float    dir[3], w0[3], w[3];          // ray vectors
    float    r, a, b;             // params to calc ray-plane intersect

    // get triangle edge vectors and plane normal
    gDiffVertex(u, tri[1], tri[0]);
    // u = T.V1 - T.V0;
    gDiffVertex(v, tri[2], tri[0]);
    // v = T.V2 - T.V0;
    gCross(u, v, n);
    //n = u * v;             // cross product
    if (n[0]==0 && n[1]==0 && n[2]==0)       // triangle is degenerate
        return 0;                 // do not deal with this case

    gDiffVertex(dir, R1, R0);
    //dir = R.P1 - R.P0;             // ray direction vector
    gDiffVertex(w0, R0, tri[0]);
    //w0 = R.P0 - T.V0;
    a = -gDot(n, w0);
    b = gDot(n,dir);
	// is ray parallel to triangle plane
    if (fabs(b) == 0) return 0;

    // get intersect point of ray with triangle plane
    r = a / b;
    // if r e[0; 1], point is on line segment

	gCopyVertex(I, R0); gMultVertex(dir, r); gAddVertex(I, dir);
    // *I = R.P0 + r * dir;           // intersect point of ray and plane

    // is I inside T?
    float    uu, uv, vv, wu, wv, D;
    uu = gDot(u,u);
    uv = gDot(u,v);
    vv = gDot(v,v);
    gDiffVertex(w, I, tri[0]);
    // w = *I - T.V0;
    wu = gDot(w,u);
    wv = gDot(w,v);
    D = uv * uv - uu * vv;

    // get and test parametric coords (a=s, b=t)
    a = (uv * wv - vv * wu) / D;
    if (a < 0.0 || a > 1.0)        // I is outside T
        return 0;
    b = (uv * wu - uu * wv) / D;
    if (b < 0.0 || (a + b) > 1.0)  // I is outside T
        return 0;

    return r;                      // I is in T
}


//----------------------------------------------------------------------------
bool TriangleAndCone(float* rkConeAxis, float* rkConeVertex, float rkConeCosAngle, float rkTriangleV[][3])
{
    // triangle is <P0,P1,P2>, edges are E0 = P1-P0, E1=P2-P0
    int iOnConeSide = 0;
    float fP0Test = (float)0.0, fP1Test = (float)0.0, fP2Test = (float)0.0;
    float fAdE, fEdE, fEdD, fC1, fC2;



    float fCosSqr = rkConeCosAngle*rkConeCosAngle;

    // test vertex P0
    float kDiff0[3];
    gDiffVertex(kDiff0, rkTriangleV[0], rkConeVertex);
    float fAdD0 = gDot(rkConeAxis, kDiff0);
    if (fAdD0 >= 0.0f)
    {
        // P0 is on cone side of plane
        fP0Test = fAdD0*fAdD0 - fCosSqr * gDot(kDiff0, kDiff0);
        if (fP0Test >= 0.0)
        {
            // P0 is inside the cone
            return true;
        }
        else
        {
            // P0 is outside the cone, but on cone side of plane
            iOnConeSide |= 1;
        }
    }
    // else P0 is not on cone side of plane

    // test vertex P1
    float kEdge0[3]; gDiffVertex(kEdge0, rkTriangleV[1], rkTriangleV[0]);
    float kDiff1[3]; gCopyVertex(kDiff1, kDiff0); gAddVertex(kDiff1, kEdge0);
    float fAdD1 = gDot(rkConeAxis, kDiff1);
    if (fAdD1 >= 0.0)
    {
        // P1 is on cone side of plane
        fP1Test = fAdD1*fAdD1 - fCosSqr* gDot(kDiff1, kDiff1);
        if (fP1Test >= 0.0)
        {
            // P1 is inside the cone
            return true;
        }
        else
        {
            // P1 is outside the cone, but on cone side of plane
            iOnConeSide |= 2;
        }
    }
    // else P1 is not on cone side of plane

    // test vertex P2
    float kEdge1[3]; gDiffVertex(kEdge1, rkTriangleV[2], rkTriangleV[0]);
    float kDiff2[3]; gCopyVertex(kDiff2, kDiff0); gAddVertex(kDiff2, kEdge1);
    float fAdD2 = gDot(rkConeAxis, kDiff2);
    if (fAdD2 >= 0.0)
    {
        // P2 is on cone side of plane
        fP2Test = fAdD2*fAdD2 - fCosSqr* gDot(kDiff2, kDiff2);
        if (fP2Test >= 0.0)
        {
            // P2 is inside the cone
            return true;
        }
        else
        {
            // P2 is outside the cone, but on cone side of plane
            iOnConeSide |= 4;
        }
    }
    // else P2 is not on cone side of plane

    // test edge <P0,P1> = E0
    if (iOnConeSide & 3)
    {
        fAdE = fAdD1 - fAdD0;
        fEdE = gDot(kEdge0, kEdge0);
        fC2 = fAdE*fAdE - fCosSqr*fEdE;
        if (fC2 < (float)0.0)
        {
            fEdD = gDot(kEdge0, kDiff0);
            fC1 = fAdE*fAdD0 - fCosSqr*fEdD;
            if (iOnConeSide & 1)
            {
                if (iOnConeSide & 2)
                {
                    // <P0,P1> fully on cone side of plane, fC0 = fP0Test
                    if ((float)0.0 <= fC1 && fC1 <= -fC2
                    &&  fC1*fC1 >= fP0Test*fC2)
                    {
                        return true;
                    }
                }
                else
                {
                    // P0 on cone side (Dot(A,P0-V) >= 0),
                    // P1 on opposite side (Dot(A,P1-V) <= 0)
                    // (Dot(A,E0) <= 0), fC0 = fP0Test
                    if ((float)0.0 <= fC1 && fC2*fAdD0 <= fC1*fAdE
                    &&  fC1*fC1 >= fP0Test*fC2)
                    {
                        return true;
                    }
                }
            }
            else
            {
                // P1 on cone side (Dot(A,P1-V) >= 0),
                // P0 on opposite side (Dot(A,P0-V) <= 0)
                // (Dot(A,E0) >= 0), fC0 = fP0Test (needs calculating)
                if (fC1 <= -fC2 && fC2*fAdD0 <= fC1*fAdE)
                {
                    fP0Test = fAdD0*fAdD0 - fCosSqr*gDot(kDiff0, kDiff0);
                    if (fC1*fC1 >= fP0Test*fC2)
                    {
                        return true;
                    }
                }
            }
        }
    }
    // else <P0,P1> does not intersect cone half space

    // test edge <P0,P2> = E1
    if (iOnConeSide & 5)
    {
        fAdE = fAdD2 - fAdD0;
        fEdE = gDot(kEdge1, kEdge1);
        fC2 = fAdE*fAdE - fCosSqr*fEdE;
        if (fC2 < 0.0f)
        {
            fEdD = gDot(kEdge1, kDiff0);
            fC1 = fAdE*fAdD0 - fCosSqr*fEdD;
            if (iOnConeSide & 1)
            {
                if (iOnConeSide & 4)
                {
                    // <P0,P2> fully on cone side of plane, fC0 = fP0Test
                    if (0.0f <= fC1 && fC1 <= -fC2
                    &&  fC1*fC1 >= fP0Test*fC2)
                    {
                        return true;
                    }
                }
                else
                {
                    // P0 on cone side (Dot(A,P0-V) >= 0),
                    // P2 on opposite side (Dot(A,P2-V) <= 0)
                    // (Dot(A,E1) <= 0), fC0 = fP0Test
                    if (0.0f <= fC1 && fC2*fAdD0 <= fC1*fAdE
                    &&  fC1*fC1 >= fP0Test*fC2)
                    {
                        return true;
                    }
                }
            }
            else
            {
                // P2 on cone side (Dot(A,P2-V) >= 0),
                // P0 on opposite side (Dot(A,P0-V) <= 0)
                // (Dot(A,E1) >= 0), fC0 = fP0Test (needs calculating)
                if (fC1 <= -fC2 && fC2*fAdD0 <= fC1*fAdE)
                {
                    fP0Test = fAdD0*fAdD0 - fCosSqr*gDot(kDiff0, kDiff0);
                    if (fC1*fC1 >= fP0Test*fC2)
                    {
                        return true;
                    }
                }
            }
        }
    }
    // else <P0,P2> does not intersect cone half space

    // test edge <P1,P2> = E1-E0 = E2
    if (iOnConeSide & 6)
    {
        float kE2[3]; gDiffVertex(kE2, kEdge1, kEdge0);
        fAdE = fAdD2 - fAdD1;
        fEdE = gDot(kE2, kE2);
        fC2 = fAdE*fAdE - fCosSqr*fEdE;
        if (fC2 < 0.0f)
        {
            fEdD = gDot(kE2, kDiff1);
            fC1 = fAdE*fAdD1 - fCosSqr*fEdD;
            if (iOnConeSide & 2)
            {
                if (iOnConeSide & 4)
                {
                    // <P1,P2> fully on cone side of plane, fC0 = fP1Test
                    if (0.0f <= fC1 && fC1 <= -fC2
                    &&  fC1*fC1 >= fP1Test*fC2)
                    {
                        return true;
                    }
                }
                else
                {
                    // P1 on cone side (Dot(A,P1-V) >= 0),
                    // P2 on opposite side (Dot(A,P2-V) <= 0)
                    // (Dot(A,E2) <= 0), fC0 = fP1Test
                    if (0.0f <= fC1 && fC2*fAdD1 <= fC1*fAdE
                    &&  fC1*fC1 >= fP1Test*fC2)
                    {
                        return true;
                    }
                }
            }
            else
            {
                // P2 on cone side (Dot(A,P2-V) >= 0),
                // P1 on opposite side (Dot(A,P1-V) <= 0)
                // (Dot(A,E2) >= 0), fC0 = fP1Test (needs calculating)
                if (fC1 <= -fC2 && fC2*fAdD1 <= fC1*fAdE)
                {
                    fP1Test = fAdD1*fAdD1 - fCosSqr*gDot(kDiff1, kDiff1);
                    if (fC1*fC1 >= fP1Test*fC2)
                    {
                        return true;
                    }
                }
            }
        }
    }
    // else <P1,P2> does not intersect cone half space

    // Test triangle <P0,P1,P2>.  It is enough to handle only the case when
    // at least one Pi is on the cone side of the plane.  In this case and
    // after the previous testing, if the triangle intersects the cone, the
    // set of intersection must contain the point of intersection between
    // the cone axis and the triangle.
    float tmp[3];
    if (iOnConeSide > 0)
    {
        float kN[3];  gCross(kEdge0, kEdge1, kN);
        float fNdA    = gDot(kN, rkConeAxis);
        float fNdD    = gDot(kN, kDiff0);
        float kU[3]; gCopyVertex(kU, rkConeAxis); gMultVertex(kU, fNdD);
					 gCopyVertex(tmp, kDiff0); gMultVertex(tmp, fNdA); gSubVertex(kU, tmp);
        float kNcU[3]; gCross(kN, kU, kNcU);

        float fNcUdE0 = gDot(kNcU, kEdge0), fNcUdE1, fNcUdE2, fNdN;
        if (fNdA >= 0.0f)
        {
            if (fNcUdE0 <= 0.0f)
            {
                fNcUdE1 = gDot(kNcU, kEdge1);
                if (fNcUdE1 >= 0.0f)
                {
                    fNcUdE2 = fNcUdE1 - fNcUdE0;
                    fNdN = gGetSQLength(kN);
                    if (fNcUdE2 <= fNdA*fNdN)
                    {
                        return true;
                    }
                }
            }
        }
        else
        {
            if (fNcUdE0 >= 0.0f)
            {
                fNcUdE1 = gDot(kNcU, kEdge1);
                if (fNcUdE1 <= 0.0f)
                {
                    fNcUdE2 = fNcUdE1 - fNcUdE0;
                    fNdN = gGetSQLength(kN);
                    if (fNcUdE2 >= fNdA*fNdN)
                    {
                        return true;
                    }
                }
            }
        }
    }

    return false;
}


// Collision between a solid sphere and a solid AABB
// The bx,by,bz params are the (half) width of each axis
bool IntersectSphereBox(float* sphere, float radius_sq, float bx, float by, float bz)
{
float  dmin=0;
	#define SQR(a) ((a)*(a))
	// Solid Box - Solid Sphere
	float* pC = sphere;
	// X
	if     ( *pC <-bx ) dmin += SQR( *pC + bx );
	else if( *pC > bx ) dmin += SQR( *pC - bx );
	++pC;
	// Y
	if     ( *pC <-by ) dmin += SQR( *pC + by );
	else if( *pC > by ) dmin += SQR( *pC - by );
	++pC;
	// Z
	if     ( *pC <-bz ) dmin += SQR( *pC + bz );
	else if( *pC > bz ) dmin += SQR( *pC - bz );
	#undef SQR

	if( dmin <= radius_sq ) return true;
return false;
} 


