#ifndef _3DMATH_H
#define _3DMATH_H

#include "vectormath.h"


#ifndef PI
#define PI   3.14159265358979323
#endif
#ifndef PI_F
#define PI_F 3.14159265358979323f
#endif

#ifndef BOG
#define BOG 57.29577951308232088 // Bogenma*BOG=Gradma
#endif
#ifndef BOG_F
#define BOG_F 57.29577951308232088f
#endif

#ifndef BOG_INV
#define BOG_INV (1.0 / BOG )
#endif

#ifndef BOG_INV_F
#define BOG_INV_F (1.0f / BOG_F )
#endif

bool SpherePolygonCollision(float vPolygon[3][3], float* vCenter, int vertexCount, float radius);


bool GFglu_UnProject(float winx, float winy, float winz, float *objx, float *objy, float *objz, const int viewport[4], const float proj[16], const float model[16]);
bool GFglu_Project(float objx, float objy, float objz,
					float* winx, float* winy, float* winz, const int viewport[4], const float proj[16], const float model[16]);


/////////////////////////////////// DISTANCE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
/////	This returns the distance between 2 3D points
#define Distance(a, b) sqrtf(DistanceSQ(a, b))
#define DistanceSQ(vPoint1, vPoint2)  \
						(	(vPoint2[0] - vPoint1[0]) * (vPoint2[0] - vPoint1[0]) +  \
						    (vPoint2[1] - vPoint1[1]) * (vPoint2[1] - vPoint1[1]) +  \
						    (vPoint2[2] - vPoint1[2]) * (vPoint2[2] - vPoint1[2]) )

///////////////////////////// POINT IN SPHERE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
/////	Is a point inside a sphere
inline int IsPointInSphere(float* pt, float* sphere, float radius_sq)
{
	return DistanceSQ(pt, sphere)<=radius_sq;
}
///////////////////////////// CUBOID AND TRIAGNLE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
/////	Does a cuboid (parallelepiped) intersect with a triangle
bool IntersectCuboidAndTriangle(					const float* Org, const float* A0, const float* A1, const float* A2,
													const float* V0, const float* V1, const float* V2);


///////////////////////////// SPHERE AND AABB \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
/////	Does a sphere intersect with an AABB
bool IntersectSphereBox(float* sphere, float radius_sq, float bx, float by, float bz);


///////////////////////////// RAY AND TRIAGNLE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
/////	Does a ray intersect with a triangle - returns 0 if not, otherwise
/////   factor of lenght l1->l2
float IntersectTriangleAndRay(const float tri[3][3], const float l1[3], const float l2[3]);


///////////////////////////// TRIAGNLE AND CONE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
/////	Does a cone intersect with a triangle - returns true/false
bool TriangleAndCone(float* rkConeAxis, float* rkConeVertex, float rkConeCosAngle, float rkTriangleV[][3]);


// From Mesa/src-glu/glu.c/transform_point.
/*
 * Transform a point (column vector) by a 4x4 matrix.  I.e.  out = m * in
 * Input:  m - the 4x4 matrix
 *         in - the 4x1 vector
 * Output:  out - the resulting 4x1 vector.
 */
//////////////////////////////////////////////////////////////////////////////
template <class T> inline void transform_point(T out[4],const T m[16],const T in[4])
{
#define M(row,col)  m[col*4+row]
  out[0] = M(0,0) * in[0] + M(0,1) * in[1] + M(0,2) * in[2] + M(0,3) * in[3];
  out[1] = M(1,0) * in[0] + M(1,1) * in[1] + M(1,2) * in[2] + M(1,3) * in[3];
  out[2] = M(2,0) * in[0] + M(2,1) * in[1] + M(2,2) * in[2] + M(2,3) * in[3];
  out[3] = M(3,0) * in[0] + M(3,1) * in[1] + M(3,2) * in[2] + M(3,3) * in[3];
#undef M
}




// From Mesa/src-mesa/m_matrix.c
/**
 * Perform a full 4x4 matrix multiplication.
 *
 * \param a matrix.
 * \param b matrix.
 * \param product will receive the product of \p a and \p b.
 *
 * \warning Is assumed that \p product != \p b. \p product == \p a is allowed.
 *
 * \note KW: 4*16 = 64 multiplications
 * 
 * \author This \c matmul was contributed by Thomas Malik
 */
template<class T> inline void matmul4( T *product, const T *a, const T *b )
{
	#define A(row,col)  a[(col<<2)+row]
	#define B(row,col)  b[(col<<2)+row]
	#define P(row,col)  product[(col<<2)+row]
	int i;
	for (i = 0; i < 4; i++)
	{
		const float ai0=A(i,0),  ai1=A(i,1),  ai2=A(i,2),  ai3=A(i,3);
		P(i,0) = ai0 * B(0,0) + ai1 * B(1,0) + ai2 * B(2,0) + ai3 * B(3,0);
		P(i,1) = ai0 * B(0,1) + ai1 * B(1,1) + ai2 * B(2,1) + ai3 * B(3,1);
		P(i,2) = ai0 * B(0,2) + ai1 * B(1,2) + ai2 * B(2,2) + ai3 * B(3,2);
		P(i,3) = ai0 * B(0,3) + ai1 * B(1,3) + ai2 * B(2,3) + ai3 * B(3,3);
	}
	#undef A
	#undef B
	#undef P
}



/**********************************************************************/
/** \name Matrix generation */
/*@{*/

/**
 * Generate a 4x4 transformation matrix from glRotate parameters, and
 * post-multiply the input matrix by it.
 *
 * \author
 * This function was contributed by Erich Boleyn (erich@uruk.org).
 * Optimizations contributed by Rudolf Opalla (rudi@khm.de).
 */
static void
matrix_rotate( float *mat,
		     float angle, float x, float y, float z )
{
   float xx, yy, zz, xy, yz, zx, xs, ys, zs, one_c, s, c;
   float m[16];
   int optimized;

   s = (float) sin( angle * BOG_INV_F );
   c = (float) cos( angle * BOG_INV_F );

	gIdentity(m);
   optimized = 0;

#define M(row,col)  m[col*4+row]

   if (x == 0.0F) {
      if (y == 0.0F) {
         if (z != 0.0F) {
            optimized = 1;
            /* rotate only around z-axis */
            M(0,0) = c;
            M(1,1) = c;
            if (z < 0.0F) {
               M(0,1) = s;
               M(1,0) = -s;
            }
            else {
               M(0,1) = -s;
               M(1,0) = s;
            }
         }
      }
      else if (z == 0.0F) {
         optimized = 1;
         /* rotate only around y-axis */
         M(0,0) = c;
         M(2,2) = c;
         if (y < 0.0F) {
            M(0,2) = -s;
            M(2,0) = s;
         }
         else {
            M(0,2) = s;
            M(2,0) = -s;
         }
      }
   }
   else if (y == 0.0F) {
      if (z == 0.0F) {
         optimized = 1;
         /* rotate only around x-axis */
         M(1,1) = c;
         M(2,2) = c;
         if (x < 0.0F) {
            M(1,2) = s;
            M(2,1) = -s;
         }
         else {
            M(1,2) = -s;
            M(2,1) = s;
         }
      }
   }

   if (!optimized) {
      const float mag = sqrtf(x * x + y * y + z * z);

      if (mag <= 1.0e-4) {
         /* no rotation, leave mat as-is */
         return;
      }

      x /= mag;
      y /= mag;
      z /= mag;


      /*
       *     Arbitrary axis rotation matrix.
       *
       *  This is composed of 5 matrices, Rz, Ry, T, Ry', Rz', multiplied
       *  like so:  Rz * Ry * T * Ry' * Rz'.  T is the final rotation
       *  (which is about the X-axis), and the two composite transforms
       *  Ry' * Rz' and Rz * Ry are (respectively) the rotations necessary
       *  from the arbitrary axis to the X-axis then back.  They are
       *  all elementary rotations.
       *
       *  Rz' is a rotation about the Z-axis, to bring the axis vector
       *  into the x-z plane.  Then Ry' is applied, rotating about the
       *  Y-axis to bring the axis vector parallel with the X-axis.  The
       *  rotation about the X-axis is then performed.  Ry and Rz are
       *  simply the respective inverse transforms to bring the arbitrary
       *  axis back to it's original orientation.  The first transforms
       *  Rz' and Ry' are considered inverses, since the data from the
       *  arbitrary axis gives you info on how to get to it, not how
       *  to get away from it, and an inverse must be applied.
       *
       *  The basic calculation used is to recognize that the arbitrary
       *  axis vector (x, y, z), since it is of unit length, actually
       *  represents the sines and cosines of the angles to rotate the
       *  X-axis to the same orientation, with theta being the angle about
       *  Z and phi the angle about Y (in the order described above)
       *  as follows:
       *
       *  cos ( theta ) = x / sqrt ( 1 - z^2 )
       *  sin ( theta ) = y / sqrt ( 1 - z^2 )
       *
       *  cos ( phi ) = sqrt ( 1 - z^2 )
       *  sin ( phi ) = z
       *
       *  Note that cos ( phi ) can further be inserted to the above
       *  formulas:
       *
       *  cos ( theta ) = x / cos ( phi )
       *  sin ( theta ) = y / sin ( phi )
       *
       *  ...etc.  Because of those relations and the standard trigonometric
       *  relations, it is pssible to reduce the transforms down to what
       *  is used below.  It may be that any primary axis chosen will give the
       *  same results (modulo a sign convention) using thie method.
       *
       *  Particularly nice is to notice that all divisions that might
       *  have caused trouble when parallel to certain planes or
       *  axis go away with care paid to reducing the expressions.
       *  After checking, it does perform correctly under all cases, since
       *  in all the cases of division where the denominator would have
       *  been zero, the numerator would have been zero as well, giving
       *  the expected result.
       */

      xx = x * x;
      yy = y * y;
      zz = z * z;
      xy = x * y;
      yz = y * z;
      zx = z * x;
      xs = x * s;
      ys = y * s;
      zs = z * s;
      one_c = 1.0f - c;

      /* We already hold the identity-matrix so we can skip some statements */
      M(0,0) = (one_c * xx) + c;
      M(0,1) = (one_c * xy) - zs;
      M(0,2) = (one_c * zx) + ys;
/*    M(0,3) = 0.0F; */

      M(1,0) = (one_c * xy) + zs;
      M(1,1) = (one_c * yy) + c;
      M(1,2) = (one_c * yz) - xs;
/*    M(1,3) = 0.0F; */

      M(2,0) = (one_c * zx) - ys;
      M(2,1) = (one_c * yz) + xs;
      M(2,2) = (one_c * zz) + c;
/*    M(2,3) = 0.0F; */

/*
      M(3,0) = 0.0F;
      M(3,1) = 0.0F;
      M(3,2) = 0.0F;
      M(3,3) = 1.0F;
*/
   }
#undef M

   matmul4( mat, mat, m);
}


#endif 

