// blowfish.cpp   C++ class implementation of the BLOWFISH encryption algorithm
// _THE BLOWFISH ENCRYPTION ALGORITHM_
// by Bruce Schneier
// Revised code--3/20/94
// Converted to C++ class 5/96, Jim Conger

#include "stdafx.h"
#include "blowfish.h"
#include "blowfish.h2.h" // Random numbers

// ------------------------------------------------------------- //


// ------------------------------------------------------------- //
CBlowFish::CBlowFish ()
// ------------------------------------------------------------- //
{
 	PArray = new DWORD [18] ;
 	SBoxes = new DWORD [4][256] ;
}

// ------------------------------------------------------------- //
CBlowFish::~CBlowFish ()
// ------------------------------------------------------------- //
{
	delete PArray ;
	delete [] SBoxes ;
}

// the low level (private) encryption function
// ------------------------------------------------------------- //
void CBlowFish::Blowfish_encipher (DWORD *xl, DWORD *xr)
// ------------------------------------------------------------- //
{
	union aword  Xl, Xr ;

	Xl.dword = *xl ;
	Xr.dword = *xr ;

	Xl.dword ^= PArray [0];
	Round (Xr, Xl, 1) ;  Round (Xl, Xr, 2) ;
	Round (Xr, Xl, 3) ;  Round (Xl, Xr, 4) ;
	Round (Xr, Xl, 5) ;  Round (Xl, Xr, 6) ;
	Round (Xr, Xl, 7) ;  Round (Xl, Xr, 8) ;
	Round (Xr, Xl, 9) ;  Round (Xl, Xr, 10) ;
	Round (Xr, Xl, 11) ; Round (Xl, Xr, 12) ;
	Round (Xr, Xl, 13) ; Round (Xl, Xr, 14) ;
	Round (Xr, Xl, 15) ; Round (Xl, Xr, 16) ;
	Xr.dword ^= PArray [17] ;

	*xr = Xl.dword ;
	*xl = Xr.dword ;
}

// the low level (private) decryption function
// ------------------------------------------------------------- //
void CBlowFish::Blowfish_decipher (DWORD *xl, DWORD *xr)
// ------------------------------------------------------------- //
{
   union aword  Xl ;
   union aword  Xr ;

   Xl.dword = *xl ;
   Xr.dword = *xr ;

   Xl.dword ^= PArray [17] ;
   Round (Xr, Xl, 16) ;  Round (Xl, Xr, 15) ;
   Round (Xr, Xl, 14) ;  Round (Xl, Xr, 13) ;
   Round (Xr, Xl, 12) ;  Round (Xl, Xr, 11) ;
   Round (Xr, Xl, 10) ;  Round (Xl, Xr, 9) ;
   Round (Xr, Xl, 8) ;   Round (Xl, Xr, 7) ;
   Round (Xr, Xl, 6) ;   Round (Xl, Xr, 5) ;
   Round (Xr, Xl, 4) ;   Round (Xl, Xr, 3) ;
   Round (Xr, Xl, 2) ;   Round (Xl, Xr, 1) ;
   Xr.dword ^= PArray[0];

   *xl = Xr.dword;
   *xr = Xl.dword;
}


	// constructs the enctryption sieve
// ------------------------------------------------------------- //
void CBlowFish::Initialize (const BYTE key[], int keybytes)
// ------------------------------------------------------------- //
{
	int  		i, j ;
	DWORD  		data, datal, datar ;
	union aword temp ;

	// first fill arrays from data tables
	for (i = 0 ; i < 18 ; i++)
	{
		PArray [i] = bf_P [i] ;
	}

	for (i = 0 ; i < 4 ; i++)
	{
	 	for (j = 0 ; j < 256 ; j++)
	 		SBoxes [i][j] = bf_S [i][j] ;
	}


	j = 0 ;
	for (i = 0 ; i < NPASS + 2 ; ++i)
	{
		temp.dword = 0 ;
		temp.w.byte0 = key[j];
		temp.w.byte1 = key[(j+1) % keybytes] ;
		temp.w.byte2 = key[(j+2) % keybytes] ;
		temp.w.byte3 = key[(j+3) % keybytes] ;
		data = temp.dword ;
		PArray [i] ^= data ;
		j = (j + 4) % keybytes ;
	}

	datal = 0 ;
	datar = 0 ;

	for (i = 0 ; i < NPASS + 2 ; i += 2)
	{
		Blowfish_encipher (&datal, &datar) ;
		PArray [i] = datal ;
		PArray [i + 1] = datar ;
	}

	for (i = 0 ; i < 4 ; ++i)
	{
		for (j = 0 ; j < 256 ; j += 2)
		{
		  Blowfish_encipher (&datal, &datar) ;
		  SBoxes [i][j] = datal ;
		  SBoxes [i][j + 1] = datar ;
		}
	}
}

	// get output length, which must be even MOD 8
// ------------------------------------------------------------- //
DWORD CBlowFish::GetOutputLength (DWORD lInputLong)
// ------------------------------------------------------------- //
{
	DWORD 	lVal ;

	lVal = lInputLong % 8 ;	// find out if uneven number of bytes at the end
	if (lVal != 0)
		return lInputLong + 8 - lVal ;
	else
		return lInputLong ;
}

	// Encode pIntput into pOutput.  Input length in lSize.  Returned value
	// is length of output which will be even MOD 8 bytes.  Input buffer and
	// output buffer can be the same, but be sure buffer length is even MOD 8.
#ifdef BLOWFISH_ALLOW_ENCODE
// ------------------------------------------------------------- //
DWORD CBlowFish::Encode (BYTE * pInput, BYTE * pOutput, DWORD lSize)
// ------------------------------------------------------------- //
{
DWORD 	lCount, lOutSize, lGoodBytes ;
BYTE	*pi, *po ;
int		i, j ;
int		SameDest = (pInput == pOutput ? 1 : 0) ;

	lOutSize = GetOutputLength (lSize) ;
	for (lCount = 0 ; lCount < lOutSize ; lCount += 8)
	{
		if (SameDest)	// if encoded data is being written into input buffer
		{
		 	if (lCount < lSize - 7)	// if not dealing with uneven bytes at end
		 	{
		 	 	Blowfish_encipher ((DWORD *) pInput,
		 	 		(DWORD *) (pInput + 4)) ;
		 	}
		 	else		// pad end of data with null bytes to complete encryption
		 	{
				po = pInput + lSize ;	// point at byte past the end of actual data
				j = (int) (lOutSize - lSize) ;	// number of bytes to set to null
				for (i = 0 ; i < j ; i++)
					*po++ = 0 ;
		 	 	Blowfish_encipher ((DWORD *) pInput,
		 	 		(DWORD *) (pInput + 4)) ;
		 	}
		 	pInput += 8 ;
		}
		else 			// output buffer not equal to input buffer, so must copy
		{               // input to output buffer prior to encrypting
		 	if (lCount < lSize - 7)	// if not dealing with uneven bytes at end
		 	{
		 		pi = pInput ;
		 		po = pOutput ;
		 		for (i = 0 ; i < 8 ; i++) // copy bytes to output
		 			*po++ = *pi++ ;
		 	 	Blowfish_encipher ((DWORD *) pOutput,	// now encrypt them
		 	 		(DWORD *) (pOutput + 4)) ;
		 	}
		 	else		// pad end of data with null bytes to complete encryption
		 	{
		 		lGoodBytes = lSize - lCount ;	// number of remaining data bytes
		 		po = pOutput ;
		 		for (i = 0 ; i < (int) lGoodBytes ; i++)
		 			*po++ = *pInput++ ;
		 		for (j = i ; j < 8 ; j++)
		 			*po++ = 0 ;
		 	 	Blowfish_encipher ((DWORD *) pOutput,
		 	 		(DWORD *) (pOutput + 4)) ;
		 	}
		 	pInput += 8 ;
		 	pOutput += 8 ;
		}
	}
	return lOutSize ;
}
#endif //#ifdef BLOWFISH_ALLOW_ENCODE


	// Decode pIntput into pOutput.  Input length in lSize.  Input buffer and
	// output buffer can be the same, but be sure buffer length is even MOD 8.
// ------------------------------------------------------------- //
void CBlowFish::Decode (BYTE * pInput, BYTE * pOutput, DWORD lSize)
// ------------------------------------------------------------- //
{
	DWORD 	lCount ;
	BYTE	*pi, *po ;
	int		i ;
	int		SameDest = (pInput == pOutput ? 1 : 0) ;

	for (lCount = 0 ; lCount < lSize ; lCount += 8)
	{
		if (SameDest)	// if encoded data is being written into input buffer
		{
	 	 	Blowfish_decipher ((DWORD *) pInput,
	 	 		(DWORD *) (pInput + 4)) ;
		 	pInput += 8 ;
		}
		else 			// output buffer not equal to input buffer
		{               // so copy input to output before decoding
	 		pi = pInput ;
	 		po = pOutput ;
	 		for (i = 0 ; i < 8 ; i++)
	 			*po++ = *pi++ ;
	 	 	Blowfish_decipher ((DWORD *) pOutput,
	 	 		(DWORD *) (pOutput + 4)) ;
		 	pInput += 8 ;
		 	pOutput += 8 ;
		}
	}
}

#ifdef BLOWFISH_ALLOW_ENCODE
#ifdef _AFX
// ------------------------------------------------------------- //
CString CBlowFish::EasyEncode(LPCTSTR mail)
// ------------------------------------------------------------- //
{
int len;
CString	as, pw;
int		brc=0;
	len=strlen(mail);
	len+=(len%8);
BYTE* text=new BYTE[len*4+64];
BYTE code[32];

	memset(text, 0, len*4+64);
	memcpy(text, mail, strlen(mail));
	memcpy(text, mail, strlen(mail));

	memset(code, 0, 32);
	memcpy(code, "d9~R.E+", 9);

	this->Initialize(code, 9);
	len=this->Encode(text, text, len);

	pw="";
	for (int i=0; i<len; i++)
	{
		as.Format("%.2X", text[i]);
		pw+=as;
		brc++;
		if(brc==10)
		{
			brc=0; pw+="\r\n";
		}
	}

delete [] text;
return pw;
}
#endif
#endif // #ifdef BLOWFISH_ALLOW_ENCODE



#ifdef _AFX
// ------------------------------------------------------------- //
CString CBlowFish::EasyDecode(LPCTSTR BFcode)
// ------------------------------------------------------------- //
{
int len;
CString	as, pw;
int		brc=0, i, j, n;

	as=BFcode;
	as.MakeUpper();
	as.Replace(" ", "");
	as.Replace("-", "");
	as.Replace("\n", "");
	as.Replace("\r", "");
	as.Replace("\t", "");
	len=as.GetLength()/2;
	len+=(len%8);
BYTE* text=new BYTE[len*4+64];
BYTE* code=new BYTE[32];

char	hex[17]="0123456789ABCDEF";

	memset(text, 0, len*4+64);
	for(i=0; i<as.GetLength()/2; i++)
	{
		n=0;
		for (j=0; j<16; j++)
		{
			if (as.GetAt(i*2)==hex[j])
				n+=(j*16);
			if (as.GetAt(i*2+1)==hex[j])
				n+=j;
		}
		text[i]=(BYTE)n;
	}
	memset(code, 0, 32);
	memcpy(code, "d9~R.E+", 9);

	this->Initialize(code, 9);
	this->Decode(text, text, len);

	pw=(LPCTSTR)text;

delete [] text;
delete [] code;
return pw;
}
#endif
