// --- DXInput.cpp

#include "stdafx.h"


#if !defined(GLB_CONSOLE)

const int MIN_QUERY_COUNT = 100;

#include "gfqueue.h"

#if !defined(WANT_SDL) && defined(WIN32)
	#include <mmsystem.h> // NT-Joystick
	#pragma comment(lib, "dinput8.lib")
	#pragma comment(lib, "winmm.lib")
#endif
#include "dxinput.h"

#if defined(GP2X) || defined(GPHWIZ)
#include <sys/mman.h>
#include <fcntl.h>
#endif

#if defined(GP2X) || defined(GPHWIZ)
	#include <sys/time.h>
#endif

#if defined(CAANOO)
	extern "C" void caanoo_sensor_poll(int* ax, int* ay, int* az);
	extern "C" void caanoo_sensor_deinit(void);
	extern "C" void caanoo_sensor_init();
#endif

#if defined(EMSCRIPTEN)
	// function glb_get_multitouch_info(imouse, int_ptr)
	extern "C" void glb_init_multitouch();
	extern "C" int  glb_get_multitouch_info(int imouse, int* p_X_Y_Button);
	extern "C" int  glb_get_gravity(float* pXYZ); // return 1/0 if gravity is available. FF26 returns 1 - bug
#endif


#if defined(TARGET_OS_IPHONE)
extern "C"
{
	void iPhoneGetAcceleration(float* x, float* y, float* z);
	int  iPhoneGetMaxMice();
	void iPhoneGetMouseData(int iMouse, int* x, int* y, int* button);
	int iPhoneIsICadeAttached();
	unsigned int iPhoneGetICadeInfo();
}
#endif


#include "gffilesystem.h"


//////////////////////////////////////////////////////////////////////////////////
// EnumJoystick
//////////////////////////////////////////////////////////////////////////////////

#if !defined(WANT_SDL) && defined(WIN32) && !defined(POCKETPC)
	struct GF_JoyEnum
	{
		int				CurDev;
		#if DIRECTINPUT_VERSION > 0x0700
			LPDIRECTINPUT8A	pDI;
			LPDIRECTINPUTDEVICE8A	lpDIDJoystick[DXIN_MAX_JOYSTICKS];
		#else
			LPDIRECTINPUT	pDI;
			LPDIRECTINPUTDEVICE		lpDIDJoystick[DXIN_MAX_JOYSTICKS];
		#endif
		TCHAR			Names[DXIN_MAX_JOYSTICKS][MAX_PATH*2+2];
		HWND			hWnd;
	};


	BOOL CALLBACK g_EnumJoystick(LPCDIDEVICEINSTANCE pdinst, LPVOID pvRef)
	{
	HRESULT		hr = DI_OK;
	GF_JoyEnum* pJoy = (GF_JoyEnum*)pvRef;

		if (pJoy->CurDev>=DXIN_MAX_JOYSTICKS) return DIENUM_STOP;
		if(pJoy->pDI->CreateDevice(pdinst->guidInstance, &pJoy->lpDIDJoystick[pJoy->CurDev], NULL) != DI_OK)
			return DIENUM_CONTINUE;

		if (pJoy->lpDIDJoystick[pJoy->CurDev]==NULL) return DIENUM_CONTINUE;
		pJoy->lpDIDJoystick[pJoy->CurDev]->SetDataFormat(&c_dfDIJoystick);
		pJoy->lpDIDJoystick[pJoy->CurDev]->SetCooperativeLevel(pJoy->hWnd, DISCL_EXCLUSIVE | DISCL_FOREGROUND);

		// Set the X-axis range (-1000 to +1000)
		DIPROPRANGE diprg;
		diprg.diph.dwSize = sizeof(diprg);
		diprg.diph.dwHeaderSize = sizeof(diprg.diph);
		diprg.diph.dwObj = DIJOFS_X;
		diprg.diph.dwHow = DIPH_BYOFFSET;
		diprg.lMin = -32768;
		diprg.lMax = +32768; // 1000

		pJoy->lpDIDJoystick[pJoy->CurDev]->SetProperty(DIPROP_RANGE, &diprg.diph);
		diprg.diph.dwObj = DIJOFS_Y;
		pJoy->lpDIDJoystick[pJoy->CurDev]->SetProperty(DIPROP_RANGE, &diprg.diph);
		diprg.diph.dwObj = DIJOFS_Z;
		pJoy->lpDIDJoystick[pJoy->CurDev]->SetProperty(DIPROP_RANGE, &diprg.diph);
		diprg.diph.dwObj = DIJOFS_RX;
		pJoy->lpDIDJoystick[pJoy->CurDev]->SetProperty(DIPROP_RANGE, &diprg.diph);
		diprg.diph.dwObj = DIJOFS_RY;
		pJoy->lpDIDJoystick[pJoy->CurDev]->SetProperty(DIPROP_RANGE, &diprg.diph);
		diprg.diph.dwObj = DIJOFS_RZ;
		pJoy->lpDIDJoystick[pJoy->CurDev]->SetProperty(DIPROP_RANGE, &diprg.diph);

		// Set X axis dead zone to 10%
		DIPROPDWORD dipdw;
		dipdw.diph.dwSize = sizeof(dipdw);
		dipdw.diph.dwHeaderSize = sizeof(dipdw.diph);
		dipdw.diph.dwObj = DIJOFS_X;
		dipdw.diph.dwHow = DIPH_BYOFFSET;
		dipdw.dwData = 0; //1000

		if(pJoy->lpDIDJoystick[pJoy->CurDev]->SetProperty(DIPROP_DEADZONE, &dipdw.diph) != DI_OK)
			return DIENUM_CONTINUE;

		// Set Y axis dead zone to 10%
		dipdw.diph.dwObj = DIJOFS_Y;
		if(pJoy->lpDIDJoystick[pJoy->CurDev]->SetProperty(DIPROP_DEADZONE, &dipdw.diph) != DI_OK)
			return DIENUM_CONTINUE;
		hr = pJoy->lpDIDJoystick[pJoy->CurDev]->Acquire();
		
		strcpy(pJoy->Names[pJoy->CurDev], pdinst->tszInstanceName);
		strcat(pJoy->Names[pJoy->CurDev], ", ");
		strcat(pJoy->Names[pJoy->CurDev], pdinst->tszProductName);

		pJoy->CurDev++;
	return DIENUM_CONTINUE;
	}
#endif // !WANT_SDL


//---------------------------------------------------------------//
// --- constructor ---
//---------------------------------------------------------------//
DXInput::DXInput()
{
	m_QueryCount = 0;
	m_iCurMouse=0;
	m_MousePos.SetAllocSteps(1,4);
	m_MousePos.SetSize(1);

	m_SystemPointer=false;
	NumJoysticks=0;

#if defined(EMSCRIPTEN)
	glb_init_multitouch();

#elif defined(GP2X) || defined(GPHWIZ)
	device=-1;
    dev_touchscreen = -1;
// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE)
#elif defined(WANT_SDL)
	m_AccelerationNativeLandscape = false;
	// SDL-port
	memset(m_pJoys, 0, sizeof(SDL_Joystick*) * DXIN_MAX_JOYSTICKS);

#else // Win32
	m_HWND=NULL;
	g_lpDID			=NULL;
	lpDInput		=NULL;
	lpDIDKeyboard	=NULL;
	lpDIDMouse		=NULL;
	for(int i=0; i<DXIN_MAX_JOYSTICKS; i++)
		lpDIDJoystick[i]=NULL;
#endif
}

//---------------------------------------------------------------//
// --- destructor --
//---------------------------------------------------------------//
DXInput::~DXInput()
{
	SystemPointer(true);
#if defined(GP2X) || defined(GPHWIZ) // GP2X
	if(device>=0) close(device);
	#if defined(GP2X)
		if(dev_touchscreen>=0) close(dev_touchscreen);
	#endif
	device=-1;
	dev_touchscreen=-1;
// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE)
#elif defined(WANT_SDL) // SDL

	#ifdef CAANOO
		caanoo_sensor_deinit();
	#endif


	for (long i=0; i<min(DXIN_MAX_JOYSTICKS, SDL_NumJoysticks()); i++)
	{
	    #if SDL_VERSION_ATLEAST(2,0,0)
			if(m_pJoys[i])
			{
				SDL_JoystickClose(m_pJoys[i]);
				m_pJoys[i] = NULL;
			}
		#else
			if(SDL_JoystickOpened(i) && m_pJoys[i])
				SDL_JoystickClose(m_pJoys[i]);
		#endif
	}
#else // Win32
	if (lpDIDMouse)
	{
		lpDIDMouse->Unacquire();
		lpDIDMouse->Release();
	}
	if(lpDIDKeyboard)
	{
		lpDIDKeyboard->Unacquire();		// kein zugriff aufs keyboard
		lpDIDKeyboard->Release();		// keyboard object lschen
	}
	for (int i=0; i<DXIN_MAX_JOYSTICKS; i++)
	if (lpDIDJoystick[i])
	{
		lpDIDJoystick[i]->Unacquire();
		lpDIDJoystick[i]->Release();
	}
	if (lpDInput)	lpDInput->Release();
#endif
}



#if defined(GP2X)
/*
 * The F-200 uses tslib data for it's touchscreen (although no modules).
 * I've ripped the basic (linear) code directly into here.
 */

// Set default a touchscreen calibration
static int ts_cal[7] = {6203, 0, -1501397, 0, -4200, 16132680, 65536};
// WIZ: -155 -21777 21672896 16368 -36 -563220 655366


static void read_calibration()
{
  struct stat sbuf;
  int pcal_fd;
  char pcalbuf[200];
  int index;
  char *tokptr;
  char *calfile = "/etc/pointercal";

  if (stat(calfile,&sbuf) == 0) {
    pcal_fd = open(calfile,O_RDONLY);
    read(pcal_fd,pcalbuf, 200);
    ts_cal[0] = atoi(strtok(pcalbuf," "));
    index = 1;
    while (index < 7) {
      tokptr = strtok(NULL," ");
      if (*tokptr != '\0') {
	ts_cal[index] = atoi(tokptr);
	index++;
      }
    }
    close(pcal_fd);
  };
}
#endif


//---------------------------------------------------------------//
// --- create DXInput objects ---
//---------------------------------------------------------------//
#ifndef WIN32
unsigned long DXInput::CreateDInput(HWND hWnd, HINSTANCE hInstance, bool WantMouse)
#else
HRESULT DXInput::CreateDInput(HWND hWnd, HINSTANCE hInstance, bool WantMouse)
#endif
{
	m_iCurMouse=0;

// -------- GP2X --------
#if defined(GP2X) || defined(GPHWIZ)
	NumJoysticks = 1;
	device = open("/dev/GPIO", O_RDWR | O_NDELAY );
	if (device < 0)
	{
		printf("GPIO device not present");
	}
	keystate=0;
	
	#ifdef GP2X
		#define DEV_TOUCHSCREEN	"/dev/touchscreen/wm97xx"
		dev_touchscreen=open( DEV_TOUCHSCREEN, O_RDONLY | O_NOCTTY); // | O_NDELAY
		if(dev_touchscreen<0)
		{
			printf("touchscreen device not present\n");
		}
		else
		{
			read_calibration();
		}
	#else
		m_SystemPointer=false;
		SDL_ShowCursor(false);
		dev_touchscreen = 1;
	#endif
// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE)
	NumJoysticks = 2; // 0: acceleration, 1: iCade
	return 0;
// -------- SDL --------
#elif defined(WANT_SDL)

	// SDL_Init(SDL_INIT_JOYSTICK); - global call now
	#if !defined(EMSCRIPTEN)
		#if SDL_MAJOR_VERSION>1 || SDL_MINOR_VERSION >= 3
			SDL_StartTextInput();
		#else
			SDL_EnableUNICODE(1);
		#endif
	#endif

	#if defined(PANDORA)
    /* http://blogs.distant-earth.com/wp/?p=293
       On the OpenPandora we need to work around an SDL assumption that 
       returns relative mouse coordinates when you get to the screen 
       edges using the touchscreen and a custom surface for the cursor. 
       The workaround is to set a blank SDL cursor and NOT disable it 
       (Hackish I know).

       The root issues lies in the Windows Manager GRAB code in SDL. 
       That is why the issue is not seen on framebuffer devices like the 
       GP2X (there is no X window manager  ). 
    */ 
    SDL_ShowCursor(SDL_ENABLE); 
    uint8_t hiddenCursorData = 0; 
    SDL_Cursor *hiddenCursor = SDL_CreateCursor(&hiddenCursorData, &hiddenCursorData, 8, 1, 0, 0); 
    SDL_SetCursor(hiddenCursor);
	#endif


	this->DXtoVK(false);
	m_SystemPointer=false;

	#if !defined( PALM_WEBOS) && !defined(PANDORA)
		SDL_ShowCursor(false);
	#endif

	NumJoysticks = min(DXIN_MAX_JOYSTICKS, SDL_NumJoysticks());
	for (long i=0; i<NumJoysticks; i++)
	{
		m_pJoys[i]=SDL_JoystickOpen(i);
	}

	#ifdef CAANOO
	g_debug("caanoo sensor init\n");
	caanoo_sensor_init();
	g_debug("caanoo sensor initalized\n");
	NumJoysticks = 2; // 0: dpad, 1: acceleration
	#endif

	return 0;

// -------- WIN32 --------
#else
	HRESULT		ret;
	HRESULT		hr;

	ret=(HRESULT)DINPUT_OK;
/*
#define DIRECTINPUT_VERSION_7 0x700
hr = DirectInputCreateEx(hInstance, DIRECTINPUT_VERSION_7, IID_IDirectInput7, (void **)&lpDInput, NULL);
*/
	// hr=DirectInput8Create(hInstance, 0x0800/*DIRECTINPUT_VERSION*/, &lpDInput, NULL);
	#if DIRECTINPUT_VERSION > 0x0700
		hr = DirectInput8Create(hInstance, DIRECTINPUT_VERSION, IID_IDirectInput8A, (void**)&lpDInput, NULL);
	#else
		#define DIRECTINPUT_VERSION_7 0x0700
		hr = DirectInputCreate(hInstance, DIRECTINPUT_VERSION_7, &lpDInput, NULL);
	#endif

	if(FAILED(hr))
	{
		g_debug("No DX_Input\n");
		ret=CANT_CREATE_DINPUT;
	}

	m_HWND=hWnd;

	// Get cursor pos and save use it in the future
	POINT pt;
	::GetCursorPos(&pt);
	::ScreenToClient(hWnd, &pt);
	MousePos& mps = GetCurMousePos();
	mps.m_MouseX=pt.x; mps.m_MouseY=pt.y;
	mps.m_VirtualX=pt.x; mps.m_VirtualY=pt.y;
	memset(m_KeyBuffer, 0, sizeof(m_KeyBuffer));

	// --- keyboard ---
	if (ret==(HRESULT)DINPUT_OK)
	{
		hr=lpDInput->CreateDevice(GUID_SysKeyboard, &lpDIDKeyboard, NULL);
		if (SUCCEEDED(hr))
		{
			lpDIDKeyboard->SetDataFormat(&c_dfDIKeyboard);
			hr = lpDIDKeyboard->SetCooperativeLevel(hWnd, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND); // DISCL_EXCLUSIVE | <- No Alt-TAB/Alt-Print
			if(SUCCEEDED(hr))	// Prg. mu im Vordergrund laufen
			{
				hr=lpDIDKeyboard->Acquire();
				if(FAILED(hr))
				{
					g_debug("can't aquire keyboard - first try\n");
				}
				/*
				if(FAILED(hr))
				{
					g_debug("No DX_KB\n");
					ret=NO_ACCESS_ON_KEYBOARD;
				}
				*/
			} // NT Cannot Create DX Keyboard
		}
		this->DXtoVK(lpDIDKeyboard!=NULL ? true : false);

		// --- mouse ---
		m_SystemPointer=false;
		if(WantMouse)
		{
			hr=lpDInput->CreateDevice(GUID_SysMouse, &lpDIDMouse,NULL);
			if (SUCCEEDED(hr))
			{
				lpDIDMouse->SetDataFormat(&c_dfDIMouse);
				#ifdef _DEBUG
					m_MouseCooLevel = DISCL_NONEXCLUSIVE | DISCL_FOREGROUND; // DISCL_EXCLUSIVE|;
				#else
					m_MouseCooLevel = DISCL_EXCLUSIVE|DISCL_FOREGROUND;
				#endif
				if(SUCCEEDED(lpDIDMouse->SetCooperativeLevel(hWnd, m_MouseCooLevel)))
				{
					hr=lpDIDMouse->Acquire();
					if(FAILED(hr))
					{
						g_debug("can't aquire mouse - first try\n");
					}
				}
				if(FAILED(hr))
				{
					g_debug("No DX_Mouse\n");
					ret=NO_ACCESS_ON_MOUSE;
				}
			}
		}
		else
			SystemPointer(true);
		// --- joystick ---
		GF_JoyEnum	joy;
		memset(&joy, 0, sizeof(joy));
		joy.hWnd=hWnd;
		joy.pDI=lpDInput;
		// #define DI8DEVCLASS_GAMECTRL        4
		hr = lpDInput->EnumDevices(	4 ,
									g_EnumJoystick,
									(LPDIRECTINPUT) &joy,
									(long unsigned int)DIEDFL_ATTACHEDONLY);
		if(SUCCEEDED(hr))
		{
			NumJoysticks=joy.CurDev;
			for (int i=0; i<NumJoysticks; i++)
			{
				lpDIDJoystick[i]=joy.lpDIDJoystick[i];
				strcpy(JoyNames[i], joy.Names[i]);
			}
		}
	}
	else // NT-Patch
	{
		NumJoysticks=joyGetNumDevs();
		lpDIDJoystick[0]=lpDIDJoystick[1]=NULL;
		JOYCAPS	jc;
		for (int i=0; i<NumJoysticks; i++)
		{
			joyGetDevCaps(i, &jc, sizeof(jc));
			strcpy(JoyNames[i], jc.szPname);
		}
		this->DXtoVK(false);
	}
	return ret;
#endif

}

//---------------------------------------------------------------//
//---------------------------------------------------------------//
void DXInput::GetJoyName(char* pName, int num)
{
	if (num<0 || num>=NumJoysticks)
	{
		pName[0]=0;

		#ifdef ANDROID
			if(num == NumJoysticks)
			{
				strcpy(pName, "Gamepad"); // Android gamepad using KEY() - spacefractal
			}
		#endif
	}
	else
	{
// -------- GP2X --------
#if defined(GP2X)
	strcpy(pName, "GP2X");
#elif defined(GPHWIZ)
	strcpy(pName, "WIZ");

#elif defined(CAANOO)
	if(num==1)
		strcpy(pName, "Accelerometer");
	else
		strcpy(pName, "CAANOO");

// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE) || defined(PALM_WEBOS) || defined(EMSCRIPTEN)
// -------- SDL --------
	switch(num)
	{
	case 0:
		strcpy(pName, "Accelerometer");
		break;
	case 1:
		strcpy(pName, "iCade");
		break;
	default:
		pName[0]='\0';
	}
#elif defined(WANT_SDL)
	#if SDL_VERSION_ATLEAST(2,0,0)
		strcpy(pName, SDL_JoystickName(m_pJoys[num]));
	#else
		strcpy(pName, SDL_JoystickName(num));
	#endif
// -------- WIN32 --------
#else
		strcpy(pName, JoyNames[num]);
#endif
	} // endif
}


//---------------------------------------------------------------//
//---------------------------------------------------------------//
int DXInput::GetNumJoysticks() // Get Numer of Found Devices
{
	#if defined(EMSCRIPTEN)
		NumJoysticks = 1; // tilt GF_TODO icade
	#endif

	// iCade "joystick" gets attached once you pressed a valid button
	#ifdef TARGET_OS_IPHONE
	if(NumJoysticks == 1 && iPhoneIsICadeAttached())
		NumJoysticks = 2;
	#endif

	#ifdef ANDROID
		return 1+NumJoysticks; // last = Android Joypad
	#endif

	return NumJoysticks;
}

//---------------------------------------------------------------//
// --- get keyboard status ---
//---------------------------------------------------------------//
void DXInput::GetKeyboardStatus(bool bForcePolling)
{
	if(bForcePolling ||++m_QueryCount > MIN_QUERY_COUNT)
		m_QueryCount = 0;
	else
		return;

// -------- GP2X --------
#if defined(GP2X) || defined(GPHWIZ)
	keystate = 0;
	if (read(device, &keystate, 4) != 4)
	{
		printf("GPIO READ ERROR\n");
	}
// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE)
// -------- SDL --------
#elif defined(WANT_SDL)
	SDL_PumpEvents();

// -------- WIN32 --------
#else
	HRESULT		hr=0;

	if(::GetActiveWindow() != m_HWND) return;
	memset(&m_KeyBuffer, 0, sizeof(m_KeyBuffer) );
	if (lpDIDKeyboard)
	{
		//Keyboardbuffer lschen (beep)
		if(GetInputState())
		{
			// avoid some "beeeeep" when you have too much input
			// This will make inkey$ broken!!
			// MSG msg; while(PeekMessage(&msg, m_HWND,WM_KEYFIRST,WM_KEYLAST,PM_REMOVE)) {}
		}

		hr=lpDIDKeyboard->Acquire();
		if FAILED(hr)
		{
			lpDIDKeyboard->SetDataFormat(&c_dfDIKeyboard);				// BYTE
			lpDIDKeyboard->SetCooperativeLevel(m_HWND,	// Prg. mu im Vordergrund
											   DISCL_NONEXCLUSIVE | DISCL_FOREGROUND);		// laufen!!
											   // DISCL_EXCLUSIVE | <- No Alt-TAB/Alt-Print
			lpDIDKeyboard->Acquire();
		}
		lpDIDKeyboard->GetDeviceState(sizeof(m_KeyBuffer), m_KeyBuffer);
	}
	else
		::GetKeyboardState(m_KeyBuffer);
#endif
}


//---------------------------------------------------------------//
//---------------------------------------------------------------//
bool DXInput::AnyKey()
{
	if (GetKeyChar()!='\0')
		return true;
return false;
}


//---------------------------------------------------------------//
//---------------------------------------------------------------//
bool DXInput::AnyMouse()
{
DIMOUSEDATA	md;
	RefreshMouseXY();
	this->GetMouseData(&md, false);
return (md.mButton1 || md.mButton2 || md.mButton3) ? 1:0;
}


#if defined(GP2X)
	typedef struct {
		long secs;
		long usecs;
	} TIMEVAL;

	typedef struct {
		unsigned short pressure;
		unsigned short x;
		unsigned short y;
		unsigned short pad;
		TIMEVAL  stamp;
	} TS_EVENT;

	inline int ReadTSData(int dev_touchscreen, TS_EVENT* ts_event)
	{
		int rv = read(dev_touchscreen, ts_event, sizeof(TS_EVENT));
		// always returns "8" ?
		// #need returens fales all the time!?
		if(rv >= sizeof(short)*3) return sizeof(TS_EVENT); else return 0;
	
		return rv>0 ? sizeof(TS_EVENT) : 0;
	}
#endif

#if defined(GPHWIZ)
	typedef struct {
		unsigned short pressure;
		unsigned short x;
		unsigned short y;
		unsigned short pad;

/*		TIMEVAL  stamp;
*/
	} TS_EVENT;

	// use SDL instead of tslib
	inline int ReadTSData(int dev, TS_EVENT* ts_event)
	{
		SDL_PumpEvents();
		int	px, py, btn;
		btn = SDL_GetMouseState(&px, &py);
		ts_event->pressure = (btn & SDL_BUTTON(1)) ? 40000 : 0;
		
		ts_event->x = px;
		ts_event->y = py;
	}
#endif



#if defined(ANDROID)
	extern "C" int  android_get_mousecount();
	extern "C" void android_get_mouse(int id, int* mx, int* my, int* b1);
#endif
#if defined(WIN32)
	extern "C" void win32_get_touches(int id, int* mx, int* my, int* b1);
#endif




#ifdef WIN32

struct GLB_OSVERSIONINFOW {
	DWORD dwOSVersionInfoSize;
	DWORD dwMajorVersion;
	DWORD dwMinorVersion;
	DWORD dwBuildNumber;
	DWORD dwPlatformId;
	WCHAR  szCSDVersion[ 128 ];     // Maintenance string for PSS usage
};

GLB_OSVERSIONINFOW GetRealOSVersion() {
	typedef LONG NTSTATUS, *PNTSTATUS;
#define STATUS_SUCCESS (0x00000000)
	typedef NTSTATUS (WINAPI* RtlGetVersionPtr)(GLB_OSVERSIONINFOW*);
	HMODULE hMod = ::GetModuleHandleW(L"ntdll.dll");
	if (hMod) {
		RtlGetVersionPtr fxPtr = (RtlGetVersionPtr)::GetProcAddress(hMod, "RtlGetVersion");
		if (fxPtr != NULL) {
			GLB_OSVERSIONINFOW rovi; memset(&rovi,0,sizeof(rovi));
			rovi.dwOSVersionInfoSize = sizeof(rovi);
			if ( STATUS_SUCCESS == fxPtr(&rovi) ) {
				return rovi;
			}
		}
	}
	GLB_OSVERSIONINFOW rovi; memset(&rovi,0,sizeof(rovi));
	return rovi;
}
#endif

//---------------------------------------------------------------//
// return multi-mouse count
//---------------------------------------------------------------//
int DXInput::GetNumMice()
{
	int num = 1;
	#if defined(EMSCRIPTEN)
		return 10;
	
	// -------- GP2X --------
	#elif defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)

	// -------- iPhone --------
	#elif defined(TARGET_OS_IPHONE)
	num= iPhoneGetMaxMice();

	#elif defined(PALM_WEBOS)
	num= 5;

	#elif defined(ANDROID)
	num= android_get_mousecount();

	// -------- SDL --------
	#elif defined(WANT_SDL)
	// -------- WIN32 --------
	#elif defined(WIN32)
		static int bIsWindows8orLater =-1;

		if(bIsWindows8orLater == -1)
		{
			GLB_OSVERSIONINFOW osver = GetRealOSVersion();
			bIsWindows8orLater = (osver.dwMajorVersion >= 6 && osver.dwMinorVersion >= 2) ? true:false;
			/*
			OSVERSIONINFO osvi;
			ZeroMemory(&osvi, sizeof(OSVERSIONINFO));
			osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
			GetVersionEx(&osvi);

			g_debug("Win: %d, %d\n", (int)osvi.dwMajorVersion, (int)osvi.dwMinorVersion );

			bIsWindows8orLater = 
			   ( (osvi.dwMajorVersion > 6) ||
			   ( (osvi.dwMajorVersion == 6) && (osvi.dwMinorVersion >= 2) ));
			*/
		}
		if(bIsWindows8orLater)
			num = 32;
	#endif

	return num;
}


//---------------------------------------------------------------//
// select one of the multiple mice (iPhone only, so far)
//---------------------------------------------------------------//
void DXInput::SetActiveMouse(int i)
{
	int num_mc = GetNumMice();
	m_iCurMouse = min(i, num_mc-1);

	if( m_iCurMouse >= (int)m_MousePos.Size())
		m_MousePos.SetSize( num_mc );
}


//---------------------------------------------------------------//
// --- get mousedata ---
// touch position for current mouse. false if not touched(clicked)
// pX, pY -> in last touch position. out new touch pos.
// RV = 1|2|4:  buttons 1,2,3 pressed
//      0:      no touch -> pX,pY is untouched
//     -1:      error
//---------------------------------------------------------------//
int DXInput::GetTouchPos(int* pX, int* pY, int* pWheel)
{
	*pWheel = 0;

#if defined(EMSCRIPTEN)
	int st[4] = {0,0,0,0}; // x,y,mb1
	int count = glb_get_multitouch_info(m_iCurMouse, &st[0]);
	if(count <= m_iCurMouse)
		return 0;
	*pX = st[0];
	*pY = st[1];
	*pWheel = 0;
	return st[2];



// -------- GP2X (F200) --------
#elif defined(GP2X) || defined(GPHWIZ)
	struct HISTORY
	{
		long x,y;
	};
	struct LASTSTATE
	{
		int x,y,b;
	};

	#ifdef GP2X
		#define MINPRESSURE 35000 // soft touch on GP2X causes terrible jitter
		#define MAX_HISTORY	7 // number of points in history before you start shifting them

		#define MAX_ERROR	3 // number of big-step reads to see it's true
		#define MAX_JITTER 16 // big step jitter value
		#define MIN_HISTORY 3 // minimum number of reads to get a safe x/y position
	#else
		/* +/- 1 pixel, quite good
		#define MINPRESSURE 0 // WIZ just returns 1 or 0
		#define MAX_HISTORY	15 // number of points in history before you start shifting them
		#define MAX_ERROR	 4
		#define MAX_JITTER  16
		#define MIN_HISTORY 12
		*/

		#define MINPRESSURE 0 // WIZ just returns 1 or 0
		#define MAX_HISTORY	15 // number of points in history before you start shifting them
		#define MAX_ERROR	 8
		#define MAX_JITTER  16
		#define MIN_HISTORY 12
		
	#endif
	#if MIN_HISTORY >= MAX_HISTORY
		#error Incorrect history size in __FILE__ __LINE__
	#endif

	static HISTORY hist[MAX_HISTORY+1], errhist[MAX_HISTORY+1];
	static int num_hist=0;
	TS_EVENT ts_event;

// only poll every 30th of a second
static LASTSTATE sLastState = {0,0,0};
static unsigned int sNextTime = 0;

	// GETTIMERALL()
	timeval now;
	gettimeofday(&now, NULL);
	unsigned int thistime = (now.tv_sec)*1000+(now.tv_usec/1000);
	
	if(thistime < sNextTime)
	{
		*pX = sLastState.x; *pY=sLastState.y;
		return sLastState.b;
	}
	sNextTime = thistime+6;

	if(dev_touchscreen>=0)
	{
		int errors = 0;
	repeat_reading:
		int nread = ReadTSData(dev_touchscreen, &ts_event);
	
		if(nread < sizeof(TS_EVENT) )
		{
			return 0;
		}

		int bPressed = 0;
		if(ts_event.pressure>MINPRESSURE)
		{
			bPressed=1;
			int dx, dy;

			#ifdef GP2X
				dx = (ts_event.x * ts_cal[0] + ts_cal[2]) >> 16;
				dy = (ts_event.y * ts_cal[4] + ts_cal[5]) >> 16;
			#else
				// what tslib does:
				/*
				int xtemp = ts_event.x; int ytemp = ts_event.y;
				dx = 	( ts_cal[2] + ts_cal[0]*xtemp + ts_cal[1]*ytemp ) / ts_cal[6];
				dy =	( ts_cal[5] + ts_cal[3]*xtemp + ts_cal[4]*ytemp ) / ts_cal[6];
				*/
				dx = ts_event.x;
				dy = ts_event.y;
			#endif
			if (dx<0 || dx>319 || dy<0 || dy>239) // bad read
			{
				*pX = sLastState.x; *pY=sLastState.y;
				return sLastState.b;
			}

			// put current read in history
			hist[num_hist].x = dx;
			hist[num_hist].y = dy;
			++num_hist;
			
			// slow movement
			if(abs(dx-sLastState.x)<MAX_JITTER && abs(dy-sLastState.y)<MAX_JITTER)
			{
			stable_move:
				// has more than one value, interpolate for stable results
				if(num_hist > MIN_HISTORY) // best to have some values at least
				{
					// shift all histroy one back. (cyclic buffer would be faster)
					if(num_hist>MAX_HISTORY-1)
					{
						for(int i=0; i<MAX_HISTORY-1; ++i)
						{
							hist[i]=hist[i+1];
						}
						num_hist = MAX_HISTORY-1;
					}

					int factor;
					int accux=0, accuy=0, scaler=0;

					// senquack, SDL, GP2X
					//  1->2->3->4->5->6->7->8->7->6->5->4->3->2->1

					int emphasis[]={8,7,6,5,4,3,2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};

					for(int i=0; i<num_hist; ++i)
					{
						factor=emphasis[num_hist - i -1];
						accux+=(factor * hist[i].x);
						accuy+=(factor * hist[i].y);
						scaler+=factor;
					}
					// interpolate new dx,dy from history
					dx = ((accux<<8)/scaler)>>8;
					dy = ((accuy<<8)/scaler)>>8;
				}
				else
				{
					goto repeat_reading;
				}
			}
			else
			{
				// big jitter
				errhist[errors].x = dx;
				errhist[errors].y = dy;
				++errors;
				
				if(errors>MAX_ERROR) // too many errors -> must be real.
				{
					for(int i=0; i<errors; ++i)
					{
						hist[i] = errhist[i];
					}
					num_hist=errors;

					// take these errors as valid input
				}
				else
				{
					goto repeat_reading;
				}
			}

			// ok, use that as position
			*pX = dx;
			*pY = dy;
			sLastState.x=dx;
			sLastState.y=dy;
		}
		else // not pressed
		{
			num_hist=0; // pen released
			bPressed = 0;
		}

		sLastState.b=bPressed;
		return bPressed;
	}

	return 0; // uh oh - no touchscreen device


// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE)
	int button=0;
	iPhoneGetMouseData(m_iCurMouse, pX, pY, &button);
	return button ? 1:0;

// -------- Android --------
#elif defined(ANDROID)
	int button=0;
	android_get_mouse(m_iCurMouse, pX, pY, &button);
	return button ? 1:0;


// -------- SDL --------
#elif defined(WANT_SDL)
	SDL_PumpEvents();
	Uint8 buttons;

	// Android: SDL 1.3 w/o multitouch :/
	#if defined (ANDROID) // #NEED multitouch?
		#define SDL_GetRelativeMouseState(i,x,y)	SDL_GetRelativeMouseState(x,y)
		#define SDL_GetMouseState(i,x,y)			SDL_GetMouseState(x,y)
	#endif

	// Absolute Movement - touches or systempointer!
	if(1) // IsSystemPointer() )
	{
		#if defined(PALM_WEBOS)
			buttons = SDL_GetMultiMouseState(m_iCurMouse, pX, pY); // mouse index -> Mutimouse support
		#else
			#if 0 // SDL_MAJOR_VERSION>1 || SDL_MINOR_VERSION >= 3
				buttons = SDL_GetMouseState(m_iCurMouse, pX, pY);
			#else
				buttons = SDL_GetMouseState(pX, pY);
			#endif
		#endif
	}
	else // Relative movement - no systempointer (when pointer goes out of window, you have to move a long way to get it back to 0)
	{
		int	ix=0, iy=0;
		#if defined(PALM_WEBOS)
			buttons = SDL_GetRelativeMultiMouseState(m_iCurMouse, &ix, &iy); // mouse index -> Mutimouse support
		#else
			#if 0 // SDL_MAJOR_VERSION>1 || SDL_MINOR_VERSION >= 3
				// SDL_GetNumMice(void);
				// SDL_SelectMouse(int index); index==-1 -> Query current
				buttons = SDL_GetRelativeMouseState(m_iCurMouse, &ix, &iy); // mouse index -> Mutimouse support
			#else
				buttons = SDL_GetRelativeMouseState(&ix, &iy);
			#endif
		#endif
		*pX +=ix;
		*pY +=iy;
	}

	// The trick is the wheel-up event is reported as button 4 and the wheel-down event is reported as mouse button 5.
	/*
	/etc/X11/XF86Config-4:
	Section "InputDevice"
			Identifier      "Configured Mouse"
			Driver          "mouse"
			Option          "Device"                "/dev/psaux"
			Option          "Protocol"              "PS/2"
			Option          "ZAxisMapping"          "4 5"  # ADD THIS LINE!!!
	EndSection
	*/
	if (buttons & SDL_BUTTON(4)) *pWheel = 1;
	else if (buttons & SDL_BUTTON(5)) *pWheel = -1;

	// Buttons
	int retv = 0;
	if (buttons & SDL_BUTTON(1)) retv|=1; // Left
	if (buttons & SDL_BUTTON(3)) retv|=2; // Right
	if (buttons & SDL_BUTTON(2)) retv|=4; // Middle
	return retv;

// -------- WIN32 --------
#else

	if(GetNumMice() > 1 && IsSystemPointer()) // Win8 WM_POINTER API
	{
		int button=0;
		win32_get_touches(m_iCurMouse, pX, pY, &button);
		return button;
	}
	else
	{
		int rv = 0;
		POINT p;
		if (lpDIDMouse)
		{
		HRESULT			hr;
		DIMOUSESTATE	Mouse;
			hr=lpDIDMouse->GetDeviceState(sizeof(DIMOUSESTATE), &Mouse);
			if(FAILED(hr))
			{
				hr = lpDIDMouse->Acquire();
				if(FAILED(hr))
				{
					hr = lpDIDMouse->SetDataFormat(&c_dfDIMouse);
					hr = lpDIDMouse->SetCooperativeLevel(m_HWND, m_MouseCooLevel);
					hr = lpDIDMouse->Acquire();
					return 0; // Uh-oh
				}
				hr=lpDIDMouse->GetDeviceState(sizeof(DIMOUSESTATE), &Mouse);
			}
			if(SUCCEEDED(hr))
			{
				*pWheel = Mouse.lZ>0 ? 1 : (Mouse.lZ<0 ? -1 : 0); // Mousewheel
				if( m_SystemPointer)
				{
					// true position
					::GetCursorPos(&p);
					::ScreenToClient(m_HWND, &p);
					*pX = p.x;
					*pY = p.y;
				}
				else
				{
					// movement
					*pX += Mouse.lX;
					*pY += Mouse.lY;
				}
				if(Mouse.rgbButtons[0]) rv|=1;
				if(Mouse.rgbButtons[1]) rv|=2;
				if(Mouse.rgbButtons[2]) rv|=4;
			}
		}
		// NT PATCH
		else
		{
			if(m_SystemPointer)
			{
				::GetCursorPos(&p);
				::ScreenToClient(m_HWND, &p);
				*pX = p.x;
				*pY = p.y;
			}
			else
			{
				::GetCursorPos(&p);
				::SetCursorPos(200, 200);
				*pX += (p.x-200);
				*pY += (p.y-200);
			}
			if(GetAsyncKeyState(VK_LBUTTON) & 0x8000) rv|=1;
			if(GetAsyncKeyState(VK_RBUTTON) & 0x8000) rv|=2;
			if(GetAsyncKeyState(VK_MBUTTON) & 0x8000) rv|=4;
		}

		// Clip mouse coords to screen
		RECT rect;
		::GetWindowRect(m_HWND, &rect);
		if(*pX<0) *pX = 0;
		if(*pY<0) *pY = 0;
		if(*pX> rect.right-rect.left) *pX = rect.right-rect.left;
		if(*pY> rect.bottom-rect.top) *pY = rect.bottom-rect.top;

		return rv;
	} // not Win8
#endif
return 0;
}

//---------------------------------------------------------------//
// --- get mousedata ---
//---------------------------------------------------------------//
void DXInput::GetMouseData(DIMOUSEDATA *Data, bool bUpdateXY)
{
	MousePos& mps = GetCurMousePos();
	int tx = mps.m_MouseX;
	int ty = mps.m_MouseY; // #need multitouch coords for Systempointer(false)!!
	int wheel=0;
	int buttons = 0;
	if(IsSystemPointer())
	{
		tx = mps.m_MouseX;
		ty = mps.m_MouseY;
		buttons = GetTouchPos(&tx, &ty, &wheel);
		Data->mAccX = tx - mps.m_MouseX;
		Data->mAccY = ty - mps.m_MouseY;
	}
	else
	{
		tx = mps.m_VirtualX;
		ty = mps.m_VirtualY;
		buttons = GetTouchPos(&tx, &ty, &wheel);
		mps.m_VirtualX = tx;
		mps.m_VirtualY = ty;
		Data->mAccX = mps.m_VirtualX - mps.m_MouseX;
		Data->mAccY = mps.m_VirtualY - mps.m_MouseY;
	}
	mps.m_MouseWheel += wheel;
	Data->mMouseWheel = mps.m_MouseWheel;

	Data->mButton1 = (buttons & 1) ? 1:0;
	Data->mButton2 = (buttons & 2) ? 1:0;
	Data->mButton3 = (buttons & 4) ? 1:0;

	if(bUpdateXY)
	{
		if(IsSystemPointer())
		{
			mps.m_MouseX = mps.m_VirtualX = tx;
			mps.m_MouseY = mps.m_VirtualY = ty;
		}
		else
		{
			mps.m_MouseX = mps.m_VirtualX;
			mps.m_MouseY = mps.m_VirtualY;
		}
		mps.m_MouseWheel = 0;
	}
}

#if defined (ANDROID)
	extern "C" int AndroidScancode(int);
#endif


//---------------------------------------------------------------//
//---------------------------------------------------------------//
bool DXInput::GetJoystickData(DIJOYSTATE* Data, int num)
{
// 2018-05-10 GF: just to be safe if there's no joystick
memset(Data, 0, sizeof(DIJOYSTATE));

#if defined (EMSCRIPTEN)
	float xyz[3] = {0.0f, 0.0f, 0.0f};
	int ok = glb_get_gravity(xyz);
	if(ok)
	{
		Data->lX= (long)(xyz[0] * 32767.0f);
		Data->lY= (long)(xyz[1] * 32767.0f);
		Data->lZ= (long)(xyz[2] * 32767.0f);
		return true;
	}
	return false;

// -------- GP2X --------
#elif defined(GP2X) || defined(GPHWIZ)
	// Bit Mask
	#define GP2X_VK_UP              (1 <<  0)
	#define GP2X_VK_UP_LEFT         (1 <<  1)
	#define GP2X_VK_LEFT            (1 <<  2)
	#define GP2X_VK_DOWN_LEFT       (1 <<  3)
	#define GP2X_VK_DOWN            (1 <<  4)
	#define GP2X_VK_DOWN_RIGHT      (1 <<  5)
	#define GP2X_VK_RIGHT           (1 <<  6)
	#define GP2X_VK_UP_RIGHT        (1 <<  7)
	#define GP2X_VK_START           (1 <<  8)
	#define GP2X_VK_SELECT          (1 <<  9)
	#define GP2X_VK_FL              (1 << 10)
	#define GP2X_VK_FR              (1 << 11)
	#define GP2X_VK_FA              (1 << 12)
	#define GP2X_VK_FB              (1 << 13)
	#define GP2X_VK_FX              (1 << 14)
	#define GP2X_VK_FY              (1 << 15)
	#define GP2X_VK_VOL_UP          (1 << 16)
	#define GP2X_VK_VOL_DOWN        (1 << 17)
	#define GP2X_VK_TAT             (1 << 18)
	#define GPK(a) ((keystate & (GP2X_VK_ ##a)) ? 1:0)
	#define GPKq(a) (keystate & (GP2X_VK_ ##a))

	Data->lX = (GPK(RIGHT)|GPK(UP_RIGHT)|GPK(DOWN_RIGHT)) - (GPK(UP_LEFT) | GPK(LEFT) | GPK(DOWN_LEFT) );
	Data->lY = (GPK(DOWN_LEFT) | GPK(DOWN) | GPK(DOWN_RIGHT) ) - (GPK(UP)|GPK(UP_RIGHT)|GPK(UP_LEFT));

	// diagonal?
	// If you're on the NNW positions, remove the minor fraction part.
	// But allow for the DPad Mod
	if(Data->lX && Data->lY)
	{
		if( (GPKq(UP) || GPKq(DOWN)) && !(GPKq(LEFT) || GPKq(RIGHT)) )
			Data->lX = 0;
		if( (GPKq(LEFT) || GPKq(RIGHT)) && !(GPKq(UP) | GPKq(DOWN)) )
			Data->lY = 0;
	}

	Data->lRx= (GPK(FR) - GPK(FL));
	Data->lRz= (GPK(VOL_DOWN) - GPK(VOL_UP));
	Data->rgbButtons[0] = GPK(FA);
	Data->rgbButtons[1] = GPK(FB);
	Data->rgbButtons[2] = GPK(FX);
	Data->rgbButtons[3] = GPK(FY);
	Data->rgbButtons[4] = GPK(TAT);
	Data->rgbButtons[5] = GPK(START);
	Data->rgbButtons[6] = GPK(SELECT);
	Data->rgbButtons[7] = GPK(VOL_DOWN);
	Data->rgbButtons[8] = GPK(VOL_UP);
	Data->rgbButtons[9] = GPK(FL);
	Data->rgbButtons[10]= GPK(FR);

// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE)
	memset(Data, 0, sizeof(DIJOYSTATE));
	if(num==0) // acceleration
	{
		float x,y,z;
		iPhoneGetAcceleration(&x, &y, &z);
		Data->lX= (float)(x * 32767.0f); // Space Fractal, NO LONG
		Data->lY= (float)(y * 32767.0f); // Space Fractal, NO LONG
		Data->lZ= (float)(z * 32767.0f); // Space Fractal, NO LONG
		// #NEED get rx,ry,rz values for gyro? (takes lots of battery)
	}
	else
	{
		/*
			The iCADE button layout is:
			0 2 4 6
			1 3 5 7

			Button mapping on iCade and iControlPad:
			iCADE       iControlPad
			0           START
			1           B
			2           SELECT
			3           Y
			4           RIGHT SHOULDER
			5           LEFT SHOULDER
			6           A
			7           X
		*/

		if(this->GetNumJoysticks() > 1)
		{
			unsigned int icade = iPhoneGetICadeInfo();

			#define ICADE(butn, code, value) if( (icade & (unsigned int)(code)) != 0 ) {butn=value;}
			ICADE(Data->lX           , 0x0008, -32767);
			ICADE(Data->lX           , 0x0002,  32767);
			ICADE(Data->lY           , 0x0001, -32767);
			ICADE(Data->lY           , 0x0004,  32767);
			ICADE(Data->rgbButtons[0], 0x0010,  1);
			ICADE(Data->rgbButtons[1], 0x0020,  1);
			ICADE(Data->rgbButtons[2], 0x0040,  1);
			ICADE(Data->rgbButtons[3], 0x0080,  1);
			ICADE(Data->rgbButtons[4], 0x0100,  1);
			ICADE(Data->rgbButtons[5], 0x0200,  1);
			ICADE(Data->rgbButtons[6], 0x0400,  1);
			ICADE(Data->rgbButtons[7], 0x0800,  1);
			#undef ICADE
		}
	}

	
// -------- SDL --------
#elif defined(WANT_SDL)
	memset(Data, 0, sizeof(DIJOYSTATE));
	int numJ = NumJoysticks;

	#ifdef ANDROID
		++numJ; // Gamepad after acceleration
	#endif

	if(num>=0 && num<numJ)
	{
		#if defined(CAANOO)
		if(num > 0) // acceleration
		{
			int ax, ay, az;
			caanoo_sensor_poll(&ax, &ay, &az);
			Data->lX=-ax;
			Data->lY=-ay;
			Data->lZ=-az;
			return true;
		}
		#endif


		Data->lX=SDL_JoystickGetAxis(m_pJoys[num], 0);
		Data->lY=SDL_JoystickGetAxis(m_pJoys[num], 1);
		Data->lZ=SDL_JoystickGetAxis(m_pJoys[num], 2);

		Data->lRx=SDL_JoystickGetAxis(m_pJoys[num], 3);
		Data->lRy=SDL_JoystickGetAxis(m_pJoys[num], 4);
		Data->lRz=SDL_JoystickGetAxis(m_pJoys[num], 5);

		#if defined(PALM_WEBOS) || defined(PALM_PIXI)
			// 1G = 32768 SDL for WebOS

			Data->lY = -Data->lY; // swap Y for the Palm Pre to match iPhone
			// HP touchpad - turned again
			if(m_AccelerationNativeLandscape)
			{
				long swp = Data->lX; Data->lX = -Data->lY; Data->lY =swp;
			}
			return true; // no buttons on acceleration
		#endif

		#if defined(ANDROID)
			// HTC Wildfire2
			// SDL2 already returns values in G factored
			if(num == 0)
			{
				Data->lX= -Data->lX;
				Data->lY= -Data->lY;
				Data->lZ= -Data->lZ;
			}

			if(num == NumJoysticks) // 0 = acceleration
			{
				// iControlPad HID Joystick Mode (=Default) (X-Button+Start)
				//      Sel=L2 104    Start=R2 105
				//          96          Y 109
				// DPAD: 98 -- 97   A 108    B 106
				//          99          X 110
				//    L 104               R 105

				// analog nubs:
				//   258      266
				// 256 257  260 261
				//   259      267
				// http://developer.android.com/reference/android/view/KeyEvent.html#KEYCODE_BUTTON_A


				// Android joypads are sent using an keycode
				Data->lX            = 32767 * (AndroidScancode( 257 ) - AndroidScancode( 256 )); // nub1
				Data->lY            = 32767 * (AndroidScancode( 258 ) - AndroidScancode( 259 )); // nub1
				Data->lRx           = 32767 * (AndroidScancode( 261 ) - AndroidScancode( 260 )); // nub2
				Data->lRy           = 32767 * (AndroidScancode( 266 ) - AndroidScancode( 267 )); // nub2
				Data->rgbButtons[0] = AndroidScancode(  14 ); // A
				Data->rgbButtons[1] = AndroidScancode(  15 ); // B
				Data->rgbButtons[2] = AndroidScancode(  16 ); // X
				Data->rgbButtons[3] = AndroidScancode( 17 ); // Y
				Data->rgbButtons[4] = AndroidScancode( 18 ); // L1
				Data->rgbButtons[5] = AndroidScancode( 19 ); // R1
				Data->rgbButtons[6] = AndroidScancode( 54 ); // Start
				Data->rgbButtons[7] = AndroidScancode( 55 ); // Select


				Data->rgdwPOV[0] = -1; // Center
				int dx = AndroidScancode( 11 ) - AndroidScancode( 10 );
				int dy = AndroidScancode( 13 ) - AndroidScancode( 12 );
				switch(dx + 8*dy)
				{
					//    x      y
					case -1 + 8*-1: Data->rgdwPOV[0] = 4500; break;
					case  0 + 8*-1: Data->rgdwPOV[0] =    0; break;
					case  1 + 8*-1: Data->rgdwPOV[0] = 31500; break;

					case -1 + 8* 1: Data->rgdwPOV[0] = 13500; break;
					case  0 + 8* 1: Data->rgdwPOV[0] = 18000; break;
					case  1 + 8* 1: Data->rgdwPOV[0] = 22500; break;
	
					case -1 + 8* 0: Data->rgdwPOV[0] =  9000; break;
					case  1 + 8* 0: Data->rgdwPOV[0] = 27000; break;
				}		


				// remapping in GameControllers.java
				/*
				public int onKeyRemap(int keyCode)
				{	switch(keyCode){
						case 21:  return 10; // DPAD LEFT
						case 22:  return 11; // DPAD RIGHT
						case 19:  return 12; // DPAD UP
						case 20:  return 13; // DPAD DOWN
						case 96:  return 14; // A BUTTON
						case 97:  return 15; // B BUTTON (Back in menues)
						case 99:  return 16; // X BUTTON
						case 100: return 17; // Y BUTTON
						case 102: return 18; // LEFT SHOULDER
						case 103: return 19; // RIGHT SHOULDER
						case 106: return 20; // LEFT STICK BUTTON
						case 107: return 21; // RIGHT STICK BUTTON
						case 108: return 54; // PAUSE/BACK
						case 109: return 55; // SELECT
						default: return 0; // NOT A KNOWN BUTTON
					}
				*/
			}
			return true;
		#endif

		#ifndef CAANOO // any Joystick device
			int numj = min(32, SDL_JoystickNumButtons(m_pJoys[num]));
			for (long i=0; i<numj; i++)
				Data->rgbButtons[i]=(unsigned char)SDL_JoystickGetButton(m_pJoys[num], i);
		#else
			Data->rgbButtons[0]=(unsigned char)SDL_JoystickGetButton(m_pJoys[num], 0);
			Data->rgbButtons[1]=(unsigned char)SDL_JoystickGetButton(m_pJoys[num], 1);
			Data->rgbButtons[2]=(unsigned char)SDL_JoystickGetButton(m_pJoys[num], 2);
			Data->rgbButtons[3]=(unsigned char)SDL_JoystickGetButton(m_pJoys[num], 3);
			Data->rgbButtons[4]=(unsigned char)SDL_JoystickGetButton(m_pJoys[num], 10); // tap DPad

			// Caanoo fix
			// helpI, helpII (Menu+Select on Wiz, Select+Home on GP2X)
			Data->rgbButtons[5] = (unsigned char)SDL_JoystickGetButton(m_pJoys[num], 8);
			Data->rgbButtons[6] = (unsigned char)SDL_JoystickGetButton(m_pJoys[num], 9);
			// vol+-
			// Can only press both "volume" buttons -> ESC
			Data->rgbButtons[7] = (unsigned char)SDL_JoystickGetButton(m_pJoys[num], 6);
			Data->rgbButtons[8] = Data->rgbButtons[7];
			// front
			Data->rgbButtons[9] = (unsigned char)SDL_JoystickGetButton(m_pJoys[num], 4);
			Data->rgbButtons[10]= (unsigned char)SDL_JoystickGetButton(m_pJoys[num], 5);
		#endif
	}
// -------- WIN32 --------
#else

	if (num<0 || num>=NumJoysticks){ return false; }
	
	if (lpDIDJoystick[num])
	{
		LPDIRECTINPUTDEVICE8A pJoy = lpDIDJoystick[num];
		HRESULT hr = pJoy->Poll();
		if(FAILED(hr)) //  == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
		{
			pJoy->SetDataFormat(&c_dfDIJoystick);
			pJoy->SetCooperativeLevel(m_HWND, DISCL_EXCLUSIVE | DISCL_FOREGROUND);
			pJoy->Acquire();
			hr = pJoy->Poll();
		}
		if(FAILED(hr))
		{
			memset(Data, 0, sizeof(DIJOYSTATE));
		}
		else
		{
			pJoy->GetDeviceState(sizeof(DIJOYSTATE), Data);
		}
	}
	else // NT PATCH
	{
		if (num>1) return false;
		JOYINFOEX	ji;
		memset(&ji, 0, sizeof(JOYINFOEX));
		ji.dwSize=sizeof(JOYINFOEX);
		ji.dwFlags=JOY_RETURNALL;
		MMRESULT	res;
		if (num==0) res=joyGetPosEx(JOYSTICKID1, &ji);
		else        res=joyGetPosEx(JOYSTICKID2, &ji);
		if (res==JOYERR_NOERROR)
		{
			Data->lX=ji.dwXpos;
			Data->lY=ji.dwYpos;
			Data->lZ=ji.dwZpos;
			Data->rgbButtons[0]=(BYTE)(ji.dwButtons&JOY_BUTTON1);
			Data->rgbButtons[1]=(BYTE)(ji.dwButtons&JOY_BUTTON2);
			Data->rgbButtons[2]=(BYTE)(ji.dwButtons&JOY_BUTTON3);
			Data->rgbButtons[3]=(BYTE)(ji.dwButtons&JOY_BUTTON4);
			Data->rgbButtons[4]=(BYTE)(ji.dwButtons&JOY_BUTTON5);
			Data->rgbButtons[5]=(BYTE)(ji.dwButtons&JOY_BUTTON6);
			Data->rgbButtons[6]=(BYTE)(ji.dwButtons&JOY_BUTTON7);
			Data->rgbButtons[7]=(BYTE)(ji.dwButtons&JOY_BUTTON8);
			Data->rgdwPOV[0]=ji.dwPOV;
		}
		else
		{
			memset(Data, 0, sizeof(DIJOYSTATE));
		}
	}
#endif

/*
	// -10000 -> +10000
	#define LIMP 24000
	#define LIMIT(a) if (a>LIMP)a=LIMP;else if(a<-LIMP) a=-LIMP; a*=10; a/=(LIMP/1000);
	LIMIT(Data->lX);
	LIMIT(Data->lY);
	LIMIT(Data->lZ);
*/
return true;
}

#if defined(TARGET_OS_IPHONE)
	extern "C" void iPhoneRumble();
#endif

// Force Feedback
void DXInput::ffRumble(int ndevice, int count, float duration, float x_motor, float y_motor)
{
// -------- GP2X --------
#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)

// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE)
	iPhoneRumble();

// -------- SDL --------
#elif defined(WANT_SDL)
// -------- WIN32 --------
#else
	LPDIRECTINPUTDEVICE8A device = lpDIDJoystick[ndevice];
	if (!device)
		return;

//DIJOYSTATE Data;
//	if(device->GetDeviceState(sizeof(Data), &Data) == (DIERR_INPUTLOST | DIERR_NOTACQUIRED))
		device->Acquire();

	DIDEVCAPS caps;
	caps.dwSize = sizeof(DIDEVCAPS);
	device->GetCapabilities(&caps);
	if (! (caps.dwFlags & DIDC_FORCEFEEDBACK)) return; // that device is too cheap

	// Reset Force Feedback
	if(device) device->SendForceFeedbackCommand(DISFFC_RESET);

	/*
	* Create a constant "rumbling" effect with the specified parameters
	* Note that on joysticks that support "real" force feedback this will
	* probably just feel like a constant pressure.
	*/
	DICONSTANTFORCE constantForce;

	/* This is about consistent with the relative strength of the motors
	* in the Logitech Cordless Rumblepad, which seems to be what the Linux
	* FF_RUMBLE API (which bzflag's is patterned after) was designed for.
	*/
	
	if(x_motor>1.0f) x_motor=1.0;
	if(x_motor<-1.0f) x_motor=-1.0f;
	if(y_motor>1.0f) y_motor=1.0;
	if(y_motor<-1.0f) y_motor=-1.0f;

	constantForce.lMagnitude = (LONG)(DI_FFNOMINALMAX * (x_motor*x_motor+y_motor*y_motor)/2.0f);
	/*
	* Build the actual effect
	*/
	DWORD axes[3] = {DIJOFS_X, DIJOFS_Y, DIJOFS_Z};
	LONG  dir[3] = {(LONG)(x_motor*32767), (LONG)(y_motor*32767), 32767};

	LPDIRECTINPUTEFFECT createdEffect;

	DIEFFECT effect;
	effect.dwSize = sizeof(DIEFFECT);
	// coordinate system really doesn't matter for rumbles but we need to specify it.
	effect.dwFlags = DIEFF_OBJECTOFFSETS | DIEFF_CARTESIAN;
	// duration
	effect.dwDuration = (DWORD)(duration * (DI_SECONDS / 1000));
	// defaults
	effect.dwSamplePeriod = 0;
	effect.dwGain = DI_FFNOMINALMAX;
	effect.dwTriggerButton = DIEB_NOTRIGGER;
	effect.dwTriggerRepeatInterval = 0;
	// all axes
	effect.cAxes = 2;
	effect.rgdwAxes = &axes[0];
	// direction doesn't matter
	effect.rglDirection = &dir[0];
	// no envelope
	effect.lpEnvelope = NULL;
	// use the constant force data
	effect.cbTypeSpecificParams = sizeof(DICONSTANTFORCE);
	effect.lpvTypeSpecificParams = &constantForce;
	// start delay
#if(DIRECTINPUT_VERSION >= 0x0600)
	effect.dwStartDelay = (DWORD)(/*delay*/ 0 * DI_SECONDS);
#endif

	// create the effect
	HRESULT hr = device->CreateEffect(GUID_ConstantForce, &effect, &createdEffect, NULL);
	// play the thing
	hr = createdEffect->Start(count, 0);
#endif
}


Queue<char> g_inkey_last_ascii_keydown;

//---------------------------------------------------------------//
// WHAT KEY IS PRESSED (INKEY$)
//---------------------------------------------------------------//
char DXInput::GetKeyChar(void)
{
#if defined( WANT_SDL ) || defined( WIN32) || defined(TARGET_OS_IPHONE)
	int count = m_inkey_chars.Size();

	char c='\0';
	if(count)
	{
		char* pC = &m_inkey_chars[0];
		c = *pC;

		// shift backwards
		while(count>1)
		{
			*pC = pC[1];
			++pC;
			--count;
		}
		m_inkey_chars.Pop();

		if(c=='\r') c='\n';
	}
	return c;
#else
	static	char lastpressed='\0';
		char asc_unshift='\0';
		this->GetKeyboardStatus();
	
		unsigned char asc='\0', i;
		for (i=0; i<255; i++) // A-Z
		{
			if(this->GetKey(i))
			{
				asc = this->Scan2Ascii(i);
				asc_unshift=asc;
				if(asc>='A' && asc<='Z') {asc_unshift = asc - 'A' + 'a';}
				if(asc!='\0')
				{
					if( (GetKey(42) || GetKey(54)) &&
						(
							(asc>='a' && asc<='z')
						)
					)
					{
						asc-=32; // Capital Letters
					}
					break;
				}
			}
		}
		if(lastpressed==asc_unshift)
		{
			asc='\0';
		}
		else
		{
			lastpressed=asc_unshift;
		}
	if(asc=='\r') asc='\n';
	return asc;
#endif
}




//---------------------------------------------------------------//
// convert dx to ascii
// -> only used for INKEY$ function
//---------------------------------------------------------------//
char DXInput::Scan2Ascii(DWORD scancode)
{
	char ch;
	switch(scancode)
	{
	case 0x0E:				// Backsp
		ch=0x08; break;
	case 0x1C: case 0x9C:	// Return
		ch=0x0A; break;
	case 69: case 58: case 42: case 54: case 29: case 56: case 221: // NUM-Lock, ShiftGR, ShiftL, ShiftR, CtrlL, Alt, Menu
		ch='\0'; break;
	default:				// Any other key

	// -------- GP2X --------
	#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
		ch=scancode;
	// -------- iPhone --------
	#elif defined(TARGET_OS_IPHONE)
	return '\0';

	// -------- SDL --------
	#elif defined(WANT_SDL)
		ch=(unsigned char)m_MapBuffer[scancode%256];
	// -------- WIN32 --------
	#else
		// with this method, we get even different keyboard layouts
		static HKL layout=GetKeyboardLayout(0);
		static unsigned char State[256];
		static DWORD lastkey=0xffff;
		unsigned short result;

		if (::GetKeyboardState(State)==FALSE)
			return 0;

		lastkey = scancode;
		UINT vk=MapVirtualKeyEx(scancode,1,layout);
		int rv = ToAsciiEx(vk,scancode, State,&result,0,layout);
		ch = (char) result;
		if (!rv) ch='\0';
	#endif
	}
	if (ch>0 && ch<0x20 && ch!=0x08 && ch!=0x0A) // backsp & return
		ch='\0';
	return ch;
}


// ------------------------------------------------------------- //
void DXInput::DXtoVK(bool DXActive)
// ------------------------------------------------------------- //
{
#ifndef DXINPUT_HAVE_NO_KEYS
// -------- GP2X --------
#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE)

// -------- SDL --------
#elif defined(WANT_SDL)

#if SDL_VERSION_ATLEAST(2,0,0)
	// These key constants were renamed for clarity or consistency. 
	#define SDLK_KP0 SDLK_KP_0
	#define SDLK_KP1 SDLK_KP_1
	#define SDLK_KP2 SDLK_KP_2
	#define SDLK_KP3 SDLK_KP_3
	#define SDLK_KP4 SDLK_KP_4
	#define SDLK_KP5 SDLK_KP_5
	#define SDLK_KP6 SDLK_KP_6
	#define SDLK_KP7 SDLK_KP_7
	#define SDLK_KP8 SDLK_KP_8
	#define SDLK_KP9 SDLK_KP_9
	#define SDLK_NUMLOCK SDLK_NUMLOCKCLEAR
	#define SDLK_SCROLLOCK SDLK_SCROLLLOCK
	#define SDLK_PRINT SDLK_PRINTSCREEN
	#define SDLK_LMETA SDLK_LGUI
	#define SDLK_RMETA SDLK_RGUI
#endif



#ifndef DIK_ESCAPE
#define DIK_ESCAPE          0x01
#endif
#define DIK_1               0x02
#define DIK_2               0x03
#define DIK_3               0x04
#define DIK_4               0x05
#define DIK_5               0x06
#define DIK_6               0x07
#define DIK_7               0x08
#define DIK_8               0x09
#define DIK_9               0x0A
#define DIK_0               0x0B
#define DIK_MINUS           0x0C    /* - on main keyboard */
#define DIK_EQUALS          0x0D
#define DIK_BACK            0x0E    /* backspace */
#define DIK_TAB             0x0F
#define DIK_Q               0x10
#define DIK_W               0x11
#define DIK_E               0x12
#define DIK_R               0x13
#define DIK_T               0x14
#define DIK_Y               0x15
#define DIK_U               0x16
#define DIK_I               0x17
#define DIK_O               0x18
#define DIK_P               0x19
#define DIK_LBRACKET        0x1A
#define DIK_RBRACKET        0x1B
#define DIK_RETURN          0x1C    /* Enter on main keyboard */
#define DIK_LCONTROL        0x1D
#define DIK_A               0x1E
#define DIK_S               0x1F
#define DIK_D               0x20
#define DIK_F               0x21
#define DIK_G               0x22
#define DIK_H               0x23
#define DIK_J               0x24
#define DIK_K               0x25
#define DIK_L               0x26
#define DIK_SEMICOLON       0x27
#define DIK_APOSTROPHE      0x28
#define DIK_GRAVE           0x29    /* accent grave */
#define DIK_LSHIFT          0x2A
#define DIK_BACKSLASH       0x2B
#define DIK_Z               0x2C
#define DIK_X               0x2D
#define DIK_C               0x2E
#define DIK_V               0x2F
#define DIK_B               0x30
#define DIK_N               0x31
#define DIK_M               0x32
#define DIK_COMMA           0x33
#define DIK_PERIOD          0x34    /* . on main keyboard */
#define DIK_SLASH           0x35    /* / on main keyboard */
#define DIK_RSHIFT          0x36
#define DIK_MULTIPLY        0x37    /* * on numeric keypad */
#define DIK_LMENU           0x38    /* left Alt */
#define DIK_SPACE           0x39
#define DIK_CAPITAL         0x3A
#define DIK_F1              0x3B
#define DIK_F2              0x3C
#define DIK_F3              0x3D
#define DIK_F4              0x3E
#define DIK_F5              0x3F
#define DIK_F6              0x40
#define DIK_F7              0x41
#define DIK_F8              0x42
#define DIK_F9              0x43
#define DIK_F10             0x44
#define DIK_NUMLOCK         0x45
#define DIK_SCROLL          0x46    /* Scroll Lock */
#define DIK_NUMPAD7         0x47
#define DIK_NUMPAD8         0x48
#define DIK_NUMPAD9         0x49
#define DIK_SUBTRACT        0x4A    /* - on numeric keypad */
#define DIK_NUMPAD4         0x4B
#define DIK_NUMPAD5         0x4C
#define DIK_NUMPAD6         0x4D
#define DIK_ADD             0x4E    /* + on numeric keypad */
#define DIK_NUMPAD1         0x4F
#define DIK_NUMPAD2         0x50
#define DIK_NUMPAD3         0x51
#define DIK_NUMPAD0         0x52
#define DIK_DECIMAL         0x53    /* . on numeric keypad */
#define DIK_OEM_102         0x56    /* < > | on UK/Germany keyboards */
#define DIK_F11             0x57
#define DIK_F12             0x58

#define DIK_F13             0x64    /*                     (NEC PC98) */
#define DIK_F14             0x65    /*                     (NEC PC98) */
#define DIK_F15             0x66    /*                     (NEC PC98) */

#define DIK_KANA            0x70    /* (Japanese keyboard)            */
#define DIK_ABNT_C1         0x73    /* / ? on Portugese (Brazilian) keyboards */
#define DIK_CONVERT         0x79    /* (Japanese keyboard)            */
#define DIK_NOCONVERT       0x7B    /* (Japanese keyboard)            */
#define DIK_YEN             0x7D    /* (Japanese keyboard)            */
#define DIK_ABNT_C2         0x7E    /* Numpad . on Portugese (Brazilian) keyboards */
#define DIK_NUMPADEQUALS    0x8D    /* = on numeric keypad (NEC PC98) */
#define DIK_PREVTRACK       0x90    /* Previous Track (DIK_CIRCUMFLEX on Japanese keyboard) */
#define DIK_AT              0x91    /*                     (NEC PC98) */
#define DIK_COLON           0x92    /*                     (NEC PC98) */
#define DIK_UNDERLINE       0x93    /*                     (NEC PC98) */
#define DIK_KANJI           0x94    /* (Japanese keyboard)            */
#define DIK_STOP            0x95    /*                     (NEC PC98) */
#define DIK_AX              0x96    /*                     (Japan AX) */
#define DIK_UNLABELED       0x97    /*                        (J3100) */
#define DIK_NEXTTRACK       0x99    /* Next Track */
#define DIK_NUMPADENTER     0x9C    /* Enter on numeric keypad */
#define DIK_RCONTROL        0x9D
#define DIK_MUTE            0xA0    /* Mute */
#define DIK_CALCULATOR      0xA1    /* Calculator */
#define DIK_PLAYPAUSE       0xA2    /* Play / Pause */
#define DIK_MEDIASTOP       0xA4    /* Media Stop */
#define DIK_VOLUMEDOWN      0xAE    /* Volume - */
#define DIK_VOLUMEUP        0xB0    /* Volume + */
#define DIK_WEBHOME         0xB2    /* Web home */
#define DIK_NUMPADCOMMA     0xB3    /* , on numeric keypad (NEC PC98) */
#define DIK_DIVIDE          0xB5    /* / on numeric keypad */
#define DIK_SYSRQ           0xB7
#define DIK_RMENU           0xB8    /* right Alt */
#define DIK_PAUSE           0xC5    /* Pause */
#define DIK_HOME            0xC7    /* Home on arrow keypad */
#define DIK_UP              0xC8    /* UpArrow on arrow keypad */
#define DIK_PRIOR           0xC9    /* PgUp on arrow keypad */
#define DIK_LEFT            0xCB    /* LeftArrow on arrow keypad */
#define DIK_RIGHT           0xCD    /* RightArrow on arrow keypad */
#define DIK_END             0xCF    /* End on arrow keypad */
#define DIK_DOWN            0xD0    /* DownArrow on arrow keypad */
#define DIK_NEXT            0xD1    /* PgDn on arrow keypad */
#define DIK_INSERT          0xD2    /* Insert on arrow keypad */
#define DIK_DELETE          0xD3    /* Delete on arrow keypad */
#define DIK_LWIN            0xDB    /* Left Windows key */
#define DIK_RWIN            0xDC    /* Right Windows key */
#define DIK_APPS            0xDD    /* AppMenu key */
#define DIK_POWER           0xDE    /* System Power */
#define DIK_SLEEP           0xDF    /* System Sleep */
#define DIK_WAKE            0xE3    /* System Wake */
#define DIK_WEBSEARCH       0xE5    /* Web Search */
#define DIK_WEBFAVORITES    0xE6    /* Web Favorites */
#define DIK_WEBREFRESH      0xE7    /* Web Refresh */
#define DIK_WEBSTOP         0xE8    /* Web Stop */
#define DIK_WEBFORWARD      0xE9    /* Web Forward */
#define DIK_WEBBACK         0xEA    /* Web Back */
#define DIK_MYCOMPUTER      0xEB    /* My Computer */
#define DIK_MAIL            0xEC    /* Mail */
#define DIK_MEDIASELECT     0xED    /* Media Select */

	long i;
	for(i=0; i<MAX_KEYS_DINPUT; i++)
		m_MapBuffer[i]=255;

	/* Defined DIK_* constants */
	m_MapBuffer[DIK_ESCAPE] = SDLK_ESCAPE;
	m_MapBuffer[DIK_1] = SDLK_1;
	m_MapBuffer[DIK_2] = SDLK_2;
	m_MapBuffer[DIK_3] = SDLK_3;
	m_MapBuffer[DIK_4] = SDLK_4;
	m_MapBuffer[DIK_5] = SDLK_5;
	m_MapBuffer[DIK_6] = SDLK_6;
	m_MapBuffer[DIK_7] = SDLK_7;
	m_MapBuffer[DIK_8] = SDLK_8;
	m_MapBuffer[DIK_9] = SDLK_9;
	m_MapBuffer[DIK_0] = SDLK_0;
	m_MapBuffer[DIK_MINUS] = SDLK_MINUS;
	m_MapBuffer[DIK_EQUALS] = SDLK_EQUALS;
	m_MapBuffer[DIK_BACK] = SDLK_BACKSPACE;
	m_MapBuffer[DIK_TAB] = SDLK_TAB;
	m_MapBuffer[DIK_Q] = SDLK_q;
	m_MapBuffer[DIK_W] = SDLK_w;
	m_MapBuffer[DIK_E] = SDLK_e;
	m_MapBuffer[DIK_R] = SDLK_r;
	m_MapBuffer[DIK_T] = SDLK_t;
	m_MapBuffer[DIK_Y] = SDLK_y;
	m_MapBuffer[DIK_U] = SDLK_u;
	m_MapBuffer[DIK_I] = SDLK_i;
	m_MapBuffer[DIK_O] = SDLK_o;
	m_MapBuffer[DIK_P] = SDLK_p;
	m_MapBuffer[DIK_LBRACKET] = SDLK_LEFTBRACKET;
	m_MapBuffer[DIK_RBRACKET] = SDLK_RIGHTBRACKET;
	m_MapBuffer[DIK_RETURN] = SDLK_RETURN;
	m_MapBuffer[DIK_LCONTROL] = SDLK_LCTRL;
	m_MapBuffer[DIK_A] = SDLK_a;
	m_MapBuffer[DIK_S] = SDLK_s;
	m_MapBuffer[DIK_D] = SDLK_d;
	m_MapBuffer[DIK_F] = SDLK_f;
	m_MapBuffer[DIK_G] = SDLK_g;
	m_MapBuffer[DIK_H] = SDLK_h;
	m_MapBuffer[DIK_J] = SDLK_j;
	m_MapBuffer[DIK_K] = SDLK_k;
	m_MapBuffer[DIK_L] = SDLK_l;
	m_MapBuffer[DIK_SEMICOLON] = SDLK_SEMICOLON;
	m_MapBuffer[DIK_APOSTROPHE] = SDLK_QUOTE;
	m_MapBuffer[DIK_GRAVE] = SDLK_BACKQUOTE;
	m_MapBuffer[DIK_LSHIFT] = SDLK_LSHIFT;
	m_MapBuffer[DIK_BACKSLASH] = SDLK_BACKSLASH;
	m_MapBuffer[DIK_OEM_102] = SDLK_BACKSLASH;
	m_MapBuffer[DIK_Z] = SDLK_z;
	m_MapBuffer[DIK_X] = SDLK_x;
	m_MapBuffer[DIK_C] = SDLK_c;
	m_MapBuffer[DIK_V] = SDLK_v;
	m_MapBuffer[DIK_B] = SDLK_b;
	m_MapBuffer[DIK_N] = SDLK_n;
	m_MapBuffer[DIK_M] = SDLK_m;
	m_MapBuffer[DIK_COMMA] = SDLK_COMMA;
	m_MapBuffer[DIK_PERIOD] = SDLK_PERIOD;
	m_MapBuffer[DIK_SLASH] = SDLK_SLASH;
	m_MapBuffer[DIK_RSHIFT] = SDLK_RSHIFT;
	m_MapBuffer[DIK_MULTIPLY] = SDLK_KP_MULTIPLY;
	m_MapBuffer[DIK_LMENU] = SDLK_LALT;
	m_MapBuffer[DIK_SPACE] = SDLK_SPACE;
	m_MapBuffer[DIK_CAPITAL] = SDLK_CAPSLOCK;
	m_MapBuffer[DIK_F1] = SDLK_F1;
	m_MapBuffer[DIK_F2] = SDLK_F2;
	m_MapBuffer[DIK_F3] = SDLK_F3;
	m_MapBuffer[DIK_F4] = SDLK_F4;
	m_MapBuffer[DIK_F5] = SDLK_F5;
	m_MapBuffer[DIK_F6] = SDLK_F6;
	m_MapBuffer[DIK_F7] = SDLK_F7;
	m_MapBuffer[DIK_F8] = SDLK_F8;
	m_MapBuffer[DIK_F9] = SDLK_F9;
	m_MapBuffer[DIK_F10] = SDLK_F10;
	m_MapBuffer[DIK_NUMLOCK] = SDLK_NUMLOCK;
	m_MapBuffer[DIK_SCROLL] = SDLK_SCROLLOCK;
	m_MapBuffer[DIK_NUMPAD7] = SDLK_KP7;
	m_MapBuffer[DIK_NUMPAD8] = SDLK_KP8;
	m_MapBuffer[DIK_NUMPAD9] = SDLK_KP9;
	m_MapBuffer[DIK_SUBTRACT] = SDLK_KP_MINUS;
	m_MapBuffer[DIK_NUMPAD4] = SDLK_KP4;
	m_MapBuffer[DIK_NUMPAD5] = SDLK_KP5;
	m_MapBuffer[DIK_NUMPAD6] = SDLK_KP6;
	m_MapBuffer[DIK_ADD] = SDLK_KP_PLUS;
	m_MapBuffer[DIK_NUMPAD1] = SDLK_KP1;
	m_MapBuffer[DIK_NUMPAD2] = SDLK_KP2;
	m_MapBuffer[DIK_NUMPAD3] = SDLK_KP3;
	m_MapBuffer[DIK_NUMPAD0] = SDLK_KP0;
	m_MapBuffer[DIK_DECIMAL] = SDLK_KP_PERIOD;
	m_MapBuffer[DIK_F11] = SDLK_F11;
	m_MapBuffer[DIK_F12] = SDLK_F12;

	m_MapBuffer[DIK_F13] = SDLK_F13;
	m_MapBuffer[DIK_F14] = SDLK_F14;
	m_MapBuffer[DIK_F15] = SDLK_F15;
	m_MapBuffer[DIK_NUMPADEQUALS] = SDLK_KP_EQUALS;
	m_MapBuffer[DIK_NUMPADENTER] = SDLK_KP_ENTER;
	m_MapBuffer[DIK_RCONTROL] = SDLK_RCTRL;
	m_MapBuffer[DIK_DIVIDE] = SDLK_KP_DIVIDE;
	m_MapBuffer[DIK_SYSRQ] = SDLK_SYSREQ;
	m_MapBuffer[DIK_RMENU] = SDLK_RALT;
	m_MapBuffer[DIK_PAUSE] = SDLK_PAUSE;
	m_MapBuffer[DIK_HOME] = SDLK_HOME;
	m_MapBuffer[DIK_UP] = SDLK_UP;
	m_MapBuffer[DIK_PRIOR] = SDLK_PAGEUP;
	m_MapBuffer[DIK_LEFT] = SDLK_LEFT;
	m_MapBuffer[DIK_RIGHT] = SDLK_RIGHT;
	m_MapBuffer[DIK_END] = SDLK_END;
	m_MapBuffer[DIK_DOWN] = SDLK_DOWN;
	m_MapBuffer[DIK_NEXT] = SDLK_PAGEDOWN;
	m_MapBuffer[DIK_INSERT] = SDLK_INSERT;
	m_MapBuffer[DIK_DELETE] = SDLK_DELETE;
	m_MapBuffer[DIK_LWIN] = SDLK_LMETA;
	m_MapBuffer[DIK_RWIN] = SDLK_RMETA;
	m_MapBuffer[DIK_APPS] = SDLK_MENU;
// -------- WIN32 --------
#else
unsigned short i;
	if (DXActive==true)
	{
		for (i=0; i<MAX_KEYS_DINPUT; i++) m_MapBuffer[i]=i; // Preset
		return;
	}

	lpDIDKeyboard=NULL;

	for (i=0; i<MAX_KEYS_DINPUT; i++) m_MapBuffer[i]=255; // Preset
	m_MapBuffer[0x00]=0x0D;
	m_MapBuffer[0x01]=0x1B;
	m_MapBuffer[0x01]=0x1B;
	m_MapBuffer[0x02]=0x31;
	m_MapBuffer[0x03]=0x32;
	m_MapBuffer[0x04]=0x33;
	m_MapBuffer[0x05]=0x34;
	m_MapBuffer[0x06]=0x35;
	m_MapBuffer[0x07]=0x36;
	m_MapBuffer[0x08]=0x37;
	m_MapBuffer[0x09]=0x38;
	m_MapBuffer[0x0A]=0x39;
	m_MapBuffer[0x0B]=0x30;
	m_MapBuffer[0x0C]=0xDB;
	m_MapBuffer[0x0D]=0xDD;
	m_MapBuffer[0x0E]=0x08;
	m_MapBuffer[0x0F]=0x09;
	m_MapBuffer[0x10]=0x51;
	m_MapBuffer[0x11]=0x57;
	m_MapBuffer[0x12]=0x45;
	m_MapBuffer[0x13]=0x52;
	m_MapBuffer[0x14]=0x54;
	m_MapBuffer[0x15]=0x5A;
	m_MapBuffer[0x16]=0x55;
	m_MapBuffer[0x17]=0x49;
	m_MapBuffer[0x18]=0x4F;
	m_MapBuffer[0x19]=0x50;
	m_MapBuffer[0x1A]=0xBA;
	m_MapBuffer[0x1B]=0xBB;
	m_MapBuffer[0x1C]=0x0D;
	m_MapBuffer[0x1D]=0x11;
	m_MapBuffer[0x1E]=0x41;
	m_MapBuffer[0x1F]=0x53;
	m_MapBuffer[0x20]=0x44;
	m_MapBuffer[0x21]=0x46;
	m_MapBuffer[0x22]=0x47;
	m_MapBuffer[0x23]=0x48;
	m_MapBuffer[0x24]=0x4A;
	m_MapBuffer[0x25]=0x4B;
	m_MapBuffer[0x26]=0x4C;
	m_MapBuffer[0x27]=0xC0;
	m_MapBuffer[0x28]=0xDE;
	m_MapBuffer[0x29]=0xDC;
	m_MapBuffer[0x2A]=0x10;
	m_MapBuffer[0x2B]=0xBF;
	m_MapBuffer[0x2C]=0x59;
	m_MapBuffer[0x2D]=0x58;
	m_MapBuffer[0x2E]=0x43;
	m_MapBuffer[0x2F]=0x56;
	m_MapBuffer[0x30]=0x42;
	m_MapBuffer[0x31]=0x4E;
	m_MapBuffer[0x32]=0x4D;
	m_MapBuffer[0x33]=0xBC;
	m_MapBuffer[0x34]=0xBE;
	m_MapBuffer[0x35]=0xBD;
	m_MapBuffer[0x36]=0x10;
	m_MapBuffer[0x37]=0x6A;
	m_MapBuffer[0x38]=0x12;
	m_MapBuffer[0x39]=0x20;
	m_MapBuffer[0x3A]=0x00;
	m_MapBuffer[0x3B]=0x70;
	m_MapBuffer[0x3B]=0x70;
	m_MapBuffer[0x3C]=0x71;
	m_MapBuffer[0x3C]=0x71;
	m_MapBuffer[0x3D]=0x72;
	m_MapBuffer[0x3E]=0x73;
	m_MapBuffer[0x3F]=0x74;
	m_MapBuffer[0x40]=0x75;
	m_MapBuffer[0x41]=0x76;
	m_MapBuffer[0x42]=0x77;
	m_MapBuffer[0x43]=0x78;
	m_MapBuffer[0x44]=0x79;
	m_MapBuffer[0x45]=0x90;
	m_MapBuffer[0x46]=0x91;
	m_MapBuffer[0x46]=0x91;
	m_MapBuffer[0x47]=0x24;
	m_MapBuffer[0x48]=0x26;
	m_MapBuffer[0x49]=0x21;
	m_MapBuffer[0x4A]=0x6D;
	m_MapBuffer[0x4B]=0x25;
	m_MapBuffer[0x4C]=0x0C;
	m_MapBuffer[0x4D]=0x27;
	m_MapBuffer[0x4E]=0x6B;
	m_MapBuffer[0x4F]=0x23;
	m_MapBuffer[0x50]=0x28;
	m_MapBuffer[0x51]=0x22;
	m_MapBuffer[0x52]=0x2D;
	m_MapBuffer[0x53]=0x2E;
	m_MapBuffer[0x56]=0xE2;
	m_MapBuffer[0x57]=0x7A;
	m_MapBuffer[0x58]=0x7B;
// Num-Enter!?
	m_MapBuffer[0x9C]=0x1C;
	m_MapBuffer[0x9D]=0x11;
	m_MapBuffer[0xB5]=0x6F;
	m_MapBuffer[0xB8]=0x12;
	m_MapBuffer[0xC7]=0x24;
	m_MapBuffer[0xC8]=0x26;
	m_MapBuffer[0xC9]=0x21;
	m_MapBuffer[0xCB]=0x25;
	m_MapBuffer[0xCD]=0x27;
	m_MapBuffer[0xCF]=0x23;
	m_MapBuffer[0xD0]=0x28;
	m_MapBuffer[0xD1]=0x22;
	m_MapBuffer[0xD2]=0x2D;
	m_MapBuffer[0xD3]=0x2E;
	m_MapBuffer[0xDB]=0x5B;
	m_MapBuffer[0xDC]=0x5C;
	m_MapBuffer[0xDD]=0x5D;
// Numpad
	m_MapBuffer[0x47]=0x67;
	m_MapBuffer[0x48]=0x68;
	m_MapBuffer[0x49]=0x69;
	m_MapBuffer[0x4B]=0x64;
	m_MapBuffer[0x4C]=0x65;
	m_MapBuffer[0x4D]=0x66;
	m_MapBuffer[0x4F]=0x61;
	m_MapBuffer[0x50]=0x62;
	m_MapBuffer[0x51]=0x63;
	m_MapBuffer[0x52]=0x60;
	m_MapBuffer[0x53]=0x6E;
#endif

#endif // DXINPUT_HAVE_NO_KEYS
}


#else // console
	void glb_dxinput_console_dummy() {};
#endif // !GLB_CONSOLE