
#include "stdafx.h"
#include "dxsound.h"
#include "gffilesystem.h"

#ifdef GPHWIZ
#include <linux/soundcard.h>
#endif

#if defined(WIN32) && !defined(WANT_SDL_MIXER)
	#pragma comment(lib, "DSound.lib")

	#pragma pack(1)
	struct Waveheader
	{
		DWORD RIFF; // BYTE        RIFF[4];          // "RIFF"=0x52494646
		DWORD       dwSize;           // Size of data to follow
		DWORD WAVE; // BYTE        WAVE[4];          // "WAVE"=0x57415645
	};
	struct SubChunk1
	{
		DWORD ID; // BYTE        fmt_[4];          // "fmt " = 0x666D7420
		DWORD       Subchunk1Size;    // usually 16. Number of bytes between 'fmt ' and 'data' text tags (usually 16)
	};

	struct FmtChunk
	{
		WORD        AudioFormat;      // 1=PCM
		WORD        wChnls;           // Number of Channels
		DWORD       dwSRate;          // Sample Rate
		DWORD       BytesPerSec;      // Sample Rate [hz]
		WORD        wBlkAlign;        // 1
		WORD        BitsPerSample;    // Sample size
	};
	#pragma pack()


// -------------------------------------------- //
// SFXSound Constructor
// -------------------------------------------- //
SFXSound::SFXSound()
{
	m_lpDS=NULL;
}


SFXSound::~SFXSound()
{
	if (m_lpDS != NULL)
	{
		m_lpDS -> Release();	
	}
	m_lpDS=NULL;
}


// -------------------------------------------- //
// Create DirectSound Object
// -------------------------------------------- //
// COM must be intitialized!
BOOL SFXSound::Create (HWND hwnd)
{
HRESULT       hr;
	hr = DirectSoundCreate(NULL, &m_lpDS, NULL);
	if (hr != DS_OK) return FALSE;

	hr = m_lpDS->SetCooperativeLevel(hwnd, DSSCL_PRIORITY);
	if (hr != DS_OK) return FALSE;

	// Get the primary buffer 
LPDIRECTSOUNDBUFFER pDSBPrimary = NULL;
DSBUFFERDESC dsbd;
    ZeroMemory( &dsbd, sizeof(DSBUFFERDESC) );
    dsbd.dwSize        = sizeof(DSBUFFERDESC);
    dsbd.dwFlags       = DSBCAPS_PRIMARYBUFFER;
    dsbd.dwBufferBytes = 0;
    dsbd.lpwfxFormat   = NULL;
       
    if( FAILED( hr = m_lpDS->CreateSoundBuffer( &dsbd, &pDSBPrimary, NULL ) ) )
        return FALSE;

    // Set primary buffer format to 22kHz and 16-bit output.
WAVEFORMATEX wfx;
    ZeroMemory( &wfx, sizeof(WAVEFORMATEX) ); 
    wfx.wFormatTag      = WAVE_FORMAT_PCM; 
    wfx.nChannels       = 2; // Stereo
    wfx.nSamplesPerSec  = 22050;
    wfx.wBitsPerSample  = 16;
    wfx.nBlockAlign     = wfx.wBitsPerSample / 8 * wfx.nChannels;
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;

    if( FAILED( hr = pDSBPrimary->SetFormat(&wfx) ) )
        return FALSE;

    RELEASE(pDSBPrimary);
    GetCaps();
return TRUE;
}// Create

void SFXSound::SetWindow(HWND hwnd)
{
	if(m_lpDS)
		m_lpDS->SetCooperativeLevel(hwnd, DSSCL_PRIORITY);
}



// -------------------------------------------- //
// Get Capabilities of Sound Device
// -------------------------------------------- //
BOOL SFXSound::GetCaps(void)
{
HRESULT hr;
	if(!m_lpDS) return FALSE;
	m_DSCaps.dwSize = sizeof(DSCAPS); 
	hr = m_lpDS->GetCaps(&m_DSCaps); 
	
	if (hr != DS_OK) return FALSE;
return true;
}


// -------------------------------------------- //
// SFXSoundBuffer Constructor
// -------------------------------------------- //
SFXSoundBuffer::SFXSoundBuffer(void)
{
	m_pDS = NULL;
	for (int i=0; i<4; i++)
		m_lpDSB[i] = NULL;
	m_Current = 0;
	m_nBuffers = 0;
}

// -------------------------------------------- //
// SFXSoundBuffer Destructor
// -------------------------------------------- //
SFXSoundBuffer::~SFXSoundBuffer(void)
{
	FreeBuffer();
}

void SFXSoundBuffer::FreeBuffer(void)
{
	this->Stop();
	for(int i = 0; i<m_nBuffers; i++)
	{
		RELEASE(m_lpDSB[i]);
		m_lpDSB[i] = NULL;
	}
	m_Current = 0;
	m_nBuffers = 0;
}


void SFXSoundBuffer::Init(void)
{m_nBuffers=m_Current=0; m_Filename=NULL; m_pDS=NULL;}

// -------------------------------------------- //
// SFXSoundBuffer CreateSoundBuffer
// -------------------------------------------- //
BOOL SFXSoundBuffer::CreateSoundBuffer(DWORD dwBufSize, DWORD dwFreq, DWORD dwBitsPerSample, DWORD dwBlkAlign, BOOL bStereo)
{
	PCMWAVEFORMAT pcmwf;
	DSBUFFERDESC dsbdesc;

	// Set up wave format structure.
	memset(&pcmwf, 0, sizeof(PCMWAVEFORMAT));
	pcmwf.wf.wFormatTag         = WAVE_FORMAT_PCM;      
	pcmwf.wf.nChannels          = bStereo ? 2 : 1;
	pcmwf.wf.nSamplesPerSec     = dwFreq;
	pcmwf.wf.nBlockAlign        = (WORD)dwBlkAlign;
	pcmwf.wf.nAvgBytesPerSec    = pcmwf.wf.nSamplesPerSec * pcmwf.wf.nBlockAlign;
	pcmwf.wBitsPerSample        = (WORD)dwBitsPerSample;

	// Set up DSBUFFERDESC structure.
	memset(&dsbdesc, 0, sizeof(DSBUFFERDESC));
	dsbdesc.dwSize              = sizeof(DSBUFFERDESC);
	dsbdesc.dwFlags             = DSBCAPS_STATIC | DSBCAPS_GETCURRENTPOSITION2
								  |DSBCAPS_CTRLPAN|DSBCAPS_CTRLVOLUME;
	dsbdesc.dwBufferBytes       = dwBufSize; 
	dsbdesc.lpwfxFormat         = (LPWAVEFORMATEX)&pcmwf;

	if (m_pDS->m_lpDS->CreateSoundBuffer(&dsbdesc, &m_lpDSB[0], NULL)!=DS_OK)
	{
		dsbdesc.dwFlags = DSBCAPS_STATIC | DSBCAPS_GETCURRENTPOSITION2; // No PAN, Vol..
		if (m_pDS->m_lpDS->CreateSoundBuffer(&dsbdesc, &m_lpDSB[0], NULL)!=DS_OK)
		{
			m_pDS=NULL;
			return FALSE;
		}
	}

	return TRUE;
}

// -------------------------------------------- //
// SFXSoundBuffer ReadData
// -------------------------------------------- //
BOOL SFXSoundBuffer::ReadData(GFFileSystem& fp, DWORD dwSize, DWORD dwPos)
{
	// Seek to correct position in file (if necessary)
	if (dwPos != 0xffffffff) 
	{
		if (!fp.SeekTo(dwPos)) return FALSE;
	}

// Lock data in buffer for writing
LPVOID pData1;
DWORD  dwData1Size;
LPVOID pData2;
DWORD  dwData2Size;
HRESULT rval;

	rval = m_lpDSB[0]->Lock(0, dwSize, &pData1, &dwData1Size, &pData2, &dwData2Size, DSBLOCK_FROMWRITECURSOR);
	if (rval != DS_OK)
	{
		return FALSE;
	}

	// Read in first chunk of data
	if (dwData1Size > 0) 
	{
		if (!fp.Read(pData1, dwData1Size))
		{
			return FALSE;
		}
	}

	// Read in second chunk if necessary
	if (dwData2Size > 0) 
	{
		if (!fp.Read(pData2, dwData2Size))
			return FALSE;
	}

	// Unlock data in buffer
	rval = m_lpDSB[0]->Unlock(pData1, dwData1Size, pData2, dwData2Size);
	if (rval != DS_OK) return FALSE;

	return TRUE;
}

// -------------------------------------------- //
// SFXSoundBuffer GetFreeBuffer
// -------------------------------------------- //
LPDIRECTSOUNDBUFFER SFXSoundBuffer::GetFreeBuffer(void)
{
DWORD				Status;
HRESULT				rval;
LPDIRECTSOUNDBUFFER	Buffer;
	if(m_lpDSB == NULL) return NULL;
	if(Buffer = m_lpDSB[m_Current])
	{
		rval = Buffer->GetStatus(&Status);
		if(rval < 0) Status = 0;
		if((Status & DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING)
		{
			if(m_nBuffers > 1)
			{
				if (++m_Current >= m_nBuffers) m_Current = 0;
				Buffer = m_lpDSB[m_Current];
				rval = Buffer->GetStatus(&Status);
				if((rval >= 0) && (Status & DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING)
				{
					Buffer->Stop();
					Buffer->SetCurrentPosition(0);
				}
			}
			else
			{
				Buffer = NULL;
			}
		}
		if(Buffer && (Status & DSBSTATUS_BUFFERLOST))
		{
			if((Buffer->Restore() < 0))
			{
				Buffer = NULL;
			}
		}
	}
return Buffer;
}

// -------------------------------------------- //
// SFXSoundBuffer LoadFromFile
// http://www.sonicspot.com/guide/wavefiles.html <-- FINALLY!
// -------------------------------------------- //
bool SFXSoundBuffer::LoadFromFile(void)
{
// Open the wave file       
GFFileSystem	file;
	if (!file.OpenIn(m_Filename)) return false;
	// FILE* fp = fopen(m_Filename, "rb"); if (fp == NULL) return false;

	// Read in the wave header          
Waveheader Wavhdr;
	if (!file.Read(&Wavhdr, sizeof(Wavhdr)))
	{
		file.Close();
		return false;
	}
	if(Wavhdr.RIFF!=0x46464952) return false;
	if(Wavhdr.WAVE!=0x45564157) return false;

	// read sub1 chunks, try to find a "FMT" chunk
	SubChunk1 ch1;
	while(file.Read(&ch1, sizeof(ch1)) && ch1.ID!=0x20746D66)
	{
		file.SeekTo(ch1.Subchunk1Size, GFFileSystem::seek_current);
	}
	if(ch1.ID!=0x20746D66)
		return false;
FmtChunk fmt;
	if(!file.Read(&fmt, sizeof(fmt)))
		return false;
	// skip extra fmt bytes
	if(ch1.Subchunk1Size>16)
	{
		file.SeekTo(ch1.Subchunk1Size - 16, GFFileSystem::seek_current);
	}

	// find the DATA chunk
	while(file.Read(&ch1, sizeof(ch1)) && ch1.ID!=0x61746164)
	{
		file.SeekTo(ch1.Subchunk1Size, GFFileSystem::seek_current);
	}
	if(ch1.ID!=0x61746164)
		return false;

	// Is this a stereo or mono file?
	BOOL bStereo = fmt.wChnls > 1 ? TRUE : FALSE;
	// Create the sound buffer for the wave file
	if (!CreateSoundBuffer(ch1.Subchunk1Size, fmt.dwSRate, fmt.BitsPerSample, fmt.wBlkAlign, bStereo))
	{
		// Close the file
		file.Close();
		return false;
	}

	// Read the data for the wave file into the sound buffer
	if (!ReadData(file, ch1.Subchunk1Size, 0xffffffff)) 
	{
		file.Close();
		return false;
	}
	// Close out the wave file
	file.Close();
return true;
}

// -------------------------------------------- //
// SFXSoundBuffer Clear
// -------------------------------------------- //
void SFXSoundBuffer::Clear()
{
	FreeBuffer();
	Init();
}

// -------------------------------------------- //
// SFXSoundBuffer Load
// -------------------------------------------- //
bool SFXSoundBuffer::Load(SFXSound* pDS, const char *Filename, int Num)
{
	// Free before reload
	Clear();

	m_pDS = pDS;
	if (Num<1) Num=1;
	if (Num>4) Num=4;
	m_nBuffers = Num;
	m_Filename = Filename;

	if(!m_pDS || !m_pDS->m_lpDS || !LoadFromFile()) // Try Load to buffer[0]
	{
		m_nBuffers=0;
		m_Filename=NULL;
		return false;
	}

	// Try duplicating if possible
	for(int i = 1; i < m_nBuffers; i++)
	{
		m_pDS->m_lpDS->DuplicateSoundBuffer(m_lpDSB[0], &m_lpDSB[i]);
	}
return true;
}

// -------------------------------------------- //
// SFXSoundBuffer Play
// volume : 0..100 (%) - linear
// -------------------------------------------- //
long SFXSoundBuffer::Play(int Pan, int volume, DWORD dwFlags)
{
	if(!m_pDS || m_nBuffers==0 || !m_pDS->m_lpDS)
	{
		return -1;
	}
HRESULT rval;
LPDIRECTSOUNDBUFFER Buffer = NULL;

	Buffer = GetFreeBuffer();
	if(Buffer == NULL)
	{
		return -1;
	}

	int vol_db = -10000; // silent == DSBVOLUME_MIN
	if (volume > 0)
		vol_db = (int)(20.0 * 100.0 * log10((double)volume/100.0));

	// int vol = ((DSBVOLUME_MAX-DSBVOLUME_MIN) / 100) * volume  + DSBVOLUME_MIN;

	Buffer->SetVolume( vol_db );
	Buffer->SetPan(Pan);

	rval = Buffer->Play(0, 0, dwFlags);
	if(rval != DS_OK) return -1;
return m_Current;
}

// -------------------------------------------- //
// SFXSoundBuffer IsPlaying
// -------------------------------------------- //
bool SFXSoundBuffer::IsPlaying(long channel)
{
	if (channel<0 || channel>=m_nBuffers) return false;
	
LPDIRECTSOUNDBUFFER	Buffer;
DWORD				Status;
HRESULT				rval;
	Buffer = m_lpDSB[channel];
	rval = Buffer->GetStatus(&Status);
	if(rval < 0) Status = 0;
	if((Status & DSBSTATUS_PLAYING) == DSBSTATUS_PLAYING)
	{
		return true;
	}
return false;
}


// -------------------------------------------- //
// SFXSoundBuffer Stop
// -------------------------------------------- //
BOOL SFXSoundBuffer::Stop(int channel)
{
	if(!m_pDS || m_nBuffers==0 || !m_pDS->m_lpDS) return FALSE;
HRESULT rval;

	if(channel<0 || channel>=m_nBuffers)
	{
		for(int i = 0; i < m_nBuffers; i++)
		{
			rval = m_lpDSB[i]->Stop();
			if(rval != DS_OK) return false;

			rval = m_lpDSB[i]->SetCurrentPosition(0);
			if(rval != DS_OK) return false;
		}
	}
	else
	{
		rval = m_lpDSB[channel]->Stop();
		if(rval != DS_OK) return false;

		rval = m_lpDSB[channel]->SetCurrentPosition(0);
		if(rval != DS_OK) return false;
	}
return true;
}

// -------------------------------------------- //
// SFXSoundBuffer SetVolume
// -------------------------------------------- //
void SFXSoundBuffer::SetVolume(LONG Volume)
{
double vol = (double)DSBVOLUME_MAX * (double) Volume / 100.0;

	if(!m_pDS || m_nBuffers==0 || !m_pDS->m_lpDS) return;
	for(int i = 0; i < m_nBuffers; i++)
		m_lpDSB[i]->SetVolume((LONG)vol);
}


#elif defined(WANT_SDL_MIXER)
//////////////////////////////////////////////////////////////////////////////////
//             S D L
//////////////////////////////////////////////////////////////////////////////////

SFXSound::SFXSound()
{
/*
#if defined(GPHWIZ)

#define SEE(a) printf("#define %-20s 0x%x\n", #a, a);

SEE( SNDCTL_DSP_SETFMT      );
SEE( SNDCTL_DSP_GETBLKSIZE  );
SEE( SNDCTL_DSP_CHANNELS    );
SEE( SNDCTL_DSP_SPEED       );
SEE( SNDCTL_DSP_SAMPLESIZE  );
SEE( SNDCTL_DSP_STEREO      );
SEE( SOUND_PCM_READ_BITS    );
SEE( AFMT_S16_NE    );


#endif

*/

	m_IsSetup=0L;
/*
	SDL_AudioSpec format;
	memset(&sounds, 0, sizeof(NOW_PLAYING)*NUM_SOUNDS);

	// Format: 16 Bit, stereo, 22 KHz
	format.freq = 22050;
	format.format = AUDIO_S16;
	format.channels = 2;
	format.samples = 512;        // good val
	format.callback = mixeAudio;
	format.userdata = NULL;
*/

int channels = 2;
int buffers=4096;
int frequency = 44100;
short format = MIX_DEFAULT_FORMAT;


#if defined(GP2X)
	frequency = 22050;
	buffers=256;
	format = AUDIO_U8;
#endif

#if defined(GPHWIZ)
	buffers=256;
	frequency = 22050;
	format = MIX_DEFAULT_FORMAT; // AUDIO_U16SYS;
#endif

#if defined(CAANOO)
	buffers=256;
	frequency = 22050;
	format = MIX_DEFAULT_FORMAT; // AUDIO_U16SYS;
#endif

#if defined(PALM_WEBOS) || defined(PALM_PIXI)
	frequency = 44100;
	buffers = 1024;
	//channels=1;
	format = MIX_DEFAULT_FORMAT;
#endif

#if defined(ANDROID)
	frequency = 44100;
	// 512 makes SFXR sounds sound distorted!
	buffers = 1024; // 512; // 1024 - the higher, the longer the pause. Oh, bad!
	//channels=1;
	format = MIX_DEFAULT_FORMAT;
#endif

	if(Mix_OpenAudio(frequency, format, channels, buffers)==0)
	{
		#if !defined(EMSCRIPTEN)
		Mix_Volume(-1, MIX_MAX_VOLUME);
		#endif
		Mix_VolumeMusic(MIX_MAX_VOLUME/2);
		Mix_AllocateChannels(16);
	//	SDL_PauseAudio(0);
		m_IsSetup=1;
		g_debug("open sound: %d Hz %d chan, buffers %d, format %d\n", frequency, channels, buffers, format);
	}
	else
	{
		printf("Can't init audio\n");
	}
}


SFXSound::~SFXSound()
{
	Mix_CloseAudio();
}

SFXSoundBuffer::SFXSoundBuffer(void)
{
	m_cvt=NULL;
}
SFXSoundBuffer::~SFXSoundBuffer(void)
{
	FreeBuffer();
}

void SFXSoundBuffer::Init(void)
{
	FreeBuffer();
}

// -------------------------------------------- //
// SFXSoundBuffer Clear
// -------------------------------------------- //
void SFXSoundBuffer::Clear()
{
	FreeBuffer();
	Init();
}


bool SFXSoundBuffer::Load(SFXSound* pDS, const char* Filename, int Num)
{
	Clear();
	m_cvt = Mix_LoadWAV(Filename);
	// Mix_LoadWAV_RW(SDL_RWops *src, int freesrc); <-- from memory block
	return m_cvt ? true : false;
}

// -------------------------------------------- //
// SFXSoundBuffer IsPlaying
// -------------------------------------------- //
bool SFXSoundBuffer::IsPlaying(long channel)
{
	if (!m_cvt) return false;
	if(channel==-1) return false;
	return Mix_Playing(channel) ? true:false;
}

long SFXSoundBuffer::Play(int Pan, int Volume, unsigned long dwFlags)
{
	if (!m_cvt)
	{
		printf("play: m_cvt=NULL\n");
		return -1;
	}
	Mix_VolumeChunk(m_cvt, (Volume*MIX_MAX_VOLUME)/100);
	long channel = Mix_PlayChannel(-1, m_cvt, 0);
return channel;
}

long SFXSoundBuffer::Stop(int channel)
{
	if (!m_cvt) return TRUE;
	int rv = Mix_HaltChannel(channel); // -1 is all
	return TRUE;
}

void SFXSoundBuffer::FreeBuffer()
{
	if (m_cvt)
		Mix_FreeChunk(m_cvt);
	m_cvt=NULL;
}

// ################################################
#elif defined(_WIN32_WCE)

	/*
	[DllImport("aygshell.dll")]

	internal static extern uint SndPlaySync(LPCTSTR file, DWORD flags);
	internal static extern uint SndOpen(LPCTSTR file, int* phSound);
	internal static extern uint SndPlayAsync(int* hSound, DWORD flags);
	internal static extern uint SndClose(int* hSound);
	internal static extern uint SndStop(int SoundScope, int* hSound);
	*/

	static HINSTANCE ghDLL=NULL;
	#define LINK(name,import) name = m_hDll ? (PFN##name)GetProcAddress( m_hDll, TEXT(import) ) : NULL;

	#define TYPDEF(rv, name, args) typedef rv ( __stdcall* PFN##name) args; PFN##name name = NULL;

	TYPDEF(DWORD,SndPlaySync, (LPCTSTR file, DWORD flags));
	TYPDEF(DWORD,SndOpen,     (LPCTSTR file, int* phSound));
	TYPDEF(DWORD,SndPlayAsync,(int* hSound, DWORD flags));
	TYPDEF(DWORD,SndClose,    (int* hSound));
	TYPDEF(DWORD,SndStop,     (int SoundScope, int* hSound));
	TYPDEF(HRESULT,SndGetWaitHandle, (int* hSound, HANDLE* phDoneEvent));

	BOOL SFXSound::Create(HWND hwnd)
	{
		if(!ghDLL)
		{
			ghDLL = LoadLibrary(TEXT("aygshell.dll"));
			if(!ghDLL) return FALSE;
			LINK(SndPlaySync, "SndPlaySync");
			LINK(SndOpen,     "SndOpen");
			LINK(SndPlayAsync,"SndPlayAsync");
			LINK(SndClose,    "SndClose");
			LINK(SndStop,     "SndStop");
		}
	return TRUE;
	}


	BOOL SFXSound::GetCaps(void)
	{
		return SndOpen ? TRUE:FALSE;
	}

	SFXSoundBuffer::SFXSoundBuffer(void){m_pSound=NULL;}
	SFXSoundBuffer::~SFXSoundBuffer(void)
	{
		if(m_pSound) SndClose(m_pSound);
		m_psound=NULL;
	}

	void SFXSoundBuffer::Init(void)
	{
		if(m_pSound) SndClose(m_pSound);
		m_psound=NULL;
	}

	void SFXSoundBuffer::Clear()
	{
		Init();
	}

	bool SFXSoundBuffer::Load(SFXSound* pDS, const char* Filename, int Num)
	{
	char fullpath[512];
		GFFileSystem::MakeAbsolutePathCE(Filename, fullpath);
		wchar_t path[512];
		mbstowcs(path, fullpath, strlen(fullpath)+1);
		// SndOpen
		return false;
	}
	long SFXSoundBuffer::Play(int Pan = 0, int Volume=100, unsigned long dwFlags = 0);
	bool SFXSoundBuffer::IsPlaying(long channel);
	long SFXSoundBuffer::Stop(void);


#elif defined(APPLE_COREAUDIO)

#ifdef MACOSX

	enum {
					kSoundEngineErrUnitialized                      = 1,
					kSoundEngineErrInvalidID                        = 2,
					kSoundEngineErrFileNotFound                     = 3,
					kSoundEngineErrInvalidFileFormat        = 4,
					kSoundEngineErrDeviceNotFound           = 5
	};
	#include "glb.h"
	#include <CoreAudio/CoreAudioTypes.h>
	#define DYLIB "SoundEngine.dylib"
	DECLARE_C(glbSoundEngine_Initialize, DYLIB, (Float32 inMixerOutputRate), OSStatus );
	DECLARE_C(glbSoundEngine_Teardown, DYLIB, (), OSStatus );
	DECLARE_C(glbSoundEngine_SetMasterVolume, DYLIB, (Float32 inValue), OSStatus );
	DECLARE_C(glbSoundEngine_SetListenerPosition, DYLIB, (Float32 inX, Float32 inY, Float32 inZ), OSStatus );
	DECLARE_C(glbSoundEngine_SetListenerGain, DYLIB, (Float32 inValue), OSStatus );
	DECLARE_C(glbSoundEngine_LoadBackgroundMusicTrack, DYLIB, (const char* inPath, Boolean inAddToQueue, Boolean inLoadAtOnce), OSStatus );
	DECLARE_C(glbSoundEngine_UnloadBackgroundMusicTrack, DYLIB, (), OSStatus );
	DECLARE_C(glbSoundEngine_StartBackgroundMusic, DYLIB, (), OSStatus );
	DECLARE_C(glbSoundEngine_IsBackgroundMusicPlaying, DYLIB, (), int );
	DECLARE_C(glbSoundEngine_StopBackgroundMusic, DYLIB, (Boolean inStopAtEnd), OSStatus );
	DECLARE_C(glbSoundEngine_PauseBackgroundMusic, DYLIB, (Boolean bPause), void );
	DECLARE_C(glbSoundEngine_SetBackgroundMusicVolume, DYLIB, (Float32 inValue), OSStatus );
	DECLARE_C(glbSoundEngine_LoadLoopingEffect, DYLIB, (const char* inLoopFilePath, const char* inAttackFilePath, const char* inDecayFilePath, UInt32* outEffectID), OSStatus );
	DECLARE_C(glbSoundEngine_LoadEffect, DYLIB, (const char* inPath, UInt32* outEffectID), OSStatus );
	DECLARE_C(glbSoundEngine_UnloadEffect, DYLIB, (UInt32 inEffectID), OSStatus );
	DECLARE_C(glbSoundEngine_StartEffect, DYLIB, (UInt32 inEffectID), OSStatus );
	DECLARE_C(glbSoundEngine_IsEffectPlaying, DYLIB, (UInt32 inEffectID), int );
	DECLARE_C(glbSoundEngine_StopEffect, DYLIB, (UInt32 inEffectID, Boolean inDoDecay), OSStatus );
	DECLARE_C(glbSoundEngine_SetEffectPitch, DYLIB, (UInt32 inEffectID, Float32 inValue), OSStatus );
	DECLARE_C(glbSoundEngine_SetEffectLevel, DYLIB, (UInt32 inEffectID, Float32 inValue), OSStatus );
	DECLARE_C(glbSoundEngine_SetEffectPosition, DYLIB, (UInt32 inEffectID, Float32 inX, Float32 inY, Float32 inZ), OSStatus );
	DECLARE_C(glbSoundEngine_SetEffectsVolume, DYLIB, (Float32 inValue), OSStatus );
	DECLARE_C(glbSoundEngine_SetMaxDistance, DYLIB, (Float32 inValue), OSStatus );
	DECLARE_C(glbSoundEngine_SetReferenceDistance, DYLIB, (Float32 inValue), OSStatus );

#else
	#include "iphonesoundengine.h"
#endif

	SFXSound::SFXSound(void)
	{
		glbSoundEngine_Initialize(0);
		glbSoundEngine_SetListenerPosition(0.0, 0.0, 0.0f);
	}
	SFXSound::~SFXSound(void)
	{
		glbSoundEngine_Teardown();
	}
	BOOL SFXSound::Create(HWND hwnd)	{return TRUE;}
	BOOL SFXSound::GetCaps(void)		{return TRUE;}

	SFXSoundBuffer::SFXSoundBuffer()		{for(int i=0; i<GLB_OPENAL_BUFFERS; ++i) effects[i]=0; }
	SFXSoundBuffer::~SFXSoundBuffer()		{Clear();}
	void SFXSoundBuffer::Init(void)			{Clear();}

	void SFXSoundBuffer::Clear()
	{
		for(int i=0; i<GLB_OPENAL_BUFFERS; ++i)
			if(effects[i])
			{
				glbSoundEngine_UnloadEffect(effects[i]);
				effects[i]=0;
			}
	}

	bool SFXSoundBuffer::Load(SFXSound* pDS, const char* Filename, int Num)
	{
		Clear();
		if(Num>GLB_OPENAL_BUFFERS) Num=GLB_OPENAL_BUFFERS;
		for(int i=0; i<Num; ++i)
		{
			OSStatus rv = glbSoundEngine_LoadEffect(Filename, &effects[i]);
			printf("loadeffect - i: %d eff[i]: %d\n", i, (int)effects[i]);
			if(rv != noErr )
			{
				g_debug("Can' init audio. Retcode: %d\n", (int)rv);
				Clear();
				return false;
			}
		}
		return true;
	}
	
	long SFXSoundBuffer::Play(int Pan, int Volume, unsigned long dwFlags)
	{
		for(int i=0; i<GLB_OPENAL_BUFFERS; ++i)
		{
			if(effects[i] && !glbSoundEngine_IsEffectPlaying(effects[i]))
			{
				glbSoundEngine_SetEffectLevel(effects[i], (float)Volume / 100.0f);
				glbSoundEngine_SetEffectPosition(effects[i], (float)Pan / (float)DXSOUND_PAN_RIGHT, 0.0f, 0.0f);
				glbSoundEngine_StartEffect(effects[i]);
				// printf("play - i: %d eff[i]: %d\n", i, effects[i]);
				return i; // buffer index
			}
			else
			{
				// printf("no play2 - i: %d eff[i]: %d (isply: %d)\n", i, effects[i], glbSoundEngine_IsEffectPlaying(effects[i]));
			}
		}
		return -1; // no free buffer
	}

	bool SFXSoundBuffer::IsPlaying(long channel)
	{
		if(channel<0 || channel>=GLB_OPENAL_BUFFERS) return false;
		return glbSoundEngine_IsEffectPlaying(effects[channel]) ? true:false;
	}
	long SFXSoundBuffer::Stop(int channel)
	{
		if(channel<0 || channel>=GLB_OPENAL_BUFFERS) return 1;
		glbSoundEngine_StopEffect(effects[channel], false);
		return 1;
	}
	
	
	bool iPhonePlayMusic(const char* pFilename, bool bLoop)
	{
		glbSoundEngine_UnloadBackgroundMusicTrack(); // full stop

		glbSoundEngine_UnloadBackgroundMusicTrack(); // just quit it and reset the position when you re-create the instance
		int rv = (noErr == glbSoundEngine_LoadBackgroundMusicTrack(pFilename, false, false)
			 && noErr == glbSoundEngine_StartBackgroundMusic());
		if(rv && !bLoop)
			glbSoundEngine_StopBackgroundMusic(true); // only play once
        if(rv) return true;
        return false;
	}

	void iPhonePauseMusic(bool bPause)
	{
		glbSoundEngine_PauseBackgroundMusic(bPause);
	}
	bool iPhoneIsMusicPlaying()
	{
		return glbSoundEngine_IsBackgroundMusicPlaying();
	}
	void iPhoneStopMusic()
	{
		glbSoundEngine_UnloadBackgroundMusicTrack(); // full stop
	}
	void iPhoneMusicVolume(float vol)
	{
		glbSoundEngine_SetBackgroundMusicVolume(vol);
	}
	
	

#elif defined(WANT_SDL_WAVSOUND)


#include <SDL/SDL.h>

#define NUM_CHANNELS 8				     /* max number of sounds we can play at once */

void audioCallback(void *userdata, Uint8 * stream, int len);

struct
{
    /* channel array holds information about currently playing sounds */
    struct
    {
        Uint8 *position;        /* what is the current position in the buffer of this sound ? */
        Uint32 remaining;       /* how many bytes remaining before we're done playing the sound ? */
        Uint32 timestamp;       /* when did this sound start playing ? */
    } channels[NUM_CHANNELS];
    SDL_AudioSpec outputSpec;   /* what audio format are we using for output? */
    int numSoundsPlaying;       /* how many sounds are currently playing */
} mixer;


SFXSound::SFXSound(void)
{
    SDL_memset(&mixer, 0, sizeof(mixer));
}

SFXSound::~SFXSound(void)
{

}



BOOL SFXSound::Create(HWND hwnd)
{
    // if (SDL_Init(SDL_INIT_AUDIO) < 0) - global call now

    /* initialize the mixer */
    SDL_memset(&mixer, 0, sizeof(mixer));
    /* setup output format */
    mixer.outputSpec.freq = 22050;
    mixer.outputSpec.format = AUDIO_S16LSB;
    mixer.outputSpec.channels = 2;
    mixer.outputSpec.samples = 256;
    mixer.outputSpec.callback = audioCallback;
    mixer.outputSpec.userdata = NULL;

    /* open audio for output */
    if (SDL_OpenAudio(&mixer.outputSpec, NULL) != 0)
    {
        printf("Opening audio failed\n");
        return FALSE;
    }
}
BOOL SFXSound::GetCaps(void)
{
	return TRUE;
}



SFXSoundBuffer::SFXSoundBuffer()
{
	buffer=NULL; length=0;
}

SFXSoundBuffer::~SFXSoundBuffer()
{
	if(buffer) Clear();
}

void SFXSoundBuffer::Init(void)
{
	if(buffer) Clear();
}

void SFXSoundBuffer::Clear()
{
	SDL_LockAudio();
	if(buffer)
	{
		SDL_free(buffer);
	}
	buffer=NULL; length=0;
	SDL_UnlockAudio();
}

bool SFXSoundBuffer::Load(SFXSound* pDS, const char* Filename, int Num)
{
	Init();
    SDL_AudioSpec spec;         /* the audio format of the .wav file */
    SDL_AudioCVT cvt;           /* used to convert .wav to output format when formats differ */
    int result;
    if (SDL_LoadWAV(Filename, &spec, &this->buffer, &this->length) == NULL)
    {
        printf("could not load .wav %s\n", Filename);
        return false;
    }
    /* build the audio converter */
    result = SDL_BuildAudioCVT(&cvt, spec.format, spec.channels, spec.freq,
                               mixer.outputSpec.format,
                               mixer.outputSpec.channels,
                               mixer.outputSpec.freq);
    if (result == -1)
    {
        printf("could not build audio CVT\n");
        return false;
    }
    else if (result != 0)
    {
        /* 
           this happens when the .wav format differs from the output format.
           we convert the .wav buffer here
         */
        cvt.buf = (Uint8 *) SDL_malloc(this->length * cvt.len_mult);       /* allocate conversion buffer */
        cvt.len = this->length;    /* set conversion buffer length */
        SDL_memcpy(cvt.buf, this->buffer, this->length);      /* copy sound to conversion buffer */
        if (SDL_ConvertAudio(&cvt) == -1)
        {     /* convert the sound */
            printf("could not convert .wav %s\n", Filename);
        }
        SDL_free(this->buffer);    /* free the original (unconverted) buffer */
        this->buffer = cvt.buf;    /* point sound buffer to converted buffer */
        this->length = cvt.len_cvt;        /* set sound buffer's new length */
    }
    return true;
}

long SFXSoundBuffer::Play(int Pan, int Volume, unsigned long dwFlags)
{
	SDL_LockAudio();
    /*
       find an empty channel to play on.
       if no channel is available, use oldest channel
     */
    int i;
    int selected_channel = -1;
    int oldest_channel = 0;

    if (mixer.numSoundsPlaying == 0) {
        /* we're playing a sound now, so start audio callback back up */
        SDL_PauseAudio(0);
    }

    /* find a sound channel to play the sound on */
    for (i = 0; i < NUM_CHANNELS; i++)
    {
        if (mixer.channels[i].position == NULL) {
            /* if no sound on this channel, select it */
            selected_channel = i;
            break;
        }
        /* if this channel's sound is older than the oldest so far, set it to oldest */
        if (mixer.channels[i].timestamp <
            mixer.channels[oldest_channel].timestamp)
            oldest_channel = i;
    }

    /* no empty channels, take the oldest one */
    if (selected_channel == -1)
        selected_channel = oldest_channel;
    else
        mixer.numSoundsPlaying++;

    /* point channel data to wav data */
    mixer.channels[selected_channel].position = this->buffer;
    mixer.channels[selected_channel].remaining = this->length;
    mixer.channels[selected_channel].timestamp = SDL_GetTicks();
	SDL_UnlockAudio();

    return selected_channel;
}

bool SFXSoundBuffer::IsPlaying(long channel)
{
	int rv=0;
	SDL_LockAudio();
	if(	channel>0 && channel<NUM_CHANNELS
		&& mixer.numSoundsPlaying && mixer.channels[channel].position)
		rv=1;
	SDL_UnlockAudio();
	return rv;
}


long SFXSoundBuffer::Stop(int channel)
{
	SDL_LockAudio();
	if(	channel>0 && channel<NUM_CHANNELS && mixer.numSoundsPlaying && mixer.channels[channel].position)
	{
		mixer.channels[channel].position = NULL;
		--mixer.numSoundsPlaying;
    }
    else
    {
		for (int i = 0; i < NUM_CHANNELS; i++)
		{
			mixer.channels[i].position = NULL;
		}
		mixer.numSoundsPlaying=0;
    }
	SDL_UnlockAudio();
}



/* 
	Called from SDL's audio system.  Supplies sound input with data by mixing together all
	currently playing sound effects.
*/
void audioCallback(void *userdata, Uint8 * stream, int len)
{
    int i;
    int copy_amt;
    SDL_memset(stream, mixer.outputSpec.silence, len);  /* initialize buffer to silence */
    /* for each channel, mix in whatever is playing on that channel */
    for (i = 0; i < NUM_CHANNELS; i++) {
        if (mixer.channels[i].position == NULL) {
            /* if no sound is playing on this channel */
            continue;           /* nothing to do for this channel */
        }

        /* copy len bytes to the buffer, unless we have fewer than len bytes remaining */
        copy_amt =
            mixer.channels[i].remaining <
            len ? mixer.channels[i].remaining : len;

        /* mix this sound effect with the output */
        SDL_MixAudioFormat(stream, mixer.channels[i].position,
                           mixer.outputSpec.format, copy_amt, 150);

        /* update buffer position in sound effect and the number of bytes left */
        mixer.channels[i].position += copy_amt;
        mixer.channels[i].remaining -= copy_amt;

        /* did we finish playing the sound effect ? */
        if (mixer.channels[i].remaining == 0) {
            mixer.channels[i].position = NULL;  /* indicates no sound playing on channel anymore */
            mixer.numSoundsPlaying--;
            if (mixer.numSoundsPlaying == 0) {
                /* if no sounds left playing, pause audio callback */
                SDL_PauseAudio(1);
            }
        }
    }
}



#else // no sound engine (console)
	void glb_dxsound_console_dummy() {};

#endif

