#include <ogg/vorbisfile.h>


class GF_oggdec
{
public:
	GF_oggdec()
	{
		quiet = 0;
		bits = 16;
		endian = 0;
		raw = 0;
		sign = 1;
	}

	int quiet;
	int bits;
	int endian;
	int raw;
	int sign;
	unsigned char headbuf[44]; /* The whole buffer */

	/*
	static void usage(void) {
	fprintf(stderr, "Usage: oggdec [flags] file1.ogg [file2.ogg ... fileN.ogg]\n"
	"\n"
	"Supported flags:\n"
	" --quiet, -Q      Quiet mode. No console output.\n"
	" --help,  -h      Produce this help message.\n"
	" --version, -v    Print out version number.\n"
	" --bits, -b       Bit depth for output (8 and 16 supported)\n"
	" --endianness, -e Output endianness for 16 bit output. 0 for\n"
	"                  little endian (default), 1 for big endian\n"
	" --sign, -s       Sign for output PCM, 0 for unsigned, 1 for\n"
	"                  signed (default 1)\n"
	" --raw, -R        Raw (headerless) output.\n"
	" --output, -o     Output to given filename. May only be used\n"
	"                  if there is only one input file, except in\n"
	"                  raw mode.\n"

	);

	}
	*/

#define WRITE_U32(buf, x) *(buf)     = (unsigned char)((x)&0xff);\
	*((buf)+1) = (unsigned char)(((x)>>8)&0xff);\
	*((buf)+2) = (unsigned char)(((x)>>16)&0xff);\
	*((buf)+3) = (unsigned char)(((x)>>24)&0xff);

#define WRITE_U16(buf, x) *(buf)     = (unsigned char)((x)&0xff);\
	*((buf)+1) = (unsigned char)(((x)>>8)&0xff);

protected:
	/* Some of this based on ao/src/ao_wav.c */
	int write_prelim_header(OggVorbis_File *vf, FILE *out, ogg_int64_t knownlength) {
		unsigned int size = 0x7fffffff;
		int channels = ov_info(vf,0)->channels;
		int samplerate = ov_info(vf,0)->rate;
		int bytespersec = channels*samplerate*bits/8;
		int align = channels*bits/8;
		int samplesize = bits;

		if(knownlength && knownlength*bits/8*channels < size)
			size = (unsigned int)(knownlength*bits/8*channels+44);

		memcpy(headbuf, "RIFF", 4);
		WRITE_U32(headbuf+4, size-8);
		memcpy(headbuf+8, "WAVE", 4);
		memcpy(headbuf+12, "fmt ", 4);
		WRITE_U32(headbuf+16, 16);
		WRITE_U16(headbuf+20, 1); /* format */
		WRITE_U16(headbuf+22, channels);
		WRITE_U32(headbuf+24, samplerate);
		WRITE_U32(headbuf+28, bytespersec);
		WRITE_U16(headbuf+32, align);
		WRITE_U16(headbuf+34, samplesize);
		memcpy(headbuf+36, "data", 4);
		WRITE_U32(headbuf+40, size - 44);

		if(fwrite(headbuf, 1, 44, out) != 44) {
			// fprintf(stderr, "ERROR: Failed to write wav header: %s\n", strerror(errno));
			return 1;
		}

		return 0;
	}

	int rewrite_header(FILE *out, unsigned int written) 
	{
		unsigned int length = written;

		length += 44;

		WRITE_U32(headbuf+4, length-8);
		WRITE_U32(headbuf+40, length-44);
		if(fseek(out, 0, SEEK_SET) != 0)
			return 1;

		if(fwrite(headbuf, 1, 44, out) != 44) {
			// fprintf(stderr, "ERROR: Failed to write wav header: %s\n", strerror(errno));
			return 1;
		}
		return 0;
	}

public:
	int decode_file(const char* infile, const char* outfile)
	{
		FILE* in, *out;
		in = fopen(infile, "rb");
		out=fopen(outfile, "wb");

		OggVorbis_File vf;
		int bs = 0;
		char buf[8192];
		int buflen = 8192;
		unsigned int written = 0;
		int ret;
		ogg_int64_t length = 0;
		ogg_int64_t done = 0;
		int size;
		int seekable = 0;
		int percent = 0;
		int channels;
		int samplerate;

		if(ov_open(in, &vf, NULL, 0) < 0) {
			// fprintf(stderr, "ERROR: Failed to open input as vorbis\n");
			fclose(in);
			return 1;
		}

		channels = ov_info(&vf,0)->channels;
		samplerate = ov_info(&vf,0)->rate;

		if(ov_seekable(&vf)) {
			int link;
			int chainsallowed;
			for(link = 0; link < ov_streams(&vf); link++) {
				if(ov_info(&vf, link)->channels = channels && 
					ov_info(&vf, link)->rate == samplerate)
				{
					chainsallowed = 1;
				}
			}

			seekable = 1;
			if(chainsallowed)
				length = ov_pcm_total(&vf, -1);
			else
				length = ov_pcm_total(&vf, 0);
			size = bits/8 * channels;
		}

		if(!raw) {
			if(write_prelim_header(&vf, out, length)) {
				ov_clear(&vf);
				return 1;
			}
		}

		while((ret = ov_read(&vf, buf, buflen, endian, bits/8, sign, &bs)) != 0) {
			if(bs != 0) {
				vorbis_info *vi = ov_info(&vf, -1);
				if(channels != vi->channels || samplerate != vi->rate) {
					// fprintf(stderr, "Logical bitstreams with changing parameters are not supported\n");
					break;
				}
			}

			if(ret < 0 ) continue;

			if(fwrite(buf, 1, ret, out) != ret)
			{
				// fprintf(stderr, "Error writing to file: %s\n", strerror(errno));
				ov_clear(&vf);
				return 1;
			}

			written += ret;
		}

		if(!raw)
			rewrite_header(out, written); /* We don't care if it fails, too late */

		ov_clear(&vf);

		return 0;
	}

};


