//-------GFFileSystem.h------------------------------//
//  Global functions for file handling
#ifndef GFFILESYSTEM_H
#define GFFILESYSTEM_H

#ifdef GLB_CONSOLE
#include <stdio.h>
#endif

#ifdef ANDROID
	#include <android/log.h>
#endif

typedef long long int64; // _INT64_SEE_HERE_
		static char int64_must_be_64bits[ (sizeof(int64)==8 )? 1 : -1];

#if defined(WIN32)
	#define GF_FILE_SYS_USE_WINAPI
#endif

// Debugging for GP2X and Linux to console - much more comfortable
#if !defined(_DEBUG) && (defined(WIN32) || defined(_WIN32_WCE) || defined(GLB_CONSOLE) ) || defined(GPHWIZ) || defined(CAANOO)
	#define g_deldebug()		GFFileSystem::DeleteDebug()
	#define g_debug(...)		{char tmp[256]; sprintf(tmp, __VA_ARGS__); GFFileSystem::WriteDebug(tmp);}
#elif defined(ANDROID)
	#define g_deldebug()
	#define g_debug(...) (__android_log_print(ANDROID_LOG_INFO, "glbasic", __VA_ARGS__))
#else
	#define g_deldebug() { /* system("clear"); */ }
	#define g_debug(...) (printf(__VA_ARGS__))
#endif


#define g_FileExist(a)		GFFileSystem::DoesFileExist(a)
#define g_msg(a)			{g_msg2("Info", a)}


// ENDIAN - ISSUES
#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
	#define GF_BIG_ENDIAN (0)
#else
	#define GF_BIG_ENDIAN BigEndian() // BIG is PPC Mac, ARM + x86 = little
	inline int BigEndian()		{const unsigned short s = 0x0001; return ! *(const unsigned char*)&s;}
#endif

#define FlipDword(d) ((d)<<24 | (((d)<<8) & 0x00FF0000) | (((d)>>8) & 0x0000FF00) | (d)>>24)
#define FlipWord(w) ( (w)<<8 | (w)>>8 )
#define FlipDouble(a) gfByteSwap((unsigned char*)&a, sizeof(a));
#define FlipFloat(a) gfByteSwap((unsigned char*)&a, sizeof(a));
inline unsigned long  ToBigEndian(unsigned long i)     {return GF_BIG_ENDIAN ? i : FlipDword(i);}
inline unsigned short ToBigEndian(unsigned short i)    {return GF_BIG_ENDIAN ? i : FlipWord(i); }
inline unsigned long  FromBigEndian(unsigned long i)   {return GF_BIG_ENDIAN ? i : FlipDword(i);}
inline unsigned short FromBigEndian(unsigned short i)  {return GF_BIG_ENDIAN ? i : FlipWord(i); }

inline unsigned long  ToLittleEndian(unsigned long i)     {return GF_BIG_ENDIAN ? FlipDword(i) : i;}
inline unsigned short ToLittleEndian(unsigned short i)    {return GF_BIG_ENDIAN ? FlipWord(i)  : i;}
inline unsigned long  FromLittleEndian(unsigned long i)   {return GF_BIG_ENDIAN ? FlipDword(i) : i;}
inline unsigned short FromLittleEndian(unsigned short i)  {return GF_BIG_ENDIAN ? FlipWord(i)  : i;}


inline void gfByteSwap(unsigned char * b, int n)
{
	unsigned char swp;
	int i = 0;
	int j = n-1;
	while (i<j) {swp=b[i]; b[i]=b[j]; b[j]=swp; i++, j--; }
}




#ifdef _WIN32_WCE
	extern char __g_CurDir[];
#endif


#if (defined(WIN32) || defined(_WIN32_WCE)) && !defined(GLB_CONSOLE)
	#ifndef UNICODE
		#define g_msg2(cp, a)			MessageBox(::GetForegroundWindow(), a, cp, MB_OK);
	#else
		#define g_msg2(cp, a)			{wchar_t b[1024], c[1024]; _mbstowcsz(b, a, strlen(a)+1); _mbstowcsz(c, cp, strlen(cp)+1); MessageBox(::GetForegroundWindow(), b, c, MB_OK);}
	#endif
#else
	inline void ConsoleMessageBox(const char* caption, const char* text)
	{
		printf(	"_======================_\n"
				"= %-20s =\n"
				"========================\n\n"
				"%-20s \n\n"
				"-===============[OK]===-\n\n",
				caption, text);
	}
	#define g_msg2(cp, a)			ConsoleMessageBox(cp, a);
#endif


#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


#ifdef GF_FILE_SYS_USE_WINAPI
	#define WIN32_LEAN_AND_MEAN		// Selten benutzte Teile der Windows-Header nicht einbinden

	#ifdef _WINDOWS_
	#include <windows.h>
	#else
	#include "afxwin.h"
	#endif
#else
	#ifndef NULL
		#define NULL (0L)
	#endif
	// #define DWORD	unsigned int
	
	#ifdef DWORD
	#undef DWORD
	#endif


	typedef unsigned int    DWORD;
	typedef unsigned char   BYTE;
	typedef unsigned short  WORD;


	#define STATIC_ASSERT( condition, name ) typedef char assert_failed_ ## name [ (condition) ? -1 : 1 ];
	STATIC_ASSERT(  sizeof(DWORD)==32, must_define_proper_32_bit_DWORD  );



	#define TRUE	1
	#define	FALSE	0
#endif

	// START TINYLIBC
	#include <string.h> // strcpy, strcat
	#include <stdio.h> // sprintf - wsprintf %g is different!?
	#include <stdlib.h>
	#include <time.h> // localtime

#ifndef _mbstowcsz
#define _mbstowcsz mbstowcs
#define _wcstombsz wcstombs
#endif
// ENDTINY

#if !(defined(WIN32) || defined(_WIN32_WCE))
   #include <sys/types.h>
   #include <sys/stat.h>
   #include <unistd.h>
   #include <dirent.h> // opendir/readdir/closedir
   #include <sys/stat.h> // Is Directory
#else
//	#define sprintf wsprintf // %g differs
#endif

#ifdef MACOSX
	/* SPI first appeared in Mac OS X 10.2 */
	extern "C" int _NSGetExecutablePath(char *buf, uint32_t *bufsize);
#endif

#ifdef TARGET_OS_IPHONE
	extern "C" const char* iPhoneGetExePath();
#endif

#ifdef ANDROID
	extern "C" const char* android_files_dir(); // no trailing slash.
	extern "C" const char* android_cache_dir(); // no trailing slash.
#endif

#ifdef MACOSX
	extern "C" const char* MacGetExePath();
#endif


#ifndef GF_FILE_SYS_USE_WINAPI
	#if defined(MACOSX) || defined(TARGET_OS_IPHONE) // apple support for 64 bit file seek is using ftello
		typedef off_t off64_t;
		#define ftello64 ftello // returns an off_t, which might be 64 bits
		#define fseeko64 fseeko
		static char filesystem_64bit_support[ (sizeof(off_t)==8 )? 1 : -1];
	#endif

	#if defined(ANDROID) || defined(EMSCRIPTEN)
		// #NEED - does not support long files!!
		#define off64_t off_t
		#define ftello64 ftello // returns an off_t, which might be 64 bits
		#define fseeko64 fseeko
	#endif

#endif

// -------------------------------------
// TIME MANAGEMENT
// -------------------------------------
inline const char* glb_timestring_from_time_t(struct tm* t)
{
	static char out[32];
	sprintf(out, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d", 1900+t->tm_year, t->tm_mon+1, t->tm_mday
				,t->tm_hour, t->tm_min, t->tm_sec);
	return out;
}


// ------------------------------------------------------------- //
// GFFileSystem Class
// ------------------------------------------------------------- //
class GFFileSystem
{
public:
	#ifdef _DEBUG
	char m_cur_filename[1024];
	#endif

	enum SEEK_FROM
	{
		seek_start = 0,
		seek_current,
		seek_end
	};


	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	GFFileSystem()
	{
		m_file=NULL;
		m_pMemFile=NULL;
		m_fullsize = m_nread=0;
		#ifdef _DEBUG
		m_cur_filename[0]='\0';
		#endif
	}
	~GFFileSystem()	{Close();}
	
	
	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	inline bool IsOpen()
	{
		return (m_file==NULL && m_pMemFile==NULL) ? FALSE : TRUE;
	}
	
	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	static int SetCurrentDir(const char* cdir)
	{
	// g_msg2("SetDir", cdir);
		#ifdef WIN32
			#ifndef _WIN32_WCE
				#ifdef UNICODE
					wchar_t wcdir[1024];
					_mbstowcsz(wcdir, cdir, strlen(cdir)+1);
					return ::SetCurrentDirectory(wcdir) ? 1 : 0;
				#else
					return ::SetCurrentDirectory(cdir) ? 1 : 0;
				#endif
			#else // PocketPC
				// relative path!
				if(cdir[0]!='\\' && cdir[0]!='/')
				{
					strcat(__g_CurDir, "\\");
					strcat(__g_CurDir, cdir);
				}
				else // abolute path, phew!
				{
					strcpy(__g_CurDir, cdir);
				}
			
				// trim trailing / character
				long npos = strlen(__g_CurDir)-1;
				if (npos>=0 && (__g_CurDir[npos] == '\\' || __g_CurDir[npos] == '/'))
					__g_CurDir[npos]='\0';
				return 1; // PocketPC - no current dir
			#endif
		#else
// g_debug("set cdir to: %s\n", cdir);
			int rv = (0 == chdir(cdir)) ? 1 : 0; // _ removed?
			return rv; 
		#endif
	}

	// ------------------------------------------------------------- //
	// C:\prog\myapp - no trailing '\'
	// ------------------------------------------------------------- //
	static void GetCurrentDir(char* cdir)
	{
		#ifdef WIN32
		#ifndef _WIN32_WCE
			#ifdef UNICODE
			wchar_t wc[1024];
			::GetCurrentDirectory(1024, wc);
			_wcstombsz(cdir, wc, wcslen(wc)+1);
			#else
			::GetCurrentDirectory(1024, cdir);
			#endif
		#else
			if (__g_CurDir[0]) // strlen()!=0
				strcpy(cdir, __g_CurDir);
			else
			{
				GetExePath(cdir);
				// Get Exe Dir
				for(long ee = strlen(cdir); ee>0; ee--)
				{
					if (cdir[ee]== '/' || cdir[ee]== '\\')
					{
						cdir[ee] = '\0';
						break;
					}
				}
			}
		#endif
		#else
			if(getcwd(cdir, 1024) == NULL)
			{
				cdir[0] = '\0';
			}
		#endif
	}

	// ------------------------------------------------------------- //
	// Pathname of .exe 
	// ------------------------------------------------------------- //
	static void GetExePath(char* cexepathname)
	{
		cexepathname[0]='\0';

		#if defined(TARGET_OS_IPHONE)
			strcpy(cexepathname, iPhoneGetExePath());
		#elif defined(ANDROID)
			strcpy(cexepathname, android_files_dir());
		#elif defined( MACOSX )
			strcpy(cexepathname, MacGetExePath() );
			return;

			char fullpath[1024], *exepath;
			uint32_t pathsz = 1024;
			char ns_path[1024];
			if(_NSGetExecutablePath(ns_path, &pathsz) == 0 && realpath(ns_path, fullpath))
			{
				strcpy(cexepathname, fullpath);
			}
			else
			{
				cexepathname[0]='\0';
			}
		#elif defined(POSIX) // LINUX
			char fullpath[1024], *exepath;

			pid_t pid = getpid();
			char link[64] = "/proc/";
			sprintf(link, "/proc/%d/exe", pid);
			int ch = readlink(link, fullpath, 512);
			if (ch != -1)
			{
				fullpath[ch] = 0;
				strcpy(cexepathname, fullpath);
			}
			else
			{
				//                          resolve ./ ../
				if((exepath = getenv("_")) && realpath(exepath, fullpath) && (exepath[0]!='\0'))
				{
					strcpy(cexepathname, fullpath);
				}
				else
				{
					strcpy(cexepathname, "./");
				}
			}
		#elif defined(WIN32)
			// Win32
			#ifndef UNICODE
				::GetModuleFileName(NULL, cexepathname, 1024);
			#else
				wchar_t wc[1024];
				// _mbstowcsz(wc, cexepathname, strlen(cexepathname)+1);
				::GetModuleFileName(NULL, wc, 1024);
				_wcstombsz(cexepathname, wc, wcslen(wc)+1);
			#endif
			// fix to forward slashes
			for(char* pc = cexepathname; *pc!='\0'; ++pc)
				if(*pc=='\\') *pc ='/';
		#endif
		
	// g_msg2("GetExePath:", cexepathname);
	}
	
	// ------------------------------------------------------------- //
	// Directory of .exe (inside the .app for GLBasic programs, no slash trailing)
	// ------------------------------------------------------------- //
	static void GetExeDir(char* cexepathname)
	{
		cexepathname[0]='\0';
		GetExePath(cexepathname);
		#if defined( MACOSX )
			// /User/hirokanakwa/Desktop/glb.app/Contents/MacOS/
			// reverse find ".app/"
			// and trim, so that .app is the end
			for(int pos = strlen(cexepathname)-4; pos>=0; --pos)
			{
				if(cexepathname[pos]=='.' && cexepathname[pos+1]=='a' && cexepathname[pos+2]=='p' && cexepathname[pos+3]=='p' && cexepathname[pos+4]=='/')
				{
					cexepathname[pos+4]='\0';
					return;
				}
			}
		#else
			for(int ee = (int)strlen(cexepathname); ee>=0; ee--)
			{
				if (cexepathname[ee]== '/' || cexepathname[ee]== '\\')
				{
					cexepathname[ee] = '\0';
					break;
				}
			}
		#endif
	// g_msg2("GetExePath:", cexepathname);
	}
	

	// Get the name you'd use for opening as input
	static bool GetNoCaseName4Input(const char* name, char* output)
	{
		#ifdef WIN32
			strcpy(output, name);
			return GFFileSystem::DoesFileExist(name);
		#else
			// See if exact name exists and simply open this one
			FILE* fp = fopen(name, "rb");
			if(fp)
			{
				fclose(fp);
				strcpy(output, name);
				return true;
			}
		
			// now see if you can find one with upper/lowercase ignorance
			struct dirent *dp;
			DIR *dirp=NULL;
			char *dir=NULL;
			char *fname=NULL;
			char *found_name=NULL;

			// Separate NAME into directory DIR and filename within the directory FNAME.
			dir = strdup (name);
			fname = strrchr (dir, '/');
			if (fname == NULL)
			{
				char temp[1024]; strcpy(temp, "./"); strcat(temp, name);
				free(dir);
				dir=strdup(temp);
				fname=dir+2;
				dir[1]='\0';
			}
			else
			{
				*fname++ = '\0';
			}
			found_name = NULL;
			dirp = opendir(dir);
			if (dirp == NULL)
			{
				goto err;
			}
			while ((dp = readdir (dirp)) != NULL)
			{
				// first match is taken
				if (strcasecmp(dp->d_name, fname) == 0)
				{
					found_name = strdup(dp->d_name);
					break;
				}
			}
			closedir (dirp);

			if (found_name == NULL) goto err;

			char *p;
			/* Copy the found name back into DIR.  We are assuming that
			   found_name is the same length as fname, which is true as
			   long as the above code is just ignoring case and not other
			   aspects of filename syntax.  */
			if(dir[0]!='.')
			{
				p = dir + strlen (dir);
				*p++ = '/';
			}
			else
				p=dir;
			strcpy (p, found_name);
			strcpy(output, dir);
			free (dir);
			free (found_name);
			return true;

		err:
			free(dir);
			return false;
		#endif
	}

	#ifndef WIN32
	// ------------------------------------------------------------- //
	/* Case-insensitive file open.  As you can see, this is an expensive
	   call.  We don't regard it as our main strategy for dealing with
	   case-insensitivity.  Returns errno code or 0 for success.  Puts the
	   new file in *FP.  NAME and MODE are as for fopen.  If PATHP is not
	   NULL, then put a malloc'd string containing the pathname as found
	   into *PATHP.  *PATHP is only set if the return value is 0.

	   Might be cleaner to separate the file finding (which just gives
	   *PATHP) from the file opening (which the caller can do).  For one
	   thing, might make it easier to know whether to put NAME or *PATHP
	   into error messages.  */
	// ------------------------------------------------------------- //
	FILE* fopen_case(const char* name, const char* mode)
	{
		char tmp[1024];
		if(GetNoCaseName4Input(name, tmp))
		{
			FILE* pF= fopen(tmp, mode);
			#ifndef GLB_CONSOLE
				if(!pF)
					printf("request: fopen_case(\"%s\", \"%s\") opens \"%s\": %s\n", name, mode, tmp, (pF ? "succeeded" : "failed") );
			#endif
			if(pF) return pF;
		}
		// for write access -> do open what you said
		else if(strchr(mode, 'w') || strchr(mode, 'a'))
		{
			FILE* pF =  fopen(name, mode);
			#ifndef GLB_CONSOLE
			if(!pF) printf("wrt-request: fopen(\"%s\", \"%s\") failed\n\n", name, mode );
			#endif
			return pF;
		}
		else
		{
			#ifndef GLB_CONSOLE
				printf("request: fopen(\"%s\", \"%s\") failed\n", name, mode);
			#endif
		}
		return NULL;
	}
	#else
		#define fopen_case	fopen
	#endif





	#ifdef WIN32
	// ------------------------------------------------------------- //
	// Make Abs path for CE systems
	// ------------------------------------------------------------- //
	static void MakeAbsolutePathCE(const char* file, char* cexepathname)
	{
		#ifdef _WIN32_WCE
			if (file[1] == ':')
			{
				// C:\xy goes \xy
				file+=2;
			}
			else if (file[0]!='\\' && file[0]!='/')
			{
				// make abs from rel
				GetCurrentDir(cexepathname);
				strcat(cexepathname, "\\");
				strcat(cexepathname, file);
			}
			else // seems good
			{
				strcpy(cexepathname, file);
			}
		#else
			strcpy(cexepathname, file);
		#endif

		#ifdef WIN32
			// Fix slashes to backslashes
			for(char* pC = cexepathname; *pC!='\0'; ++pC)
			{
				if(*pC=='/') *pC='\\';
			}
		#endif
	}


	static void StrLastError()
	{
		#ifdef _DEBUG
			LPVOID lpMsgBuf;
			FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
				NULL, ::GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
				(LPTSTR) &lpMsgBuf,	0, NULL);
			#ifndef GLB_CONSOLE
				printf("Error: %d %s\n", (int)::GetLastError(), (LPCTSTR)lpMsgBuf);
			#endif
			LocalFree( lpMsgBuf );
		#endif
	}
	
	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	static HANDLE CreateFileCE(const char* file, DWORD desAccess, DWORD ShareMode, LPSECURITY_ATTRIBUTES secat, DWORD creat, DWORD flag, HANDLE htmp)
	{
		char cexepathname[1024];
		MakeAbsolutePathCE(file, cexepathname);
		file = cexepathname;

		#ifdef UNICODE
			wchar_t wc[1024];
			_mbstowcsz(wc, file, strlen(file)+1);
			HANDLE ha = CreateFile(wc, desAccess, ShareMode, secat, creat, flag, htmp);
		#else
			HANDLE ha = CreateFile(file, desAccess, ShareMode, secat, creat, flag, htmp);
			#ifdef _DEBUG
			if(ha==INVALID_HANDLE_VALUE)
			{
				char cdir[1024];
				GetCurrentDir(cdir);
				StrLastError();
			}
			#endif
		#endif
		return ha;
	}
	#endif

	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	bool OpenIn(const char* file)
	{
		if (m_file || m_pMemFile) this->Close();

		#ifdef _DEBUG
		strcpy(m_cur_filename, file);
		#endif
		if(file[0]=='\0') return false;
		
		if(strncmp(file, "mem://", 6)==0)
		{
			long length = strtol(file+6, NULL, 16);
			#pragma warning(disable: 4311 4312)
			char* pMem = (char*)strtol(file+15,NULL, 16);
			return OpenMem(pMem, (int)length);
		}
		
	#ifdef GF_FILE_SYS_USE_WINAPI
		#ifdef _WIN32_WCE
			DWORD access = GENERIC_READ;
		#else
			DWORD access = 0x0001; // FILE_READ_DATA;
		#endif
		m_file = CreateFileCE(file, access,
									0,
									NULL,
									OPEN_EXISTING,
									0, // FILE_ATTRIBUTE_NORMAL,
									NULL);
		if (m_file==INVALID_HANDLE_VALUE) m_file=NULL;
	#else
		m_file=fopen_case(file, "rb");
	#endif
		m_nread=0;
		m_fullsize=GetSize();
		return m_file ? true : false;
	}

	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	bool OpenOut(const char* file)
	{
		if (m_file || m_pMemFile) this->Close();
		if(strncmp(file, "mem://", 6)==0) return false;

		#ifdef _DEBUG
		strcpy(m_cur_filename, file);
		#endif
		
		// delete before you write
		GFFileSystem::Delete(file);

	#ifdef GF_FILE_SYS_USE_WINAPI
		#ifdef _WIN32_WCE
			m_file = CreateFileCE(file, GENERIC_READ|GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		#else
			m_file = CreateFileCE(file, GENERIC_READ|GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0L /*FILE_ATTRIBUTE_NORMAL*/, NULL);
		#endif
		if (m_file==INVALID_HANDLE_VALUE) m_file=NULL;
	#else
		m_file=fopen_case(file, "wb");
		if(!m_file)
			m_file=fopen(file, "wb");
		#ifndef GLB_CONSOLE
		if(!m_file)
			printf("can't write to file: %s\n", file);
		#endif
	#endif
		m_nread=0; m_fullsize=0;
		return m_file ? true : false;
	}
		
	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	static bool MkDir(const char* file)
	{
	#if defined( GF_FILE_SYS_USE_WINAPI ) || defined(_DEBUG)
		char cfullpathname[1024];
		MakeAbsolutePathCE(file, cfullpathname);
		file = cfullpathname;

		#ifdef UNICODE
			wchar_t wc[1024];
			_mbstowcsz(wc, cfullpathname, strlen(cfullpathname)+1);
			return ::CreateDirectory(wc, NULL) ? true:false;
		#else
			return ::CreateDirectory(cfullpathname, NULL) ? true:false;
		#endif
	#else
		return (mkdir(file, (mode_t)0755) == 0) ? true:false;
	#endif
	}


	// ------------------------------------------------------------- //
	// Move a file or directory.
	// ------------------------------------------------------------- //
	static bool MoveFile(const char* pRead, const char* pWrite)
	{
		if(!DoesDirExist(pRead)  && !DoesFileExist(pRead))  {return false;}
		if( DoesDirExist(pWrite) ||  DoesFileExist(pWrite)) {return false;}

		#ifdef GF_FILE_SYS_USE_WINAPI
			#ifdef UNICODE
				wchar_t wc[1024], wc2[1024];
				_mbstowcsz(wc, pRead, strlen(pRead)+1);
				_mbstowcsz(wc2, pWrite, strlen(pWrite)+1);

				return ::MoveFile(wc, wc2) ? true:false;
			#else
				return ::MoveFile(pRead, pWrite) ? true:false;
			#endif
			
		#else
			char pRealRead[1024]="";
			if(GetNoCaseName4Input(pRead, pRealRead))
			{
				if(0== rename(pRealRead, pWrite))
					return true;
			}
		#endif
		return false;
	}

	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	bool OpenAppend(const char* file)
	{
		if (m_file || m_pMemFile) this->Close();
		if(strncmp(file, "mem://", 6)==0) return false;

		#ifdef _DEBUG
		strcpy(m_cur_filename, file);
		#endif

	#ifdef GF_FILE_SYS_USE_WINAPI
		m_file = CreateFileCE(file, GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (m_file==INVALID_HANDLE_VALUE)
		{
			m_file=NULL;
		}
		else
		{
			SetFilePointer(m_file, 0, NULL, FILE_END);
		}
	#else
		m_file=fopen_case(file, "ab");
	#endif
		m_nread=0; m_fullsize=0;
		return m_file ? true : false;
	}

	static void FormatMemFile(char* pFilename, const char* pMem, int length)
	{
		#pragma warning(disable: 4311)
		sprintf(pFilename, "mem://%8x.%p", length, pMem);
	}

	// ------------------------------------------------------------- //
	// Create a readonly file from memory. Will neither new[], nor delete[] this
	// buffer after work!!
	// ------------------------------------------------------------- //
	bool OpenMem(char* pMemory, int length)
	{
		if(m_file || m_pMemFile) Close();
		m_pMemFile = pMemory;
		m_fullsize = length;
		m_nread=0;

		#ifdef _DEBUG
		strcpy(m_cur_filename, "memory file");
		#endif

		return true;
	}

	// ------------------------------------------------------------- //
	// File must have been opened before!
	// ------------------------------------------------------------- //
	int64 GetSize()
	{
		if (!m_file && !m_pMemFile) return 0;
		if(m_pMemFile)
			return m_fullsize;
	int64 file_size;
	#ifdef GF_FILE_SYS_USE_WINAPI
		DWORD szhigh;
		DWORD low=::GetFileSize(m_file, &szhigh);
		if(low == 0xffffffff)
			return 0;
		file_size = szhigh;
		file_size <<=32;
		file_size += low;
	#else
		long cur = ftell(m_file);
		fseek(m_file, 0, SEEK_END);
		file_size = (int64) ftello64(m_file); // MUST be ftello64 - you must define it if ftell returns 64 bits
		//file_size = (int64)ftell(m_file);
		fseek(m_file, cur, SEEK_SET);
	#endif
	return file_size;
	}

	// ------------------------------------------------------------- //
	// Tell how many bytes have been read already
	// ------------------------------------------------------------- //
	int64 Tell()
	{
		return m_nread;
	}

	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	bool SeekTo(int64 pos, SEEK_FROM sf=seek_start)
	{
		if (!m_file && !m_pMemFile)
			return false;

		switch(sf)
		{
		case GFFileSystem::seek_current:
			m_nread+=pos; break;
		case GFFileSystem::seek_end:
			m_nread=m_fullsize + pos; break;
		default:
			m_nread=pos;
		}

	#ifdef GF_FILE_SYS_USE_WINAPI
		if(m_file)
		{
			#ifndef INVALID_SET_FILE_POINTER
			#define INVALID_SET_FILE_POINTER ((DWORD)0xFFFFFFFF)
			#endif
			LONG disthigh = (LONG)((m_nread>>32)&0xffffffff);
			return (INVALID_SET_FILE_POINTER == ::SetFilePointer(m_file, (LONG)(m_nread&0xffffffff), &disthigh, FILE_BEGIN)) ? false : true;
		}
	#else
		if(m_file)
			// MUST be fseeko64 - define it if fseek works on 64 bits
			return (fseeko64(m_file, (off64_t)m_nread, SEEK_SET) == 0) ? true : false;
	#endif
		// memory pointer
		if(m_nread<0 || m_nread>m_fullsize)
			return false;
		return true;
	}

	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	void Close()
	{
		#ifdef _DEBUG
		strcpy(m_cur_filename, "");
		#endif

		if (m_file)
		{
			#ifdef GF_FILE_SYS_USE_WINAPI
			CloseHandle(m_file);
			#else
			fclose(m_file);
			#endif
			m_file = NULL;
			m_nread=0; m_fullsize=0;
		}
		if(m_pMemFile)
		{
			m_pMemFile=NULL;
			m_nread=0; m_fullsize=0;
		}
	}
	// ------------------------------------------------------------- //

	// ------------------------------------------------------------- //
	bool Write(const void* buffer, int bytes)
	// ------------------------------------------------------------- //
	{
	if(!bytes) return true; // crash on unix
	#ifdef GF_FILE_SYS_USE_WINAPI
		DWORD	nread;
		if (m_file)
		{
			WriteFile(m_file, buffer, bytes, &nread, NULL);
			return (nread == (DWORD)bytes) ? true : false;
		}
	#else
		if (m_file && fwrite(buffer, bytes, 1, m_file) == 1) return true;
	#endif
	return false;
	}

	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	bool WriteLine(const char* line)
	{
		Write((void*)line, (int)strlen(line));
		return Write((void*)"\r\n", 2);
	}

	// ------------------------------------------------------------- //
	// ------------------------------------------------------------- //
	bool Read(void* buffer, int bytes)
	{
		if(m_file)
		{
			m_nread+=bytes;
		#ifdef GF_FILE_SYS_USE_WINAPI
			DWORD	nread=0;
			if (m_file && ::ReadFile(m_file, buffer, bytes, &nread, NULL) && (nread == (DWORD)bytes))
				return true;
		#else
			if (m_file && fread(buffer, bytes, 1, m_file)==1)
				return true;
		#endif
		}
	
		// might be memfile
		if(m_pMemFile)
		{
			if(m_nread+bytes>m_fullsize)
			{
				bytes = (int)(m_fullsize - m_nread);
				if(bytes<0) {return false;}
			}
			memcpy(buffer, m_pMemFile + m_nread, bytes);
			m_nread+=bytes;
			return true;
		}
	return false;
	}
	
	// ------------------------------------------------------------- //
	// Read double, fix endianess and stuff
	// ------------------------------------------------------------- //
	bool ReadDbl(double& dbl)
	{
		bool rv = Read(&dbl, 8);
		// floats swapped
		#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
			unsigned int* pL = ((unsigned int*)&dbl);
			unsigned int* pR = ((unsigned int*)&dbl); ++pR;
			unsigned int swp = *pL; *pL = *pR; *pR = swp;
		#else
		if(GF_BIG_ENDIAN)
		{
			FlipDouble(dbl);
		}
		#endif
		return rv;
	}

	// ------------------------------------------------------------- //
	// Write double, fix endianess and stuff
	// ------------------------------------------------------------- //
	bool WriteDbl(double dbl)
	{
		// floats swapped
		#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
			char buffer[8]; memcpy(buffer, &dbl, 8);
			char swp;
			#define SWP(a,b) {swp=buffer[a]; buffer[a]=buffer[b]; buffer[b]=swp;}
			SWP(0,4); SWP(1,5); SWP(2,6); SWP(3,7);
			return Write(buffer, 8);
		#else
			if(GF_BIG_ENDIAN)
			{
				FlipDouble(dbl);
			}
			return Write(&dbl, 8);
		#endif
	}

	// ------------------------------------------------------------- //
	// Read float, fix endianess and stuff
	// ------------------------------------------------------------- //
	bool ReadFlt(float& flt)
	{
		bool rv = Read(&flt, 4);
		if(GF_BIG_ENDIAN) FlipFloat(flt);
		// #NEED see what GP2X has here
		return rv;
	}

	// ------------------------------------------------------------- //
	// Write float, fix endianess and stuff
	// ------------------------------------------------------------- //
	bool WriteFlt(float flt)
	{
		// #NEED see what GP2X has here
		if(GF_BIG_ENDIAN) FlipFloat(flt);
		return Write(&flt, 4);
	}

	// ------------------------------------------------------------- //
	// Read 32 bit integer (from little endian file)
	// ------------------------------------------------------------- //
	template<class T>bool ReadLong(T& i)
	{
		unsigned int io;
		if(!Read((void*) &io, 4)) return false;
		if(GF_BIG_ENDIAN) i = (T)FlipDword(io); else i = (T)io;
		return true;
	}

	// ------------------------------------------------------------- //
	// Write 32 bit integer (from little endian file)
	// ------------------------------------------------------------- //
	template<class T>bool WriteLong(T i)
	{
		if(GF_BIG_ENDIAN) i = (T)FlipDword(i);
		return Write((void*)&i, 4);
	}

	// ------------------------------------------------------------- //
	// Read 16 bit integer (from little endian file)
	// ------------------------------------------------------------- //
	template<class T>bool ReadWord(T& i)
	{
		unsigned short io;
		if(!Read((void*) &io, 2)) return false;
		if(GF_BIG_ENDIAN) i = (T)FlipWord(io); else i =(T)io;
		return true;
	}

	// ------------------------------------------------------------- //
	// Write 16 bit integer (from little endian file)
	// ------------------------------------------------------------- //
	template<class T>bool WriteWord(T i)
	{
		if(GF_BIG_ENDIAN) i = (T)FlipWord(i);
		return Write((void*)&i, 2);
	}
	
	// ------------------------------------------------------------- //
	// Read (including) next '\n' truncating '\r\n'
	// ------------------------------------------------------------- //
	bool ReadLine(void* buffer, int extbuflen)
	{
	char*	pBuffer = (char*)buffer;
	int readlen = 64; // bytes at once
	if(!m_file && !m_pMemFile) return false;

	if(m_file)
	{
		for(int bytes=extbuflen; bytes>0; bytes-=readlen)
		{
			int rlen = readlen<bytes ? readlen : bytes;

			DWORD nread;
			#ifdef GF_FILE_SYS_USE_WINAPI
				if(!ReadFile(m_file, pBuffer, rlen, &nread, NULL))
					return false;
			#else
				nread = (DWORD)fread(pBuffer, 1, rlen,m_file);
			#endif
			m_nread += nread;
			if(!nread)
				return false;

			if((int)nread<rlen)
				pBuffer[nread]='\n';

			char* pEnd = pBuffer+rlen, *pC= pBuffer;
			do
			{
				if (*pC == '\n')
				{
					// rewind what's been read to much
					SeekTo(1 - (int)nread + (int)(pC-pBuffer), seek_current);
					*pC = '\0';
					if(pC>pBuffer && pC[-1] == '\r') pC[-1]='\0';
					return true;
				}
			} while(++pC < pEnd);
			pBuffer+=rlen;
		}
	}
	else // memfile
	{
		pBuffer = m_pMemFile + m_nread;
		int nread=0;
		for(char* pStop = pBuffer; *pStop!='\n' && m_nread<m_fullsize && nread<extbuflen; )
		{
			++nread;
			++pStop;
		}
		if(nread)
		{
			m_nread+=nread;
			memcpy(buffer, pBuffer, nread);
			pBuffer[nread] = '\0';
			if(nread>1 && pBuffer[nread-1]=='\r') pBuffer[nread-1]='\0';
			return true;
		}
		else
			return false;
	}
	return true;
	}

	// ------------------------------------------------------------- //
	// Is there something left to read?
	// ------------------------------------------------------------- //
	inline bool Eof()
	{
		return m_fullsize > m_nread ? false:true;
	}

	//---------------------------------------------------------------//
	//---------------------------------------------------------------//
	static void Delete(const char* filename)
	{
		#ifdef GF_FILE_SYS_USE_WINAPI
			char cexepathname[1024];
			MakeAbsolutePathCE(filename, cexepathname);
			filename = cexepathname;
			#ifndef UNICODE
				BOOL ok=FALSE;
				DWORD att=::GetFileAttributes(cexepathname);
				if(att == INVALID_FILE_ATTRIBUTES)
					return;

				if(att & FILE_ATTRIBUTE_DIRECTORY)
				{
					// if dir == currentdir, the system holds a handle to the directory (error: file in use by another process...)
					char curdir[1024];
					GFFileSystem::GetCurrentDir(curdir);
					if(strcmp(curdir, cexepathname)==0)
					{
						strcat(curdir, "\\.."); // go one up with the current dir
						GFFileSystem::SetCurrentDir(curdir);
					}

					::SetFileAttributes(cexepathname, FILE_ATTRIBUTE_NORMAL);

					HANDLE DirectoryHandle;
					DirectoryHandle = CreateFile(cexepathname,
												FILE_LIST_DIRECTORY,
												FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
												NULL,
												OPEN_EXISTING,
												FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OVERLAPPED,
												NULL);
					ok=FALSE;
					if( DirectoryHandle != INVALID_HANDLE_VALUE )
					{
						ok = ::RemoveDirectory(cexepathname);
						CloseHandle(DirectoryHandle);
					}
					// StrLastError();
				}
				else
				{
					ok = ::DeleteFile(cexepathname);
					if(!ok)
					{
						StrLastError();
					}
				}
			#else
				wchar_t* wc = new wchar_t[1024];
				_mbstowcsz(wc, cexepathname, strlen(cexepathname)+1);

				DWORD att=::GetFileAttributes(wc);
				if(att & FILE_ATTRIBUTE_DIRECTORY)
					::RemoveDirectoryW(wc);
				else
					::DeleteFileW(wc);
				delete[] wc;
			#endif
		#else
			remove(filename);
		#endif
	}

	//---------------------------------------------------------------//
	// Get Temp folder - ends with a slash
	//---------------------------------------------------------------//
	static void GetTempFolder(long size, char* filename)
	{
		#ifdef _WIN32_WCE
			strcpy(filename, "/Temp/");
			return;
		#elif defined (ANDROID)
			strcpy(filename, android_cache_dir());
			if(filename[strlen(filename)-1] != '/')
				strcat(filename, "/");
			return;
		#else
			#ifdef WIN32
				#ifndef UNICODE
					::GetTempPath(size,filename); // C:\Temp
				#else
					wchar_t wc[1024];
					::GetTempPath(1024,wc); // \Temp\...
					long len = (long)wcslen(wc)+1;
					if (len>size)
					{
						len=size;
						wc[size-1] = TEXT('\0');
					}
					_wcstombsz(filename, wc, len);
				#endif
				// Fix slashes to backslashes
				for(char* pC = filename; *pC!='\0'; ++pC)
				{
					if(*pC=='\\') *pC='/';
				}
				if(filename[strlen(filename)-1] != '/')
					strcat(filename, "/");
			#else
				#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
					strcpy(filename, "/mnt/sd/"); // don't write to nand!!
				#else
					strcpy(filename, "/tmp/");
				#endif
			#endif
		#endif
	}

	//---------------------------------------------------------------//
	//---------------------------------------------------------------//
	static void DeleteDebug()
	{
		char filename[1024];
		GFFileSystem::GetTempFolder(1024, filename);
		strcat(filename, "ErrorLog.txt");
		::GFFileSystem::Delete(filename);
		WriteDebug("Start new log...\n");
	}

	//---------------------------------------------------------------//
	//---------------------------------------------------------------//
	static void WriteDebug(const char* text)
	{
		char filename[1024];
		GFFileSystem::GetTempFolder(1024, filename);

		GFFileSystem	fs;
		strcat(filename, "ErrorLog.txt");
		if (!text) text="NULL";
		for(int retry=0; retry<5; ++retry)
		{
			if(fs.OpenAppend(filename))
			{
				fs.Write((void*)text, (int)strlen(text));
				fs.Close();
				break;
			}
		}
		// g_msg(text);
	}


	#ifdef UNICODE
	static char* m_itoa(int value, char *buffer)
	{
		char tmp[33], *tp = tmp, *sp;int i, sign;	unsigned v;
		sign = (value < 0);
		if (sign)v = -value;else v = (unsigned)value;
		while (v || tp == tmp)
		{i = v % 10; v = v / 10; *tp++ = (char)(i+'0');}
		sp = buffer;

		if (sign)*sp++ = '-';
		while (tp > tmp)*sp++ = *--tp;
		*sp = 0;
		return buffer;
	}
	#endif

	//---------------------------------------------------------------//
	//---------------------------------------------------------------//
	static bool DoesFileExist(const char* filename)
	{
		if(strncmp(filename, "mem://", 6)==0)
			return true;

		#ifdef GF_FILE_SYS_USE_WINAPI
			char cexepathname[1024];
			MakeAbsolutePathCE(filename, cexepathname);
			filename = cexepathname;

			#ifdef UNICODE
				wchar_t wc[1024];
				_mbstowcsz(wc, filename, strlen(filename)+1);
				DWORD att = ::GetFileAttributes(wc);
			#else
				DWORD att = ::GetFileAttributes(filename);
			#endif
			if(att == INVALID_FILE_ATTRIBUTES)
				return false;
			if(!(att & FILE_ATTRIBUTE_DIRECTORY))
				return true;
		#else
			GFFileSystem	fs;
			if(fs.OpenIn(filename))
			{
				fs.Close();
				return true;
			}
		#endif
		return false;
	}

	//---------------------------------------------------------------//
	//---------------------------------------------------------------//
	static bool DoesDirExist(const char* dirname)
	{
		#ifdef GF_FILE_SYS_USE_WINAPI
			char cexepathname[1024];
			MakeAbsolutePathCE(dirname, cexepathname);
			dirname = cexepathname;

			#ifdef UNICODE
				wchar_t wc[1024];
				_mbstowcsz(wc, dirname, strlen(dirname)+1);
				DWORD att = ::GetFileAttributes(wc);
			#else
				DWORD att = ::GetFileAttributes(dirname);
			#endif
			if(att == INVALID_FILE_ATTRIBUTES)
				return false;
			if(att & FILE_ATTRIBUTE_DIRECTORY)
				return true;
			return false;
		#else
		   struct stat statbuf;
		   if (stat(dirname, &statbuf) == -1)
			  return false;
		   else
			  return S_ISDIR(statbuf.st_mode) ? true : false;
		#endif
	}



	//---------------------------------------------------------------//
	//---------------------------------------------------------------//
	static const char* GetFileTime(const char* file)
	{
		//static char buff[128];
		time_t filetime;
		memset(&filetime, 0, sizeof(filetime));

		if(strncmp(file, "mem://", 6)!=0)
		{
			#ifdef GF_FILE_SYS_USE_WINAPI
				#ifdef _WIN32_WCE
					DWORD access = GENERIC_READ;
				#else
					DWORD access = 0x0001; // FILE_READ_DATA;
				#endif
				HANDLE hfile = CreateFileCE(file, access,
											0,
											NULL,
											OPEN_EXISTING,
											0, // FILE_ATTRIBUTE_NORMAL,
											NULL);
				FILETIME ft;
				memset(&ft, 0, sizeof(ft));
				BOOL ok;
				ok = ::GetFileTime(hfile, NULL, NULL, &ft);
				::CloseHandle(hfile);
				if(ok)
				{
					// FILETIME to time_t
				   ULARGE_INTEGER ull;
				   DWORD* pLow = (DWORD*)&ft;
				   DWORD* pHigh = &pLow[1];
				   ull.LowPart = *pLow;
				   ull.HighPart = *pHigh;
				   filetime = ull.QuadPart / 10000000ULL - 11644473600ULL;
				}

			#else
				struct stat sb;
				if(0 == stat(file, &sb))
				{
					filetime = sb.st_mtime;
				}
			#endif
		}

		if(filetime == 0)
		{
			return "";
		}

		struct tm *newtime;
		newtime = localtime( &filetime ); /* Convert to local time. */
		return glb_timestring_from_time_t(newtime);
	}


protected:
#ifdef GF_FILE_SYS_USE_WINAPI
	HANDLE	m_file;
#else
	FILE*	m_file;
#endif

	char* m_pMemFile; // m_fullsize must be set, too

	int64 m_fullsize, m_nread;
};


#endif // GFFILESYSTEM

