#ifndef GF_SHOE_BOX_INCLUDED
#define GF_SHOE_BOX_INCLUDED
#include "gfcompression.h"
#include "gffilesystem.h"
#include <string.h>

#ifndef EXCLUDE_SHOEBOX_PACKING
	#include <vector>
	#include <string>
#endif


#ifndef WIN32
#define _stricmp strcasecmp
#endif

#ifndef MAX_PATH
#define MAX_PATH 260
#endif

class GFShoeBox
{
public:
	GFShoeBox()
	{
		m_pZipMemory=NULL;
		m_FileName[0]='\0';
		m_LastOpened[0]='\0';
		m_pFiles=NULL;
		m_GoodBox = true;
	}
	~GFShoeBox()
	{
		Clear();
	}
	void Clear()
	{
		m_GoodBox = true;
		if(m_pZipMemory)
		{
			delete[] m_pZipMemory;
			m_pZipMemory = NULL;
		}
		else if (m_LastOpened[0]!='\0')
		{
			GFFileSystem::Delete(m_LastOpened);
		}
		while(m_pFiles)
		{
			FILEENTRY* pF = m_pFiles;
			m_pFiles = m_pFiles->pNext;
			delete pF;
			pF=NULL;
		}
		m_pFiles=NULL;
		m_FileName[0]='\0';
		m_LastOpened[0]='\0';
	}

	void SetBox(const char* name)
	{
		Clear();
		strcpy(m_FileName, name);
	} 
#ifndef EXCLUDE_SHOEBOX_PACKING
#if !defined( WANT_SDL) && defined(WIN32)

	std::vector<std::string> gSubDirs;

	// Pack a folder's content in the shoebox
	long PackFolder(long& curpos, GFFileSystem& outfile, const char* szDirPath)
	{
	WIN32_FIND_DATA FileData;
	HANDLE hSearch;
	BOOL fFinished = FALSE;
	char	szFindWhat[1024];
	unsigned char*	pBuffer;
	GFCompression*	zip=new GFCompression();
	FILEENTRY*		pFe;
	GFFileSystem	file;
		
		// Start searching for files in the current directory. 
		strcpy(szFindWhat, szDirPath);
		strcat(szFindWhat, "\\*.*");
		#ifndef _WIN32_WCE
			hSearch = FindFirstFile(szFindWhat, &FileData); 
		#else
			wchar_t wc[1024];
			_mbstowcsz(wc, szFindWhat, strlen(szFindWhat));
			hSearch = FindFirstFile(wc, &FileData); 
		#endif
		if (hSearch == INVALID_HANDLE_VALUE) 
			return 0;
		 
		while (!fFinished)
		{
			if (FileData.cFileName[0]=='.') goto skipfile;
			if (FileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				char szTemp[1024];
				strcpy(szTemp, szDirPath);
				strcat(szTemp, "\\");
				#ifndef _WIN32_WCE
					strcat(szTemp, FileData.cFileName);
				#else
					char mb[1024];
					_wcstombsz(mb, FileData.cFileName, wcslen(wc)+1);
					strcat(szTemp, mb);
				#endif

				gSubDirs.push_back(szTemp);
				// PackFolder(curpos, outfile, szTemp); // stack overflow!
				goto skipfile;
			}
			strcpy(szFindWhat, szDirPath);
			strcat(szFindWhat, "\\");
			#ifndef _WIN32_WCE
				strcat(szFindWhat, FileData.cFileName);
			#else
				char mb[1024];
				_wcstombsz(mb, FileData.cFileName, wcslen(wc)+1);
				strcat(szFindWhat, mb);
			#endif
			if(!file.OpenIn(szFindWhat)) goto skipfile;
			pBuffer = new unsigned char[FileData.nFileSizeLow];
			file.Read(pBuffer, FileData.nFileSizeLow);
			// fread(pBuffer, 1, FileData.nFileSizeLow, file); fclose(file);
			file.Close();
			zip->Pack(pBuffer, FileData.nFileSizeLow);
			outfile.Write((void*)zip->GetResult(), zip->GetResultLength());
			delete [] pBuffer;

			pFe = new FILEENTRY;
			pFe->pNext=NULL;
			pFe->flen = zip->GetResultLength(); // Ignore size high. Come on 2gigs is enough!
			pFe->fpos=curpos;
			#ifndef _WIN32_WCE
				strcpy(pFe->name, FileData.cFileName);
			#else
				_wcstombsz(mb, FileData.cFileName, wcslen(wc)+1);
				strcpy(pFe->name, mb);
			#endif
			curpos+=pFe->flen;
			if (!m_pFiles)
				m_pFiles = pFe;
			else
			{
				FILEENTRY*	pFeList = m_pFiles;
				while(pFeList->pNext) pFeList = pFeList->pNext;
				pFeList->pNext = pFe;
			}
			skipfile:
			if (!FindNextFile(hSearch, &FileData))
				fFinished = TRUE;
		}
		// Close the search handle. 
		FindClose(hSearch);
		return 1;
	}


	// CreateFrom("E:\\Temp"); // Creates: E:\\Temp.sbx from E:\Temp\*.* files
	int	CreateFrom(const char* szDirPathe)
	{
	char	szFindWhat[1024];
	char	szDirPath[1024];
	long	curpos=0;
	GFFileSystem	outfile;
	FILEENTRY*		pFe;
		Clear();
		strcpy(szDirPath, szDirPathe);
		if(szDirPath[strlen(szDirPath)-1]=='\\')
			szDirPath[strlen(szDirPath)-1]='\0';

		strcpy(szFindWhat, szDirPath); strcat(szFindWhat, ".sbx");
		if (!outfile.OpenOut(szFindWhat)) return 0;
		
		gSubDirs.clear();
		if(!PackFolder(curpos, outfile, szDirPath))
		{
			printf("error\n");
			Clear();
			return 0;
		}
		while(gSubDirs.size())
		{
			std::string dir = gSubDirs.back();
			gSubDirs.pop_back();
			if(!PackFolder(curpos, outfile, dir.c_str()))
			{
				printf("error\n");
				Clear();
				gSubDirs.clear();
				return 0;
			}
		}

		// Write file table
		pFe=m_pFiles;
		unsigned char	nlen; // Len of file name (max 256 bytes)

		// Calculate the size of the file we're going to produce
		long filelength=curpos;
		for(pFe=m_pFiles; pFe; pFe = pFe->pNext)
		{
			filelength += (1 + (unsigned char)strlen(pFe->name) + sizeof(pFe->fpos)+sizeof(pFe->flen));
		}
		filelength+=(3 + sizeof(curpos) + 1); // Table position

		// Write the fileentries
		for(pFe=m_pFiles; pFe; pFe = pFe->pNext)
		{
			nlen = (unsigned char) strlen(pFe->name);
			outfile.Write(&nlen, 1);
			outfile.Write(pFe->name, nlen);
			pFe->fpos -= filelength;
			outfile.Write(&pFe->fpos, sizeof(pFe->fpos));
			outfile.Write(&pFe->flen, sizeof(pFe->flen));
		}
		nlen = 0;
		outfile.Write(&nlen, 1); // Write a 0 lenght entry to indicate end of file
		// outfile.Write(&curpos, sizeof(curpos)); // Start of file table @ last pos
		curpos-=filelength;
		outfile.Write((void*)"sbX", 3);				// File "header"
		outfile.Write(&curpos, sizeof(curpos)); // Start of file table @ last pos
		outfile.Close();
		return 1;
	}
#endif // WANT_SDL
#endif // EXCLUDE_SHOEBOX_PACKING

	// Get the file-list as a \n terminated string.
	// If argument == NULL, the lenght will be calculated
	int ListFiles(char* pszFoundFiles=NULL)
	{
	long			pos, datalength=0;
	GFFileSystem	file;

		if (pszFoundFiles) *pszFoundFiles = '\0';
		// Not there - try to uncompress
		if(strlen(m_FileName))
		{
			if (!file.OpenIn(m_FileName)) return 0; //fopen(m_FileName, "rb");
		}
		else
			return 0;
		file.SeekTo((int)file.GetSize()-7, GFFileSystem::seek_start); // Pos of file table
		

		// fseek(file, -(long)sizeof(pos), GFFileSystem::end); // Pos of file table
		char header[4];
		if(file.Read(header, 3) && file.ReadLong(pos)) // fread(&pos, sizeof(pos), 1, file))
		{
			// compatible file?
			if(header[0]!='s' || header[1]!='b' || header[2]!='X')
			{
				return 0;
			}
			if (pos<0)
				file.SeekTo((int)file.GetSize() + pos, GFFileSystem::seek_start);
			else // old version for compatibility
				file.SeekTo(pos);
			// read file entries
			for(;;)
			{
				FILEENTRY	fe;
				unsigned char	nlen;
				if (!file.Read(&nlen, 1))
					break;
				// fread(&nlen, 1, 1, file);
				// if (feof(file) || ferror(file) || nlen==0) break; // Last file | error
				if(nlen==0) break; // Last file
				file.Read(fe.name, nlen); fe.name[nlen]='\0';
				file.ReadLong(fe.fpos);
				file.ReadLong(fe.flen);
				// file.Read(&fe.fpos, sizeof(fe.fpos));
				// file.Read(&fe.flen, sizeof(fe.flen));

				//fread(fe.name, nlen, 1, file); fe.name[nlen]='\0';
				//fread(&fe.fpos, sizeof(fe.fpos), 1, file);
				//fread(&fe.flen, sizeof(fe.flen), 1, file);
				datalength+=((long)strlen(fe.name) + 1);
				if (pszFoundFiles)
				{
					strcat(pszFoundFiles, fe.name);
					strcat(pszFoundFiles, "\n");
				}
			}
		}
		file.Close();
		//fclose(file);
		return (int)datalength;
	}
	
	// Will delete last extracted file and extract a new file
	// FindOrExtract("Data\\File.dat", data); might return:
	// Data\File.dat if exists, or
	// Temp\anyfile.tmp if extracted
	int FindOrExtract(const char* szFileName, char* pszFoundName, bool bAllowMemfile)
	{
		if (m_pZipMemory)
		{
			delete[] m_pZipMemory;
			m_pZipMemory = NULL;
		}
		if (m_LastOpened[0]!='\0')
		{
			GFFileSystem::Delete(m_LastOpened);
			m_LastOpened[0]='\0';
		}
		if (szFileName == NULL || szFileName[0]=='\0')
		{
			pszFoundName[0]='\0';
			return 0;
		}
	
		char szFoundName[1024];
		szFoundName[0]='\0';
		// try opening that file for input, fix case if required
		if(GFFileSystem::GetNoCaseName4Input(szFileName, pszFoundName)) //  (GFFileSystem::DoesFileExist(szFileName))
		{
			// strcpy(pszFoundName, szFileName);
			return 1;
		}
		// Not there - try to uncompress
		if(!m_GoodBox)
		{
			return 0;
		}

		GFFileSystem file;
		long	pos;
		if(m_FileName[0]) // strlen()!=0
		{
			if(!file.OpenIn(m_FileName))
			{
				m_GoodBox = false;
				return 0;
			}
		}
		else
		{
			m_GoodBox = false;
			return 0;
		}

		char header[4];

		file.SeekTo(-7, GFFileSystem::seek_end); // Pos of file table
//		long full = file.GetSize();
//		file.SeekTo(full-7); // Pos of file table
		header[3]='\0';
		if(file.Read(header, 3) )
		{
			// compatible file?
			if(header[0]!='s' || header[1]!='b' || header[2]!='X')
			{
				m_GoodBox = false;
				return 0;
			}

			if(file.ReadLong(pos))
			{
				if (pos<0)
				{
					file.SeekTo(pos, GFFileSystem::seek_end);
				}
				else // old version for compatibility
				{
					file.SeekTo(pos);
				}
				// read file entries
				for(;;)
				{
					FILEENTRY	fe;
					unsigned char	nlen;
					if (!file.Read(&nlen, 1)) break;
					if(nlen==0) break; // Last entry
					file.Read(fe.name, nlen);
					fe.name[nlen]='\0';

					file.ReadLong(fe.fpos);
					file.ReadLong(fe.flen);
					// file.Read(&fe.fpos, sizeof(fe.fpos));
					// file.Read(&fe.flen, sizeof(fe.flen));
					
					int orglen=0;
					// get file-only name
					const char* pEnd = szFileName+strlen(szFileName)-1;
					while(pEnd>=szFileName && *pEnd!='/' && *pEnd!='\\')
					{
						--pEnd;
						++orglen;
					}
					// last entry has filename ""
					if ((int)strlen(fe.name)>=orglen && _stricmp(fe.name, szFileName + strlen(szFileName)-orglen)==0) // _stricmp
					{
						if (pos<0)
							file.SeekTo(fe.fpos, GFFileSystem::seek_end);
						else // old version for compatibility
							file.SeekTo(fe.fpos);
						unsigned char* pBuffer = new unsigned char[((fe.flen+4)/4) * 4];
						file.Read(pBuffer, (int)fe.flen);
						file.Close();
						
						GFCompression	zip;
						zip.UnPack(pBuffer, (int)fe.flen);
						delete[] pBuffer; pBuffer=NULL;

						GFFileSystem::GetTempFolder(1024, szFoundName);
						char fname[1024];
						// ~sbxAA
						strcpy(fname, "sbx__");
						fname[3] = (char) ((unsigned char)fe.fpos%26)+'A';
						fname[4] = (char) ((unsigned char)fe.flen%26)+'A';
						strcat(fname, fe.name);
						strcat(szFoundName, fname);
						
						if(bAllowMemfile)
						{
							m_pZipMemory = (char*)zip.GetResult();
							// make a GFFileSystem compatible "filename"
							GFFileSystem::FormatMemFile(pszFoundName, m_pZipMemory, (int)zip.GetResultLength());
							// must call delete yourself!
							zip.DontDeleteOnDtor();
							return true;
						}
						else
						{
							GFFileSystem outfile;
							if(outfile.OpenOut(szFoundName))
							{
								outfile.Write((void*)zip.GetResult(), (int)zip.GetResultLength());
								outfile.Close();
								strcpy(m_LastOpened, szFoundName);
								strcpy(pszFoundName, szFoundName);
								return 1;
							}
							else
							{
								return 0;
							}
						}

					}
				}
			}
		}
		file.Close();
		pszFoundName[0]='\0';
		return 0;
	}

private:
	struct FILEENTRY
	{
		char name[MAX_PATH];
		long fpos;
		long flen;
		FILEENTRY* pNext;
	};

	bool m_GoodBox;
	char m_FileName[1024];
	char m_LastOpened[1024];
	char* m_pZipMemory;
	FILEENTRY*	m_pFiles;
};

#endif // GF_SHOE_BOX_INCLUDED
