// Implementation of GLBasic commands

#include "stdafx.h"

#ifdef MACOSX
	typedef int BOOL;
	extern "C" BOOL NSApplicationLoad();
#endif

#if !defined( WANT_SDL)
	#ifdef WIN32
		#define VC_EXTRALEAN
		#define _NO_OLDNAMES // Allow _yn() but not yn() in Math.h
		#if _MSC_VER > 1000
		#pragma once
		#endif // _MSC_VER > 1000

		// Video playback
		#define NOMMREG
		#define NOMSACM
		#ifndef GLB_CONSOLE
			#include <vfw.h>
			#pragma comment(lib, "VFW32.lib")
		#endif
		#define strlen	lstrlen // winAPI instead of stdlib


		#include <io.h> // _open_osfhandle

	#endif
#else
	//	#include <process.h> // system() -> stdio.h / stdlib.h??
	#ifndef GLB_CONSOLE
		#pragma comment (lib, "SDL.lib")
		#pragma comment (lib, "SDL_mixer.lib")
		#pragma comment (lib, "SDLmain.lib")
	//	#pragma comment (lib, "mikmod.lib")
	//	#pragma comment (lib, "native_midi.lib")
	//	#pragma comment (lib, "ogg_static.lib")
	//	#pragma comment (lib, "timidity.lib")
	//	#pragma comment (lib, "vorbis_static.lib")
	//	#pragma comment (lib, "vorbisfile_static.lib")
	#endif
#endif

#ifdef WANT_SDL_MIXER
	#ifndef GLB_CONSOLE
		#pragma comment (lib, "SDL_mixer.lib")
	#endif
#endif


#ifdef GLB_CONSOLE
	#ifdef WIN32
		#include <conio.h>
		static unsigned long __g_ConsoleCP, __g_ConsoleCPOut;
	#else
		// must fake _getch()
		#include <termios.h>
		#include <unistd.h>
	#endif
#endif

//#ifndef WANT_SDL_MIXER
//	#include <process.h>
//#endif

#include <math.h>
#ifndef GLB_CONSOLE
	#include "dxsound.h"
	#include "dxinput.h"
	#include "openglrainbows.h"
	#include "rainbows3d.h"
	// #include "openglfont.h"
#endif
#include "gffilesystem.h"
#include "gfqueue.h"


//#include <stdio.h> // sprintf
#include "gfrand.h"

#include "glb.h"

#ifdef WANT_SHOEBOX
	#include "gfshoebox.h"
	extern GFShoeBox *__g_ShoeBox_Data;
	extern GFShoeBox *__g_ShoeBox_Media;
#endif


#ifdef GF_DEBUG_NEW
	// This must delete after all d'tors have been deleted
    AllocList allocList; // stdafx.h
#endif

// Networking
#include "network/clientserver.h"


#if defined(POSIX)
	void Sleep(unsigned long mssec) {msleep((int)mssec);}
/*
	#include <time.h>
	#include <errno.h>

	void Sleep(unsigned long mmsec)
	{
		struct timespec     tv;
		tv.tv_sec = mmsec / 1000;
		tv.tv_nsec = (mmsec % 1000) * 1000;
		nanosleep(&tv, NULL);
	}
*/
#endif


#define MAX_FONTS		16


namespace __GLBASIC__
{
// GLOBAL OBJECTS
#ifndef GLB_CONSOLE
	DXInput*		pDXin=NULL;
	SFXSound*		pDXsnd=NULL;
	SFXSoundBuffer*	DXwav=NULL;
	extern OpenGLRainbows*	__g_pRbow;
	extern Rainbows3D*		__g_pRbow3D;
#endif
extern GFHawkNetwork*	pGFNet;

extern int				__nCurFont;
extern DGInt			__g_FirstTime;	// The time the program has started
extern DGInt			__g_LastTime;	// To get the time elapsed within the SHOWSCREEN cmds
extern DGInt			__g_ThisTime;	// To get the time elapsed within the SHOWSCREEN cmds

DGNat			__g_debugfile;
extern DGNat			__g_debugline;
extern DGNat			__g_error;
extern DGNat			__g_errorline;
static int		__IsDemoVersion=FALSE;
int				__g_HadBreakPoint=0; // had a breakpoint -> force window to top on SHOWSCREEN
int				__DG_SCHOOLVER;
DGInt			__HR;

extern int			__g_AllowEscape;
extern int			__g_paused;
extern int			__g_allowPause;


DGStr&			__g_Caption(){static DGStr cap; return cap;}
bool			__g_Did3D=false;
bool			__g_DidNET=false;

#define __g_LicenseBits	OpenGLRainbows10FreeMemoriesEiiiii // harder to find in HTML5
int				__g_LicenseBits; // replaced by: "XVSDVS%d", g_ProgramFeature (==3,2)
								// g_allow3D =(nfeature & 1) ? 1 : 0;
								// g_allowNET=(nfeature & 2) ? 1 : 0;


#if defined(TARGET_OS_IPHONE)
extern "C"
{
	extern void* gpGLB_MainWnd;
}
#endif


	HWND		__hWnd=NULL;
#if !defined(WANT_SDL) && defined(WIN32)
	HINSTANCE	__hInstance=NULL;
#endif

#ifndef WANT_SDL_MIXER
	#ifndef GLB_CONSOLE
	HWND		wnd_mp3=NULL;
	#endif
#endif

void __SetDemoVersion()
{
	__IsDemoVersion = TRUE;
}

extern "C" void glb_do_pause(int pause);
extern "C" int glb_allow_escape();

BOOL glb_prepare_reading_file(DGStr& filename);


void* GLBASIC_HWND() {return (void*)__hWnd;}


// ------------------------------------------------------------- //
// __EndProgram
// ------------------------------------------------------------- //
extern "C" void __GLB_close_all_filehandles();


extern "C" void __ShutDownGLBasic()
{
	static int washere=0;
	if(!washere)
	{
		washere=1;
		CALLBYNAME(DGStr("GLB_ON_QUIT"));
	}



	#ifndef GLB_CONSOLE

	if(!__g_pRbow) return; // already shut down.
	g_debug("Shut down GLB\n");

	#endif
	STOPMUSIC();


#if defined(WIN32) && !defined(POCKETPC)
	// restore start directory - for profiler output
	DGStr exe;
	GFFileSystem::GetExeDir(exe.getbuffer(1024));
	exe.releasebuffer();
	GFFileSystem::SetCurrentDir(exe.c_str());
#endif

	__GLB_close_all_filehandles(); // write profiler output

	// shutdown network, quickly - iPhone must release sockets
	if (pGFNet) delete pGFNet;			pGFNet=NULL;

	// be safe to really, really shut down the network commands
	for(int safe=0; safe<16; ++safe)
	{
		netShutdown();
	}

#ifdef GLB_CONSOLE
	#ifdef WIN32
		// restore console code pages
		SetConsoleCP(__g_ConsoleCP);
		SetConsoleOutputCP(__g_ConsoleCPOut);
	#endif
#endif

#if !defined(WANT_SDL) && defined (WIN32)
  	// timeEndPeriod(1);
#endif

#ifdef WANT_SDL
	#ifndef GLB_CONSOLE
		#ifndef ANDROID
		SDL_WM_GrabInput(SDL_GRAB_OFF);
		#endif
		// SDL_Quit(); in endprogram - AFTER PDL_Quit()
	#endif
#endif
	#ifndef GLB_CONSOLE
		if (pDXin) delete pDXin;			pDXin=NULL;
		if (DXwav) delete[] DXwav;			DXwav=NULL;
		if (pDXsnd)delete pDXsnd;			pDXsnd=NULL;
		if (__g_pRbow) delete __g_pRbow;	__g_pRbow=NULL;
		if (__g_pRbow3D) delete __g_pRbow3D;__g_pRbow3D=NULL;
	#endif

	if (__g_ShoeBox_Media) delete __g_ShoeBox_Media;	__g_ShoeBox_Media=NULL;
	if (__g_ShoeBox_Data) delete __g_ShoeBox_Data;		__g_ShoeBox_Data=NULL;
#if !defined(WANT_SDL) && defined(WIN32)
	#ifndef GLB_CONSOLE
		if(__hWnd) DestroyWindow(__hWnd);	__hWnd=NULL;
	#endif
#endif
g_debug("glb is shut down\n");
}

// !!!! DIRECT SHOW WORKS NOW !!
/*
#include <dshow.h>

void DoPlay(LPCTSTR szFileName)
{
	IGraphBuilder *pGraphBuilder;
	IMediaControl *pMediaControl;
	IMediaSeeking *pMediaSeeking;
	IMediaEventEx *pMediaEventEx;

	HRESULT hr;

	// Get an instance of the graph builder
	CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER, IID_IGraphBuilder, (void**)&pGraphBuilder);

	if (!pGraphBuilder)
	{
		MessageBox(NULL, "Critical error, unable to create graph builder", "Error", MB_OK);
	}

	// Get the references to interfaces
	hr = pGraphBuilder->QueryInterface(IID_IMediaControl, (void **)&pMediaControl);
	hr = pGraphBuilder->QueryInterface(IID_IMediaSeeking, (void**)&pMediaSeeking);
	hr = pGraphBuilder->QueryInterface(IID_IMediaEventEx, (void**)&pMediaEventEx);

	if (!pMediaControl || !pMediaSeeking || !pMediaEventEx)
	{
		MessageBox(NULL, "Critical error, error getting references", "Error", MB_OK);
	}



	WCHAR wFileName[1024];
	char totalTime[1024];
	LONGLONG lDuration100NanoSecs = 0;
	long temporary;

	// convert file name to wide char
	mbstowcs(wFileName, szFileName, 1024);

	// create the filter graph
	pGraphBuilder->RenderFile(wFileName, NULL);

	// get the time in units of 100 ns
	pMediaSeeking->GetDuration(&lDuration100NanoSecs);

	// hOwner stores the handle to the dialog
	hOwner = owner;

	// initialise is called everytime that a user selects a media file, so by set this button to Pause
	SetWindowText(GetDlgItem(hOwner, IDC_PLAYPAUSE), PAUSE);

	totalDuration = 0;
	timeElapsed = 0;

	// set our dialog to receive events from graph manager
	pMediaEventEx->SetNotifyWindow((OAHWND)hOwner, WM_GRAPHNOTIFY, 0);

}
*/


// ------------------------------------------------------------- //
void __dLn	(int nfile, int num)
// ------------------------------------------------------------- //
{
//	printf("\r%d,%d,", nfile, num);
	__g_debugfile=nfile;
	__g_debugline=num;
}


#if !defined( WANT_SDL) && defined(WIN32)
#ifndef GLB_CONSOLE


// *** Windows 8 WM_POITNER mutlitouch messages ***
#ifndef WM_POINTERUP
	#define WM_POINTERUPDATE              0x0245
	#define WM_POINTERDOWN                0x0246
	#define WM_POINTERUP                  0x0247
	typedef enum tagPOINTER_INPUT_TYPE { 
	  PT_POINTER   = 0x00000001,
	  PT_TOUCH     = 0x00000002,
	  PT_PEN       = 0x00000003,
	  PT_MOUSE     = 0x00000004,
	  PT_TOUCHPAD  = 0x00000005
	} POINTER_INPUT_TYPE;
	typedef enum _POINTER_BUTTON_CHANGE_TYPE { 
	  POINTER_CHANGE_NONE               = 0,
	  POINTER_CHANGE_FIRSTBUTTON_DOWN  , // might be wrong - enum was broken in MSDN
	  POINTER_CHANGE_FIRSTBUTTON_UP    ,
	  POINTER_CHANGE_SECONDBUTTON_DOWN ,
	  POINTER_CHANGE_SECONDBUTTON_UP   ,
	  POINTER_CHANGE_THIRDBUTTON_DOWN  ,
	  POINTER_CHANGE_THIRDBUTTON_UP    ,
	  POINTER_CHANGE_FOURTHBUTTON_DOWN ,
	  POINTER_CHANGE_FOURTHBUTTON_UP   ,
	  POINTER_CHANGE_FIFTHBUTTON_DOWN  ,
	  POINTER_CHANGE_FIFTHBUTTON_UP    
	} POINTER_BUTTON_CHANGE_TYPE;

	typedef struct tagPOINTER_INFO {
	  POINTER_INPUT_TYPE         pointerType;
	  UINT32                     pointerId;
	  UINT32                     frameId;
	  UINT32 /* ? */              pointerFlags;
	  HANDLE                     sourceDevice;
	  HWND                       hwndTarget;
	  POINT                      ptPixelLocation;
	  POINT                      ptHimetricLocation;
	  POINT                      ptPixelLocationRaw;
	  POINT                      ptHimetricLocationRaw;
	  DWORD                      dwTime;
	  UINT32                     historyCount;
	  INT32                      inputData;
	  DWORD                      dwKeyStates;
	  UINT64                     PerformanceCount;
	  POINTER_BUTTON_CHANGE_TYPE ButtonChangeType;
	} POINTER_INFO;
	typedef struct tagPOINTER_TOUCH_INFO {
	  POINTER_INFO pointerInfo;
	  UINT32 /*Touch Flags*/  touchFlags;
	  UINT32 /*Touch Mask */  touchMask;
	  RECT         rcContact;
	  RECT         rcContactRaw;
	  UINT32       orientation;
	  UINT32       pressure;
	} POINTER_TOUCH_INFO;
	#define GET_POINTERID_WPARAM(wParam) (LOWORD (wParam))
	#define GET_X_LPARAM(lp)                        ((int)(short)LOWORD(lp))
	#define GET_Y_LPARAM(lp)                        ((int)(short)HIWORD(lp))
	#define IS_POINTER_FLAG_SET_WPARAM(wParam, flag) (((DWORD)HIWORD (wParam) &(flag)) == (flag))
	#define IS_POINTER_PRIMARY_WPARAM(wParam) IS_POINTER_FLAG_SET_WPARAM (wParam, POINTER_MESSAGE_FLAG_PRIMARY)
	#define IS_POINTER_FIRSTBUTTON_WPARAM(wParam) IS_POINTER_FLAG_SET_WPARAM (wParam, POINTER_MESSAGE_FLAG_FIRSTBUTTON)
	#define IS_POINTER_SECONDBUTTON_WPARAM(wParam) IS_POINTER_FLAG_SET_WPARAM (wParam, POINTER_MESSAGE_FLAG_SECONDBUTTON)
#endif

struct Win32_TouchPos // Windows 8+ WM_POINTERUP messages for multitouch/multiple mice
{
	Win32_TouchPos() {m_X=0; m_Y=0; m_Down=0; m_id=0;}
	int m_X, m_Y, m_Down; // m_Down 1=left/touch |2=right |4=middle
	int m_id;
}
gWin32_TouchPos[32];

extern "C" void win32_get_touches(int id, int* mx, int* my, int* b1)
{
	*mx=gWin32_TouchPos[id].m_X;
	*my=gWin32_TouchPos[id].m_Y;
	*b1=gWin32_TouchPos[id].m_Down;
}


DECLARE(GetPointerTouchInfo, "user32.dll", (UINT32, POINTER_TOUCH_INFO*), BOOL);


// iDirection: 1 down, -1 up, 0 mode
void HandleWM_POINTER(int iDirection, WPARAM wParam, LPARAM lParam)
{
	int pointerId = GET_POINTERID_WPARAM(wParam);

	if(pointerId==0)
		return;

	POINTER_TOUCH_INFO info;
	GetPointerTouchInfo(pointerId, &info);

	for(int i=0; i<32 * 2; ++i)
	{
		if((i<32 && gWin32_TouchPos[i].m_id == pointerId) // update
			|| ( i>=32 && gWin32_TouchPos[i-32].m_id == 0)) // new pointer down
		{
			if(i>=32) i-=32; // had to use an empty slot instead of an update.

			int button=0;
			if(info.pointerInfo.pointerFlags & 0x00000010) // POINTER_FLAG_FIRSTBUTTON)
				button |=1;
			if(info.pointerInfo.pointerFlags & 0x00000020) // POINTER_FLAG_SECONDBUTTON)
				button |=2;
			if(info.pointerInfo.pointerFlags & 0x00000040) // POINTER_FLAG_THIRDBUTTON)
				button |=4;

			switch(iDirection)
			{
			case 0: // update
				gWin32_TouchPos[i].m_Down = button;
				break;
			case 1:
				gWin32_TouchPos[i].m_Down |= button;
				gWin32_TouchPos[i].m_id = pointerId; break;
			case -1:
				gWin32_TouchPos[i].m_Down &= button;
				gWin32_TouchPos[i].m_id = 0; break;
			}
			::ScreenToClient((HWND)GLBASIC_HWND(), &info.pointerInfo.ptPixelLocation);
			gWin32_TouchPos[i].m_X = info.pointerInfo.ptPixelLocation.x;
			gWin32_TouchPos[i].m_Y = info.pointerInfo.ptPixelLocation.y;

			break; // !! must
		}
	}
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
LRESULT CALLBACK __WndProc_(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
	// Screensaver
	case WM_SYSCOMMAND:
		if((wParam&0xFFF0)==SC_SCREENSAVE || (wParam&0xFFF0)==SC_MONITORPOWER)
			return 0L;
		if((wParam&0xFFF0)==SC_CLOSE)
		{
			PostQuitMessage(0);
		}
		break;
	case WM_SYSKEYUP:
		return 0; // ignore Sysmenu on F10/LAlt
		break;
	case WM_CLOSE:
		__EndProgram();
		DestroyWindow(hWnd);
		PostQuitMessage(0);
	case WM_ACTIVATE:
		{
			switch(wParam & 0xffff)
			{
				case WA_ACTIVE: case WA_CLICKACTIVE:
					if(!__DG_DEBUG)
					{
						glb_do_pause(false);
						if(!__g_paused && __g_allowPause)
							::SetWindowText(__hWnd, __g_Caption().c_str());
					}
					break;
				case WA_INACTIVE:
					if(!__DG_DEBUG)
					{
						glb_do_pause(true);
						if(__g_paused && __g_allowPause)
							::SetWindowText(__hWnd, (__g_Caption() + CGStr(" <pause>")).c_str());
					}
					break;
			}
		}
		return 0L;
	case WM_ERASEBKGND:
		return 1;
//	case WM_MOVE: case WM_CREATE: case WM_SIZE:
//		::InvalidateRect(0, NULL, TRUE);

//		return 0L;


	case WM_POINTERUP:
		HandleWM_POINTER(-1, wParam, lParam);
		break;
	case WM_POINTERDOWN:
		HandleWM_POINTER(1, wParam, lParam);
		break;
	case WM_POINTERUPDATE:
		HandleWM_POINTER(0, wParam, lParam);
		break;


		break;
	}
	return (DefWindowProc(hWnd, message, wParam, lParam));
}
#endif
#endif


// ------------------------------------------------------------- //
// __AllocConsole - allocate a console
// ------------------------------------------------------------- //

#ifndef GLB_CONSOLE
	// extern "C" int _open_osfhandle(intptr_t, int); <io.h>
	void __AllocConsole()
	{
	#if defined (WIN32) && !defined(POCKETPC)
	int hCrtIn, hCrtOut;
	FILE *conIn, *conOut;
		AllocConsole();
		#pragma warning(disable: 4311)
		hCrtIn  = _open_osfhandle (size_t(GetStdHandle(STD_INPUT_HANDLE )), 0x4000);
		hCrtOut = _open_osfhandle (size_t(GetStdHandle(STD_OUTPUT_HANDLE)), 0x4000);
		conIn = _fdopen( hCrtIn, "r" );
		conOut = _fdopen( hCrtOut, "w" );
		*stdin = *conIn;
		*stdout = *conOut;
	#endif
	}
#else

#ifdef WIN32
	HWND GetConsoleHwnd()
	{
		#define MY_BUFSIZE 1024
		HWND hwndFound;
		char pszNewWindowTitle[MY_BUFSIZE];
		char pszOldWindowTitle[MY_BUFSIZE];
		GetConsoleTitle(pszOldWindowTitle, MY_BUFSIZE);
		wsprintf(pszNewWindowTitle,"%d/%d/%d",
		GetTickCount(),
		GetCurrentProcessId(), (int)RND(32767));
		SetConsoleTitle(pszNewWindowTitle);
		Sleep(40);
		hwndFound=FindWindow(NULL, pszNewWindowTitle);
		SetConsoleTitle(pszOldWindowTitle);
		return hwndFound;
	}
#endif

#endif


/*
// Demo char sequence - replaced by GLBasic IDE if not demo anymore
static char _g_glb_prtc[16]="DvSDVsdVS0e7"; // don't make it const - can't replace in binary otherwise
// replaced by: "XVSDVS%d", g_ProgramFeature (==3,2)

#define MAKE_DEMO_VARIABLE() {volatile int fake = 0; if(fake==0){++fake;} ++fake; __IsDemoVersion= _g_glb_prtc[0]=='B'+fake ? TRUE : __IsDemoVersion;}
*/

#define PrepLic	OpenGLRainbows10AllocMemoriesEiii // harder to find in HTML5
void PrepLic();

#if defined( WIN32) && !defined(_WIN32_WCE)
DECLARE(EnableMouseInPointer, "User32.dll", (BOOL), BOOL );
#endif
// ------------------------------------------------------------- //
// __GlobalInit - Create and handle Window Stuff
// ------------------------------------------------------------- //
extern "C" int __GlobalInit(HINSTANCE hInst, int nCmdShow, const char* Caption)
{
	/*
	char force_demo_ok[1024];
	strcpy(force_demo_ok, _g_glb_prtc);
	sprintf(force_demo_ok, "%s", _g_glb_prtc);
	*/
	g_deldebug();


#ifdef MACOSX
#ifndef GLB_CONSOLE
	::NSApplicationLoad();
#endif
#endif

#ifdef _DEBUG
#ifndef GLB_CONSOLE
	__AllocConsole();
#endif
#endif


#if defined(WANT_SDL) || defined(WANT_SDL_MIXER)
	Uint32 flags = 0;
	#if defined(WANT_SDL_MIXER)
		flags |= SDL_INIT_AUDIO;
	#endif

	if(!SDL_Init(SDL_INIT_JOYSTICK|SDL_INIT_TIMER |SDL_INIT_VIDEO | flags) ) // segfault w/o SDL_VIDEO on RaspberryPi
	{
		g_debug("Can't init SDL.\n%s\n", (const char*)SDL_GetError());
	}
#endif



	__g_allowPause=true;
	PrepLic();

	// must reset. Win8 loads a DLL at runtime
	__g_Did3D = false;
	__g_DidNET= false;

	
#if (!defined(WIN32) && !defined(LINUX) && !defined(MACOSX)) || defined(_WIN32_WCE)
	if((__g_LicenseBits & 2) == 0) // need NET for other platforms
		__SetDemoVersion();
#endif




#ifndef _DEBUG
	#ifndef WIN32
		atexit(__EndProgram);
	#endif
	
#else
//	__IsDemoVersion=FALSE;
//	__g_LicenseBits= 0xff;
#endif

	// before Rbow
	g_debug("timer\n");
	SetupAccurateTimer();

	g_debug("rbow\n");

// Create window
#ifndef GLB_CONSOLE
	__g_pRbow = new OpenGLRainbows;

	 // SPACE FRACTAL :: Removed ANDROID define, needs calling -1 on Android to fix the Orientation
	#if defined(TARGET_OS_IPHONE) || defined(PALM_WEBOS) || (defined(_DEBUG) && defined(HAVE_OPENGL_ES))
		if(__DG_RESX>__DG_RESY)
		{
			__g_pRbow->m_landscape=1; // 90 CCW
		}
	#endif
	
	g_debug("rbow init\n");
	if(__g_pRbow->Init(	hInst,
						(unsigned int)__DG_RESX,
						(unsigned int)__DG_RESY, 0, 
						(unsigned int)__DG_FRAMERATE,
						(int)__DG_MULTISAMPLE,
						__DG_FULLSCREEN?true:false)
						)
	{
		#if !defined( WANT_SDL) && defined(WIN32)
		__hWnd = __g_pRbow->m_WND.hwnd;
		SetWindowText(__hWnd, Caption);
		#endif
		
		#if defined(TARGET_OS_IPHONE)
		__hWnd = (HWND)gpGLB_MainWnd;
		#endif
	}
	else
	{
		return 0;
	}
#else
	#ifdef WIN32
	// __hWnd = ::GetConsoleWindow(); 2k,xp,Vista - only
	__hWnd = GetConsoleHwnd();
	#endif
#endif

	g_debug("Cptn\n");
	__g_Caption() = CGStr(Caption);

#ifdef _DEBUG
	GFsrand(1);
//	__DG_FULLSCREEN=FALSE;
#else
	GFsrand((long)GETTIMERALL());
#endif
	g_debug("Network\n");
	pGFNet=new GFHawkNetwork;

// Allocate Memory on heap:
#ifndef GLB_CONSOLE
	__g_pRbow3D = new Rainbows3D(__g_pRbow);

	g_debug("Input\n");
	pDXin=new  DXInput;

	g_debug("Window mode\n");
	if(__DG_FULLSCREEN)
		__g_pRbow->DoFullscreen();
	else
		__g_pRbow->DoWindowed();

	g_debug("Create DXin\n");
	#ifdef WANT_SDL
		pDXin->CreateDInput(0L, 0L, __DG_WANTMOUSE ? true : false);
		#ifndef ANDROID
		SDL_WM_SetCaption(Caption, "");
		if (__DG_WANTMOUSE)
		{
			g_debug("sdl_grab\n");
			SDL_WM_GrabInput(SDL_GRAB_ON);
		}
		#endif
	#else
		pDXin->CreateDInput((HWND)__hWnd, hInst, __DG_WANTMOUSE ? true : false);
	#endif

	#if defined( WANT_SDL) && !defined(GP2X) && !defined(GPHWIZ)
		pDXin->m_AccelerationNativeLandscape = __g_pRbow->m_physical_device_screen_is_landscape?true:false;
	#endif

	// Set Window pointers again - just to be sure
	#if defined(WIN32) && !defined(WANT_SDL) && !defined(POCKETPC)
		g_debug("reptr\n");
		pDXin->SetWindow(__g_pRbow->m_WND.hwnd);
	#endif
#endif

	__nCurFont=0;
	DGStr exepath;

	// Set current dir for Linux when double clicking with KDE
	// The same for Win32 when associated with a file extension
	g_debug("getexe\n");
	GFFileSystem::GetExeDir(exepath.getbuffer(1024));
	exepath.releasebuffer();
	g_debug("cd\n");
	GFFileSystem::SetCurrentDir(exepath.c_str());
	g_debug("exepath=curdir= %s\n", exepath.c_str() );

	#ifdef MACOSX
		exepath += CGStr("/Contents/Resources");
		GFFileSystem::SetCurrentDir( exepath.c_str());
		g_debug("curdir= %s\n", exepath.c_str() );
	#endif


	#ifdef WANT_SHOEBOX
	__g_ShoeBox_Media = new GFShoeBox;
	__g_ShoeBox_Data  = new GFShoeBox;
	#endif

	// Set .exe = default ShoeBox
	g_debug("shoeboxing...\n" );
	GFFileSystem::GetExePath(exepath.getbuffer(1024) );
	exepath.releasebuffer();
	SETSHOEBOX(exepath, exepath);


	g_debug("init gettimer\n" );
	__g_FirstTime=__g_LastTime=__g_ThisTime=AccurateTime();
#ifndef GLB_CONSOLE
	g_debug("clear screen\n" );
	CLEARSCREEN();

	g_debug("flip\n" );
	__g_pRbow->FlipScreens();
	g_debug("mk2D\n" );
	__g_pRbow->Make2DMode();

	#define CENTER(a, y) PRINT(DGStr(a), (sx-LEN(CGStr(a))*fx)/2, y);
	DGInt sx, sy, fx, fy;
	__g_pRbow->DefaultFont(0); // default font
	GETSCREENSIZE(sx, sy);
	GETFONTSIZE(fx, fy);
	if(__DG_SCHOOLVER)
	{
		int paused=0;
		for(DGInt alpha=0.01f; alpha<=2; alpha+=.039f)
		{
			ALPHAMODE(alpha<1 ? alpha : 2.1-alpha);
			if(alpha>=1 && !paused)
			{
				paused=1;
				ALPHAMODE(0);
			}
			CENTER("School License", (sy-fy)/2)
			CENTER("www.GLBasic.com", sy-fy);
			SHOWSCREEN();
			SLEEP(20);
			if(paused==1)
			{
				paused=2;
				SLEEP(500);
			}
		}
		ALPHAMODE(0);
		CLEARSCREEN();
	}

	//if(!android_copy_asset_done())
	//{
	//	DGStr waiting = "ooooooo";
	//	int iwait = 0;
	//	do
	//	{
	//		iwait = (1+iwait)%waiting.len();
	//		waiting = LCASE_Str(waiting);
	//		waiting[iwait] = 'O';
	//		GETSCREENSIZE(sx, sy);
	//		CENTER(waiting, (sy-fy)/2)
	//		SHOWSCREEN();
	//		SLEEP(250);
	//	} while (!android_copy_asset_done());
	//}
	//#endif
	//#undef CENTER


	// load smalfont.bmp / smalfont.png
	g_debug("finding font...\n" );

	DGStr files[4];
	files[0] = "Media/smalfont.png";
	files[1] = "Media/smalfont.bmp";
	files[2] = "smalfont.png";
	files[3] = "smalfont.bmp";

	for(int i=0; i<3; ++i)
	{
		if(glb_prepare_reading_file(files[i]))
		{
			LOADFONT(files[i].c_str(), 0);
			break;
		}
	}
	SETFONT(0);
#else // console
	if(__DG_SCHOOLVER)
	{
		printf("School License\n\n");
	}
	if(__IsDemoVersion)
	{
		printf("DEMO version\n...waiting 5 seconds.\n");
		SLEEP(4959);
	}
#endif



	__g_paused=FALSE;

#if !defined( WANT_SDL) && defined(WIN32)
	// timeBeginPeriod(1);
	SetCursor((HCURSOR)LoadCursor(hInst, IDC_ARROW));
#endif
	__g_FirstTime=__g_LastTime=__g_ThisTime=AccurateTime();


	// Set Priority: - causes trouble and trouble!
/*
	if (!__DG_DEBUG)
	{
		HANDLE hProc=GetCurrentProcess(); // ChokeXP
		::SetPriorityClass(hProc, 0x00008000); // ABOVE_NORMAL_PRIORITY_CLASS - Never ever try realtime on NT systems
	}
*/	

	// Hide if desired (for creation of DLLs, e.g.)
#ifndef GLB_CONSOLE
	#ifdef WIN32
	if(nCmdShow == SW_HIDE)
	{
		__g_pRbow->HideWindow();
		ShowWindow(__hWnd, SW_HIDE);
	}	
	#endif

	pDXin->SystemPointer(!__DG_WANTMOUSE);

	#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
		system("killall -9 gp2xmenu"); // WIZ has menu in background?
	#endif
#else // Set console codepage to ansi
	#ifdef WIN32
		__g_ConsoleCP = ::GetConsoleCP();
		__g_ConsoleCPOut = ::GetConsoleOutputCP();
		SetConsoleCP(1252);
		SetConsoleOutputCP(1252);
	#endif
#endif

	DLL_DELAY_LOAD_NOW();


	#if defined( WIN32) && !defined(_WIN32_WCE) && !defined(GLB_CONSOLE)
	if(EnableMouseInPointer)
	{
		g_debug("Win8+ mouse as pointer\n");
		EnableMouseInPointer(TRUE);
	}
	else
	{
		STDERR(DGStr("!!! EnableMouseInPointer() not found in user32.dll. If this is >=Win8, you get no mouse messages.\n") );
	}
	#endif


	// Mac loads sound dll later
	// Don't change order of sound+buffers!
	#ifndef GLB_CONSOLE
		pDXsnd=new SFXSound;
		DXwav=new  SFXSoundBuffer[MAX_WAV_SAMPLES];
		pDXsnd->Create(__hWnd);

		// Set the main window pointers again.
		#if defined(WIN32) && !defined(WANT_SDL_MIXER)
			pDXsnd->SetWindow(__g_pRbow->m_WND.hwnd);
		#endif
	#endif

	#if defined(ANDROID) // Space Fractal - Added few things to prevent crash
		GFFileSystem::SetCurrentDir("files");
		int okfolder=CREATEDIR(CGStr("Media"));
		SETORIENTATION(-1);
	#endif

	g_debug("Init Finalized\n");
return 0;
}


#if defined(WIN32) && !defined(WANT_SDL_MIXER)
#ifndef GLB_CONSOLE
DGInt gMusicRepeat=0;

	UINT_PTR gTimerIndex = 0;
	#define EVENT_TIMER_MP3_LOOP	(WM_USER+132)
	VOID CALLBACK MusicTimer(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
	{
		MCIWndSeek(wnd_mp3, MCIWND_START);
		MCIWndPlay(wnd_mp3);
	}
#endif
#endif



#ifdef WIN32
	#define GLB_WIN32_MSG "GLB_WIN32_MSG"
	void PeekMessageCallback()
	{
		typedef DGInt (*a_sub)();
		static a_sub mysub = (a_sub)GETSUBADDRESS(GLB_WIN32_MSG);
		if(mysub) mysub();
	}
#endif


// ------------------------------------------------------------- //
#ifndef GLB_CONSOLE
void(*gpWinMsgCallback)() = NULL;

inline unsigned int _WinMSG(bool allow_to_sleep)
{
	#if defined(WIN32) && !defined(WANT_SDL_MIXER)
		if(gMusicRepeat > 0) // loop music, not paused
		{
			DGInt time = AccurateTime();
			if(time >= gMusicRepeat)
			{
				MCIWndSetTimeFormat(wnd_mp3, "ms");
				UINT milliseconds = max(500, MCIWndGetLength(wnd_mp3));
				gMusicRepeat = AccurateTime() + milliseconds;

				MCIWndPlay(wnd_mp3);
				MCIWndSeek(wnd_mp3, MCIWND_START);
				MCIWndPlay(wnd_mp3);
			}
		}

		static int bHasCallBack= -1;
		if(bHasCallBack == -1)
		{
			if(GETSUBADDRESS(GLB_WIN32_MSG))
			{
				bHasCallBack=1;
				gpWinMsgCallback = (void(*)())GETSUBADDRESS(GLB_WIN32_MSG);
			}
			else
				bHasCallBack=0;
		}

		return __g_pRbow->WinMessages(gpWinMsgCallback, allow_to_sleep);
	#else
		return __g_pRbow->WinMessages(gpWinMsgCallback, allow_to_sleep);
	#endif
}
#endif



unsigned int gKeyStatusCounter = 0;


// ------------------------------------------------------------- //
#if defined(ANDROID)
	extern "C" int AndroidScancode(int);
#endif

DGNat	KEY(DGNat key)
{
#if defined(ANDROID)
	return AndroidScancode(key);
#else
	#ifndef GLB_CONSOLE
		if(++gKeyStatusCounter > 128)
		{	
			gKeyStatusCounter=0;
			pDXin->GetKeyboardStatus();
		}
		return (DGNat)(pDXin->GetKey((unsigned char)key));
	#else
		return 0;
	#endif
#endif
}




// GFX-FUNCTIONS

#if defined(TARGET_OS_IPHONE)
	extern "C" void iPhonePlayMovie(const char* pMoviePath);
#endif

// ------------------------------------------------------------- //
DGNat PlayVideo(const DGStr& filename, int loop)
{
#ifndef GLB_CONSOLE
#if !defined( WANT_SDL) && defined(WIN32)
	HWND	wnd_video=NULL;
	DGStr s(filename);
	if(!__g_ShoeBox_Media->FindOrExtract(filename.c_str(), s.getbuffer(1024), false))
		return CMP_NO_FILE;
	s.releasebuffer();
	wnd_video = MCIWndCreate(__hWnd, __hInstance, WS_CHILD | WS_VISIBLE | MCIWNDF_NOMENU |
													MCIWNDF_NOPLAYBAR | MCIWNDF_NOERRORDLG |
													MCIWNDF_NOAUTOSIZEMOVIE | MCIWNDF_NOAUTOSIZEWINDOW
													, s.c_str());
	if (wnd_video)
	{
		long end = MCIWndGetStart(wnd_video) + MCIWndGetLength(wnd_video);

		// Fullscreen
		RECT	r2, rect;
		MCIWndGetSource(wnd_video, &r2);
		::GetClientRect(__hWnd, &rect);

		double sc1, sc2;
		sc1 = (double)(rect.right-rect.left) / (double)(r2.right-r2.left);
		sc2 = (double)(rect.bottom-rect.top)/ (double)(r2.bottom-r2.top);
		sc1=MIN(sc1, sc2);
		r2.right=(LONG)((r2.right-r2.left) *sc1);
		r2.left = 0;
		r2.bottom = rect.top + (LONG)((r2.bottom-r2.top)*sc1);
		r2.top = 0;
		OffsetRect(&rect, -rect.left, -rect.top);
		OffsetRect(&r2, (rect.right-r2.right)/2, (rect.bottom-r2.bottom)/2);
		MCIWndPutDest(wnd_video, &r2); // new RECT
		SetWindowPos(wnd_video,           // window to resize
						NULL,                          // z-order: don't care
						r2.left, r2.top, r2.right-r2.left, r2.bottom-r2.top,
						SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);
		// Fullscreen - end
		while (pDXin->AnyKey() || pDXin->AnyMouse()) ;

		if (MCIWndPlay(wnd_video)==0)
		{
			long cpos = 0;

			while(cpos<end)
			{
				cpos = MCIWndGetPosition(wnd_video);
				if(pDXin->AnyKey() || pDXin->AnyMouse())
					break;
				if (!_WinMSG(false)) __EndProgram();
				if (loop && cpos>=end)
				{
					MCIWndHome(wnd_video);
					MCIWndPlay(wnd_video);
					cpos = -1;
				}
			}
		}
		MCIWndStop(wnd_video);
		MCIWndDestroy(wnd_video);
	}
	else
		return CMP_NO_FILE;
	return CMP_OK;

#elif defined(TARGET_OS_IPHONE)
	iPhonePlayMovie(filename.c_str());
	return CMP_OK;
#else
	printf("PLAYMOVIE + LOOPMOVIE only on Win32 so far...\n");
	return CMP_OK;
#endif

#else
	return 0;
#endif
}


// ------------------------------------------------------------- //
void	PLAYMOVIE(const DGStr& s)
{
	PlayVideo(s, 0);
}

// ------------------------------------------------------------- //
void	LOOPMOVIE(const DGStr& s)
{
	PlayVideo(s, 1);
}

#ifdef WANT_SDL_MIXER

#ifdef PALM_WEBOS
	extern "C" int PDL_NotifyMusicPlaying(SDL_bool MusicPlaying);
#endif

#ifndef GLB_CONSOLE
	Mix_Music*	g_pMixMusic=NULL;
#endif
#endif

DGInt gMusicVolume=-1;

// ------------------------------------------------------------- //
void	PLAYMUSIC(const DGStr& s, DGNat bLoop)
{
	__Error(CMP_OK);
#ifndef GLB_CONSOLE
	STOPMUSIC();
	static DGStr filename;
	filename = s;
	if(!glb_prepare_reading_file(filename)) // Space Fractal :: Forgotten for this command as well.
	{	__Error(CMP_NO_FILE);
		return;
	}

	#if !defined(WANT_SDL_MIXER) && defined(WIN32)

		// Make a window for MCI
	/*
		wnd_mp3 = MCIWndCreate(__hWnd, __hInstance, 
								WS_CHILD | WS_VISIBLE | MCIWNDF_NOMENU |
														MCIWNDF_NOPLAYBAR | MCIWNDF_NOERRORDLG |
														MCIWNDF_NOAUTOSIZEMOVIE | MCIWNDF_NOAUTOSIZEWINDOW| MCIWNDF_NOTIFYPOS,
								filename.c_str());
	*/
		bool systemptr = pDXin->IsSystemPointer();
		pDXin->SystemPointer(false);
		wnd_mp3 = MCIWndCreate(NULL, NULL, WS_POPUP| // 
											MCIWNDF_NOERRORDLG |
											MCIWNDF_NOPLAYBAR|MCIWNDF_NOMENU
											, filename.c_str());
		
		
		MCIWndSetTimeFormat(wnd_mp3, "ms");
		UINT milliseconds = max(500, MCIWndGetLength(wnd_mp3));

	//	gTimerIndex = ::SetTimer(__hWnd, EVENT_TIMER_MP3_LOOP, milliseconds, MusicTimer);
		if(bLoop)
			gMusicRepeat = AccurateTime() + milliseconds;
		else
			gMusicRepeat = 0;

	/*
		UpdateWindow(wnd_mp3);
		ShowWindow(wnd_mp3, SW_SHOW);
		RECT r2 = {0,-30,100,-1};
		MCIWndPutDest(wnd_mp3, &r2); // new RECT

		SetWindowPos(wnd_mp3,           // window to resize
						NULL,                          // z-order: don't care
						r2.left, r2.top, r2.right-r2.left, r2.bottom-r2.top,
						SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);

	//	EnableWindow(wnd_mp3, FALSE);
	*/
		MCIWndSeek(wnd_mp3, MCIWND_START);
		MCIWndPlay(wnd_mp3);
		::SetFocus(__hWnd);

		pDXin->SystemPointer(systemptr);
	#elif defined(WANT_SDL_MIXER)
		#ifdef PALM_WEBOS
			PDL_NotifyMusicPlaying(SDL_TRUE);
		#endif
		g_pMixMusic = Mix_LoadMUS(filename.c_str());

		if (g_pMixMusic)
		{
			Mix_PlayMusic(g_pMixMusic, bLoop ? -1 : 1);
			Mix_VolumeMusic(SDL_MIX_MAXVOLUME);
		}
		else
		{
			g_debug("PLAYMUSIC: %s\n", Mix_GetError());
		}
	#elif defined(APPLE_COREAUDIO)

		iPhonePlayMusic(filename.c_str(), bLoop);

	#endif
		MUSICVOLUME(1.0);

		glb_do_pause(FALSE);
		if(!pDXin->IsSystemPointer())
		{
			pDXin->SystemPointer(true);
			pDXin->SystemPointer(false);
		}
#endif // console
}

DGNat	ISMUSICPLAYING()
{
#ifndef GLB_CONSOLE
	#if !defined(WANT_SDL_MIXER) && defined(WIN32)
		if(wnd_mp3)
		{
			if(gMusicRepeat>0) return TRUE; // loops, so, yes, plays
			LONG pos = MCIWndGetPosition(wnd_mp3);
			if (pos >= MCIWndGetEnd(wnd_mp3))
				return FALSE;
			return TRUE;
		}
		return FALSE;
	#elif defined(WANT_SDL_MIXER)
		return Mix_PlayingMusic() ? TRUE : FALSE;
	#elif defined(TARGET_OS_IPHONE)
		return iPhoneIsMusicPlaying();
	#endif
#else
	return 0;
#endif
}

void PAUSEMUSIC(DGNat bPause)
{
#ifndef GLB_CONSOLE
	#if !defined(WANT_SDL_MIXER) && defined(WIN32)
		static DGInt pause_time=0;
		if(wnd_mp3)
		{
			if(bPause)
			{
				if(gMusicRepeat>0)
				{
					pause_time=AccurateTime();
					gMusicRepeat = -gMusicRepeat;
				}
				MCIWndPause(wnd_mp3);
			}
			else
			{
				if(gMusicRepeat < 0)
				{
					gMusicRepeat = -gMusicRepeat;
					gMusicRepeat += (AccurateTime() - pause_time);
				}
				MCIWndResume(wnd_mp3);
			}
		}
	#elif defined(WANT_SDL_MIXER)
		if(bPause)
		{
			if( Mix_PlayingMusic() ) Mix_PauseMusic();
		}
		else
		{
			Mix_ResumeMusic();
		}
	#elif defined(TARGET_OS_IPHONE)
		iPhonePauseMusic(bPause ? true:false);
	#endif
#endif
}


// ------------------------------------------------------------- //
void MUSICVOLUME(DGInt volume)
{
#ifndef GLB_CONSOLE
	if(gMusicVolume == volume) return;
	gMusicVolume=volume;
	#if !defined(WANT_SDL_MIXER) && defined(WIN32)
		if(wnd_mp3)
		{
			MCIWndSetVolume(wnd_mp3, (LONG)(1000.0 * volume));
		}
	#elif defined(WANT_SDL_MIXER)
		Mix_VolumeMusic( (int)(SDL_MIX_MAXVOLUME * volume));
	#elif defined(TARGET_OS_IPHONE)
		iPhoneMusicVolume(volume);
	#endif
#endif
}


// ------------------------------------------------------------- //
void	STOPMUSIC()
{
#ifndef GLB_CONSOLE
	gMusicVolume=-1;
#if !defined(WANT_SDL_MIXER) && defined(WIN32)
//	if(gTimerIndex) ::KillTimer(__hWnd, EVENT_TIMER_MP3_LOOP); gTimerIndex=0;
	gMusicRepeat=0;
	MCIWndStop(wnd_mp3);
	MCIWndDestroy(wnd_mp3);
	wnd_mp3=NULL;
#elif defined(WANT_SDL_MIXER)
	#ifdef PALM_WEBOS
		PDL_NotifyMusicPlaying(SDL_FALSE);
	#endif

	if(g_pMixMusic)
	{
		Mix_HaltMusic();
		Mix_FreeMusic(g_pMixMusic);
	}
	g_pMixMusic=NULL;
#elif defined(TARGET_OS_IPHONE)
	iPhoneStopMusic();
#endif
#endif
}

// ------------------------------------------------------------- //
void	SETSCREEN(DGNat w, DGNat h, DGNat fs)
{
#ifndef GLB_CONSOLE
	#ifdef TARGET_OS_IPHONE
		fs=1;
	#endif

	#if !defined(GPHWIZ) && !defined(GP2X) && !defined(CAANOO)
		__DG_RESX = (int)w;
		__DG_RESY = (int)h;
		__DG_FULLSCREEN = (int)fs;
		__g_pRbow->SetScreen(__DG_RESX, __DG_RESY, 0, __DG_FULLSCREEN);
		pDXin->SystemPointer( pDXin->IsSystemPointer() );
	#endif

#if defined(WIN32) && !defined(WANT_SDL) && !defined(POCKETPC)
	pDXin->SetWindow(__g_pRbow->m_WND.hwnd);
	// hwnd might have changed!
	__hWnd = __g_pRbow->m_WND.hwnd;
#endif
#if defined(WIN32) && !defined(WANT_SDL_MIXER)
	pDXsnd->SetWindow(__g_pRbow->m_WND.hwnd);
#endif
// otherwise no window
for(int i=0; i<5; ++i)
	{SHOWSCREEN();}
#endif
}

extern void X_MIPMAPPING(DGNat state);




// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void DemoWatermark()
{
#ifndef GLB_CONSOLE
	if (__IsDemoVersion)
	{
		static	int _dmx=0, _dmy=0;
		#define DEMO_DURATION (60000*5) // 5 minutes
		static DGInt _demo_end = 0;
		static DGInt _demo_start = 0;

		if(_demo_end == 0)
		{
			_demo_end = AccurateTime() + DEMO_DURATION;
			_demo_start = AccurateTime();
		}
		DGInt timeall = AccurateTime();
		// factor of demo time (1.0 ... 0.0)
		DGInt percent = (_demo_end - timeall) / DEMO_DURATION;
		DGInt fade=-1;
		//if(percent<0.875 && percent>0.128)
			fade = COS(percent * DEMO_DURATION/60000 * 4 * 360);
		
		// Blinking DEMO
		if(fade<0)
		{
			_dmx = (int)((timeall - _demo_start)/10);
			if(_dmx>100) _dmx=100;
			__g_pRbow->Make2DMode();

			#if defined(TINYGL) || defined(TRENKI3D)
				__g_pRbow->ShowDemo((int)((_dmx>>2) - 32), (float)fade);
				const int logoy=14;
			#else
				__g_pRbow->ShowDemo((int)_dmx, (float)fade);
				const int logoy=84;
			#endif

			if(percent<0)
			{
				ALPHAMODE(0);
				DRAWRECT(0,logoy,64,8,RGB(255,255,255));
				DRAWRECT(1,logoy+1,62,6,RGB(255,0,0));
				__g_pRbow->FlipScreens(gpWinMsgCallback);
				SLEEP(1000);
				__EndProgram();
			}

			ALPHAMODE(fade);
			DRAWRECT(0,logoy, 64,8,RGB(0,0,0));
			DRAWRECT(0,logoy, (DGNat)(1+64*percent),8,RGB((DGNat)MIN(255.0,MAX(0.0,256-percent*512)),(DGNat)MIN(255.0, MAX(0.0,percent*512)),0));

			DRAWLINE(0, logoy,  64,logoy  ,RGB(255,255,255));
			DRAWLINE(0, logoy,   0,logoy+8,RGB(255,255,255));
			DRAWLINE(0 ,logoy+8,64,logoy+8,RGB(255,255,255));
			DRAWLINE(64,logoy,  64,logoy+8,RGB(255,255,255));
			ALPHAMODE(0);
		}
	}
#endif
}



const int MIN_FRAMES_BEFORE_HIBERNATE = 30; // 30x SHOWSCREEN, then allow first try to hibernate -> more fluid experience
static int __g_nFlipSinceHibernate = 0; // number of SHOWSCREENS since last HIBERNATE -> make more responsive for a short time after a mousemove


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void    glb_showscreen_finish_frame()
{
	if(__g_Did3D  && ((__g_LicenseBits & 1) == 0)) __SetDemoVersion();
	if(__g_DidNET && ((__g_LicenseBits & 2) == 0)) __SetDemoVersion();

#ifndef GLB_CONSOLE
	DemoWatermark();

	// don't activate the window with every ondebug
	// makes debugging less flickering
	#if defined(WIN32) && !defined(POCKETPC)
	if(__g_HadBreakPoint)
	{
		printf("activate\n");
		// BringWintowToFront has a "flash taskbar/prevent steal focus" issue
		// Here's the fix
		//Attach foreground window thread
		//to our thread
		HWND hwnd = (HWND)GLBASIC_HWND();
		DWORD nfore = GetWindowThreadProcessId(::GetForegroundWindow(),NULL);
		DWORD nmine = GetCurrentThreadId();
		if( nfore != nmine)
		{
			AttachThreadInput(nfore,nmine,TRUE);
			::ShowWindow(hwnd, SW_SHOW);
			::BringWindowToTop(hwnd);
			::SetActiveWindow(hwnd);
			//Detach the attached thread
			AttachThreadInput(nfore, nmine,FALSE);
		}
		else
		{
			::ShowWindow(hwnd, SW_SHOW);
			::BringWindowToTop(hwnd);
			::SetActiveWindow(hwnd);
		}
		printf("ok\n");
		
		__g_HadBreakPoint=FALSE;
	}
	#endif
	
	__g_LastTime = __g_ThisTime; __g_ThisTime = AccurateTime();

	// Repeat Music
	while(__g_paused)
	{
		__g_nFlipSinceHibernate = MIN_FRAMES_BEFORE_HIBERNATE+1; // allow hibernation
		HIBERNATE();
	}

	#if !defined(ANDROID) // Space Fractal, Not doing this on Android. Using Java system instead.
		gKeyStatusCounter = 0;
		pDXin->GetKeyboardStatus(); // Update keyboard status for next frames
	#endif
	pDXin->RefreshMouseXY();

	// #if defined(HAVE_OPENGL_ES) && defined(WANT_SDL)
	//	SDL_Delay(1); // will wait 10 ms, maybe? Yikes!
	// #endif

	if ((__g_AllowEscape
		&& (pDXin->GetKey(129) || pDXin->GetKey(1 /*DIK_ESCAPE*/ )) ) // ESC
		#ifdef MACOSX
			|| (pDXin->GetKey(219) && pDXin->GetKey(16))                  // Command+Q
		#else
			|| (pDXin->GetKey(56) && pDXin->GetKey(62))                   // ALT-F4
		#endif
		) // ESC | Alt+F4
		goto eXit;

	__g_pRbow3D->DoShadows(false); // finish shadows if any

	_WinMSG(false); // at least do it once, so you have music repeating

	if(!__g_pRbow->FlipScreens(gpWinMsgCallback))
		goto eXit;

return;
eXit:
	__Error(CMP_OK);
	__EndProgram();
#endif
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void    glb_showscreen_init_next_frame()
{
#ifndef GLB_CONSOLE
	VIEWPORT(0,0,0,0); // this calls the prescaler // __g_pRbow->Make2DMode();
	__g_pRbow->PrepareBackbuffer(); // only prepare as if showscreen was done

	__g_pRbow->SetAlphaTesting(1.0f); // no testing
	__g_pRbow->DoAlpha(-1.0f);

	// __g_pRbow->PasteSurface(0);

	X_MIPMAPPING(0);
	__g_pRbow3D->SetShader(-1);

	// SPACEFRACTAL_SEE_HERE - you deleted this line?
	pDXin->m_QueryCount = 100000; // reset keyboard counter on next KEY() call

	++__g_nFlipSinceHibernate;
#endif
}

// *** SPACE FRACTAL FIXES 2017.3.1 ****
#ifdef ANDROID
extern "C"
{ 	void pausemode();
}
#endif

#ifdef TARGET_OS_IPHONE
    extern "C" // **** Space Fractal :: few extern functions is required.
    { 	int iPhonePausemode();
        int iPhonePauseresume();
        int isIphone();
    }
#endif
// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void	SHOWSCREEN()
{
	#if 1 // !defined(EMSCRIPTEN) || defined(TRENKI3D) // don't manually flip in WebGL
	glb_showscreen_finish_frame();
	glb_showscreen_init_next_frame();
	#endif
	#if defined(ANDROID) // Space Fractal :: Fixes for the HIBERNATE issue. Pausemode cannot directly been called from Java, which might crash Virtual Machine instead.
		pausemode(); 
	#else // Space Fractal :: HIBERNATE can also cause some platforms to crash when called from a thread. To prevent it, uses __g_paused_safe=1 for FALSE and __g_paused_safe=2 for TRUE instead.
	
        int __g_paused_safe=0;
		
		#ifdef TARGET_OS_IPHONE
			__g_paused_safe=iPhonePausemode();
			if (isIphone()==3) SETORIENTATION(-1);
		#endif
		
		if(__g_paused_safe>0)
		{	
			if (__g_paused_safe==2) glb_do_pause(TRUE);
			__g_paused_safe=0;
		}
	#endif
}		

// ------------------------------------------------------------- //
void BUSY()
{
	__g_nFlipSinceHibernate = 0; // next few frames will not hibernate
}

// ------------------------------------------------------------- //
void	HIBERNATE()
{
#ifndef GLB_CONSOLE
	// one message loop is a must -> some use while(KEY(123)){HIBERNATE(); }
	if (!_WinMSG(false))
	{
		g_debug("end since WinMSG()==0\n");
		__EndProgram();
	}


	// do nothing more if the last action (mousemove, key, ...) was not long ago
	if(__g_nFlipSinceHibernate < MIN_FRAMES_BEFORE_HIBERNATE)
	{
		return;
	}


DGNat oldmouse[4];
DGNat v1,v2,v3,v4;
int   nj;
	MOUSESTATE(oldmouse[0], oldmouse[1], oldmouse[2], oldmouse[3]);
	for(int nt=0; nt<100; ++nt) // see sleep below -> loop 1 sec w/o updating the display
	{
		if (!_WinMSG(true)) // here you are allowed to sleep until the next message arrives
		{
			g_debug("end since WinMSG()==0\n");
			__EndProgram();
		}

		// Stop in pause mode
		if(__g_paused)
		{
			g_debug("pause loop in showscreen/hibernate\n");
			while(__g_paused && !__g_HadBreakPoint)
			{
				SLEEP(500);
				if (!_WinMSG(true))
				{
					g_debug("end since WinMSG()==0\n");
					__Error(CMP_OK); __EndProgram();
				}
#if defined(ANDROID)
					pausemode(); // Space Fractal - This reequired to get it resumed.
#endif
#if defined(TARGET_OS_IPHONE) // Space Fractal - Required to wake up app from a freeze: ******
                int gl_resume=iPhonePauseresume();
                __g_paused=1;
                if (gl_resume==1) __g_paused=0;
#endif
                SLEEP(500);
                CALLBYNAME(DGStr("GLB_ON_HIDE"));
			}
			g_debug("pause loop ended\n");
            CALLBYNAME(DGStr("GLB_ON_RESUME"));
            __g_paused=0;
            //return;
		}

		pDXin->RefreshMouseXY();
		if (__g_AllowEscape && (pDXin->GetKey(129) || pDXin->GetKey(1 /*DIK_ESCAPE*/ )) ) __EndProgram();

		// KEYS
		for(nj=0; nj<255; ++nj)
		{
			if(KEY(nj))
			{
				__g_nFlipSinceHibernate = 0;
				return;
			}
		}
		
		// MOUSE
		MOUSESTATE(v1, v2, v3, v4);
		if(v1!=oldmouse[0] || v2!=oldmouse[1] || v3!=0 || v4!=0 /*hover=no sleep*/)
		{
			oldmouse[0]=v1;
			oldmouse[1]=v2;
			oldmouse[2]=v3;
			oldmouse[3]=v4;
			__g_nFlipSinceHibernate = 0;
			return;
		}

		// JOYSTYX
		#if 0
		#if !defined(ANDROID) && !defined(PALM_WEBOS) && !defined(PALM_PIXI) && !defined(TARGET_OS_IPHONE)
			nj = (int)GETNUMJOYSTICKS();
			bool bAction=false;
			for(--nj; nj>=0 && !bAction; --nj)
			{
				for(i=0; i<32 && !bAction; ++i)
				{
					if(GETJOYBUTTON(nj, i)) bAction=true;
				}
				if(ABS(GETJOYX(nj))>0.1) bAction=true;
				if(ABS(GETJOYY(nj))>0.1) bAction=true;
				if(ABS(GETJOYZ(nj))>0.1) bAction=true;
				if(GETDIGIX(nj))         bAction=true;
				if(GETDIGIY(nj))         bAction=true;
			}
			if(bAction)
			{
				__g_nFlipSinceHibernate = 0;
				return;
			}
		#endif
		#endif

		// really PAUSE
		SLEEP(10);
	}
#endif
}


// INPUT/OUTPUT
// ------------------------------------------------------------- //
void	INPUT(DGStr& s, DGNat x, DGNat y, DGNat bKerning)
{
#ifndef GLB_CONSOLE
DGStr	str=s, show;
char	c[3];
int		sleeper=0;
int		blink=0;
	c[0]=0; c[1]=0; c[2]=0;
	__g_pRbow->GrabSurface(3);
	__g_pRbow->PasteSurface(3);
	__g_pRbow->Text2D(__nCurFont, s.c_str(), (float)x, (float)y, 0.0f, bKerning, TRUE);
//	pFnt[__nCurFont].Print(s.c_str(), (int)x, (int)y);
	SHOWSCREEN();
	while(true)
	{
		pDXin->GetKeyboardStatus();
		c[0]=pDXin->GetKeyChar();
		if (c[0]=='\0')
		{
			sleeper++; if (sleeper>20) {sleeper=0; blink=1-blink;}
		}
		else if (c[0]=='\r' || c[0]=='\n')
			break; // Return|Enter|ESC
		else if (c[0]==0x08) // Back
		{
			if (s.len()>0) s.left(s.len()-1);
		}
		else
		{
			s+=c;
		}
		__g_pRbow->PasteSurface(3);
		show=s;
		if (blink) show+="_";
		PRINT(show, x, y, bKerning);
		SHOWSCREEN();
	}
	__g_pRbow->FreeSurface(3);
#endif
}

// ------------------------------------------------------------- //
void	INPUT(DGInt& s, DGNat x, DGNat y, DGNat bKerning) {DGStr	str; INPUT(str, x, y, bKerning); s=str;}
void	INPUT(DGNat& s, DGNat x, DGNat y, DGNat bKerning) {DGStr	str; INPUT(str, x, y, bKerning); s=str;}




// ------------------------------------------------------------- //
#ifdef GLB_CONSOLE


static char mygetch()
{
#ifdef WIN32
	if(!_kbhit()) return '\0'; // no key pressed -> don't freeze program.

	char rv = (char)_getch();
	if (rv == 3) // Ctrl+C
	{
		END();
	}
	return rv;
#else


// GF: not cleanly defined sometimes
struct gf_termios {
	tcflag_t c_iflag;
	tcflag_t c_oflag;
	tcflag_t c_cflag;
	tcflag_t c_lflag;
	cc_t c_cc[NCCS];
	speed_t c_ispeed;
	speed_t c_ospeed;
};

	gf_termios oldt, newt;
	char ch=0;
		tcgetattr( STDIN_FILENO, (termios*)&oldt );
		newt = oldt;
		newt.c_lflag &= ~( ICANON | ECHO );
		tcsetattr( STDIN_FILENO, TCSANOW, (termios*)&newt );
		ch = (char)getchar();
		tcsetattr( STDIN_FILENO, TCSANOW, (termios*)&oldt );
	return ch;
#endif
}
#endif



// ------------------------------------------------------------- //
#ifndef GLB_CONSOLE
void DXinp_push_inkey(char ic)
{
	if(!ic) return;

	if(ic==13) ic=10; // \r is \n for me

	if(ic < 32 && ic>0) // special chars
	{
		if(ic!=8 && ic!=9 && ic!=10) // back tab \n
			return;
	}

	// avoid buffer overflow -> ingore new chars when old ones are there
	if(pDXin->m_inkey_chars.Size() < 16)
		pDXin->m_inkey_chars.Push(ic);
}
#endif


// ------------------------------------------------------------- //
DGStr	INKEY_Str()
{
#ifndef GLB_CONSOLE
char	c[2];
DGStr	str;
	c[0]=pDXin->GetKeyChar();
	c[1]='\0';
	str=(const char*)c;
return str;
#else
	char str[4];
	str[1]='\0';
	str[0] = mygetch();
	
	if(str[0]==(char)0xE0)
	{
		str[2]='\0';
		str[0]=(char)0xE0;
		str[1]=mygetch();
	}

	return DGStr(str);
#endif
}

void	JOYSTATE(DGNat& jx, DGNat& jy, DGNat& b1, DGNat& b2)
{
#ifndef GLB_CONSOLE
DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, 0);
	jx=(DGNat) (int)(js.lX/30000.0f);
	jy=(DGNat) (int)(js.lY/30000.0f);
	b1=(DGNat) js.rgbButtons[0];
	b2=(DGNat) js.rgbButtons[1];
#endif
}

// ------------------------------------------------------------- //
void	JOYSTATE(DGInt& jx, DGInt& jy, DGInt& b1, DGInt& b2)
{
#ifndef GLB_CONSOLE
DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, 0);
	jx=(DGInt) (int)(js.lX/30000.0);
	jy=(DGInt) (int)(js.lY/30000.0);
	b1=(DGInt) js.rgbButtons[0];
	b2=(DGInt) js.rgbButtons[1];
#endif
}

#if defined (TARGET_OS_IPHONE) || defined(ANDROID) || defined(PALM_WEBOS)
	#define JOY_AXIS_LIMIT 10.0 // 10x "g"
#else
	#define JOY_AXIS_LIMIT 1.0  // joystick range
#endif

DGInt	GETJOYX(DGNat nJoy)
{
#ifndef GLB_CONSOLE
DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, (int)nJoy);

	// ICADE - do _not_ rotate the controller - just the screen
	#ifdef TARGET_OS_IPHONE
		if(nJoy>0) return MAX(-JOY_AXIS_LIMIT, MIN(JOY_AXIS_LIMIT, js.lX/32767.0f));
	#endif
	__g_pRbow->Rot_Gravity2Display((int)js.lX, (int)js.lY, (int*)&js.lX, (int*)&js.lY);

	return MAX(-JOY_AXIS_LIMIT, MIN(JOY_AXIS_LIMIT, js.lX/32767.0f));
#else
	return 0;
#endif
}

DGInt	GETJOYY(DGNat nJoy)
{
#ifndef GLB_CONSOLE
DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, (int)nJoy);

	// ICADE - do _not_ rotate the controller - just the screen
	#ifdef TARGET_OS_IPHONE
		if(nJoy>0) return MAX(-JOY_AXIS_LIMIT, MIN(JOY_AXIS_LIMIT, js.lY/32767.0));
	#endif

	__g_pRbow->Rot_Gravity2Display((int)js.lX, (int)js.lY, (int*)&js.lX, (int*)&js.lY);
	return MAX(-JOY_AXIS_LIMIT, MIN(JOY_AXIS_LIMIT, js.lY/32767.0f));
#else
	return 0;
#endif
}

DGInt	GETJOYZ(DGNat nJoy)
{
#ifndef GLB_CONSOLE
DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, (int)nJoy);
return MAX(-JOY_AXIS_LIMIT, MIN(JOY_AXIS_LIMIT, js.lZ/32767.0f));
#else
	return 0;
#endif
}
DGInt	GETJOYRX(DGNat nJoy)
{
#ifndef GLB_CONSOLE
DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, (int)nJoy);
	__g_pRbow->Rot_Gravity2Display((int)js.lRx, (int)js.lRy, (int*)&js.lRx, (int*)&js.lRy);
	return MAX(-1.0, MIN(1.0, js.lRx/32767.0f));
#else
	return 0;
#endif
}
DGInt	GETJOYRY(DGNat nJoy)
{
#ifndef GLB_CONSOLE
DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, (int)nJoy);
	__g_pRbow->Rot_Gravity2Display((int)js.lRx, (int)js.lRy, (int*)&js.lRx, (int*)&js.lRy);
return MAX(-1.0, MIN(1.0, js.lRy/32767.0f));
#else
	return 0;
#endif
}
DGInt	GETJOYRZ(DGNat nJoy)
{
#ifndef GLB_CONSOLE
DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, (int)nJoy);
return MAX(-1.0, MIN(1.0, js.lRz/32767.0f));
#else
	return 0;
#endif
}

DGNat	GETDIGIX(DGNat nJoy)
{
#ifndef GLB_CONSOLE
int rx=0, ry=0;

DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, (int)nJoy);
	switch(js.rgdwPOV[0])
	{
		case  4500:
		case  9000:
		case 13500:
			rx=1; break;
		case 31500:
		case 22500:
		case 27000:
			rx=-1; break;
	}
	__g_pRbow->Rot_Display2Screen(rx, ry, &rx, &ry);
	return rx;
#else
	return 0;
#endif
}

DGNat	GETDIGIY(DGNat nJoy)
{
#ifndef GLB_CONSOLE
int rx=0, ry=0;
DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, (int)nJoy);
	switch(js.rgdwPOV[0])
	{
		case 31500:
		case  4500:
		case 0: ry=-1; break;

		case 13500:
		case 22500:
		case 18000: ry=1;break;
	}
	__g_pRbow->Rot_Display2Screen(rx, ry, &rx, &ry);
	return ry;
#else
	return 0;
#endif
}

DGNat	GETJOYBUTTON(DGNat nJoy, DGNat Button)
{
#ifndef GLB_CONSOLE
	if (Button<0 || Button>=32) return 0;
	DIJOYSTATE	js;
	pDXin->GetJoystickData(&js, (int)nJoy);
	return js.rgbButtons[(int)Button];
#else
	return 0;
#endif
}

DGStr	GETJOYNAME_Str(DGNat nJoy)
{
#ifndef GLB_CONSOLE
DGStr	str;
	pDXin->GetJoyName(str.getbuffer(522), (int)nJoy);
	str.releasebuffer();
return str;
#else
	return 0;
#endif
}

DGNat	GETNUMJOYSTICKS()
{
#ifndef GLB_CONSOLE
	return pDXin->GetNumJoysticks();
#else
	return 0;
#endif
}

void	FORCEFEEDBACK(DGNat nJoy, DGInt duration, DGInt x_motor, DGInt y_motor)
{
#ifndef GLB_CONSOLE
	pDXin->ffRumble((int)nJoy, 1, (float)duration, (float)x_motor, (float)y_motor);
#endif
}

DGNat	GETMOUSECOUNT()
{
#ifndef GLB_CONSOLE
	return pDXin->GetNumMice();
#else
	return 0;
#endif
}
void	SETACTIVEMOUSE(DGNat i)
{
#ifndef GLB_CONSOLE
	int imax = pDXin->GetNumMice();
	if (i>=imax) i=0;
	pDXin->SetActiveMouse(i);
#endif
}


DGNat MOUSEAXIS(DGNat nAxis)
{
#ifndef GLB_CONSOLE
	DIMOUSEDATA	di;
	pDXin->GetMouseData(&di, false);

	int acx, acy;
	__g_pRbow->Rot_Display2Screen(di.mAccX, di.mAccY, &acx, &acy);
	switch(nAxis)
	{
	case 0: return (DGNat)(acx);
	case 1: return (DGNat)(acy);
	case 2: return (DGNat)di.mMouseWheel;
	case 3: return di.mButton1 ? 1 : 0;
	case 4: return di.mButton2 ? 1 : 0;
	case 5: return di.mButton3 ? 1 : 0;
	}
return 0;
#else
	return 0;
#endif
}

// ------------------------------------------------------------- //
void	MOUSESTATE(DGInt& jx, DGInt& jy, DGInt& b1, DGInt& b2)
{
#ifndef GLB_CONSOLE
	DGNat a,b,c,d;
	MOUSESTATE(a,b,c,d);
	jx=(DGInt)a; jy=(DGInt)b; b1=(DGInt)c; b2=(DGInt)d;
#endif
}

void	MOUSESTATE(DGNat& jx, DGNat& jy, DGNat& b1, DGNat& b2)
{
#ifndef GLB_CONSOLE
	DIMOUSEDATA	di;
	pDXin->GetMouseData(&di, true);

	b1=di.mButton1 ? 1 : 0;
	b2=di.mButton2 ? 1 : 0;

	DXInput::MousePos& mps = pDXin->GetCurMousePos();
	jx=(DGNat)mps.m_MouseX;
	jy=(DGNat)mps.m_MouseY;

	// rotate to screen coords
	__g_pRbow->Rot_Display2Screen(jx, jy, &jx, &jy);


	// clip
	DGNat scx, scy;

	// GETSCREENSIZE(scx, scy); <- might be offscreen size
	if(__g_pRbow->m_landscape==1 || __g_pRbow->m_landscape==3)
	{
		scy = (DGNat)__g_pRbow->m_WND.width;
		scx = (DGNat)__g_pRbow->m_WND.height;
	}
	else
	{
		scx = (DGNat)__g_pRbow->m_WND.width;
		scy = (DGNat)__g_pRbow->m_WND.height;
	}

	if(jx<0) jx=0;
	if(jy<0) jy=0;
	if(jx>=scx) jx=scx;
	if(jy>=scy) jy=scy;

	if(__g_CurrentScreen == -1 && __g_Prescaler.scale != 0.0f)
	{
		jx-=__g_Prescaler.offx;
		jy-=__g_Prescaler.offy;
		jx=(DGNat)((float)jx/__g_Prescaler.scale);
		jy=(DGNat)((float)jy/__g_Prescaler.scale);
	}



	// rotate back into mouse coords for clipping
	// __g_pRbow->Rot_Screen2Display(jx, jy, &mps.m_MouseX, &mps.m_MouseY);

#endif
}

void	SYSTEMPOINTER(DGInt onoff)
{
#ifndef GLB_CONSOLE
	pDXin->SystemPointer(onoff ? true:false);
#endif
}


// ------------------------------------------------------------- //
void	SETMOUSE(DGNat mx, DGNat my)
{
#ifndef GLB_CONSOLE

	int scx=mx, scy=my;
	__g_pRbow->Rot_Screen2Display(mx, my, &scx, &scy);
	
	pDXin->SetMouse((int)scx, (int)scy);
#endif
}

// ------------------------------------------------------------- //
void	MOUSEWAIT()
{
#ifndef GLB_CONSOLE
	__g_nFlipSinceHibernate = MIN_FRAMES_BEFORE_HIBERNATE+1; // allow to hibernate
	while(pDXin->AnyMouse()) {HIBERNATE();}

	while(pDXin->AnyMouse() == 0)
	{
		// pDXin->GetKeyboardStatus(); if (__g_AllowEscape && (pDXin->GetKey(129) || pDXin->GetKey(1 /*DIK_ESCAPE*/ )) ) __EndProgram();
		HIBERNATE();
	}
	while(pDXin->AnyMouse()) {HIBERNATE();} // wait for release
	__g_nFlipSinceHibernate = 0; // no hibernate for a short time
#endif
}

// ------------------------------------------------------------- //
void	KEYWAIT()
{
#if defined( TARGET_OS_IPHONE )
	MOUSEWAIT();
	return;
#endif
#ifndef GLB_CONSOLE
	__g_nFlipSinceHibernate = MIN_FRAMES_BEFORE_HIBERNATE+1; // allow to hibernate


	int iwait=-1;
	while(iwait==-1)
	{
		pDXin->GetKeyboardStatus(true);
		for (int i = 1; i <= 255; i++)
		{
			if(KEY(i))
				{ iwait = i; break; }
		}
		__g_pRbow->WinMessages(NULL, true);
	}
	pDXin->GetKeyboardStatus(true);
	while (KEY(iwait) != 0)
	{
		SLEEP(100);
		__g_pRbow->WinMessages(NULL, false);
		pDXin->GetKeyboardStatus(true);
	}


	__g_nFlipSinceHibernate = 0; // no hibernate for a short time

#else // Console
	#if defined( WIN32 ) && !defined( _WIN32_WCE)
		char rv = (char)_getch(); // windows function that does not type the chracter to stdout
	#else
		char rv = getchar(); // std library. Will print the input character, though :(
	#endif
	if (rv == 3) // Ctrl+C
	{
		END();
	}
#endif
}



// ------------------------------------------------------------- //
DGNat	GENSOUND()
{
	CLEARERROR();
	#ifndef GLB_CONSOLE
	for(DGNat i=MAX_WAV_SAMPLES-1; i>=0; --i)
	{
		if( !DXwav[i].IsLoaded() )
			return i;
	}
	#endif
	__Error(CMP_OUT_OF_MEMORY);
	return -1;
}

// ------------------------------------------------------------- //
void	LOADSOUND(const DGStr& s, DGNat num, DGNat buffers)
{
	CLEARERROR();
#ifndef GLB_CONSOLE
	// Buffer check = Internal
	if(num < 0 || num >= (DGNat)MAX_WAV_SAMPLES)
	{
		__Error(CMP_OUT_OF_MEMORY);
		return;
	}

	DGStr file(s);
	if(!glb_prepare_reading_file(file))
	{
		__Error(CMP_NO_FILE);
		return;
	}

	// No error, just free mem
	if (!DXwav[(int)num].Load(pDXsnd, file.c_str(), (int)buffers))
	{
		__Error(CMP_NO_FILE);
	}
#endif
}

// ------------------------------------------------------------- //
DGNat	PLAYSOUND(DGNat num, DGInt fade, DGInt volume)
{
#ifndef GLB_CONSOLE
	if(num<0 || num>=MAX_WAV_SAMPLES)
	{
		__Error(CMP_OUT_OF_MEMORY);
		return 0;
	}
	fade =MAX(-1.0, MIN(fade, 1.0));
	int ivolume= (int)(MAX(0, MIN(100.0*volume, 100.0)));
	int ifade   =(int)(fade*(abs(DXSOUND_PAN_RIGHT)));
	DGNat rv;

	rv = (DGNat)DXwav[(int)num].Play(ifade, ivolume);
	return rv<0 ? rv : (num << 8) + rv; // low byte = channel, above is sound index
#else
	return 0;
#endif
}

// ------------------------------------------------------------- //
DGNat	SOUNDPLAYING(DGNat channel)
{
#ifndef GLB_CONSOLE
int iwav, n;

	n = (int)channel;
	iwav = n >> 8;
	if(iwav<0 || iwav>=MAX_WAV_SAMPLES)
	{
		__Error(CMP_OUT_OF_MEMORY);
		return FALSE;
	}
	return DXwav[iwav].IsPlaying( n & 0xff );
#else
	return 0;
#endif
}

// ------------------------------------------------------------- //
void STOPSOUND(DGNat channel)
{
#ifndef GLB_CONSOLE
int iwav, n;

	if(channel<0) return;
	n = (int)channel;
	iwav = n >> 8;
	DXwav[iwav].Stop( n & 0xff );
#endif
}


// ------------------------------------------------------------- //
void	HUSH()
{
#ifndef GLB_CONSOLE
	for (int i=0; i<MAX_WAV_SAMPLES; i++) {DXwav[i].Stop();}
#endif
}

// ------------------------------------------------------------- //
#if defined(ANDROID) // Space Fractal, this one disable the Back button, require a Java Call
	extern "C" int AndroidEscape(int);
#endif

void	ALLOWESCAPE(DGNat i)
{	
	#if defined(ANDROID)
		AndroidEscape(i);
	#endif
	__g_AllowEscape = i ? TRUE : FALSE;
}


// ------------------------------------------------------------- //
DGStr	FILEREQUEST_Str(DGNat open, const DGStr& filter)
// ------------------------------------------------------------- //
{
DGStr			retstr;
#if !defined(WANT_SDL) && defined(WIN32)
	int				i;
	char			filt[1024];
	char			dir[1024];
	char			file[1024];
	OPENFILENAME	ofn;
		memset(&ofn, 0, sizeof(ofn));

		memset(file, 0, 1024);
		memset(dir,  0, 1024);
		memset(filt, 0, 1024);

		strcpy(filt, filter.c_str());
		for (i=0; i<(int)filter.len(); i++)
		{
			if (filt[i]=='|')
			{
				filt[i]='\0';
			}
		}
		filt[filter.len()]='\0';
		ofn.lpstrFilter=filt;

		// Get CurDir
		::GetCurrentDirectory(1024, dir);
		ofn.lpstrInitialDir=dir;

		ofn.lStructSize=sizeof(ofn);
		ofn.hwndOwner=(HWND)GLBASIC_HWND();
		ofn.lpstrFile=file;
		ofn.nMaxFile=1024;
		ofn.Flags=OFN_HIDEREADONLY|OFN_EXPLORER;
	//	if (open)
	//		ofn.Flags|=OFN_FILEMUSTEXIST;

		#ifndef GLB_CONSOLE
			__g_pRbow->HideWindow();
		#endif

		BOOL ok=FALSE;
		if (open)
		{
			ok=GetOpenFileName(&ofn);
		}
		else
		{
			ok=GetSaveFileName(&ofn);
		}

		::SetCurrentDirectory(dir);
		#ifndef GLB_CONSOLE
			if (__DG_FULLSCREEN)
			{
				SETSCREEN(__DG_RESX, __DG_RESY, TRUE);
			}
			else
			{
				__g_pRbow->DoWindowed();
			}
		#endif

		retstr=(const char*)ofn.lpstrFile;
		if (!ok || retstr[0]=='\0')
			retstr="NO_FILE";
		else
		{
			for (i=0; i<strlen(retstr.c_str()); i++)
			{
				if (retstr[i]=='\\') retstr[i]= '/';
			}
		}
#else // SDL-Version - Doh!!
	retstr = "File:";
	INPUT(retstr, 0, 100);
#endif
	return retstr;
}


// --------------------------------------------------------------- //
// Stop process and hand over to debugger
// --------------------------------------------------------------- //
#ifdef WIN32
#pragma comment(lib, "kernel32.lib")
extern "C" int __stdcall IsDebuggerPresent();
#endif
void __PauseAndDebug(void* paddress)
{
#if defined (WIN32)


	#ifndef _DEBUG
		if (!IsDebuggerPresent())
		{
			return;
		}
		

		#pragma warning(disable: 4311)
		#if defined(_WIN64)
			static char s[64+8];
			for(int i=0; i<64+8; ++i) s[i]=0;
			s[0]='B'; s[1]='P';
			_itoa(reinterpret_cast<size_t>(paddress), s+2, 16);
		#else
			static char s[32+8];
			for(int i=0; i<32+8; ++i) s[i]=0;
			s[0]='B'; s[1]='P';
			_itoa(reinterpret_cast<int>(paddress), s+2, 16);
		#endif


		// Set manually to make debugable in VC
		// __g_dbg.firsttime=0;
		#ifndef GLB_CONSOLE
			pDXin->UnaquireMouse();
//			::ShowWindow(__hWnd, SW_HIDE);
		#endif

			::OutputDebugStringA(s);

		#ifndef GLB_CONSOLE
	/*
			::ShowWindow(__hWnd, SW_SHOW);
			::BringWindowToTop(__hWnd);
			::SetActiveWindow(__hWnd);
	*/
			__g_paused = FALSE;
			pDXin->GetKeyboardStatus();
			pDXin->RefreshMouseXY();
		#endif

		__g_HadBreakPoint=TRUE; // force window to top next time you do a SHOWSCREEN
	#endif
#endif
}


}; // namespace

#ifdef PALM_WEBOS
	extern "C" void PDL_Quit();
#endif



// Called in any case the PRG will be stopped
extern "C" void __EndProgram()
{
static char ending=0; if(ending)return;	ending=1; // might be called recursively. Bad!

	if(__DG_DEBUG)
	{
		g_debug("-End On Line\n");
		g_debug("%d\n", (int)__g_error);
		g_debug("%d\n", (int)__g_errorline -1 );
		g_debug("%d\n", (int)__g_debugline);
		g_debug("%d\n", (int)__g_debugfile);
	}
	else
	{
//		g_deldebug();
	}
	
	#if defined(POSIX)
	if(__g_error != CMP_OK)
	{
		printf("GLB Error: %x\n", __g_error);
	}
	#endif

	__ShutDownGLBasic();

	#ifdef PALM_WEBOS
	// crashes like mad
	printf("PDL_Quit\n");
	PDL_Quit(); // before SDL_Quit
	printf("</PDL_Quit>\n");
	#endif

	#ifdef WANT_SDL
		SDL_Quit();
	#endif



#if !defined(WANT_SDL) && defined(WIN32)
	#ifndef GLB_CONSOLE
	// PostQuitMessage(0);
	::InvalidateRect(0, NULL, TRUE);
	#endif
#endif



#if defined(_DEBUG) && (defined(GF_WANT_DEBUG_NEW) || defined(MSVC_DEBUG_NEW_USED))
	class DD{public: DD()
					{}
					 ~DD()
	{
		#ifdef GF_WANT_DEBUG_NEW
		__debug_new_DumpUnfreed();
		#endif
		#ifdef _MSC_VER
			_CrtDumpMemoryLeaks();
		#endif
	}
	};
	
	static DD __dd__;
#endif

#if defined(GLB_CONSOLE) || defined(POSIX)
	fflush( stdout );
#endif


#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
	// sync the written data to the disk
	system("sync");


/*
	// use a script instead to be safe
	DGStr exepath;
	GFFileSystem::GetExeDir(exepath.getbuffer(1024));
	exepath.releasebuffer();
	GFFileSystem::SetCurrentDir(exepath.c_str());

	if(GFFileSystem::DoesFileExist("gp2xmenu.gpe"))
		execl("gp2xmenu.pge","gp2xmenu.gpe",NULL);
	if(GFFileSystem::DoesFileExist("gp2xmenu.gpu"))
		execl("gp2xmenu.pgu","gp2xmenu.gpu",NULL);
	if(GFFileSystem::DoesFileExist("gp2xmenu"))
		execl("gp2xmenu","gp2xmenu",NULL);

	chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
*/
	fflush( stdout );
	SLEEP(500);
#endif
	printf("exit\n");

#if defined(EMSCRIPTEN)
	SETLOOPSUB("");
	printf("Exit with *(int*)0=0 - sorry\n");
	int* pNull1 = NULL;
	*pNull1=0;
#endif

	exit(0);

	//printf("Exit with *(int*)0=0 - sorry\n");
	//int* pNull = NULL;
	//*pNull=0;
	// ExitProcess(0); // fails with memory exception
}
