#if !defined(AFX_MESHOBJECT_H__A54DCD42_E1B4_11D5_B6EB_003042FFFD01__INCLUDED_)
#define AFX_MESHOBJECT_H__A54DCD42_E1B4_11D5_B6EB_003042FFFD01__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "gfqueue.h"
#include "vectormath.h"

#define NUMVERTEXNORMALS	162
extern float gAvertexnormals[NUMVERTEXNORMALS][3];


#define	COL_NEW_TRI		0xFFFFFFFF // COL_NEW_TRI=Triangles COL_NEW_STRIP=TriStrips COL_NEW_FAN=TriFan
#define COL_NEW_STRIP	0xFFFFFFFE
#define COL_NEW_FAN		0xFFFFFFFD
#define COL_NEW_ANY		0xFFFFFFFC // Bigger than this-> a new part starts


// Data
class POLY_NODE
{
public:
	// KEEP this order for load/save/memcpy operations!!
	float			pos[3];
	float			normal[3];
	float			tex[2];
	// Additional (not saved) data
	float			bump_light[3]; // offset vertex glColor
	float			tangent[3];
	float			binormal[3];
	inline void col(unsigned long c)
	{
		dwcol = c;
		if(c<COL_NEW_ANY)
		{
			#ifdef TINYGL // #NEED this is strange!
			#define LIMITCOL 253
			fcol[0] =  (c      &0xff)/ 255.0f;
			fcol[1] = ((c>> 8) &0xff)/ 255.0f;
			fcol[2] = ((c>> 16)&0xff)/ 255.0f;
			dwcol = (unsigned int)((int)(fcol[2]*LIMITCOL)*(256*256) + (int)(fcol[1]*LIMITCOL)*256 + (int)(fcol[0]*LIMITCOL));
			#else
			fcol[0] =  (c      &0xff)/ 255.0f;
			fcol[1] = ((c>> 8) &0xff)/ 255.0f;
			fcol[2] = ((c>> 16)&0xff)/ 255.0f;
			#endif
			fcol[3]=1.0f;
		}
		else
		{
			fcol[0]=fcol[1]=fcol[2]=fcol[3]=1.0f;
		}
	}
	inline unsigned long col()
	{
		return dwcol;
	}
	
	unsigned long	dwcol;
	unsigned char	icol; // Required for cel shading
	float			fcol[4]; // ES requires "4" of them :(
};

// compressed node (stored in file)
struct XPOLY_NODE
{
	unsigned char	pos[3];
	unsigned char	inormal, itangent, ibinormal;
	float			tangent[3];
	float			binormal[3];
	unsigned char	icol; // Index in color table
};

class POLY_FRAME
{
public:
	Queue <POLY_NODE>	nodes;		// The Nodes of the mesh (uncomressed xnodes -> triangle strip/fan info, too)
	Queue <POLY_NODE>	triangles;	// For fast VBO rendering. Every 3 nodes are one triangle.

	float				max_radius;	// For pre-calc of collisions - around center of mass
	float				center_of_mass[3];
	POLY_FRAME();
	~POLY_FRAME();
	POLY_FRAME& operator = (const POLY_FRAME& f);
};

class MeshObject
{
public:
	// Data
	Queue <POLY_FRAME>	m_Frames; // The Animation Frames
	Queue <POLY_NODE>	m_tri_animation; // triangles for interoplated animation
//	Queue <float>		m_Texture[2]; // The Texture coords (once for model, thus not part of node)
	Queue <unsigned long> m_ColTable; // color table, for cel shading & saving
	long				m_HasNormals; // Does the object already have calculated normals?
	long				m_HasBumpNormals; // Does the object already have calculated bump normals?

	#if defined(TINYGL) || defined(TRENKI3D)
	Queue<int>			m_glListIDs;
	#endif

public:
	// Functions
	MeshObject();
	~MeshObject();
public:
	bool				Save(const char* file);
	bool				Load(const char* file, bool uncompress=false); // Want uncompression for adding further frames later?

	void				Clear(bool free_mem);
	virtual void		PrepareTriangles()		{}

	inline int			NumFrames()				{return m_Frames.Size();}
	inline int			NumNodes(int frame)		{return m_Frames.Size()<=frame ? 0 : m_Frames.GetAt(frame)->nodes.Size();}
	inline void			PushNode(int frame, float pos[3], float tex[2], unsigned long col)
						{
						POLY_NODE pn;
							pn.pos[0]=pos[0]; pn.pos[1]=pos[1]; pn.pos[2]=pos[2];
							pn.tex[0]=tex[0]; pn.tex[1]=tex[1];
							
							pn.fcol[0] =  (col      &0xff)/ 255.0f;
							pn.fcol[1] = ((col>> 8) &0xff)/ 255.0f;
							pn.fcol[2] = ((col>> 16)&0xff)/ 255.0f;
							m_Frames.GetAt(frame)->nodes.Push(pn);
						}

	void				AutoNormalsAfterCompression(bool wantnormals, bool surfacenormals, bool wantbumpvertices);
	// void				Compress(void);
	void				BuildColorTable();
	inline void			Uncompress(Queue<XPOLY_NODE>& xnodes, float* pU, float* pV, float* move, float* scale, POLY_FRAME* pFrame)
						{
							pFrame->nodes.SetSize(xnodes.Size());
							for(int j=0; j<(int)xnodes.Size(); ++j)
							{
								InternUncompress(xnodes, pU, pV, move, scale, pFrame, j);
							}
						}
protected:
	inline void			InternUncompress(Queue<XPOLY_NODE>&	xnodes, float* pU, float* pV, float* move, float* scale, POLY_FRAME* pFrame, int index)
						{
						POLY_NODE*				pNode;
						XPOLY_NODE*	pXNode;
						int i;
							pXNode=&xnodes[index];
							pNode =&pFrame->nodes[index];
							for (i=0; i<3; i++)
							{
								pNode->pos[i]=((float)pXNode->pos[i]-move[i]) * scale[i];
								pNode->normal  [i]=gAvertexnormals[pXNode->inormal  ][i];
								pNode->tangent [i]=gAvertexnormals[pXNode->itangent ][i];
								pNode->binormal[i]=gAvertexnormals[pXNode->ibinormal][i];
							}
							pNode->col(*m_ColTable.GetAt(pXNode->icol) );
						/*
							pNode->fcol[0] = ((unsigned char) (pNode->col())) / 255.0f;
							pNode->fcol[1] = ((unsigned char) (((unsigned short) (pNode->col())) >> 8)) / 255.0f;
							pNode->fcol[2] = ((unsigned char) ((pNode->col()) >> 16)) / 255.0f;
						*/
							pNode->icol=pXNode->icol; // For cel shading
							pNode->tex[0]=  pU[index];
							pNode->tex[1]= -pV[index]; // texture flipped due to grab sprite compatibility
						}
};


//------------------------------------------------------------------//
// Get Normal Index closest to the given vertex
//------------------------------------------------------------------//
inline int	gGetNormalIndex(const float* normal)
{
int				maxindex;
float			maxphi=10000.0f, ang;
float			v1[3], v2[3];
	maxindex=0;
	// Normal
	gCopyVertex(v1, normal);
	gUnify(v1);
	for (int j=0; j<NUMVERTEXNORMALS; j++)
	{
		v2[0]=-gAvertexnormals[j][0];
		v2[1]=-gAvertexnormals[j][1];
		v2[2]=-gAvertexnormals[j][2];
		gAddVertex(v2, v1);
		ang = (float)fabs(gDot(v1, v2));
		if (ang<=maxphi)
		{
			maxindex=j;
			if (ang==0.0f) // found exact - why try further
				break;
			maxphi = ang;
		}
	}
return maxindex;
}

#endif


