#ifndef _OGL_IMAGE_H_INC_
#define _OGL_IMAGE_H_INC_

// An IMAGE with RGBA content. A=255 -> solid
struct OPENGL_IMAGE
{
	OPENGL_IMAGE() {m_sizex=m_sizey=m_mem=0; m_txdata=NULL; m_bSourceHadAlpha=true; }
	~OPENGL_IMAGE() {clear();}
	bool resize(int x, int y) // clears all data. Colour=black, transparent
	{
		int alloc = x*y*4;
		if(alloc>m_mem)
		{
			clear();
			m_mem = alloc;
			m_txdata = new unsigned char[alloc];
			if(!m_txdata) return false;
		}
		m_sizex=x; m_sizey=y;
		memset(m_txdata, 0, alloc);
		return true;
	}


	void clear() {m_sizex=m_sizey=m_mem=0; if(m_txdata) delete[] m_txdata; m_txdata=NULL; m_bSourceHadAlpha=true;}
	// Rainbows.cpp
	bool LoadImageBMP(const char* szFileName, COLORREF cookie=-1);
	bool LoadImagePNG(const char* szFileName, COLORREF cookie=-1);
	bool LoadImageJPG(const char* szFileName);

	void FixNeighbourAlpha(); // for cookie-coloured images
	
	void MakeDefFont();
	void ScaleX2(); // scale by 2x2

	inline unsigned char& operator[](int pos)
	{
		#ifdef _DEBUG
		if(pos<0 || pos>=m_mem)
			*(int*)0=42;
		#endif
		return *(m_txdata + pos);
	}
	OPENGL_IMAGE& operator=(const OPENGL_IMAGE& o) {resize(o.m_sizex, o.m_sizey); memcpy(m_txdata, o.m_txdata, m_sizex*m_sizey*4); return *this;}
	
	// flip upside down
	void flipy()
	{
		if(m_sizex<=0) return;
		int width = m_sizex*4;
		unsigned char* pLine = new unsigned char[width];
		
		for(int y=0; y<m_sizey/2; ++y)
		{
			memcpy(pLine, &m_txdata[y*width], width);
			memcpy(&m_txdata[y*width], &m_txdata[(m_sizey-1-y)*width], width);
			memcpy(&m_txdata[(m_sizey-1-y)*width], pLine, width);
		}
		delete[] pLine;
	}

	void swap(OPENGL_IMAGE& i2)
	{
		#define SSSW(tp, x) {tp sw=x; x=i2.x; i2.x=sw;}
		SSSW(int, m_sizex);
		SSSW(int, m_sizey);
		SSSW(unsigned char*, m_txdata);
		SSSW(int, m_mem);
		SSSW(bool, m_bSourceHadAlpha);
	}
	void swap_red_blue()
	{
		if(!m_mem) return;
		unsigned char* pEnd = m_txdata + 4*m_sizex*m_sizey;
		unsigned char* pRGBA = m_txdata;
		unsigned char swp;
		while(pRGBA<pEnd)
		{
			swp=pRGBA[0]; pRGBA[0]=pRGBA[2]; pRGBA[2]=swp;
			pRGBA+=4;
		}
	}

	// make bigger
	bool expand(int w, int h)
	{
		if(w==m_sizex && h==m_sizey) return true;
		OPENGL_IMAGE i2;
		i2.resize(w,h);
		w=min(w, m_sizex);
		h=min(h, m_sizey);
		int linew = w*4;

		unsigned char* pSrc = (unsigned char*)m_txdata;
		unsigned char* pDst = (unsigned char*)i2.m_txdata;
		for(int y=0; y<h; ++y)
		{
			memcpy(pDst, pSrc, linew);
			pDst+=(i2.m_sizex*4);
			pSrc+=(m_sizex*4);
		}
		this->swap(i2);
		return true;
	}

	bool clip(int w, int h) // clip right and top (only make smaller!) for n^2->odd size
	{
		if(w==m_sizex && h==m_sizey) return true;
		if(w>m_sizex  || h>m_sizey)  return false;
		int* pSrc = (int*)m_txdata;
		int* pDst = (int*)m_txdata;
		const int skip = m_sizex-w;
		for(int y=0; y<h; ++y)
		{
			for(int x=0; x<w; ++x)
			{
				*pDst++ = *pSrc++;
			}
			pSrc += skip;
		}
		m_sizex=w;
		m_sizey=h;
		return true;
	}


/* unused, but might work
	void rotate90left()
	{
		OPENGL_IMAGE i2;
		i2.resize(m_sizey, m_sizex);
		
		unsigned int* pDst = (unsigned int*)i2.m_txdata;
		unsigned int* pSrc = (unsigned int*)this->m_txdata;
		register unsigned int* pLine = pDst + (m_sizex-1)*m_sizey;
		for(register int y=0; y<m_sizey; ++y)
		{
			pDst = pLine++;

			#define D_ONE {*pDst = *pSrc++; pDst-=m_sizey;}
			for(register int x=0; x<m_sizex; ++x)
			{
				D_ONE
			}
		}
		*this = i2;
	}
*/
	
	// rotate 90 (needed for landscape iPad grabscreen)
	void rotate90right()
	{
		OPENGL_IMAGE i2;
		i2.resize(m_sizey, m_sizex);
		
		unsigned int* pDst = (unsigned int*)i2.m_txdata;
		unsigned int* pSrc = (unsigned int*)this->m_txdata;
		unsigned int* pLine = pDst + m_sizey-1;
		for(int y=0; y<m_sizey; ++y)
		{
			pDst = pLine--;
			for(int x=0; x<m_sizex; ++x)
			{
				*pDst = *pSrc++; pDst+=m_sizey;
			}
		}
		*this = i2;
	}
	
	void debug()
	{
		g_debug("image: %dx%d (had alpha %d)\n", m_sizex, m_sizey, (int)m_bSourceHadAlpha);
		for(int y=0; y<m_sizey; ++y)
		{
			for(int x=0; x<m_sizex; ++x)
			{
				int* rgb = (int*)&m_txdata[(x+y*m_sizex) * 4];
				int r=GetRValue(*rgb);
				int g=GetGValue(*rgb);
				int b=GetBValue(*rgb);
				int rgb8 = (r<<16)+(b<<8)+g;
				g_debug("%.3x,", rgb8);
			}
			g_debug("\n");
		}
	}
	
	bool	m_bSourceHadAlpha; // the source of this image (BMP/PNG) had alpha information?
	int m_sizex, m_sizey;
	unsigned char* m_txdata;
	int m_mem;
};

#endif
