#ifndef BESTPLANNER_H
#define BESTPLANNER_H
/**
  @class BestPlanner
  @brief "Best-first" path-finding algorithm.

  In "best-first" path-finding, each possible node is given a heuristic weight
  based on its distance from the end point.  The node deemed as "closest" is
  examined first.  As such, walls are taken into account, but ground weighting
  is ignored.  The path found is typically not too optimal except in the case
  of maps with little or no weighting or obstacles.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.

  @date  Created: Thu Mar 20 2003
*/

#include "planner.h"
#include <vector>

//----------------------------------------------------------------------------
class BestPlanner : public Planner
{
public:
	//------------------------------------------------------------------------
	/** @name Running */
	//@{

	/// Initialize the path-finding algorithm.
	virtual void initialize ();
	/// Run a single time slice of the path-finding algorithm.
	virtual bool run ();

	//@}
};

#endif	// BESTPLANNER_H
