#ifndef DIJKSTRAPLANNER_H
#define DIJKSTRAPLANNER_H
/**
  @class DijkstraPlanner
  @brief Dijkstra's path-finding algorithm.

  In Dijkstra's path-finding, the next node checked out of potential path
  nodes is the one with the shortest cost so far.  Ground costs are taken into
  account as each node is processed.  This algorithm will give the optimum
  path no matter what, but can be very slow in searching.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.

  @date  Created: Sun Mar 23 2003
*/

#include "planner.h"
#include <vector>

//----------------------------------------------------------------------------
class DijkstraPlanner : public Planner
{
public:
	//------------------------------------------------------------------------
	/** @name Running */
	//@{

	/// Initialize the path-finding algorithm.
	virtual void initialize ();
	/// Run a single time slice of the path-finding algorithm.
	virtual bool run ();

	//@}
};

#endif	// DIJKSTRAPLANNER_H
