// Rainbows3D.cpp: Implementierung der Klasse Rainbows3D.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#ifndef GLB_CONSOLE

#include "rainbows3d.h"
#include "openglrainbows.h"

#include <math.h>			// sqrt
#include "gffilesystem.h"


#if defined(HAVE_OPENGL) || (defined(HAVE_OPENGL_ES) && defined(_DEBUG))
	// OpenGL extensions required
	extern PFNGLACTIVETEXTUREARBPROC		glActiveTextureArb; // Texture Unit
	extern PFNGLCLIENTACTIVETEXTUREARBPROC	glClientActiveTextureArb;  // Vertex Array Unit
	extern PFNGLMULTITEXCOORD2FARBPROC		glMultiTexCoord2fArb;
	// VBO Extension Function Pointers
	extern PFNGLGENBUFFERSARBPROC			glGenBuffersArb;			// VBO Name Generation Procedure
	extern PFNGLBINDBUFFERARBPROC			glBindBufferArb;			// VBO Bind Procedure
	extern PFNGLBUFFERDATAARBPROC			glBufferDataArb;			// VBO Data Loading Procedure
	extern PFNGLDELETEBUFFERSARBPROC		glDeleteBuffersArb;			// VBO Deletion Procedure
	extern PFNGLMAPBUFFERARBPROC			glMapBufferArb;
	extern PFNGLUNMAPBUFFERARBPROC			glUnmapBufferArb;
	extern PFNGLGETBUFFERPARAMETERIVARBPROC	glGetBufferParameterivArb;

	// GLSL
	extern PFNGLCREATEPROGRAMOBJECTARBPROC     glCreateProgramObjectArb;
	extern PFNGLDELETEOBJECTARBPROC            glDeleteObjectArb;
	extern PFNGLCREATESHADEROBJECTARBPROC      glCreateShaderObjectArb;
	extern PFNGLSHADERSOURCEARBPROC            glShaderSourceArb;
	extern PFNGLCOMPILESHADERARBPROC           glCompileShaderArb;
	extern PFNGLGETOBJECTPARAMETERIVARBPROC    glGetObjectParameterivArb;
	extern PFNGLATTACHOBJECTARBPROC            glAttachObjectArb;
	extern PFNGLGETINFOLOGARBPROC              glGetInfoLogArb;
	extern PFNGLLINKPROGRAMARBPROC             glLinkProgramArb;
	extern PFNGLUSEPROGRAMOBJECTARBPROC        glUseProgramObjectArb;
	extern PFNGLGETUNIFORMLOCATIONARBPROC      glGetUniformLocationArb;
	extern PFNGLUNIFORM1FARBPROC               glUniform1fArb;
	extern PFNGLUNIFORM1IARBPROC               glUniform1iArb;
	extern PFNGLGETATTRIBLOCATIONARBPROC       glGetAttribLocationArb;
	extern PFNGLVERTEXATTRIB1FARBPROC          glVertexAttrib1fArb;
#endif


// GLOBALS
int				gCurObject=-1;		// For Creation



// ------------------------------------------------------------------ //
// Preapare a triangle VBO buffer for the objects
// created frron nodes, not xnodes!!
// ------------------------------------------------------------------ //
void MeshObjectX::PrepareTriangles()
{
	POLY_NODE*	nods[3];
	int i;
	if (!m_Frames.Size()) return;

	for (i=0; i<m_Frames.Size(); i++)
	{
		m_Frames.GetAt(i)->triangles.Clear(false);
		m_Frames.GetAt(i)->triangles.AllocAtLeast(m_Frames.GetAt(i)->nodes.Size()/3);
		long ndrawn=0; unsigned long mode=COL_NEW_TRI; unsigned long col;
		long odd=0;
		for (int j=0; j<m_Frames.GetAt(i)->nodes.Size(); j++)
		{
			POLY_FRAME& frm = m_Frames[i];
			POLY_NODE& node = frm.nodes[j];
			col = m_ColTable[node.icol];
			switch (col)
			{
			case COL_NEW_TRI:  ndrawn=0; mode=COL_NEW_TRI;   continue;
			case COL_NEW_STRIP:ndrawn=0; mode=COL_NEW_STRIP; odd=0; continue;
			case COL_NEW_FAN:  ndrawn=0; mode=COL_NEW_FAN;   continue;
			default:
				;
				// color is set to col
			}
			nods[ndrawn] = &frm.nodes[j];
			// Uncompress(i, j, &nods[ndrawn]);
			ndrawn++;
			if (ndrawn>2)
			{
				m_Frames[i].triangles.Push(*nods[0]);
				m_Frames[i].triangles.Push(*nods[1]);
				m_Frames[i].triangles.Push(*nods[2]);
				switch(mode)
				{
					case COL_NEW_TRI:  ndrawn=0; break;
					case COL_NEW_STRIP:
						/*
							1    3     5

							2    4     6

							123
							324
							345
						*/
						ndrawn=2;
						odd=1-odd;
						if (odd)
						{
							nods[0]=nods[2];
						}
						else
						{
							nods[1]=nods[2];
						}
						break;
					case COL_NEW_FAN:
						ndrawn=2;
						nods[1] = nods[2];
						break;
				}
			}
		}
	}
	
	m_tri_animation = m_Frames[0].triangles;
	/*
	m_tri_animation.Clear(false);
	m_tri_animation.AllocAtLeast(m_Frames[0].triangles.Size());
	for (i=0; i<m_Frames[0].triangles.Size(); i++)
		m_tri_animation.Push(m_Frames[0].triangles[i]);
	*/
	
	// See if we need a color pointer, or just white
	CheckIfColoringNeeded();
	

	#if defined(TINYGL)
		// Make a display list for TinyGL
		m_glListIDs.SetSize(m_Frames.Size());
		int liststart = glGenLists(m_Frames.Size());
		for(i=0; i<m_Frames.Size(); ++i)
		{
			m_glListIDs[i] = liststart + i;
			glNewList(m_glListIDs[i], GL_COMPILE);
			glVertexPointer(3, GL_FLOAT, sizeof(POLY_NODE), m_Frames[i].triangles[0].pos);
			glTexCoordPointer(2, GL_FLOAT, sizeof(POLY_NODE), m_Frames[i].triangles[0].tex);
			glColorPointer(3, GL_FLOAT, sizeof(POLY_NODE), m_Frames[i].triangles[0].fcol);
			glDrawArrays(GL_TRIANGLES, 0, m_Frames[i].triangles.Size());
			glEndList();
		}
	#elif defined(TRENKI3D)
		// Make a display list for TinyGL
		m_glListIDs.SetSize(m_Frames.Size());
		int liststart = tglGenLists(m_Frames.Size());
		struct VEC{
		float xyz[3], n[3], uv[2];
		unsigned char rgba[4];
		};
		
		VEC* pV = new VEC[m_Frames[0].triangles.Size()];
		for(i=0; i<m_Frames.Size(); ++i)
		{
			m_glListIDs[i] = liststart + i;
			int useRGBA = 0;
			for(int it=0; it<(int)m_Frames[i].triangles.Size(); ++it)
			{
				pV[it].xyz[0] = m_Frames[i].triangles[it].pos[0];
				pV[it].xyz[1] = m_Frames[i].triangles[it].pos[1];
				pV[it].xyz[2] = m_Frames[i].triangles[it].pos[2];
				pV[it].n[0] = m_Frames[i].triangles[it].normal[0];
				pV[it].n[1] = m_Frames[i].triangles[it].normal[1];
				pV[it].n[2] = m_Frames[i].triangles[it].normal[2];
				pV[it].uv[0] = m_Frames[i].triangles[it].tex[0];
				pV[it].uv[1] = m_Frames[i].triangles[it].tex[1];

				pV[it].rgba[0] = (unsigned char)(m_Frames[i].triangles[it].dwcol>>16) & 0xff;
				pV[it].rgba[1] = (unsigned char)(m_Frames[i].triangles[it].dwcol>> 8) & 0xff;
				pV[it].rgba[2] = (unsigned char)(m_Frames[i].triangles[it].dwcol    ) & 0xff;
				pV[it].rgba[3] = 255;
				if(pV[it].rgba[0] != 255 || pV[it].rgba[1] != 255 || pV[it].rgba[2] != 255)
					useRGBA = 1;
			}

			tglNewList(m_glListIDs[i]);
			tglDrawArraysXYZ_UV_RGBA(GL_TRIANGLES, m_Frames[i].triangles.Size(), &pV[0].xyz[0], 1,1,1, useRGBA, 1);
			tglEndList();
		}
		delete[] pV;
	#endif
}

Queue<unsigned char> gCelShadeDate;
// --------------------------------------------------------------- //
// Set Cel Shading Policy
// --------------------------------------------------------------- //
void Rainbows3D::SetCelShades(unsigned char* pScales, int nscales)
{
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
	if(nscales<=0)
	{
		gCelShadeDate.SetSize(1);
		gCelShadeDate[0]=255;
		return;
	}
	gCelShadeDate.SetSize(nscales);
int i;
	for(i=0; i<nscales; ++i)
	{
		gCelShadeDate[i] = *pScales++;
	}
	for(i=0; i<GLR_MAX_OBJECTS; ++i)
	{
		for(int j=0; j<(int)m_Objects[i].m_objects.Size(); ++j)
		{
			if(m_Objects[i].m_objects[j].obj.CelShadeTexture)
			{
				glDeleteTextures(1, &m_Objects[i].m_objects[j].obj.CelShadeTexture);
				m_Objects[i].m_objects[j].obj.CelShadeTexture=0;
			}
		}
	}
#endif
}

// --------------------------------------------------------------- //
// Make Cel-Shading Texture (for coloured objects and gray-scaling)
// --------------------------------------------------------------- //
void MeshObjectX::MakeCelShadeTexutre(bool bgra_ext)
{
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
	if(CelShadeTexture) return; // See also Clear() - I assume this is already done, here

	int sy = m_ColTable.Size(), sx = 256;
	// Create 2^n size
	for (int exp=1; exp<32768; exp *= 2) if (exp >= sy){sy = exp; break;}
	CelShadeColorScale = 1.0f/ (float)sy;

    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
	#if defined(HAVE_OPENGL)
    glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
    glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
    glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	#endif
	// Generate & Bind into Memory
	glGenTextures(1, &CelShadeTexture);
	glBindTexture(GL_TEXTURE_2D, CelShadeTexture);
	// Set Texture mapping parameters
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); // GL_NEAREST); // too hard edges
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); // GL_NEAREST);
	if(gCelShadeDate.Size()==0)
	{
		gCelShadeDate.SetSize(16);
		gCelShadeDate[ 0]=100;
		gCelShadeDate[ 1]=160;
		gCelShadeDate[ 2]=160;
		gCelShadeDate[ 3]=160;
		gCelShadeDate[ 4]=160;
		gCelShadeDate[ 5]=160;
		gCelShadeDate[ 6]=160;
		gCelShadeDate[ 7]=207;
		gCelShadeDate[ 8]=255;
		gCelShadeDate[ 9]=255;
		gCelShadeDate[10]=255;
		gCelShadeDate[11]=255;
		gCelShadeDate[12]=255;
		gCelShadeDate[13]=255;
		gCelShadeDate[14]=255;
		gCelShadeDate[15]=255;
	}

	sx=64;
	while(sx < 4*gCelShadeDate.Size() && sx < 512)
	{
		sx*=2;
	}

	unsigned char* data = new unsigned char[sx*sy*4];
//	unsigned char CelShadeDate[16] = {0, 0, 140, 160, 160, 160, 160, 180, 255, 255, 255, 255, 255, 255, 255, 255};


	for (int y=0; y<sy; ++y)
	{
		unsigned long r,g,b;
		unsigned long col = *m_ColTable.GetAt(y<m_ColTable.Size() ? y : 0);
		for(int x=0; x<sx; ++x)
		{
			if(col < COL_NEW_ANY)
			{
				r = (unsigned char) col;
				g = (unsigned char) (((unsigned short) col) >> 8);
				b = (unsigned char) (col >> 16);
			}
			else // New group - use white color
			{
				r=g=b=255;
			}
			int ix = (x * gCelShadeDate.Size()) / sx;

			r*=gCelShadeDate[ix];
			g*=gCelShadeDate[ix];
			b*=gCelShadeDate[ix];
			data[(x + y*sx)*4 + 0]=(unsigned char)(r/255);
			data[(x + y*sx)*4 + 1]=(unsigned char)(g/255);
			data[(x + y*sx)*4 + 2]=(unsigned char)(b/255);
			data[(x + y*sx)*4 + 3]=255;
		}
	}
	
	if(bgra_ext)
	{
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, sx, sy, 0, GL_BGRA_EXT, GL_UNSIGNED_BYTE, data);
	}
	else
	{
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, sx, sy, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
	}
	#if defined(HAVE_OPENGL)
	#elif defined(HAVE_OPENGL_ES)
	#endif
	delete [] data;
#endif
}

// --------------------------------------------------- //
// Calculate max radius for sphere - for quick collision detection
// !!! Requires XNODES to be stored !!!
// relative to center_of_mass
// --------------------------------------------------- //
void MeshObjectX::CalcMaxRadius()
{
	int i;
	for (int j=0; j<m_Frames.Size(); j++)
	{
		// calc size of max radius + set for frame
		float maxlen=0.0f;
		float len;
		float tmp[3];
		POLY_FRAME* pFrm = m_Frames.GetAt(j);
		pFrm->center_of_mass[0]=0; pFrm->center_of_mass[1]=0; pFrm->center_of_mass[2]=0;
		for (i=0; i<pFrm->triangles.Size(); i++)
		{
			gAddVertex(pFrm->center_of_mass, pFrm->triangles.GetAt(i)->pos);
		}
		gMultVertex(pFrm->center_of_mass, 1.0f / (float)pFrm->triangles.Size());
		for (i=0; i<pFrm->triangles.Size(); i++)
		{
			gDiffVertex(tmp, pFrm->center_of_mass, pFrm->triangles.GetAt(i)->pos);
			len=::gGetLength(tmp);
			if (len>maxlen)
				maxlen=len;
		}
		pFrm->max_radius=(float)fabs(maxlen);
	}
	
}

// --------------------------------------------------- //
// Create a list of edges and their neighbours
// for faster stencil shadow volume calculation
// --------------------------------------------------- //
void MeshObjectX::CreateEdges()
{
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
	if (m_Edges.Size()) return; // already got these
	if (m_Frames.Size()==0) return;
	
POLY_FRAME* pFr = &m_Frames[0];

	int num_tri = (int)pFr->triangles.Size()/3;
	m_Edges.SetSize(num_tri*3);
	
	//set the neighbour indices to be -1
	int i;
	for(i=0; i<num_tri*3; ++i)
	{
		m_Edges[i].neighbourIndex=-1;
	}
	
int* indices=new int[num_tri*3];
	// build an index list - we have to uncompress each node and
	// compare it with the vertex of the triangle!
	for(i=0; i<num_tri*3; ++i)
	{
		indices[i] = i;
		float* pPos = pFr->triangles[i].pos;
		for(int j=0; j<i; ++j)
		{
			float* pPosj = pFr->triangles[j].pos;
			if(	fabs(pPosj[0] - pPos[0])<1e-5 &&
				fabs(pPosj[1] - pPos[1])<1e-5 &&
				fabs(pPosj[2] - pPos[2])<1e-5  )
			{
				indices[i] = j;
				break;
			}
		}
	}

	//loop through triangles
	for(i=0; i<num_tri-1; ++i)
	{
		//loop through edges on the first triangle
		for(int edgeI=0; edgeI<3; ++edgeI)
		{
			//continue if this edge already has a neighbour set
			if(m_Edges[i*3+edgeI].neighbourIndex!=-1)
				continue;

			//loop through triangles with greater indices than this one
			for(int j=i+1; j<num_tri; ++j)
			{
				//loop through edges on triangle j
				for(int edgeJ=0; edgeJ<3; ++edgeJ)
				{
					//get the vertex indices on each edge
					int edgeI1=indices[i*3+edgeI];
					int edgeI2=indices[i*3+(edgeI+1)%3];
					int edgeJ1=indices[j*3+edgeJ];
					int edgeJ2=indices[j*3+(edgeJ+1)%3];

					//if these are the same (possibly reversed order), these faces are neighbours
					if(		(edgeI1==edgeJ1 && edgeI2==edgeJ2)
						||	(edgeI1==edgeJ2 && edgeI2==edgeJ1))
					{
						m_Edges[i*3+edgeI].neighbourIndex=j;
						m_Edges[j*3+edgeJ].neighbourIndex=i;
					}
				}
			}
		}
	}
	delete[] indices;
#endif
}


void MeshObjectX::Scale(float sx,float sy, float sz)
{
	int i;
	for (int j=0; j<m_Frames.Size(); ++j)
	{
		#define SCV(v) v[0]*=sx; v[1]*=sy; v[2]*=sz;
		#define SCNV(v) v[0]*=uni[0]; v[1]*=uni[1]; v[2]*=uni[2];
		POLY_FRAME* pFrm = m_Frames.GetAt(j);
		for (i=0; i<pFrm->triangles.Size(); ++i)
		{
			POLY_NODE *p = pFrm->triangles.GetAt(i);
			SCV(p->pos);
			SCV(p->normal);
			SCV(p->binormal);
			SCV(p->tangent);
		}
		
		float uni[3];
		gSetVertex(uni, sx,sy,sz);
		gUnify(uni);

		for(i=0; i<pFrm->nodes.Size(); ++i)
		{
			POLY_NODE *p = pFrm->nodes.GetAt(i);
			SCV(p->pos);
			SCNV(p->normal);
			SCNV(p->binormal);
			SCNV(p->tangent);
		}
		#undef SCV
		#undef SCNV
	}
	CalcMaxRadius();
}

// --------------------------------------------------- //
// calculate silhouette edges for stencil shadows
// --------------------------------------------------- //
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
void MeshObjectX::CalculateSilhouetteEdges(const POLY_NODE* pNodes, float* lightPosition)
{
	//Calculate which faces face the light
unsigned int numTriangles = m_Frames[0].triangles.Size() / 3;
	m_isTriFacingLight.SetSize(numTriangles);
unsigned int i;
	for(i=0; i<numTriangles; i++)
	{
		/*
		if(planeEquations[i].ClassifyPoint(lightPosition)==POINT_IN_FRONT_OF_PLANE)
			isFacingLight[i]=true;
		else
			isFacingLight[i]=false;
		*/
		// i take just the 3 normals of the triangle
		// and see if it's lit
		
		float dir[3], dir2[3], dir3[3];
		gDiffVertex(dir3, pNodes[i*3  ].pos, pNodes[i*3+1].pos);
		gDiffVertex(dir2, pNodes[i*3+2].pos, pNodes[i*3+1].pos);
		gCross(dir3, dir2, dir); // normal _direction_ to triangle
		
		// we need the _direction_ triangle->light!
		gCopyVertex(dir2, pNodes[i*3  ].pos);
		gAddVertex(dir2, pNodes[i*3+1].pos);
		gAddVertex(dir2, pNodes[i*3+2].pos);
		gMultVertex(dir2, -1.0f/3.0f);
		gAddVertex(dir2, lightPosition);

		if(gDot(dir, dir2) >= 0)
			m_isTriFacingLight[i]=true;
		else
			m_isTriFacingLight[i]=false;
	}
	//loop through edges
	for(i=0; i<numTriangles*3; ++i)
	{
		// this face is facing the light
		// if the neighbouring face is not facing the light, or there is no neighbouring face,
		// then this is a silhouette edge
		if(	m_isTriFacingLight[i/3] &&
			(m_Edges[i].neighbourIndex==-1 || !m_isTriFacingLight[m_Edges[i].neighbourIndex]))
		{
			m_Edges[i].isSilhouetteEdge=true;
			continue;
		}
		// if this face is not facing the light, not a silhouette edge
		m_Edges[i].isSilhouetteEdge=false;
	}
}
#endif


// --------------------------------------------------- //
// Draw the shadow volume object
// --------------------------------------------------- //
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
void MeshObjectX::DrawInfiniteShadowVolume(const POLY_NODE* pNodes, float* lightPosition, bool drawCaps)
{
// lightPosition -> mult with inverted modelview
unsigned int numTriangles = m_Frames[0].triangles.Size() / 3;

	glbDisable(GL_CULL_FACE);
	glDisable(GL_TEXTURE_2D);
	glbDisable(GL_LIGHTING);
	glDepthFunc(GL_LESS); // otherwise shadow on top of flat surfaces
	
	for(int pass=0; pass<2; ++pass)
	{
		// zpass
		switch (pass)
		{
		case 0:
			glbEnable(GL_CULL_FACE);
			glStencilFunc(GL_ALWAYS, 0, ~0);
			glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
			glCullFace(GL_BACK);
			break;
		case 1:
			//Decrement stencil buffer for back face depth pass
			glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
			glCullFace(GL_FRONT);
			break;
		}

		glBegin(GL_QUADS);
		{
			for(unsigned int i=0; i<numTriangles; ++i)
			{
				//if this face does not face the light, continue
				if(!m_isTriFacingLight[i])
					continue;

				//Loop through edges on this face
				for(int j=0; j<3; ++j)
				{
					//Draw the shadow volume "edge" if this is a silhouette edge
					if(m_Edges[i*3+j].isSilhouetteEdge)
					{
						// float* vertex1=vertices[indices[i*3+j]].position;
						// float* vertex2=vertices[indices[i*3+(j+1)%3]].position;

						const float* vertex1 = pNodes[i*3+j].pos;
						const float* vertex2 = pNodes[i*3+(j+1)%3].pos;

						glVertex3fv(vertex2);
						glVertex3fv(vertex1);
						glVertex4f(	vertex1[0]-lightPosition[0],
									vertex1[1]-lightPosition[1],
									vertex1[2]-lightPosition[2],0.0f);
						glVertex4f(	vertex2[0]-lightPosition[0],
									vertex2[1]-lightPosition[1],
									vertex2[2]-lightPosition[2],0.0f);
					}
				}
			}
		}
		glEnd();

		//Draw caps if required
		if(drawCaps)
		{
			glBegin(GL_TRIANGLES);
			{
				for(unsigned int i=0; i<numTriangles; ++i)
				{
					for(int j=0; j<3; ++j)
					{
						// float* vertex=vertices[indices[i*3+j]].position;
						const float* vertex = pNodes[i*3+j].pos;
						
						if(m_isTriFacingLight[i])
						{
							glVertex3fv(vertex);
						}
						else
						{
							glVertex4f(	vertex[0]-lightPosition[0],
										vertex[1]-lightPosition[1],
										vertex[2]-lightPosition[2], 0.0f);
						}
					}
				}
			}
			glEnd();
		}
	}
	// glDepthFunc(GL_LESS);
}
#endif




void MeshObjectX::Clear(bool free_mem)
{
	MeshObject::Clear(free_mem);

	#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
		m_Edges.Clear(free_mem);
		m_isTriFacingLight.Clear(free_mem);

		m_Edges.SetAllocSteps(64,64);
		m_isTriFacingLight.SetAllocSteps(64,64);
	#endif

	#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
		if (m_VBO_id && glBindBufferArb!=NULL)
		{
			glDeleteBuffersArb(1, &m_VBO_id);
		}
		if (m_VBO_id_morphing && glBindBufferArb!=NULL)
		{
			glDeleteBuffersArb(1, &m_VBO_id_morphing);
		}
	//	if (glCallList) glDeleteLists(glCallList, 1); glCallList=0;

		if(CelShadeTexture) // See also MakeCelShadeTexture()
		{
			glDeleteTextures(1, &CelShadeTexture);
			CelShadeTexture=0L;
		}

		m_VBO_id = 0;
		m_VBO_id_morphing = 0;
	#endif
}



//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Rainbows3D::Rainbows3D(OpenGLRainbows* pR)
{
	m_iCurObjFrame = 0;
	m_LightsOn=0;
	m_AutoNormalMode=2;
	m_pRBOW = pR;
	m_nTransformations=0;
	m_shadowmode=false;

	m_Objects.SetSize(GLR_MAX_OBJECTS);
	#if defined(HAVE_OPENGL)
	memset(&m_Shaders, 0, sizeof(m_Shaders));
	#endif
	pR->m_pRbow3D = this;
}

Rainbows3D::~Rainbows3D()
{
	m_Objects.Clear(true);
	#if defined(HAVE_OPENGL)
	for(int i=0; i<MAX_SHADERS; ++i)
		FreeShader(i);
	#endif
}

void Rainbows3D::ClearZBuffer()
{
	#ifdef TRENKI3D
	Screen.ClearZBuffer();
	#else
	DoShadows(false);
	glClear(GL_DEPTH_BUFFER_BIT);
	#endif
}


// --------------------------------------------------------------- //
// Start / Draw Shadows
// --------------------------------------------------------------- //
void Rainbows3D::DoShadows(bool mode)
{
	if (mode == m_shadowmode) return; // nothing to do
	m_shadowmode = mode;

#if !defined(TINYGL) && !defined(TRENKI3D)
	if(mode) // start drawing in stencil mode
	{
		glShadeModel(GL_FLAT);
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE); // Don't draw r,g,b,a - only z

		glbEnable(GL_STENCIL_TEST);
		glDepthMask(GL_FALSE); // Do not write into depth buffer
		glStencilFunc(GL_ALWAYS, 0, 0);
		// Draw the whole model front+backfaces, always inverting the stencil buffer
		// If only one face has been drawn, the buffer == -1, otherwise == 0
		// Where 0, it's infront/behind the shadow
//		glbDisable(GL_CULL_FACE);
		glStencilOp(GL_KEEP, GL_KEEP, GL_INVERT); // If visible, add one to the stencil
//		glCallList(SHADOWVOL);
//		glbEnable(GL_CULL_FACE);
	}
	else // finally draw the shadow
	{
		// Now turn everything back to normal
		glCullFace(GL_BACK);
		glDepthMask(GL_TRUE);
		glDepthFunc(GL_LEQUAL);
		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
		// Wherever stencil == -1, draw the shadow

//		glbEnable(GL_BLEND);
//		glBlendFunc(GL_SRC_COLOR, GL_DST_COLOR);

		
		// We're doing a 2D viewport here!
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
			m_pRBOW->Make2DMode();
			m_pRBOW->DoAlpha(-.5);

			glbEnable(GL_STENCIL_TEST);
			// glStencilFunc(GL_EQUAL, 1,1); glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP); // one pass method
			glStencilFunc(GL_NOTEQUAL, 0, ~0); glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP); // two pass method
			m_pRBOW->DrawRect(0, 0, m_pRBOW->m_WND.viewport.width, m_pRBOW->m_WND.viewport.height, m_BumpLight.dwcol);
			m_pRBOW->DoAlpha(0);
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		
		// glCallList(SHADOWVOL); // render the shadow volume(s) bounding surfaces
//		glbDisable(GL_BLEND);
		
		// Reset to normal
		glShadeModel(GL_SMOOTH);
		
		this->SetDepthTesting(true);
		
		glbDisable(GL_STENCIL_TEST);
		glClear(GL_STENCIL_BUFFER_BIT);

		glbEnable(GL_CULL_FACE);
		glCullFace(GL_BACK);

		// restore culling as set before shadows
		CullMode(m_CullMode);
	}
#endif
}



// ------------------------------------------------------------- //
// Make a Fog
// ------------------------------------------------------------- //
void Rainbows3D::Fog(unsigned long col, int exp_mode, float p1, float p2)
{
#if !defined(TINYGL) && !defined(TRENKI3D)
	if(col==-1)
	{
		glbDisable(GL_FOG);
	}
	else
	{
		glbEnable(GL_FOG);
		if(exp_mode)
		{
			// this fog is very silly to use
			glFogf(GL_FOG_MODE, GL_EXP); glFogf(GL_FOG_DENSITY, p1);
		}
		else
		{
			// nice to use - but they claim it's ugly...
			glFogf(GL_FOG_MODE, GL_LINEAR);
			// 0,0 -> use znear, zfar
			if(p1==0.0f && p2==0.0f)
			{
				p1 = m_Camera.znear;
				p2 = m_Camera.zfar;
			}
			glFogf(GL_FOG_START, p1);
			glFogf(GL_FOG_END, p2);
		}

		float fcl[4] = {GetRValue(col)/255.0f, GetGValue(col)/255.0f, GetBValue(col)/255.0f, 0.0f};
		glFogfv(GL_FOG_COLOR, fcl);
	}
#endif
}


//------------------------------------------------------------------//
// Build the current matrix in the stack
//------------------------------------------------------------------//
void Rainbows3D::BuildMatrix()
{
TRANSFORMATION& tr = m_Transformations[m_nTransformations];
	if(tr.m_IsGoodMatrix)
		return;

	// for each tr in m_Transformations...
	float* pM = tr.m_Matrix;
	float m[16];

	// Multiply
	for (int i=0; i<tr.m_nMult; ++i)
	{
		// glMultMatrixf(tr.m_Mults[i]);
		matmul4(pM, pM, tr.m_Mults[i]);
	}

	// Move
	// glTranslatef(tr.m_Translate[0], tr.m_Translate[1], tr.m_Translate[2]);
	float* pT = tr.m_Translate;
	if(pT[0] || pT[1] || pT[2])
	{
		gIdentity(m);
		m[12]=pT[0];
		m[13]=pT[1];
		m[14]=pT[2];
		matmul4(pM, pM, m);
	}

	// Rotate
	for (int i=tr.m_nRotate-1; i>=0; i--)
	{
		// glRotatef(tr.m_Rotate[i][0], tr.m_Rotate[i][1], tr.m_Rotate[i][2], tr.m_Rotate[i][3]);
		float* pR = tr.m_Rotate[i];
		matrix_rotate(pM, pR[0], pR[1], pR[2], pR[3]);
	}
	
	// Zoom
	// glScalef(tr.m_Scale[0], tr.m_Scale[1], tr.m_Scale[2]);
	if(tr.m_Scale[0]!=1.0f || tr.m_Scale[1]!=1.0f || tr.m_Scale[2]!=1.0)
	{
		gIdentity(m);
		m[0]=tr.m_Scale[0]; m[5]=tr.m_Scale[1]; m[10]=tr.m_Scale[2];
		matmul4(pM, pM, m);
	}
	tr.m_IsGoodMatrix=1;
}

//------------------------------------------------------------------//
// Rotate and Scale current Matrix
//------------------------------------------------------------------//
void Rainbows3D::TransformMatrix()
{
	BuildMatrix();
	#ifdef TRENKI3D
		tglMultMatrixf(m_Transformations[m_nTransformations].m_Matrix);
	#else
		glMultMatrixf(m_Transformations[m_nTransformations].m_Matrix);
	#endif
}


//------------------------------------------------------------------//
// Specify texture offset
//------------------------------------------------------------------//
void Rainbows3D::SetTextureOffset(float du, float dv)
{
	m_TexOffset[0] = du; m_TexOffset[1] = dv;
}


//------------------------------------------------------------------//
// Render a model
//------------------------------------------------------------------//
void Rainbows3D::RenderAnim(int num, int from, int to, float percent, bool fullAnim)
{
	X_WORLD& wld = m_Objects[num];
	if(wld.is_multiworld)
	{
		for(int i=0; i<wld.m_objects.Size(); ++i)
		{
			X_WORLD::DATA& dat = wld.m_objects[i];
			if(wld.m_textures.Size())
			{
				#ifdef TINYGL
				RenderMD2_TinyGL(dat.obj, from, to, percent, fullAnim);
				#elif defined(TRENKI3D)
				RenderMD2_Trenki(dat.obj, from, to, percent, fullAnim);
				#else
				m_pRBOW->SetTexture3D(dat.gl_tex, dat.gl_tex2, 0, &wld.m_textures[0]);
				RenderMD2(dat.obj, from, to, percent, fullAnim, &wld.m_textures[0]);
				#endif
			}
			else
			{
				m_pRBOW->SetTexture3D(-1, -1, 0);
				#ifdef TINYGL
				RenderMD2_TinyGL(dat.obj, from, to, percent, fullAnim);
				#elif defined(TRENKI3D)
				RenderMD2_Trenki(dat.obj, from, to, percent, fullAnim);
				#else
				RenderMD2(dat.obj, from, to, percent, fullAnim, &wld.m_textures[0]);
				#endif
			}
		}
	}
	else if(wld.m_objects.Size())
	{
		#ifdef TINYGL
			RenderMD2_TinyGL(wld.m_objects[0].obj, from, to, percent, fullAnim);
		#elif defined(TRENKI3D)
			RenderMD2_Trenki(wld.m_objects[0].obj, from, to, percent, fullAnim);
		#else
			RenderMD2(wld.m_objects[0].obj, from, to, percent, fullAnim);
		#endif
	}
}



#ifdef _DEBUG
	// #define _NO_VBO
#endif

//  #define _NO_VBO // Vertex arrays - no VBO

#ifdef TRENKI3D

void Rainbows3D::RenderMD2_Trenki(MeshObjectX& the_obj, int from, int to, float percent, bool fullAnim)
{
	if(!the_obj.m_Frames.Size()) return;
	tglPushMatrix();
		TransformMatrix();
		tglCallLists(the_obj.m_glListIDs[from],1);
		Screen.SetZBufferUsed(1);
	tglPopMatrix();
}

#else // TRENKI3D

// Render a Model
// Take care of texture state yourself before this call!
void Rainbows3D::RenderMD2_TinyGL(MeshObjectX& the_obj, int from, int to, float percent, bool fullAnim, GLuint* pTextures)
{
	if(!the_obj.m_Frames.Size()) return;
#ifdef TINYGL
	from = from + (int)((to-from)*percent);
	if (from<0) from=0;
	if (from>=the_obj.NumFrames()) from=the_obj.NumFrames()-1;
	
	glEnableClientState( GL_VERTEX_ARRAY );

	if(m_pRBOW->m_Texture3D[0]>=0)
	{
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glbEnable(GL_TEXTURE_2D);
		// int id = m_pRBOW->m_Texture[m_pRBOW->m_Texture3D[1]].mipstate[m_pRBOW->m_Texture3D[2]];
		if(pTextures)
			glBindTexture(GL_TEXTURE_2D, pTextures[m_pRBOW->m_Texture3D[0]]);
		else
			glBindTexture(GL_TEXTURE_2D, m_pRBOW->m_Texture[m_pRBOW->m_Texture3D[0]].mipstate[m_pRBOW->m_Texture3D[2]]);
	}
	if (the_obj.m_NeedColoring)
	{
 		glEnableClientState(GL_COLOR_ARRAY);
	}

	//glbEnable(GL_COLOR_MATERIAL);
	glPushMatrix();
		TransformMatrix();
		glCallList(the_obj.m_glListIDs[from]);
	glPopMatrix();

	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState( GL_VERTEX_ARRAY );
	glbColor4f(1.0f, 1.0f, 1.0f, m_pRBOW->m_falpha);
#endif
}


//------------------------------------------------------------------//
// Render a MD2 Object (with Dot3 Bump Mapping if required)
//------------------------------------------------------------------//
void Rainbows3D::RenderMD2(MeshObjectX& the_obj, int from, int to, float percent, bool fullAnim, GLuint* pTextures)
{
#ifndef TINYGL
	if (the_obj.NumNodes(0)==0 || the_obj.NumFrames()==0) return;
	POLY_FRAME				*cFrame, *nFrame;
	POLY_NODE				*cNode, *nNode, *tNode, *tNodes_all;
	int						index;
	float					invM[16], lv0[3];
	unsigned int			*pVBO_id;

	pVBO_id = &the_obj.m_VBO_id_morphing;

	// Find out the frames to be interpolated
	if (fullAnim)
	{
		if(percent>1.0f || percent<0.0f) percent=fabsf(fmodf(percent, 1.0f));
		float	pc;
		pc=(float)from+(float)(to-from)*percent;
		index=(int)pc;
		if (to>from)
		{
			percent=1.0f - (pc-(float)index);
			from=index;
			to=index+1;
		}
		else if(to<from)
		{
			percent=pc-(float)index;
			from=index+1;
			to=index;
		}
		else // from and to are the same - no interpolation
		{
			percent=0.0f;
		}
	}
	else
	{
		// this is soo stupid!! #NEED
		int sw=from; from=to; to=sw;
	}

	// For safety reasons check range
	if (from<0) from=0;
	if (to<0) to=0;
	if (to>=the_obj.NumFrames()) to=the_obj.NumFrames()-1;
	if (from>=the_obj.NumFrames()) from=the_obj.NumFrames()-1;

	//Get the current frame and gl command information
	cFrame=the_obj.m_Frames.GetAt(from);
	nFrame=the_obj.m_Frames.GetAt(to);


	// Very quick object in view test to safe a lot of time
/*
	if(m_nTransformations==0 && m_Transformations[0].m_nMult==0)
	{
		gCopyVertex(lv0, m_Transformations[0].m_Translate);
		gAddVertex(lv0, cFrame->center_of_mass);

		if(!gSphereInCone(	m_Camera.eye, m_Camera.lookat, m_Camera.tan_fov,
							lv0,
							cFrame->max_radius * max(m_Transformations[0].m_Scale[0], max(m_Transformations[0].m_Scale[1], m_Transformations[0].m_Scale[2]))))
			return;
	}
*/




	BuildMatrix();
	gCopyVertex(lv0,cFrame->center_of_mass);
	gTransformVector(m_Transformations[m_nTransformations].m_Matrix, lv0);
	float sx,sy,sz;
	GetMatrixScalings(sx,sy,sz);

	if(!SphereInFrustum(lv0[0], lv0[1], lv0[2],
						cFrame->max_radius*max(sx,max(sy,sz)) ))
	{
		// never clip the shadow volume
		if (! (m_BumpLight.active && m_BumpLight.type == 3))
			return;
	}

	// Don't remove this!
	// if(the_obj.m_NeedColoring || m_pRBOW->m_falpha!=1.0f)
	{
		// colour value might be wrong from last rendered model.
		glbColor4f(1.0f, 1.0f, 1.0f, m_pRBOW->m_falpha);
		glEnable(GL_COLOR_MATERIAL);
	}

	// Create Normals and Bump-Vertices if not done already
	if(m_BumpLight.active) // now we need both of them
	{
		switch(m_BumpLight.type)
		{
		case 3:
			the_obj.AutoNormalsAfterCompression(true, false, false);
			the_obj.CreateEdges();
			break;
		case 2: // Cel Shading
			the_obj.AutoNormalsAfterCompression(true, false, false);
			the_obj.MakeCelShadeTexutre(m_pRBOW->m_bSupportBGRA_ext);
			break;
		case 1: // Bump mapping
			the_obj.AutoNormalsAfterCompression(true, m_AutoNormalMode==2 ? true : false, true);
			break;
		}
	}
	// Not yet created triangles due to normal calculations?
	if (the_obj.m_tri_animation.Size()==0)
		the_obj.PrepareTriangles();
	
	// Move, Rotate & Scale
	tNodes_all = the_obj.m_tri_animation.GetAt(0);
	// Modelview matrix and inverted light matrix
	// -> modelspace light position
	if (m_BumpLight.active)
	{
		gCopyMatrix(invM, m_Transformations[m_nTransformations].m_Matrix);
		float inv2[16];
		gCopyMatrix(inv2, invM);

		// gInvertMatrix4x4(invM);
		gInvertMatrix4x4_full(invM); // due to anisotrope scaling
		
		gCopyVertex(lv0, m_BumpLight.pos);
		// float ll = gGetLength(lv0);
		gTransformVector(invM, lv0);
		// gMultVertex(lv0, ll); // bad position
	}
	// No bumping, no animation, no spheremapping? Try pre-rendered Objects
	else if (percent==0.0f && m_TexOffset[0]==0.0f && m_TexOffset[1]==0.0f)
	{
		tNodes_all = cFrame->triangles.GetAt(0);
		if (from == 0) // First frame? Create/Render a AGP memory VBO!
		{
			pVBO_id = & the_obj.m_VBO_id;
		}
		goto just_draw;
	}

	// Prepare light normals + keyframe processing
	float lv[3], tc[3];

	cNode = cFrame->triangles.GetAt(0); // a bit quicker
	tNode = the_obj.m_tri_animation.GetAt(0);
	for (index=0; index<(int)the_obj.m_tri_animation.Size(); index++)
	{
		// cNode = cFrame->triangles.GetAt(index);
		// tNode = the_obj.m_tri_animation.GetAt(index);

		// Still frame or interpolation?
		// Really interpolate
		if (percent==0.0f)
		{
			*tNode = *cNode; // Copy static vertex to tri_animation
			nNode = cNode;
		}
		else
		{
			nNode = nFrame->triangles.GetAt(index);
			tNode->pos[0]=	(nNode->pos[0]+
								(cNode->pos[0]-nNode->pos[0])*percent);

			tNode->pos[1]=	(nNode->pos[1]+  // -2
								(cNode->pos[1]-nNode->pos[1])*percent);

			tNode->pos[2]=	(nNode->pos[2]+ // 1
								(cNode->pos[2]-nNode->pos[2])*percent);
			// Interpolate Normals
			tNode->normal[0]=	(nNode->normal[0]+
								(cNode->normal[0]-nNode->normal[0])*percent);

			tNode->normal[1]=	(nNode->normal[1]+ // -2
								(cNode->normal[1]-nNode->normal[1])*percent);

			tNode->normal[2]=	(nNode->normal[2]+ // 1
								(cNode->normal[2]-nNode->normal[2])*percent);
		}

		// Special light (Bump Mapping / Cel shading...)
		if (m_BumpLight.active && ((void*)glMultiTexCoord2fArb)) // Has Light and extension provided
		{
			// lv = lightpos - vertex_triangle_point
			gDiffVertex(lv, lv0, tNode->pos);
			gUnify(lv);
			switch(m_BumpLight.type)
			{
			case 1: // Dot 3 Bump Mapping
				gCopyVertex(tc, tNode->pos);
				lv[0]=(lv0[0]-tc[0]);
				lv[1]=(lv0[1]-tc[1]);
				lv[2]=(lv0[2]-tc[2]);
				gUnify(lv);
				// destryoing invM here -> need a temporary matrix
				// for each vertex
				invM[0]=nNode->tangent[0];
				invM[1]=nNode->binormal[0];
				invM[2]=tNode->normal[0];
				invM[3]=nNode->tangent[1];
				invM[4]=nNode->binormal[1];
				invM[5]=tNode->normal[1];
				invM[6]=nNode->tangent[2];
				invM[7]=nNode->binormal[2];
				invM[8]=tNode->normal[2];

				// This is not a color, it's the displacement direction!
				tNode->bump_light[0]=((lv[0]*invM[0])+(lv[1]*invM[3])+(lv[2]*invM[6]))*0.5f+0.5f;
				tNode->bump_light[1]=((lv[0]*invM[1])+(lv[1]*invM[4])+(lv[2]*invM[7]))*0.5f+0.5f;
				tNode->bump_light[2]=((lv[0]*invM[2])+(lv[1]*invM[5])+(lv[2]*invM[8]))*0.5f+0.5f;
				break;
			case 2: // Cel Shading
				// y=color, x=shading position
				tNode->binormal[1] = (float)((tNode->icol)+0.5f)*the_obj.CelShadeColorScale;
				
				gDiffVertex(tc, tNode->pos, tNode->normal);
				gUnify(tc);
									
				// u = light' o Normal; light' = light x matrix
				// tNode->binormal[0]=gDot(lv0, tNode->normal);
				tNode->binormal[0]=min(0.95f, max(0.05f, gDot(lv, tc)));
				// the texture at 1.0 has a bit of black from 0.0!?
				break;
			}
		}

		// Texture Offset
		tNode->tex[0]+=m_TexOffset[0]; tNode->tex[1]+=m_TexOffset[1];


		++cNode; // a big quicker - but does not work!
		++tNode;
	} // Next Vertex



	// draw shadow volume of animated mesh
	// then return
	if (m_BumpLight.active && m_BumpLight.type == 3)
	{
		#if defined(HAVE_OPENGL)
		glPushMatrix();
			TransformMatrix();
			the_obj.CalculateSilhouetteEdges(&cFrame->triangles[0], lv0);

			m_pRBOW->SetTexture3D(-1, -1, 0); // no texture for shadow
			//                               &cFrame->triangles[0]
			the_obj.DrawInfiniteShadowVolume(the_obj.m_tri_animation.GetAt(0), lv0, true);
		glPopMatrix();
		#endif
		return;
	}



just_draw:

#ifdef _NO_VBO
	glBindBufferArb = NULL;
#endif

#define BUFFER_OFFSET(i) (  ((void*)glBindBufferArb) ? /*NULL based*/ ((void*)((size_t)(i) - (size_t)(tNodes_all))) : /*real pointer*/ ((void*)(i))  )

	if ((void*)glBindBufferArb)
	{
		// Use the morphing object here
		int newsize = sizeof(POLY_NODE) * the_obj.m_tri_animation.Size();
		if (*pVBO_id==0)
		{
			glGenBuffersArb(1, pVBO_id);
			glBindBufferArb(GL_ARRAY_BUFFER_ARB, *pVBO_id);
			glBufferDataArb(GL_ARRAY_BUFFER_ARB, newsize,
							tNodes_all,
							(pVBO_id==&the_obj.m_VBO_id_morphing) ? GL_STREAM_DRAW_ARB : GL_STATIC_DRAW_ARB);
		}
		// already there, but dynamic
		else if(pVBO_id==&the_obj.m_VBO_id_morphing)
		{
			glBindBufferArb(GL_ARRAY_BUFFER_ARB, *pVBO_id);
			glBufferDataArb(GL_ARRAY_BUFFER_ARB, newsize, tNodes_all, (GLenum)GL_STREAM_DRAW_ARB);
		}
		else // Static Object, already created
		{
			glBindBufferArb( GL_ARRAY_BUFFER_ARB, *pVBO_id);
		}
	}

	glEnableClientState( GL_VERTEX_ARRAY );
	glVertexPointer(  3, GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(tNodes_all->pos));
	if (m_BumpLight.active && (void*)glClientActiveTextureArb) // bump mapping / cel shading
	{
		switch(m_BumpLight.type)
		{
		case 1: // bump mapping
			#if defined(HAVE_OPENGL)
				// Setup Blending Texture settings
				glClientActiveTextureArb(GL_TEXTURE1_ARB);
				glActiveTextureArb(GL_TEXTURE1_ARB);
					glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
					glEnableClientState(GL_TEXTURE_COORD_ARRAY);
					glTexCoordPointer(2, GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(tNodes_all->tex));

				glClientActiveTextureArb(GL_TEXTURE0_ARB);
				glActiveTextureArb(GL_TEXTURE0_ARB);
					glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
					glTexEnvf(GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT,  GL_DOT3_RGB_EXT);
					glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE0_RGB_EXT,  GL_PRIMARY_COLOR_EXT);
					glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND0_RGB_EXT, GL_SRC_COLOR);
					glTexEnvf(GL_TEXTURE_ENV, GL_SOURCE1_RGB_EXT,  GL_TEXTURE);
					glTexEnvf(GL_TEXTURE_ENV, GL_OPERAND1_RGB_EXT, GL_SRC_COLOR);
					glEnableClientState(GL_TEXTURE_COORD_ARRAY);
					glTexCoordPointer(2, GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(tNodes_all->tex));

				glEnableClientState(GL_COLOR_ARRAY);
					glColorPointer(3, GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(tNodes_all->bump_light));
			#endif
			break;
		case 2: // cel shading
			// Cel Shading Texture Unit
			glClientActiveTextureArb(GL_TEXTURE1_ARB);
			glActiveTextureArb(GL_TEXTURE1_ARB); // Textures
				glBindTexture(GL_TEXTURE_2D, the_obj.CelShadeTexture);
				glEnable(GL_TEXTURE_2D);
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(2, GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(tNodes_all->binormal));
				glEnable(GL_TEXTURE_2D);

			// restore default texture unit
			glClientActiveTextureArb(GL_TEXTURE0_ARB);
			glActiveTextureArb(GL_TEXTURE0_ARB); // Textures

			// Image Texture
			if(m_pRBOW->m_Texture3D[0]>=0)
			{
				glEnable(GL_TEXTURE_2D);
				// #NPBUFFER
				if(pTextures)
				{
					glBindTexture(GL_TEXTURE_2D, pTextures[m_pRBOW->m_Texture3D[0]]);
				}
				else
				{
					glBindTexture(GL_TEXTURE_2D, m_pRBOW->m_Texture[m_pRBOW->m_Texture3D[0]].mipstate[m_pRBOW->m_Texture3D[2]]);
				}
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(2, GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(tNodes_all->tex));
			}
			break;
		}
	}
	else // vanilla textured rendering
	{
		// has a texture set
		if(m_pRBOW->m_Texture3D[0]>=0)
		{
			// has 2nd texture unit set? - Multitexturing
			if(m_pRBOW->m_Texture3D[1]>=0 && (void*)glClientActiveTextureArb)
			{
				glClientActiveTextureArb(GL_TEXTURE1_ARB);
				glActiveTextureArb(GL_TEXTURE1_ARB);
				// #NPBUFFER
				if(pTextures)
				{
					glBindTexture(GL_TEXTURE_2D, pTextures[m_pRBOW->m_Texture3D[1]]);
				}
				else
				{
					glBindTexture(GL_TEXTURE_2D, m_pRBOW->m_Texture[m_pRBOW->m_Texture3D[1]].mipstate[m_pRBOW->m_Texture3D[2]]);
				}
				glEnableClientState(GL_TEXTURE_COORD_ARRAY);
				glTexCoordPointer(2, GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(tNodes_all->tex));			
			}

			// do texture unit 0
			if((void*)glClientActiveTextureArb)
			{
				glClientActiveTextureArb(GL_TEXTURE0_ARB); // tex coords
				glActiveTextureArb(GL_TEXTURE0_ARB);       // tex unit
			}
			// #NPBUFFER
			if(pTextures)
			{
				glBindTexture(GL_TEXTURE_2D, pTextures[m_pRBOW->m_Texture3D[0]]);
			}
			else
			{
				glBindTexture(GL_TEXTURE_2D, m_pRBOW->m_Texture[m_pRBOW->m_Texture3D[0]].mipstate[m_pRBOW->m_Texture3D[2]]);
			}
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(tNodes_all->tex));
		}
		if (the_obj.m_NeedColoring)
		{
	 		glEnableClientState(GL_COLOR_ARRAY);
 			glColorPointer(4, GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(((BYTE*)&tNodes_all->fcol )));
		}
	}

	//glbEnable(GL_COLOR_MATERIAL);
	{glPushMatrix();
		TransformMatrix();
		
		if(m_LightsOn || m_SphereMapping)
		{
			glEnableClientState(GL_NORMAL_ARRAY);
			glNormalPointer(GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(((BYTE*)&tNodes_all->normal)));
		}
		
		#if !defined(TINYGL) && !defined(TRENKI3D) && !defined(HAVE_OPENGL_ES)
		if(m_SphereMapping)
		{
			glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
			glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
			glbEnable(GL_TEXTURE_GEN_S); glbEnable(GL_TEXTURE_GEN_T);
		}
		#endif

		//
		// -= Render the Object itself =-
		// Draw All Of The Triangles At Once
		// 
		//GL_TRIANGLES
		glDrawArrays(GL_TRIANGLES, 0, the_obj.m_tri_animation.Size());

		#if !defined(TINYGL) && !defined(TRENKI3D) && !defined(HAVE_OPENGL_ES)
		if(m_SphereMapping)
		{
			glbDisable(GL_TEXTURE_GEN_S); glbDisable(GL_TEXTURE_GEN_T);
		}
		#endif

		// Disable Pointers
		glDisableClientState( GL_VERTEX_ARRAY );					// Disable Vertex Arrays
		glDisableClientState( GL_TEXTURE_COORD_ARRAY );				// Disable Texture Coord Arrays
		glDisableClientState( GL_COLOR_ARRAY  );					// Disable Color Arrays
		glDisableClientState( GL_NORMAL_ARRAY );
		if((void*)glBindBufferArb) // allow normal vertex arrays (POLYSPRITE)
			glBindBufferArb( GL_ARRAY_BUFFER_ARB, 0);

		///////////////////////////////////////////////////////////////
		// Cel Shading: Outlines
		///////////////////////////////////////////////////////////////
		#if !defined(TINYGL) && !defined(TRENKI3D)
		if(m_BumpLight.active && m_BumpLight.type==2 && m_CullMode!=0)
		{
			// blow up the vertices along their normals by given ratio
			
			// get the distance camera -> object
			// 'm' is inverse of camera
			gCopyVertex(lv0, m_Camera.eye);
			gTransformVector(invM, lv0);
			gSubVertex(lv0, cFrame->center_of_mass);

			// 0.003f = relative line width
			float scale = m_Camera.tan_fov * gGetLength(lv0) * 0.003f;


			#if 0 // 2018-11-08 GF: I found that X_SCALING 0.32 (Rock'N'Bolt) looked weirb, but w/o this code it's OK
			// the scaling is a problem, since it get's rotated
			// you have to scal*scal, since scaling is still active
			// at this point, so we have to "undo" it twice
			for(int i=0; i<=m_nTransformations; ++i)
			{
				// float matscale = m_Transformations[i].m_Scale[1]; // #NEED
				float fsx,fsy,fsz;
				if(GetMatrixScalings(fsx,fsy,fsz))
				{
					float matscale = (fsx+fsy+fsz)/3.0f;
				//	scale /= ( matscale ); // matscale *
				}
			}
			#endif

			if(scale>cFrame->max_radius/2.0f) goto skip_outlines; // Object is too small - would be a black dot
			if(scale<1E-7f) scale=1E-7f; // Limit for numerical problems otherwise

			this->CullMode(-m_CullMode); // reverse
			float norm[3];
			for (int n=0; n<the_obj.m_tri_animation.Size(); ++n)
			{
				gCopyVertex(norm, the_obj.m_tri_animation.GetAt(n)->normal);
				gMultVertex(norm, scale);
				gAddVertex(the_obj.m_tri_animation.GetAt(n)->pos, norm);
			}

			if ((void*)glBindBufferArb)
			{
				// Use the morphing object here
				pVBO_id = &the_obj.m_VBO_id_morphing;
				int newsize = sizeof(POLY_NODE) * the_obj.m_tri_animation.Size();
				if (*pVBO_id==0)
				{
					glGenBuffersArb(1, pVBO_id);
					glBindBufferArb(GL_ARRAY_BUFFER_ARB, *pVBO_id);
					glBufferDataArb(GL_ARRAY_BUFFER_ARB, newsize, &the_obj.m_tri_animation[0], GL_STREAM_DRAW_ARB);
				}
				else
				{
					// check data size in VBO is same as input array, if not return 0 and delete VBO
					int bufferSize = 0;
					glBindBufferArb(GL_ARRAY_BUFFER_ARB, *pVBO_id);
					glBufferDataArb(GL_ARRAY_BUFFER_ARB, newsize, &the_obj.m_tri_animation[0], (GLenum)GL_STREAM_DRAW_ARB);
				}
			}

// goto skip_outlines;
			glEnableClientState( GL_VERTEX_ARRAY );
			glVertexPointer(  3, GL_FLOAT, sizeof(POLY_NODE), BUFFER_OFFSET(the_obj.m_tri_animation.GetAt(0)->pos));

			
			if((void*)glActiveTextureArb)
			{
				glActiveTextureArb(GL_TEXTURE1_ARB); // Texture unit
				glDisable(GL_TEXTURE_2D);
				glActiveTextureArb(GL_TEXTURE0_ARB);
				glDisable(GL_TEXTURE_2D);
			}
			else
			{
				glDisable(GL_TEXTURE_2D);
			}
			glbDisable(GL_LIGHTING);
			glbEnable(GL_COLOR_MATERIAL);
			glbColor4f(0.0f,0.0f,0.0f,1.0f);
			glDepthFunc(GL_LEQUAL);
			
			// This looks much better, but is slow as hell!
			//#define SMOOTH_LINES
			#ifdef SMOOTH_LINES
				glLineWidth(5);
				glPolygonMode(GL_BACK, GL_LINE); // Backfacing polygons as lines
				glPolygonMode(GL_FRONT, GL_LINE); // Backfacing polygons as lines
			#endif
			glDrawArrays(GL_TRIANGLES, 0, the_obj.m_tri_animation.Size());	// Draw All Of The Triangles At Once
			#ifdef SMOOTH_LINES
				glLineWidth(1);
				glPolygonMode(GL_BACK, GL_FILL); // Backfacing polygons as lines
				glPolygonMode(GL_FRONT, GL_FILL); // Backfacing polygons as lines
			#endif

			glDepthFunc(GL_LESS);
			// Disable Pointers
			glDisableClientState( GL_VERTEX_ARRAY );								// Disable Vertex Arrays
			// no need to un-scale the vertices again. Vertices are only valid for this frame
			m_pRBOW->SetTexture3D(m_pRBOW->m_Texture3D[0], m_pRBOW->m_Texture3D[1], m_pRBOW->m_Texture3D[2]);
			this->CullMode(-m_CullMode); // re-reverse

			if((void*)glBindBufferArb) // allow normal vertex arrays (POLYSPRITE)
				glBindBufferArb( GL_ARRAY_BUFFER_ARB, 0);
		}
	skip_outlines:
		#endif

		// Reset blending modes for both textures
		if ((void*)glActiveTextureArb)
		{
			glActiveTextureArb(GL_TEXTURE1_ARB);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glActiveTextureArb(GL_TEXTURE0_ARB);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		}
	glPopMatrix();}

// Debug Normals
#if defined(_DEBUG) && !defined(TINYGL) && !defined(TRENKI3D)
 #define DEBUG_MODEL_NORMALS
	#ifdef DEBUG_MODEL_NORMALS
	glPushMatrix();
		TransformMatrix();
		glbDisable(GL_FOG);
		// glbEnable(GL_COLOR_MATERIAL);
		glDisable(GL_TEXTURE_2D);
		glBegin(GL_LINES);
		float sc=cFrame->max_radius/4.0f;
		for (int i=0; i<the_obj.m_tri_animation.Size(); i++)
		{
			float v[3];
			cNode = the_obj.m_tri_animation.GetAt(i);

			// Add normal Vertex - red
			glbColor4f(1,0,0,1);
			glVertex3fv(cNode->pos);
			gCopyVertex(v, cNode->normal); gMultVertex(v, sc); gAddVertex(v, cNode->pos);
			glVertex3fv(v);
			glbColor4f(1,1,1,1);
			glVertex3fv(v);
			gCopyVertex(v, cNode->normal); gMultVertex(v, sc*2); gAddVertex(v, cNode->pos);
			glVertex3fv(v);
			
/*			// Add tangent Vertex - green
			glbColor4f(0,1,0,1);
			glVertex3fv(cNode->pos);
			gCopyVertex(v, cNode->tangent); gMultVertex(v, sc); gAddVertex(v, cNode->pos);
			glVertex3fv(v);

			// Add binormal Vertex - blue
			glbColor4f(0,0,1,1);
			glVertex3fv(cNode->pos);
			gCopyVertex(v, cNode->binormal); gMultVertex(v, sc); gAddVertex(v, cNode->pos);
			glVertex3fv(v);
			// Add bump-displacement Vertex - yellow
			glbColor4f(1,1,0,1);
			glVertex3fv(cNode->pos);
			gCopyVertex(v, cNode->bump_light); gMultVertex(v, sc); gAddVertex(v, cNode->pos);
			glVertex3fv(v);
*/
		}
		glEnd();
		glEnable(GL_TEXTURE_2D);
	glPopMatrix();
	#endif
#endif

	glDisable(GL_COLOR_MATERIAL);
	glbColor4f(1.0f, 1.0f, 1.0f, m_pRBOW->m_falpha);
#endif // !TINYGL
}
#endif // TRENKI3D



// ------------------------------------------------------------- //
// do Frustum and store the matrix (EGL has no glGetFloatv(XX_MATRIX)
// ------------------------------------------------------------- //
void glbFrustum(float left, float right, float bottom, float top, float znear, float zfar, float m[16])
{
	// float m [16];
	float _2n = 2.0f * znear;
	float _1over_rml = 1.0f / (right - left);
	float _1over_fmn = 1.0f / (zfar - znear);
	float _1over_tmb = 1.0f / (top - bottom);

	m[0] = _2n * _1over_rml;
	m[1] = 0.0f;
	m[2] = 0.0f;
	m[3] = 0.0f;

	m[4] = 0.0f;
	m[5] = _2n * _1over_tmb;
	m[6] = 0.0f;
	m[7] = 0.0f;

	m[8] = (right + left) * _1over_rml;
	m[9] = (top + bottom) * _1over_tmb;
	m[10] = (-(zfar + znear)) * _1over_fmn;
	m[11] = -1.0f;

	m[12] = 0.0f;
	m[13] = 0.0f;
	m[14] = - _2n * zfar * _1over_fmn;
	m[15] = 0.0f;

	#if defined(HAVE_OPENGL)||defined(HAVE_OPENGL_ES)
	// glLoadMatrixf(m);
	glMultMatrixf(m);
	#endif
}

void glbOrtho(float l, float r, float b, float t, float n, float f, float m[16])
{
	m[0]=2/(r - l);
	m[1]=0;
	m[2]=0;
	m[3]=-(r + l)/(r - l);
	m[4]=0;
	m[5]=2/(t - b);
	m[6]=0;
	m[7]=-(t + b)/(t - b);
	m[8]=0;
	m[9]=0;
	m[10]= -2/(f - n);
	m[11]=-(f + n)/(f - n);
	m[12]=0;
	m[13]=0;
	m[14]=0;
	m[15]=1;
	
	#if defined(HAVE_OPENGL)||defined(HAVE_OPENGL_ES)
	// glLoadMatrixf(m);
	glMultMatrixf(m);
	#endif
}


// ------------------------------------------------------------- //
// Toggle Z buffer
// ------------------------------------------------------------- //
void Rainbows3D::SetDepthTesting(bool enable)
{
#ifdef TRENKI3D
	tglEnableDepthTest(enable ? 1:0);
#else
	if(enable)
	{
		glbEnable(GL_DEPTH_TEST); glDepthFunc(GL_LESS);
	}
	else
	{
		glbDisable(GL_DEPTH_TEST);
	}
#endif
}


// ------------------------------------------------------------- //
// Start a 3D Viewport
// ------------------------------------------------------------- //
void Rainbows3D::Make3DViewport(float zNear, float zFar, float fov)
{
int	height, width;
	m_pRBOW->m_bDirtyZBuffer = true;

	DoShadows(false); // draw shadow from last scene -> must be called in MAKE2D & SHOWSCREEN, too

	width =m_pRBOW->m_WND.width;
	height=m_pRBOW->m_WND.height;

	if(height==0)										//Prevent a divide by zero (bad)
		height=1;										//Making height equal one


	m_Camera.tan_fov=tanf( (min(89.0f, fov+5.0f))*BOG_INV_F );// 5% difference due to fast sqrt, qtan is a bit inaccurate... add 5 degress to be sure
	m_Camera.znear = zNear;
	m_Camera.zfar=zFar;
	m_Camera.up[0]=m_Camera.up[2]=0.0f; m_Camera.up[1]=1.0f;



#ifdef TRENKI3D
	// float	aspect=(float)m_pRBOW->m_WND.viewport.width/(float)m_pRBOW->m_WND.viewport.height;
	float	aspect=(float)width/(float)height;
	tgluPerspective(fov, aspect, zNear, zFar);
	tglLoadIdentity();
	
#else

/*
	if(m_pRBOW->m_landscape && m_pRBOW->m_CurrentFBO==-1)
	{
		glViewport(	m_pRBOW->m_WND.width - 1 - m_pRBOW->m_WND.viewport.y -  m_pRBOW->m_WND.viewport.height,
					m_pRBOW->m_WND.height-m_pRBOW->m_WND.viewport.width-m_pRBOW->m_WND.viewport.x,
					m_pRBOW->m_WND.viewport.height,
					m_pRBOW->m_WND.viewport.width);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glRotatef(-90  ,0,0,1);
	}
	else
	{
		glViewport(	m_pRBOW->m_WND.viewport.x,
					m_pRBOW->m_WND.height-m_pRBOW->m_WND.viewport.height-m_pRBOW->m_WND.viewport.y,
					m_pRBOW->m_WND.viewport.width,
					m_pRBOW->m_WND.viewport.height);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
	}
*/


	int x = m_pRBOW->m_WND.viewport.x;
	int y = m_pRBOW->m_WND.viewport.y;
	int w = m_pRBOW->m_WND.viewport.width;
	int h = m_pRBOW->m_WND.viewport.height;

	if(m_pRBOW->m_CurrentFBO==-1)
	{
		glMatrixMode(GL_PROJECTION);
		switch(m_pRBOW->m_landscape)
		{
		default:	
			glLoadIdentity();
			break;
		case 1:
			{
			float matrix[16] = {0,-1,0,0, 1,0,0,0,  0,0,1,0,  0,0,0,1 };
			glLoadMatrixf(matrix);
			break;
			}
		case 2:
			{
			float matrix[16] = {-1,0,0,0, 0,-1,0,0,  0,0,1,0,  0,0,0,1 };
			glLoadMatrixf(matrix);
			break;
			}
		case 3:
			{
			float matrix[16] = {0, 1,0,0, -1,0,0,0,  0,0,1,0,  0,0,0,1 };
			glLoadMatrixf(matrix);
			break;
			}
		}

		int vpx, vpy;
		m_pRBOW->Rot_Screen2Display(x, y, w, h, &vpx, &vpy, &w, &h);

		if(m_pRBOW->m_landscape==1 || m_pRBOW->m_landscape==3)
		{
			int swp=w; w=h; h=swp;
		}

		glViewport(vpx, m_pRBOW->m_WND.height-h-vpy, w, h);
		glMatrixMode(GL_PROJECTION);
	}
	else
	{
		int fb_height = m_pRBOW->m_Texture[m_pRBOW->m_FBuffer[m_pRBOW->m_CurrentFBO].m_SpriteIndex].bh;
		glViewport(	m_pRBOW->m_WND.viewport.x,
					fb_height-m_pRBOW->m_WND.viewport.height-m_pRBOW->m_WND.viewport.y,
					m_pRBOW->m_WND.viewport.width,
					m_pRBOW->m_WND.viewport.height);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
	}

	// float	aspect=(float)m_pRBOW->m_WND.viewport.width/(float)m_pRBOW->m_WND.viewport.height;
	float	aspect=(float)w/(float)h;

// gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,1f,1000.0f);
// void GLAPIENTRY gluPerspective(GLdouble fov=45.0, GLdouble aspect, GLdouble zNear, GLdouble zFar)
// {
	// GLdouble

	float xmin, xmax, ymin, ymax;
	if (fov>0.0f) // perspective view
	{
		ymax = zNear * qtan(fov * (PI_F) / 360.0f);
		ymin = -ymax;
		xmin = ymin * aspect;
		xmax = ymax * aspect;
		glbFrustum(xmin, xmax, ymin, ymax, zNear, zFar, m_ProjectionMatrix);
	}
	else // Ortho view
	{
		ymax = zNear * qtan(-fov * (PI_F) / 360.0f) * 100.0f;
		ymin = -ymax;
		xmin = ymin * aspect;
		xmax = ymax * aspect;
		glbOrtho(xmin, xmax, ymin, ymax, zNear, zFar, m_ProjectionMatrix);
		m_Camera.tan_fov = tanf(89.0f*BOG_INV_F); // for clipping of objects + width of cartoon lines
	}

	glMatrixMode(GL_MODELVIEW);							//Select the modelview matrix
	glLoadIdentity();									//Reset The modelview matrix

// Options for 3D Mode:
// -----------------------------------


	glbEnable(GL_CULL_FACE);
	glFrontFace(GL_CCW);	// Counter clock-wise polygons face out

	glbDisable(GL_FOG);
	glbDisable(GL_LINE_SMOOTH);
	glbDisable(GL_POINT_SMOOTH);
	
	#if defined(HAVE_OPENGL)
	glbDisable(GL_POLYGON_SMOOTH); // Ugly slow otherwise!
	glbDisable(GL_POLYGON_OFFSET_POINT);
	glbDisable(GL_AUTO_NORMAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	#endif

	glbDisable(GL_STENCIL_TEST);
	glbDisable(GL_NORMALIZE);
	glbDisable(GL_LIGHTING);
	// Enable Antialaising
	glShadeModel(GL_SMOOTH);
	//glbEnable(GL_COLOR_MATERIAL);
	glbEnable(GL_DITHER);
	// glGetFloatv(GL_MODELVIEW_MATRIX, (float*)m_CameraMatrix); -> SetCamera does it
	
	glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
#endif


	// Z-Buffer
	this->SetDepthTesting(true);

	m_SphereMapping=false;
	m_BumpLight.active=false;
	// clear transformations (if any)
	while(PopTransformation()) {}

	// Disable culling
	CullMode(0);

	SetTextureOffset(0,0);
	m_pRBOW->SetTexture3D(-1, -1, 0);					// No Texture
	
	SetCamera(0,0,0, 0,0,-1);
}





void LookAt( float eyex, float eyey, float eyez, float centerx, float centery, float centerz, float upx, float upy, float upz, float m[16])
{
#ifdef TRENKI3D
	tglLookAt(eyex,eyey,eyez, centerx,centery,centerz, upx,upy,upz);
#else


// float m[16];
float x[3], y[3], z[3];
float mag;
	// Make rotation matrix

	// Z vector
	z[0] = eyex - centerx;
	z[1] = eyey - centery;
	z[2] = eyez - centerz;
	mag = (float)sqrt(z[0]*z[0] + z[1]*z[1] + z[2]*z[2]);

	if (mag)
	{ // mpichler, 19950515
		z[0] /= mag;
		z[1] /= mag;
		z[2] /= mag;
	}


// Up = 0, 1, 0
/*
	// X vector = Y cross Z
	x[0] = z[2];
	x[1] = 0;
	x[2] = -z[0];

	y[0] =  z[1]*x[2];
	y[1] = -z[0]*x[2] + z[2]*x[0];
	y[2] = -z[1]*x[0];
*/

// otherwise:
	// Y vector
	y[0] = upx;
	y[1] = upy;
	y[2] = upz;

	// X vector = Y cross Z
	x[0] = y[1]*z[2] - y[2]*z[1];
	x[1] = -y[0]*z[2] + y[2]*z[0];
	x[2] = y[0]*z[1] - y[1]*z[0];

	// Recompute Y = Z cross X
	y[0] = z[1]*x[2] - z[2]*x[1];
	y[1] = -z[0]*x[2] + z[2]*x[0];
	y[2] = z[0]*x[1] - z[1]*x[0];
// end otherwise


	// cross product gives area of parallelogram, which is < 1.0 for
	// non-perpendicular unit-length vectors; so normalize x, y here
	mag = (float)sqrt( x[0]*x[0] + x[1]*x[1] + x[2]*x[2] );
	if (mag)
	{
		x[0] /= mag;
		x[1] /= mag;
		x[2] /= mag;
	}

	mag = (float)sqrt( y[0]*y[0] + y[1]*y[1] + y[2]*y[2] );
	if (mag)
	{
		y[0] /= mag;
		y[1] /= mag;
		y[2] /= mag;
	}


	#if 0 // old way - we NEED the matrix in "m", mind!
		float mrot[16] = {x[0],y[0],z[0],0,   x[1],y[1],z[1],0,   x[2],y[2],z[2],0,   0,0,0,1};
		#define M(row,col) mrot[col*4 + row]
		M(0,0) = x[0]; M(0,1) = x[1]; M(0,2) = x[2]; M(0,3) = 0.0;
		M(1,0) = y[0]; M(1,1) = y[1]; M(1,2) = y[2]; M(1,3) = 0.0;
		M(2,0) = z[0]; M(2,1) = z[1]; M(2,2) = z[2]; M(2,3) = 0.0;
		M(3,0) = 0.0; M(3,1) = 0.0; M(3,2) = 0.0; M(3,3) = 1.0;
		#undef M
		glLoadIdentity();
		glMultMatrixf( mrot );
		glTranslatef(-eyex, -eyey, -eyez);
	#else
		float mrot[16] = {x[0],y[0],z[0],0,   x[1],y[1],z[1],0,   x[2],y[2],z[2],0,   0,0,0,1};

		// Translate Eye to Origin
		float mtrans[16] = {1,0,0,0,   0,1,0,0,  0,0,1,0,  -eyex,-eyey,-eyez,1};
		matmul4(m, mrot, mtrans);

		// glLoadIdentity();
		glLoadMatrixf( m );
	#endif


#endif
}


//------------------------------------------------------------------//
// Set Camera UP
//------------------------------------------------------------------//
void Rainbows3D::SetCameraUp(float x, float y, float z)
{
	m_Camera.up[0]=x; m_Camera.up[1]=y; m_Camera.up[2]=z;
}


//------------------------------------------------------------------//
// Set & Point Camera - POSXYZ - PointToXYZ
//------------------------------------------------------------------//
void Rainbows3D::SetCamera(float px, float py, float pz, float rx, float ry, float rz)
{
	m_Camera.eye[0]=px; m_Camera.eye[1]=py; m_Camera.eye[2]=pz;
	m_Camera.lookat[0]=rx; m_Camera.lookat[1]=ry; m_Camera.lookat[2]=rz;
	LookAt(px, py, pz, rx, ry, rz, m_Camera.up[0], m_Camera.up[1], m_Camera.up[2], m_ModelViewCameraMatrix);

	while(PopTransformation()) {}

	// prepare frustum culling
	ExtractFrustum();
}

// ------------------------------------------------------------------//
// Create an ambient light
// ------------------------------------------------------------------//
void Rainbows3D::AmbientLt(int num, unsigned long col)
{
	SpotLt(num, col, m_Camera.eye[0], m_Camera.eye[1], m_Camera.eye[2], 
					-m_Camera.eye[0],-m_Camera.eye[1],-m_Camera.eye[2]
					, 180, 0);
/*
	num = num%8; // Just to be sure
	GLfloat light_ambient[] = {	((BYTE) (col))/255.0f,
								((BYTE) (((WORD) (col)) >> 8))/255.0f,
								((BYTE) ((col) >> 16))/255.0f,
								0.0f};
	// Ambient Light Color
	glLightfv(GL_LIGHT0+num, GL_AMBIENT, light_ambient);

	// Lights ON
	glbEnable(GL_LIGHT0+num);
	glbEnable(GL_LIGHTING);
	m_LightsOn |= (1<<num);
*/
}

// ------------------------------------------------------------------//
// Create a spotlight
// ------------------------------------------------------------------//
void Rainbows3D::SpotLt(int num, unsigned long col, float x, float y, float z, float dx, float dy, float dz, float cutoff, int ShowDebug)
{
#if !defined(TINYGL) && !defined(TRENKI3D)

	// no bump mapping on ES
	#if defined(HAVE_OPENGL_ES)
		if(num==-1) return;
	#endif

	if (num<0)
	{
		m_BumpLight.active=(fabs(cutoff)<1e-5f ? false : true);

		// store pos+colour for active light -> keep when turn off (for shadows)
		if(m_BumpLight.active)
		{
			m_BumpLight.pos[0]=x;
			m_BumpLight.pos[1]=y;
			m_BumpLight.pos[2]=z;
			m_BumpLight.dwcol = col;
		}

		switch(num)
		{
			case -1: // bump
				m_BumpLight.type=1; break;
			default:
			case -2: // cartoon
				m_BumpLight.type=2; break;
			case -3: // shadow
				m_BumpLight.type=3;
				DoShadows(m_BumpLight.active);
				break;
		}
		return;
	}

	num = num%8; // Just to be sure
	if (cutoff==0.0f)
	{
		glDisable(GL_LIGHT0+num);
		m_LightsOn &= ~(1<<num);
		if(!m_LightsOn) glbDisable(GL_LIGHTING);
		return;
	}

	// Color + Direction
	static GLfloat light_off[]={0,0,0,1};
	static GLfloat light_on[]={1,1,1,1};
	GLfloat light_pos[] = { x,  y,  z, 1.0f/*fix*/}; // w=0.0: Directional light - position == direction(pretty stupid)
	GLfloat light_dir[] = {dx, dy, dz, 0.0};
	GLfloat light_diffuse[] = {	((BYTE) (col))/255.0f, // Red
								((BYTE) (((WORD) (col)) >> 8))/255.0f, // Green
								((BYTE) ((col) >> 16))/255.0f, // Blue
								1.0f};
	glbEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0+num);
	m_LightsOn |= (1<<num);

	//glbEnable(GL_COLOR_MATERIAL);
	#if defined(HAVE_OPENGL)
	glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
	#endif

	glMatrixMode(GL_MODELVIEW);
	{glPushMatrix();
//		glLoadIdentity();
//	glLoadMatrixf(m_CameraMatrix);
	// no direction, no pos
	if (x==0 && y==0 && z==0 && dx==0 && dy==0 && dz==0)
	{
		float d[]={0,0,1,0};
		glLightfv(GL_LIGHT0+num, GL_POSITION, d);

		glLightfv(GL_LIGHT0+num, GL_DIFFUSE, light_diffuse);
		glLightfv(GL_LIGHT0+num, GL_SPECULAR, light_off);
		glLightf(GL_LIGHT0+num, GL_SPOT_CUTOFF, 180);
	return;
	}

	if(cutoff==180) // parallel light (direction, no position)
	{
		glLightf (GL_LIGHT0+num, GL_SPOT_CUTOFF, 180);
		gMultVertex(light_dir, -1);
		gUnify(light_dir);
		light_dir[3] = 0;
		glLightfv(GL_LIGHT0+num, GL_POSITION, light_dir);
		glLightfv(GL_LIGHT0+num, GL_DIFFUSE, light_diffuse);
		glLightfv(GL_LIGHT0+num, GL_SPECULAR, light_off);
	}
	else if (cutoff <= 90) // spot light
	{
		glLightf (GL_LIGHT0+num, GL_SPOT_CUTOFF,    cutoff);
		gUnify(light_dir);
		glLightfv(GL_LIGHT0+num, GL_POSITION,       light_pos);
		glLightfv(GL_LIGHT0+num, GL_SPOT_DIRECTION, light_dir);
		glLightfv(GL_LIGHT0+num, GL_DIFFUSE, light_diffuse);
		glLightfv(GL_LIGHT0+num, GL_SPECULAR, light_diffuse);
	}
	else // (360) point light
	{
		glLightf(GL_LIGHT0+num, GL_SPOT_CUTOFF, 180);
		glLightfv(GL_LIGHT0+num, GL_POSITION, light_pos);
		glLightfv(GL_LIGHT0+num, GL_DIFFUSE, light_diffuse);
		glLightfv(GL_LIGHT0+num, GL_SPECULAR, light_diffuse);
	}
	}glPopMatrix();


	if (ShowDebug)
	{
		DrawDot(light_pos, RGB(255,200,128), 40);
		if (cutoff<=90)
		{
			float vec[3];
			gCopyVertex(vec, light_dir);
			gUnify(vec);
			gMultVertex(vec, 500);
			gAddVertex(vec, light_pos);
			DrawLine(light_pos, vec, RGB(225,170,108), 3);
		}
	}
#endif
}





//------------------------------------------------------------------//
// Draw the Coord-Axis
//------------------------------------------------------------------//
void Rainbows3D::DrawCoordAxis(float x, float y, float z)
{
#if !defined(TINYGL) && !defined(TRENKI3D)
	glPushMatrix();
		glLoadMatrixf((const float*) m_ModelViewCameraMatrix);

		//glbEnable(GL_COLOR_MATERIAL);
		glbEnable( GL_LINE_SMOOTH );
		if(m_pRBOW->m_Texture3D[0]>=0) glDisable(GL_TEXTURE_2D);
		if(m_LightsOn) glbDisable(GL_LIGHTING);

		// Move
		glTranslatef(x, y, z);
		float scale = 20.0f;
		glScalef(scale,scale,scale);

		glEnableClientState( GL_VERTEX_ARRAY );

		// Draw
GLfloat coox[]={
				0.0,  0.0,  0.0,
				1.0,  0.0,  0.0,
				0.75, 0.25, 0.0,
				0.75,-0.25, 0.0,
				1.0,  0.0,  0.0,
				0.75, 0.0, 0.25,
				0.75, 0.0,-0.25,
				1.0,  0.0,  0.0};
GLfloat cooy[]={
				0.0,   0.0,  0.00,
				0.0,   1.0,  0.00,
				0.0,   0.75, 0.25,
				0.0,   0.75,-0.25,
				0.0,   1.0,  0.00,
				0.25,  0.75, 0.00,
				-0.25, 0.75, 0.00,
				0.0,   1.0,  0.00};
GLfloat cooz[]={
				0.00, 0.00, 0.00,
				0.00, 0.00, 1.00,
				0.25, 0.00, 0.75,
				0.25, 0.00, 0.75,
				0.00, 0.00, 1.00,
				0.00, 0.25, 0.75,
				0.00,-0.25, 0.75,
				0.00, 0.00, 1.00};


		glbColor4f(1.0f, 0, 0, 1.0f);
		glVertexPointer(  3, GL_FLOAT, 0, (char*)&coox[0]);
		glDrawArrays(GL_LINE_STRIP, 0, sizeof(coox)/(3*sizeof(GLfloat)));

		glbColor4f(0,     1.0f, 0, 1.0f);
		glVertexPointer(  3, GL_FLOAT, 0, (char*)&cooy[0]);
		glDrawArrays(GL_LINE_STRIP, 0, sizeof(cooy)/(3*sizeof(GLfloat)));

		glbColor4f(0, 0, 1.0f,1.0f);
		glVertexPointer(  3, GL_FLOAT, 0, (char*)&cooz[0]);
		glDrawArrays(GL_LINE_STRIP, 0, sizeof(cooz)/(3*sizeof(GLfloat)));

		glDisableClientState( GL_VERTEX_ARRAY );

		if(m_LightsOn) glbEnable(GL_LIGHTING);

		glbDisable( GL_LINE_SMOOTH );
		glbDisable(GL_COLOR_MATERIAL);
		if(m_pRBOW->m_Texture3D[0]>=0)
			glEnable(GL_TEXTURE_2D);
		glbColor4f(1.0f, 1.0f, 1.0f, m_pRBOW->m_falpha);

	glPopMatrix();
#endif
}




//------------------------------------------------------------------//
// Start Creating a new 3D Object by hand-drawing it
//------------------------------------------------------------------//
void Rainbows3D::UserObjectStart(int num, int type)
{
	m_iCurObjFrame = 0;
	m_Objects[num].Clear(false);
	m_Objects[num].m_objects.AllocAtLeast(1);

	gCurObject=num;

POLY_FRAME	Frame;
	m_Objects[num].m_objects.SetSize(1);
	m_Objects[num].m_objects[0].obj.m_Frames.Push(Frame);

	UserObjectSetNormal(0,0,1);

	// New Object
	UserObjectNewGroup(type);
}

//------------------------------------------------------------------//
// switch to another frame
//------------------------------------------------------------------//
void Rainbows3D::UserObjectSetFrame(int frm)
{
	if(frm<0) return;

	Queue<POLY_FRAME>& frames = m_Objects[gCurObject].m_objects[0].obj.m_Frames;

	if(frames.Size() <= frm)
	{
		frames.SetSize(frm+1);
	}

	// if it's an empty frame, push the "group type" of the startobj call
	if(frames[frm].nodes.Size() == 0)
	{
		frames[frm].nodes.Push(m_Objects[gCurObject].m_objects[0].obj.m_Frames[0].nodes[0] );
	}

	m_iCurObjFrame = frm;
}

//------------------------------------------------------------------//
// Restart new Stripe at Hand Drawing
//------------------------------------------------------------------//
void Rainbows3D::UserObjectNewGroup(int type)
{
	// Determine how to start -> See also UserObjectNewGroup!
	unsigned int col = COL_NEW_STRIP;
	switch (type)
	{
	case 0: // Triangles
		col=COL_NEW_TRI;
		break;
	case 1:
		col=COL_NEW_STRIP;
		break;
	case 2:
		col=COL_NEW_FAN;
		break;
	}
POLY_NODE	Node;
	memset(&Node, 0, sizeof(Node));
	Node.col(col);
	m_Objects[gCurObject].m_objects[0].obj.m_Frames.GetAt(m_iCurObjFrame)->nodes.Push(Node);
}

//------------------------------------------------------------------//
// Add a Vertex to a Hand Drawing
//------------------------------------------------------------------//
void Rainbows3D::UserObjectAddVertex(float x, float y, float z, float tx, float ty, unsigned long col)
{
	if (gCurObject!=-1)
	{
		POLY_NODE&	Node = m_UserObjectNode;
		// And Store it
		Node.pos[0]=x;
		Node.pos[1]=y;
		Node.pos[2]=z;
		Node.tex[0]=  tx; // tx
		Node.tex[1]= -ty; // ty  - texture flipped due to grab sprite compatibility
		Node.col(col);
		m_Objects[gCurObject].m_objects[0].obj.m_Frames.GetAt(m_iCurObjFrame)->nodes.Push(Node);
		printf("%d %d\n", m_iCurObjFrame, m_Objects[gCurObject].m_objects[0].obj.m_Frames.GetAt(m_iCurObjFrame)->nodes.Size());
	}
}

//------------------------------------------------------------------//
// Add a normal direction to a Hand Drawing with Autonormals 0
//------------------------------------------------------------------//
void Rainbows3D::UserObjectSetNormal(float x, float y, float z)
{
	if (gCurObject!=-1)
	{
		POLY_NODE&	Node = m_UserObjectNode;
		// And Store it
		Node.normal[0]=x;
		Node.normal[1]=y;
		Node.normal[2]=z;
	}
}


//------------------------------------------------------------------//
// Finish a Hand Drawing
//------------------------------------------------------------------//
void Rainbows3D::UserObjectEnd()
{
Queue <POLY_NODE>*	pN;
	MeshObjectX& the_obj = m_Objects[gCurObject].m_objects[0].obj;
	m_Objects[gCurObject].is_multiworld=false;
	pN=&the_obj.m_Frames.GetAt(0)->nodes;

	if(the_obj.m_Frames.Size()==0 || the_obj.m_Frames.GetAt(0)->nodes.Size()<3)
	{
		the_obj.Clear(true);
		m_Objects[gCurObject].is_loaded = false;
	}
	else
	{
		bool bOK = true;

		m_Objects[gCurObject].is_loaded=true;

		int nof_nodes0 = the_obj.m_Frames[0].nodes.Size();
		for(int f=0; f<the_obj.NumFrames(); ++f)
		{
			if(the_obj.m_Frames[f].nodes.Size() != nof_nodes0)
			{
				WIN32_ONLY( DEBUG(CGStr("Frame ")+DGNat(f)+CGStr(" has ")+DGNat(the_obj.m_Frames[f].nodes.Size())
									+CGStr(" nodes, but must have (like frame 0) ") + DGNat(nof_nodes0) + CGStr(" nodes\n") ));
				m_Objects[gCurObject].is_loaded=false;
				bOK = false;
				break;
			}
		}

		// Compress Data
		if(bOK)
		{
			the_obj.BuildColorTable();
			switch (m_AutoNormalMode)
			{
				case 0: // No normals (or user defined)
					break;
				case 1: // Flat normals
					the_obj.AutoNormalsAfterCompression(true, false, false);
					break;
				case 2: // smooth normals
					the_obj.AutoNormalsAfterCompression(true, true, false);
					break;
			}
	//		PreRenderObject(gCurObject);

			// Not yet created triangles due to normal calculations?
			if (the_obj.m_tri_animation.Size()==0)
				the_obj.PrepareTriangles(); // For VBO rendering
			the_obj.CalcMaxRadius();
		}
	}
	gCurObject=-1;
}


#ifdef _DEBUG
// #define DEBUG_COLLISIONS -> glb3d does this for you if __DG_DEBUG
#endif


//------------------------------------------------------------------//
// Draw a Dot in World Coordinates
//------------------------------------------------------------------//
void Rainbows3D::DrawDot(const float *pos, unsigned long col, float size)
{
#if !defined(TINYGL) && !defined(TRENKI3D)
	glPushMatrix();
		// Do not gTransformMatrix()
		glbEnable(GL_COLOR_MATERIAL);
		glbColor4f(	   (col      &0xff)/ 255.0f,
					  ((col>> 8) &0xff)/ 255.0f,
					  ((col>> 16)&0xff)/ 255.0f, 1.0f);
		glbEnable( GL_POINT_SMOOTH );
		if(m_LightsOn) glbDisable(GL_LIGHTING);
		m_pRBOW->SetTexture3D(-1, -1, 0);
		glPointSize( size );

		glBegin(GL_POINTS);
		glVertex3fv(pos);
		glEnd();

		glPointSize( 1 );
		glbDisable( GL_POINT_SMOOTH );
		glbColor4f(1.0f, 1.0f, 1.0f, m_pRBOW->m_falpha);
		glbDisable(GL_COLOR_MATERIAL);
		if(m_LightsOn) glbEnable(GL_LIGHTING);
	glPopMatrix();
#endif
}

//------------------------------------------------------------------//
// Draw a Line in World Coordinates
//------------------------------------------------------------------//
void Rainbows3D::DrawLine(const float* P1, const float* P2, unsigned long col, float width)
{
#if !defined(TINYGL) && !defined(TRENKI3D)
	glPushMatrix();
		glbEnable(GL_COLOR_MATERIAL); // Noo-hooo
		// Do not gTransformMatrix()
		glbColor4f(	 (col      &0xff)/ 255.0f,
					((col>> 8) &0xff)/ 255.0f,
					((col>> 16)&0xff)/ 255.0f, 1.0f);
		if(m_LightsOn) glbDisable(GL_LIGHTING);
		m_pRBOW->SetTexture3D(-1, -1, 0);

		glLineWidth( width );
		#if defined(HAVE_OPENGL_ES)
			float pts[] = {P1[0], P1[1], P1[2], P2[0], P2[1], P2[2]};
			glEnableClientState( GL_VERTEX_ARRAY );
			glVertexPointer( 3, GL_FLOAT, 0, (char*)pts);
			glDrawArrays(GL_LINES, 0, 2);
			glDisableClientState( GL_VERTEX_ARRAY );
		#else
			glbEnable( GL_LINE_SMOOTH );
			glBegin(GL_LINES);
			glVertex3fv(P1); glVertex3fv(P2);
			glEnd();
		#endif
		glLineWidth( 1 );

		glbDisable( GL_LINE_SMOOTH );
		glbColor4f(1.0f, 1.0f, 1.0f, m_pRBOW->m_falpha);
		glbDisable(GL_COLOR_MATERIAL);
		if(m_LightsOn) glbEnable(GL_LIGHTING);
	glPopMatrix();
#endif
}

// --------------------------------------------------------------- //
// Subdivide an octahedron to match a sphere and draw it
// --------------------------------------------------------------- //
inline void SubdivideSphereGrid(float* org, float R, float* v1, float* v2, float* v3, int depth)
{
#if !defined(TINYGL) && !defined(TRENKI3D)
	float v12[3], v23[3], v31[3];
	if (depth == 0)
	{
		float tmp[3];
		glBegin(GL_LINE_LOOP);
		gCopyVertex(tmp, v1); gMultVertex(tmp, R); gAddVertex(tmp, org); glVertex3fv(tmp);
		gCopyVertex(tmp, v2); gMultVertex(tmp, R); gAddVertex(tmp, org); glVertex3fv(tmp);
		gCopyVertex(tmp, v3); gMultVertex(tmp, R); gAddVertex(tmp, org); glVertex3fv(tmp);
		glEnd();
		return;
	}
	gCopyVertex(v12, v1); gAddVertex(v12, v2);
	gCopyVertex(v23, v2); gAddVertex(v23, v3);
	gCopyVertex(v31, v3); gAddVertex(v31, v1);
	gUnify(v12); gUnify(v23); gUnify(v31);

	SubdivideSphereGrid(org,R, v1, v12, v31, depth-1);
	SubdivideSphereGrid(org,R, v2, v23, v12, depth-1);
	SubdivideSphereGrid(org,R, v3, v31, v23, depth-1);
	SubdivideSphereGrid(org,R, v12,v23, v31, depth-1);
#endif
}

// --------------------------------------------------------------- //
// Draw a tesselated sphere
// --------------------------------------------------------------- //
void Rainbows3D::DrawSphereGrid(float* org, float R, unsigned long col, short width)
{
#if !defined(TINYGL) && !defined(TRENKI3D)
	// User Constants
#define X .525731112119133606f
#define Z .850650808352039932f
	static float vdata[12][3] = {    
		{-X, 0.0, Z}, {X, 0.0, Z}, {-X, 0.0, -Z}, {X, 0.0, -Z},    
		{0.0, Z, X}, {0.0, Z, -X}, {0.0, -Z, X}, {0.0, -Z, -X},    
		{Z, X, 0.0}, {-Z, X, 0.0}, {Z, -X, 0.0}, {-Z, -X, 0.0} 
	};
#undef X
#undef Y

	static int tindices[20][3] = { 
		{0,4,1}, {0,9,4}, {9,5,4}, {4,5,8}, {4,8,1},    
		{8,10,1}, {8,3,10}, {5,3,8}, {5,2,3}, {2,7,3},    
		{7,10,3}, {7,6,10}, {7,11,6}, {11,0,6}, {0,1,6}, 
		{6,1,10}, {9,0,11}, {9,11,2}, {9,2,5}, {7,2,11} };


	if(m_LightsOn) glbDisable(GL_LIGHTING);
	m_pRBOW->SetTexture3D(-1, -1, 0);
	glbDisable(GL_BLEND);
	glbEnable(GL_COLOR_MATERIAL);
	glLineWidth( width );
	glPushMatrix();
	glbColor4f(	((BYTE) (col)) / 255.0f,
				((BYTE) (((WORD) (col)) >> 8)) / 255.0f,
				((BYTE) ((col) >> 16)) / 255.0f, 1.0f);

	for (long i=0; i<20; i++) 
	{
		float v1[3], v2[3], v3[3];
		gCopyVertex(v1, vdata[tindices[i][0]]);
		gCopyVertex(v2, vdata[tindices[i][1]]);
		gCopyVertex(v3, vdata[tindices[i][2]]);
		SubdivideSphereGrid(org,R, v1, v2, v3, 1); // depth -> [0; 5]
	}
	glLineWidth( 1 );
	glbDisable(GL_COLOR_MATERIAL);
	glPopMatrix();
	if(m_LightsOn) glbEnable(GL_LIGHTING);
	glbColor4f(1.0f, 1.0f, 1.0f, m_pRBOW->m_falpha);
#endif
}


// --------------------------------------------------------------- //
// Draw an AABB test object
// C1 = center of the AABB
// C2,3,4 are the direction vertices to 3 bounding plane's centers
// --------------------------------------------------------------- //
void Rainbows3D::DrawAABB(const float *C1, const float* C2, const float* C3, const float* C4, unsigned long col, short width)
{
#if !defined(TINYGL) && !defined(TRENKI3D)
float tmp[3], nC1[3], nC2[3], nC3[3], nC4[3];
		m_pRBOW->SetTexture3D(-1, -1, 0);
		glbDisable(GL_BLEND);
		glbEnable(GL_COLOR_MATERIAL);
		glLineWidth( width );
		glPushMatrix();
		// Do not gTransformMatrix()
		glbColor4f(	((BYTE) (col)) / 255.0f,
					((BYTE) (((WORD) (col)) >> 8)) / 255.0f,
					((BYTE) ((col) >> 16)) / 255.0f, 1.0f);
		if(m_LightsOn) glbDisable(GL_LIGHTING);

		
		// nC1 = lower, left corner of box
		gDiffVertex(nC1, C1 , C2); gSubVertex(nC1, C3); gSubVertex(nC1, C4);
		
		gCopyVertex(nC2, C2); gMultVertex(nC2,2.0f);
		gCopyVertex(nC3, C3); gMultVertex(nC3,2.0f);
		gCopyVertex(nC4, C4); gMultVertex(nC4,2.0f);
		
		glBegin(GL_LINES);
			#define DLINE(from, delta, fac) glVertex3fv(from); gCopyVertex(tmp, delta); gMultVertex(tmp, fac); gAddVertex(tmp, from); glVertex3fv(tmp);
			DLINE(nC1, C2, 2.0f);
			DLINE(nC1, C3, 2.0f);
			DLINE(nC1, C4, 2.0f);
			gAddVertex(nC1, C2); gAddVertex(nC1, C2);
			DLINE(nC1, C3, 2.0f);
			DLINE(nC1, C4, 2.0f);
			gAddVertex(nC1, C3); gAddVertex(nC1, C3);
			DLINE(nC1, C2, -2.0f);
			DLINE(nC1, C4,  2.0f);
			gAddVertex(nC1, C4); gAddVertex(nC1, C4);
			DLINE(nC1, C2, -2.0f);
			DLINE(nC1, C3, -2.0f);
			gSubVertex(nC1, C2); gSubVertex(nC1, C2);
			DLINE(nC1, C3, -2.0f);
			DLINE(nC1, C4, -2.0f);
			gSubVertex(nC1, C3); gSubVertex(nC1, C3);
			DLINE(nC1, C2, 2.0f);
		glEnd();

		glLineWidth( 1 );
		glbDisable(GL_COLOR_MATERIAL);

		glPopMatrix();
		if(m_LightsOn) glbEnable(GL_LIGHTING);
		glbColor4f(1.0f, 1.0f, 1.0f, m_pRBOW->m_falpha);
#endif
}

int g_CollisionTriangleIndex=-1; // index of the first triangle's vertex in the triangles list
int   CollisionGetLastCollisionTriangleIndex()
{
	return g_CollisionTriangleIndex / 3; // Get the FACE index (3 triangles is one face.)
}


bool Rainbows3D::CollisionAABB(int id, int frame, const float *C1, const float* C2, const float* C3, const float* C4)
{
	X_WORLD& wld = m_Objects[id];
	for(int i=0; i<wld.m_objects.Size(); ++i)
		if(CollisionAABB(wld.m_objects[i].obj, frame, C1, C2, C3, C4)) return true;
return false;
}
bool Rainbows3D::CollisionSphere(int id, int frame, float radius, float x, float y, float z)
{
	X_WORLD& wld = m_Objects[id];
	for(int i=0; i<wld.m_objects.Size(); ++i)
		if(CollisionSphere(wld.m_objects[i].obj, frame, radius, x,y,z)) return true;
return false;
}
float Rainbows3D::CollisionRay(int id, int frame, float x, float y, float z, float dirx, float diry, float dirz)
{
	X_WORLD& wld = m_Objects[id];
float rfac_neg = 0.0f, rfac_pos=0.0, fac;
	for(int i=0; i<wld.m_objects.Size(); ++i)
	{
		fac=CollisionRay(wld.m_objects[i].obj, frame, x,y,z,dirx,diry,dirz);
		if (fac!= 0.0f)
		{
			if(fac > 0.0f && (fac < rfac_pos || rfac_pos == 0.0f))
				rfac_pos = fac; // found a better positive value
			else
			if(fac < 0.0f && (fac > rfac_neg || rfac_neg == 0.0f))
				rfac_neg=fac; // found a better negative value
		}
	}
return rfac_pos == 0.0f ? rfac_neg : rfac_pos;
}



//------------------------------------------------------------------//
// Collision detection - axis aligned bounding box, object's triangles
// C1 = center of the AABB
// C2,3,4 are the direction vertices to 3 bounding plane's centers
//------------------------------------------------------------------//
bool Rainbows3D::CollisionAABB(MeshObjectX& the_obj, int frame, const float *C1, const float* C2, const float* C3, const float* C4)
{
g_CollisionTriangleIndex=-1;
	if (the_obj.m_Frames.Size()<=frame) return false;
	if (the_obj.m_Frames.GetAt(frame)->triangles.Size()<3) return false;
	
float	m[16];
float	nC1[3], nC2[3], nC3[3], nC4[3];
POLY_FRAME*	pFrm = the_obj.m_Frames.GetAt(frame);
	// Shift & rotate the AABB into the model's position
	// This way we don't have to transform each of the object's points
/*
	glPushMatrix();
		glLoadIdentity();
		TransformMatrix();
		glGetFloatv(GL_MODELVIEW_MATRIX, m);
	glPopMatrix();
	gInvertMatrix4x4(m);
*/
	BuildMatrix();
	gCopyMatrix(m, m_Transformations[m_nTransformations].m_Matrix);
	gInvertMatrix4x4(m);

	// This one is absolute
	gCopyVertex(nC1, C1);
	//gTransformVector(m_CameraMatrix, nC1);
	gTransformVector(m, nC1);
	
	// Make a very quick sphere-sphere test
	gCopyVertex(nC2, C2); gAddVertex(nC2, C3); gAddVertex(nC2, C4);
	// (gDot(a,a) = Squared length or radius
	if(!IsPointInSphere(pFrm->center_of_mass, nC1, gDot(nC2, nC2) + SQUARE(pFrm->max_radius))) return false;
	
	// These are realative to nC1 - we must transform an absolute point however!
	//#define TRANSF(out, in) gCopyVertex(out, in); gAddVertex(out, C1); gTransformVector(m_CameraMatrix, out); gTransformVector(m, out); gSubVertex(out, nC1);
	#define TRANSF(out, in) gCopyVertex(out, in); gAddVertex(out, C1); gTransformVector(m, out); gSubVertex(out, nC1);
	TRANSF(nC2, C2); TRANSF(nC3, C3); TRANSF(nC4, C4);

	int num_pts = (int)pFrm->triangles.Size();
	for(int i=0; i<num_pts; i+=3)
	{
		if (	IntersectCuboidAndTriangle(nC1, nC2, nC3, nC4,	pFrm->triangles.GetAt(i)->pos,
																pFrm->triangles.GetAt(i+1)->pos,
																pFrm->triangles.GetAt(i+2)->pos))
			{
				g_CollisionTriangleIndex=i;	
				return true;
			}
	}
return false;
}


//------------------------------------------------------------------//
// Collision detection - sphere, object's triangles
//------------------------------------------------------------------//
bool Rainbows3D::CollisionSphere(MeshObjectX& the_obj, int frame, float radius, float x, float y, float z)
{
g_CollisionTriangleIndex=-1;
	if (the_obj.m_Frames.Size()<=frame) return false;
	if (the_obj.m_Frames.GetAt(frame)->triangles.Size()<3) return false;
	POLY_FRAME* pFrm = the_obj.m_Frames.GetAt(frame);
	
float	invM[16];
float sphere[3];
POLY_NODE*	pNode;
float	tri[3][3];
int		i;

	// Shift & rotate the sphere into the model's position
	// This way we don't have to transform each of the object's points
	sphere[0]=x; sphere[1]=y; sphere[2]=z;
	BuildMatrix();
	gCopyMatrix(invM, m_Transformations[m_nTransformations].m_Matrix);
	gInvertMatrix4x4(invM);
//	gTransformVector(m_CameraMatrix, sphere); // Move sphere to it's position due to camera
	gTransformVector(invM, sphere);
	
	// Quick test: Do the bounding spheres of the object and the test sphere collide?
	if(!IsPointInSphere(pFrm->center_of_mass, sphere, SQUARE(radius + pFrm->max_radius))) return false;

	// Special case: All triangles do not intersect, but lie completely inside sphere!
	// Here we test only for the first node to be inside.
	// This is very quick, so I do it at the very beginning
	if(IsPointInSphere(pFrm->triangles.GetAt(0)->pos, sphere, SQUARE(radius))) return true;

	for(i=0; i<(int)pFrm->triangles.Size(); )
	{
		pNode = pFrm->triangles.GetAt(i++);
		tri[0][0]=pNode->pos[0];
		tri[0][1]=pNode->pos[1];
		tri[0][2]=pNode->pos[2];
		pNode = pFrm->triangles.GetAt(i++);
		tri[1][0]=pNode->pos[0];
		tri[1][1]=pNode->pos[1];
		tri[1][2]=pNode->pos[2];
		pNode = pFrm->triangles.GetAt(i++);
		tri[2][0]=pNode->pos[0];
		tri[2][1]=pNode->pos[1];
		tri[2][2]=pNode->pos[2];
		if (::SpherePolygonCollision(tri, sphere, 3, radius) )
		{
	// Vizualize Hitting Triangle
#ifdef DEBUG_COLLISIONS
	sphere[0]=x; sphere[1]=y; sphere[2]=z;
	DrawSphereGrid(sphere, radius, RGB(255,0,0));
#endif
			g_CollisionTriangleIndex=i-3;
			return true;
		}
	}
#ifdef DEBUG_COLLISIONS
	sphere[0]=x; sphere[1]=y; sphere[2]=z;
	DrawSphereGrid(sphere, radius, RGB(0,255,0));
#endif
return false;
}

//------------------------------------------------------------------//
// Collision detection - ray, object's triangles
//------------------------------------------------------------------//
float Rainbows3D::CollisionRay(MeshObjectX& the_obj, int frame, float x, float y, float z, float dirx, float diry, float dirz)
{
	int collision_pos=-1;
	int collision_neg=-1;

	if (the_obj.m_Frames.Size()<=frame) return false;
	if (the_obj.m_Frames.GetAt(frame)->triangles.Size()<3) return false;
	POLY_FRAME* pFrm = the_obj.m_Frames.GetAt(frame);
	
float		invM[16];
float		pt1[3], pt2[3];
POLY_NODE*	pNode;
float		tri[3][3];
int			i;

	// Shift & rotate the line into the model's position
	// This way we don't have to transform each of the object's points
	pt1[0]=x; pt1[1]=y; pt1[2]=z;
	pt2[0]=dirx; pt2[1]=diry; pt2[2]=dirz;
	gUnify(pt2);
	gAddVertex(pt2, pt1);
	BuildMatrix();
	gCopyMatrix(invM, m_Transformations[m_nTransformations].m_Matrix);
	gInvertMatrix4x4(invM);
//	gTransformVector(m_CameraMatrix, pt1); // Move point to it's position due to camera
	gTransformVector(invM, pt1);
//	gTransformVector(m_CameraMatrix, pt2);
	gTransformVector(invM, pt2);

float rfac_neg = 0.0f, rfac_pos=0.0, fac;
	for(i=0; i<(int)pFrm->triangles.Size(); i+=3)
	{
		pNode = pFrm->triangles.GetAt(i);
		tri[0][0]=pNode->pos[0];
		tri[0][1]=pNode->pos[1];
		tri[0][2]=pNode->pos[2];
		pNode = pFrm->triangles.GetAt(i+1);
		tri[1][0]=pNode->pos[0];
		tri[1][1]=pNode->pos[1];
		tri[1][2]=pNode->pos[2];
		pNode = pFrm->triangles.GetAt(i+2);
		tri[2][0]=pNode->pos[0];
		tri[2][1]=pNode->pos[1];
		tri[2][2]=pNode->pos[2];
		fac=::IntersectTriangleAndRay(tri, pt1, pt2);
	
		if (fac!= 0.0f)
		{
			if(fac > 0.0f && (fac < rfac_pos || rfac_pos == 0.0f))
			{
				rfac_pos = fac; // found a better positive value
				collision_pos=i;
			}
			else
			if(fac < 0.0f && (fac > rfac_neg || rfac_neg == 0.0f))
			{
				rfac_neg=fac; // found a better negative value
				collision_neg=i;
			}
		}

#ifdef DEBUG_COLLISIONS
		if (fac!=0)
		{
			glbColor4f(1,0,0,1);
			glbDisable(GL_BLEND);
			glbEnable(GL_COLOR_MATERIAL);
			glLineWidth( 5 );
			glPushMatrix();
			TransformMatrix();
			glBegin(GL_TRIANGLES);//GL_LINE_STRIP);
			glVertex3fv(tri[0]);
			glVertex3fv(tri[1]);
			glVertex3fv(tri[2]);
			//glVertex3fv(tri[0]);
			glEnd();
			glPopMatrix();
			glLineWidth( 1 );
			glbDisable(GL_COLOR_MATERIAL);
		}
#endif
	}

	if(rfac_pos == 0.0f)
	{
		g_CollisionTriangleIndex = collision_neg;
		return rfac_neg;
	}
	g_CollisionTriangleIndex = collision_pos;
	return rfac_pos;
}


//------------------------------------------------------------------//
// Collision detection - Cone, object's triangles
//------------------------------------------------------------------//
float Rainbows3D::CollisionCone(int id, int frame, float x, float y, float z, float dirx, float diry, float dirz, float angle)
{
	X_WORLD& wld = m_Objects[id];
	for(int i=0; i<wld.m_objects.Size(); ++i)
		if(CollisionCone(wld.m_objects[i].obj, frame, x, y, z, dirx, diry, dirz, angle)) return true;
return false;
}


//------------------------------------------------------------------//
//------------------------------------------------------------------//
float Rainbows3D::CollisionCone(MeshObjectX& the_obj, int frame, float x, float y, float z, float dirx, float diry, float dirz, float angle)
{
float c_start[3], c_end[3];
	c_start[0]=x; c_start[1]=y; c_start[2]=z;
	c_end[0]=x+dirx*10000.0f; c_end[1]=y+diry*10000.0f; c_end[2]=z+dirz*10000.0f;

	angle *= BOG_INV_F;

	if (the_obj.m_Frames.Size()<=frame) return false;
	if (the_obj.m_Frames.GetAt(frame)->triangles.Size()<3) return false;
	POLY_FRAME* pFrm = the_obj.m_Frames.GetAt(frame);
	
	if(::gSphereInCone(c_start, c_end, tanf(angle), pFrm->center_of_mass, pFrm->max_radius))
	{
	float		invM[16];
	float		pt1[3], pt2[3];
	POLY_NODE*	pNode;
	float		tri[3][3];
	int			i;

		// Shift & rotate the line into the model's position
		// This way we don't have to transform each of the object's points
		pt1[0]=x; pt1[1]=y; pt1[2]=z;
		pt2[0]=dirx; pt2[1]=diry; pt2[2]=dirz;
		gUnify(pt2);
		gAddVertex(pt2, pt1);
		BuildMatrix();
		gCopyMatrix(invM, m_Transformations[m_nTransformations].m_Matrix);
		gInvertMatrix4x4(invM);
	//	gTransformVector(m_CameraMatrix, pt1); // Move point to it's position due to camera
		gTransformVector(invM, pt1);
	//	gTransformVector(m_CameraMatrix, pt2);
		gTransformVector(invM, pt2);


	float cos_angle = cosf(angle);
	float axis[3];
		axis[0] = dirx; axis[1]=diry; axis[2]=dirz;
		for(i=0; i<(int)pFrm->triangles.Size(); i+=3)
		{
			pNode = pFrm->triangles.GetAt(i);
			tri[0][0]=pNode->pos[0];
			tri[0][1]=pNode->pos[1];
			tri[0][2]=pNode->pos[2];
			pNode = pFrm->triangles.GetAt(i+1);
			tri[1][0]=pNode->pos[0];
			tri[1][1]=pNode->pos[1];
			tri[1][2]=pNode->pos[2];
			pNode = pFrm->triangles.GetAt(i+2);
			tri[2][0]=pNode->pos[0];
			tri[2][1]=pNode->pos[1];
			tri[2][2]=pNode->pos[2];
			if (TriangleAndCone(axis, c_start, cos_angle, tri))
				return true;
		}
	}
return false;
}




// --------------------------------
// sphere/AABB collision. AABB in current matrix transformation
// --------------------------------
bool  Rainbows3D::CollisionSphereBox(float x, float y, float z, float r, float bx, float by, float bz)
{
float sphere[3], invM[16];

	// Shift & rotate the sphere into the model's position
	// This way we don't have to transform each of the object's points
	sphere[0]=x; sphere[1]=y; sphere[2]=z;
	BuildMatrix();
	gCopyMatrix(invM, m_Transformations[m_nTransformations].m_Matrix);
	gInvertMatrix4x4(invM);
//	gTransformVector(m_CameraMatrix, sphere); // Move sphere to it's position due to camera
	gTransformVector(invM, sphere);

	// Check for collision
	return ::IntersectSphereBox(sphere, r*r, bx, by, bz);
}




//------------------------------------------------------------------//
// Load an Object in Rainbows-Format
//------------------------------------------------------------------//
extern bool Load_DDW(const char* filename, X_WORLD& world, COLORREF cookiecol);
bool Rainbows3D::LoadObject(const char* filename, int ID)
{
	m_Objects[ID].Clear(true);
	m_Objects[ID].is_loaded=false;
	GFFileSystem fs;
	if (fs.OpenIn(filename))
	{
		char type[4];
		fs.Read(type, 4);
		fs.Close();
		
		if(type[0]==';' && type[1]=='D' && type[2]=='D' && type[3]=='W')
		{
			return Load_DDW(filename, m_Objects[ID], m_pRBOW->m_CookieColor);
		}
	}
	else
		return false; // can't read - no file no object

	m_Objects[ID].m_objects.SetSize(1);
	if (!m_Objects[ID].m_objects[0].obj.Load(filename))
	{
		m_Objects[ID].Clear(true);
		return false;
	}
	if (m_Objects[ID].m_objects[0].obj.NumNodes(0)==0)
	{
		m_Objects[ID].Clear(true);
		return false;
	}
	m_Objects[ID].is_loaded=true;
	m_Objects[ID].is_multiworld=false;

	m_Objects[ID].m_objects[0].obj.PrepareTriangles(); // For VBO rendering
//	PreRenderObject(ID);		// Display list
	m_Objects[ID].m_objects[0].obj.CalcMaxRadius();
return true;
}

//------------------------------------------------------------------//
// Free RAM memory of large objects. Will no longer X_ANIM, nor X_COLxx
//------------------------------------------------------------------//
void Rainbows3D::FreeExtraObject(int id)
{
	for(int i=0; i<m_Objects[id].m_objects.Size(); ++i)
	{
		MeshObjectX& obj = m_Objects[id].m_objects[i].obj;
		obj.m_ColTable.Clear(true);
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
		obj.m_Edges.Clear(true);
#endif
		if(obj.m_Frames.Size())
		{
			POLY_FRAME fr;
			fr = obj.m_Frames[0];
			fr.nodes.SetSize(3);
			fr.triangles.SetSize(1);

			obj.m_Frames.Clear(true);
			obj.m_Frames.SetAllocSteps(1,1);
			obj.m_Frames.Push(fr);
		}
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
		obj.m_isTriFacingLight.Clear(true);
#endif
		obj.m_tri_animation.Clear(true);
		obj.m_tri_animation.SetSize(3);
	}
}


//------------------------------------------------------------------//
// Save an Object in Rainbows-Format
//------------------------------------------------------------------//
bool Rainbows3D::SaveObject(const char* filename, int ID)
{
	m_Objects[ID].m_objects[0].obj.Save(filename);
return true;
}


void Rainbows3D::CullMode(int mode)
{
#ifdef TRENKI3D
	m_CullMode = (char)mode;
	switch(mode)
	{
	case 0:
		tglCullFace(GL_NONE);
		break;
	case -1:
		tglCullFace(GL_FRONT);
		break;
	case 1:
		tglCullFace(GL_BACK);
		break;
	}
	
#else
	m_CullMode = (char)mode;
	switch(mode)
	{
	case 0:
		glbDisable(GL_CULL_FACE);
		break;
	case -1:
		glbEnable(GL_CULL_FACE);
		glFrontFace(GL_CCW);
		break;
	case 1:
		glFrontFace(GL_CW);
		glbEnable(GL_CULL_FACE);
		break;
	}
#endif
}


// --------------------------------
// GLSL Support
// --------------------------------
#if defined(HAVE_OPENGL)
void Rainbows3D::FreeShader(int index)
{
	if(index<0 || index>=MAX_SHADERS || !glDeleteObjectArb) return;	
	if(m_Shaders[index][0])
		glDeleteObjectArb(m_Shaders[index][0]);
	if(m_Shaders[index][1])
		glDeleteObjectArb(m_Shaders[index][1]);
	if(m_Shaders[index][2])
		glDeleteObjectArb(m_Shaders[index][2]);
	m_Shaders[index][0] = m_Shaders[index][1] = m_Shaders[index][2] = 0;
}

// --------------------------------
// Check to see if it compiled.
// --------------------------------
bool Rainbows3D::VerifyShader(GLhandleARB shader, bool compiler, void(*printout)(const char*) )
{
GLsizei compiled=0, length;
GLenum status =  GL_OBJECT_LINK_STATUS_ARB;
	if(compiler)
		status = GL_OBJECT_COMPILE_STATUS_ARB;
	glGetObjectParameterivArb(shader, status, &compiled);
	if (compiled == 0)
	{
		if(printout)
		{
			glGetObjectParameterivArb(shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, &length);
			GLcharARB* log = new GLcharARB[length];
			glGetInfoLogArb(shader, length, &length, log);
			printout(log);
			delete[] log;
		}
		return FALSE;
	}
return TRUE;
}

// --------------------------------
// --------------------------------
bool Rainbows3D::LoadShader(int index, const char* szVertex, const char* szFragment, void(*printout)(const char*) )
{
	if(index<0 || index>=MAX_SHADERS)
		return FALSE;
	if(!glCreateShaderObjectArb || !glCompileShaderArb || !glLinkProgramArb)
	{
		static int no_shader = 1;
		if (printout && no_shader)
		{
			no_shader = 0;
			printout("No OpenGL 2.0 Shader Support");
		}
		return FALSE;
	}
	FreeShader(index);
	
bool good = false;
GFFileSystem fs;
char *pFrag=NULL, *pVert=NULL;
	if(szVertex && fs.OpenIn(szVertex))
	{
		int len = (int)fs.GetSize();
		pVert = new char[len+1];
		if(fs.Read(pVert, len))
		{
			pVert[len]='\0';
			m_Shaders[index][1] = glCreateShaderObjectArb(GL_VERTEX_SHADER_ARB);
			const GLcharARB* pA = pVert;
			glShaderSourceArb(m_Shaders[index][1], 1, &pA, NULL);
			glCompileShaderArb(m_Shaders[index][1]);
			if(!VerifyShader(m_Shaders[index][1], TRUE, printout))
			{
				if(printout)
				{
					printout("vertex shader failed");
				}
				good=false;
			}
			else
				good=true;
		}
	}

	if(szFragment && fs.OpenIn(szFragment))
	{
		int len = (int)fs.GetSize();
		pFrag = new char[len+1];
		if(fs.Read(pFrag, len))
		{
			pFrag[len]='\0';
			m_Shaders[index][2] = glCreateShaderObjectArb(GL_FRAGMENT_SHADER_ARB);
			const GLcharARB* pA = pFrag;
			glShaderSourceArb(m_Shaders[index][2], 1, &pA, NULL);
			glCompileShaderArb(m_Shaders[index][2]);
			if(!VerifyShader(m_Shaders[index][2], TRUE, printout))
			{
				if(printout)
				{
					printout("fragment shader failed");
				}
				good=false;
			}
			else
				good=true;
		}
	}
	
	if(good && (m_Shaders[index][1] || m_Shaders[index][2]))
	{
		m_Shaders[index][0] = glCreateProgramObjectArb();
		if(m_Shaders[index][1]) glAttachObjectArb(m_Shaders[index][0], m_Shaders[index][1]);
		if(m_Shaders[index][2]) glAttachObjectArb(m_Shaders[index][0], m_Shaders[index][2]);
		glLinkProgramArb(m_Shaders[index][0]);
		if(!VerifyShader(m_Shaders[index][0], FALSE, printout))
		{
			if(printout)
			{
				printout("link shader failed");
			}
			good=false;
		}
		else
			good=true;
	}
	if(pVert) delete[] pVert;
	if(pFrag) delete[] pFrag;
return good;
}




// --------------------------------
// --------------------------------
void Rainbows3D::SetShader(int index)
{
	if(!glUseProgramObjectArb) return;
	if(index<0 || index>=MAX_SHADERS)
	{
		glUseProgramObjectArb(0);
		m_BoundShaderIndex=-1;
		return;	
	}
	m_BoundShaderIndex=index;
	glUseProgramObjectArb(m_Shaders[index][0]);
}

// --------------------------------
// --------------------------------
void Rainbows3D::PutShaderUniform1f(const char* name, float val, void(*printout)(const char*))
{
	if(!glGetUniformLocationArb) return;
	if(m_BoundShaderIndex<0 || m_BoundShaderIndex>=MAX_SHADERS)
		return;	
	GLint pos = glGetUniformLocationArb(m_Shaders[m_BoundShaderIndex][0], name);
	if(pos!=-1)
		glUniform1fArb(pos, val);
	else
	{
		if(printout)
		{
			printout("undefined uniform:");
			printout(name);
		}
	}
}

// --------------------------------
// --------------------------------
void Rainbows3D::PutShaderUniform1i(const char* name, int val, void(*printout)(const char*))
{
	if(!glGetUniformLocationArb) return;
	if(m_BoundShaderIndex<0 || m_BoundShaderIndex>=MAX_SHADERS)
		return;	
	GLint pos = glGetUniformLocationArb(m_Shaders[m_BoundShaderIndex][0], name);
	if(pos!=-1)
		glUniform1iArb(pos, val);
	else
	{
		if(printout)
		{
			printout("undefined uniform:");
			printout(name);
		}
	}
}

// --------------------------------
// --------------------------------
void Rainbows3D::PutShaderAttrib1f(const char* name, float val, void(*printout)(const char*))
{
	if(!glUseProgramObjectArb) return;
	if(m_BoundShaderIndex<0 || m_BoundShaderIndex>=MAX_SHADERS)
		return;	
	GLint pos = glGetAttribLocationArb(m_Shaders[m_BoundShaderIndex][0], name);
	if(pos!=-1)
		glVertexAttrib1fArb(pos, val);
	else
	{
		if(printout)
		{
			printout("undefined attrib:");
			printout(name);
		}
	}
}

#endif // #if defined(HAVE_OPENGL)



// --------------------------------
// For Frustum Culling
// --------------------------------
void Rainbows3D::ExtractFrustum()
{
#ifndef TRENKI3D
   float   proj[16];
   float   modl[16];
   float   clip[16];
   float   t;

   /* Get the current PROJECTION matrix from OpenGL */
   gCopyMatrix(proj, m_ProjectionMatrix);
   // glGetFloatv( GL_PROJECTION_MATRIX, proj );


   /* Get the current MODELVIEW matrix from OpenGL */
   // float m2[16], m3[16];
   gCopyMatrix(modl, m_ModelViewCameraMatrix);
   // gCopyMatrix(m3, this->GetFinalMatrix());
   // glGetFloatv( GL_MODELVIEW_MATRIX, modl );

 
   /* Combine the two matrices (multiply projection by modelview) */
   clip[ 0] = modl[ 0] * proj[ 0] + modl[ 1] * proj[ 4] + modl[ 2] * proj[ 8] + modl[ 3] * proj[12];
   clip[ 1] = modl[ 0] * proj[ 1] + modl[ 1] * proj[ 5] + modl[ 2] * proj[ 9] + modl[ 3] * proj[13];
   clip[ 2] = modl[ 0] * proj[ 2] + modl[ 1] * proj[ 6] + modl[ 2] * proj[10] + modl[ 3] * proj[14];
   clip[ 3] = modl[ 0] * proj[ 3] + modl[ 1] * proj[ 7] + modl[ 2] * proj[11] + modl[ 3] * proj[15];

   clip[ 4] = modl[ 4] * proj[ 0] + modl[ 5] * proj[ 4] + modl[ 6] * proj[ 8] + modl[ 7] * proj[12];
   clip[ 5] = modl[ 4] * proj[ 1] + modl[ 5] * proj[ 5] + modl[ 6] * proj[ 9] + modl[ 7] * proj[13];
   clip[ 6] = modl[ 4] * proj[ 2] + modl[ 5] * proj[ 6] + modl[ 6] * proj[10] + modl[ 7] * proj[14];
   clip[ 7] = modl[ 4] * proj[ 3] + modl[ 5] * proj[ 7] + modl[ 6] * proj[11] + modl[ 7] * proj[15];

   clip[ 8] = modl[ 8] * proj[ 0] + modl[ 9] * proj[ 4] + modl[10] * proj[ 8] + modl[11] * proj[12];
   clip[ 9] = modl[ 8] * proj[ 1] + modl[ 9] * proj[ 5] + modl[10] * proj[ 9] + modl[11] * proj[13];
   clip[10] = modl[ 8] * proj[ 2] + modl[ 9] * proj[ 6] + modl[10] * proj[10] + modl[11] * proj[14];
   clip[11] = modl[ 8] * proj[ 3] + modl[ 9] * proj[ 7] + modl[10] * proj[11] + modl[11] * proj[15];

   clip[12] = modl[12] * proj[ 0] + modl[13] * proj[ 4] + modl[14] * proj[ 8] + modl[15] * proj[12];
   clip[13] = modl[12] * proj[ 1] + modl[13] * proj[ 5] + modl[14] * proj[ 9] + modl[15] * proj[13];
   clip[14] = modl[12] * proj[ 2] + modl[13] * proj[ 6] + modl[14] * proj[10] + modl[15] * proj[14];
   clip[15] = modl[12] * proj[ 3] + modl[13] * proj[ 7] + modl[14] * proj[11] + modl[15] * proj[15];

   /* Extract the numbers for the RIGHT plane */
   m_frustum[0][0] = clip[ 3] - clip[ 0];
   m_frustum[0][1] = clip[ 7] - clip[ 4];
   m_frustum[0][2] = clip[11] - clip[ 8];
   m_frustum[0][3] = clip[15] - clip[12];

   /* Normalize the result */
   t = sqrtf( m_frustum[0][0] * m_frustum[0][0] + m_frustum[0][1] * m_frustum[0][1] + m_frustum[0][2] * m_frustum[0][2] );
   m_frustum[0][0] /= t;
   m_frustum[0][1] /= t;
   m_frustum[0][2] /= t;
   m_frustum[0][3] /= t;

   /* Extract the numbers for the LEFT plane */
   m_frustum[1][0] = clip[ 3] + clip[ 0];
   m_frustum[1][1] = clip[ 7] + clip[ 4];
   m_frustum[1][2] = clip[11] + clip[ 8];
   m_frustum[1][3] = clip[15] + clip[12];

   /* Normalize the result */
   t = sqrtf( m_frustum[1][0] * m_frustum[1][0] + m_frustum[1][1] * m_frustum[1][1] + m_frustum[1][2] * m_frustum[1][2] );
   m_frustum[1][0] /= t;
   m_frustum[1][1] /= t;
   m_frustum[1][2] /= t;
   m_frustum[1][3] /= t;

   /* Extract the BOTTOM plane */
   m_frustum[2][0] = clip[ 3] + clip[ 1];
   m_frustum[2][1] = clip[ 7] + clip[ 5];
   m_frustum[2][2] = clip[11] + clip[ 9];
   m_frustum[2][3] = clip[15] + clip[13];

   /* Normalize the result */
   t = sqrtf( m_frustum[2][0] * m_frustum[2][0] + m_frustum[2][1] * m_frustum[2][1] + m_frustum[2][2] * m_frustum[2][2] );
   m_frustum[2][0] /= t;
   m_frustum[2][1] /= t;
   m_frustum[2][2] /= t;
   m_frustum[2][3] /= t;

   /* Extract the TOP plane */
   m_frustum[3][0] = clip[ 3] - clip[ 1];
   m_frustum[3][1] = clip[ 7] - clip[ 5];
   m_frustum[3][2] = clip[11] - clip[ 9];
   m_frustum[3][3] = clip[15] - clip[13];

   /* Normalize the result */
   t = sqrtf( m_frustum[3][0] * m_frustum[3][0] + m_frustum[3][1] * m_frustum[3][1] + m_frustum[3][2] * m_frustum[3][2] );
   m_frustum[3][0] /= t;
   m_frustum[3][1] /= t;
   m_frustum[3][2] /= t;
   m_frustum[3][3] /= t;

   /* Extract the FAR plane */
   m_frustum[4][0] = clip[ 3] - clip[ 2];
   m_frustum[4][1] = clip[ 7] - clip[ 6];
   m_frustum[4][2] = clip[11] - clip[10];
   m_frustum[4][3] = clip[15] - clip[14];

   /* Normalize the result */
   t = sqrtf( m_frustum[4][0] * m_frustum[4][0] + m_frustum[4][1] * m_frustum[4][1] + m_frustum[4][2] * m_frustum[4][2] );
   m_frustum[4][0] /= t;
   m_frustum[4][1] /= t;
   m_frustum[4][2] /= t;
   m_frustum[4][3] /= t;

   /* Extract the NEAR plane */
   m_frustum[5][0] = clip[ 3] + clip[ 2];
   m_frustum[5][1] = clip[ 7] + clip[ 6];
   m_frustum[5][2] = clip[11] + clip[10];
   m_frustum[5][3] = clip[15] + clip[14];

   /* Normalize the result */
   t = sqrtf( m_frustum[5][0] * m_frustum[5][0] + m_frustum[5][1] * m_frustum[5][1] + m_frustum[5][2] * m_frustum[5][2] );
   m_frustum[5][0] /= t;
   m_frustum[5][1] /= t;
   m_frustum[5][2] /= t;
   m_frustum[5][3] /= t;
#endif
}

// --------------------------------
// Quick Frustum culling code
// --------------------------------
bool Rainbows3D::SphereInFrustum(float x, float y, float z, float radius )
{
#ifdef TRENKI3D
	return true;
#else
	if(m_frustum[0][0]*x + m_frustum[0][1]*y + m_frustum[0][2]*z + m_frustum[0][3] < -radius) return false;
	if(m_frustum[1][0]*x + m_frustum[1][1]*y + m_frustum[1][2]*z + m_frustum[1][3] < -radius) return false;
	if(m_frustum[2][0]*x + m_frustum[2][1]*y + m_frustum[2][2]*z + m_frustum[2][3] < -radius) return false;
	if(m_frustum[3][0]*x + m_frustum[3][1]*y + m_frustum[3][2]*z + m_frustum[3][3] < -radius) return false;
	if(m_frustum[4][0]*x + m_frustum[4][1]*y + m_frustum[4][2]*z + m_frustum[4][3] < -radius) return false;
	if(m_frustum[5][0]*x + m_frustum[5][1]*y + m_frustum[5][2]*z + m_frustum[5][3] < -radius) return false;
	return true;
#endif
}

#else // console
	void glb_rainbows3d_console_dummy() {};
#endif // GLB_CONSOLE
