// Rainbows3D.h: Schnittstelle fr die Klasse Rainbows3D.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_RAINBOWS3D_H__A54DCD42_E1B4_11D5_B6EB_003042FFFD01__INCLUDED_)
#define AFX_RAINBOWS3D_H__A54DCD42_E1B4_11D5_B6EB_003042FFFD01__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#define GLR_MAX_OBJECTS		0x1000
#define MAX_ROTATIONS		16
#define MAX_SHADERS			0xff

#include "3dmath.h"
#include "meshobject.h"

// MeshObject Extension with what you don't need for import/export filter
class MeshObjectX : public MeshObject
{
public:
	MeshObjectX():		 MeshObject(), m_NeedColoring(0)
						#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
						, m_VBO_id(0), m_VBO_id_morphing(0)
						, CelShadeTexture(0)
						#endif
						{
						}
	~MeshObjectX()		{Clear(true);}


	int					m_NeedColoring; // Or is every used color white?
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
	unsigned int		m_VBO_id; // VBO for static frame 0 - in AGP memory! Muhahaha
	unsigned int		m_VBO_id_morphing; // VBO for dnamic(temp) frame
#endif

#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
	unsigned int		CelShadeTexture; // One line for each color
	float				CelShadeColorScale;

	struct EDGE
	{
		int neighbourIndex; // index of triangle in m_tri_animation
		bool isSilhouetteEdge;
	};
	Queue<EDGE> m_Edges;            // for each triangle in m_Triangles 3 edges. v0->v1, v1->v2, v2-v0
	Queue<char> m_isTriFacingLight; // is triangle[i] (index/3) facing the light(0,1) (for shadow volumes)
#endif

	void Clear(bool free_mem);
	void PrepareTriangles();
	void CalcMaxRadius();
	void MakeCelShadeTexutre(bool bgra_ext);
	void CreateEdges();
	void Scale(float sx,float sy, float sz);

#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
	void CalculateSilhouetteEdges(const POLY_NODE* pNodes, float* lightPosition); // frame=-1 for m_tri_anim
	void DrawInfiniteShadowVolume(const POLY_NODE* pNodes, float* lightPosition, bool drawCaps);
#endif
	// Does the model need coloring
	inline void CheckIfColoringNeeded()
						{
							m_NeedColoring = 0;
							int col, sz = m_ColTable.Size();
							for (int i=0; i<sz; i++)
							{
								col = *m_ColTable.GetAt(i);
								if (col<COL_NEW_ANY && col!=0x00ffffff)
								{
									m_NeedColoring=1;
									return;
								}
							}
						}
};


class X_WORLD
{
public:
	struct DATA
	{
		MeshObjectX obj;
		int gl_tex, gl_tex2;
	};
	X_WORLD()					{Clear(true);}
	~X_WORLD()					{Clear(true);}
	void Clear(bool free_mem)	{
									for(int i=0; i<(int)m_objects.Size(); ++i)
										m_objects[i].obj.Clear(free_mem);
									m_objects.Clear(free_mem);
									is_loaded=false; is_multiworld=false;
									#ifndef TRENKI3D
									if(m_textures.Size())
										glDeleteTextures(m_textures.Size(), (GLuint*) &m_textures[0]);
									#endif
									m_textures.Clear(free_mem);
									m_objects.SetAllocSteps(1,1);
									m_textures.SetAllocSteps(1,1);
								}
	void Scale(float sx, float sy, float sz) {for(int i=0; i<(int)m_objects.Size(); ++i)m_objects[i].obj.Scale(sx,sy,sz);}
	bool			is_loaded;
	bool			is_multiworld;

	Queue<DATA>		m_objects;
	Queue<unsigned int>		m_textures;
};


int   CollisionGetLastCollisionTriangleIndex();
class OpenGLRainbows;
//------------------------------------------------------------------//
//- 3D OpenGL Helper Class    --------------------------------------//
//------------------------------------------------------------------//
class Rainbows3D
{
// Member Variables...
	public:
		Queue<X_WORLD> m_Objects;
	private:

	int	m_iCurObjFrame; // when creating object


// Class Functions
public:
	// Setup
	Rainbows3D(OpenGLRainbows* pR);
	virtual ~Rainbows3D();
	void ClearZBuffer();

private:
	bool  CollisionAABB(MeshObjectX& the_obj, int frame, const float *C1, const float* C2, const float* C3, const float* C4);
	bool  CollisionSphere(MeshObjectX& the_obj, int frame, float radius, float x, float y, float z);
	float CollisionRay(MeshObjectX& the_obj, int frame, float x, float y, float z, float dirx, float diry, float dirz);
	float CollisionCone(MeshObjectX& the_obj, int frame, float x, float y, float z, float dirx, float diry, float dirz, float angle);
public:
	bool  CollisionAABB(int id, int frame, const float *C1, const float* C2, const float* C3, const float* C4);
	bool  CollisionSphere(int id, int frame, float radius, float x, float y, float z);
	float CollisionRay(int id, int frame, float x, float y, float z, float dirx, float diry, float dirz);
	float CollisionCone(int id, int frame, float x, float y, float z, float dirx, float diry, float dirz, float angle);
	bool  CollisionSphereBox(float x, float y, float z, float r, float bx, float by, float bz); // sphere/AABB collision. AABB in current matrix transformation
	


	// Rendering
	void Fog(unsigned long col=-1, int exp_mode=0, float p1=0, float p2=0);
	void Make3DViewport(float zNear=1.0, float zFar=1000.0, float fov=45.0);
	void SetCameraUp(float x, float y, float z);
	void SetCamera(float px, float py, float pz, float rx, float ry, float rz);
	void RenderAnim(int num, int from, int to, float percent=0.0f, bool fullAnim=true);

	// frustum culling
	float m_frustum[6][4];
	void ExtractFrustum();
	bool SphereInFrustum(float x, float y, float z, float radius );


	// GLSL
#if defined(HAVE_OPENGL)
	private:
		unsigned int m_Shaders[MAX_SHADERS][3]; // [id][prog,vert,frag]
		int m_BoundShaderIndex;
	public:
		bool IsShaderActive()				{return m_BoundShaderIndex>=0 ? true:false;}
		void SetShader(int index);
		void FreeShader(int index);
		bool LoadShader(int index, const char* szVertex, const char* szFragment, void(*printout)(const char*)=NULL);
		void PutShaderUniform1f(const char* name, float val, void(*printout)(const char*)=NULL);
		void PutShaderUniform1i(const char* name, int val, void(*printout)(const char*)=NULL);
		void PutShaderAttrib1f(const char* name, float val, void(*printout)(const char*)=NULL);
	private:
		bool VerifyShader(GLhandleARB shader, bool compiler, void(*printout)(const char*));
#else
		inline bool IsShaderActive()				{return false;}
		inline void SetShader(int) {}
		inline void FreeShader(int){}
		inline bool LoadShader(int, const char*, const char*, void(*)(const char*)=NULL) {return false;}
		inline void PutShaderUniform1f(const char* name, float val, void(*printout)(const char*)=NULL) {}
		inline void PutShaderUniform1i(const char* name, int val, void(*printout)(const char*)=NULL) {}
		inline void PutShaderAttrib1f(const char* name, float val, void(*printout)(const char*)=NULL)  {}
#endif

private:
	#ifdef TRENKI3D
	void RenderMD2_Trenki(MeshObjectX& the_obj, int from, int to, float percent, bool fullAnim);
	#else
	void RenderMD2       (MeshObjectX& the_obj, int from, int to, float percent, bool fullAnim, GLuint* pTextures=NULL);
	void RenderMD2_TinyGL(MeshObjectX& the_obj, int from, int to, float percent, bool fullAnim, GLuint* pTextures=NULL);
	#endif
public:
	void DrawCoordAxis(float x, float y, float z);

	// User Objects
	POLY_NODE	m_UserObjectNode;
	void UserObjectStart(int num, int type=1);
	void UserObjectSetFrame(int frm);
	void UserObjectNewGroup(int type=1);
	void UserObjectSetNormal(float x, float y, float z); // if you don't do auto-normals, you can use these before you add a vertex
	void UserObjectAddVertex(float x, float y, float z, float tx, float ty, unsigned long col);
	void UserObjectEnd(void);
//	void PreRenderObject(int ID);

	// Modes
	void CullMode(int mode);
	void DrawSphereGrid(float*org, float R, unsigned long col, short width = 5);
	void DrawDot(const float *pos, unsigned long col, float size);
	void DrawLine(const float* P1, const float* P2, unsigned long col, float width = 5);
	void DrawAABB(const float *C1, const float* C2, const float* C3, const float* C4, unsigned long col, short width = 5);
	void UserObjectAutoNormalMode(long mode=2)				{m_AutoNormalMode = mode%3;}
	void SetCelShades(unsigned char* pScales, int nscales);


	void DoShadows(bool mode); // true: start shadow objects. false: flush and draw shadows
	bool m_shadowmode;

	// Lights
	void LightsOff(void) {m_LightsOn=0; m_BumpLight.active=0;}
	void SpotLt(int num, unsigned long col, float x, float y, float z,
							float px, float py, float pz, float cutoff, int ShowDebug=0);
	void AmbientLt(int num, unsigned long col);

	void SetDepthTesting(bool enable);


	// File IO
	bool LoadObject(const char* filename, int ID); // Rainbow-Objects
	void FreeExtraObject(int id); // free RAM memory of object, be able to draw but no longer x_col and x_anim!
	bool SaveObject(const char* filename, int ID);

	// Inline Transformations
	inline void SetRotation(float angle, float x, float y, float z)
	{
		m_Transformations[m_nTransformations].m_IsGoodMatrix=0;
		if(m_Transformations[m_nTransformations].m_nRotate<MAX_ROTATIONS)
		{
			float* pR = m_Transformations[m_nTransformations].m_Rotate[m_Transformations[m_nTransformations].m_nRotate];
			pR[0]=angle; pR[1]=x; pR[2]=y; pR[3]=z;
			m_Transformations[m_nTransformations].m_nRotate++;
		}
	}
	#define MAT(m,x,y) m[(y)+(x)*4]
	inline void SetScaling(float sx, float sy, float sz)
	{
		if(m_nTransformations)
		{
			gCopyMatrix(m_Transformations[m_nTransformations].m_Matrix,
						m_Transformations[m_nTransformations-1].m_Matrix);
		}
		else
		{
			gIdentity(m_Transformations[m_nTransformations].m_Matrix);
		}
	
		m_Transformations[m_nTransformations].m_IsGoodMatrix=0;
		m_Transformations[m_nTransformations].m_nRotate=0;
		m_Transformations[m_nTransformations].m_nMult=0;
		m_Transformations[m_nTransformations].m_Scale[0]=sx;
		m_Transformations[m_nTransformations].m_Scale[1]=sy;
		m_Transformations[m_nTransformations].m_Scale[2]=sz;
	}
	inline void SetTranslation(float x, float y, float z)
	{
		if(m_nTransformations)
		{
			gCopyMatrix(m_Transformations[m_nTransformations].m_Matrix,
						m_Transformations[m_nTransformations-1].m_Matrix);
		}
		else
		{
			gIdentity(m_Transformations[m_nTransformations].m_Matrix);
		}
	
		m_Transformations[m_nTransformations].m_IsGoodMatrix=0;
		m_Transformations[m_nTransformations].m_nRotate=0;
		m_Transformations[m_nTransformations].m_nMult=0;
		m_Transformations[m_nTransformations].m_Translate[0]=x;
		m_Transformations[m_nTransformations].m_Translate[1]=y;
		m_Transformations[m_nTransformations].m_Translate[2]=z;
	}
	inline bool GetMatrixScalings(float& sx, float& sy, float&sz)
	{
		BuildMatrix();
		float* pM = m_Transformations[m_nTransformations].m_Matrix;
		if(		MAT(pM,0,0)==1 && MAT(pM,1,0)==0 && MAT(pM,2,0)==0
			&&	MAT(pM,0,1)==0 && MAT(pM,1,1)==1 && MAT(pM,2,1)==0
			&&	MAT(pM,0,2)==0 && MAT(pM,1,2)==0 && MAT(pM,2,2)==1)
		{
			sx=sy=sz=1.0f;
			return false;
		}
		else
		{
			// Juck! Get the scalings now
			sx = sqrtf(MAT(pM,0,0)*MAT(pM,0,0) + MAT(pM,0,1)*MAT(pM,0,1) + MAT(pM,0,2)*MAT(pM,0,2));
			sy = sqrtf(MAT(pM,1,0)*MAT(pM,1,0) + MAT(pM,1,1)*MAT(pM,1,1) + MAT(pM,1,2)*MAT(pM,1,2));
			sz = sqrtf(MAT(pM,2,0)*MAT(pM,2,0) + MAT(pM,2,1)*MAT(pM,2,1) + MAT(pM,2,2)*MAT(pM,2,2));
			return true;
		}
	}
	

	inline void SetMultMatrix(const float* m)
	{
		m_Transformations[m_nTransformations].m_IsGoodMatrix=0;
		gCopyMatrix(m_Transformations[m_nTransformations].m_Mults[m_Transformations[m_nTransformations].m_nMult++], m);
	}

	// return pointer to mat[0] for final computed matrix
	const float* GetFinalMatrix()
	{
		BuildMatrix();
		return m_Transformations[m_nTransformations].m_Matrix;
	}

	// pointer to mat[0] of model view camera matrix
	const float* GetCameraMatrix()	{return m_ModelViewCameraMatrix;}


	#undef MAT
	
	void PushTransformation()
	{
		if(m_nTransformations>=15)
			return;

		BuildMatrix(); // a must - the new one is based on this
		TRANSFORMATION& tr = m_Transformations[ ++m_nTransformations];
		tr.clear(false);
		gCopyMatrix(tr.m_Matrix,
					m_Transformations[m_nTransformations-1].m_Matrix);
	}
	bool PopTransformation()
	{
		if(m_nTransformations>0) {--m_nTransformations; return true;}
		m_Transformations[0].clear();
		return false;
	}
	void BuildMatrix();





	void SetTextureOffset(float du, float dv);
	void SphereMapping(bool onoff) {m_SphereMapping = onoff;}

private:
	char m_CullMode;			// 0 None, +1 Clockwise, -1, CounterClockwise
	// Transformations
	void	TransformMatrix();

	struct TRANSFORMATION
	{
		void	clear(bool identity=true)	{
							m_nMult=0; m_nRotate=0;
							m_Scale[0]=m_Scale[1]=m_Scale[2]=1.0f;
							m_Translate[0]=m_Translate[1]=m_Translate[2]=0.0f;
							m_IsGoodMatrix=1;
							if(identity)
							{
								gIdentity(m_Matrix);
							}
						}
		int		m_nMult;
		int		m_nRotate;
		float	m_Rotate[MAX_ROTATIONS][4];
		float	m_Scale[3];
		float	m_Translate[3];
		float	m_Mults[MAX_ROTATIONS][16];
		float	m_Matrix[16];
		int		m_IsGoodMatrix;
	}m_Transformations[16];
	int m_nTransformations;
	
	
public:
	float   m_ProjectionMatrix[16];    // GL_PROJECTION_MATRIX   
	float	m_ModelViewCameraMatrix[16]; // GL_MODELVIEW_MATRIX after LookAt
private:
	float	m_InvCameraMatrix[16]; // Inverted GL_MODELVIEW_MATRIX after LookAt
	float	m_TexOffset[2];
	unsigned char	m_LightsOn; // Bit pattern for active GL_LIGHTs
	bool	m_SphereMapping;
	OpenGLRainbows*	m_pRBOW;
	struct BUMP_LIGHT_DATA
	{
		bool	active;
		char	type;			// 1 bump, 2 cartoon, 3 shadow
		float	pos[3];
		unsigned long dwcol; // colour as RGB() value
	} m_BumpLight;
	struct CAMERA_DATA
	{
		float up[3];
		float eye[3];
		float lookat[3];
		float tan_fov;			// tangens of fov
		float znear;
		float zfar;
	} m_Camera;
	struct TWO_FLOATS
	{
		float f[2];
	};
	long	m_AutoNormalMode; // 0=none, 1=flat, 2=surface normals
};

#endif // !defined(AFX_RAINBOWS3D_H__A54DCD42_E1B4_11D5_B6EB_003042FFFD01__INCLUDED_)
