#import <QuartzCore/QuartzCore.h>
#import <OpenGLES/EAGLDrawable.h>
#import <sys/utsname.h>
#import "iPhoneEAGLView.h"

#ifdef __cplusplus
extern "C" {
#endif

char** gpArgv=NULL;
int    gArgc=0;
int    gScaleFactor = 1; // 1 lowres, 2 retina
int    _isActive = 0;
EAGLView* gpView=NULL;

void iPhoneMain(int argc, char* argv[]);
void iPhonePumpEvents(unsigned long millisecs);

void __GLB_Defaults();
void __EndProgram(void);
void glbasic_notify_screen_change(int w, int h);
void glbasic_get_screen_size(int* w, int* h);
int iphone_get_dpi();

#ifdef GLB_IN_A_THREAD
	HThreadID gMainThreadID = NULL;
	HTmutex gMainMutex;

	void iPhoneMainThread(void* pVoid)
	{
		NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];

		iPhoneMain(gArgc, gpArgv);
		__EndProgram();

		[pool release];
	}

#endif




#ifdef __cplusplus
} // extern "C"
#endif


/* A class extension to declare private methods */
@interface EAGLView ()

@property (nonatomic, retain) EAGLContext *context;
@property (nonatomic, assign) NSTimer *animationTimer;

- (BOOL) createFramebuffer;
- (void) destroyFramebuffer;

@end


@implementation EAGLView

@synthesize context;
@synthesize animationTimer;
@synthesize animationInterval;

float gMouseScale=1.0f;

/* You must implement this method */
+ (Class)layerClass {
    return [CAEAGLLayer class];
}

- (id) initWithFrame:(CGRect)frame
{
	printf("init with frame.\n");

	[self initWithFrame:frame pixelFormat:GL_RGBA // statt GL_RGB565_OES __SEE_565_IPHONE__
						depthFormat:GL_DEPTH_COMPONENT24_OES preserveBackbuffer:YES];
	return self;
}

- (id) initWithFrame:(CGRect)frame pixelFormat:(GLuint)format 
{
	return [self initWithFrame:frame pixelFormat:format depthFormat:0 preserveBackbuffer:YES];
}


- (id) initWithFrame:(CGRect)frame pixelFormat:(GLuint)format depthFormat:(GLuint)depth preserveBackbuffer:(BOOL)retained
{
#ifdef GLB_IN_A_THREAD
	htMutexInit(&gMainMutex);
#endif
	gpView = self;
	if((self = [super initWithFrame:frame])) {
		//if ((self = [super initWithCoder:coder])) {
		/* Get the layer */
		CAEAGLLayer *eaglLayer = (CAEAGLLayer *)[self layer];

		eaglLayer.opaque = YES;
		eaglLayer.drawableProperties = [NSDictionary dictionaryWithObjectsAndKeys:
										[NSNumber numberWithBool:NO],
										[NSNumber numberWithBool:NO], /* GF: kEAGLDrawablePropertyRetainedBacking, */
										kEAGLColorFormatRGBA8, /* GF: statt  kEAGLColorFormatRGB565  __SEE_565_IPHONE__ */
										kEAGLDrawablePropertyColorFormat, nil];

		context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];

		if (!context || ![EAGLContext setCurrentContext:context]) {
			[self release];
			return nil;
		}

		animationInterval = 1.0 / 4.0;
	}
	return self;
}



- (void)drawView
{
static int bRunning = 0;
	if(!bRunning)
	{
		bRunning = 1;
		#ifdef GLB_IN_A_THREAD
				gMainThreadID = htThreadCreate(iPhoneMainThread, NULL, 0);
		#else
			iPhoneMain(gArgc, gpArgv);
			__EndProgram();
		#endif
	}
}

-(void) swapBuffers
{
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, viewRenderbuffer);
    [context presentRenderbuffer:GL_RENDERBUFFER_OES];
    [EAGLContext setCurrentContext:context];
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, viewFramebuffer);
}


- (void)layoutSubviews {
    [EAGLContext setCurrentContext:context];
    [self destroyFramebuffer];
    [self createFramebuffer];
//    [self drawView];
}


- (BOOL)createFramebuffer {
	// *** RETINA DISPLAY *** UIView scaling
	__GLB_Defaults(); // must set what the project options say

	int dg_resx=0, dg_resy=0;
	glbasic_get_screen_size(&dg_resx, &dg_resy);
	if(dg_resx == 0)
	{
		printf("__GLB_Defaults was NOT CALLED!\n");
		*(int*)0=0;
	}

    // Space Fractal, this is now required, due various resolutions and now longer check glbasic ide max defaults.
    // You will end to have a window on some devices if we not just using the max resolution.
	if([[UIScreen mainScreen] respondsToSelector: NSSelectorFromString(@"scale")])
	{
		if([self respondsToSelector: NSSelectorFromString(@"contentScaleFactor")])
		{
			float use_scale = [[UIScreen mainScreen] scale];
			gScaleFactor = use_scale == 2.0f ? 2:1;
            self.contentScaleFactor = use_scale;
			gMouseScale = use_scale;
			printf("scaleing available: %f\n", use_scale);
		}
	}

	CAEAGLLayer*			eaglLayer = (CAEAGLLayer*)[self layer];
	CGSize					newSize;
	newSize = [eaglLayer bounds].size;
	newSize.width = roundf(newSize.width);
	newSize.height = roundf(newSize.height);

    backingWidth = newSize.width;
	backingHeight= newSize.height;
	
	glGenFramebuffersOES(1, &viewFramebuffer);
    glGenRenderbuffersOES(1, &viewRenderbuffer);
    
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, viewFramebuffer);
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, viewRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:(CAEAGLLayer*) [self layer] ];
    glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, viewRenderbuffer);
    
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES,  &backingWidth);
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight);
    

	glGenRenderbuffersOES(1, &depthRenderbuffer);
        glBindRenderbufferOES(GL_RENDERBUFFER_OES, depthRenderbuffer);
        glRenderbufferStorageOES(GL_RENDERBUFFER_OES, GL_DEPTH_COMPONENT24_OES, backingWidth, backingHeight);
        glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_DEPTH_ATTACHMENT_OES, GL_RENDERBUFFER_OES, depthRenderbuffer);

	printf("desktop backing: %d x %d\n", backingWidth, backingHeight);

	dg_resx = backingWidth;
	dg_resy = backingHeight;
	printf("mainScreen: %dx%d\n", dg_resx, dg_resy);
	glbasic_notify_screen_change(dg_resx, dg_resy);

    if(glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES) {
        NSLog(@"failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
        return NO;
    }

	// pre-flip
	[EAGLContext setCurrentContext:context];
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, viewFramebuffer);

    return YES;
}


- (void)destroyFramebuffer {
    
    glDeleteFramebuffersOES(1, &viewFramebuffer);
    viewFramebuffer = 0;
    glDeleteRenderbuffersOES(1, &viewRenderbuffer);
    viewRenderbuffer = 0;
    
    if(depthRenderbuffer) {
        glDeleteRenderbuffersOES(1, &depthRenderbuffer);
        depthRenderbuffer = 0;
    }
}


- (void)startAnimation {
 self.animationTimer = [NSTimer scheduledTimerWithTimeInterval:animationInterval target:self selector:@selector(drawView) userInfo:nil repeats:NO];
}


- (void)stopAnimation {
    self.animationTimer = nil;
}


- (void)setAnimationTimer:(NSTimer *)newTimer {
    [animationTimer invalidate];
    animationTimer = newTimer;
}


- (void)setAnimationInterval:(NSTimeInterval)interval {
    
    animationInterval = interval;
    if (animationTimer) {
        [self stopAnimation];
        [self startAnimation];
    }
}


- (void)dealloc {
    
    [self stopAnimation];
    
    if ([EAGLContext currentContext] == context) {
        [EAGLContext setCurrentContext:nil];
    }
    
    [context release];  
    [super dealloc];
}

- (void)didEnterBackground {
    printf("didEnterBackground\n");
    if (_isActive)
        [self resignFirstResponder];
}

- (void)didBecomeActive {
    printf("didBecomeActive\n");
    if (_isActive)
        [self becomeFirstResponder];
}

- (BOOL)canBecomeFirstResponder {
    return YES;
}

- (void)setActive:(BOOL)value {
    printf("setActive\n");
    if (_isActive == value) return;
    _isActive = value;
    if (_isActive) {
        [self becomeFirstResponder];
    } else {
        [self resignFirstResponder];
    }
}

// MultiTouch
- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event {
	/* associate touches with mice, so long as we have slots */
	int i;
#ifdef GLB_IN_A_THREAD
	if(htMutexLock(&gMainMutex)==0)
#endif
	{
		for (UITouch *touch in touches)
		{
			struct GLBMOUSE* pMouse = &mice[0]; --pMouse;
			for(i=0; touch && i < MAX_SIMULTANEOUS_TOUCHES; i++)
			{
				++pMouse;
				/* check if this mouse is already tracking a touch */
				if (pMouse->driverdata != NULL) {
					continue;
				}
				/*
					mouse not associated with anything right now,
					associate the touch with this mouse
				*/
				
				/* select this slot's mouse */
				pMouse->driverdata = touch;

				CGPoint locationInView = [touch locationInView: self];
				locationInView.x *= gMouseScale;
				locationInView.y *= gMouseScale;
						
				/* send moved event */
				pMouse->x = (int)locationInView.x;
				pMouse->y = (int)locationInView.y;
				pMouse->button=1;
				/* grab next touch */
				// touch = (UITouch*)[enumerator nextObject];
				break;
			}
		}

#ifdef GLB_IN_A_THREAD
		htMutexUnlock(&gMainMutex);
#endif
	}
    [super touchesEnded: touches withEvent: event];
}

- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event {
#ifdef GLB_IN_A_THREAD
	if(htMutexLock(&gMainMutex)==0)
#endif
	{
		for (UITouch *touch in touches)
		{
		
			int i;
			/* search for the mouse slot associated with this touch */
			struct GLBMOUSE* pMouse = &mice[0];
			for (i=0; i<MAX_SIMULTANEOUS_TOUCHES; i++)
			{
				if (touch == pMouse->driverdata) // mice[i].button && mice[i].x==prevlocationInView.x && mice[i].y==prevlocationInView.y)
				{
					/* found the mouse associate with the touch */
					/* send mouse up */
					pMouse->button=0;
					pMouse->driverdata=NULL;
					break;
				}
				++pMouse;
			}

		}
#ifdef GLB_IN_A_THREAD
		htMutexUnlock(&gMainMutex);
#endif
	}
    [super touchesEnded: touches withEvent: event];
}

- (void)touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event {
	/*
		this can happen if the user puts more than 5 touches on the screen
		at once, or perhaps in other circumstances.  Usually (it seems)
		all active touches are canceled.
	*/
	[self touchesEnded: touches withEvent: event];
}

- (void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event {
	for (UITouch *touch in touches)
	{
		/* try to find the mouse associated with this touch */
		int i, found=0;
		/* search for the mouse slot associated with this touch */
#ifdef GLB_IN_A_THREAD
		if(htMutexLock(&gMainMutex)==0)
#endif
		{
			struct GLBMOUSE* pMouse = &mice[0];
			for (i=0; i<MAX_SIMULTANEOUS_TOUCHES; i++)
			{
				if (pMouse->driverdata==touch) // mice[i].button && mice[i].x==prevlocationInView.x && mice[i].y==prevlocationInView.y)
				{
					/* found proper mouse */
					CGPoint locationInView = [touch locationInView: self];
					
					locationInView.x *= gMouseScale;
					locationInView.y *= gMouseScale;
					
					/* send moved event */
					pMouse->x=locationInView.x;
					pMouse->y=locationInView.y;

					/* discontinue search */
					found=1;
					break;
				}
				++pMouse;
			}
			
			// oh - make all invalid
			if(!found)
			{
				struct GLBMOUSE* pMouse = &mice[0];
				for (i=0; i<MAX_SIMULTANEOUS_TOUCHES; i++)
				{
					pMouse->button=0; ++pMouse;
				}
			}
#ifdef GLB_IN_A_THREAD
			htMutexUnlock(&gMainMutex);
#endif
		}
	}
    [super touchesEnded: touches withEvent: event];

}

@end



// EXPORTS:
#ifdef __cplusplus
extern "C" {
#endif

int  iPhoneGetMaxMice()
{
	return MAX_SIMULTANEOUS_TOUCHES;
}

void iPhoneSwapBuffers()
{
	if(gpView)
		[gpView swapBuffers];
}



// DPI/PPI
int iphone_get_dpi()
{
	struct utsname systemInfo;
	uname(&systemInfo);
	char *name = systemInfo.machine;
 
	int ppi = 163;
	if ((strstr(name, "iPod1") != NULL) || (strstr(name, "iPod2") != NULL) || (strstr(name, "iPod3") != NULL) ) {
		// older ipod touches
		ppi = 163;
	} else if ((strstr(name, "iPhone1") != NULL) || (strstr(name, "iPhone2") != NULL)) {
		// older non-retina iphones
		ppi = 163;
	} else if ((strstr(name, "iPad1") != NULL) || (strstr(name, "iPad2") != NULL)) {
		// ipad 1, ipad 2
		ppi = 132;

		// ipad mini
		if(strstr(name, "iPad2,5") != NULL || strstr(name, "iPad2,6") != NULL || strstr(name, "iPad2,7") != NULL)
			ppi = 163;

	} else if (strstr(name, "iPad3") != NULL) {
		// ipad 3
		ppi = 264;
		if(gScaleFactor==1) ppi/=2;
	} else {
		// iphone 4/4s,5, ipod touch 4g,5 ( or simulator )
		ppi = 326;
		if(gScaleFactor==1) ppi/=2;
	}
	return ppi;
}


#ifdef __cplusplus
} // extern "C"
#endif
