/* DaCarSoft: file to create a view controller that will contain the iCadeView to let wrappers work in any orientation */

#import "iphoneviewcontroller.h"

#define SYSTEM_VERSION_GREATER_THAN_OR_EQUAL_TO(v)  ([[[UIDevice currentDevice] systemVersion] compare:v options:NSNumericSearch] != NSOrderedAscending)


@implementation iPhoneViewController

int fScaleFactor = 0;
int landscapesupported = 0;
int portraitsupported = 0;

- (void)viewDidLoad{
	float use_scale = [[UIScreen mainScreen] scale]; // 1=iPod, 2=retina
	fScaleFactor = use_scale == 2.0f ? 2:1;

}

-(BOOL) shouldAutorotate {
    if (SYSTEM_VERSION_GREATER_THAN_OR_EQUAL_TO(@"8.0"))
    {   return YES;
    }
    return NO;
}

- (BOOL)canBecomeFirstResponder {
    if (SYSTEM_VERSION_GREATER_THAN_OR_EQUAL_TO(@"8.0"))
    {   return YES;
    }
    return NO;
}

- (void)dealloc {
    
    [super dealloc];
    
}

- (BOOL)shouldAutorotateToInterfaceOrientation:(UIInterfaceOrientation)interfaceOrientation
{

    if (portraitsupported==0 && landscapesupported==0)
    {   NSArray* SupportedUIOrientations = [[NSBundle mainBundle] objectForInfoDictionaryKey:@"UISupportedInterfaceOrientations"]; /* DaCarSoft */
        for (NSString* iObject in SupportedUIOrientations) { /* DaCarSoft */
            if ([iObject isEqualToString: @"UIInterfaceOrientationPortrait"]) { /* DaCarSoft */
                portraitsupported = 1;
            } /* DaCarSoft */
            if ([iObject isEqualToString: @"UIInterfaceOrientationPortraitUpsideDown"]) { /* DaCarSoft */
                portraitsupported = 1;
            } /* DaCarSoft */
            if ([iObject isEqualToString: @"UIInterfaceOrientationLandscapeLeft"]) { /* DaCarSoft */
                landscapesupported = 1;
            } /* DaCarSoft */
            if ([iObject isEqualToString: @"UIInterfaceOrientationLandscapeRight"]) { /* DaCarSoft */
                landscapesupported = 1;
            } /* DaCarSoft */
        } /* DaCarSoft */
    }
    
    if (interfaceOrientation == UIDeviceOrientationPortrait) {
        if (portraitsupported==0) return NO;
        [[UIApplication sharedApplication] setStatusBarOrientation:UIInterfaceOrientationPortrait];
        return YES;
    }
    if (interfaceOrientation  == UIDeviceOrientationLandscapeLeft) {
        if (landscapesupported==0) return NO;
        [[UIApplication sharedApplication] setStatusBarOrientation:UIInterfaceOrientationLandscapeRight];
        return YES;
    }
    if (interfaceOrientation  == UIDeviceOrientationPortraitUpsideDown) {
        if (portraitsupported==0) return NO;
        [[UIApplication sharedApplication] setStatusBarOrientation:UIInterfaceOrientationPortraitUpsideDown];
        return YES;
    }
    if (interfaceOrientation == UIDeviceOrientationLandscapeRight) {
        if (landscapesupported==0) return NO;
        [[UIApplication sharedApplication] setStatusBarOrientation:UIInterfaceOrientationLandscapeLeft];
        return YES;
    }
    return NO;
}

- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{   /* associate touches with mice, so long as we have slots */
    int i;
#ifdef GLB_IN_A_THREAD
    if(htMutexLock(&gMainMutex)==0)
#endif
    {   for (UITouch *touch in touches)
        {
            struct GLBMOUSEVIEW* pMouse = &mice[0]; --pMouse;
            for(i=0; touch && i < MAX_SIMULTANEOUS_TOUCHES; i++)
            {
                ++pMouse;
                /* check if this mouse is already tracking a touch */
                if (pMouse->driverdata != NULL) {
                    continue;
                }
                /*
                 mouse not associated with anything right now,
                 associate the touch with this mouse
                 */
                
                /* select this slot's mouse */
                pMouse->driverdata = touch;
                
                CGPoint locationInView = [touch locationInView: self.view];
                locationInView.x *= fScaleFactor;
                locationInView.y *= fScaleFactor;
                
                /* send moved event */
                pMouse->x = (int)locationInView.x;
                pMouse->y = (int)locationInView.y;
                pMouse->button=1;
                /* grab next touch */
                // touch = (UITouch*)[enumerator nextObject];
                break;
            }
        }
        
#ifdef GLB_IN_A_THREAD
        htMutexUnlock(&gMainMutex);
#endif
    }
    [super touchesEnded: touches withEvent: event];
}

- (void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event
{
    for (UITouch *touch in touches)
    {
        /* try to find the mouse associated with this touch */
        int i, found=0;
        /* search for the mouse slot associated with this touch */
#ifdef GLB_IN_A_THREAD
        if(htMutexLock(&gMainMutex)==0)
#endif
        {
            struct GLBMOUSEVIEW* pMouse = &mice[0];
            for (i=0; i<MAX_SIMULTANEOUS_TOUCHES; i++)
            {
                if (pMouse->driverdata==touch) // mice[i].button && mice[i].x==prevlocationInView.x && mice[i].y==prevlocationInView.y)
                {
                    /* found proper mouse */
                    CGPoint locationInView = [touch locationInView: self.view];
                    
                    locationInView.x *= fScaleFactor;
                    locationInView.y *= fScaleFactor;
                    
                    /* send moved event */
                    pMouse->x=locationInView.x;
                    pMouse->y=locationInView.y;
                    
                    /* discontinue search */
                    found=1;
                    break;
                }
                ++pMouse;
            }
            
            // oh - make all invalid
            if(!found)
            {
                struct GLBMOUSEVIEW* pMouse = &mice[0];
                for (i=0; i<MAX_SIMULTANEOUS_TOUCHES; i++)
                {
                    pMouse->button=0; ++pMouse;
                }
            }
#ifdef GLB_IN_A_THREAD
            htMutexUnlock(&gMainMutex);
#endif
        }
    }
    [super touchesEnded: touches withEvent: event];
}

- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event
{
#ifdef GLB_IN_A_THREAD
    if(htMutexLock(&gMainMutex)==0)
#endif
    {
        for (UITouch *touch in touches)
        {
            
            int i;
            /* search for the mouse slot associated with this touch */
            struct GLBMOUSEVIEW* pMouse = &mice[0];
            for (i=0; i<MAX_SIMULTANEOUS_TOUCHES; i++)
            {
                if (touch == pMouse->driverdata) // mice[i].button && mice[i].x==prevlocationInView.x && mice[i].y==prevlocationInView.y)
                {
                    /* found the mouse associate with the touch */
                    /* send mouse up */
                    pMouse->button=0;
                    pMouse->driverdata=NULL;
                    break;
                }
                ++pMouse;
            }
            
        }
#ifdef GLB_IN_A_THREAD
        htMutexUnlock(&gMainMutex);
#endif
    }
    [super touchesEnded: touches withEvent: event];
}

@end
