#ifndef ASTARPLANNER_H
#define ASTARPLANNER_H
/**
  @class AStarPlanner
  @brief A* path-finding algorithm.

  In A* path-finding, the next node to check is chosen based on a combination
  of the node's cost up to that point and a heuristic weighting of how close
  it is to the goal.  This is essentially a combination of Dijkstra's and
  best-first.  The result is fairly optimal, and it runs faster than
  Dijkstra's algorithm alone.  The degree to which the algorithm takes into
  account the heuristic weight can be altered so as to balance the speed of
  the search with its accuracy.

  Copyright (C) 2003 Ted Cipicchio <ted@thereisnospork.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  IN THE SOFTWARE.

  @date  Created: Sun Mar 23 2003
*/

#include "planner.h"
//#include <vector>

//----------------------------------------------------------------------------
class AStarPlanner : public Planner
{
public:
	//------------------------------------------------------------------------
	/** @name Running */
	//@{

	/// Initialize the path-finding algorithm.
	virtual void initialize ();
	/// Run a single time slice of the path-finding algorithm.
	virtual bool run ();

	//@}
};

#endif	// ASTARPLANNER_H
