// blowfish.h     interface file for blowfish.cpp
// _THE BLOWFISH ENCRYPTION ALGORITHM_
// by Bruce Schneier
// Revised code--3/20/94
// Converted to C++ class 5/96, Jim Conger

// If you wnat to allow encode-code in project, define this:
// BLOWFISH_ALLOW_ENCODE


#ifndef BLOWFISH_INCLUDED
#define BLOWFISH_INCLUDED

#define MAXKEYBYTES 	56		// 448 bits max
#define NPASS           16		// SBox passes


typedef unsigned char BYTE;
typedef unsigned short WORD;
#ifdef _WIN32_WCE
	typedef long unsigned int DWORD;
#else
	#ifdef _WIN32
		typedef unsigned long DWORD;
	#else
		// *** SPACE FRACTAL FIXES 2017.3.1 **** MUST TO BEEN INT, OR ITS BORKS ON 64BIT.
		typedef unsigned int DWORD;
	#endif
#endif


#define BLOWFISH_STATIC_ASSERT( condition, name ) typedef char assert_failed_ ## name [ (condition) ? -1 : 1 ];
BLOWFISH_STATIC_ASSERT(  sizeof(DWORD)==32, must_define_proper_32_bit_DWORD  );
BLOWFISH_STATIC_ASSERT(  sizeof(WORD) ==16, must_define_proper_16_bit_WORD  );
BLOWFISH_STATIC_ASSERT(  sizeof(BYTE) == 8, must_define_proper_8_bit_BYTE  );




// choose a byte order for your hardware
#if !defined(BLOWFISH_BYTES_ABCD) && !defined(BLOWFISH_BYTES_BADC) && !defined(BLOWFISH_BYTES_DCBA)

	#ifdef __BIG_ENDIAN__
		#define BLOWFISH_BYTES_ABCD
	#else
		#define BLOWFISH_BYTES_DCBA // default to intel
	#endif

#endif


// DCBA - little endian - intel
#ifdef BLOWFISH_BYTES_DCBA
	union aword {
	  DWORD dword;
	  BYTE byte [4];
	  struct
	  {
		unsigned int byte3:8;
		unsigned int byte2:8;
		unsigned int byte1:8;
		unsigned int byte0:8;
	  } w;
	};
#endif

// ABCD - big endian - motorola
#ifdef BLOWFISH_BYTES_ABCD
	union aword {
	  DWORD dword;
	  BYTE byte [4];
	  struct {
	    unsigned int byte0:8;
	    unsigned int byte1:8;
	    unsigned int byte2:8;
	    unsigned int byte3:8;
	  } w;
	};
#endif

// BADC - vax
#ifdef BLOWFISH_BYTES_BADC
	union aword {
	  DWORD dword;
	  BYTE byte [4];
	  struct {
	    unsigned int byte1:8;
	    unsigned int byte0:8;
	    unsigned int byte3:8;
	    unsigned int byte2:8;
	  } w;
	};
#endif

class CBlowFish
{
private:
	DWORD 		* PArray ;
	DWORD		(* SBoxes)[256];
inline DWORD	F(aword x)							{return (((SBoxes[0][x.w.byte0] + SBoxes[1][x.w.byte1]) ^ SBoxes[2][x.w.byte2]) + SBoxes[3][x.w.byte3]);}
inline void		Round(aword& a, aword b, DWORD n)	{a.dword ^= F(b) ^ PArray[n];}


#ifdef BLOWFISH_ALLOW_ENCODE
public:
	DWORD		Encode (BYTE * pInput, BYTE * pOutput, DWORD lSize) ;
	#ifdef _AFX
	CString		EasyEncode(LPCTSTR mail);
	#endif
private:
#endif
	void 		Blowfish_encipher (DWORD *xl, DWORD *xr) ;
	void 		Blowfish_decipher (DWORD *xl, DWORD *xr) ;

public:
				CBlowFish () ;
				~CBlowFish () ;
	void 		Initialize (const BYTE key[], int keybytes) ;
	DWORD		GetOutputLength (DWORD lInputLong) ;
	void		Decode (BYTE * pInput, BYTE * pOutput, DWORD lSize) ;
	#ifdef _AFX
	CString		EasyDecode(LPCTSTR BFcode);
	#endif
} ;


#endif

