/*
------------------------------------------------------------------------------
rand.c: By Bob Jenkins.  My random number generator, ISAAC.  Public Domain.
MODIFIED:
  960327: Creation (addition of randinit, really)
  970719: use context, not global variables, for internal state
  980324: added main (ifdef'ed out), also rearranged randinit()
  010626: Note that this is public domain
------------------------------------------------------------------------------
*/

// Useage:
// GFsrand(GetTickCount()); unsigned long rn = GFrand();

extern void GFsrand(unsigned long any_number);
extern unsigned long GFrand();


#ifndef GFRAND_INCLUDED
#define GFRAND_INCLUDED

#define RANDSIZL   (8)  /* I recommend 8 for crypto, 4 for simulations */
#define RANDSIZ    (1<<RANDSIZL)


/* context of random number generator */
struct randctx
{
	unsigned long randcnt;
	unsigned long randrsl[RANDSIZ];
	unsigned long randmem[RANDSIZ];
	unsigned long randa;
	unsigned long randb;
	unsigned long randc;
};
typedef struct randctx  randctx;
randctx __g_ctx;


#define GFrand_ind(mm,x)  (*(unsigned long *)((unsigned char *)(mm) + ((x) & ((RANDSIZ-1)<<2))))
#define GFrngstep(mix,a,b,mm,m,m2,r,x) \
{ \
  x = *m;  \
  a = (a^(mix)) + *(m2++); \
  *(m++) = y = GFrand_ind(mm,x) + a + b; \
  *(r++) = b = GFrand_ind(mm,y>>RANDSIZL) + x; \
}

unsigned long GFrand()
{
   unsigned long a,b,x,y,*m,*mm,*m2,*r,*mend;
   mm=__g_ctx.randmem; r=__g_ctx.randrsl;
   a = __g_ctx.randa; b = __g_ctx.randb + (++__g_ctx.randc);
   for (m = mm, mend = m2 = m+(RANDSIZ/2); m<mend; )
   {
      GFrngstep( a<<13, a, b, mm, m, m2, r, x);
      GFrngstep( a>>6 , a, b, mm, m, m2, r, x);
      GFrngstep( a<<2 , a, b, mm, m, m2, r, x);
      GFrngstep( a>>16, a, b, mm, m, m2, r, x);
   }
   for (m2 = mm; m2<mend; )
   {
      GFrngstep( a<<13, a, b, mm, m, m2, r, x);
      GFrngstep( a>>6 , a, b, mm, m, m2, r, x);
      GFrngstep( a<<2 , a, b, mm, m, m2, r, x);
      GFrngstep( a>>16, a, b, mm, m, m2, r, x);
   }
   __g_ctx.randb = b; __g_ctx.randa = a;
return __g_ctx.randrsl[0];
}

#define GFrand_mix(a,b,c,d,e,f,g,h) \
{ \
   a^=b<<11; d+=a; b+=c; \
   b^=c>>2;  e+=b; c+=d; \
   c^=d<<8;  f+=c; d+=e; \
   d^=e>>16; g+=d; e+=f; \
   e^=f<<10; h+=e; f+=g; \
   f^=g>>4;  a+=f; g+=h; \
   g^=h<<8;  b+=g; h+=a; \
   h^=a>>9;  c+=h; a+=b; \
}

/* if (flag==TRUE), then use the contents of randrsl[] to initialize mm[]. */
void GFsrand(unsigned long any_number)
{
   unsigned long i;
   unsigned long a,b,c,d,e,f,g,h;
   unsigned long *m,*r;
   
   memset(&__g_ctx, 0,sizeof(__g_ctx));
   
   __g_ctx.randa = __g_ctx.randb = __g_ctx.randc = 0;
   m=__g_ctx.randmem;
   r=__g_ctx.randrsl;
   a=b=c=d=e=f=g=h=0x9e3779b9;  /* the golden ratio */

    /* fill in mm[] with messy stuff */
    for (i=0; i<RANDSIZ; i+=8)
    {
    a=a+any_number;
    GFrand_mix(a,b,c,d,e,f,g,h);
    m[i  ]=a; m[i+1]=b; m[i+2]=c; m[i+3]=d;
    m[i+4]=e; m[i+5]=f; m[i+6]=g; m[i+7]=h;
    }

   for (i=0; i<4; ++i)          /* scramble it */
     GFrand_mix(a,b,c,d,e,f,g,h);

     /* initialize using the contents of r[] as the seed */
     for (i=0; i<RANDSIZ; i+=8)
     {
       a+=r[i  ]; b+=r[i+1]; c+=r[i+2]; d+=r[i+3];
       e+=r[i+4]; f+=r[i+5]; g+=r[i+6]; h+=r[i+7];
       GFrand_mix(a,b,c,d,e,f,g,h);
       m[i  ]=a; m[i+1]=b; m[i+2]=c; m[i+3]=d;
       m[i+4]=e; m[i+5]=f; m[i+6]=g; m[i+7]=h;
     }
     /* do a second pass to make all of the seed affect all of m */
     for (i=0; i<RANDSIZ; i+=8)
     {
       a+=m[i  ]; b+=m[i+1]; c+=m[i+2]; d+=m[i+3];
       e+=m[i+4]; f+=m[i+5]; g+=m[i+6]; h+=m[i+7];
       GFrand_mix(a,b,c,d,e,f,g,h);
       m[i  ]=a; m[i+1]=b; m[i+2]=c; m[i+3]=d;
       m[i+4]=e; m[i+5]=f; m[i+6]=g; m[i+7]=h;
     }
   GFrand();		          /* fill in the first set of results */
   __g_ctx.randcnt=RANDSIZ;  /* prepare to use the first set of results */
}

#endif // #define GFRAND_INCLUDED

