
#ifndef _DDD_ASCII_H_
#define _DDD_ASCII_H_

#include "meshobject.h"
#include "gffilesystem.h"

class DDD_ascii
{
public:
	DDD_ascii() {}
/*
DDDA 1.1 # DDDA version_number

# texture coordinates

4            # texture coordinates count
0.000 0.000  # uv of vertex 0
0.000 0.000  # uv of vertex 1
0.000 0.000  # uv 2
0.000 0.000  # uv 3

# faces

2            # number of faces
255 255 255  # material color for face
3  0 0 3 0 2 0# number_of_nodes n1 t1 n2 t2 ni ti (n=position index, t=texture index)
255 255 255  # RGB
3  0 2 1     # n.o.n. n1 t1 n2 t2 ...

# animation frames

2            # number of keyframes
4            # nof_vertices_for_keyframe 0
0.0 1.8  0.6   0 1 0 # x1 y1 z1 nx ny nz
2.0 1.5  0.9   0 1 0 # x2 y2 z2 nx ny nz
1.9 0.1 -0.2   0 1 0
0.0 0.1 -0.3   0 1 0
4            # nof_vertices_for_keyframe 1
0.0 1.9 0.0   0 1 0
2.0 1.9 0.0   0 0 1
1.9 0.0 0.1   0 0 1
0.0 0.0 0.0   0 1 0
*/	
	#define DDDA_VERSION 1.1f
	bool Read(const char* fn)
	{
		int i;
		model.faces.Clear();
		model.frames.Clear();
		model.texcoo.Clear();

		float version;
		GFFileSystem file;
		if(!file.OpenIn(fn)) return false;

		if(strcmp(getword(file), "DDDA")!=0) return false;
		version = getflt(file);
		if(version> DDDA_VERSION) return false;
		
		// number of uv
		int numv = (int)getflt(file);
		for(i=0; i<numv; ++i)
		{
			TEXCOO n;
			n.u = getflt(file);
			n.v = getflt(file);
			model.texcoo.Push(n);
		}
		// Faces
		int numf = (int)getflt(file);
		for(i=0; i<numf; ++i)
		{
			FACE face;
			face.r = (int)getflt(file); face.g =(int)getflt(file); face.b = (int)getflt(file);
			int numn = (int)getflt(file);
			for(int j=0; j<numn; ++j)
			{
				int k = (int)getflt(file);
				face.nodes.Push(k);
				k = (int)getflt(file);
				face.tex_nodes.Push(k);
			}
			// printf("%d RGB: %f %f %f - %d\n", i, face.r, face.g, face.b, numn);
			model.faces.Push(face);
			face.nodes.Clear();
		}
		// keyframes
		int numkf = (int)getflt(file);
		for(i=0; i<numkf; ++i)
		{
			FRAME frame;
			numf = (int)getflt(file);
/*
			if(numf > (int)model.texcoo.Size())
			{
				model.texcoo.Resize(numf);
//				return false;
			}
*/
			for(int j=0; j<numf; ++j)
			{
				VERT vert;
				vert.x = getflt(file);
				vert.y = getflt(file);
				vert.z = getflt(file);
				if(version>=1.1f)
				{
					vert.nx = getflt(file);
					vert.ny = getflt(file);
					vert.nz = getflt(file);
				}
				frame.pos.Push(vert);
			}
			model.frames.Push(frame);
		}
		file.Close();
		return true;
	}
	void Make(MeshObject& obj)
	{
		POLY_NODE node, node_fan;
		POLY_FRAME frame;
		memset(&node_fan, 0, sizeof(node_fan));
		node_fan.col(COL_NEW_FAN);

		obj.Clear(true);
		obj.m_HasNormals=1;
		
		for(int kf=0; kf<(int)model.frames.Size(); ++kf)
		{
			for(int f=0; f<(int)model.faces.Size(); ++f)
			{
				frame.nodes.Push(node_fan);
				for(int n=0; n<(int)model.faces[f].nodes.Size(); ++n)
				{
					// i = node index for this face's vertex
					int i = model.faces[f].nodes[n];
					node.pos[0] = model.frames[kf].pos[i].x;
					node.pos[1] = model.frames[kf].pos[i].y;
					node.pos[2] = model.frames[kf].pos[i].z;
					node.normal[0] = model.frames[kf].pos[i].nx;
					node.normal[1] = model.frames[kf].pos[i].ny;
					node.normal[2] = model.frames[kf].pos[i].nz;
					// might have no texture coords
					if(model.texcoo.Size())
					{
						int j = model.faces[f].tex_nodes[n];
						node.tex[0]  =  model.texcoo[j].u;
						node.tex[1]  = -model.texcoo[j].v; // swap vertival to stay grabsprite compatible
						// printf("%d %d %.1f %.1f %.1f - %.1f %.1f\n", i, j, node.pos[0], node.pos[1], node.pos[2], node.tex[0], node.tex[1]);
					}					
					node.col(RGB(model.faces[f].r, model.faces[f].g, model.faces[f].b) );
					frame.nodes.Push(node);
				}
			}
			obj.m_Frames.Push(frame);
			frame.nodes.Clear(false);
		}
		obj.BuildColorTable();
	}

char a_word[1024];
	const char* getword(GFFileSystem& file)
	{
		memset(a_word, 0, sizeof(a_word));
		char* pC = a_word;
		char* pStart=NULL;
		bool is_word=false;
		char  read;

		while(file.Read(&read, 1) )
		{
			bool space=false;
			*pC = read;
			if(*pC==' ' || *pC =='\t' || *pC=='\n' || *pC=='\r')
				space=true;
			if(*pC=='#')
			{
				--pC;
				space=true;
				eat(file);
			}
			if(is_word && space) break;
			if(!space) {is_word = true; if(!pStart) pStart=pC;}
			++pC;
		}
		*pC = '\0';
		return pStart;
	}

	float getflt(GFFileSystem& file)
	{
		const char* pW = getword(file);
		if(pW)
			return (float)atof(pW);
		else
			return 0;
	}
	void eat(GFFileSystem& file)
	{
		char tmp[512];
		file.ReadLine(tmp,512);
	}

	struct VERT
	{
		float x,y,z, nx,ny,nz;
	};
	struct TEXCOO
	{
		float u,v;
	};
	struct FACE
	{
		int r,g,b;
		Queue<int> nodes;
		Queue<int> tex_nodes;
	};
	
	struct FRAME
	{
		Queue<VERT> pos;
	};

	struct MODEL
	{
		Queue<FACE>   faces;
		Queue<TEXCOO> texcoo;
		Queue<FRAME>  frames;
	} model;
};




#endif // _DDD_ASCII_H_

