// --- DXInput.h
#if !defined(DIRECTINPUT_CLASS_INCLUDED)
#define DIRECTINPUT_CLASS_INCLUDED

#define	DXIN_MAX_JOYSTICKS		10
#include "gfqueue.h"

// --- defines ---
#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)

#elif defined(TARGET_OS_IPHONE)
	
#elif defined(WANT_SDL) // SDL - Linux
	#ifdef TARGET_OS_IPHONE
		#define DXINPUT_HAVE_NO_KEYS
	#endif

	#include <SDL/SDL.h>
	#include <SDL/SDL_events.h>
	#include <SDL/SDL_timer.h>
	#include "gffilesystem.h"
	#ifndef SDLK_LAST
	#define SDLK_LAST 324
	#endif
	#define MAX_KEYS_DINPUT	(SDLK_LAST+1)

	#undef AnyKey


#elif defined(WIN32) // Win32
	#define MAX_KEYS_DINPUT			256

	#define DIRECTINPUT_VERSION		0x0800

	#include <dinput.h>
	//#include <dinputd.h>

	#define HR(error)  MAKE_HRESULT(1, 0x000, error)

	#define DINPUT_OK				1
	#define DINPUT_ERROR			0
	#define CANT_CREATE_DINPUT		HR(10)
	#define NO_ACCESS_ON_KEYBOARD	HR(11)
	#define NO_ACCESS_ON_MOUSE		HR(12)

	BOOL FAR PASCAL					g_EnumJoystick(LPCDIDEVICEINSTANCE pdinst, LPVOID pvRef);
	#if DIRECTINPUT_VERSION > 0x0700
		static	LPDIRECTINPUTDEVICE8A	g_lpDID=NULL;
	#else
		static	LPDIRECTINPUTDEVICE		g_lpDID=NULL;
	#endif

#endif


#if !defined( DIRECTINPUT_VERSION )
	typedef struct DIJOYSTATE {
		#if defined(GLBASIC_32BIT)
			long    lX;                     /* x-axis position              */
			long    lY;                     /* y-axis position              */
			long    lZ;                     /* z-axis position              */
		#else
			int    lX;                     /* x-axis position              */
			int    lY;                     /* y-axis position              */
			int    lZ;                     /* z-axis position              */
		#endif
		long    lRx;                    /* x-axis rotation              */
		long    lRy;                    /* y-axis rotation              */
		long    lRz;                    /* z-axis rotation              */
		long    rglSlider[2];           /* extra axes positions         */
		unsigned long    rgdwPOV[4];    /* POV directions   1/100 deg from north, CW   */
		unsigned char    rgbButtons[32];/* 32 buttons                   */
	} DIJOYSTATE, *LPDIJOYSTATE;
#endif

// --- structure for mousedata ---
struct DIMOUSEDATA
{
	short mButton1;
	short mButton2;
	short mButton3;
	short mAccX;
	short mAccY;
	short mMouseWheel; // (-)up, (+)down
};


// --- class constructor ---
class DXInput
{
public:
	DXInput();
	~DXInput();

#ifdef WIN32
	HRESULT	CreateDInput(HWND hWnd, HINSTANCE hInstance, bool WantMouse=true); // Init DInput
	void SetWindow(HWND hwnd)
	{
		#if defined(GP2X) || defined(GPHWIZ)||defined(CAANOO)
		#elif defined(WANT_SDL)
		#else
			m_HWND=hwnd;
		#endif
	}
#else
	unsigned long CreateDInput(HWND hWnd, HINSTANCE hInstance, bool WantMouse=true); // Init DInput
#endif
	void	GetKeyboardStatus(bool bForcePolling=false); // Freeze current buffer
	bool	AnyKey(void); // Is Any Key Pressed?

	inline bool	GetKey(unsigned char Key) // Get DI-Key Status
	{
		// -------- GP2X --------
		#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
			DIJOYSTATE data;
			GetJoystickData(&data, 0);
			switch(Key)
			{
				case 203: return data.lX<0 ? 1:0;
				case 205: return data.lX>0 ? 1:0;
				case 200: return data.lY<0 ? 1:0;
				case 208: return data.lY>0 ? 1:0;
				// ABXY = ADSW and Y,X,LCtrl,Caps
				case 44:
				case 30: return data.rgbButtons[0];
				case 45:
				case 32: return data.rgbButtons[1];
				case 29:
				case 31: return data.rgbButtons[2];
				case 58:
				case 17: return data.rgbButtons[3];
				
				// Lshift, RShift  = Flippers
				case 42: return data.rgbButtons[9];
				case 54: return data.rgbButtons[10];
				
				// Vol-,+ = PgDn,Up
				case 209:return data.rgbButtons[7];
				case 201:return data.rgbButtons[8];
				// Start=enter, Select=Tab
				case 28: return data.rgbButtons[5];
				case 15: return data.rgbButtons[6];
				// D-Pad button = Space
				case 57: return data.rgbButtons[4];
				// ESC
				case 1:
					// Vol+ & Vol- = ESC
					return (data.rgbButtons[7] && data.rgbButtons[8])?1:0;
				default:
					return 0;
			}

/*
DPad:    Cursor Keys/Space
Volume:  PageDn/PageUp
Front:   LShift/RShift
Start:   Enter
Select:  Tab
A,B,X,Y: A,D,S,W / Y,X,LCtr,CapsLock
Vol +&-: ESC
*/
		// -------- iPhone --------
		#elif defined(TARGET_OS_IPHONE)
			return false; // no keys

		// -------- SDL --------
		#elif defined(WANT_SDL)
			SDL_PumpEvents();
			Uint8 *pKeyState=NULL;
			// scheint SDL_GetKeyboardState nicht zu haben
			#if SDL_MAJOR_VERSION>1 || SDL_MINOR_VERSION >= 3
				pKeyState = SDL_GetKeyboardState(NULL);
			#else
				pKeyState = SDL_GetKeyState(NULL);
			#endif
			if(!pKeyState) return 0;

			#if defined(ANDROID)
				// __ANDROID_ESC_KEY_HACK__
				pKeyState[m_MapBuffer[1]] = pKeyState[m_MapBuffer[219]] || pKeyState[m_MapBuffer[117]]; // map silly BACK button (left windows key) to ESC
			#endif

			#if defined(PALM_WEBOS)
				SDL_Event event;
				if(SDL_PeepEvents(&event, 1, SDL_PEEKEVENT, SDL_EVENTMASK (SDL_KEYDOWN)))
				{
					enum
					{
						PDLK_GESTURE_AREA       = 231,
						PDLK_GESTURE_BACK       = 27,
						PDLK_GESTURE_FORWARD    = 229
					};

					pKeyState[m_MapBuffer[0x01]] = 0;
					pKeyState[m_MapBuffer[0x0F]] = 0; // tab
					switch (event.type)         
					{         
						case SDL_KEYDOWN:
							if (event.key.keysym.sym == PDLK_GESTURE_BACK)         
							{
								pKeyState[m_MapBuffer[0x01]] = SDL_PRESSED; // esc
							}       
							else if (event.key.keysym.sym == PDLK_GESTURE_FORWARD)         
							{
								pKeyState[m_MapBuffer[0x0F]] = SDL_PRESSED; // tab
							}
						break;         
    				}
				}
			#endif
				
			// int qp = pKeyState[SDLK_q];
			return pKeyState[m_MapBuffer[Key]]==SDL_PRESSED ? true : false;
		// -------- WIN32 --------
		#else
			GetKeyboardStatus();
			if (lpDIDKeyboard)
				return ((m_KeyBuffer[Key] & 0x80) ? true : false);
			else
			{
				return (m_KeyBuffer[m_MapBuffer[Key]]& 0x8000) ? true : false;
				// return (::GetAsyncKeyState(m_MapBuffer[Key])& 0x8000) ? true : false;
			}
		#endif
	}

	int     m_QueryCount; // don't update buffers on every function call
	Queue<char> m_inkey_chars;
	char	GetKeyChar(void); // Get the char that has been pressed

	int		GetNumMice();
	void	SetActiveMouse(int i);
	
	int		GetTouchPos(int* pX, int* pY, int* pWheel); // touch position for current mouse. false if not touched(clicked)
	void	GetMouseData(DIMOUSEDATA *Data, bool bUpdateXY); // Freeze Mousedata to stuct


	bool	AnyMouse(); // Is any Mousebutton down?

	inline void	UnaquireMouse() {
	#if defined(WIN32) && !defined(WANT_SDL)
			if(lpDIDMouse) lpDIDMouse->Unacquire();
	#endif
	}


	bool	GetJoystickData(DIJOYSTATE* Data, int num); // Get Josytick Information
	int		GetNumJoysticks(); // Get Numer of Found Devices
	void	GetJoyName(char* pName, int num); // Name of Device
	void	ffRumble(int ndevice, int count, float duration /*[ms]*/, float x_motor /*[-1.0; 1.0]*/, float y_motor);

	bool	IsSystemPointer()					{return m_SystemPointer;}
	void	SystemPointer(bool sp)
												{
													// -------- Touchscreen devices --------
													#if defined(EMSCRIPTEN) || defined(GP2X) || defined(GPHWIZ) || defined(ANDROID) || defined(TARGET_OS_IPHONE) || defined(PALM_WEBOS) || defined(PANDORA) || defined(CAANOO)
														sp = true;
									
													// -------- SDL (PC hardware) --------
													#elif defined(WANT_SDL)
														SDL_WM_GrabInput(sp ? SDL_GRAB_OFF : SDL_GRAB_ON);
														SDL_ShowCursor(sp?1:0);
													// -------- WIN32 --------
													#else
														if(lpDIDMouse /* && sp!=m_SystemPointer*/ )
														{
															lpDIDMouse->Unacquire();
															m_MouseCooLevel = sp ? DISCL_NONEXCLUSIVE|DISCL_FOREGROUND : DISCL_EXCLUSIVE|DISCL_FOREGROUND;
															lpDIDMouse->SetCooperativeLevel(m_HWND, m_MouseCooLevel);
															lpDIDMouse->Acquire();
														}
														if(sp)
														{
															// Set Pointer to where it is in GLBasic
															if(sp && !m_SystemPointer)
															{
																POINT pt;
																MousePos& mps = GetCurMousePos();
																pt.x = mps.m_MouseX; pt.y=mps.m_MouseY;
																::ClientToScreen(m_HWND, &pt);
																::SetCursorPos(pt.x, pt.y);
															}
															::SetCursor(::LoadCursor(NULL, IDC_ARROW));
														}
														else
														{
															::SetCursor(NULL);
														}
													#endif
													m_SystemPointer = sp;
												}
	void	RefreshMouseXY()					{
													int iold = m_iCurMouse;
													static DIMOUSEDATA d;
													for(int i=0; i<(int)m_MousePos.Size(); ++i)
													{
														m_iCurMouse = i; // update all mouse(touch) positions. Fixed? http://www.glbasic.com/forum/index.php?topic=11094
														GetMouseData(&d, true);
													}
													m_iCurMouse = iold;
												}
	void	SetMouse(int x, int y)
												{
													MousePos& mps = GetCurMousePos();
													mps.m_MouseX=x; mps.m_MouseY=y; mps.m_VirtualX=x; mps.m_VirtualY=y;
												}
	
	struct MousePos
	{
		MousePos() {m_VirtualX=m_VirtualY=0; m_MouseX=m_MouseY=100; m_MouseWheel=0;}

		int m_VirtualX, m_VirtualY; // position where accx will move the mouse pointer on next refresh
		int m_MouseX, m_MouseY, m_MouseWheel;
	};
	Queue<MousePos> m_MousePos;
	MousePos& GetCurMousePos()					{if(m_iCurMouse>=0 && m_iCurMouse<m_MousePos.Size()) {return m_MousePos[m_iCurMouse];} static MousePos mp; return mp;}
private:
	int m_iCurMouse;

public:
#if defined(GP2X) || defined(GPHWIZ)
	unsigned long keystate;
	int				device;
	int				dev_touchscreen;
#elif defined(WANT_SDL)	
	private:
//		static Uint8  SDL_KeyState[SDLK_LAST];
		SDL_Joystick* m_pJoys[DXIN_MAX_JOYSTICKS];
	public:
		bool m_AccelerationNativeLandscape; // make true for HP touchpad ^X <Y turned to ^Y >X



#elif defined(WIN32) // Win32
	private:
		#if DIRECTINPUT_VERSION > 0x0700
			LPDIRECTINPUT8			lpDInput;
			LPDIRECTINPUTDEVICE8A	lpDIDKeyboard;
			LPDIRECTINPUTDEVICE8A	lpDIDMouse;
			LPDIRECTINPUTDEVICE8A	lpDIDJoystick[DXIN_MAX_JOYSTICKS];
		#else
			LPDIRECTINPUT			lpDInput;
			LPDIRECTINPUTDEVICE	lpDIDKeyboard;
			LPDIRECTINPUTDEVICE	lpDIDMouse;
			LPDIRECTINPUTDEVICE2	lpDIDJoystick[DXIN_MAX_JOYSTICKS];
		#endif
		TCHAR					JoyNames[DXIN_MAX_JOYSTICKS][MAX_PATH*2+2];
		BYTE					m_KeyBuffer[MAX_KEYS_DINPUT]; // must be static size
		HWND					m_HWND;
#endif


protected:
	void					DXtoVK(bool DXActive);
	char					Scan2Ascii(DWORD scancode);

	bool					m_SystemPointer;
	unsigned long			m_MouseCooLevel;
	int						NumJoysticks;
#if !defined(GP2X) && !defined(GPHWIZ) && !defined(TARGET_OS_IPHONE) && !defined(CAANOO)
	unsigned short			m_MapBuffer[MAX_KEYS_DINPUT]; // must be static size
#endif
};

#endif // DIRECTINPUT_CLASS_INCLUDED

