#ifndef SFXSOUND_H
#define SFXSOUND_H

// WANT_NOSOUND
// WANT_SDL_MIXER
// _WIN32_WCE
// APPLE_COREAUDIO
// WANT_SDL_WAVSOUND

#ifdef WANT_NOSOUND
#define DXSOUND_PAN_RIGHT 0x1000 // maximum right value (0=left?)

struct SFXSound
{
	BOOL Create(HWND hwnd){return FALSE;}
	BOOL GetCaps(void){return FALSE;}
};

struct SFXSoundBuffer
{
	void Init(void){}
	void Clear(); {}
	bool Load(SFXSound* pDS, const char* Filename, int Num){return 0;}
	long Play(int Pan = 0, int Volume=100, unsigned long dwFlags = 0){return 0;}
	bool IsPlaying(long channel){return 0;}
	long Stop(int channel=-1){return 0;}
	BOOL IsLoaded()		{return FALSE;}
};



#else // ! WANT_NOSOUND

#if defined(WIN32) && !defined(WANT_SDL_MIXER)


#define DXSOUND_PAN_RIGHT DSBPAN_RIGHT


#ifndef RELEASE
 #define RELEASE(x) if(x != NULL) { x->Release(); x = NULL; }
#endif

#include "mmsystem.h"
#include "dsound.h" //link to dxguid.lib, dsound.lib
// #include "stdio.h"
#include "gffilesystem.h"

//////////////////////////////////////////////////////////////////////////////////
// SFXSound Class
//////////////////////////////////////////////////////////////////////////////////
class SFXSound
{
public:
	SFXSound(void);
	~SFXSound(void);

	BOOL Create(HWND hwnd);
	void SetWindow(HWND hwnd);
	LPDIRECTSOUND m_lpDS;
	DSCAPS m_DSCaps;
private:
	BOOL GetCaps(void);
};

//////////////////////////////////////////////////////////////////////////////////
// SFXSoundBuffer Class
//////////////////////////////////////////////////////////////////////////////////
class SFXSoundBuffer
{
private:
	int m_nBuffers;
	int m_Current;
	const char* m_Filename;
	SFXSound* m_pDS;
	LPDIRECTSOUNDBUFFER m_lpDSB[4];

public:
	SFXSoundBuffer(void);
	~SFXSoundBuffer(void);

	void Init(void);

	void Clear();
	bool Load(SFXSound* pDS, const char* Filename, int Num);
	long Play(int Pan = 0, int Volume=100, DWORD dwFlags = 0);
	bool IsPlaying(long channel);
	BOOL Stop(int channel=-1);
	BOOL IsLoaded()		{return m_nBuffers>0 ? TRUE:FALSE;}

private:
	void FreeBuffer(void);
	void SetVolume(LONG Volume);
	bool LoadFromFile(void);
	BOOL CreateSoundBuffer(DWORD dwBufSize, DWORD dwFreq, DWORD dwBitsPerSample,
												 DWORD dwBlkAlign, BOOL bStereo);
	BOOL ReadData(GFFileSystem& fp, DWORD dwSize, DWORD dwPos);
	LPDIRECTSOUNDBUFFER GetFreeBuffer(void);
};


#elif defined(WANT_SDL_MIXER)

#include <SDL/SDL.h>
#include <SDL/SDL_mixer.h>
#include "gffilesystem.h"

#define DXSOUND_PAN_RIGHT 255


class SFXSound
{
public:
	SFXSound(void);
	~SFXSound(void);

	BOOL Create(HWND hwnd) {return TRUE;}
	BOOL GetCaps(void) {return TRUE;}
	long m_IsSetup;
};

class SFXSoundBuffer
{
public:
	SFXSoundBuffer(void);
	~SFXSoundBuffer(void);

	void Init(void);
	void Clear();

	bool Load(SFXSound* pDS, const char* Filename, int Num);
	long Play(int Pan = 0, int Volume=100, unsigned long dwFlags = 0);
	bool IsPlaying(long channel);
	long Stop(int channel=-1);
	BOOL IsLoaded()		{return m_cvt ? TRUE:FALSE;}
private:
	void FreeBuffer(void);
	Mix_Chunk* m_cvt;
};

#elif defined(_WIN32_WCE)

/*
[DllImport("aygshell.dll")]

internal static extern uint SndPlaySync(string file, uint flags);
internal static extern uint SndOpen(string file, ref IntPtr phSound);
internal static extern uint SndPlayAsync(IntPtr hSound, uint flags);
internal static extern uint SndClose(IntPtr hSound);
internal static extern uint SndStop(int SoundScope, IntPtr hSound);
*/
class SFXSound
{
public:
	SFXSound(void);
	~SFXSound(void);

	BOOL Create(HWND hwnd);
	BOOL GetCaps(void);
};

class SFXSoundBuffer
{
public:
	SFXSoundBuffer(void);
	~SFXSoundBuffer(void);

	void Init(void);
	void Clear();
	bool Load(SFXSound* pDS, const char* Filename, int Num);
	long Play(int Pan = 0, int Volume=100, unsigned long dwFlags = 0);
	bool IsPlaying(long channel);
	long Stop(int channel=-1);
	BOOL IsLoaded()		{return m_pSound ? TRUE:FALSE;}
	int* m_pSound;
};


#elif defined(APPLE_COREAUDIO)
	#define GLB_OPENAL_BUFFERS 4
        #if defined(__LP64__) || defined(GLBASIC_64BIT)
            typedef unsigned int                    UInt32;
            typedef signed int                      SInt32;
        #else
            typedef unsigned long                   UInt32;
            typedef signed long                     SInt32;
        #endif
	#define DXSOUND_PAN_RIGHT 0x1000 // what?
	class SFXSound
	{
	public:
		SFXSound(void);
		~SFXSound(void);
		BOOL Create(HWND hwnd);
		BOOL GetCaps(void);
	};

	class SFXSoundBuffer
	{
	public:
		SFXSoundBuffer();
		~SFXSoundBuffer();
		void Init(void);
		void Clear();
		bool Load(SFXSound* pDS, const char* Filename, int Num);
		long Play(int Pan = 0, int Volume=100, unsigned long dwFlags = 0);
		bool IsPlaying(long channel);
		long Stop(int channel=-1);
		BOOL IsLoaded()		{return effects[0]>0 ? TRUE:FALSE;}
	private:
		UInt32 effects[GLB_OPENAL_BUFFERS];
	};

	bool iPhonePlayMusic(const char* pFilename, bool bLoop);
	bool iPhoneIsMusicPlaying();
	void iPhonePauseMusic(bool bPause);
	void iPhoneStopMusic();
	void iPhoneMusicVolume(float vol);

#elif defined(WANT_SDL_WAVSOUND)
	#define DXSOUND_PAN_RIGHT 0x1000 // what?
	class SFXSound
	{
	public:
		SFXSound(void);
		~SFXSound(void);
		BOOL Create(HWND hwnd);
		BOOL GetCaps(void);
	};

	class SFXSoundBuffer
	{
	public:
		SFXSoundBuffer();
		~SFXSoundBuffer();
		void Init(void);
		void Clear();
		bool Load(SFXSound* pDS, const char* Filename, int Num);
		long Play(int Pan = 0, int Volume=100, unsigned long dwFlags = 0);
		bool IsPlaying(long channel);
		long Stop(int channel=-1);
		BOOL IsLoaded()		{return length ? TRUE:FALSE;}
	private:
		Uint8 *buffer;              /* audio buffer for sound file */
		Uint32 length;              /* length of the buffer (in bytes) */
	};



#endif

#endif



#endif



