
#ifndef GF_SOCKETS_H_INC_
#define GF_SOCKETS_H_INC_


#ifdef _SS_MAXSIZE
	typedef sockaddr_storage* LP_SOCKADDR_STORAGE;
#else
	typedef void* LP_SOCKADDR_STORAGE;
#endif


// constants
#define NET_INVALID	-1	// invalid socket index


typedef int NETsocket;

struct NETaddress
{
	int  m_ipFamily; // AF_INET or AF_INET6
	char m_ip[16];   // in network format (big endian) (IPv4 or IPv6)
	unsigned short m_port; // nw format
	bool operator == (const NETaddress& a)const {return m_ipFamily==a.m_ipFamily && m_ip==a.m_ip && m_port==a.m_port;}
	bool operator != (const NETaddress& a)const {return m_ipFamily!=a.m_ipFamily || m_ip!=a.m_ip || m_port!=a.m_port;}

	static int get_family_ofIPv(int v=4);

	NETaddress() {memset(this, 0, sizeof(NETaddress));}
	NETaddress(int iFamily, const void* pIP, unsigned short port) {assign(iFamily, pIP, port);}

	NETaddress& operator=(const NETaddress& a){memcpy(this, &a, sizeof(NETaddress)); return *this;}

	void assign(int iFamily, const void* pIP, unsigned short port_hostmachine_format);
	void assign(LP_SOCKADDR_STORAGE ptr_sockaddr_storage);

	void fill_sockaddr_storage(LP_SOCKADDR_STORAGE ptr_sockaddr_storage)const;
};

const NETaddress NET_ADDR_ANY    =NETaddress(0, NULL, 0);
const NETaddress NET_ADDR_INVALID=NETaddress(-1, NULL, 0);

// Init network. Please call a netShutdown for every netInit when you're done.
bool netInit();
void netSetVersion(int iv); // only accepted values are 4: IPv4 and 6: IPv6
void netShutdown();

// create an address from a ip or computer name
// server - name of computer or IP. NULL or "" for ANY
// port   - port for IP. 0 for have network find any free one (epiphermal)
NETaddress netAddress(const char* server=NULL, unsigned short port=0);
NETaddress netAddressIP4(unsigned char ip1=127, unsigned char ip2=0, unsigned char ip3=0, unsigned char ip4=1, unsigned short port=0);


// Find remote connection (or last UDP recv)
NETaddress netGetRemoteAddress(NETsocket isock);

// for better debugging
const char*    netGetIPstr(const NETaddress& address);

//unsigned long  netGetIP   (const NETaddress& address); // in host format
unsigned short netGetPort (const NETaddress& address); // in host format


// See if the socket is ready to read
// Returns:
//  1: ok
// -1: error
// -2: would block
int netSelect(NETsocket isock);


// create a TCP socket that listens to a port.
// the socket will be put to non-blocking mode, initially.
// returns socket index, or NET_INVALID on error
// port - port to accept connections on local machine
NETsocket netListen(unsigned short port);

// accept a connection on a listening socket
// returns a new socket (must call netClose when you're done with it)
// or NET_INVALID on error.
// the socket will be put to non-blocking mode, initially.
// isock - NETsocket returned by netListen()
NETsocket netAccept(NETsocket isock, NETaddress* pAddress=NULL);

// create a TCP socket that connects to a url / port.
// the socket will be put to non-blocking mode, initially.
// returns socket index, or NET_INVALID if error
// port      - port connect to remote machine.
NETsocket netConnect(const NETaddress& server, int timeout_ms=5000);


// Create and bind a new socket used to send and receive to/from another
// computer. If you leave server=NULL or "", you can only receive with this
// "socket". Otherwise you can send as well.
// server     - computer name, or ip string to write and receive to/from
// port       - port to read from
NETsocket netSocketUdp(unsigned short port);


// close a socket when you don't need it anymore
// the engine will probably reuse this socket the next time
// a new socket is needed.
// isock - a valid NETsocket. Set it to NET_INVALID afterwards, don't use it again.
int netCloseSocket(NETsocket isock);

// set a socket to blocking or non-blocking mode
void netSetBlocking(NETsocket isock, bool block);

// Send a buffer to the remote IP+port of a socket
// -1 : error
// >=0: ok, sent all.
int netSend(NETsocket isock, const char* buf, int len, NETaddress* pUDPtarget=NULL);

// receive from a socket's local IP (specified by bind() )
// -2 : EWOULDBLOCK -> just ignore for non blocking reads
// -1 : error
//  0 : sender quit
//  n : ok, sent all.
int netRecv(NETsocket isock, char* buf, int len, bool force_read_all=false);

// Get a list of all local IP addresses for this computer
const char* netGetLocalIPs(int* count, const char* delimiter="\n");

// Get the error and a description
const char* netLastError();

// Get the error as a number, only
int netLastErrNum();

// Set a user defined last error
void netStoErr(int err);

#endif // GF_SOCKETS_H_INC_
