#if !defined(AFX_QUEUE_H__A54DCD42_E1B4_11D5_B6EB_003042FFFD01__INCLUDED_)
#define AFX_QUEUE_H__A54DCD42_E1B4_11D5_B6EB_003042FFFD01__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

template <class QueueItem> class Queue
{
private:
	QueueItem* m_buffer;
	int m_count;
	int m_bufsize;
	int m_alloc_step, m_alloc_step_first;

public:
	Queue()
	{
		m_count=0;
		m_bufsize=0;
		m_buffer=NULL;
		#ifdef POCKETPC
			m_alloc_step_first=4;
			m_alloc_step = 10;
		#else
			m_alloc_step_first=512;
			m_alloc_step = 128;
		#endif
	}
	#if 0
		Queue(const Queue& q) {*this=q;}
		inline Queue& operator=(const Queue& q)
		{
			AllocAtLeast(q.m_bufsize);
			m_count = q.m_count;
			for(int i=0; i<m_count; ++i)
				m_buffer[i] = q.m_buffer[i];
			return *this;
		}
	#else
		Queue(const Queue& q) : m_bufsize(0), m_buffer(NULL), m_count(0)
		{
			#ifdef POCKETPC
				m_alloc_step_first=4;
				m_alloc_step = 10;
			#else
				m_alloc_step_first=512;
				m_alloc_step = 128;
			#endif

			AllocAtLeast(q.m_bufsize);
			m_count = q.m_count;

			for(int i=0; i<m_count; ++i)
				m_buffer[i] = q.m_buffer[i];
			m_alloc_step_first = q.m_alloc_step_first;
			m_alloc_step = q.m_alloc_step;
		}
		template<typename T> static void member_swap(T& a, T& b) {T t(a);a=b;b=t;}
		friend void swap(Queue& a, Queue& b)
		{
			member_swap(a.m_buffer,           b.m_buffer);
			member_swap(a.m_count,            b.m_count);
			member_swap(a.m_bufsize,          b.m_bufsize);
			member_swap(a.m_alloc_step,       b.m_alloc_step);
			member_swap(a.m_alloc_step_first, b.m_alloc_step_first);
		}
		inline Queue& operator=(Queue q)
		{
			swap(*this, q);
			return *this;
		}
	#endif

	void SetAllocSteps(int firsttime, int next_alloc)
	{
		m_alloc_step = firsttime;
		m_alloc_step_first=next_alloc;
	}
	
	inline bool AllocAtLeast(int size)
	{
		if (m_bufsize<=size)
		{
			if(size<=m_alloc_step_first)
				m_bufsize=m_alloc_step_first;
			else
				m_bufsize=m_alloc_step*(1+size/m_alloc_step);
			#ifdef _DEBUG
				if(m_bufsize < size) *(int*)0=0;
			#endif
				
			QueueItem* newbuffer=new QueueItem[m_bufsize];
			if(!newbuffer)
			{
				#ifdef _DEBUG
				*(int*)0=0;
				#endif
				Clear(true);
				return false;
			}
			if (m_buffer)
			{
				// Only for structs! Not for heap memory!
				// memcpy(newbuffer, m_buffer, sizeof(QueueItem)*m_count);
				for (int i=0; i<m_count; i++)
					newbuffer[i]=m_buffer[i];
				delete [] m_buffer;
			}
			m_buffer=newbuffer;
		}
		return true;
	}
	
	inline bool Push(const QueueItem& item)
	{
		if(m_bufsize <= m_count)
			if(!AllocAtLeast(m_count+1))
				return false;
		m_buffer[m_count++]=item;
		return true;
	}

	void Pop(void) 
	{
		if (m_count>0) --m_count; // Don't really free the memory.
	}

	// keep old elements (copy might occour)
	inline void SetSize(int count)
	{
		AllocAtLeast(count);
		m_count = count;
	}
	
	inline int Size(void) const
	{
		 return m_count;
	}

	void Clear(bool free_mem=true)
	{
		if (free_mem && m_buffer)
		{
			delete[] m_buffer;
			m_buffer=NULL; m_bufsize=0;
		}
		m_count=0;
	}
	inline QueueItem*	GetAt(int i) const
	{
		#ifdef _DEBUG
		if(i<0 || i>=m_count) return 0L;
		#endif
		return &m_buffer[i];
	}

	inline QueueItem& operator[](int i)
	{
		#ifdef _DEBUG
		if(i<0 || i>=m_count) *(int*)0=0;
		#endif
		return m_buffer[i];
	}
	inline const QueueItem& operator[](int i)const
	{
		#ifdef _DEBUG
		if(i<0 || i>=m_count) *(int*)0=0;
		#endif
		return m_buffer[i];
	}

	
	inline void Remove(int index, bool allow_insert_last=true)
	{
		#ifdef _DEBUG
		if(index<0 || index>=m_count) *(int*)0=0;
		#endif
		--m_count;
		if(allow_insert_last)
		{
			m_buffer[index] = m_buffer[m_count];
		}
		else
		{
			for(int i=index; i<m_count; ++i)
			{
				m_buffer[i] = m_buffer[i+1];
			}
		}
	}

	~Queue() 
	{
		Clear(true);
	}
};

#endif // ndef Included
