// Implementation of GLBasic commands
#include "stdafx.h"
#include <time.h>
#include <locale.h>
#include <ctype.h>
#include <wchar.h>

#ifdef _WIN32_WCE
	#define UNICODE_ONLY
	#include <sys/wcebase.h>
	#include <windows.h>
	#define WINSHELLAPI
	#include <shellapi.h> // ShellExecuteEx
#elif defined (WIN32)
	#ifndef _WIN32_IE
	#define _WIN32_IE 0x0400
	#endif
	#include <shlobj.h> // SHGetSpecialFolderPath
	#include <shellapi.h> // ShellExecuteEx
#endif

#ifndef GLB_CONSOLE
	#include "openglrainbows.h"
	#include "rainbows3d.h"

	// PLATFORMINFO - requires F200 touchscreen knowledge
	#ifdef GP2X
	#include "dxinput.h"
	namespace __GLBASIC__
	{
		extern DXInput*		pDXin;
	};
	#endif
#else
	#ifndef WIN32
		// get console size
		#include <sys/ioctl.h>
		#include <stdio.h>
	#endif
#endif


#ifdef WANT_SHOEBOX
	#include "gfshoebox.h"
	extern GFShoeBox *__g_ShoeBox_Data;
	extern GFShoeBox *__g_ShoeBox_Media;
#endif

#ifndef WANT_SDL
	#ifdef WIN32
	#ifndef POCKETPC
		#define VC_EXTRALEAN
		#define _NO_OLDNAMES // Allow _yn() but not yn() in Math.h
		#if _MSC_VER > 1000
		#pragma once
		#endif // _MSC_VER > 1000

		// Video playback
		#ifndef GLB_CONSOLE
			#define NOMMREG
			#define NOMSACM
			#include <vfw.h>
			#pragma comment(lib, "VFW32.lib")
		#endif
		#define strlen	lstrlen // winAPI instead of stdlib
	#endif
	#endif
#else
	#ifndef GLB_CONSOLE
	//	#include <process.h> // system() -> stdio.h / stdlib.h??
		#pragma comment (lib, "SDL.lib")
		#pragma comment (lib, "SDL_mixer.lib")
		#pragma comment (lib, "SDLmain.lib")
	//	#pragma comment (lib, "mikmod.lib")
	//	#pragma comment (lib, "native_midi.lib")
	//	#pragma comment (lib, "ogg_static.lib")
	//	#pragma comment (lib, "timidity.lib")
	//	#pragma comment (lib, "vorbis_static.lib")
	//	#pragma comment (lib, "vorbisfile_static.lib")
	#endif
	#include <SDL/SDL.h>
#endif
#ifdef WANT_SDL_MIXER
	#ifndef GLB_CONSOLE
		#pragma comment (lib, "SDL_mixer.lib")
	#endif
#endif

#include <math.h>
#include "gfshoebox.h"
#include "gffilesystem.h"
#include "gfqueue.h"

// #include "gfrand.h"
extern void GFsrand(unsigned long any_number);
extern unsigned long GFrand();

// Networking
#include "network/clientserver.h"

// Path finding:
#include "pathfinding/astarplanner.h"
/*
#include "pathfinding/bestplanner.h"
#include "pathfinding/dijkstraplanner.h"
#include "pathfinding/breadthplanner.h"
*/

#include "glb.h"
#ifndef GLB_CONSOLE
	#include "glb3d.h"
#endif
#include "gfini.h"
#include "3dmath.h"
#include "blowfish.h"


#ifdef WIN32
	#ifdef POCKETPC
		extern char* gcvt(double,int,char *);
		#define _gcvt gcvt

		inline char* _itoa(int value, char *buffer, int /*radix = 10 */)
		{
			char tmp[33];
			char *tp = tmp;
			int i;
			unsigned v;
			int sign;
			char *sp;

			sign = (value < 0);
			if (sign)
				v = -value;
			else
				v = (unsigned)value;
			while (v || tp == tmp)
			{
				i = v % 10;
				v = v / 10;
				// if (i < 10)
					*tp++ = i+'0';
				// else *tp++ = i + 'a' - 10;
			}
			sp = buffer;

			if (sign)
				*sp++ = '-';
			while (tp > tmp)
				*sp++ = *--tp;
			*sp = 0;
			return buffer;
		}

	#else // WIN32-PC
	#endif // POCKETPC

#else // Linux - std-C
	char * _gcvt (double value, int ndigits, char *buf)
	{
	char *p = buf;

	sprintf(buf, "%-#.*g", ndigits, value);

	/* It seems they expect us to return .XXXX instead of 0.XXXX  */
	if (*p == '-')
		p++;
	if (*p == '0' && p[1] == '.')
		memmove (p, p + 1, strlen (p + 1) + 1);

	/* They want Xe-YY, not X.e-YY, and XXXX instead of XXXX.  */
	p = strchr (buf, 'e');
	if (!p)
		{
		p = buf + strlen (buf);
		/* They don't want trailing zeroes.  */
		while (p[-1] == '0' && p > buf + 2)
		*--p = '\0';
		}
	if (p > buf && p[-1] == '.')
		memmove (p - 1, p, strlen (p) + 1);
	return buf;
	}

	inline void _itoa(int num, char* data, int base)
	{
		sprintf(data, "%d", num);
	}
#endif
	inline void floattochar(double f, char* data)
	{
		_gcvt(f, 10, data);
		// linux bug
		if(data[0]=='.' && data[1]=='0' && data[2]=='\0')
		{
			data[0]='0'; data[1]='\0';
			return;
		}
		if(data[strlen(data)-1]=='.')
			data[strlen(data)-1]='\0';
	}

#ifdef WANT_SHOEBOX
	GFShoeBox *__g_ShoeBox_Data;
	GFShoeBox *__g_ShoeBox_Media;
#endif

//  *** DLLCALL ***
#if defined(POSIX) || defined(LINUX)
	#include <dlfcn.h>
	#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO) || defined(PALM_WEBOS) // webos - disallows that
		#define LoadLibrary(a)      NULL
		#define GetProcAddress(a,b) NULL
		#define FreeLibrary(a)      {}
	#else
		#define LoadLibrary(a)      dlopen(a, RTLD_LAZY)
		#define GetProcAddress(a,b) dlsym(a,b)
		#define FreeLibrary(a)      dlclose(a);
	#endif
	#define HMODULE   void*
	#define HINSTANCE void*
#endif

// GP2X memory registers
#ifdef GP2X
extern volatile unsigned short *gp2x_memreg16;
#endif



// ---------- __GLBASIC__ ------------------
namespace __GLBASIC__
{
int				__DG_RESX=0; // window screen res
int				__DG_RESY=0;
int				__DG_DESKTOPX=0;
int				__DG_DESKTOPY=0;
int				__DG_MULTISAMPLE=0;
int				__DG_DEBUG=0;
int				__DG_FRAMERATE=0;
int				__DG_WANTMOUSE=0;
int				__DG_FULLSCREEN=0;
int             __IS_INITDONE=0; // Space Fractal - Required by Android


DGInt			__g_FirstTime;	// The time the program has started
DGInt			__g_LastTime;	// To get the time elapsed within the SHOWSCREEN cmds
DGInt			__g_ThisTime;	// To get the time elapsed within the SHOWSCREEN cmds
#ifdef ANDROID
	// SPACE FRACTAL: Gernot, on Android, This >MUST< ALLWAYS been false, which is REQUIRED by newer Android versions and by some venders (Amazon AppStore).
	// We need to could monitor BACK button manually and required let programmer implement this button in correct way. Android is NOT Windows!!!
	int				__g_AllowEscape = FALSE; 
#else
	int				__g_AllowEscape = TRUE; // GF: Must be this way. Don't break backwards compatibility. USE ALLOWESCAPE FALSE in your GLB source code instead.
#endif
int				__g_paused=FALSE;
int				__g_allowPause=TRUE;
int				__g_CurrentScreen=-1; // USESCREEN arg

GLB_PRESCALER	__g_Prescaler;

extern bool		__g_DidNET;
extern bool		__g_Did3D;
int				__nCurFont;
GFHawkNetwork*	pGFNet=NULL;

#ifndef GLB_CONSOLE
	OpenGLRainbows*	__g_pRbow;
	Rainbows3D*		__g_pRbow3D;
	HINSTANCE		__g_HInstance;
#endif

/* PUSHSTATE
struct glb_engine_state
{
	DGNat vpx, vpy, vpw, vph; // viewport size (GETVIEWPORT)
	DGNat smoothshading;
	DGNat usescreen;
	DGStr currentdir;
	DGNat font;
	DGNat orientation;
	DGInt alphamode;
	DGNat transparency;

};
*/
void    glb_showscreen_finish_frame();
void    glb_showscreen_init_next_frame();


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
extern "C" int glb_load_save_textures_to_tmp(int bLoad)
{
#ifndef GLB_CONSOLE
	return __g_pRbow->LoadSaveTextureToTmp(bLoad?true:false) ? 1:0;
#endif
return 0;
}



// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
extern "C" void glb_do_pause(int pause)
{
	if(!__g_allowPause)
		pause = FALSE;

	if(__g_paused != (pause?TRUE:FALSE) )
	{
		__g_paused = pause?TRUE:FALSE;
		if(pause)
		{
			CALLBYNAME(DGStr("GLB_ON_PAUSE"));
		}
		else
		{
			CALLBYNAME(DGStr("GLB_ON_RESUME"));
		}
	}
	__g_paused = pause?TRUE:FALSE;
}

extern "C" int glb_allow_escape()
{
	return __g_AllowEscape ? TRUE:FALSE;
}


extern "C" void glbasic_notify_screen_change(int w, int h)
{
	g_debug("glb_notify_screen_size %dx%d\n",w ,h);
	__DG_DESKTOPX = w;
	__DG_DESKTOPY = h;
	__DG_RESX = w;
	__DG_RESY = h;

#ifndef GLB_CONSOLE
	if(__g_pRbow)
	{
		__g_pRbow->m_WND.width = w;
		__g_pRbow->m_WND.height= h;
	//	__g_pRbow->SetScreen(w,h,0, __g_pRbow->m_WND.fullscreen); // would change rotation
	}
#endif
}

extern "C" void glbasic_get_screen_size(int* w, int* h)
{
	if(w) *w = __DG_RESX;
	if(h) *h = __DG_RESY;
}



// ------------------------------------------------------------- //
// unpack/unzip the file before you try to read access it
BOOL glb_prepare_reading_file(DGStr& filename)
{	if(filename.len()==0)
		return FALSE;

	// STDOUT(CGStr("glb_prepare_reading_file: ")+filename);
#ifdef WANT_SHOEBOX
	DGStr TempFileName;
	if(!__g_ShoeBox_Data->FindOrExtract(filename, TempFileName.getbuffer(1024), true))
	{
	#ifdef ANDROID
				TempFileName.releasebuffer();
		DGStr checkfilename=filename;
// GF: 2016-08-11 absolute path check is done in CheckAssetFile()
//				checkfilename=REPLACE_Str(GETCURRENTDIR_Str()+filename, DGStr(android_files_dir()), CGStr(""));
//				checkfilename=REPLACE_Str(checkfilename, CGStr("//"), CGStr("/"));
//				checkfilename=REPLACE_Str(checkfilename, CGStr("/Media/"), CGStr("Media/"));
				//STDOUT(checkfilename);
		if(!DOESFILEEXIST(checkfilename)) // This will unpack the file from the APK, if fails to extract from above.
				{	return FALSE;
		}
				//STDOUT(CGStr("3: "));
		return TRUE;
	#else
				TempFileName.releasebuffer();
				//STDOUT(CGStr("2: "));
		return FALSE;
	#endif
	}

	TempFileName.releasebuffer();
	filename = TempFileName;
#endif

	return TRUE;
}




// ------------------------------------------------------------- //
void	AUTOPAUSE(DGNat mode)
{
#if defined (TARGET_OS_IPHONE)
    mode=TRUE; // AUTOPAUSE REQUIRED TO BEEN OFF ON IOS TO AVOID RESUME ISSUES!!!
#endif
	__g_allowPause = mode ? TRUE:FALSE;

	if(!__g_allowPause && __g_paused)
		glb_do_pause(FALSE);
}




// ERROR HANDLING
// -------------------------------------------
/*
#define	CMP_OK					(0x00000)
#define	CMP_FOR_WITHOUT_NEXT	(0x00001)
#define	CMP_SYNTAX_ERROR		(0x00002)
#define CMP_NO_FILE				(0x00003)
#define CMP_WRONG_ARGUMENT		(0x00004)
#define CMP_STRING_TOO_LONG		(0x00005)
#define CMP_DIVISION_BY_ZERO	(0x00006)
#define CMP_OUT_OF_MEMORY		(0x00007)
#define CMP_WRONG_DIMENSION		(0x00008)
#define CMP_OUT_OF_DIMENSION	(0x00009)
#define CMP_OUT_OF_DATA			(0x00014)
*/
DGNat			__g_debugline;
DGNat			__g_error = CMP_OK;
DGNat			__g_errorline;

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGNat	__Error(DGNat code)
{
	if(code != CMP_OK)
	{	__g_error=code; // Set for GETLASTERROR()
		__g_errorline=__g_debugline;
		#ifdef WIN32 // Space Fractal :: This can been insanly annoying on iOS and Android, due there can been many of those. So Windos only
		STDOUT(DGStr("error: ") + HEX_Str(code) + DGStr("\n"));
		#endif
	}

return code;
}


void	CLEARERROR()
{
	__g_error=CMP_OK;
}

DGStr	GETLASTERROR_Str()
{
	const char* pMsg="unknown error";
	switch(__g_error)
	{
		case CMP_OK:      pMsg="success"; break;
		case CMP_NO_FILE: pMsg="no such file"; break;
		case CMP_OUT_OF_DIMENSION: pMsg="out of bounds"; break;
		case CMP_OUT_OF_MEMORY: pMsg="out of memory"; break;
		case CMP_OUT_OF_DATA: pMsg="out of data"; break;
	}
	char err[1024];
	sprintf(err, "%.4d %s", __g_error, pMsg);
	return DGStr(err);
}

	//  *** DLLCALL ***
	struct __DELAY_DLL
	{
		const char* dll, *func;
		void** pPointer;
	};

	class __DLLCALLS
	{
	public:
		__DLLCALLS(){}
		~__DLLCALLS()
		{
			links.Clear();
			for(int i=0; i<modules.Size(); ++i)
			{
				FreeLibrary((HINSTANCE)modules[i]);
			}
			modules.Clear();
		}
		Queue<void*> modules;
		Queue<DGStr> dlls;
		Queue<__DELAY_DLL> links;
	};
	__DLLCALLS& __dll_calls()
	{
		static __DLLCALLS ddd;
		return ddd; 
	}

	void DLLCALL(const char* dll, const char* func, void** pPointer)
	{
		__g_Did3D = true;
		__g_DidNET= true;
		*pPointer = NULL;
		__DELAY_DLL lnk;
		lnk.dll=dll;
		lnk.func=func;
		lnk.pPointer=pPointer;
		__dll_calls().links.Push(lnk);
	}

	void DLL_DELAY_CALL(const char* dll, const char* func, void** pPointer)
	{
		HMODULE pDll=NULL;
		// No case in Win32
		DGStr lowerdll = LCASE_Str(CGStr(dll));
		// Find the dll already loaded
		for(int i=0; i<__dll_calls().dlls.Size(); ++i)
		{
			if(*__dll_calls().dlls.GetAt(i) == lowerdll)
			{
				pDll = (HMODULE)*__dll_calls().modules.GetAt(i);
				break;
			}
		}
		
		char* pFuncName=(char*)func;
		#ifdef UNICODE
			wchar_t pDllName[1024];
			wchar_t pFuncNameW[256];
			mbstowcs(pDllName,  dll, strlen(dll)+1);
			mbstowcs(pFuncNameW, func, strlen(func)+1);
		#else
			const char* pDllName = dll;
		#endif
		// must load dll
		if(!pDll)
		{
			pDll = LoadLibrary(pDllName);

			#ifdef MACOSX
			if(!pDll)
			{
				char exe[1024];
				GFFileSystem::GetExeDir(exe);
				strcat(exe, "/Contents/MacOS/");
				strcat(exe, pDllName);
				pDll = LoadLibrary(exe);
				if(!pDll)
				{
					g_debug("Error: %d %s\n", errno, strerror(errno));
					g_debug("Retry: %s = %p\n", exe, pDll);
				}
			}
			#endif


			// loaded? store pointer to it and get proc address
			if(pDll)
			{
				__dll_calls().modules.Push(pDll);
				__dll_calls().dlls.Push(lowerdll);
				#if defined( _DEBUG)
				g_debug("LoadLibrary: %s = %p\n", dll, pDll);
				#endif
			}
			else
			{
				DGStr err = CGStr("DECLARE module not found: ") + lowerdll;
				WIN32_ONLY( DEBUG(err+CGStr("\n")) );
				g_debug("%s\n", err.c_str());
			}
		}
		if(pDll)
		{
			#if defined(UNICODE) && defined(_WIN32_WCE)
				*pPointer = (void*)GetProcAddressW(pDll, pFuncNameW);
			#else
				*pPointer = (void*)GetProcAddress(pDll, pFuncName);
			#endif
			if(!*pPointer)
			{
				DGStr err = CGStr("DECLARE entry not found: ")+CGStr(func);
				WIN32_ONLY( DEBUG(err+CGStr("\n")) );
				g_debug("%s\n", err.c_str());
				#if defined( _DEBUG) || !defined(WIN32)
				g_debug("Can't find \"%s\" in %s\n", pFuncName, pDllName);
				#ifdef _DEBUG
				GFFileSystem::StrLastError();
				#endif
				#endif
			}
		}
	}
	void DLL_DELAY_LOAD_NOW()
	{
		if(__dll_calls().links.Size())
		{
			g_debug("Delay load DLLs\n");
			for(int i=0; i<__dll_calls().links.Size(); ++i)
			{
				__DELAY_DLL* call = __dll_calls().links.GetAt(i);
				DLL_DELAY_CALL(call->dll, call->func, call->pPointer);
			}
		}
	}
	// ERROR_ALREADY_EXISTS
	// 0x000000b7
	
DGStr __Bytes2Hex(const unsigned char* pHex, int nBytes, int bUppercase=TRUE)
{
	DGStr str;
	if(!nBytes) return str;

	str.alloc(nBytes*2, false);
	const char *hexs;
	if(bUppercase)
		hexs = "0123456789ABCDEF";
	else
		hexs = "0123456789abcdef";
	char a_word[3]; a_word[2]='\0';
	int j=0;
	for(int i=0; i<(int)nBytes; ++i)
	{
		str[j++] = hexs[(*pHex>>4)];
		str[j++] = hexs[*pHex & 0x0f];
		++pHex;
	}
	str[j] = '\0';
	str.releasebuffer();
return str;
}


extern const char* GLB_LIC_MAIL; // = "allen.edward.webb@googlemail.com";
// code : abcd is: ba0dc0 (0=0 character)
extern const char* GLB_LIC_CODE; // = "61156fd561821633999dfd33259570f73d64c795cb18ea98ed43b143e53e8250f4cbf06cdffdc549";



#define __g_LicenseBits	OpenGLRainbows10FreeMemoriesEiiiii // harder to find in HTML5
extern int __g_LicenseBits;

// CF072101E981B41ADEFB6C617216EB4DFEE462B33934A0AD  = 48
#define PrepLic	OpenGLRainbows10AllocMemoriesEiii // harder to find in HTML5
void PrepLic()
{
	int len, stlen;

	stlen=(int)strlen(GLB_LIC_CODE);
	if(stlen == 0) return;

	len=(stlen*2)/3;
	len+=(len%8);
	BYTE* text=new BYTE[len*4];

	memset(text, 0, len*4);
	for(int i=0; i<stlen/3; i++)
	{
		int n=0;
		#define HEXOF(a) ((a)>'9' ? ((a)-'a'+10) : ((a)-'0'))
		n=HEXOF(GLB_LIC_CODE[i*3+1])*16 + HEXOF(GLB_LIC_CODE[i*3]); // Code has each byte flipped
		text[i]=(BYTE)n;
	}
	#undef HEXOF

	CBlowFish fish;
	BYTE code[32];
	memcpy(code, "dP~P.3+"  "\xd6" "\xdf" /*""*/ , 9+1); // "d9~R.E+"
	code[1]='9';
	code[3]='R';
	code[5]='E';
	
	fish.Initialize(code, 9);
	fish.Decode(text, text, len/2);
	
	__g_LicenseBits = 0;

	char* mail = new char[strlen(GLB_LIC_MAIL)/2+1];
	for(int i=0; i<(int)strlen(GLB_LIC_MAIL)/2; i++)
	{
		mail[i] = GLB_LIC_MAIL[i+i];
		mail[i+1]='\0';
	}

	if(stricmp((const char*)(text+4), mail)==0)
	{
		// D3D,D3N,DNC,DNT,SCO
		// 3D:  3N,3D
		// NET: 3N,NT
		if(strncmp((const char*)(text), "SCO", 3)==0)
		{
			__DG_SCHOOLVER = 1;
			__g_LicenseBits |= 3;
		}
		if(text[1] == '3') // 3D
			__g_LicenseBits |= 1;
		if(text[1] == 'N' || text[2] == 'N') // NET
			__g_LicenseBits |= 2;
	}
	delete [] mail;
	delete [] text;
}

DGStr ENCRYPT_Str(DGStr code, DGStr text)
{
DGStr out;
	CBlowFish fish;
//	code += CGStr("x"); // you no make glbasic license code
	fish.Initialize((const BYTE*)code.c_str(), code.len());
	int len = (int)fish.GetOutputLength(text.len()+1);

	BYTE* pOut = (BYTE*)out.getbuffer(len+1);
	len = (int)fish.Encode((BYTE*)text.c_str(), pOut, text.len());
	pOut[len] = '\0';
	out.setlength(len);
	out = __Bytes2Hex((const BYTE*)out.c_str(), len, FALSE);
	return out;
}

DGStr DECRYPT_Str(DGStr code, DGStr text)
{
	text.lower();
//	code += CGStr("x");

	// Hex to binary
	int stlen=text.len();
	int len=stlen/2;
	len+=(len%8);
	BYTE* btext=new BYTE[len*4];

	memset(btext, 0, len*4);
	for(int i=0; i<stlen/2; i++)
	{
		int n=0;
		#define HEXOF(a) ((a)>'9' ? ((a)-'a'+10) : ((a)-'0'))
		n=HEXOF(text.c_str()[i*2])*16 + HEXOF(text.c_str()[i*2+1]);
		btext[i]=(BYTE)n;
	}
	#undef HEXOF

	CBlowFish fish;
	fish.Initialize((const BYTE*)code.c_str(), code.len());
	
	DGStr out;
	fish.Decode((BYTE*)btext, (BYTE*)btext, len);
	btext[len] = '\0';
	out.assign((const char*)btext);
	return out;
}


int __gf_TwoHex2Int(char c1, char c2) {
    int Hi = c1, Lo = c2 , Result;
    
    if ('0'<=Hi && Hi<='9')      Hi -= '0';
    else if ('a'<=Hi && Hi<='f') Hi -= ('a'-10);
    else if ('A'<=Hi && Hi<='F') Hi -= ('A'-10);
    
    if ('0'<=Lo && Lo<='9')      Lo -= '0';
    else if ('a'<=Lo && Lo<='f') Lo -= ('a'-10);
    else if ('A'<=Lo && Lo<='F') Lo -= ('A'-10);

    Result = Lo + 16*Hi;
    return Result;
}

inline bool __gf_isxdigit(char a)
{
	return ((a>='0' && a<='9') || (a>='a'&&a<='f') || (a>='A'&&a<='F'));
}





// http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8
void _utf8decode_glb(DGStr& str)
{
DGStr dest;
    int i = 0;
    const char* s = str.c_str();

    while (i<(int)str.len())
    {
        const wchar_t c = (wchar_t)(unsigned char)s[i++];

        // U-0 to U-7F 
        if ((c&0x80) == 0x00)
        {
            dest.addchar((char)c);
            continue;
        }

        // U-80 to U-7FF
        else if ((c&0xE0) == 0xC0)
        {
            if (i<(int)str.len())
            {
                const wchar_t d = s[i++];
                dest.addchar(( (c&0x1f)<<6 | (d&0x3f) ) );
                continue;
            }
        }

        // U-800 to U-FFFF
        else if ((c&0xF0) == 0xE0)
        {
            if (i+1<(int)str.len())
            {
                const wchar_t d = s[i++];
                const wchar_t e = s[i++];
                dest.addchar(( (c&0x0f)<<12 | (d&0x3f)<<6 | (e&0x3f) ));
                continue;
            }
        }
		// three continuation (U-10000 to U-10ffff) - required for greek!
		else if ((c & 0xF8) == 0xF0)
		{
            if (i+2<(int)str.len())
            {
                const wchar_t d = s[i++];
                const wchar_t e = s[i++];
                const wchar_t f = s[i++];
                dest.addchar(( ((c&0x0f)<<18 | (d&0x3f)<<12 | (e&0x3f)<<6 | (f&0x3f)) + 655536 ));
                continue;
            }
		}
    }
    str=dest;
}





DGStr URLENCODE_Str(const DGStr& SrcString)
{
    DGStr DestString;

    int len = LEN(SrcString);
    for(int i=0; i<len; i++)
    {
        unsigned char c = (unsigned char)SrcString[i];
        //if( isalnum(c) )  // Alpha Numeric Char
        if (  (c>=48 && c<= 57) || ( c>= 65 && c<= 90) || (c>= 97 && c<= 122) ) {
            DestString.addchar(c);
        } 
//        else if(c == ' ') { // Space Char
//            DestString.addchar('+');
//        } 
        else {              // Other Char (also space goes %20 )
            char temp[16];
			if(c > 127)
			{
				// simple ISO8859-1 to utf-8
				sprintf(temp, "%%%02x%%%02x", 0xc2 + (c > 0xbf), (c & 0x3f)+0x80);
				DestString += temp;
			}
			else
			{
				sprintf(temp, "%%%02x", (unsigned char)c);
				DestString += temp;
			}
        }
    }
    return DestString;
}

DGStr URLDECODE_Str(const DGStr& SrcString)
{
    DGStr DestString;
    int len = LEN(SrcString);
    for(int i=0; i<len; i++)
    {
        if(SrcString[i]== '%'){      // Non Alpha Numeric Char
            i++;
            if (__gf_isxdigit(SrcString[i]) && __gf_isxdigit(SrcString[i+1])) {
				unsigned char code = __gf_TwoHex2Int(SrcString[i], SrcString[i+1]);
                DestString.addchar(code);
                i++;
            }
        }
        else if(SrcString[i]=='+') { // Space Char - old style
            DestString.addchar(' ');
        }
        else {                       // Other Char
            DestString.addchar(SrcString[i]);
        }
    }
	_utf8decode_glb(DestString);
    return DestString;
}





#if 0
#if defined(EMSCRIPTEN)
	#define mymemcpy memcpy
#else
	#pragma warning(disable: 4311)
	static void* mymemcpy(void* dst0, const void* src0, size_t len0)
	{
		// Nonzero if either X or Y is not aligned on a "long" boundary.
		#define MEMCPY_UNALIGNED(X, Y) \
		  (((long)X & (sizeof (long) - 1)) | ((long)Y & (sizeof (long) - 1)))

		// How many bytes are copied each iteration of the 4X unrolled loop.
		#define MEMCPY_BIGBLOCKSIZE    (sizeof (long) << 2)

		// How many bytes are copied each iteration of the word copy loop.
		#define MEMCPY_LITTLEBLOCKSIZE (sizeof (long))
		// Threshhold for punting to the byte copier.
		#define MEMCPY_TOO_SMALL(LEN)  ((LEN) < MEMCPY_BIGBLOCKSIZE)

	  char *dst = (char*)dst0;
	  const char *src = (const char*)src0;
	  long *aligned_dst;
	  const long *aligned_src;
	  int   len =  (int)len0;

	  // If the size is small, or either SRC or DST is unaligned,
	  // then punt into the byte copy loop.  This should be rare.
	  if (!MEMCPY_TOO_SMALL(len) && !MEMCPY_UNALIGNED (src, dst))
		{
		  aligned_dst = (long*)dst;
		  aligned_src = (long*)src;

		  // Copy 4X long words at a time if possible.
		  while (len >= MEMCPY_BIGBLOCKSIZE)
			{
			  *aligned_dst++ = *aligned_src++;
			  *aligned_dst++ = *aligned_src++;
			  *aligned_dst++ = *aligned_src++;
			  *aligned_dst++ = *aligned_src++;
			  len -= MEMCPY_BIGBLOCKSIZE;
			}

		  // Copy one long word at a time if possible.
		  while (len >= MEMCPY_LITTLEBLOCKSIZE)
			{
			  *aligned_dst++ = *aligned_src++;
			  len -= MEMCPY_LITTLEBLOCKSIZE;
			}

		   // Pick up any residual with a byte copier.
		  dst = (char*)aligned_dst;
		  src = (char*)aligned_src;
		}
	  while (len--)
		*dst++ = *src++;

	  return dst0;
	}
#endif
#else
	#define mymemcpy memcpy // x64 and whatnot

#endif
	
#if defined(__arm__)
	#define STR_BLOCK_SIZE	2 // *4 bits to shift (4 chars)
#else
	#define STR_BLOCK_SIZE	4 // *16 (16 chars)
#endif


static char DGSTR_NULL[4] = {'\0', '\0', '\0', '\0'};


// faster than the create temp and swap (can reuse old memory a lot)
DGStr& DGStr::operator=(const DGStr& a)
{
	if(a.m_dat == this->m_dat)
		return *this;

	if(a.m_length)
	{
		alloc(a.m_length, false);
		m_length=a.m_length;
		mymemcpy(m_dat, a.m_dat, m_length+1);
	}
	else
	{
		if(m_dat != DGSTR_NULL)
			m_dat[0] = '\0';
		m_length=0;
	}
	return *this;
}

// not supported - ignore, just put it in
#if 1
// #if defined(ANDROID) || defined(EMSCRIPTEN) || defined(PANDORA) || defined(RASPBERRY_PI) || defined(TARGET_OS_IPHONE)

// taken from http://source.winehq.org/source/dlls/ntdll/string.c
char * _i64toa(
    int64 value,    /* [I] Value to be converted */
    char *str,      /* [O] Destination for the converted value */
    int radix)      /* [I] Number base for conversion */
{
    uint64 val;
    int negative;
    char buffer[65];
    char *pos;
    int digit;

    if (value < 0 && radix == 10) {
        negative = 1;
        val = -value;
    } else {
        negative = 0;
        val = value;
    } /* if */

    pos = &buffer[64];
    *pos = '\0';

    do {
        digit = val % radix;
        val = val / radix;
        if (digit < 10) {
            *--pos = '0' + digit;
        } else {
            *--pos = 'a' + digit - 10;
        } /* if */
    } while (val != 0L);

    if (negative) {
        *--pos = '-';
    } /* if */

    memcpy(str, pos, &buffer[64] - pos + 1);
    return str;
}

int64 _atoi64( const char *str )
{
    int64 RunningTotal = 0;
    char bMinus = 0;

    while (*str == ' ' || (*str >= '\011' && *str <= '\015')) {
        str++;
    } /* while */

    if (*str == '+') {
        str++;
    } else if (*str == '-') {
        bMinus = 1;
        str++;
    } /* if */

    while (*str >= '\0' && *str <= '9') {
        RunningTotal = RunningTotal * 10 + *str; // - '';
        str++;
    } /* while */

    return bMinus ? -RunningTotal : RunningTotal;
}

#endif


DGStr::DGStr()				:m_length(0),m_buflen(0),m_dat(DGSTR_NULL)	{} // don't !! memcpy the '\0' character. Might break some times!?
DGStr::DGStr(const char* c)	:m_length(0),m_buflen(0),m_dat(DGSTR_NULL)	{unsigned int nlen=(unsigned int)strlen(c);        if(nlen){alloc(nlen,false);      m_length=nlen;        mymemcpy(m_dat, c,         m_length); m_dat[m_length]='\0';}}
DGStr::DGStr(const wchar_t* c):m_length(0),m_buflen(0),m_dat(DGSTR_NULL)	{assign(c);}
DGStr::DGStr(CGStr a)		:m_length(0),m_buflen(0),m_dat(DGSTR_NULL)	{unsigned int nlen=(unsigned int)strlen(a.c_str());if(nlen){alloc(nlen, false);     m_length=nlen;        mymemcpy(m_dat, a.c_str(), m_length); m_dat[m_length]='\0';}}
DGStr::DGStr(const DGStr& a):m_length(0),m_buflen(0),m_dat(DGSTR_NULL)	{if(a.m_length){alloc(a.m_length,false); m_length=a.m_length; mymemcpy(m_dat, a.m_dat,   m_length); m_dat[m_length]='\0';}  }
DGStr::DGStr(int a)			:m_length(0),m_buflen(0),m_dat(DGSTR_NULL)	{alloc(15,false); _itoa(a,m_dat, 10);   releasebuffer();}
DGStr::DGStr(int64 a)		:m_length(0),m_buflen(0),m_dat(DGSTR_NULL)	{alloc(63,false); _i64toa(a,m_dat, 10);   releasebuffer();}
DGStr::DGStr(float a)		:m_length(0),m_buflen(0),m_dat(DGSTR_NULL)	{alloc(63,false); floattochar(a,m_dat); releasebuffer();} // gcvt(a, b, c) == sprintf(c, "%bg", a);
DGStr::DGStr(double a)		:m_length(0),m_buflen(0),m_dat(DGSTR_NULL)	{alloc(63,false); floattochar(a,m_dat); releasebuffer();}
																		
void DGStr::clear()														{
																			if(m_dat!=DGSTR_NULL) delete[]m_dat;
																			m_dat = DGSTR_NULL;
																			m_length=0; m_buflen=0;
																		}
void DGStr::releasebuffer()
{
	for(m_length=0; m_dat[m_length]!='\0' && m_length<m_buflen; ++m_length)
	{}
	if(m_length>=m_buflen) // uh-oh! non 0 terminated string!
	{
		#ifdef _DEBUG
		*(long*)0=42;
		#endif
		if(m_length>0)
			m_dat[ --m_length ] = '\0';
	}
}



void DGStr::alloc(unsigned int n, bool keep_content)
{
	if(n<=0)
		return;
	++n; // space for '\0' character
	if(m_buflen<n)
	{
		m_buflen = (((n)>>STR_BLOCK_SIZE)+1) << STR_BLOCK_SIZE;  // (1+n/STR_BLOCK_SIZE) * STR_BLOCK_SIZE;
		char* pD = new char[m_buflen];
		if(m_length==0 || !keep_content)
		{
			#if STR_BLOCK_SIZE>1
				*(long*)pD = 0;
			#else
				*pD='\0';
			#endif
		}
		else
		{
			mymemcpy(pD, m_dat, m_length); // strcpy(pD, dat);
			pD[m_length] = '\0';
		}
		if(m_dat!=DGSTR_NULL)
		{
			delete[] m_dat;
		}
		m_dat = pD;
	}
}

void DGStr::assignlowmem(const char* c)	{
											if(m_dat!=DGSTR_NULL) delete[]m_dat;
											m_length = (unsigned int)strlen(c);
											if(m_length)
											{
												m_buflen=m_length+1;
												m_dat=new char[m_buflen];
												mymemcpy(m_dat, c, m_length);
												m_dat[m_length]='\0';
											}
											else
											{
												m_dat = DGSTR_NULL;
												m_buflen=0;
											}
										}


#ifdef _DEBUG
	char DGStr::operator[](int n)const
	{
		if(n && n>=(int)m_buflen)
			*(int*)0=42;
		return m_dat[n];
	}
	char& DGStr::operator[](int n)
	{
		if(n && n>=(int)m_buflen)
			*(int*)0=42;
		return m_dat[n];
	}
#endif


void DGStr::upper()
{
	char* pEnd = m_dat + m_length;
	for (char* pC = m_dat; pC<pEnd; ++pC)
	{
		*pC = toupper(*pC);
	}
}

void DGStr::lower()
{
	char* pEnd = m_dat + m_length;
	for (char* pC = m_dat; pC<pEnd; ++pC)
	{
		*pC = tolower(*pC);
	}
}

/* - others seem faster
DGStr& DGStr::operator=(CGStr a)		{m_length=(unsigned int)strlen(a.c_str()); if(m_length) {alloc(m_length,false); memcpy(m_dat, a.c_str(), m_length+1);}else{clear();} return *this;}
DGStr& DGStr::operator=(const char* c)	{m_length=(unsigned int)strlen(c);         if(m_length) {alloc(m_length,false); memcpy(m_dat, c, m_length+1);        }else{clear();} return *this;}
DGStr& DGStr::operator=(int a)			{alloc(16,false); _itoa(a, m_dat, 10);  releasebuffer(); return *this;}
DGStr& DGStr::operator=(float a)		{alloc(64,false); floattochar(a, m_dat);releasebuffer(); return *this;} // gcvt(a, b, c) == sprintf(c, "%bg", a);
DGStr& DGStr::operator=(double a)		{alloc(64,false); floattochar(a, m_dat);releasebuffer(); return *this;}
*/

// the DGStr a operator seems faster
// DGStr& DGStr::operator=(const DGStr& a)	{m_length=a.m_length;                      if(m_length) {alloc(m_length,false); memcpy(m_dat, a.m_dat, m_length+1);  }else{clear();} return *this;}

#define MK_OPERATOR_CL(zz) \
DGStr& DGStr::operator zz(const char* str){unsigned int slen = (unsigned int)strlen(str); \
											if(slen)  {alloc(m_length+slen,      true);mymemcpy(m_dat+m_length, str,     slen);       m_length+=slen;       m_dat[m_length]='\0';} return *this;} \
DGStr& DGStr::operator zz(const DGStr& a) {if(a.len()){alloc(m_length+a.m_length,true);mymemcpy(m_dat+m_length, a.m_dat, a.m_length); m_length+=a.m_length; m_dat[m_length]='\0';} return *this;} \
DGStr& DGStr::operator zz(CGStr a)		  {*this += a.m_dat; return *this;} \
DGStr& DGStr::operator zz(int a)		  {DGStr b(a); *this+=b; return *this;} \
DGStr& DGStr::operator zz(int64 a)		  {DGStr b(a); *this+=b; return *this;} \
DGStr& DGStr::operator zz(float a)		  {DGStr b(a); *this+=b; return *this;} \
DGStr& DGStr::operator zz(double a)		  {DGStr b(a); *this+=b; return *this;}
MK_OPERATOR_CL(+=)
MK_OPERATOR_CL(-=)
MK_OPERATOR_CL(*=)
MK_OPERATOR_CL(/=)
MK_OPERATOR_CL(^=)

//DGStr operator+(DGStr& a, const CGStr& b) {strcat(a.dat, b.dat); DGStr c; c = (DGStr)b; return c;}
#define MK_OPERATOR(zz, zzz) \
DGStr operator zz (int a, CGStr b)			{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (int a, const DGStr& b)	{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (int64 a, CGStr b)		{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (int64 a, const DGStr& b)	{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (float a, CGStr b)		{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (float a, const DGStr& b)	{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (double a, CGStr b)		{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (double a, const DGStr& b){DGStr c(a); c zzz b; return c;} \
DGStr operator zz (CGStr a, CGStr b)		{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (CGStr a, const DGStr& b)	{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (CGStr a, int b)			{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (CGStr a, int64 b)		{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (CGStr a, float b)		{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (CGStr a, double b)		{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (const DGStr& a, CGStr b)	{DGStr c(a); c zzz b;  return c;} \
DGStr operator zz (const DGStr& a, const DGStr& b){DGStr c(a); c zzz b; return c;} \
DGStr operator zz (const DGStr& a, int b)	{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (const DGStr& a, int64 b)	{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (const DGStr& a, float b)	{DGStr c(a); c zzz b; return c;} \
DGStr operator zz (const DGStr& a, double b){DGStr c(a); c zzz b; return c;}
MK_OPERATOR(+, +=)
MK_OPERATOR(-, -=)
MK_OPERATOR(*, *=)
MK_OPERATOR(/, /=)
MK_OPERATOR(^, ^=)

#define NUMNUM(zz)	return a zz b ? 1 : 0;
#define STRNUM(zz)	return a.number() zz b ? 1 : 0;
#define STRNAT(zz)	return a.natural() zz b ? 1 : 0;
#define STRNAT64(zz)return a.natural64() zz b ? 1 : 0;
#define NUMSTR(zz)	return a zz b.number() ? 1 : 0;
#define NATSTR(zz)	return a zz b.natural() ? 1 : 0;
#define NAT64STR(zz)return a zz b.natural64() ? 1 : 0;
#define STRSTR(zz)	return (strcmp(a.c_str(), b.c_str()) zz 0) ? 1 : 0;
#define STRNSTR(zz, nn)	return (memcmp(a.c_str(), b.c_str(), nn) zz 0) ? 1 : 0;

#undef MK_OPERATOR
#define MK_OPERATOR(zz) \
bool operator zz (int a, CGStr b)			{NATSTR(zz)} \
bool operator zz (int a, const DGStr& b)	{NATSTR(zz)} \
bool operator zz (int64 a, CGStr b)			{NAT64STR(zz)} \
bool operator zz (int64 a, const DGStr& b)	{NAT64STR(zz)} \
bool operator zz (float a, CGStr b)			{NUMSTR(zz)} \
bool operator zz (float a, const DGStr& b)	{NUMSTR(zz)} \
bool operator zz (double a, CGStr b)		{NUMSTR(zz)} \
bool operator zz (double a, const DGStr& b)	{NUMSTR(zz)} \
bool operator zz (CGStr a, CGStr b)			{STRSTR(zz)} \
bool operator zz (CGStr a, const DGStr& b)	{STRSTR(zz)} \
bool operator zz (CGStr a, int b)			{STRNAT(zz)} \
bool operator zz (CGStr a, int64 b)			{STRNAT64(zz)} \
bool operator zz (CGStr a, float b)			{STRNUM(zz)} \
bool operator zz (CGStr a, double b)		{STRNUM(zz)} \
bool operator zz (const DGStr& a, CGStr b)	{STRSTR(zz)} \
bool operator zz (const DGStr& a, const DGStr& b)	{STRSTR(zz)} \
bool operator zz (const DGStr& a, int b)	{STRNAT(zz)} \
bool operator zz (const DGStr& a, int64 b)	{STRNAT64(zz)} \
bool operator zz (const DGStr& a, float b)	{STRNUM(zz)} \
bool operator zz (const DGStr& a, double b)	{STRNUM(zz)}
MK_OPERATOR(<=)
MK_OPERATOR(>=)
MK_OPERATOR(!=)
MK_OPERATOR(<)
MK_OPERATOR(>)
#undef MK_OPERATOR

// The == operator should check for equal length first - much faster!
bool operator == (const DGStr& a, const DGStr& b)	{if(a.len()!=b.len()) return 0; STRNSTR(==, a.len());}
bool operator == (CGStr a, const DGStr& b)			{if(a.len()!=b.len()) return 0; STRNSTR(==, a.len());}
bool operator == (const DGStr& a, CGStr b)			{if(a.len()!=b.len()) return 0; STRNSTR(==, a.len());}
bool operator == (int a, CGStr b)			{NATSTR(==)}
bool operator == (int a, const DGStr& b)	{NATSTR(==)}
bool operator == (int64 a, CGStr b)			{NAT64STR(==)}
bool operator == (int64 a, const DGStr& b)	{NAT64STR(==)}
bool operator == (float a, CGStr b)			{NUMSTR(==)}
bool operator == (float a, const DGStr& b)	{NUMSTR(==)}
bool operator == (double a, CGStr b)		{NUMSTR(==)}
bool operator == (double a, const DGStr& b)	{NUMSTR(==)}
bool operator == (CGStr a, CGStr b)			{STRSTR(==)}
bool operator == (CGStr a, int b)			{STRNAT(==)}
bool operator == (CGStr a, int64 b)			{STRNAT64(==)}
bool operator == (CGStr a, float b)			{STRNUM(==)}
bool operator == (CGStr a, double b)		{STRNUM(==)}
bool operator == (const DGStr& a, int b)	{STRNAT(==)}
bool operator == (const DGStr& a, int64 b)	{STRNAT64(==)}
bool operator == (const DGStr& a, float b)	{STRNUM(==)}
bool operator == (const DGStr& a, double b)	{STRNUM(==)}

// returns a static temp buffer -> don't use twice in one call!
static wchar_t g_w_str_buffer[512];
const wchar_t* DGStr::w_str()const
{
	mbstowcs(g_w_str_buffer, c_str(), len()+1);
	return g_w_str_buffer;
}

// DGStr::operator const char*()const	{return this->dat;}
void	DGStr::left(long i)				{if((int)i>=0 && (int)i<(long)m_length) {m_dat[(int)i]='\0'; m_length=(int)i;} }

char*	DGStr::getbuffer(unsigned int size) {alloc(size,true); return m_dat;}

int		DGStr::compare(const char* s)const	{return strcmp(m_dat, s);}
int		DGStr::compare(DGInt i)const	{DGInt a=(DGInt)this->number();  if(a<i) return -1; else if(a>i) return 1; return 0;}
int		DGStr::compare(DGNat i)const	{DGNat a=(DGNat)this->natural(); if(a<i) return -1; else if(a>i) return 1; return 0;}

void	DGStr::trimright()
{
	if(m_length<=0) return;
	signed int i, end; // SIGNED!!
	end=m_length-1;
	for(i=end; i>=0; --i)
	{
		if ((unsigned)m_dat[i]<32) // trim any non visible characters
		{
			m_dat[i]='\0';
			--m_length;
		}
		else
			break;
	}
}
DGInt CGStr::number()const					{return (DGInt)atof(m_dat);}
DGInt DGStr::number()const					{return (DGInt)atof(m_dat);}
DGNat CGStr::natural()const					{return (DGNat)atoi(m_dat);}
DGNat DGStr::natural()const					{return (DGNat)atoi(m_dat);}
int64 CGStr::natural64()const				{return (int64)_atoi64(m_dat);}
int64 DGStr::natural64()const				{return (int64)_atoi64(m_dat);}
unsigned int CGStr::len()const				{return (unsigned int)strlen(m_dat);}

void  DGStr::assign(const char* c)			{DGStr tmp(c); swap(tmp);}
void DGStr::assign(const wchar_t* c)
{
	unsigned int len = (unsigned)wcslen((const wchar_t*)c) + 1;
	alloc(len, false);
	_wcstombsz(m_dat, (const wchar_t*)c, len);
	m_length=len;
}
// append a single character
void DGStr::addchar(char c)
{
	alloc(m_length+1, true);
	m_dat[m_length++]=c;
	m_dat[m_length]='\0';
}



/* - can't do '\0' characters
inline const char* __qstrstr(const char *phaystack, const char *pneedle)
{
  register const unsigned char *haystack, *needle;
  register unsigned b, c;

  haystack = (const unsigned char *) phaystack;
  needle = (const unsigned char *) pneedle;

  b = *needle;
  if (b != '\0')
    {
      haystack--;				// possible ANSI violation
      do
	{
	  c = *++haystack;
	  if (c == '\0')
	    goto ret0;
	}
      while (c != b);

      c = *++needle;
      if (c == '\0')
	goto foundneedle;
      ++needle;
      goto jin;

      for (;;)
        {
          register unsigned a;
	  register const unsigned char *rhaystack, *rneedle;

	  do
	    {
	      a = *++haystack;
	      if (a == '\0')
		goto ret0;
	      if (a == b)
		break;
	      a = *++haystack;
	      if (a == '\0')
		goto ret0;
shloop:;    }
          while (a != b);

jin:	  a = *++haystack;
	  if (a == '\0')
	    goto ret0;

	  if (a != c)
	    goto shloop;

	  rhaystack = haystack-- + 1;
	  rneedle = needle;
	  a = *rneedle;

	  if (*rhaystack == a)
	    do
	      {
		if (a == '\0')
		  goto foundneedle;
		++rhaystack;
		a = *++needle;
		if (*rhaystack != a)
		  break;
		if (a == '\0')
		  goto foundneedle;
		++rhaystack;
		a = *++needle;
	      }
	    while (*rhaystack == a);

	  needle = rneedle;		   // took the register-poor approach

	  if (a == '\0')
	    break;
        }
    }
foundneedle:
  return (const char*) haystack;
ret0:
  return NULL;
}
*/

inline int __qmemcmp(const unsigned char* p1, const unsigned char* p2, unsigned int len)
{
	const unsigned int* pi1=(const unsigned int*)p1;
	const unsigned int* pi2=(const unsigned int*)p2;
	while(len>3)
	{
		if(*pi1++ != *pi2++)
			return 1;
		len-=4;
	}
	if(len > 0)
	{
		p1=(const unsigned char*)pi1;
		p2=(const unsigned char*)pi2;

		if(*p1 != *p2) return 1;
		--len; ++p1; ++p2;
		if(len > 0)
		{
			if(*p1 != *p2) return 1;
			--len; ++p1; ++p2;
			if(len > 0)
			{
				if(*p1 != *p2) return 1;
			}
		}
	}
	return 0;
}

/* The constant UCHAR_MAX is assumed to contain the maximum
 * value of the input character type. Typically it's 255.
 * size_t is an unsigned type for representing sizes.
 * If your system doesn't have it, replace with
 * unsigned int.
 */
 
/* Returns a pointer to the first occurrence of "needle"
 * within "haystack", or NULL if not found. Works like 
 * memmem().
 */
 
 

int __qmemmem(const unsigned char* haystack, unsigned int hlen, const unsigned char* needle, unsigned int  nlen)
{
    /* Sanity checks on the parameters */
    if (nlen <= 0 || !haystack || !needle)
        return -1;

    unsigned int scan = 0;
    unsigned int bad_char_skip[0xff + 1]; /* Officially called:
                                          * bad character shift */
 
     if(hlen==0) return  0;
     if(nlen>hlen)  return -1;
     if(hlen == nlen)
		return __qmemcmp(haystack, needle, nlen)==0 ?  0 : -1;

 const unsigned char* start=haystack;
 
    /* ---- Preprocess ---- */
    /* Initialize the table to default value */
    /* When a character is encountered that does not occur
     * in the needle, we can safely skip ahead for the whole
     * length of the needle.
     */
    for (scan = 0; scan <= 0xff; scan++)
        bad_char_skip[scan] = nlen;
 
    /* C arrays have the first byte at [0], therefore:
     * [nlen - 1] is the last byte of the array. */
    unsigned int last = nlen - 1;
 
    /* Then populate it with the analysis of the needle */
    for (scan = 0; scan < last; scan++)
        bad_char_skip[needle[scan]] = last - scan;
 
    /* ---- Do the matching ---- */
 
    /* Search the haystack, while the needle can still be within it. */
    while (hlen >= nlen)
    {
        /* scan from the end of the needle */
        for (scan = last; haystack[scan] == needle[scan]; scan = scan - 1)
            if (scan == 0) /* If the first byte matches, we've found it. */
                return (int)(haystack-start);
 
        /* otherwise, we need to skip some bytes and start again. 
           Note that here we are getting the skip value based on the last byte
           of needle, no matter where we didn't match. So if needle is: "abcd"
           then we are skipping based on 'd' and that value will be 4, and
           for "abcdd" we again skip on 'd' but the value will be only 1.
           The alternative of pretending that the mismatched character was 
           the last character is slower in the normal case (Eg. finding 
           "abcd" in "...azcd..." gives 4 by using 'd' but only 
           4-2==2 using 'z'. */
        hlen     -= bad_char_skip[haystack[last]];
        haystack += bad_char_skip[haystack[last]];
    }
 
    return -1;
}






// find c2 in c1, starting at nskip in c1
inline int __qinstr(const char* c1, int len1, const char* c2, int len2, int nskip=0)
{
	if(nskip >= len1) return -1;

	int pos = __qmemmem((const unsigned char*)c1+nskip, len1-nskip, (const unsigned char*)c2, len2);
	if(pos==-1) return -1;
return (int) pos+ nskip;
}

// reverse find c2 in c1, starting at nskip in c1
inline int __rev_qinstr(const char* c1, int len1, const char* c2, int len2, int iafter=-1)
{
	if(iafter <0) iafter = len1-len2;
	if(iafter<0) return -1; // needle>haystack

	int ibest=-1;
	int istart=0;
	for(;;)
	{
 
//  0123456789012345678901234567890123456789012 // 43 len
// "the quick brown fox jumps over the lazy dog"

		int pos = __qmemmem((const unsigned char*)c1+istart, len1-istart, (const unsigned char*)c2, len2);
		if(pos<0) break;
		pos+=istart;
		if(pos>iafter) break;
		ibest=pos;
		istart=pos+1;
	}
	return ibest;
}


// replace in this. Keep *this untouched
DGStr DGStr::replace(const DGStr& sOld, const DGStr& sNew)const
{
	DGStr str(*this);
int	pos=0;
int qpos;

unsigned int	lold=sOld.len();
unsigned int	lnew=sNew.len();
const bool onebyte = (lold==1 && lnew==1) ? true : false;
	for(;;)
	{
		qpos = __qinstr(str.c_str() + pos, str.len()-pos,  sOld.c_str(), lold);
		if (qpos != -1)
		{
			pos+=qpos;
			if(onebyte)
			{
				str.m_dat[pos] = sNew.m_dat[0];
				++pos;
			}
			else if (lold==lnew)
			{
				mymemcpy(str.m_dat+pos, sNew.c_str(), lnew);
				pos+=(lnew);
			}
			else
			{
				str = MID_Str(str, 0, pos) + sNew + MID_Str(str, pos+lold, str.len());
				pos+=(lnew-lold+1);
			}
		}
		else
			break;
	}
return str;
}



// ------------------------------------------------------------- //
// find string in string
// ------------------------------------------------------------- //
DGNat INSTR(CGStr        haystack, CGStr        needle, DGNat start) {return __qinstr(haystack.c_str(), haystack.len(), needle.c_str(), needle.len(), start);}
DGNat INSTR(CGStr        haystack, const DGStr& needle, DGNat start) {return __qinstr(haystack.c_str(), haystack.len(), needle.c_str(), needle.len(), start);}
DGNat INSTR(const DGStr& haystack, CGStr        needle, DGNat start) {return __qinstr(haystack.c_str(), haystack.len(), needle.c_str(), needle.len(), start);}
DGNat INSTR(const DGStr& haystack, const DGStr& needle, DGNat start) {return __qinstr(haystack.c_str(), haystack.len(), needle.c_str(), needle.len(), start);}
// can't handle '\0':
//DGNat INSTR(const char* s1, const char* s2, DGNat start)		{return __qinstr(s1, s2, start);}



// ------------------------------------------------------------- //
// reverse find string in string
// ------------------------------------------------------------- //
DGNat REVINSTR(CGStr        haystack, CGStr        needle, DGNat start)		{return __rev_qinstr(haystack.c_str(), haystack.len(), needle.c_str(), needle.len(), start);}
DGNat REVINSTR(CGStr        haystack, const DGStr& needle, DGNat start)		{return __rev_qinstr(haystack.c_str(), haystack.len(), needle.c_str(), needle.len(), start);}
DGNat REVINSTR(const DGStr& haystack, CGStr        needle, DGNat start)		{return __rev_qinstr(haystack.c_str(), haystack.len(), needle.c_str(), needle.len(), start);}
DGNat REVINSTR(const DGStr& haystack, const DGStr& needle, DGNat start)		{return __rev_qinstr(haystack.c_str(), haystack.len(), needle.c_str(), needle.len(), start);}
// can't handle '\0':
//DGNat INSTR(const char* s1, const char* s2, DGNat start)		{return __rev_qinstr(s1, s2, start);}



// ------------------------------------------------------------- //
// find first char in string
// ------------------------------------------------------------- //
const char* __findOneOf(const char* str, const char* fnd)
{
	while(*str!='\0')
	{
		const char* nfnd = fnd;
		do
		{
			if(*str == *nfnd++) return str;
		} while(*nfnd!='\0');
		++str;
	};
return NULL;
}

// ------------------------------------------------------------- //
// find first char not in string
// ------------------------------------------------------------- //
const char* __findOneNotOf(const char* str, const char* fnd)
{
	while(*str!='\0')
	{
		const char* nfnd = fnd;
		do
		{
			if(*str == *nfnd++) goto skip;
		} while(*nfnd!='\0');
		return str;
		skip:
		++str;
	};
return NULL;
}

// ------------------------------------------------------------- //
// Split String into Array
// ------------------------------------------------------------- //
DGNat SPLITSTR(const DGStr& instr, DGStrArray& arr, const DGStr& splitter, DGNat bSkipEmtpy)
{
const char* pCur, *pEnd, *pTotalEnd;
// Queue<DGStr> vals;
static DGStr		 tmp, str;	
	str = instr;
	pCur = str.c_str();
	pEnd = pCur;
	// avoid byte garbage if str==""
	arr.dim=0;
	arr.max1=0;
	arr.max2=0;
	arr.max3=0;
	arr.max4=0;
	arr.fac2=0;
	arr.fac3=0;
	arr.fac4=0;
	arr.count=0;
	// DIM(arr,0);

	pTotalEnd = pCur + str.len();
	for( ;pCur<pTotalEnd; )
	{
		// skip splitters
		if(bSkipEmtpy)
			pEnd = __findOneNotOf(pCur, splitter.c_str());
		else
			pEnd = pCur;
		if(pEnd) pCur = pEnd;
		
		// find next splitter
		pEnd = __findOneOf(pCur, splitter.c_str());
		if(pEnd)
		{
			// *pEnd = '\0';
			if(*pCur!='\0')
			{
				int len = (int)(pEnd - pCur);
				if(len || !bSkipEmtpy)
				{
					mymemcpy(tmp.getbuffer(len+1), pCur, len);
					tmp[len] = '\0';
					tmp.setlength(len);

					// tmp = pCur;
					DIMPUSH(arr, tmp);
					// vals.Push(tmp);
				}
			}
			pCur = ++pEnd;
		}
		else
		{
			if(pCur && *pCur !='\0')
			{
				tmp = pCur;
				DIMPUSH(arr, tmp);
				// vals.Push(tmp);
			}
			break;
		}
	}
/*
	DIM(arr, vals.Size());
	for(unsigned int i=0; i<vals.Size(); ++i)
	{
		arr(i) = *vals.GetAt(i);
	}
*/
return LEN(arr);
}
//DGInt SPLITSTR(CGStr str, DGStrArray& arr, CGStr splitter) {return SPLITSTR(str, arr, splitter);}

DGStr UCASE_Str(const DGStr& org)
{
	DGStr rv = org;
	rv.upper();
return rv;
}

DGStr LCASE_Str(const DGStr& org)
{
	DGStr rv = org;
	rv.lower();
return rv;
}


#if defined(LINUX) || defined(GPHWIZ) || defined(CAANOO) || defined(PALM_WEBOS)||defined(PALM_PIXI)
	extern "C" int ioctl(int fildes, unsigned long request, ...);
	extern "C" int open(const char *path, int flags, unsigned int mode=0);
	#define	O_RDONLY	 0
	#define O_RDWR       2
	#define I2C_RDWR 0x0707
	#define O_NONBLOCK	 04000
	#define HDIO_GET_IDENTITY 0x307
	struct hd_driveid {
		unsigned int	config;			unsigned int	cyls;			unsigned int	reserved2;		unsigned int	heads;		
		unsigned int	track_bytes;	unsigned int	sector_bytes;	unsigned int	sectors;		unsigned int	vendor0;	
		unsigned int	vendor1;		unsigned int	vendor2;		unsigned char	serial_no[20];	unsigned int	buf_type;
		unsigned int	buf_size;		unsigned int	ecc_bytes;		unsigned char	fw_rev[8];		unsigned char	model[40];	
		unsigned char	max_multsect;	unsigned char	vendor3;		unsigned int	dword_io;		unsigned char	vendor4;	
		unsigned char	capability;		unsigned int	reserved50;		unsigned char	vendor5;		unsigned char	tPIO;		
		unsigned char	vendor6;		unsigned char	tDMA;			unsigned int	field_valid;	unsigned int	cur_cyls;	
		unsigned int	cur_heads;		unsigned int	cur_sectors;	unsigned int	cur_capacity0;	unsigned int	cur_capacity1;	
		unsigned char	multsect;		unsigned char	multsect_valid;	unsigned int	lba_capacity;	unsigned int	dma_1word;	
		unsigned int	dma_mword;		unsigned int  eide_pio_modes;	unsigned int  eide_dma_min;	unsigned int  eide_dma_time;	
		unsigned int  eide_pio;       unsigned int  eide_pio_iordy;	unsigned int	words69_70[2];	unsigned int	words71_74[4];	
		unsigned int  queue_depth;	unsigned int  words76_79[4];	unsigned int  major_rev_num;	unsigned int  minor_rev_num;	
		unsigned int  command_set_1;	unsigned int  command_set_2;	unsigned int  cfsse;			unsigned int  cfs_enable_1;	
		unsigned int  cfs_enable_2;	unsigned int  csf_default;	unsigned int  dma_ultra;		unsigned int	word89;		
		unsigned int	word90;			unsigned int	CurAPMvalues;	unsigned int	word92;			unsigned int	hw_config;	
		unsigned int  words94_125[32];unsigned int	last_lun;		unsigned int	word127;		unsigned int	dlf;
		unsigned int  csfo;			unsigned int	words130_155[26];unsigned int	word156;		unsigned int	words157_159[3];
		unsigned int	words160_255[95];
	};

	static int linux_battery_level()
	{
		int blevel=100;
		FILE *battery;
		int i;
		char b[128];

		battery = fopen ("/proc/apm", "r");
		if(battery != NULL)
		{
			int bloading; // is loading?
			fscanf(battery, "%s %s %s %x %s %x %d%% %s %s",
				b, b, b, &bloading, b, &i, &blevel, b, b);
			//     1.16 1.2 0x03 0x01      0x00 0x01 99%      -1  ?
			fclose(battery);

			// Bit 7 is set if we have a battery (laptop).
			if ((i & 0x80) != 0) blevel = 100;
		}
	return blevel;
	}
#endif // LINUX battery level

#ifdef GP2X
#define I2CCON   (0xD00>>1)
#define I2CSTAT  (0xD02>>1)
#define I2CADDR  (0xD04>>1)
#define I2CDATA  (0xD06>>1)

#define DEVICE  0xA0

int WaitOnI2CInterrupt()
{int n = 10000000;while(n > 0){if(gp2x_memreg16[I2CCON] & 0x10) // wait for interrupt pending flag
      return 1; n--;}return 0;}


// returns 1 on success, 0 on failure
int FetchGP2XSerial(char *pszSerial)
{
  unsigned short usCount = 0;

  // initialize registers
  gp2x_memreg16[I2CCON] = 0xEF; // set clocks, enable interrupt, clear pending flag
  gp2x_memreg16[I2CADDR] = DEVICE; // set address
  gp2x_memreg16[I2CSTAT] = 0x10; // enable I2C

  *pszSerial = 0;

  // tell the eeprom what address we're interested in
  gp2x_memreg16[I2CDATA] = DEVICE;
  gp2x_memreg16[I2CSTAT] = 0xF0; // Master mode, data write
  if(WaitOnI2CInterrupt())
  {
    gp2x_memreg16[I2CDATA] = 0x00; // high byte of address
    gp2x_memreg16[I2CCON] = 0xEF;
    if(WaitOnI2CInterrupt())
    {
      gp2x_memreg16[I2CDATA] = 0x20; // address of serial number
      gp2x_memreg16[I2CCON] = 0xEF;
      if(WaitOnI2CInterrupt())
      {
        // issue the read command
        gp2x_memreg16[I2CDATA] = DEVICE;
        gp2x_memreg16[I2CSTAT] = 0xB0; // Master mode, data read
        gp2x_memreg16[I2CCON] = 0xEF;
        if(WaitOnI2CInterrupt())
        {
          // read the data bytes
          for(usCount = 0; usCount < 24; usCount++)
          {
            gp2x_memreg16[I2CCON] = 0xEF;
            if(!WaitOnI2CInterrupt())
              break;
            *pszSerial++ = (char) gp2x_memreg16[I2CDATA];
            *pszSerial = 0;
          }
        }
      }
    }
  }
  // shut down
  gp2x_memreg16[I2CSTAT] = 0x90;
  gp2x_memreg16[I2CCON] = 0xEF;
  usleep(10000);
  // leave the I2C in a quiet state
  gp2x_memreg16[I2CCON] = 0x40; // disable interrupt
  gp2x_memreg16[I2CSTAT] = 0x00; // disable I2C

  if(usCount == 24)
    return 1;
  return 0;
}
#endif

#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
int gp2x_wiz_battery()
{
	int devbatt;
	unsigned short batt=0;
	devbatt = open("/dev/batt", O_RDONLY);
	if(devbatt<0) // failed (F220 e.g.)
	{
		typedef struct
		{
			unsigned short batt;
			unsigned short remocon;
		} MMSP2ADC;

		// The value batt has one of the following:
		#define BATT_LEVEL_HIGH 0
		#define BATT_LEVEL_MID 1
		#define BATT_LEVEL_LOW 2
		#define BATT_LEVEL_EMPTY 3

		int rv;
		MMSP2ADC val;
		val.batt=BATT_LEVEL_EMPTY;
		devbatt = open("/dev/mmsp2adc", O_RDONLY);
		read (devbatt, &val, sizeof(MMSP2ADC));
		close(devbatt);
		switch(val.batt)
		{
		case BATT_LEVEL_HIGH: return 100;
		case BATT_LEVEL_MID: return 67;
		case BATT_LEVEL_LOW: return 33;
		}
		return 0;
	}
	read (devbatt, &batt, 2);
	close (devbatt);

	/*
		>900 - AC Adapter 
		>745 - Full batteries
		678-745 - Medium batteries
		<678    - Empty batteries
	*/
	// above 745 -> return "full"
	if(batt>745) return 100;
	// between full and empty -> linear interpolation
	if(batt>678)
	{
		return 100.0f * ((float)batt-745.0f) / (745.0f - 678.0f);
	}
	return 0; // empty
}
#endif


#if defined (TARGET_OS_IPHONE)
extern "C"
{
	int   isIphone();
	const char* iPhoneVersion(); // "2.2.1" e.g.
	//const char* iPhoneDeviceID();
	float iPhoneBattery();
	void  iPhoneGetCountryInto(char * str);
	const char* iPhoneGetDocumentsPath();
	const char* iPhoneGetTempPath();
	int   iphone_get_dpi();
}
#endif


#ifdef PALM_WEBOS
extern "C"
{
	typedef enum
	{
		PDL_NOERROR         = 0,
		PDL_EMEMORY,        /* Memory failure */
		PDL_ECONNECTION,    /* Connection error */
		PDL_INVALIDINPUT,   /* Problem with the sent-in params */
		PDL_EOTHER,         /* Other error */
		PDL_UNINIT,         /* The function relies on a client-called init function that has never been called */
		PDL_NOTALLOWED,     /* A configurable option prevents the operation from succeeding (Ex: Location Services disabled by user = GPS won't work) */
		PDL_LICENSEFAILURE, /* The license of the application is invalid or missing. */
		PDL_STRINGTOOSMALL, /* A sent-in string that was to be filled was too small for the data needed */
		PDL_SYSTEMERROR_FILE, /* A OS data file or database external to PDL is missing or malformed*/
		PDL_SYSTEMERROR_NET,  /* Networking is disabled or inavailable*/
		PDL_APPINFO      	 /* PDL was unable to find the appinfo.json file, and thus does know know any of 
    						 the information therein, nor does it know the top level of the app's directory tree. */
	} PDL_Err;
	const char *PDL_GetHardware();
	PDL_Err PDL_GetUniqueID(char *buffer, int bufferLen);
	PDL_Err PDL_GetDataFilePath(const char *dataFileName, char *buffer, int bufferLen);
	PDL_Err PDL_GetLanguage(char *buffer, int bufferLen);
	PDL_Err PDL_LaunchBrowser(const char* Url);
	PDL_Err PDL_Init(unsigned int flags);
	void PDL_Quit();
}
#endif

#ifdef ANDROID
extern "C"
{	int android_battery_level();
	const char* android_device_id();
	const char* android_lang();
	const char* android_ext_storage_pub_dir();
	int android_keyboard();
	int android_get_dpi();
	const char*  android_JAVACALL(const char* string); // Space Fractal
	DGInt CheckAssetFile(DGStr, int); // Space Fractal
	DGStr JavaCall_Str(DGStr); // Space Fractal
}
#endif

#ifdef MACOSX
	extern "C" const char* MacGetDocumentsPath();
	extern "C" const char* MacSettingsDir();
#endif


// -------------------------------------
// PLATFORMINFO$()
// -------------------------------------
DGStr PLATFORMINFO_Str(DGStr WhatInfo)
{
DGStr origInfo = WhatInfo;
	WhatInfo = UCASE_Str(WhatInfo);
	if(WhatInfo[0] == '\0')
	{
		#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
			return DGStr("GP2X");
		#elif defined(PANDORA)
			return DGStr("PANDORA");
		#elif defined(RASPBERRY_PI)
			return DGStr("ARM-LINUX-EGL");
		#elif defined(POCKETPC)
			return DGStr("WINCE");
		#elif defined(TARGET_OS_IPHONE)
			return DGStr("IPHONE");
		#elif defined(MACOSX)
			return DGStr("MACOSX");
		#elif defined(PALM_WEBOS)
			return DGStr("WEBOS");
		#elif defined(ANDROID)
			return DGStr("ANDROID");
		#elif defined(LINUX)
			return DGStr("LINUX");
		#elif defined(EMSCRIPTEN)
			return DGStr("JAVASCRIPT");
		#else
			return DGStr("WIN32");
		#endif
	}
	else if(WhatInfo.compare("DEVICE")==0)
	{
		#ifdef POCKETPC
			unsigned char szPlatform[512];
			#ifndef SPI_GETPLATFORMTYPE
			#define SPI_GETPLATFORMTYPE 257
			#endif
			if(SystemParametersInfo(SPI_GETPLATFORMTYPE, 512, szPlatform, 0))
			{
				DGStr platform;
				#ifdef UNICODE
				platform.assign((const wchar_t*)szPlatform);
				#else
				platform.assign((const char*)szPlatform);
				#endif
				if(platform== CGStr("Smartphone")) return DGStr("SMARTPHONE");
				if(platform== CGStr("PocketPC"))   return DGStr("POCKETPC");
			}
			return DGStr("WINCE");
		#elif defined(GP2X)
			if(pDXin->dev_touchscreen >= 0)
			{
				return DGStr("F200");
			}
			else
			{
				return DGStr("F100");
			}
		#elif defined(GPHWIZ)
			return DGStr("WIZ");
		#elif defined(CAANOO)
			return DGStr("CAANOO");
		#elif defined(TARGET_OS_IPHONE) // SPACE FRACTAL
			switch(isIphone())
			{
			case 1:
				return DGStr("IPHONE");
			case 2:
				return DGStr("IPAD");
			case 3:
				return DGStr("OLDIOS");
			case 4:
				return DGStr("TVOS");
			}
			return DGStr("IPOD TOUCH");
		#elif defined(PALM_WEBOS)
			DGStr hw = UCASE_Str(DGStr(PDL_GetHardware()));
			if (hw == CGStr("UNKNOWN"))
			{
				// can be veer, pre2, pre3 or touchpad. PDL_GetHardwareID since SDK 2.0 - shit!
				DGNat sx, sy;
				GETSCREENSIZE(sx, sy);
				if(sx>700 && sy>700)
					return DGStr("TOUCHPAD"); // easy.
				if(sx>400 || sy>400) // not a pixi/veer
					return DGStr("PRE");
				return DGStr("PIXI");
			}
			return hw;
		#elif defined(ANDROID)
			if(android_keyboard() )
				return DGStr("KEYBOARD");
			return DGStr("TOUCHSCREEN");
		#else
			return DGStr("DESKTOP");
		#endif
	}
	else if( strncmp(WhatInfo.c_str(), "GLEXT:",6) == 0)
	{
		#ifndef GLB_CONSOLE
		void* pFoo = (void*)glGFGetProcAddressArb(origInfo.c_str() + 6);
		if(pFoo != NULL)
			return DGStr("1");
		#endif
		return DGStr("0");
	}
	else if(WhatInfo.compare("ID")==0)
	{
		#if defined(_WIN32_WCE)
			// #define CTL_CODE(DeviceType, Function, Method, Access) (((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method))
			// #define IOCTL_HAL_GET_DEVICEID CTL_CODE(FILE_DEVICE_HAL, 21, METHOD_BUFFERED, FILE_ANY_ACCESS)
			#define IOCTL_HAL_GET_DEVICEID 0x01010054

			unsigned char buffer[512];
			DWORD bytesReturned=0;
			DGStr retval;
			// extern BOOL __stdcall KernelIoControl(DWORD dwIoControlCode, LPVOID lpInBuf, DWORD nInBufSize, LPVOID lpOutBuf, DWORD nOutBufSize, LPDWORD lpBytesReturned);
			HINSTANCE hi = LoadLibraryW(TEXT("coredll.dll"));
			if(hi)
			{
				typedef BOOL (*KernelIoControlptr)(DWORD, LPVOID, DWORD, LPVOID, DWORD, LPDWORD);
				KernelIoControlptr KernelIoControl = (KernelIoControlptr) GetProcAddressW(hi, (LPSTR)TEXT("KernelIoControl"));
				if(KernelIoControl)
				{
					if(!KernelIoControl(IOCTL_HAL_GET_DEVICEID, NULL, 0, (void*)buffer, 512, &bytesReturned))
					{
						// Some older PPC devices only return the ID if the buffer == 16 bytes long
						if(!KernelIoControl(IOCTL_HAL_GET_DEVICEID, 0, 0, (void*)buffer, 16, &bytesReturned)) retval = DGStr("FA17ED");
						else retval = __Bytes2Hex(buffer, bytesReturned);
					}
					else retval = __Bytes2Hex(buffer, bytesReturned);
				}
				FreeLibrary(hi);
				return retval;
			}
		#elif defined(GP2X)
			char buffer[32]; memset(buffer, 0, 32);
			if(FetchGP2XSerial(buffer))
				return CGStr(buffer);
		#elif defined(GPHWIZ) || defined(CAANOO)
			{
				GFFileSystem fs;
				int rnd = (unsigned short)(GETTIMERALL() + 0x194ad2f3);
				if(fs.OpenIn("/etc/devid"))
				{
					fs.ReadLong(rnd);
					fs.Close();
				}
				else
				{
					if(fs.OpenOut("/etc/devid"))
					{
						fs.WriteLong(rnd);
					}
				}
				return DGStr(rnd);
			}
		#elif defined(PALM_WEBOS)
			DGStr buf;
			PDL_GetUniqueID(buf.getbuffer(255), 255);
			buf.releasebuffer();
			if(LEN(buf))
				return buf;
		#elif defined(LINUX)
			hd_driveid id;
			int fd = open("/dev/hda", O_RDONLY|O_NONBLOCK);
			if (fd >= 0 && !ioctl(fd, HDIO_GET_IDENTITY, &id))
			{
				return DGStr((const char*)id.serial_no);
			}
		#elif defined(WIN32) // win32
			DWORD dwSerial=0;
			::GetVolumeInformation(_T("C:"), NULL, 0, &dwSerial, NULL, NULL, NULL, 0);
			if(dwSerial)
			{
				dwSerial ^= 0x194ad2f3;
				return __Bytes2Hex((unsigned char*)&dwSerial, 4);
			}
//		#elif defined(TARGET_OS_IPHONE)
//			return DGStr( iPhoneDeviceID()); - no longer allowed by Apple
		#elif defined(ANDROID)
			return DGStr(android_device_id());
		#elif defined(EMSCRIPTEN)
			return DGStr("NO_DATA"); // GF_TODO
		#endif

		// x-platform fallback: use a random number stored in a file in APPDATA (per App on iOS)
		DGStr tmpfile = PLATFORMINFO_Str(CGStr("APPDATA")) + CGStr("/glbasic.deviceid");
		GFFileSystem fs;
		if(DOESFILEEXIST(tmpfile))
		{
			if(fs.OpenIn(tmpfile))
			{
				DGStr id;
				int sz = (int)fs.GetSize();
				fs.Read(id.getbuffer(sz+1), sz);
				id.getbuffer(sz+1)[sz]='\0';
				id.releasebuffer();
				id  =TRIM_Str(id);
				fs.Close();
				if(id.len()) return id;
			}
		}
		DGStr a,b,c,d;
		DGNat n = (rand()<<16) | RND(32767);
		a=__Bytes2Hex((unsigned char*)&n, 4);
		time_t long_time;
		time( &long_time );                /* Get time as long integer. */
		b=__Bytes2Hex((unsigned char*)&long_time, 4);
		n = (RND(32767)<<16) | rand();
		c=__Bytes2Hex((unsigned char*)&n, 4);
		n = (DGNat)GETTIMERALL() + (RND(32767)<<16);
		d=__Bytes2Hex((unsigned char*)&n, 4);
		a = a+b+c+d;

		if(fs.OpenOut(tmpfile))
		{
			if(fs.Write(a.c_str(), a.len()))
			{
				fs.Close();
				return a;
			}
		}

		// nothing worked :(
		return DGStr("FA17ED");
	}
	else if(WhatInfo.compare("TIME")==0)
	{
		struct tm *newtime;
		time_t long_time;
		time( &long_time );                /* Get time as long integer. */
		newtime = localtime( &long_time ); /* Convert to local time. */
		
		#define FORMAT2DIGITS(a) (INTEGER(a)>9 ? DGStr(INTEGER(a)) : (DGStr("0")+DGStr(INTEGER(a))))
		
		return DGStr(1900+newtime->tm_year)+CGStr("-")+FORMAT2DIGITS(newtime->tm_mon+1)+CGStr("-")+FORMAT2DIGITS(newtime->tm_mday)+
				CGStr(" ")+FORMAT2DIGITS(newtime->tm_hour)+DGStr(":")+FORMAT2DIGITS(newtime->tm_min)+CGStr(":")+FORMAT2DIGITS(newtime->tm_sec);
		#undef FORMAT2DIGITS
	}
	else if(WhatInfo.compare("UTCTIME")==0)
	{
		struct tm *newtime;
		time_t long_time;
		time( &long_time );                /* Get time as long integer. */
		newtime = gmtime( &long_time );    /* Convert to universal time. */
		
		#define FORMAT2DIGITS(a) (INTEGER(a)>9 ? DGStr(INTEGER(a)) : (DGStr("0")+DGStr(INTEGER(a))))
		
		return DGStr(1900+newtime->tm_year)+CGStr("-")+FORMAT2DIGITS(newtime->tm_mon+1)+CGStr("-")+FORMAT2DIGITS(newtime->tm_mday)+
				CGStr(" ")+FORMAT2DIGITS(newtime->tm_hour)+DGStr(":")+FORMAT2DIGITS(newtime->tm_min)+CGStr(":")+FORMAT2DIGITS(newtime->tm_sec);
		#undef FORMAT2DIGITS
	}
	else if (WhatInfo.compare("BATTERY")==0)
	{
		#if defined(_WIN32_WCE)
			SYSTEM_POWER_STATUS_EX2 pwr;
			GetSystemPowerStatusEx2(&pwr, sizeof(pwr), TRUE);
			return DGStr(MIN(100, DGNat(pwr.BatteryLifePercent)));
		#elif defined(WIN32)
			SYSTEM_POWER_STATUS pwr;
			GetSystemPowerStatus(&pwr);
			return DGStr(MIN(100, DGNat(pwr.BatteryLifePercent)));
		#elif defined(GP2X)
			return DGStr(DGNat(gp2x_wiz_battery()));
		#elif defined(TARGET_OS_IPHONE)
			return DGStr((DGNat)(iPhoneBattery()*100.0));
		#elif defined(ANDROID)
			return DGStr((DGNat)android_battery_level());
		#elif defined(EMSCRIPTEN)
			return DGStr("NO_DATA"); // GF_TODO
		#elif defined(LINUX) || defined(GPHWIZ) || defined(CAANOO)
			return DGStr(DGNat(linux_battery_level() ));
		#endif
		return DGStr(100);
	}
	else if (WhatInfo.compare("COMPILED")==0)
	{
		return __CompileTime();
	}
	else if (WhatInfo.compare("HOSTID")==0)
	{
		return __BuildSerial();
	}
	else if(WhatInfo.compare("VERSION")==0)
	{
		return __GetPrpjectVersionGLB();
	}
	else if(WhatInfo.compare("LOCALE")==0)
	{
		const char* pLang="us";
		#if defined(TARGET_OS_IPHONE)
			char cco[4];
			cco[0]='\0';
			iPhoneGetCountryInto(cco);
			pLang = cco;
		#elif defined(PALM_WEBOS)
			DGStr buf;
			int rv = PDL_GetLanguage(buf.getbuffer(64), 63);
			if(PDL_NOERROR != buf)
			{
				buf="us";
			}
			buf.releasebuffer();
			return buf;
		#elif defined(WIN32)
			#ifdef _WIN32_WCE
			unsigned int id = 0x0c09; // "en" #NEED - missing in header
			#else
			unsigned int id = GetSystemDefaultUILanguage();
			#endif

			switch(id)
			{
			case 0x0436: pLang = "AF"; break;
			case 0x041c: pLang = "SQ"; break;
			case 0x1401: pLang = "AR"; break;
			case 0x3c01: pLang = "AR"; break;
			case 0x0c01: pLang = "AR"; break;
			case 0x0801: pLang = "AR"; break;
			case 0x2c01: pLang = "AR"; break;
			case 0x3401: pLang = "AR"; break;
			case 0x3001: pLang = "AR"; break;
			case 0x1001: pLang = "AR"; break;
			case 0x1801: pLang = "AR"; break;
			case 0x2001: pLang = "AR"; break;
			case 0x4001: pLang = "AR"; break;
			case 0x0401: pLang = "AR"; break;
			case 0x2801: pLang = "AR"; break;
			case 0x1c01: pLang = "AR"; break;
			case 0x3801: pLang = "AR"; break;
			case 0x2401: pLang = "AR"; break;
			case 0x042b: pLang = "HY"; break;
			case 0x044d: pLang = "AS"; break;
			case 0x082c: pLang = "AZ"; break;
			case 0x042c: pLang = "AZ"; break;
			case 0x042d: pLang = "EU"; break;
			case 0x0423: pLang = "BE"; break;
			case 0x0445: pLang = "BE"; break;
			case 0x0402: pLang = "BG"; break;
			case 0x0403: pLang = "CA"; break;
			case 0x0c04: pLang = "ZH"; break;
			case 0x1404: pLang = "ZH"; break;
			case 0x0804: pLang = "CH"; break;
			case 0x1004: pLang = "ZH"; break;
			case 0x0404: pLang = "CH"; break;
			case 0x0827: pLang = "LT"; break;
			case 0x041a: pLang = "HR"; break;
			case 0x0405: pLang = "CS"; break;
			case 0x0406: pLang = "DA"; break;
			case 0x0465: pLang = "DI"; break;
			case 0x0813: pLang = "NL"; break;
			case 0x0413: pLang = "NL"; break;
			case 0x0c09: pLang = "EN"; break;
			case 0x2809: pLang = "EN"; break;
			case 0x1009: pLang = "EN"; break;
			case 0x2409: pLang = "EN"; break;
			case 0x1809: pLang = "EN"; break;
			case 0x2009: pLang = "EN"; break;
			case 0x1409: pLang = "EN"; break;
			case 0x3409: pLang = "EN"; break;
			case 0x1c09: pLang = "EN"; break;
			case 0x2c09: pLang = "EN"; break;
			case 0x0809: pLang = "EN"; break;
			case 0x0409: pLang = "UK"; break; // wibble on uk computer - should be us
			case 0x3009: pLang = "EN"; break;
			case 0x0425: pLang = "ET"; break;
			case 0x0438: pLang = "FO"; break;
			case 0x0429: pLang = "FA"; break;
			case 0x040b: pLang = "FI"; break;
			case 0x080c: pLang = "FR"; break;
			case 0x0c0c: pLang = "FR"; break;
			case 0x040c: pLang = "FR"; break;
			case 0x140c: pLang = "FR"; break;
			case 0x180c: pLang = "FR"; break;
			case 0x100c: pLang = "FR"; break;
			case 0x042f: pLang = "MK"; break;
			case 0x0456: pLang = "GL"; break;
			case 0x0437: pLang = "KA"; break;
			case 0x0c07: pLang = "DE"; break;
			case 0x0407: pLang = "DE"; break;
			case 0x1407: pLang = "DE"; break;
			case 0x1007: pLang = "DE"; break;
			case 0x0807: pLang = "DE"; break;
			case 0x0408: pLang = "EL"; break;
			case 0x0447: pLang = "GU"; break;
			case 0x040d: pLang = "HE"; break;
			case 0x0439: pLang = "HI"; break;
			case 0x040e: pLang = "HU"; break;
			case 0x040f: pLang = "IS"; break;
			case 0x0421: pLang = "IN"; break;
			case 0x0410: pLang = "IT"; break;
			case 0x0810: pLang = "IT"; break;
			case 0x0411: pLang = "JP"; break;
			case 0x044b: pLang = "KA"; break;
			case 0x043f: pLang = "KK"; break;
			case 0x0457: pLang = "KN"; break;
			case 0x0412: pLang = "KO"; break;
			case 0x0440: pLang = "KY"; break;
			case 0x0426: pLang = "LV"; break;
			case 0x0427: pLang = "LT"; break;
			case 0x083e: pLang = "MS"; break;
			case 0x043e: pLang = "MS"; break;
			case 0x044c: pLang = "MA"; break;
			case 0x044e: pLang = "MA"; break;
			case 0x0450: pLang = "MO"; break;
			case 0x0414: pLang = "NO"; break;
			case 0x0814: pLang = "NO"; break;
			case 0x0448: pLang = "OR"; break;
			case 0x0415: pLang = "PL"; break;
			case 0x0416: pLang = "PT"; break;
			case 0x0816: pLang = "PT"; break;
			case 0x0446: pLang = "PA"; break;
			case 0x0418: pLang = "RO"; break;
			case 0x0419: pLang = "RU"; break;
			case 0x044f: pLang = "SA"; break;
			case 0x0c1a: pLang = "SR"; break;
			case 0x081a: pLang = "SR"; break;
			case 0x041b: pLang = "SK"; break;
			case 0x0424: pLang = "SL"; break;
			case 0x2c0a: pLang = "ES"; break;
			case 0x400a: pLang = "ES"; break;
			case 0x340a: pLang = "ES"; break;
			case 0x240a: pLang = "ES"; break;
			case 0x140a: pLang = "ES"; break;
			case 0x1c0a: pLang = "ES"; break;
			case 0x300a: pLang = "ES"; break;
			case 0x440a: pLang = "ES"; break;
			case 0x100a: pLang = "ES"; break;
			case 0x480a: pLang = "ES"; break;
			case 0x080a: pLang = "ES"; break;
			case 0x4c0a: pLang = "ES"; break;
			case 0x180a: pLang = "ES"; break;
			case 0x3c0a: pLang = "ES"; break;
			case 0x280a: pLang = "ES"; break;
			case 0x500a: pLang = "ES"; break;
			case 0x040a: pLang = "ES"; break;
			case 0x0c0a: pLang = "ES"; break;
			case 0x380a: pLang = "ES"; break;
			case 0x200a: pLang = "ES"; break;
			case 0x0441: pLang = "SW"; break;
			case 0x081d: pLang = "SV"; break;
			case 0x041d: pLang = "SV"; break;
			case 0x045a: pLang = "SY"; break;
			case 0x0449: pLang = "TA"; break;
			case 0x0444: pLang = "TT"; break;
			case 0x044a: pLang = "TE"; break;
			case 0x041e: pLang = "TH"; break;
			case 0x041f: pLang = "TR"; break;
			case 0x0422: pLang = "UK"; break;
			case 0x0420: pLang = "UR"; break;
			case 0x0820: pLang = "UR"; break;
			case 0x0843: pLang = "UZ"; break;
			case 0x0443: pLang = "UZ"; break;
			case 0x042a: pLang = "VI"; break;
			}
		#elif defined(ANDROID)
			pLang = android_lang();
		#else
			pLang = getenv("LANG");
			if(!pLang) pLang="us";
		#endif
		DGStr c = pLang;
		c.lower();

		if(c=="eng" || c=="english-uk") c="uk";
		c.left(2);
		if(c == "c")  c="us";
		if(c == "el") c="gr";
		if(c == "du") c="nl";
		if(c == "am") c="us"; // america or armenia?
		if(c == "en") c="us"; // any english -> go usa
		if(c == "ja") c="jp";
		if(c == "po") c="pl";
		if(c.len()!=2) return "us";
		return c;
	}
	else if(WhatInfo.compare("DOCUMENTS")==0)
	{
		#if defined(_WIN32_WCE)
			return DGStr("/My Documents");
		#elif defined(WIN32)
			DGStr path;
			#ifdef UNICODE
				wchar_t wpath[1024];
				BOOL ok = SHGetSpecialFolderPath((HWND)GLBASIC_HWND(), wpath, CSIDL_PERSONAL, FALSE);
				path.assign((const wchar_t*)wpath);
			#else
				BOOL ok = SHGetSpecialFolderPath((HWND)GLBASIC_HWND(), path.getbuffer(1024), CSIDL_PERSONAL, FALSE);
				path.releasebuffer();
			#endif
			if(!ok) return DGStr("NO_DATA");
			path = REPLACE_Str(path, CGStr("\\"), CGStr("/"));
			return path;
		#elif defined(TARGET_OS_IPHONE)
			return DGStr(iPhoneGetDocumentsPath());
		#elif defined(MACOSX)
			return DGStr(MacGetDocumentsPath());
		#elif defined(ANDROID)
			return DGStr(android_ext_storage_pub_dir());
		#elif defined(EMSCRIPTEN)
			return DGStr("/tmp"); // GF_TODO
		#elif defined(POSIX)
			DGStr home = GETENV_Str("HOME");

			#if defined(PALM_WEBOS)
			if (home == CGStr("/home/root")) // started with root shell
			{
				GFFileSystem::GetExeDir(home.getbuffer(1024)); // Yikes!!
				home.releasebuffer();
				home = RTRIM_Str(home, "/");
			}
			#endif
			if(LEN(home)==0)
				return DGStr("~");
			return home;
		#endif
	}
	else if(WhatInfo.compare("APPDATA")==0)
	{
		DGStr path;
		#if defined(WIN32) && !defined(_WIN32_WCE)
			#ifdef UNICODE
				wchar_t wpath[1024];
				BOOL ok = SHGetSpecialFolderPath((HWND)GLBASIC_HWND(), wpath, CSIDL_LOCAL_APPDATA, FALSE);
				path.assign((const wchar_t*)wpath);
			#else
				BOOL ok = SHGetSpecialFolderPath((HWND)GLBASIC_HWND(), path.getbuffer(1024), CSIDL_LOCAL_APPDATA, FALSE);
				path.releasebuffer();
			#endif
			if(!ok) return DGStr("NO_DATA");
			path = REPLACE_Str(path, CGStr("\\"), CGStr("/"));
			path += CGStr("/GLBasic");
		#elif defined(MACOSX)
			path = DGStr(MacSettingsDir());
		#elif defined(ANDROID)
			path = DGStr(android_files_dir()) + CGStr("/.glbasic");
		#else
			// Default to documents directory
			path = PLATFORMINFO_Str(CGStr("DOCUMENTS")) + CGStr("/.glbasic");
		#endif

		if(!GFFileSystem::DoesDirExist(path.c_str()) )
			GFFileSystem::MkDir(path.c_str());
		return path;
	}
	else if(WhatInfo.compare("TEMP")==0)
	{
		DGStr path;
		#ifdef WIN32
			#ifdef UNICODE
				wchar_t wpath[1024] = L"";
				::GetTempPathW(1024, wpath);
				path.assign((const wchar_t*)wpath);
			#else
				::GetTempPathA(1024, path.getbuffer(1024));
				path.releasebuffer();
			#endif
			if(LEN(path)==0)
			{
				GFFileSystem::MkDir("C:\\Temp");
				return DGStr("C:/Temp");
			}
			path = REPLACE_Str(path, CGStr("\\"), CGStr("/"));
			return path;
		#elif defined(TARGET_OS_IPHONE)
			return DGStr(iPhoneGetTempPath() );
		#elif defined(ANDROID)
			return DGStr( android_cache_dir() );
		#elif defined(GPHWIZ) || defined(GP2X) || defined(CAANOO)
			return DGStr("/mnt/sd"); // don't write to nand
		#elif defined(PALM_WEBOS)
			DGStr tmp = PLATFORMINFO_Str(DGStr("DOCUMENTS")) + CGStr("/temp");
			GFFileSystem::MkDir(tmp.c_str());
			return tmp;
		#elif defined(EMSCRIPTEN)
			return DGStr("/tmp");
		#elif defined(MACOSX) || defined(POSIX) // would also work on iPhone, but better be safe
			DGStr tmp = GETENV_Str(CGStr("TMPDIR"));
			if(LEN(tmp)==0)
				return DGStr("/tmp");
			return tmp;
		#else
			#error no TEMP preprocessor define in platforminfo
		#endif
	}
	else if(WhatInfo.compare("DPI")==0)
	{
		DGNat dpi = 96;
		#if defined( WIN32) && !defined(GLB_CONSOLE)
			HDC screen = ::GetDC(0);
			dpi = GetDeviceCaps (screen, LOGPIXELSX);
			::ReleaseDC (0, screen);
		#elif defined(TARGET_OS_IPHONE)
			dpi = iphone_get_dpi();
		#elif defined(ANDROID)
			dpi = android_get_dpi();
		#elif defined(GPHWIZ) || defined(GP2X) || defined(CAANOO)
			dpi = 114;
		#elif defined(PALM_WEBOS)
			dpi = 186;
		#elif defined(MACOSX) || defined(POSIX) // would also work on iPhone, but better be safe
		#else // GF_TODO EMSCRIPTEN
		#endif

		if(dpi < 96) dpi = 96; // that's a bug obviously! MAC from 1987 had 72

		return DGStr(dpi);
	}
return DGStr("NO_DATA");
}



extern "C"
{
#if defined (POSIX)
  extern char **environ;
  #define envptr environ
#elif defined (_WIN32_WCE)
  extern char **environ;
  #define envptr environ
#else
  extern char **_environ;
  #define envptr _environ
#endif
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGStr   GETENV_Str(DGStr name)
{
	if(LEN(name)==0)
	{
		char **next = envptr;
		DGStr all;
		while (*next) 
		{
			all+=*next;
			all+=CGStr("\n");
			++next;
		}
		return all;
	}
	else
	{
		char* pEnv = getenv(name.c_str());
		if(!pEnv)
			return DGStr("");
		return DGStr((const char*)pEnv);
	}
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void	PUTENV(DGStr name, DGStr value)
{
	#ifdef _WIN32
	#define putenv _putenv
	#endif
	name += CGStr("=") + value;
	putenv(name);
}


////////////////////////
// DISC ACCESS FUNCTIONS
////////////////////////

// ------------------------------------------------------------- //
// Attaching a shoebox
// ------------------------------------------------------------- //
void SETSHOEBOX(const DGStr& data, const DGStr& media)
{ // Space Fractal, if the shoebox file is in the apk, make sure its got extracted from that.
#ifdef ANDROID
	DOESFILEEXIST(data);
	DOESFILEEXIST(media);
#endif

#ifdef WANT_SHOEBOX
	__g_ShoeBox_Data->SetBox(data.c_str());
	__g_ShoeBox_Media->SetBox(media.c_str());
#endif
}

// ------------------------------------------------------------- //
// Class for buffered file access
// ------------------------------------------------------------- //
class FAST_FILE
{
public:
	FAST_FILE()			{m_filename[0]='\0'; FillBuffer(NULL);}
	~FAST_FILE()		{FillBuffer(NULL);}
	// ------------------------------------------------------------- //
	// Read file to a buffer if buffer is different to file
	// ------------------------------------------------------------- //
	DGInt FillBuffer(const char* filename)
	{
		CLEARERROR();
		if (filename == NULL || strcmp(m_filename, filename)!=0)
		{
			GFFileSystem	file;
			// Write changed Data - not updated in shoebox, data get's void
			if (m_filename != NULL && strlen(m_filename) && m_changed)
			{
				if(file.OpenOut(m_filename))
				{
					for(long i=0; i<=m_maxline; i++)
						file.WriteLine(m_buffer[i]);
					file.Close();
				}
			}

			// Clear Data
			memset(m_buffer, 0, sizeof(m_buffer));
			m_changed=0;
			m_maxline=0;
			m_filename[0] = '\0';
			if(filename!=NULL)
				strcpy(m_filename, filename);
			else
				return CMP_OK; // Accessing the file that's already in buffer

			// Read Data - even from shoebox

			char TempFileName[1024];
			strcpy(TempFileName, m_filename);

			if (file.OpenIn(TempFileName))
			{
				long i=0;
				for(;;)
				{
					if (!file.ReadLine(m_buffer[i], 1024))
						break;
					m_maxline=i;
					i++;
					if (i>255) break;
				}
				file.Close();
			}
			else
			{
				return __Error(CMP_NO_FILE);
			}
		}
	return CMP_OK;
	}

	char m_filename[1024];
	char m_buffer[256][1024];
	long m_maxline;
	char m_changed;
};

inline FAST_FILE& __get_FastFile(){static FAST_FILE ff; return ff;}
#define __g_FastFile __get_FastFile()


// PUTFILE
// ------------------------------------------------------------- //
DGInt PutFile(const char* filename, int line, const DGStr& data)
{
	CLEARERROR();
	if (line > 255 || line < 0) return __Error(CMP_OUT_OF_MEMORY);

	__g_FastFile.FillBuffer(filename);

	strcpy(__g_FastFile.m_buffer[line], data.c_str());
	if (line>__g_FastFile.m_maxline)
		__g_FastFile.m_maxline = line;
	__g_FastFile.m_changed=1;
return CMP_OK;
}

// GETFILE
// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGInt GetFile(const char* filename, int line, DGStr& data)
{	
	CLEARERROR();
	if (line > 255 || line < 0)
	{
		data="NO_DATA";
		return __Error(CMP_OUT_OF_MEMORY);
	}

	

	DGInt nerr=__g_FastFile.FillBuffer(filename);
	if (nerr!=CMP_OK)
	{
		data="NO_FILE";
		return nerr;
	}

	data=(const char*)__g_FastFile.m_buffer[line];
	data.trimright();
	if(data.len()==0 || data[0]=='\0')
		data="NO_DATA";
return CMP_OK;
}

// ------------------------------------------------------------- //
// New File IO
// ------------------------------------------------------------- //
// number of file handles (OPENFILE)
#define MAX_IO_FILES	8
GFFileSystem __g_Files[MAX_IO_FILES];

DGNat	GENFILE()
{
	for(int channel=0; channel<MAX_IO_FILES; ++channel)
	{
		if(!__g_Files[channel].IsOpen())
			return channel;
	}
	return -1;
}


int64	GETFILESIZE(const DGStr& file)
{	
	CLEARERROR();
	GFFileSystem fs;
	if(fs.OpenIn(file.c_str()))
	{
		return (int64) fs.GetSize();
	}
	else
	{
		__Error(CMP_NO_FILE);
	}
return 0;
}

DGStr	GETFILETIME_Str(const DGStr& str)
{
	CLEARERROR();
	GFFileSystem fs;
	DGStr ss = fs.GetFileTime(str.c_str());
	if(ss.len() == 0)
	{
		__Error(CMP_NO_FILE);
	}
return ss;
}

DGNat	ENDOFFILE(DGNat channel)
{

	if(!__g_Files[(int)channel].IsOpen())
	{
		__Error(CMP_NO_FILE);
	}
	return (DGNat)__g_Files[(int)channel].Eof();
}

DGNat	OPENFILE(DGNat channel, const DGStr& file, DGNat mode)
{
	CLEARERROR();
	if(channel>=MAX_IO_FILES || channel<0)
	{
		__Error(CMP_OUT_OF_MEMORY);
		return 0;
	}
	__g_Files[(int)channel].Close();
	if(mode>0) // READ
	{
		DGStr s(file);
		if(!glb_prepare_reading_file(s))
		{
			__Error(CMP_NO_FILE);
			return FALSE;
		}
		bool rv = __g_Files[(int)channel].OpenIn(s.c_str());
		return rv ? TRUE : FALSE;	
	}
	else // WRITE/APPEND
	{
		if (mode==0) // append
		{
			int rv = __g_Files[(int)channel].OpenOut(file.c_str());
			if(!rv)
				__Error(CMP_NO_FILE);
			return rv;
		}
		else
		{
			int rv = __g_Files[(int)channel].OpenAppend(file.c_str());
			if(!rv)
				__Error(CMP_NO_FILE);
			return rv;
		}
	}
}



void	CLOSEFILE(DGNat channel)
{
	CLEARERROR();
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
		__g_Files[(int)channel].Close();
	#ifdef GP2X
	system("sync");
	#endif
}

void	READBYTE(DGNat channel, DGNat& v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	signed char b=0;
	if(!__g_Files[(int)channel].Read(&b, 1)) {__Error(CMP_NO_FILE); __EndProgram();}
	v=(DGNat) b;
}

void	READWORD(DGNat channel, DGNat& v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	signed short b=0;
	if(!__g_Files[(int)channel].ReadWord(b)) {__Error(CMP_NO_FILE); __EndProgram();}
	v=(DGNat) b;
}

void	READLONG(DGNat channel, DGNat& v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	signed long b=0;
	if(!__g_Files[(int)channel].ReadLong(b)) {__Error(CMP_NO_FILE); __EndProgram();}
	v=(DGNat) b;
}

void	READUBYTE(DGNat channel, DGNat& v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	unsigned char b=0;
	if(!__g_Files[(int)channel].Read(&b, 1)) {__Error(CMP_NO_FILE); __EndProgram();}
	v=(DGNat) b;
}

void	READUWORD(DGNat channel, DGNat& v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	unsigned short b=0;
	if(!__g_Files[(int)channel].ReadWord(b)) {__Error(CMP_NO_FILE); __EndProgram();}
	v=(DGNat) b;
}

void	READULONG(DGNat channel, DGNat& v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	unsigned long b=0;
	if(!__g_Files[(int)channel].ReadLong(b)) {__Error(CMP_NO_FILE); __EndProgram();}
	v=(DGNat) b;
}

void	READIEEE(DGNat channel, DGInt& v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
double b=0;
	if(!__g_Files[(int)channel].ReadDbl(b)) {__Error(CMP_NO_FILE); __EndProgram();}
	v=(DGInt) b;
}

void	READSHORTIEEE(DGNat channel, DGInt& v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
float b=0;
	if(!__g_Files[(int)channel].ReadFlt(b)) {__Error(CMP_NO_FILE); __EndProgram();}
	v=(DGInt) b;
}


void	READSTR(DGNat channel, DGStr& v, DGNat nchar)
{
	if(nchar<=0)
	{
		v=CGStr(""); return;
	}
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}

	memset(v.getbuffer(nchar+32), 0, nchar+32);
	if(!__g_Files[(int)channel].Read(v.getbuffer(nchar+32), nchar)) {__Error(CMP_NO_FILE); __EndProgram();}
	v.releasebuffer();
	v.setlength(nchar); // allow \0 characters
}


void READLINE(DGNat channel, DGStr& v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	memset(v.getbuffer(1024), 0, 1024);
	if(!__g_Files[(int)channel].ReadLine(v.getbuffer(1024), 1024)) {__Error(CMP_NO_FILE); __EndProgram();}
	v.releasebuffer();
}

void READLINE(DGNat channel, DGInt& v)
{
	DGStr str;
	READLINE(channel, str);
	v = (DGInt) str;
}
void READLINE(DGNat channel, DGNat& v)
{
	DGStr str;
	READLINE(channel, str);
	v = (DGNat) str;
}


void	WRITEBYTE(DGNat channel, DGNat v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	char b=(char) v;
	if(!__g_Files[(int)channel].Write(&b, 1)) {__Error(CMP_NO_FILE); __EndProgram();}
}

void	WRITEWORD(DGNat channel, DGNat v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	short b=(short) v;
	if(!__g_Files[(int)channel].WriteWord(b)) {__Error(CMP_NO_FILE); __EndProgram();}
}

void	WRITELONG(DGNat channel, DGNat v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	long b=(long) v;
	if(!__g_Files[(int)channel].WriteLong(b)) {__Error(CMP_NO_FILE); __EndProgram();}
}
void	WRITEUBYTE(DGNat channel, DGNat v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
unsigned char b=(unsigned char) v;
	if(!__g_Files[(int)channel].Write(&b, 1)) {__Error(CMP_NO_FILE); __EndProgram();}
}

void	WRITEUWORD(DGNat channel, DGNat v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
unsigned short b=(unsigned short) v;
	if(!__g_Files[(int)channel].WriteWord(b)) {__Error(CMP_NO_FILE); __EndProgram();}
}

void	WRITEULONG(DGNat channel, DGNat v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
unsigned long b=(unsigned long) v;
	if(!__g_Files[(int)channel].WriteLong(b)) {__Error(CMP_NO_FILE); __EndProgram();}
}

void	WRITEIEEE(DGNat channel, DGInt v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
double b=(double) v;
	if(!__g_Files[(int)channel].WriteDbl(b)) {__Error(CMP_NO_FILE); __EndProgram();}
}

void	WRITESHORTIEEE(DGNat channel, DGInt v)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
float b=(float) v;
	if(!__g_Files[(int)channel].WriteFlt(b)) {__Error(CMP_NO_FILE); __EndProgram();}
}

void	WRITESTR(DGNat channel, const DGStr& str)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	if(!__g_Files[(int)channel].Write((void*)str.c_str(), str.len())) {__Error(CMP_NO_FILE); __EndProgram();}
}

void	WRITELINE(DGNat channel, const DGStr& str)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	if(		!__g_Files[(int)channel].Write((void*)str.c_str(), str.len())
		||	!__g_Files[(int)channel].Write((void*)"\r\n", 2)) {__Error(CMP_NO_FILE); __EndProgram();}
}


int64	FILEPOSITION(DGNat channel)
{
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}
	return __g_Files[(int)channel].Tell();
}

void	FILESEEK(DGNat channel, int64 bytes, DGNat direction)
{
	// direction: -1=from end, 0=from start, 1=from current position
	if(channel>=MAX_IO_FILES || channel<0){__Error(CMP_OUT_OF_MEMORY);__EndProgram();}

	GFFileSystem::SEEK_FROM dir = GFFileSystem::seek_start;
	if(direction<0)
		dir = GFFileSystem::seek_end;
	else if(direction>0)
		dir=GFFileSystem::seek_current;
	__g_Files[(int)channel].SeekTo(bytes, dir);
}



// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void ReverseSlashes(DGStr& str)
{
	for (int i=0; i<(int)strlen(str.c_str()); i++)
	{
		if (str[i]=='\\')
			str[i] = '/';
	}
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void	KILLFILE(const DGStr& file)
{
	GFFileSystem::Delete(file.c_str());
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGNat CREATEDIR(const DGStr& dir)
{
	return GFFileSystem::MkDir(dir.c_str()) ? TRUE:FALSE;
}

//---------------------------------------------------------------//
//---------------------------------------------------------------//
DGStr REPLACE_Str(DGStr str, const DGStr& sOld, const DGStr& sNew)
{
	return str.replace(sOld, sNew);
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void	COPYFILE(const DGStr& src, const DGStr& dst)
{
	CLEARERROR();
	GFFileSystem in, out;
	
	DGStr thefile(src);
	if(!glb_prepare_reading_file(thefile))
	return;
	
	if (in.OpenIn(thefile.c_str()) && out.OpenOut(dst.c_str()))
	{
		unsigned long sz = (unsigned long)in.GetSize();
		const int bufsz = 4096;
		char buffer[bufsz];
		for(unsigned long i=0; i<sz; i+=bufsz)
		{
			unsigned long nread = bufsz;
			if(i+nread>=sz) nread=sz-i;
			if(!in.Read(buffer, (int)nread) || !out.Write(buffer, (int)nread))
			{
				in.Close();
				out.Close();
				GFFileSystem::Delete(dst.c_str());
			}
		}
	}
	else
	{
		__Error(CMP_NO_FILE);
	}
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void MOVEFILE(const DGStr & src, const DGStr & dst)
{
	CLEARERROR();
	GFFileSystem fs;
	if(!fs.MoveFile(src.c_str(), dst.c_str()))
	{
		__Error(CMP_NO_FILE);
	}
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGNat	DOESFILEEXIST(const DGStr& file)
{
	#if defined(ANDROID) // Copy AssestFile from APK if required.
		int filesize=CheckAssetFile(file, 0); if (filesize>0) return TRUE;
			filesize=CheckAssetFile(file, 0); if (filesize>0) return TRUE;
		return GFFileSystem::DoesFileExist(file.c_str()) ? TRUE:FALSE;
	#else
		return GFFileSystem::DoesFileExist(file.c_str()) ? TRUE:FALSE;
	#endif
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGNat	DOESDIREXIST(const DGStr& dir)
{	
	#if defined(ANDROID)
		DGStr thedir = RTRIM_Str(dir, "/"); // Trailing slash is bad for Android
		int	ok=CGStr(JavaCall_Str(CGStr("dircopy:")+thedir));
	#endif

	return GFFileSystem::DoesDirExist(dir.c_str()) ? TRUE:FALSE;
}



// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGStr	GETCURRENTDIR_Str()
{
DGStr			str;
char			dir[MAX_PATH+1];
	GFFileSystem::GetCurrentDir(dir);
	str=(const char*)dir;
	for (int i=0; i<(int)strlen(str.c_str()); i++)
	{
		if (str[i]=='\\')
			str[i] = '/';
	}
	ReverseSlashes(str);
	if (str[(int)strlen(dir)-1]!='/') str+="/";
//	STDOUT(DGStr("GETCURRENTDIR:")+str+DGStr("\n"));
return str;
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGInt	SETCURRENTDIR(DGStr dir)
{
	CLEARERROR();
	#if defined(ANDROID) // Space Fractal - required to prevent crash in DOESFILEEXIST
		if (dir==CGStr("Media") OR dir==CGStr("Media/"))
		{	int ok=CREATEDIR(CGStr("Media"));
		}
		CREATEDIR(CGStr(dir)); // Space Fractal - Needs create dir, since its can been empty, if not extracted from APK.
	#endif

	#if defined(TARGET_OS_IPHONE)
		CREATEDIR(CGStr(dir)); // Space Fractal - Odd, but seen this was required, because im did have some strange GETCURRENTDIR() bug.
	#endif
	
	#if defined(WIN32)
		dir = dir.replace(DGStr("/"), DGStr("\\"));
	#endif

	int rv = GFFileSystem::SetCurrentDir(dir.c_str());
	if(rv)
	{
		return TRUE;
	}
	else
	{
		__Error(CMP_NO_FILE);
		return FALSE;
	}
}

#if 0
// ------------------------------------------------------------- //
// check if a wildcard match is valid
// for *.* return all files, even w/o '.'
// ------------------------------------------------------------- //
bool szWildMatch(const char* pat, const char* str)
{
#define GF_lower(a) ((a>='A' && a<='Z') ? ((a)+'a'-'A') : (a))
   const char* s, *p;
   bool star = false;
   
   // for *.*, return everything
   if(pat[0]=='*' && pat[1]=='.' && pat[2]=='*' && pat[3]=='\0')
	return true;

loopStart:
   for (s=str,p=pat; *s; ++s,++p)
   {
      switch (*p)
      {
         case '?':
            if (*s == '.') goto starCheck;
            break;
         case '*':
            star = true;
            str = s, pat = p;
            do { ++pat; } while (*pat == '*');
            if (!*pat) return true;
            goto loopStart;
         default:
            if (GF_lower(*s) != GF_lower(*p))
               goto starCheck;
            break;
      } /* endswitch */
   } /* endfor */
   while (*p == '*') ++p;
   return (!*p);
starCheck:
   if (!star) return false;
   str++;
   goto loopStart;
}
#else // much better version

//---------------------------------------------------------------//
// This function compares text strings, one of which can have wildcards ('*').
// http://www.drdobbs.com/architecture-and-design/matching-wildcards-an-algorithm/210200888
//---------------------------------------------------------------//
bool szWildMatch(
	const char* pWildText,            // A (potentially) corresponding string with wildcards
	const char* pTameText             // A string without wildcards
	)
{
#define GF_lower(a) ((a>='A' && a<='Z') ? ((a)+'a'-'A') : (a))

	// for * and *.*, return everything
	if(pWildText[0]=='*' && pWildText[1]=='\0')
		return true;
	if(pWildText[0]=='*' && pWildText[1]=='.' && pWildText[2]=='*' && pWildText[3]=='\0')
		return true;


	bool bMatch = TRUE;
	const char* pAfterLastWild = NULL; // The location after the last '*', if weve encountered one
	const char* pAfterLastTame = NULL; // The location in the tame string, from which we started after last wildcard
	char t, w;

	// Walk the text strings one character at a time.
	for(;;)
	{
		t = *pTameText;
		w = *pWildText;

		// How do you match a unique text string?
		if (!t)
		{
			// Easy: unique up on it!
			if (!w )
			{
				break;                                   // "x" matches "x"
			}
			else if (w == '*')
			{
				pWildText++;
				continue;                           // "x*" matches "x" or "xy"
			}
			else if (pAfterLastTame)
			{
				if (!(*pAfterLastTame))
				{
					bMatch = false;
					break;
				}
				pTameText = pAfterLastTame++;
				pWildText = pAfterLastWild;
				continue;
			}

			bMatch = false;
			break;                                           // "x" doesn't match "xy"
		}
		else
		{
			// Lowercase the characters to be compared.
			t = GF_lower(t);
			w = GF_lower(w);

			// How do you match a tame text string?
			if (t != w)
			{
				// The tame way: unique up on it!
				if (w == '*')
				{
					pAfterLastWild = ++pWildText;
					pAfterLastTame = pTameText;
					w = *pWildText;

					if (!w )
					{
						break;                           // "*" matches "x"
					}
					continue;                           // "*y" matches "xy"
				}
				else if (pAfterLastWild)
				{
					if (pAfterLastWild != pWildText)
					{
						pWildText = pAfterLastWild;
						w = *pWildText;

						w = GF_lower(w);

						if (t == w)
						{
							pWildText++;
						}
					}
					pTameText++;
					continue;                           // "*sip*" matches "mississippi"
				}
				else
				{
					bMatch = false;
					break;                                   // "x" doesn't match "y"
				}
			}
		}

		pTameText++;
		pWildText++;
	}

	return bMatch;
}

#endif



#ifndef WIN32
long IsDirectory(const char* path)
{
struct stat stbuf;
	if( stat(path, &stbuf) )
		return 0;
	return ((stbuf.st_mode & S_IFMT)==S_IFDIR) ? 1:0;
}
#endif


DGInt __CompareNoCase(DGStr& a, DGStr& b)
{
	return (DGNat)stricmp(a.c_str(), b.c_str());
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGInt	GETFILELIST(const DGStr& wildcard, DGStrArray& files)
{
Queue<DGStr>	qfiles, qdirs;
DGStr			str;
	str.alloc(1024,false);
	


#ifdef WIN32
	HANDLE			hfnd;
		WIN32_FIND_DATA	fnd;
	
	#if defined(_WIN32_WCE) || defined(POCKETPC)// no relative paths
		DGStr csdir;
		GFFileSystem::GetCurrentDir(csdir.getbuffer(1024));
		csdir.releasebuffer();
		csdir = csdir.replace(DGStr("/"), DGStr("\\"));
		csdir += DGStr("\\*.*");
		#ifdef UNICODE
			wchar_t wdir[512]=_T("*.*");
			mbstowcs(wdir, csdir.c_str(), 512);
		#else
			const char* wdir = csdir.c_str();
		#endif
		hfnd = FindFirstFile(wdir, &fnd);
	#else
		DGStr cur_dir = GETCURRENTDIR_Str(); // has appended "/"
		cur_dir.replace(DGStr("/"), DGStr("\\"));
		hfnd = FindFirstFile((cur_dir + CGStr("*.*")).c_str(), &fnd);

		// this might be a reparse point that denies access to directory listing
		// to prevent legacy apps to double backup that dir.
		if(hfnd == INVALID_HANDLE_VALUE)
		{
			// no UNICODE here - we're on Win32
			DWORD att = ::GetFileAttributes(cur_dir.c_str());

			// It's a reparse point - see if we can list the target
			if(att!=INVALID_FILE_ATTRIBUTES && (att & FILE_ATTRIBUTE_REPARSE_POINT) )
			{
				HANDLE hDir = ::CreateFile(cur_dir.c_str(), FILE_READ_ATTRIBUTES, FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE, NULL, OPEN_EXISTING, FILE_FLAG_OPEN_REPARSE_POINT | FILE_FLAG_BACKUP_SEMANTICS, NULL);
				BYTE* buf = new BYTE[16*1024]; // We need a large buffer
				DWORD dwRet=0;

				// #define FSCTL_GET_REPARSE_POINT 0x000900A8
				if (::DeviceIoControl(hDir, 0x000900A8, NULL, 0, buf, 16*1024, &dwRet, NULL))
				{
					LPCWSTR pPath = (LPCWSTR)(buf+16); // ReparseBuffer.ReparseTarget;
					if (wcsncmp(pPath, L"\\??\\", 4) == 0) pPath += 4;  // Skip 'non-parsed' prefix
					::WideCharToMultiByte(CP_ACP, 0, pPath, -1, cur_dir.getbuffer(1024), 1024, NULL, NULL);
					cur_dir.releasebuffer();
					cur_dir += CGStr("\\");
				}
				delete[] buf; buf=NULL;
				::CloseHandle(hDir);
			}
			FindClose(hfnd);
			hfnd = FindFirstFile((cur_dir + CGStr("*.*")).c_str(), &fnd);
		}
	#endif
	
	#ifdef _DEBUG
	if(hfnd == INVALID_HANDLE_VALUE)
		GFFileSystem::StrLastError();
	#endif
	
	while(hfnd!=INVALID_HANDLE_VALUE)
	{
		#ifdef UNICODE
			int nlen=(int)wcslen(fnd.cFileName)+1;
			_wcstombsz(str.getbuffer(nlen), fnd.cFileName, nlen);
			str.releasebuffer();
		#else
			str = (const char*)fnd.cFileName;
		#endif
		
		if (fnd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			//if((fnd.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) == 0)
			qdirs.Push(str);
		}
		else if (szWildMatch(wildcard.c_str(), str.c_str()))
		{
			qfiles.Push(str);
		}
		if (FindNextFile(hfnd, &fnd) == 0)
		{
			FindClose(hfnd);
			break;
		}
	}
#else // Linux

#if defined(ANDROID)
	DGStr cdir = RTRIM_Str(GETCURRENTDIR_Str(), DGStr("/") );
	STDOUT(DGStr("GETFILELIST(\"") + cdir + DGStr("\")\n") );
	DGNat bDoesDirExist = DOESDIREXIST( cdir ); // prepare unpacking stuff from APK (in Media folder...)
	STDOUT(DGStr("DOESDIREXIST() = ") + DGStr( bDoesDirExist ) + DGStr("\n") );
#endif

#if (!defined TARGET_OS_IPHONE)
    struct stat stbuf;
#endif
    
DIR*	dirp;
	dirp = opendir(".");
	if (dirp)
    {
		struct dirent *dp;
		while ((dp = readdir (dirp)) != NULL)
		{
			str = (const char*)dp->d_name;
			if (IsDirectory(str.c_str()))
			{
				qdirs.Push(str);
			}
			else if (szWildMatch(wildcard.c_str(), str.c_str() ))
			{
				qfiles.Push(str);
			}
		}
		closedir (dirp);
    }
#endif

	if(qdirs.Size())
		sortarray(&qdirs[0], &qdirs[qdirs.Size()-1], ADDRESSOF(__CompareNoCase));
	if(qfiles.Size())
		sortarray(&qfiles[0], &qfiles[qfiles.Size()-1], ADDRESSOF(__CompareNoCase));

	DIM(files, qfiles.Size()+qdirs.Size());
	long i;
	for (i=0; i<(long)qdirs.Size(); i++)
	{
		files((int)i) = *qdirs.GetAt((int)i);
	}
	for (i=0; i<(long)qfiles.Size(); i++)
	{
		files((int)i+qdirs.Size()) = *qfiles.GetAt((int)i);
	}
return (65536*qdirs.Size() + qfiles.Size());
}


// INI File Handling
static GFini&	__g_Ini()
{
	static GFini ini; return ini;
}


void	INIOPEN(const DGStr& filename)
{
	// Write last file's changes before opening a new file
	if(__g_Ini().m_changed)
	{
		__g_Ini().Write();
		#ifdef GP2X
		system("sync");
		#endif
	}

	if( LEN(filename ) )
	{
		__g_Ini().Read(filename);
	}
	else
	{
		__g_Ini().Clear();
	}
}

DGStr	INIGET_Str(const DGStr& section, const DGStr& key, DGStr nodata)
{
	return __g_Ini().GetVal(section, key, nodata);
}


void	INIPUT(const DGStr& section, const DGStr& key, const DGStr& value)
{
	if(LEN(key)==0)
	{
		__g_Ini().RemoveSection(section);
	}
	else
	{
		__g_Ini().SetVal(section, key, value);
	}
}

// SIN / COS / KEY / RND
DGInt	SIN(DGInt angle)
{
	#ifdef _DEBUG
		DGInt cc= (DGInt)(sin((double)angle*BOG_INV));
		return cc;
	#else
		return (DGInt)(sin((double)angle*BOG_INV));
	#endif
}

DGInt	COS(DGInt angle)
{
	#ifdef _DEBUG
		DGInt cc= (DGInt)(cos((double)angle*BOG_INV));
		return cc;
	#else
		return (DGInt)(cos((double)angle*BOG_INV));
	#endif
}

DGInt	TAN(DGInt angle)
{
	return (DGInt)(tan((double)angle*BOG_INV));
}


void SEEDRND(DGNat seed)
{
 	GFsrand((long)seed);
}


DGNat	RND(DGNat max)
{
	if(max<=0) return 0;
	return (DGNat)(GFrand()%((unsigned long)max+1));
}

// strlen
DGNat	LEN(DGInt s)
{
	return (DGNat)((DGStr)s).len();
}
// strlen
DGNat	LEN(DGNat s)
{
	return (DGNat)((DGStr)s).len();
}
// MATH-Stuff
DGInt	POW(DGInt n, DGInt e)
{
	return (DGInt)pow((double)n, (double)e);
}
DGNat	MOD(DGNat n, DGNat m)
{
	if (m==0) return 0;
	return n % m;
}
DGInt	FMOD(DGInt n, DGInt m)
{
	if (m==0) return 0;
	return (DGInt) fmod(n, m);
}

DGInt	FLOOR(DGInt n)  {return (DGInt)floor(n);}
DGInt	CEIL(DGInt n)	{return (DGInt)ceil(n);}


DGInt	ATAN(DGInt dx, DGInt dy)
{
	return (DGInt) (BOG*atan2((double)dx, (double)dy));
}
DGInt	ASIN(DGInt n)
{
	// Disallow overflow problems
	if (n>=1.0) return 90.0;
	if (n<=-1.0) return -90.0;
	return (DGInt) (BOG*asin((double)n));
}
DGInt	ACOS(DGInt n)
{
	// Disallow overflow problems
	if (n>=1.0) return 0.0;
	if (n<=-1.0) return 180.0;
	return (DGInt) (BOG*acos((double)n));
}
DGInt	SQR(DGInt n)
{
	return (DGInt) sqrt((double)n);
}
DGInt	LOGN(DGInt n)
{
	return (DGInt)log( n );
}





// ------------------------------------------------------------- //
// Copy array
// ------------------------------------------------------------- //
DGIntArray& DGIntArray::operator=(const DGIntArray& a)
{
	DIM(*this, a.max1, a.max2, a.max3, a.max4);
	int	siz=(int)(max1*MAX(1,max2)*MAX(1,max3)*MAX(1,max4));
	this->count = siz;
	const DGInt* pR=a.pData;
	DGInt* pW=this->pData;
	while(siz)
	{
		*pW++ = *pR++; --siz;
	}
	// memcpy(pData, a.pData, sizeof(DGInt)*siz);
return *this;
}
DGNatArray& DGNatArray::operator=(const DGNatArray& a)
{
	DIM(*this, a.max1, a.max2, a.max3, a.max4);
	int	siz=(int)(max1*MAX(1,max2)*MAX(1,max3)*MAX(1,max4));
	this->count = siz;
	const DGNat* pR=a.pData;
	DGNat* pW=this->pData;
	while(siz)
	{
		*pW++ = *pR++; --siz;
	}
	// memcpy(pData, a.pData, sizeof(DGNat)*siz);
return *this;
}


// ------------------------------------------------------------- //
// Determine DIM bounds
// ------------------------------------------------------------- //
DGNat	BOUNDS(const DGIntArray& arr, DGNat nDim)
{
	switch((int)nDim)
	{
		case 0: return arr.max1; break;
		case 1: return arr.max2; break;
		case 2: return arr.max3; break;
		case 3: return arr.max4; break;
	}
return 0;
}
DGNat	BOUNDS(const DGNatArray& arr, DGNat nDim)
{
	switch(nDim)
	{
		case 0: return arr.max1; break;
		case 1: return arr.max2; break;
		case 2: return arr.max3; break;
		case 3: return arr.max4; break;
	}
return 0;
}



void	FIXDIM(DGIntArray& arr, DGNat xt1, DGNat xt2, DGNat xt3, DGNat xt4)
{
int clr=0;
	if (xt1<1)
	{
		xt1=0; // xt2=0; xt3=0; xt4=0;
		clr=1;
	}
	int	siz=(int)(xt1*MAX(1,xt2)*MAX(1,xt3)*MAX(1,xt4));
	if (siz>arr.mem || clr)
	{
		if (arr.pData!=NULL
		#if DGARRAY_STATIC_SIZE>0
		 && arr.pData!=&arr.pDataStatic[0]
		#endif
		 ) delete [] arr.pData;

		#if DGARRAY_STATIC_SIZE>0
		if (siz<=DGARRAY_STATIC_SIZE)
		{
			arr.pData = &arr.pDataStatic[0];
			arr.mem=DGARRAY_STATIC_SIZE;
		}
		else
		#endif
		{
			arr.pData=new DGInt[(int)siz];
			if (arr.pData==NULL) {__Error(CMP_OUT_OF_MEMORY); __EndProgram();}
			arr.mem=siz;
		}
	}
	DGInt* pW=arr.pData; while(siz) {*pW++=DGInt(0); --siz;}
	// memset(arr.pData, 0, sizeof(DGInt)*(int)arr.mem);
	if (xt4) arr.dim=4; else if (xt3) arr.dim=3; else if (xt2) arr.dim=2; else if(xt1) arr.dim=1; else arr.dim=0;
	arr.fac2=(int)xt1;
	arr.fac3=arr.fac2*(int)xt2;
	arr.fac4=arr.fac3*(int)xt3;
	arr.max1=(int)xt1; arr.max2=(int)xt2; arr.max3=(int)xt3; arr.max4=(int)xt4;
	arr.count = arr.max1*MAX(1,arr.max2)*MAX(1,arr.max3)*MAX(1,arr.max4);
}

void	FIXDIM(DGNatArray& arr, DGNat xt1, DGNat xt2, DGNat xt3, DGNat xt4)
{
int clr=0;
	if (xt1<1)
	{
		xt1=0; // xt2=0; xt3=0; xt4=0;
		clr=1;
	}
	int	siz=(int)(xt1*MAX(1,xt2)*MAX(1,xt3)*MAX(1,xt4));
	if (siz>arr.mem || clr)
	{
		if (arr.pData!=NULL
		#if DGARRAY_STATIC_SIZE>0
		 && arr.pData!=&arr.pDataStatic[0]
		#endif
		 ) delete [] arr.pData;

		#if DGARRAY_STATIC_SIZE>0
		if (siz<=DGARRAY_STATIC_SIZE)
		{
			arr.pData = &arr.pDataStatic[0];
			arr.mem=DGARRAY_STATIC_SIZE;
		}
		else
		#endif
		{
			arr.pData=new DGNat[siz];
			if (arr.pData==NULL) {__Error(CMP_OUT_OF_MEMORY); __EndProgram();}
			arr.mem=siz;
		}
	}
	DGNat* pW=arr.pData; while(siz) {*pW++=(0); --siz;}
	// memset(arr.pData, 0, sizeof(DGNat)*(int)arr.mem);
	if (xt4) arr.dim=4; else if (xt3) arr.dim=3; else if (xt2) arr.dim=2; else if(xt1) arr.dim=1; else arr.dim=0;
	arr.fac2=(int)xt1;
	arr.fac3=arr.fac2*(int)xt2;
	arr.fac4=arr.fac3*(int)xt3;
	arr.max1=(int)xt1; arr.max2=(int)xt2; arr.max3=(int)xt3; arr.max4=(int)xt4;
	arr.count = arr.max1*MAX(1,arr.max2)*MAX(1,arr.max3)*MAX(1,arr.max4);
}

void	DIM(DGIntArray& arr, DGNat xt1, DGNat xt2, DGNat xt3, DGNat xt4)
{
int clr=0;
	if (xt1<1)
	{
		xt1=0; // xt2=0; xt3=0; xt4=0;
		clr=1;
	}
	int	siz=(int)(xt1*MAX(1,xt2)*MAX(1,xt3)*MAX(1,xt4));
	if (siz>arr.mem || clr)
	{
		if (arr.pData!=NULL
		#if DGARRAY_STATIC_SIZE>0
		 && arr.pData!=&arr.pDataStatic[0]
		#endif
		 ) delete [] arr.pData;

		#if DGARRAY_STATIC_SIZE>0
		if (siz<=DGARRAY_STATIC_SIZE)
		{
			arr.pData = &arr.pDataStatic[0];
			arr.mem=DGARRAY_STATIC_SIZE;
		}
		else
		#endif
		{
			siz=((int)(siz/8)+1 )*8;
			arr.pData=new DGInt[siz];
			if (arr.pData==NULL) {__Error(CMP_OUT_OF_MEMORY); __EndProgram();}
			arr.mem=siz;
		}
	}
	DGInt* pW=arr.pData; while(siz) {*pW++=DGInt(0); --siz;}
	// memset(arr.pData, 0, sizeof(DGInt)*(int)arr.mem);
	if (xt4) arr.dim=4; else if (xt3) arr.dim=3; else if (xt2) arr.dim=2; else if(xt1) arr.dim=1; else arr.dim=0;
	arr.fac2=(int)xt1;
	arr.fac3=arr.fac2*(int)xt2;
	arr.fac4=arr.fac3*(int)xt3;
	arr.max1=(int)xt1; arr.max2=(int)xt2; arr.max3=(int)xt3; arr.max4=(int)xt4;
	arr.count = arr.max1*MAX(1,arr.max2)*MAX(1,arr.max3)*MAX(1,arr.max4);
}

void	DIM(DGNatArray& arr, DGNat xt1, DGNat xt2, DGNat xt3, DGNat xt4)
{
int clr=0;
	if (xt1<1)
	{
		xt1=0; // xt2=0; xt3=0; xt4=0;
		clr=1;
	}
	int	siz=(int)(xt1*MAX(1,xt2)*MAX(1,xt3)*MAX(1,xt4));
	if (siz>arr.mem || clr)
	{
		if (arr.pData!=NULL
		#if DGARRAY_STATIC_SIZE>0
		 && arr.pData!=&arr.pDataStatic[0]
		#endif
		 ) delete [] arr.pData;
		#if DGARRAY_STATIC_SIZE>0
		if (siz<=DGARRAY_STATIC_SIZE)
		{
			arr.pData = &arr.pDataStatic[0];
			arr.mem=DGARRAY_STATIC_SIZE;
		}
		else
		#endif
		{
			siz=((int)(siz/8)+1 )*8;
			arr.pData=new DGNat[siz];
			if (arr.pData==NULL) {__Error(CMP_OUT_OF_MEMORY); __EndProgram();}
			arr.mem=siz;
		}
	}
	DGNat* pW=arr.pData; while(siz) {*pW++=0; --siz;}
	// memset(arr.pData, 0, sizeof(DGNat)*(int)arr.mem);
	if (xt4) arr.dim=4; else if (xt3) arr.dim=3; else if (xt2) arr.dim=2; else if(xt1) arr.dim=1; else arr.dim=0;
	arr.fac2=(int)xt1;
	arr.fac3=arr.fac2*(int)xt2;
	arr.fac4=arr.fac3*(int)xt3;
	arr.max1=(int)xt1; arr.max2=(int)xt2; arr.max3=(int)xt3; arr.max4=(int)xt4;
	arr.count = arr.max1*MAX(1,arr.max2)*MAX(1,arr.max3)*MAX(1,arr.max4);
}

void	REDIM(DGIntArray& arr, DGNat xt1, DGNat xt2, DGNat xt3, DGNat xt4)
{
	// just reset dimensions
	if(xt1<=0)
	{
		arr.max1=0; arr.max2=0; arr.max3=0; arr.max4=0; arr.fac2=0; arr.fac3=0; arr.fac4=0; arr.dim=0; arr.count=0;
		return;
	}
DGIntArray tmp;
DGInt*     ptr=NULL;
	DIM(tmp, xt1, xt2, xt3, xt4);
	
int ma, mb, mc, md;
	ma = (int)MIN(arr.max1, xt1); mb = (int)MIN(arr.max2, xt2); mc = (int)MIN(arr.max3, xt3); md = (int)MIN(arr.max4, xt4);
	
	for(int a=0; a<ma; ++a)
	{
		if(mb)
			for(int b=0; b<mb; ++b)
			{
				if(mc)
					for(int c=0; c<mc; ++c)
					{
						if(md)
							for(int d=0; d<md; ++d)
							{
								tmp(a,b,c,d) = arr(a,b,c,d);
							}
						else
							tmp(a,b,c) = arr(a,b,c);
					}
				else
					tmp(a,b) = arr(a,b);
			}
		else
			tmp(a) = arr(a);
	}
	
	#if DGARRAY_STATIC_SIZE>0
	if(arr.pData==&arr.pDataStatic[0] || tmp.pData==&tmp.pDataStatic[0])
	{
		// need a full copy, one of the pointers is static
		arr = tmp;
	}
	else
	#endif
	{
		// they are dynamical - copy the variables, only
		arr.dim  = tmp.dim;
		arr.mem  = tmp.mem;
		arr.fac2 = tmp.fac2; arr.fac3 = tmp.fac3; arr.fac4 = tmp.fac4;
		arr.max1 = tmp.max1; arr.max2 = tmp.max2; arr.max3 = tmp.max3; arr.max4 = tmp.max4;
		arr.count= tmp.count;
		// can't do that on static memory...
		ptr = arr.pData; arr.pData=tmp.pData; tmp.pData=ptr;
	}
}

void	REDIM(DGNatArray& arr, DGNat xt1, DGNat xt2, DGNat xt3, DGNat xt4)
{
	// just reset dimensions
	if(xt1<=0)
	{
		arr.max1=0; arr.max2=0; arr.max3=0; arr.max4=0; arr.fac2=0; arr.fac3=0; arr.fac4=0; arr.dim=0; arr.count=0;
		return;
	}
DGNatArray tmp;
DGNat*     ptr=NULL;
	DIM(tmp, xt1, xt2, xt3, xt4);
	
int ma, mb, mc, md;
	ma = (int)MIN(arr.max1, xt1); mb = (int)MIN(arr.max2, xt2); mc = (int)MIN(arr.max3, xt3); md = (int)MIN(arr.max4, xt4);
	
	for(int a=0; a<ma; ++a)
	{
		if(mb)
			for(int b=0; b<mb; ++b)
			{
				if(mc)
					for(int c=0; c<mc; ++c)
					{
						if(md)
							for(int d=0; d<md; ++d)
							{
								tmp(a,b,c,d) = arr(a,b,c,d);
							}
						else
							tmp(a,b,c) = arr(a,b,c);
					}
				else
					tmp(a,b) = arr(a,b);
			}
		else
			tmp(a) = arr(a);
	}
	
	#if DGARRAY_STATIC_SIZE>0
	if(arr.pData==&arr.pDataStatic[0] || tmp.pData==&tmp.pDataStatic[0])
	{
		// need a full copy, one of the pointers is static
		arr = tmp;
	}
	else
	#endif
	{
		// they are dynamical - copy the variables, only
		arr.dim  = tmp.dim;
		arr.mem  = tmp.mem;
		arr.fac2 = tmp.fac2; arr.fac3 = tmp.fac3; arr.fac4 = tmp.fac4;
		arr.max1 = tmp.max1; arr.max2 = tmp.max2; arr.max3 = tmp.max3; arr.max4 = tmp.max4;
		arr.count= tmp.count;
		// can't do that on static memory...
		ptr = arr.pData; arr.pData=tmp.pData; tmp.pData=ptr;
	}
}

// *** DATA ***
void DIMDATA(DGIntArray& arr, int count, DGInt* pData)
{
	DIM(arr, count);
	for(int i=0; i<count; ++i)
		arr(i) = pData[i];
}
void DIMDATA(DGNatArray& arr, int count, DGNat* pData)
{
	DIM(arr, count);
	for(int i=0; i<count; ++i)
		arr(i) = pData[i];
}

GFDataStoreageContainer* __gpDataContainer=NULL;
void RESTORE(GFDataStoreageContainer& container)
{
	__gpDataContainer=&container;
	__gpDataContainer->index=0;
}

void READ(DGStr& str)
{
	if(__gpDataContainer && __gpDataContainer->index < __gpDataContainer->count)
	{
		
		if(__gpDataContainer->data[__gpDataContainer->index].pC)
			str= DGStr(CGStr(__gpDataContainer->data[__gpDataContainer->index].pC));
		else
			str= DGStr(__gpDataContainer->data[__gpDataContainer->index].n);
		++__gpDataContainer->index;
	}
	else
	{
		__Error(CMP_OUT_OF_DATA); __EndProgram();
	}
}

void READ(DGInt& i)
{
	if(__gpDataContainer && __gpDataContainer->index < __gpDataContainer->count)
	{
		if(__gpDataContainer->data[__gpDataContainer->index].pC)
			i= DGInt(DGStr(__gpDataContainer->data[__gpDataContainer->index].pC));
		else
			i= __gpDataContainer->data[__gpDataContainer->index].n;
		++__gpDataContainer->index;
	}
	else
	{
		__Error(CMP_OUT_OF_DATA); __EndProgram();
	}
}

void READ(DGNat& i)
{
	if(__gpDataContainer && __gpDataContainer->index < __gpDataContainer->count)
	{
		if(__gpDataContainer->data[__gpDataContainer->index].pC)
			i= DGNat(DGStr(__gpDataContainer->data[__gpDataContainer->index].pC));
		else
			i= DGNat(__gpDataContainer->data[__gpDataContainer->index].n);
		++__gpDataContainer->index;
	}
	else
	{
		__Error(CMP_OUT_OF_DATA); __EndProgram();
	}
}


// PRG CONTROL
// ------------------------------------------------------------- //
void	END()
// ------------------------------------------------------------- //
{
	SETLOOPSUB(DGStr(""));
	__Error(CMP_OK);
	__EndProgram();
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
extern const char* __g_pSubCalls[];
void*	GETSUBADDRESS(const DGStr& sub_name)
{
	const char** pC = __g_pSubCalls;
	while(*pC)
	{
		void* pSub = (void*)*pC++;
		const char* pName = *pC++;
		if(sub_name == CGStr(pName))
		{
			return pSub;
		}
	}
	return NULL;
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGNat	CALLBYNAME(const DGStr& sub_name)
{
typedef void (*a_sub)();
	a_sub mysub = (a_sub)GETSUBADDRESS(sub_name);
	
	if(mysub)
	{
		mysub();
		return TRUE;
	}
	else
	{
		// Some info for this bad crash
		if(sub_name == CGStr("GLB_ON_QUIT")) return FALSE;
		if(sub_name == CGStr("GLB_ON_HIDE")) return FALSE;
		if(sub_name == CGStr("GLB_WIN32_MSG")) return FALSE;
		
		WIN32_ONLY( DEBUG(CGStr("CALLBYNAME \"") + sub_name + CGStr("\" *** failed ***\n")) );
		return FALSE;
	}
}



#if defined(EMSCRIPTEN)
	extern "C"
	{
		extern void emscripten_set_main_loop(void (*func)(), int fps);
		extern void emscripten_pause_main_loop();
		extern void emscripten_resume_main_loop();
		extern void emscripten_cancel_main_loop();
		extern void emscripten_run_script(const char*);
	}


	typedef void (*a_sub)();
	a_sub glb_cur_main_sub = NULL;
	void glb_emscripten_loop_wrapper()
	{
		static int is_running = 0;
		if(is_running)
		{
			g_debug("skipping loop frame - too slow\n");
			return;
		}
		is_running = 1;
		if(glb_cur_main_sub)
		{

			#if defined(TRENKI3D)
				glb_cur_main_sub();
				// SHOWSCREEN as desired
			#else
				glb_showscreen_finish_frame(); // this is not really required, I think
				glb_showscreen_init_next_frame();
				glb_cur_main_sub();
			#endif

		// I had this with WebGL and it worked... Required?
		//	SHOWSCREEN(FALSE); // do the dummy initializing _before_ the new loop.

		//	#if defined(TRENKI3D)
		//		SHOWSCREEN(TRUE); // SDL+canvas, please swap
		//	#endif
		}
		is_running = 0;
	}

	// ------------------------------------------------------------- //
	// SetLoopSub
	// ------------------------------------------------------------- //
	void	SETLOOPSUB(const DGStr& sub_name)
	{
		glb_cur_main_sub = (a_sub)GETSUBADDRESS(sub_name);
		if(glb_cur_main_sub)
		{
			g_debug("emsc-loop: %s\n", sub_name.c_str() );
			emscripten_set_main_loop(glb_emscripten_loop_wrapper, 60); // sub, fps
		}
		else
		{
			emscripten_set_main_loop(NULL, 2000); // sub, fps
		}
	}

#else
	DGStr __g_NextLoopSub;
	DGStr __g_CurLoopSub;
	// ------------------------------------------------------------- //
	// SetLoopSub
	// ------------------------------------------------------------- //
	void	SETLOOPSUB(const DGStr& sub_name)
	{
		if(__g_CurLoopSub.len() )
		{
			__g_NextLoopSub = sub_name;
			__g_CurLoopSub.clear();
			return; // the current run loop must do this
		}

		__g_NextLoopSub.clear();
		__g_CurLoopSub = sub_name;

		// ok, just wanted to clear
		if(!__g_CurLoopSub.len() )
			return;


	typedef void (*a_sub)();
		a_sub mysub = (a_sub)GETSUBADDRESS(__g_CurLoopSub);


		while(mysub != NULL)
		{
			mysub();

			if(__g_CurLoopSub.len() == 0)
			{
				__g_CurLoopSub = __g_NextLoopSub;
				if(__g_CurLoopSub.len() == 0)
					return; // called SETLOOPSUB("")
				mysub = (a_sub)GETSUBADDRESS(__g_CurLoopSub);
			}
		}
	}
#endif



// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGStr	MID_Str(const DGStr& s, DGNat start, DGNat len)
{
DGStr	mid;
	int slen=(int)s.len();
	if(len < 0) len = slen; // mid(...-1) -> until the end
	if ((int)(len+start)>slen) len=slen-start;
	if (len<=0 || start>=slen || start<0) {return mid;}// return empty str
	
	mymemcpy(mid.getbuffer(len+1), s.c_str()+start, len);
	mid.setlength(len);
	mid[len] = '\0';
return mid;
}


DGStr	LTRIM_Str(const DGStr& str, const DGStr& trim)
{
    //int itrim = -1;
	const char* pC = str.c_str();
	unsigned int lstr = str.len();
	for(unsigned int i=0; i<lstr; ++i)
	{
		for(unsigned int j=0; j<trim.len(); ++j)
		{
			if(*pC == trim[j])
				goto do_trim;
		}
		// no more char to trim found!
		if(i>0)
		{
			return MID_Str(str, i);
		}
		// string unchanged
		return str;
		do_trim:
		++pC;
	}
	// reached end -> return empty string
	return DGStr("");
}

DGStr	RTRIM_Str(const DGStr& str, const DGStr& trim)
{
    //int itrim = -1;
	int lstr = str.len();
	const char* pC = &(str.c_str()[lstr-1]);
	for(int i=lstr-1; i>=0; --i)
	{
		for(unsigned int j=0; j<trim.len(); ++j)
		{
			if(*pC == trim[j])
				goto do_trim;
		}
		// no more char to trim found!
		if(i<lstr-1)
		{
			return MID_Str(str, 0, i+1);
		}
		// string unchanged
		return str;
		do_trim:
		--pC;
	}
	// reached start -> return empty string
	return DGStr("");
}




// ------------------------------------------------------------- //
// ASC
// ------------------------------------------------------------- //
DGNat ASC(const DGStr& s)
{
	return (DGNat)(unsigned char)s.c_str()[0];
}
DGNat ASC(const DGStr& s, DGNat index)
{
	if(index<0 || index>=(DGNat)s.len())
		return 0;
	return (DGNat)(unsigned char)s.c_str()[index];
}

// ------------------------------------------------------------- //
// CHR$
// ------------------------------------------------------------- //
DGStr CHR_Str(DGNat i)
{
static DGStr s("x"); // len=1, buffer=2
	s[0] = (char)((unsigned char)i);
return s;
}

// ------------------------------------------------------------- //
// HEX$
// ------------------------------------------------------------- //
DGStr	HEX_Str(DGNat n)
{
DGStr str;
	sprintf(str.getbuffer(16), "%x", (int)n);
	str.releasebuffer();
	return str;
}

// ------------------------------------------------------------- //
// FORMAT$
// ------------------------------------------------------------- //
DGStr FORMAT_Str(DGInt minchars, DGInt numaftercomma, DGInt i)
{
static char fmt[32];
static DGStr str;
	sprintf(fmt, "%%%d.%df", (int)minchars, (int)numaftercomma);

#ifdef _DEBUG
	int ireq = _scprintf(fmt, i);
	if(ireq >= 128 )
		*(int*)0=42;
#endif
	sprintf(str.getbuffer(128), fmt, i);
	str.releasebuffer();
	return str;
}

DGStr	FORMAT_Str(DGInt minchars, DGInt numaftercomma, DGNat i)
{
static char fmt[32];
static DGStr str;
	sprintf(fmt, "%%%dd", (int)minchars);

#ifdef _DEBUG
	int ireq = _scprintf(fmt, i);
	if(ireq >= 128 )
		*(int*)0=42;
#endif
	sprintf(str.getbuffer(128), fmt, i);
	str.releasebuffer();
	return str;
}


DGStr	FORMAT_Str(DGInt minchars, DGInt numaftercomma, int64 i)
{
static char fmt[32];
static DGStr str;
	sprintf(fmt, "%%%dlld", (int)minchars);

#ifdef _DEBUG
	int ireq = _scprintf(fmt, i);
	if(ireq >= 128 )
		*(int*)0=42;
#endif
	sprintf(str.getbuffer(128), fmt, i);
	str.releasebuffer();
	return str;
}



// ------------------------------------------------------------- //
// Print to GLBasic output window (This command CANNOT been used on iPHONE at all, its allready defined and used by Apple)
// ------------------------------------------------------------- //
#if defined(WIN32) && !defined(POCKETPC)
extern "C" int __stdcall IsDebuggerPresent();
#endif

// #ifndef TARGET_OS_IPHONE // GF: 2016-10-20
void DEBUG(const DGStr& s)
{
#if defined(WIN32) && !defined(POCKETPC)
	if (IsDebuggerPresent())
	{
		DGStr text = CGStr("TX") + s;
		::OutputDebugStringA(text.c_str());
	}

	#ifdef _DEBUG
		printf("%s", s.c_str());
	#endif

#else
	#ifdef POSIX
	printf("%s", s.c_str());
	#else
	g_debug("%s", s.c_str());
	#endif
#endif
}
// #endif

// ------------------------------------------------------------- //
void	ASSERT(DGNat eq)
{
	if(eq) return;
	WIN32_ONLY( DEBUG(CGStr("Assertion: Value = ") + DGStr(eq) + CGStr("\n")) );
	
	__Error(CMP_ASSERTION_FAILED);
	__EndProgram();
}



// FILE IO
// ------------------------------------------------------------- //
void	GETFILE(const DGStr& file, DGInt line, DGInt& num)
{
DGStr thefile(file);
if(!glb_prepare_reading_file(thefile))
	return;
DGStr	s;	::GetFile(thefile.c_str(), (int)line, s);	num=(DGInt) s;
}
void	GETFILE(const DGStr& file, DGInt line, DGNat& num)
{
DGStr thefile(file);
if(!glb_prepare_reading_file(thefile))
	return;
DGStr	s;	::GetFile(thefile.c_str(), (int)line, s);	num=(DGNat) s;
}

// ------------------------------------------------------------- //
void	GETFILE(const DGStr& file, DGInt line, DGStr& str)
{
DGStr thefile(file);
if(!glb_prepare_reading_file(thefile))
	return;
	::GetFile(thefile.c_str(), (int)line, str);
}

// ------------------------------------------------------------- //
void	PUTFILE(const DGStr& file, DGInt line, DGInt num)
{
DGStr	str;
	str=num;
	::PutFile(file.c_str(), (int)line, str);
}

// ------------------------------------------------------------- //
void	PUTFILE(const DGStr& file, DGInt line, DGStr str)
{
	::PutFile(file.c_str(), (int)line, str);
}

DGStr&			__glb_cmdline(){static DGStr s; return s;}
DGStr&			__glb_AppName(){static DGStr s; return s;}

// iphone fix way to handle urls
extern "C" void __glb_set_cmd_line(const char* pcmd)
{
	__glb_cmdline() = pcmd;
}

// ------------------------------------------------------------- //
DGStr	GETCOMMANDLINE_Str()
{
return __glb_cmdline();
}


// ------------------------------------------------------------- //
DGNat BOXCOLL(DGNat x1, DGNat y1, DGNat w1, DGNat h1, DGNat x2, DGNat y2, DGNat w2, DGNat h2)
{
	#define qmax_(a, b) (a>b ? a : b)
	#define qmin_(a, b) (a<b ? a : b)

	DGNat xe1=x1+w1, xe2=x2+w2, ye1=y1+h1, ye2=y2+h2;
	// only for integers!!
//	if(w1>0) --xe1; else if(w1<0) ++xe1;
//	if(w2>0) --xe2; else if(w2<0) ++xe2;
//	if(h1>0) --ye1; else if(h1<0) ++ye1;
//	if(h2>0) --ye2; else if(h2<0) ++ye2;

	if(
			(qmax_(qmin_(x1, xe1), qmin_(x2, xe2)) <= qmin_(qmax_(x1, xe1), qmax_(x2, xe2))) \
		&&	(qmax_(qmin_(y1, ye1), qmin_(y2, ye2)) <= qmin_(qmax_(y1, ye1), qmax_(y2, ye2)))
	)
		return TRUE;
	return FALSE;

	#undef qmax_
	#undef qmin_
	#undef __RectRectIntersection
}

// ------------------------------------------------------------- //
// GLBasic Console
// ------------------------------------------------------------- //

// ------------------------------------------------------------- //
void	STDOUT(const DGStr& str)
{
	#ifdef ANDROID
	g_debug("%s", str.c_str()); // log print
	#else
	printf("%s", str.c_str());
	fflush(stdout);
	#endif
}

// ------------------------------------------------------------- //
void	STDERR(const DGStr& str)
{
	#ifdef ANDROID
	g_debug("%s", str.c_str()); // log print
	#else
	fprintf(stderr, "%s", str.c_str());
	fflush(stderr);
	#endif
}

// ------------------------------------------------------------- //
DGStr	STDIN_Str()
{
	char buffer[4096];
	buffer[0] = '\0';
	// cin.getline(buffer, 4095, '\n');
	fgets(buffer, 4095, stdin);
	DGStr str(buffer);
	
	// remove '\n' character
	if(str.len() && str[str.len()-1]=='\n')
	{
		// --str.length;
		str.setlength(str.len()-1);
		str[str.len()] = '\0';
	}
	return str;
}





// ------------------------------------------------------------- //
// GLBasic NX
// ------------------------------------------------------------- //

// ------------------------------------------------------------- //
DGNat	NETHOSTGAME(DGNat port_tcp)
{
	__g_DidNET = true;

	if(port_tcp<=0) port_tcp = DEFAULT_PORT;
	
	if(!pGFNet->StartServer((unsigned short)port_tcp) )
		return 0;
return 1;
}

// ------------------------------------------------------------- //
DGNat	NETJOINGAME(const DGStr& address, DGNat port_tcp)
{
	__g_DidNET = true;

	if(port_tcp<=0) port_tcp = DEFAULT_PORT;

	if(!pGFNet->ConnectClient(address.c_str(), (unsigned short)port_tcp ))
		return 0;
	return 1;
}

// ------------------------------------------------------------- //
void	NETSHUTDOWN()
{
	pGFNet->Stop();
}


// ------------------------------------------------------------- //
void	NETALLOWJOINING(DGNat allow)
{
	pGFNet->AllowJoining(allow ? true:false);
}

// ------------------------------------------------------------- //
DGNat	NETCREATEPLAYER(const DGStr& Name)
{
DGNat	id;
	id = (DGNat)pGFNet->CreatePlayer(Name.c_str());
return id;
}

// ------------------------------------------------------------- //
void	NETDESTROYPLAYER(DGNat Player)
{
	pGFNet->DeletePlayer((unsigned short)Player);
}

// ------------------------------------------------------------- //
DGNat	NETSENDMSG(DGNat Player, DGNat Recipent, const DGStr& Messy)
{
	if (Recipent==0)
		Recipent=ID_ALL_PLAYERS;
	return (DGNat) pGFNet->SendASCIIMessage((unsigned short)Player, (unsigned short)Recipent, Messy.c_str());
}


static long __g_last_msg_sender=0;
// ------------------------------------------------------------- //
DGStr	NETGETMSG_Str(DGNat Player)
{
	GFHawkNetwork::Message msg;
	if(pGFNet->PopMessage((unsigned short)Player, msg))
	{
		__g_last_msg_sender=msg.id_from;
		DGStr s = msg.pBuffer;
		return s;
	}
return DGStr("");
}

// ------------------------------------------------------------- //
DGNat	NETGETSENDER()
{
	return (DGNat)__g_last_msg_sender;
}

// ------------------------------------------------------------- //
DGStr	NETPLAYERNAME_Str(DGNat Player)
{
DGStr	s;
	if(pGFNet->GetPlayerName((unsigned short)Player, s.getbuffer(256)))
			s.releasebuffer();
	else
		s="NO_DATA";
return s;
}

// ------------------------------------------------------------- //
DGNat	NETNUMPLAYERS()
{
	return pGFNet->NumPlayers();
}

// ------------------------------------------------------------- //
DGNat	NETGETQUITPLAYER()
{
	return (DGNat)pGFNet->GetQuitPlayer();
}

// ------------------------------------------------------------- //
DGNat	NETGETPLAYERID(DGNat num)
{
	return (DGNat)pGFNet->GetPlayerIdByIndex((unsigned short)num);
}

// ------------------------------------------------------------- //
DGStr	NETGETIP_Str()
{
char* pNames = new char[4096];
DGStr s;
	int n = pGFNet->GetLocalIPs(pNames, 4096);
	
	if(n)
		s = pNames;
	delete [] pNames;
	return s;	
}

// ------------------------------------------------------------- //
DGStr	NETGETLASTERROR_Str()
{
	return (DGStr)netLastError();
}

// ------------------------------------------------------------- //
DGNat	NETWEBGET(const DGStr& server, const DGStr& path, DGNat port, const DGStr& local, DGNat timeout_ms)
{
	return (DGNat)pGFNet->WGet(server.c_str(), path.c_str(), (int)port, local.c_str(), NULL, 0, timeout_ms);
}

// ------------------------------------------------------------- //
DGStr	NETWEBGET_Str(const DGStr& server, const DGStr& path, DGNat port, DGNat max_length, DGNat timeout_ms)
{
DGStr	str;
	pGFNet->WGet(server.c_str(), path.c_str(), (unsigned short)port, NULL, str.getbuffer(max_length+2), max_length+1, timeout_ms);
	str.releasebuffer();
	return str;
}

#if defined(TARGET_OS_IPHONE)
	extern "C" void iPhoneOpenUrl(const char* pUrl);
#endif
#if defined(ANDROID)
	extern "C" int android_openURL(DGStr url);
#endif

extern "C" void __ShutDownGLBasic();
// ------------------------------------------------------------- //
void	NETWEBEND(DGStr url)
{
	if(glb_allow_escape())
		__ShutDownGLBasic();

	#ifdef WIN32
		#if defined(UNICODE)
			wchar_t pUrl[1024];
			// convert from multibyte char string to wide char string
			mbstowcs(pUrl, (const char *)url.c_str(), url.len()+1);
		#else
			const char* pUrl = url.c_str();
		#endif
		SHELLEXECUTEINFO cShellExecuteInfo;
		memset(&cShellExecuteInfo, 0, sizeof(SHELLEXECUTEINFO));
		cShellExecuteInfo.cbSize = sizeof(SHELLEXECUTEINFO);
		cShellExecuteInfo.fMask = SEE_MASK_NOCLOSEPROCESS;
		cShellExecuteInfo.hwnd = NULL;
		cShellExecuteInfo.lpVerb = _T("Open");
		#if defined(_WIN32_WCE)
			cShellExecuteInfo.lpFile = _T("\\Windows\\IExplore.exe");
			cShellExecuteInfo.lpParameters = pUrl;
		#else
			cShellExecuteInfo.lpFile =pUrl;
		#endif
		cShellExecuteInfo.nShow = SW_SHOWNORMAL;
		ShellExecuteEx(&cShellExecuteInfo);
	#elif defined(TARGET_OS_IPHONE)
		printf("exiting with url %s\nIgnore the bad exec that will come now ;)", url.c_str());
		iPhoneOpenUrl(url.c_str());
		// exit(0);
	#elif defined(PALM_WEBOS)
		int rv = PDL_LaunchBrowser(url.c_str());
		printf("exiting with url %s\nIgnore the segfault that will come now ;)", url.c_str());
	#elif defined(ANDROID)
		int rv = android_openURL(url);
		g_debug("check url back");
		//		g_debug("open url: %s %s\n", url, rv?"ok":"fail");
	#elif defined(LINUX)
		const char* browser = "/usr/bin/sensible-browser";
		execl(browser, browser, url.c_str(), NULL);
	#elif defined(MACOSX)
		const char* browser = "/usr/bin/open";
		execl(browser, browser, url.c_str(), NULL);
	#endif

	if(glb_allow_escape())
		__EndProgram();
}

// ------------------------------------------------------------- //
DGNat   NETISACTIVE()
{
	return pGFNet->Running() ? TRUE:FALSE;
}

// ------------------------------------------------------------- //
// Socket commands TCP
// ------------------------------------------------------------- //

static NETaddress glb_netaddress_from_dgstr(const DGStr& ip, DGNat port)
{
	NETaddress addr;
	if(ip.len() == 4)
	{
		addr.assign(NETaddress::get_family_ofIPv(4), (void*)ip.c_str(), port );
	}
	else if(ip.len() == 16)
	{
		addr.assign(NETaddress::get_family_ofIPv(6), (void*)ip.c_str(), port );
	}
	else
	{
		__Error(CMP_WRONG_ARGUMENT);
		return NET_ADDR_INVALID;
	}
	return addr;
}

static DGStr glb_ipstr_from_netaddress(const NETaddress& add)
{
	DGStr out;
	if(add.m_ipFamily == NETaddress::get_family_ofIPv(6))
	{
		memcpy(out.getbuffer(17), add.m_ip, 16);
		out.setlength(16);
	}
	else
	{
		memcpy(out.getbuffer(5), add.m_ip, 4);
		out.setlength(4);
	}
	return out;
}


// ------------------------------------------------------------- //
void  SOCK_SET_IP_VERSION(DGNat ipV)
{
	if(ipV!=4 && ipV!=6)
	{
		__Error(CMP_WRONG_ARGUMENT);
		ipV=4;
	}	
	netSetVersion(ipV);
}

// ------------------------------------------------------------- //
DGNat SOCK_TCPLISTEN(DGNat port)
	{return (DGNat)netListen((unsigned short) port);}

// ------------------------------------------------------------- //
DGNat SOCK_TCPCONNECT(const DGStr& server, DGNat port, DGNat timeout)
{
	NETaddress sv = netAddress(server.c_str(), port);
	if(sv== NET_ADDR_INVALID)
		return NET_INVALID;
	return (DGNat)netConnect(sv, timeout);
}

// ------------------------------------------------------------- //
// return ip in host format
// port in host format
DGNat SOCK_TCPACCEPT(DGNat socket, DGStr& ip)
{
	ip.clear();

	NETaddress addr;
	NETsocket sock = netAccept(socket, &addr);
	if(sock!=NET_INVALID)
	{
		ip = glb_ipstr_from_netaddress(addr);
		// ip = ne148tGetIPstr(addr); // readable form

		// ip = FromBigEndian((unsigned long)netGetIP(addr));
		// ip = (int)netGetIP(addr); // already _is_ local format
	}
	return (DGNat)sock;
}

// ------------------------------------------------------------- //
DGNat SOCK_TCPSEND(DGNat sock_to, const DGStr& msg)
{
	return netSend(sock_to, msg.c_str(), msg.len());
}


// ------------------------------------------------------------- //
// Socket commands UDP
// ------------------------------------------------------------- //

// ------------------------------------------------------------- //
DGNat SOCK_UDPOPEN(DGNat port)
{
	return (DGNat) netSocketUdp((unsigned short) port);
}


// ip in host format -> use SOCK_GETIP
// port in host format
// ------------------------------------------------------------- //
DGNat SOCK_UDPSEND(DGNat sock, const DGStr& msg, DGStr ip, DGNat port)
{
	NETaddress addr = glb_netaddress_from_dgstr(ip, port);
	return netSend(sock, msg.c_str(), msg.len(), &addr);
}


// ------------------------------------------------------------- //
DGNat SOCK_RECV(DGNat sock_from, DGStr& msg, DGNat length)
{
	int rv = netRecv(sock_from, msg.getbuffer(length+1), length);
	if(rv>0)
	{
		msg[rv] = '\0';
//		msg.releasebuffer(); // that's just plain slow - might need to check against rv>=length?
		msg.setlength(rv); // you can include '\0' characters in GLB!
	}
	else
	{
		msg.releasebuffer();
		msg.assign("");
		msg.setlength(0);
	}
	return rv;
}

// ------------------------------------------------------------- //
// Socket commands TCP/UDP
// ------------------------------------------------------------- //


// ------------------------------------------------------------- //
// "glbasic.com" -> binary IP
DGStr SOCK_GETHOSTIP_Str(const DGStr& server)
{
	NETaddress add = netAddress(server.c_str());
	return glb_ipstr_from_netaddress(add);
}

// ------------------------------------------------------------- //
// binary IP -> readable format
DGStr SOCK_GETIP_Str(const DGStr& ip)
{
	NETaddress add = glb_netaddress_from_dgstr(ip, 0);
	return DGStr(netGetIPstr(add));
}

// ------------------------------------------------------------- //
DGStr SOCK_GETREMOTEIP_Str(DGNat sock)
{
	DGStr out;
	NETaddress add = netGetRemoteAddress(sock);

	return glb_ipstr_from_netaddress(add);
}

// ------------------------------------------------------------- //
DGNat SOCK_GETREMOTEPORT(DGNat sock)
{
	return (DGNat)FromBigEndian((unsigned short)netGetRemoteAddress(sock).m_port);
}


// ------------------------------------------------------------- //
DGNat SOCK_CLOSE(DGNat socket)
{
	return netCloseSocket(socket) == 0 ? TRUE:FALSE;
}


// ------------------------------------------------------------- //
void SOCK_SETBLOCKING(DGNat socket, DGNat block)
{
	netSetBlocking(socket, block?true:false);
}

// ------------------------------------------------------------- //
DGNat SOCK_INIT()
{
	__g_DidNET= true;
	return netInit() ? TRUE:FALSE;
}

// ------------------------------------------------------------- //
void SOCK_SHUTDOWN()
{
	netShutdown();
}


// natural array for map
DGNat FINDPATH(DGNatArray& map, DGNatArray& solutions, DGInt heuristic, DGNat xstart, DGNat ystart, DGNat xend, DGNat yend)
{
	DGIntArray fmap;
	DIM(fmap, LEN(map), BOUNDS(map,1));
	for(int x=0; x<LEN(map); ++x)
	for(int y=0; y<BOUNDS(map,1); ++y)
	{
		fmap(x,y) = map(x,y);
	}
	return FINDPATH(fmap, solutions, heuristic, xstart, ystart, xend, yend);
}

// DGInt array for return values
DGNat FINDPATH(DGIntArray& map, DGIntArray& solutions, DGInt heuristic, DGNat xstart, DGNat ystart, DGNat xend, DGNat yend)
{
	DGNatArray nsol;
	DGNat rv = FINDPATH(map, nsol, heuristic, xstart, ystart, xend, yend);
	DIM(solutions, LEN(nsol), 2);
	for(int i=0; i<LEN(nsol); ++i)
	{
		solutions(i,0) = nsol(i,0);
		solutions(i,1) = nsol(i,1);
	}
	return rv;
}

// ------------------------------------------------------------- //
// FINDPATH - fastest method
// ------------------------------------------------------------- //
DGNat FINDPATH(DGIntArray& map, DGNatArray& solutions, DGInt heuristic, DGNat xstart, DGNat ystart, DGNat xend, DGNat yend)
{
	AStarPlanner			p1;
	Planner::Description	desc;
	memset(&desc, 0, sizeof(desc));

	desc.bDiagonalCost=true;

	desc.startX = (int)xstart;
	desc.startY = (int)ystart;
	desc.endX = (int)xend;
	desc.endY = (int)yend;

	desc.width =map.max1;
	desc.height=map.max2;

	desc.hashPrecision = 8;
	desc.heuristicWeight = (float)MAX(0.0, MIN(heuristic, 1.0));

	desc.can_walk_diagonally = false;

	#ifdef __arm__
	// DGInt != double
	double* pDataTmp = NULL;
		pDataTmp = new double[map.max1*map.max2+1];
		int j=0;
		for (int x=0; x<map.max1; ++x)
		for (int y=0; y<map.max2; ++y)
			{pDataTmp[j] = map.pData[j]; ++j;}
		desc.pMapData = pDataTmp;
	#else
		desc.pMapData = (double*)map.pData;
	#endif

	DGNat rv=0;
	if(!p1.createMap(desc)) return FALSE;

	p1.initialize();
	if (p1.run())
	{
		const Node* pNode = p1.getGoal();
		int numnodes = 0;
		while(pNode)
		{
			numnodes++;
			pNode = pNode->pParent;
		}
		if (numnodes==0)
		{
			DIM(solutions, 0);
			rv = 0;
			goto okay;
		}
		DIM(solutions, (int)numnodes, 2);
		pNode = p1.getGoal();
		while(pNode)
		{
			numnodes--;
			solutions((int)numnodes, 0) = pNode->x;
			solutions((int)numnodes, 1) = pNode->y;
			pNode = pNode->pParent;
		}
		p1.cleanup();
		rv = solutions.max1;
		goto okay;
	}
	p1.cleanup();

okay:
	#ifdef __arm__
	if(pDataTmp) delete [] pDataTmp;
	#endif
	
return rv;
}



// ------------------------
// GRAPHICS
// ------------------------

// ------------------------------------------------------------- //

void	CLEARSCREEN(DGNat col)
{
#ifndef GLB_CONSOLE
	__g_pRbow->FreeSurface(0);
	__g_pRbow->ClearScreen(col);
#else
	#ifdef WIN32
		system("cls");
	#else
		system("clear");
	#endif
#endif
}

// ------------------------------------------------------------- //
void	BLENDSCREEN(const DGStr& s, DGNat msecs)
{
	CLEARERROR();
#ifndef GLB_CONSOLE
	DGStr file(s);

	if(!glb_prepare_reading_file(file))
	{
		__Error(CMP_NO_FILE);
		CLEARSCREEN();
		return;
	}

	__g_pRbow->GrabSurface(2);
	__g_pRbow->LoadSurface(1, file.c_str());
	CLEARSCREEN();


float fade;
const float fulltime = msecs;

	float endtime = (float)GETTIMERALL() + fulltime;
	fade=0.9f;
	while(fade<1.0 && fade>0)
	{
		ALPHAMODE(fade);
		__g_pRbow->PasteSurface(2);
		ALPHAMODE(1.0f-fade);
		__g_pRbow->PasteSurface(1);
		SHOWSCREEN();
		fade = (float)((endtime - GETTIMERALL()) / fulltime);
	}

	ALPHAMODE(0);
	__g_pRbow->PasteSurface(1);
	USEASBMP();

	__g_pRbow->FreeSurface(1);
	__g_pRbow->FreeSurface(2);

	SHOWSCREEN();
#endif
}


// ------------------------------------------------------------- //
void	PRINT(const DGStr& s, DGInt x, DGInt y, DGNat kerning)
{
#ifndef GLB_CONSOLE
	__g_pRbow->Text2D((int)__nCurFont, s.c_str(), (float)x, (float)y, 0.0f, kerning, TRUE);
#endif
}


// Drawing text width
DGNat	LEN(const DGStr& s, DGNat kerning)
{
#ifndef GLB_CONSOLE
	return __g_pRbow->Text2DWidth((int)__nCurFont, s.c_str(), kerning);
#endif
return LEN(s);
}

DGNat	LEN(const CGStr& s, DGNat kerning)
{
#ifndef GLB_CONSOLE
	return __g_pRbow->Text2DWidth((int)__nCurFont, s.c_str(), kerning);
#endif
return LEN(s);
}



// ------------------------------------------------------------- //
void	LOADANIM(DGStr s, DGNat n, DGNat width, DGNat height)
{
#ifndef GLB_CONSOLE
	LOADSPRITE(s, n);
	if(__g_pRbow->GetSpriteSize((int)n).cx>0)
		__g_pRbow->TileSprite((int)n, (int)width, (int)height);
#endif
}

// ------------------------------------------------------------- //
void	SETSPRITEANIM(DGNat n, DGNat width, DGNat height)
{
#ifndef GLB_CONSOLE
	if(__g_pRbow->GetSpriteSize((int)n).cx>0)
		__g_pRbow->TileSprite((int)n, (int)width, (int)height);
#endif
}

// ------------------------------------------------------------- //
void	DRAWANIM(DGNat num, DGNat iTile, DGNat x, DGNat y)
{
#ifndef GLB_CONSOLE
	__g_pRbow->DrawTile(num, iTile, x,y);
#endif
}

// ------------------------------------------------------------- //
void	ROTOZOOMANIM(DGNat num, DGNat itile, DGNat x, DGNat y, DGInt phi, DGInt scale)
{
#ifndef GLB_CONSOLE
	__g_pRbow->RotoZoomTile((int)num, (int)itile, (int)x, (int)y, (float)phi, (float)scale);
#endif
}


void	STRETCHANIM(DGNat num, DGNat itile, DGNat x, DGNat y, DGNat w, DGNat h)
{
#ifndef GLB_CONSOLE
	__g_pRbow->StretchTile((int)num, (int)itile, (int)x, (int)y, (int)w, (int)h);
#endif
}


// ------------------------------------------------------------- //
DGNat	ANIMCOLL(DGNat ani1, DGNat tile1, DGNat x1, DGNat y1,  DGNat ani2, DGNat tile2, DGNat x2, DGNat y2)
{
#ifndef GLB_CONSOLE
	return __g_pRbow->TileCollision((int)ani1, (int)tile1, (int)x1, (int)y1,
									(int)ani2, (int)tile2, (int)x2, (int)y2);
#else
	return 0;
#endif
}

// ------------------------------------------------------------- //
void	SMOOTHSHADING(DGNat smooth)
{
#ifndef GLB_CONSOLE
	__g_pRbow->SetSmoothShading(smooth ? true:false);
#endif
}

// ------------------------------------------------------------- //
void	ALPHAMODE(DGInt alpha)
{
#ifndef GLB_CONSOLE
	__g_pRbow->DoAlpha((float)alpha);
#endif
}

// ------------------------------------------------------------- //
void	ALPHATESTING(DGInt alpha)
{
#ifndef GLB_CONSOLE
	__g_pRbow->SetAlphaTesting((float)alpha);
#endif
}

// ------------------------------------------------------------- //
void	ROTOSPRITE(DGNat num, DGNat x, DGNat y, DGInt angle)
{
#ifndef GLB_CONSOLE
	__g_pRbow->RotoSprite((int)num, (int)x, (int)y, (float)angle, 1.0f);
#endif
}

// ------------------------------------------------------------- //
void	ROTOZOOMSPRITE(DGNat num, DGNat x, DGNat y, DGInt angle, DGInt size) // size[%]
{
#ifndef GLB_CONSOLE
	__g_pRbow->RotoSprite((int)num, (int)x, (int)y, (float)angle, (float)size);
#endif
}

// ------------------------------------------------------------- //
void	DRAWSPRITE(DGNat num, DGNat x, DGNat y)
{
#ifndef GLB_CONSOLE
	__g_pRbow->DrawSprite(num, x, y);
#endif
}

// ------------------------------------------------------------- //
void STRETCHSPRITE (DGNat num, DGNat px1, DGNat py1, DGNat wx, DGNat wy)
{
#ifndef GLB_CONSOLE
	__g_pRbow->StretchSprite((int)num, (int)px1, (int)py1, (int)wx, (int)wy);
#endif
}

// ------------------------------------------------------------- //
void	ZOOMSPRITE(DGNat num, DGNat x, DGNat y, DGInt xsize, DGInt ysize)
{
#ifndef GLB_CONSOLE
	xsize=((xsize*__g_pRbow->m_Texture[(int)num].bw));
	ysize=((ysize*__g_pRbow->m_Texture[(int)num].bh));

	x-=((DGNat)(xsize-(DGInt)__g_pRbow->m_Texture[(int)num].bw)/2);
	y-=((DGNat)(ysize-(DGInt)__g_pRbow->m_Texture[(int)num].bh)/2);

	__g_pRbow->StretchSprite((int)num, (int)x, (int)y, (int)xsize, (int)ysize);
#endif
}

// ------------------------------------------------------------- //
void GRABSPRITE(DGNat num, DGNat x, DGNat y, DGNat wx, DGNat wy)
{
#ifndef GLB_CONSOLE
	if(__g_Prescaler.scale > 1.0f)
	{
		int spx=wx, spy=wy;
		x+=__g_Prescaler.offx;
		y+=__g_Prescaler.offy;
		wx = (int)(float(wx + 0.99) * __g_Prescaler.scale);
		wy = (int)(float(wy + 0.99) * __g_Prescaler.scale);


		int dmy = GENSPRITE();

		__g_pRbow->GrabSprite(dmy, x, y, wx, wy);
		int fbo = GENSCREEN();
		CREATESCREEN(fbo, num, spx, spy);
		USESCREEN(fbo);
		ALPHAMODE(-1);
		ALPHATESTING(1); // solid
		STRETCHSPRITE(dmy, 0,0,spx,spy);
		USESCREEN(-1);
		LOADSPRITE(DGStr(""), dmy); // clear dummy
	}
	else // no prescaler
	{
		__g_pRbow->GrabSprite((int)num, (int)x, (int)y, (int)wx, (int)wy);
	}

#endif
}

// ------------------------------------------------------------- //
void	USEASBMP()
{
#ifndef GLB_CONSOLE
	__g_pRbow->GrabSurface(0);
#endif
}

// ------------------------------------------------------------- //
void	LOADBMP(const DGStr& s)
{
	CLEARERROR();

#ifndef GLB_CONSOLE
	DGStr file(s);

	if(!glb_prepare_reading_file(file))
	{
		__Error(CMP_NO_FILE);
		CLEARSCREEN();
		return;
	}

	if (!__g_pRbow->LoadSurface(0, file.c_str()))
	{
		__Error(CMP_NO_FILE);
		CLEARSCREEN();
		return;
	}
	__g_pRbow->PasteSurface(0);
#endif
}

// ------------------------------------------------------------- //
void	SETFONT(DGNat i, DGInt scale)
{
#ifndef GLB_CONSOLE
	if(i<0 || i>=GLR_MAX_FONT)
	{
		__Error(CMP_OUT_OF_MEMORY);
		return;
	}
	__nCurFont = i;
	__g_pRbow->m_FontScaling[i] = scale;
#endif
}
// ------------------------------------------------------------- //
DGNat	GETCURRENTFONT() { return __nCurFont;}


// ------------------------------------------------------------- //
void	SETTRANSPARENCY(DGNat col)
{
#ifndef GLB_CONSOLE
	__g_pRbow->SetColorKey((COLORREF) col);
#endif
}


// ------------------------------------------------------------- //
void	LOADFONT(const DGStr& s, DGNat num)
{
	CLEARERROR();

#ifndef GLB_CONSOLE
	if(LEN(s) == 0)
	{
		// clear the font out - not visible anymore
		__g_pRbow->LoadFontPIX(num, "");
		return;
	}

	SETFONT(num, 1.0);
	DGStr file(s);
	if(!glb_prepare_reading_file(file))
	{
		__Error(CMP_NO_FILE);
		__g_pRbow->DefaultFont((int)num);
		return;
	}

	if (!__g_pRbow->LoadFontPIX(__nCurFont, file.c_str()))
	{
		__Error(CMP_NO_FILE);
		__g_pRbow->DefaultFont((int)num);
		return;
	}
#endif
}

// ------------------------------------------------------------- //
void	LOADFONT(const short* ttf, DGNat i)
{
	CLEARERROR();
#ifndef GLB_CONSOLE
	__g_pRbow->LoadFontTTF(i, ttf);
#endif
}

// ------------------------------------------------------------- //
void	LOADSPRITE(const DGStr& s, DGNat n)
{
	CLEARERROR();
#ifndef GLB_CONSOLE
	DGStr file(s);
	__g_pRbow->FreeSprite((int)n);

	if(s.len() == 0)
	{
		return;
	}
	if(!glb_prepare_reading_file(file))
	{
		__Error(CMP_NO_FILE);
		return;
	}

	if(!__g_pRbow->LoadSprite((int)n, file.c_str(), __g_pRbow->m_CookieColor, false))
	{
		__Error(CMP_NO_FILE);
	}


#endif
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
DGNat	MEM2SPRITE(DGNatArray& pixRGBA, DGNat num, DGNat w, DGNat h)
{
#ifdef GLB_CONSOLE
	return FALSE;
#else
	int sz = w*h;
	if(LEN(pixRGBA) != sz) return FALSE;

	OPENGL_IMAGE img;
	img.resize(w,h);
	unsigned char* pOut = img.m_txdata;
	const DGNat* pIn = &pixRGBA(0);
	for(int i=0; i<sz; ++i)
	{
		unsigned int pixl = *pIn++;
	
		// printf("Rgba: %x %x %x %x\n", (pixl & 0xff), ((pixl>>8) & 0xff), ((pixl>>16) & 0xff), ((pixl>>24) & 0xff));
		*pOut++ = (pixl & 0xff);
		*pOut++ = ((pixl>>8) & 0xff);
		*pOut++ = ((pixl>>16) & 0xff);
		*pOut++ = ((pixl>>24) & 0xff);
	}
	// printf("\n");
	return __g_pRbow->LoadSprite(num, img, __g_pRbow->m_CookieColor, false) ? TRUE:FALSE;
#endif
}


// ------------------------------------------------------------- //
DGNat SPRITE2MEM(DGNatArray& pixRGBA, DGNat num)
{
#ifdef GLB_CONSOLE
	return FALSE;
#else
DGNat w, h;
	GETSPRITESIZE(num, w,h);
	int sz = w*h;
	if(sz==0)
		return FALSE;
	OPENGL_IMAGE img; // keep memory for next run

	bool ok = __g_pRbow->GetSpriteRGBA(num, img);
	if(!ok)
		return FALSE;

	FIXDIM(pixRGBA, sz);
	
	const unsigned char* pOut = img.m_txdata;
	DGNat* pIn = &pixRGBA(0);
	for(int i=0; i<sz; ++i)
	{
		int r = *pOut++;
		int g = *pOut++;
		int b = *pOut++;
		int a = *pOut++;
		// printf("rgba: %x %x %x %x\n", r,g,b,a);
		*pIn++ = r+(g<<8)+(b<<16)+(a<<24);
	}
	// printf("\n");
	return TRUE;
#endif
}


// ------------------------------------------------------------- //
DGNat LOADSPRITEMEM(const DGStr& s, DGNat& w, DGNat& h, DGNatArray& pixRGBA)
{
#ifdef GLB_CONSOLE
	return FALSE;
#else
	DGStr file(s);
	if(!glb_prepare_reading_file(file))
	{
		__Error(CMP_NO_FILE);
		return FALSE;
	}

	OPENGL_IMAGE img;
	if(!img.LoadImageBMP(file.c_str(), __g_pRbow->m_CookieColor) && !img.LoadImagePNG(file.c_str(), __g_pRbow->m_CookieColor) && !img.LoadImageJPG(file.c_str()))
	{
		return FALSE;
	}
	if(!img.m_bSourceHadAlpha)
		img.FixNeighbourAlpha();

	w = img.m_sizex;
	h = img.m_sizey;
	DGNat sz = w*h;	

	if(sz==0)
		return FALSE;

	FIXDIM(pixRGBA, sz);
	
	const unsigned char* pOut = img.m_txdata;
	DGNat* pIn = &pixRGBA(0);
	for(int i=0; i<sz; ++i)
	{
		int r = *pOut++;
		int g = *pOut++;
		int b = *pOut++;
		int a = *pOut++;
		*pIn++ = r+(g<<8)+(b<<16)+(a<<24);
	}
	return TRUE;
#endif
}



// ------------------------------------------------------------- //
#ifndef GLB_CONSOLE
unsigned int get_sprite_texture(int id)
{
	return (unsigned int)__g_pRbow->m_Texture[id].mipstate[0];
}
#endif

// ------------------------------------------------------------- //
void	LOADBUMPTEXTURE(const DGStr& s, DGNat n)
{
	CLEARERROR();
#ifndef GLB_CONSOLE
	__g_pRbow->FreeSprite((int)n);
	DGStr file(s);
	if(!glb_prepare_reading_file(file))
	{
		__Error(CMP_NO_FILE);
		return;
	}

	if(!__g_pRbow->LoadSprite((int)n, file.c_str(), __g_pRbow->m_CookieColor, true))
	{
		__Error(CMP_NO_FILE);
	}
#endif
}
// ------------------------------------------------------------- //
void	SAVEBMP(const DGStr& s)
{
	CLEARERROR();
#ifndef GLB_CONSOLE
	if(!__g_pRbow->SaveBMPFile(s.c_str()))
	{
		__Error(CMP_NO_FILE);
	}
#endif
}

// ------------------------------------------------------------- //
void	SAVESPRITE(const DGStr& s, DGNat n)
{
	CLEARERROR();
#ifndef GLB_CONSOLE
	if(!__g_pRbow->SaveBMPSprite((int)n, s.c_str()))
	{
		__Error(CMP_NO_FILE);
	}
#endif
}

// ------------------------------------------------------------- //
DGNat	SPRCOLL(DGNat n1, DGNat x1, DGNat y1, DGNat n2, DGNat x2, DGNat y2)
{
#ifndef GLB_CONSOLE
	return __g_pRbow->SpriteCollision((int)n1, (int)n2, (int)x1, (int)y1, (int)x2, (int)y2);
#else
	return 0;
#endif
}

// ------------------------------------------------------------- //
void	DRAWLINE(DGNat sx, DGNat sy, DGNat ex, DGNat ey, DGNat col)
{
#ifndef GLB_CONSOLE
	__g_pRbow->DrawLine((int)sx, (int)sy, (int)ex, (int)ey, (COLORREF)col);
#endif
}

// ------------------------------------------------------------- //
void	SETPIXEL(DGNat x,  DGNat y,  DGNat col)
{
#ifndef GLB_CONSOLE
	__g_pRbow->DrawPixel((int)x, (int)y, (COLORREF)col);
#endif
}

// ------------------------------------------------------------- //
void	DRAWRECT(DGNat sx, DGNat sy, DGNat wx, DGNat wy, DGNat col)
{
#ifndef GLB_CONSOLE
	__g_pRbow->DrawRect((int)sx,
					(int)sy,
					(int)wx,
					(int)wy,
					(COLORREF)col);
#endif
}

DGNat	GETPIXEL(DGNat x, DGNat y)
{
#ifndef GLB_CONSOLE
	return (DGNat) __g_pRbow->GetPixel((int)x, (int)y);
#else
	return 0;
#endif
}


// ------------------------------------------------------------- //
void	GETSPRITESIZE(DGInt id, DGInt& wx, DGInt& wy)
{
#ifndef GLB_CONSOLE
	wx = (DGInt)__g_pRbow->GetSpriteSize((int)id).cx;
	wy = (DGInt)__g_pRbow->GetSpriteSize((int)id).cy;
#endif
}
// ------------------------------------------------------------- //
void	GETSPRITESIZE(DGNat id, DGNat& wx, DGNat& wy)
{
#ifndef GLB_CONSOLE
	wx = (DGNat)__g_pRbow->GetSpriteSize((int)id).cx;
	wy = (DGNat)__g_pRbow->GetSpriteSize((int)id).cy;
#endif
}

// ------------------------------------------------------------- //
void	GETFONTSIZE(DGInt& wx, DGInt& wy)
{
#ifndef GLB_CONSOLE
	SIZE sz = __g_pRbow->GetFontSize(__nCurFont);
	float sc = __g_pRbow->m_FontScaling[__nCurFont];
	wx = (DGInt) (sz.cx*sc); wy = (DGInt) (sz.cy*sc);
#endif
}
// ------------------------------------------------------------- //
void	GETFONTSIZE(DGNat& wx, DGNat& wy)
{
#ifndef GLB_CONSOLE
	SIZE sz = __g_pRbow->GetFontSize(__nCurFont);
	float sc = __g_pRbow->m_FontScaling[__nCurFont];
	wx = (DGNat) (sz.cx*sc); wy = (DGNat) (sz.cy*sc);
#endif
}

// ------------------------------------------------------------- //
void	GETSCREENSIZE(DGInt& wx, DGInt& wy)
{
	DGNat w, h;
	GETSCREENSIZE(w,h);
	wx=w; wy=h;
}
// ------------------------------------------------------------- //
void	GETSCREENSIZE(DGNat& wx, DGNat& wy)
{
#ifndef GLB_CONSOLE

	// GETSCREENSIZE_SEE_ALSO
	if(__g_pRbow->m_CurrentFBO<0)
	{
		if(__g_pRbow->m_landscape==1 || __g_pRbow->m_landscape==3)
		{
			wy = (DGNat)__g_pRbow->m_WND.virtw;
			wx = (DGNat)__g_pRbow->m_WND.virth;
		}
		else
		{
			wx = (DGNat)__g_pRbow->m_WND.virtw;
			wy = (DGNat)__g_pRbow->m_WND.virth;
		}
	}
	else
	{
		wx = __g_pRbow->m_Texture[__g_pRbow->m_FBuffer[__g_pRbow->m_CurrentFBO].m_SpriteIndex].bw;
		wy = __g_pRbow->m_Texture[__g_pRbow->m_FBuffer[__g_pRbow->m_CurrentFBO].m_SpriteIndex].bh;
	}

#else
	#ifdef WIN32
		CONSOLE_SCREEN_BUFFER_INFO info;
		::GetConsoleScreenBufferInfo(::GetStdHandle(STD_OUTPUT_HANDLE), &info);
		wx = info.dwMaximumWindowSize.X;
		wy = info.srWindow.Bottom-info.srWindow.Top+1;
	#else
		// at least this
		wx=80;
		wy=24;

		winsize ws;
		if (ioctl(0,TIOCGWINSZ,&ws)==0)
		{
			wx = ws.ws_row;
			wy = ws.ws_col;
		}
	#endif
#endif
}

void	SETORIENTATION(DGNat imode)
{
#if defined(ANDROID) // SPACE FRACTAL - Fix Orientation on Android when -1 is used).
	if (imode==-1)
	{	SETORIENTATION(JavaCall_Str(CGStr("getOrientation:")));
	}
#endif

	
#ifndef GLB_CONSOLE
	__g_pRbow->m_landscape = MAX(0, MIN(3, imode) );
	VIEWPORT(0,0,0,0);
#endif
}

DGNat	GETORIENTATION()
{
#ifndef GLB_CONSOLE
	return __g_pRbow->m_landscape;
#endif
return 0;
}


// ------------------------------------------------------------- //
void	GETDESKTOPSIZE(DGInt& wx, DGInt& wy)
	{wx = __DG_DESKTOPX; wy=__DG_DESKTOPY;}
void	GETDESKTOPSIZE(DGNat& wx, DGNat& wy)
	{wx = __DG_DESKTOPX; wy=__DG_DESKTOPY;}


void STDCOLOR(DGNat iForeground, DGNat iBackground)
{
#ifdef GLB_CONSOLE
	static int map_col[] = {0,4,2,6,1,5,3,7,
							100,104,102,106,101,105,103,107};
	if(iForeground<0 || iForeground>15)
		iForeground = 15;
	if(iBackground<0 || iBackground>15)
		iBackground=0;

	#ifdef WIN32
		// 0=black, 1=blue,        2=green,        3=cyan,        4=red,        5=purple,        6=yellow,        7=ltgray
		// 8=gray   9=light blue, 10=light green, 11=light cyan, 12=light red, 13=light purple, 14=light yellow, 15=white
		SetConsoleTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), iForeground+iBackground*16);
	#else
	/*
		const char *const green = "\033[0;40;32m";

		\033[30m set foreground color to black
		\033[31m set foreground color to red
		\033[32m set foreground color to green
		\033[33m set foreground color to yellow
		\033[34m set foreground color to blue
		\033[35m set foreground color to magenta (purple)
		\033[36m set foreground color to cyan
		\033[37m set foreground color to white

		\033[40m set background color to black
		\033[41m set background color to red
		\033[42m set background color to green
		\033[43m set background color to yellow
		\033[44m set background color to blue
		\033[45m set background color to magenta (purple)
		\033[46m set background color to cyan
		\033[47m set background color to white

		\033[1;30m set foreground color to dark gray
		\033[1;31m set foreground color to light red
		\033[1;32m set foreground color to light green
		\033[1;33m set foreground color to yellow
		\033[1;34m set foreground color to light blue
		\033[1;35m set foreground color to light magenta (purple)
		\033[1;36m set foreground color to light cyan
		\033[1;37m set foreground color to white

		\033[1;40m set background color to dark gray
		\033[1;41m set background color to light red
		\033[1;42m set background color to light green
		\033[1;43m set background color to yellow
		\033[1;44m set background color to light blue
		\033[1;45m set background color to light magenta (purple)
		\033[1;46m set background color to light cyan
		\033[1;47m set background color to white
	*/
		int icol = map_col[iForeground];
		if(icol<99)
			printf("\033[%dm", icol+30);
		else
			printf("\033{1;%dm", icol+30-100);
		icol = map_col[iBackground];
		if(icol<99)
			printf("\033[%dm", icol+40);
		else
			printf("\033{1;%dm", icol+40-100);
	#endif
#endif
}





// ------------------------------------------------------------- //
void PRESCALER(DGInt dx, DGInt dy)
{
#ifndef GLB_CONSOLE
	#ifdef TRENKI3D
	return;
	#endif

	if(dx>0 && dy>0)
	{
		__g_Prescaler.scale = 0.0f; // act with real device
		DGInt deskx, desky;
		USESCREEN(-1);
		GETSCREENSIZE(deskx, desky);
		float sc = MIN(deskx/dx, desky/dy);

		if(sc >= 10.0f) sc = floor(sc);
		else if(sc >= 8.0f) sc=8.0f;
		else if(sc >= 4.0f) sc=4.0f;
		else if(sc >= 3.0f) sc=3.0f;
		else if(sc >= 2.0f) sc=2.0f;
		else if(sc >= 1.5f) sc=1.5f;
		else sc = 1.0f; // I won't DOWNSCALE!

		__g_Prescaler.offx = (deskx - sc*dx) / 2.0f;
		__g_Prescaler.offy = (desky - sc*dy) / 2.0f;
		__g_Prescaler.dx = dx;
		__g_Prescaler.dy = dy;
		__g_Prescaler.scale=sc;
		__g_pRbow->m_WND.virtw = dx;
		__g_pRbow->m_WND.virth = dy;
	}
	else
	{
		__g_pRbow->m_WND.virth = 0;
		__g_pRbow->m_WND.virtw = 0;
		__g_Prescaler.scale=0.0;
	}
	VIEWPORT(0,0,0,0); // OK, apply this
#endif
}

// ------------------------------------------------------------- //
void	VIEWPORT(DGNat x, DGNat y, DGNat w, DGNat h)
{
#ifndef GLB_CONSOLE
	float scale=1.0f;
	if(__g_Prescaler.scale != 0.0f && __g_CurrentScreen==-1)
	{
		if(w<=0)
		{
			w = __g_Prescaler.dx;
			h = __g_Prescaler.dy;
		}

		x*=__g_Prescaler.scale;
		y*=__g_Prescaler.scale;
		x+=__g_Prescaler.offx;
		y+=__g_Prescaler.offy;

		w*=__g_Prescaler.scale;
		h*=__g_Prescaler.scale;
		scale = __g_Prescaler.scale;
	}

	#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
		__g_pRbow3D->DoShadows(false); // finish shadows if any
	#endif
	__g_pRbow->Make2DViewport((int)x, (int)y, (int)w, (int)h, scale);
#endif
}

// ------------------------------------------------------------- //
void	GETVIEWPORT(DGNat& x, DGNat& y, DGNat& w, DGNat& h)
{
#ifndef GLB_CONSOLE
	x = __g_pRbow->m_WND.viewport.x;
	y = __g_pRbow->m_WND.viewport.y;
	w = __g_pRbow->m_WND.viewport.width;
	h = __g_pRbow->m_WND.viewport.height;
	if(__g_Prescaler.scale != 0.0f && __g_CurrentScreen==-1)
	{
		w/=__g_Prescaler.scale;
		h/=__g_Prescaler.scale;
		x-=__g_Prescaler.offx;
		y-=__g_Prescaler.offy;
	}
#endif
}
void	GETVIEWPORT(DGInt& x, DGInt& y, DGInt& w, DGInt& h)
{
#ifndef GLB_CONSOLE
	x = (DGInt)__g_pRbow->m_WND.viewport.x;
	y = (DGInt)__g_pRbow->m_WND.viewport.y;
	w = (DGInt)__g_pRbow->m_WND.viewport.width;
	h = (DGInt)__g_pRbow->m_WND.viewport.height;
	if(__g_Prescaler.scale != 0.0f && __g_CurrentScreen==-1)
	{
		w/=__g_Prescaler.scale;
		h/=__g_Prescaler.scale;
		x-=__g_Prescaler.offx;
		y-=__g_Prescaler.offy;
	}
#endif
}



// ------------------------------------------------------------- //
void	CREATESCREEN(DGNat iScreen, DGNat iSprite, DGNat width, DGNat height)
{
#ifndef GLB_CONSOLE
	__g_pRbow->FBO_create((int)iScreen, (int)iSprite, (int)width, (int)height);
#endif
}

// ------------------------------------------------------------- //
void	USESCREEN(DGNat iScreen)
{
#ifndef GLB_CONSOLE
	if(iScreen<0)
	{
		__g_CurrentScreen=-1;
		__g_pRbow->FBO_end();
	}
	else
	{
		__g_CurrentScreen = iScreen;
		__g_pRbow->FBO_start((int)iScreen);
	}
	VIEWPORT(0,0,0,0); // this calls the prescaler // __g_pRbow->Make2DMode();
#endif
}


// ------------------------------------------------------------- //
void	STARTPOLY(DGNat nTexture, DGNat bMode)
{
#ifndef GLB_CONSOLE
	__g_pRbow->PolyStart((int)nTexture, bMode);
#endif
}

// ------------------------------------------------------------- //
void	POLYVECTOR(DGInt x, DGInt y, DGInt tx, DGInt ty, DGNat col)
{
#ifndef GLB_CONSOLE
	__g_pRbow->PolyPoint((float)x, (float)y, (float)tx, (float)ty, (COLORREF)col);
#endif
}

// ------------------------------------------------------------- //
void	POLYNEWSTRIP()
{
#ifndef GLB_CONSOLE
	__g_pRbow->PolyNewStrip();
#endif
}


// ------------------------------------------------------------- //
void	ENDPOLY()
{
#ifndef GLB_CONSOLE
	__g_pRbow->PolyEnd();
#endif
}


// Get an empty slot for objects
DGNat	GENSPRITE()
{
	#ifndef GLB_CONSOLE
		return __g_pRbow->GenSprite();
	#endif
	return -1;
}

DGNat	GENFONT()
{
	#ifndef GLB_CONSOLE
		return __g_pRbow->GenFont();
	#endif
	return -1;
}

DGNat	GENSCREEN()
{
	#ifndef GLB_CONSOLE
		return __g_pRbow->GenFBuffer();
	#endif
	return -1;
}



// ------------------------------------------------------------- //
void	LIMITFPS(DGInt n, DGNat bVsync)
{
#ifndef GLB_CONSOLE
	__g_pRbow->SetMaxFPS((int)n, bVsync?true:false);
#endif
}

// ------------------------------------------------------------- //
DGNat	ISFULLSCREEN()
{
#ifndef GLB_CONSOLE
	return __g_pRbow->m_WND.fullscreen ? TRUE:FALSE;
#else
	return 0;
#endif
}

// ------------------------------------------------------------- //
DGInt	GETTIMER()
{
// that might be much too short: __g_pRbow->AccurateTime() - __g_ThisTime
DGInt	dt = __g_ThisTime - __g_LastTime;
	return dt<=0 ? 1 : dt;
}

// ------------------------------------------------------------- //
DGInt	GETTIMERALL()
{
	return ::AccurateTime() - __g_FirstTime; // __g_ThisTime
}


#if defined (TARGET_OS_IPHONE)
}
	extern "C"
	{
		void iPhonePumpEvents(unsigned long ms);
	}
namespace __GLBASIC__ {
#endif


// ------------------------------------------------------------- //
void	SLEEP(DGNat ms)
{
#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
	DGInt end=GETTIMERALL() + ms;
	while(GETTIMERALL() < end) {};
#elif defined(TARGET_OS_IPHONE)
	iPhonePumpEvents(ms);
#else
	msleep((int)ms);
#endif
}


// ------------------------------------------------------------- //
//  SHELLEND - execute and exit
// ------------------------------------------------------------- //
// extern "C" int execl(const char*, const char*, ...);
void	SHELLEND(DGStr cmd)
{
	__ShutDownGLBasic();

#ifdef WIN32
	// win32-pc
	DGInt retval;
	SHELLCMD(cmd, FALSE, TRUE, retval);
	exit(0);
#else
	// *** SPACE FRACTAL FIXES 2017.3.1 ****
	#ifndef TVOS // SPACE FRACTAL, Not supported on TVOS.
		execl(cmd.c_str(), cmd.c_str(), NULL);
		
		g_debug("failed to call: %s\n", cmd.c_str());
		
		// ok, guys, that point should not be reached at all
		exit(0);
	#endif
#endif
}


// ------------------------------------------------------------- //
//  SHELLCMD - execute something new
// ------------------------------------------------------------- //
DGInt	SHELLCMD(DGStr cmd, DGInt wait, DGInt show, DGInt& retval)
{
#ifdef WIN32
	#ifdef UNICODE
		wchar_t pCmd[1024];
        // convert from multibyte char string to wide char string
        mbstowcs(pCmd, (const char *)cmd.c_str(), cmd.len()+1);
	#else
		char* pCmd = cmd.getbuffer(1024);
	#endif


	// win32-pc
	STARTUPINFO			si;
	PROCESS_INFORMATION	pi;
	DGNat furet = TRUE;
		retval=0;
		memset(&si, 0, sizeof(STARTUPINFO));
		si.cb=sizeof(STARTUPINFO);
		si.dwFlags=STARTF_USESHOWWINDOW;
		si.wShowWindow=show ? SW_SHOWNORMAL : SW_HIDE;

		#ifndef GLB_CONSOLE
		bool was_fs = 0;
		if(__g_pRbow && __g_pRbow->m_WND.fullscreen)
		{
			was_fs=true;
			__g_pRbow->DoWindowed();
		}
		#endif

		if (::CreateProcess(NULL, pCmd, NULL, NULL, FALSE, NORMAL_PRIORITY_CLASS, NULL, NULL, &si, &pi))
		{
			cmd.releasebuffer(); // just to be safe in multibyte mode
			::CloseHandle(pi.hThread);
			if (wait)
			{
				DWORD dwCode = STILL_ACTIVE;
				while(dwCode==STILL_ACTIVE)
				{
					::WaitForSingleObject(pi.hProcess, INFINITE);
					GetExitCodeProcess(pi.hProcess, &dwCode);
					// if (dwCode!=STILL_ACTIVE) Sleep(1000);
					retval = dwCode;
				}
			}
			::CloseHandle(pi.hProcess);
		}
		else
			furet = FALSE;

		if(show)
		{
			::SetActiveWindow( (HWND)GLBASIC_HWND() );
			::SetFocus( (HWND)GLBASIC_HWND() );
		}
		#ifndef GLB_CONSOLE
		if(was_fs && __g_pRbow)
			__g_pRbow->DoFullscreen();
		#endif

	return furet;
#else
	#ifndef GLB_CONSOLE
	bool was_fs = __g_pRbow->m_WND.fullscreen;
	__g_pRbow->DoWindowed();
	#endif

	int rv=1;
	if(!wait)
	{
		cmd+=CGStr(" &"); // "/bin/sh text.sh &" (do not wait)
	}

    #if !defined(TARGET_OS_IPHONE)
        retval = system(cmd.c_str());
    #endif

	#ifndef GLB_CONSOLE
	if(was_fs)
		__g_pRbow->DoFullscreen();
	#endif
	return rv;
#endif
}




}; // namepsace GLBASIC


// -------- WIN32 --------
#ifdef WIN32
	static DGInt	g_perf_cnt=0.0;				    // performance timer frequency
	static BOOL		g_perf_flag=FALSE;				// flag determining which timer to use
	//---------------------------------------------------------------//
	// Replacement to timeGetTime if available
	//---------------------------------------------------------------//
	DGInt AccurateTime()
	{
		#if defined(POCKETPC) || defined(_WIN32_WCE)
			static DWORD starttime = 0;
			if(starttime == 0)
				starttime = GetTickCount();
			return (DGInt)(GetTickCount()-starttime);
		#else
			static double starttime = 0.0;
			LONGLONG cur_time;       // current time
			// use the appropriate method to get time
			if (g_perf_flag)
			{
				QueryPerformanceCounter((LARGE_INTEGER *) &cur_time);
				if(starttime == 0)
				{
					starttime = (double)(cur_time*1000)/g_perf_cnt;
				}
				return (double)(cur_time*1000)/g_perf_cnt - starttime;
			}
			if(starttime == 0)
				starttime = GetTickCount();
			return (double)GetTickCount() - starttime;
		#endif
	}
// -------- WIZ --------
#elif defined(GPHWIZ)
	extern "C" unsigned int wiz_timer_get_ticks_us();
	DGInt AccurateTime()
	{
		static unsigned long starttime= 0;
		if(starttime == 0) starttime = wiz_timer_get_ticks_us();
		return 	(DGInt)( (wiz_timer_get_ticks_us()-starttime)/1000.0f);
	}
// -------- SDL --------
#elif defined(WANT_SDL)
	DGInt AccurateTime()
	{
		static unsigned long starttime= 0;
		if(starttime == 0) starttime = SDL_GetTicks();
		return (DGInt)(SDL_GetTicks()-starttime);
	}
// -------- GP2X --------
#else
	DGInt AccurateTime()
	{
		static long starttime = 0;
		timeval now;
		gettimeofday(&now, NULL);
 		if(starttime==0) starttime = now.tv_sec;
		
		return DGInt( (now.tv_sec-starttime)*1000.0f + now.tv_usec * 0.001f ); // usec = microsecs
	}
#endif

//---------------------------------------------------------------//
// Setup HP Timer
//---------------------------------------------------------------//
void SetupAccurateTimer()
{
// -------- WIN32 --------
#if defined(WIN32) && !defined(POCKETPC) && !defined(_WIN32_WCE)
	g_perf_flag=FALSE;
	LONGLONG	cnt;
	// is there a performance counter available?
	if (QueryPerformanceFrequency((LARGE_INTEGER *) &cnt))
	{
		if (cnt>0)
		{
			g_perf_cnt = (double)cnt;
			g_perf_flag=TRUE;
		}
	}
// -------- WIZ/CAANOO --------
#elif defined(GPHWIZ) || defined(CAANOO)

// -------- SDL --------
#elif defined(WANT_SDL)
	// SDL_Init(SDL_INIT_TIMER);  - global call now

// -------- Linux -------
#elif defined(POSIX)


#endif
}


void __GQuitNProfile(); // write profile file on demand (gpc_temp0.cpp)

// close all file handles
extern "C" void __GLB_close_all_filehandles()
{
	for(int channel=0; channel<MAX_IO_FILES; ++channel)
	{
		__g_Files[channel].Close();
	}
	__GQuitNProfile();
}


// ENTRY POINTS
namespace __GLBASIC__
{
extern int __MainGameSub_();
void __AllocConsole();
}


#if (defined( WIN32 ) || defined(_WIN32_WCE) )&& !defined( GLB_CONSOLE )
	#if defined(__cplusplus) // || defined(MACOSX) || defined(TARGET_OS_IPHONE)
	extern "C"
	#endif
	int APIENTRY WinMain(HINSTANCE hInst, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
	{
		#ifdef _WIN32_WCE
			#ifdef UNICODE
				__glb_cmdline().assign((const wchar_t*)lpCmdLine);
			#else
				__glb_cmdline().assign(lpCmdLine);
			#endif
			__glb_cmdline()=CGStr("glbasic.exe ") + __glb_cmdline();
		#else
			#ifdef UNICODE
				__glb_cmdline().assign((const wchar_t*)::GetCommandLine());
			#else
				__glb_cmdline().assign(::GetCommandLine());
			#endif
		#endif

		
		__GLB_Defaults();
		__GlobalInit(hInst, nCmdShow, __glb_AppName().c_str());

		__g_debugline=__g_error=__g_errorline=0;
		int rv = __MainGameSub_();

		DGStr glb_on_loop = "GLB_ON_LOOP";
		if(GETSUBADDRESS(glb_on_loop) != NULL)
		{
			SETLOOPSUB(glb_on_loop);
		}
		__EndProgram();

	return rv;
	}
	
	int APIENTRY DllMain(HINSTANCE hInst, DWORD reason, LPVOID /*reserved*/)
	{
		__GLB_Defaults();
		__GlobalInit(hInst, SW_SHOWNORMAL, __glb_AppName().c_str());

		__g_debugline=__g_error=__g_errorline=0;
		return TRUE;
	}

#else
	#if defined(TARGET_OS_IPHONE)
	#define main iPhoneMain // make a "main" callback from this main, and call from the real main()
	#endif

	#if defined(__cplusplus) || defined(MACOSX) || defined(TARGET_OS_IPHONE) || defined(ANDROID)
	extern "C"
	#endif
	int main(int argc, char* argv[])
	{
		for (long nc=0; nc<argc; nc++)
		{
			if(nc==0 || strchr(argv[nc], ' ')!=NULL) // has space in argument, or is first argument
			{
				__glb_cmdline()+=CGStr("\"") + DGStr(argv[nc]) + CGStr("\"");
			}
			else
			{
				__glb_cmdline()+=argv[nc];
			}
			if(nc<argc-1)
				__glb_cmdline()+=" ";
		}

		#ifdef PALM_WEBOS
			PDL_Init(0);
		#endif

		__GLB_Defaults();
	
		__GlobalInit(0L, 0, __glb_AppName().c_str());
		__g_debugline=__g_error=__g_errorline=0;
		int rv = __MainGameSub_();

		DGStr glb_on_loop = "GLB_ON_LOOP";
		if(GETSUBADDRESS(glb_on_loop) != NULL)
		{
			SETLOOPSUB(glb_on_loop);
		}

		// emscripten has set the main loop sub and must return here - don't tear engine down!
		#if defined(EMSCRIPTEN)
			return 0;
		#else
			__EndProgram();
		#endif

	
		return rv;
	}
#endif

