// --------------------------------------------------------------
// OpenGLRainbows.cpp: Implementierung der Klasse OpenGLRainbows.
// --------------------------------------------------------------

#include "stdafx.h"
#ifndef GLB_CONSOLE

#ifdef TINYGL
	#include <GL/zgl.h>
#endif


#if defined(GPHWIZ)
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#endif


#include "openglrainbows.h"
#include "rainbows3d.h"
#include "gfqueue.h"
#include "3dmath.h"
#include "vectormath.h"

#define GLB_MIN_TEX_SIZE 16 // 16 required for iPhone - can't get alpha for 1x1 images, else

#if defined(TINYGL) || defined(TRENKI3D)
#include "16bitcolor.h"
#endif

#ifndef POCKETPC
#define HAS_PNG_LIB
#endif


#if defined(WIN32) && (defined(HAVE_OPENGL) || (defined(HAVE_OPENGL_ES) && defined(_DEBUG)))
	#pragma comment(lib, "OpenGL32.lib")
	#pragma warning(disable: 4100) // unref. formal param
#endif

#include <stdlib.h> // rand

#if defined(HAVE_OPENGL) || (defined(HAVE_OPENGL_ES) && defined(_DEBUG))
// Don't use the glext.h name! Make lowercase "Arb" e.g.
// otherwise linker errors on OS-X...



// OpenGL extension functions
PFNGLACTIVETEXTUREARBPROC	glActiveTextureArb		=NULL;
PFNGLMULTITEXCOORD2FARBPROC	glMultiTexCoord2fArb	=NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC glClientActiveTextureArb				=NULL;

typedef BOOL (APIENTRYP PFWGL_SWAP_CTRL) (int onoff);
PFWGL_SWAP_CTRL gf_wglSwapIntervalEXT=NULL;



// VBO Extension Function Pointers
PFNGLGENBUFFERSARBPROC glGenBuffersArb = NULL;					// VBO Name Generation Procedure
PFNGLBINDBUFFERARBPROC glBindBufferArb = NULL;					// VBO Bind Procedure
PFNGLBUFFERDATAARBPROC glBufferDataArb = NULL;					// VBO Data Loading Procedure
PFNGLDELETEBUFFERSARBPROC glDeleteBuffersArb = NULL;			// VBO Deletion Procedure
PFNGLMAPBUFFERARBPROC	glMapBufferArb = NULL;
PFNGLUNMAPBUFFERARBPROC	glUnmapBufferArb = NULL;
PFNGLGETBUFFERPARAMETERIVARBPROC	glGetBufferParameterivArb = NULL;

// GLSL
PFNGLCREATEPROGRAMOBJECTARBPROC     glCreateProgramObjectArb = NULL;
PFNGLDELETEOBJECTARBPROC            glDeleteObjectArb = NULL;
PFNGLCREATESHADEROBJECTARBPROC      glCreateShaderObjectArb = NULL;
PFNGLSHADERSOURCEARBPROC            glShaderSourceArb = NULL;
PFNGLCOMPILESHADERARBPROC           glCompileShaderArb = NULL;
PFNGLGETOBJECTPARAMETERIVARBPROC    glGetObjectParameterivArb = NULL;
PFNGLATTACHOBJECTARBPROC            glAttachObjectArb = NULL;
PFNGLGETINFOLOGARBPROC              glGetInfoLogArb = NULL;
PFNGLLINKPROGRAMARBPROC             glLinkProgramArb = NULL;
PFNGLUSEPROGRAMOBJECTARBPROC        glUseProgramObjectArb = NULL;
PFNGLGETUNIFORMLOCATIONARBPROC      glGetUniformLocationArb = NULL;
PFNGLUNIFORM1FARBPROC               glUniform1fArb = NULL;
PFNGLUNIFORM1IARBPROC               glUniform1iArb = NULL;
PFNGLGETATTRIBLOCATIONARBPROC       glGetAttribLocationArb = NULL;
PFNGLVERTEXATTRIB1FARBPROC          glVertexAttrib1fArb = NULL;
#endif

#if defined(HAVE_OPENGL_ES) && defined(GPHWIZ)
	#if !defined( WANT_SDL) && !defined(TARGET_OS_IPHONE)
		extern "C"
		{
			void OS_InitFakeOS( void) ;
			NativeWindowType OS_CreateWindow();
		}
	#endif
#endif


// SEARCH_GLB_ENABLE
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
	GLB_ENABLE_STATE __g_EnableState;
#endif


extern DGInt AccurateTime(); // in glb-full.cpp



// fails on Raspberry GF_TODO #NEED
#ifndef TRENKI3D
void my_glOrtho(float left, float right, float bottom, float top, float nearZ, float farZ)
{
	//glOrtho(left, right, bottom, top, nearZ, farZ);
	//return;
float m[16];
#define M(row,col)  m[col*4+row]
   M(0,0) = 2.0F / (right-left);
   M(0,1) = 0.0F;
   M(0,2) = 0.0F;
   M(0,3) = -(right+left) / (right-left);

   M(1,0) = 0.0F;
   M(1,1) = 2.0F / (top-bottom);
   M(1,2) = 0.0F;
   M(1,3) = -(top+bottom) / (top-bottom);

   M(2,0) = 0.0F;
   M(2,1) = 0.0F;
   M(2,2) = -2.0F / (farZ-nearZ);
   M(2,3) = -(farZ+nearZ) / (farZ-nearZ);

   M(3,0) = 0.0F;
   M(3,1) = 0.0F;
   M(3,2) = 0.0F;
   M(3,3) = 1.0F;
#undef M
   //glLoadMatrixf(m); NO! result is M*Original!
   glMultMatrixf(m);
}
#endif

#if defined(HAVE_OPENGL_ES) && !defined(_DEBUG)
	//	#define glOrtho glOrthof // openGL|ES
	#define glOrtho my_glOrtho // openGL|ES
#endif




//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////


//---------------------------------------------------------------//
//---------------------------------------------------------------//
OpenGLRainbows::OpenGLRainbows()
{
	m_physical_device_screen_is_landscape = 0;
	m_landscape=0;
	m_CookieColor = 0x008000ff; // RGB(255,0,128);
	m_CookieFont  = 0x00000000;
	m_TextColor = 0xffffff;
	m_Texture = new GLR_TEXTURE[GLR_MAX_TEXTURES+1]; // +1 for FBO emergency mode

#if ! (defined( WANT_SDL ) || defined(TINYGL) || defined(TRENKI3D) ) && defined(WIN32)
	m_hRC=NULL;
#endif

#if !defined(TINYGL) && !defined(TRENKI3D)
	// m_IsActive=false;
	// No Crash when releasing a texture
	// memset(m_Texture, 0, sizeof(GLR_TEXTURE) * GLR_MAX_TEXTURES);
	#ifdef SPRITE_ALPHA_PIX_AS_VECTOR
	#else
		memset(SpritePix, 0, sizeof(BYTE*) * (GLR_MAX_TEXTURES+1));
	#endif

	// BYTE*			SpritePix[GLR_MAX_TEXTURES+1]; // Pixels for Sprcoll
	// memset(this->m_Surface, 0, sizeof(OPENGL_RAINBOWS_SURFACE)*GLR_MAX_SURFACE);
	m_SmoothShading = true;

	#ifdef GLB_SURFACE_OLDSCHOOL
	#else // jus use a sprite
		for(int i=0; i<GLR_MAX_SURFACE; ++i) m_BackSurfaceSpriteID[i] = -1;
	#endif

#endif

	memset(&m_WND, 0, sizeof(m_WND) );

#ifdef WIN32
	m_WND.hwnd=NULL;
#endif




	// this->FreeAll();
	m_falpha = 0.0f;
	m_CurrentFBO=-1;
	m_bHasClearColor = true; // black is default
	m_bDirtyZBuffer = true;
	m_bSupportBGRA_ext = false;
	m_pRbow3D=NULL;


	// delete the tmp file from last crash?!
	GFFileSystem::Delete(	TmpFileForTextureStore() );

}

//---------------------------------------------------------------//
//---------------------------------------------------------------//
OpenGLRainbows::~OpenGLRainbows()
{
	// delete the tmp file for the next run!
	GFFileSystem::Delete(	TmpFileForTextureStore() );

	// (SDL is already closed!)

	FBO_end(); for(int i=0; i<GLR_MAX_FRAMEBUFFER; ++i){m_FBuffer[i].Clear();}

	Close();

	#if !(defined( WANT_SDL ) || defined(TINYGL) || defined(TRENKI3D) ) && defined(WIN32) && defined(HAVE_OPENGL)
	if(m_hRC)
		wglDeleteContext(m_hRC);
	#endif

	if(m_Texture)
		delete[]m_Texture;
	m_Texture=NULL;
}

#if (defined(PANDORA) || defined(RASPBERRY_PI) )   && defined(HAVE_OPENGL_ES)
extern "C" 
{
	void NativeEGLInitParams(int set_width, int set_height);
	void NativeEGLInfo(int* pwidth, int* pheight);

	int  NativeEGLInitOpenGL();
	void NativeEGLTerminateOpenGL();
	int  NativeEGLSwapBuffers();
	void glb_pi_test();

}
#endif



//---------------------------------------------------------------//
//---------------------------------------------------------------//
void OpenGLRainbows::Close()
{
	FreeAll();

#if (defined(PANDORA) || defined(RASPBERRY_PI) ) && defined(HAVE_OPENGL_ES)
	NativeEGLTerminateOpenGL();
#elif defined(TARGET_OS_IPHONE)
	// move along

#elif defined( WANT_SDL)
    #if SDL_VERSION_ATLEAST(2,0,0) 
		#if defined(HAVE_OPENGL_ES)
			SDL_GL_DeleteContext(m_WND.sdl_glcontext);
		#endif
		SDL_DestroyWindow(m_WND.sdl_window);
	#endif
#else

	#if defined(HAVE_OPENGL_ES) && !defined(_DEBUG)
		if(m_WND.glesDisplay)
		{
			eglMakeCurrent(m_WND.glesDisplay, 0, 0, 0);  
			if(m_WND.glesContext) eglDestroyContext(m_WND.glesDisplay, m_WND.glesContext);
			if(m_WND.glesSurface) eglDestroySurface(m_WND.glesDisplay, m_WND.glesSurface);
			eglTerminate(m_WND.glesDisplay);
			m_WND.glesContext=NULL;
			m_WND.glesSurface=NULL;
			m_WND.glesDisplay=NULL;
		}  
	#endif

	#ifdef WIN32
		this->DoWindowed();
		#if defined(HAVE_OPENGL)
			wglMakeCurrent(::GetDC(m_WND.hwnd), NULL);
		#endif
		::InvalidateRect(0, NULL, TRUE);
		::DestroyWindow(m_WND.hwnd);
		m_WND.hwnd=NULL;
		m_WND.width=0;
		m_WND.height=0;
		m_WND.hwnd = NULL;
	#endif
#endif

}


#ifdef WANT_SDL
extern "C" void glb_do_pause(int pause);

//---------------------------------------------------------------//
//---------------------------------------------------------------//
int HandleEvent(SDL_Event *event)
{
	int done;

	done = 0;
	switch( event->type ) {
		#ifndef ANDROID
	    case SDL_ACTIVEEVENT: // deprecated in 2.0 use SDL_WINDOWEVENT
			// printf("SDL_ACTIVEVENT\n");
			glb_do_pause(event->active.gain ? 0 : 1);

			// See what happened
			/*
					printf( "app %s ", event->active.gain ? "gained" : "lost" );
					if ( event->active.state & SDL_APPACTIVE ) {
						printf( "active " );
					} else if ( event->active.state & SDL_APPMOUSEFOCUS ) {
						printf( "mouse " );
					} else if ( event->active.state & SDL_APPINPUTFOCUS ) {
						printf( "input " );
					}
					printf( "focus\n" );
					break;
			*/
			break;
		#endif

/*
	    case SDL_KEYDOWN: // - showscreen does that! Takes care of allowesc
			if ( event->key.keysym.sym == SDLK_ESCAPE)
			{
				done = 1;
			}
		break;
*/
	    case SDL_QUIT:
			done = 1;
		break;
	}
	return(done);
}
#endif


#if defined (TARGET_OS_IPHONE)
	extern "C"
	{
		void iPhonePumpEvents(unsigned long ms);
		void iPhoneSwapBuffers();
		char iPhoneGetInkeyStr();
	}
#endif


//---------------------------------------------------------------//
// return: 0: Exit app, 1:run
//---------------------------------------------------------------//
#ifdef TARGET_OS_IPHONE
extern "C" int gKillProgramNow;
#endif

#if !defined(POCKETPC)
	namespace __GLBASIC__ {
	extern void DXinp_push_inkey(char ic);
	}
#endif

unsigned int OpenGLRainbows::WinMessages( void(*peekmsg)() , bool allow_to_sleep)
{
	#ifdef WANT_SDL
		// Check if there's a pending event.
		SDL_Event event;
		int icount=0;
		while(SDL_PollEvent( &event ) && ++icount<256)
		{
			switch(event.type)
			{
			case SDL_KEYDOWN:
				switch(event.key.keysym.sym)
				{
				case SDLK_BACKSPACE: //SDLK_DELETE should be SDLK_BACKSPACE. But on the Mac it's twisted :/
					DXinp_push_inkey('\b'); break;
					break;

				// non-inkey$ keys:
				case SDLK_UP:
				case SDLK_DOWN:
				case SDLK_RIGHT:
				case SDLK_LEFT:
				case SDLK_INSERT:
				case SDLK_HOME:
				case SDLK_END:
				case SDLK_PAGEUP:
				case SDLK_PAGEDOWN:
				case 14:  // backspace 2nd key?
				case 211: // del
					break;
				default:
					if((event.key.keysym.unicode & 0xff00) == 0)
						DXinp_push_inkey(event.key.keysym.unicode & 0x00ff);
				}
				break;
			case SDL_QUIT:
				g_debug("SDL_QUIT event\n");
				return 0; // Exit App
			}

			/*
			if(HandleEvent(&event))
			{
				return 0; // Exit App
			}
			*/
		}

		if(allow_to_sleep)
		{
			if(SDL_WaitEvent(&event))
			{
				SDL_PushEvent(&event);
				return WinMessages(peekmsg, false); // handle the event
			}
		}


	#elif defined (TARGET_OS_IPHONE)
		if(gKillProgramNow)
			return 0;
		char ke = iPhoneGetInkeyStr();
		if (ke)
			DXinp_push_inkey(ke);
		iPhonePumpEvents(0);
	#elif defined(WIN32) && !defined(POCKETPC) // WinCE has it's own message loop for WM_CHAR!

			if(allow_to_sleep)
				::WaitMessage();

			// Peek & Translate Win Message
			MSG		m;
			if(peekmsg != NULL)
			{
				while(::PeekMessage(&m, m_WND.hwnd, 0, 0, PM_NOREMOVE))
				{
					peekmsg();
					::PeekMessage(&m, m_WND.hwnd, 0, 0, PM_REMOVE);
					if(m.message == WM_CHAR) DXinp_push_inkey(m.wParam & 0xff);
					if(m.message == WM_KEYDOWN) TranslateMessage(&m); // sends WM_CHAR
					DispatchMessage(&m);
				}
			}
			else
			{
				while(::PeekMessage(&m, m_WND.hwnd, 0, 0, PM_REMOVE))
				{
					// printf("Msg: %d %x\n", m.message, m.message);
					if(m.message == WM_CHAR) DXinp_push_inkey(m.wParam&0xff);
					if(m.message == WM_KEYDOWN) TranslateMessage(&m); // sends WM_CHAR
					DispatchMessage(&m);
				}
			}
	#endif
return 1;
}


#if defined(POSIX)
void Sleep(unsigned long msec); // glb_pc.cpp
#endif



//---------------------------------------------------------------//
// get temp filename
//---------------------------------------------------------------//
const char* OpenGLRainbows::TmpFileForTextureStore()
{
	static int app_id = int(((size_t)this + rand()) & 0xffffffff);
	#ifdef ANDROID
	// /tmp is app-specific.
		app_id = 0;
	#endif
	static char tmp[1024];
	char tmp2[1024]="/tmp";
	GFFileSystem::GetTempFolder(1024, tmp2);
	sprintf(tmp, "%sglbtxs%x.bin", tmp2, (int)app_id);
	return tmp;
}


//---------------------------------------------------------------//
// save texture when app looses focus. Free memory, too
//---------------------------------------------------------------//
bool OpenGLRainbows::LoadSaveTextureToTmp(bool bLoad)
{
	GFFileSystem fs;

	if(bLoad)
	{
		g_debug("restoring textures from temp file\n");
		if(!fs.OpenIn(TmpFileForTextureStore())) return false;
	}
	else
	{
		g_debug("move textures from mem to temp file\n   \"%s\"\n", TmpFileForTextureStore());
		if(!fs.OpenOut(TmpFileForTextureStore())) return false;
	}

	OPENGL_IMAGE img;
	for(int i=0; i<GLR_MAX_TEXTURES; ++i)
	{
		if(bLoad)
		{
			if(fs.Eof()) break;
			int j=0, dx=0, dy=0;
			fs.Read(&j, sizeof(int));
			fs.Read(&dx, sizeof(int));
			fs.Read(&dy, sizeof(int));
			img.resize(dx, dy);
			fs.Read(img.m_txdata, dx*dy*4);
			this->LoadSprite(j, img, this->m_CookieColor, false);
		}
		else
		{
			if(this->GetSpriteSize(i).cx > 0)
			{
				fs.Write(&i, sizeof(int));
				this->GetSpriteRGBA(i, img);
				int dx = img.m_sizex, dy=img.m_sizey;
				fs.Write(&dx, sizeof(int));
				fs.Write(&dy, sizeof(int));
				fs.Write(img.m_txdata, img.m_sizex*img.m_sizey*4);
				// this->FreeSprite(i); // no. I'm bad!
			}
		}
	}
	fs.Close();

	if(bLoad)
	{
		fs.Delete(TmpFileForTextureStore());
	}
	g_debug("   ...done\n");
	return true;
}


//---------------------------------------------------------------//
// Prepare the back and Z buffers for the next frame
//---------------------------------------------------------------//
void OpenGLRainbows::PrepareBackbuffer()
{
#if defined(TINYGL) || defined(TRENKI3D)
	if(m_Surface[0].m_Height)
	{
		m_Surface[0].MemCpyToScreen();
	}
	else
	{
		if(m_bHasClearColor)
		{
			Screen.Clear();
		}
	}
	#ifdef TRENKI3D
		if(m_bDirtyZBuffer)
		{
			m_bDirtyZBuffer=false;
			Screen.ClearZBuffer();
		}
	#endif
	
#else // GL or EGL
	#ifdef PALM_WEBOS
	// glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_FALSE);
	#endif


#ifdef GLB_SURFACE_OLDSCHOOL
	// background image or clear color?
	if(m_Surface[0].usex>0)
	{
		if(m_bDirtyZBuffer)
		{
			m_bDirtyZBuffer=false; glClear(GL_DEPTH_BUFFER_BIT);
		}
		PasteSurface(0);
	}
	else if(m_bHasClearColor || m_bDirtyZBuffer)
	{
		GLbitfield bits=0;
		if(m_bHasClearColor)  bits |= GL_COLOR_BUFFER_BIT;
		if(m_bDirtyZBuffer)  {bits |= GL_DEPTH_BUFFER_BIT; 	m_bDirtyZBuffer=false;}
		glClear(bits);
	}
#else // just use a sprite
	if(m_BackSurfaceSpriteID[0] != -1)
	{
		PasteSurface(0);
	}
	else if(m_bHasClearColor || m_bDirtyZBuffer)
	{
		GLbitfield bits=0;
		if(m_bHasClearColor)  bits |= GL_COLOR_BUFFER_BIT;
		if(m_bDirtyZBuffer)  {bits |= GL_DEPTH_BUFFER_BIT; 	m_bDirtyZBuffer=false;}
		glClear(bits);
	}
#endif


#endif
}


//---------------------------------------------------------------//
//---------------------------------------------------------------//
int OpenGLRainbows::FlipScreens(void (*peekmsg)(void) )
{
	// Make sure FBO access is finished
 	FBO_end();

#if defined(TINYGL) || defined(TRENKI3D)
	Screen.Show();
#else // GL or EGL
	//	MakeAutoStereo();
	#ifdef WANT_SDL
		#if (defined(PANDORA) || defined(RASPBERRY_PI) ) && defined(HAVE_OPENGL_ES)
			NativeEGLSwapBuffers();

		#elif SDL_VERSION_ATLEAST(2,0,0)
			// SDL20
			// http://wiki.libsdl.org/moin.cgi/SDL_GL_CreateContext?highlight=%28%5CbCategoryVideo%5Cb%29%7C%28CategoryEnum%29%7C%28CategoryStruct%29%7C%28SGFunctions%29
			//#elif defined(ANDROID) SDL2.0  // #elif defined( HAVE_OPENGL_ES)
			SDL_GL_SwapWindow(m_WND.sdl_window);
		#else
			// Linux/Mac
			#if defined( HAVE_OPENGL) ||  defined( HAVE_OPENGL_ES)
				SDL_GL_SwapBuffers( );
			#endif
		#endif
	#else
		#if defined(TARGET_OS_IPHONE)
			iPhoneSwapBuffers();
		#elif defined(HAVE_OPENGL_ES) && !defined(_DEBUG)
			eglSwapBuffers(m_WND.glesDisplay, m_WND.glesSurface);
		#else
			static HDC hDC = ::GetDC(m_WND.hwnd);
			SwapBuffers(hDC);
			// ::ReleaseDC(m_WND.hwnd, hDC);
		#endif
	#endif

#endif

	PrepareBackbuffer();


	DGInt	cur_time;
	DGInt slp=0;
	// is it time to render the frame?
	do
	{
		if (!this->WinMessages(peekmsg, false))
		{
			return 0; // Exit was pressed
		}
		cur_time=AccurateTime();

		slp = m_next_time - cur_time;
		if(slp>7)
		{
#if defined(TARGET_OS_IPHONE)
			iPhonePumpEvents((int)slp-3); // give more response to the GUI if we can
#else
			slp -= 3;
			::Sleep( (DWORD)slp );
#endif				
			cur_time+=slp;
		}
	}while(slp > 1);


	// If we get more than a frame ahead, allow us to drop one
	// Otherwise, we will never catch up if we let the error
	// accumulate, and message handling will suffer
	if(m_FrameRate>0)
	{
#if 1
		// set time for next frame
		m_next_time += m_FrameRate;
		if(m_next_time < cur_time)
		{
			m_next_time += m_FrameRate;
			// too big skip required
			if(m_next_time < cur_time)
				m_next_time = cur_time + m_FrameRate;
		}
#else
	// set time for next frame
	m_next_time += m_FrameRate;
	// If we get more than a frame ahead, allow us to drop one
	// Otherwise, we will never catch up if we let the error
	// accumulate, and message handling will suffer
	if(m_FrameRate>0)
	{
		while(m_next_time < cur_time)
			m_next_time += m_FrameRate;
	}
#endif
	}
return 1;
}

#if defined(HAVE_OPENGL) && defined(WIN32)
	// wglext.h
	#define WGL_STENCIL_BITS_ARB           0x2023
	#define WGL_DEPTH_BITS_ARB             0x2022
	#define WGL_SUPPORT_OPENGL_ARB         0x2010
	#define WGL_ACCELERATION_ARB           0x2003
	#define WGL_FULL_ACCELERATION_ARB      0x2027
	#define WGL_COLOR_BITS_ARB             0x2014
	#define WGL_ALPHA_BITS_ARB             0x201B
	#define WGL_DEPTH_BITS_ARB             0x2022
	#define WGL_STENCIL_BITS_ARB           0x2023
	#define WGL_DRAW_TO_WINDOW_ARB         0x2001
	#define WGL_DOUBLE_BUFFER_ARB          0x2011
	#define WGL_STEREO_ARB                 0x2012
	#ifndef WGL_ARB_multisample
	#define WGL_SAMPLE_BUFFERS_ARB         0x2041
	#define WGL_SAMPLES_ARB                0x2042
	#endif




bool WGLisExtensionSupported(const char *extension)
{
	const size_t extlen = strlen(extension);
	const char *supported = NULL;

	// Try To Use wglGetExtensionStringARB On Current DC, If Possible
	PROC wglGetExtString = (PROC) glGFGetProcAddressArb("wglGetExtensionsStringARB");

	if (wglGetExtString)
		supported = ((char*(__stdcall*)(HDC))wglGetExtString)(wglGetCurrentDC());

	// If That Failed, Try Standard Opengl Extensions String
	if (supported == NULL)
		supported = (char*)glGetString(GL_EXTENSIONS);

	// If That Failed Too, Must Be No Extensions Supported
	if (supported == NULL)
		return false;

	// Begin Examination At Start Of String, Increment By 1 On False Match
	for (const char* p = supported; ; p++)
	{
		// Advance p Up To The Next Possible Match
		p = strstr(p, extension);
		if (p == NULL)
			return false;						// No Match
		if ((p==supported || p[-1]==' ') && (p[extlen]=='\0' || p[extlen]==' '))
			return true;						// Match
	}
}

// find a pixelformat
int ChooseMultisamplePixelFormat(HDC hDC,PIXELFORMATDESCRIPTOR pfd, int multisamples)
{  
	// See If The String Exists In WGL!
	if (!WGLisExtensionSupported("WGL_ARB_multisample"))
	{
		return 0;
	}

	#ifndef PFNWGLCHOOSEPIXELFORMATARBPROC
	typedef BOOL (WINAPI * PFNWGLCHOOSEPIXELFORMATARBPROC) (HDC hdc, const int *piAttribIList, const FLOAT *pfAttribFList, UINT nMaxFormats, int* piFormats, UINT* nNumFormats);
	#endif
	// Get Our Pixel Format
	PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatArb =
		(PFNWGLCHOOSEPIXELFORMATARBPROC)glGFGetProcAddressArb("wglChoosePixelFormatARB");

	if (!wglChoosePixelFormatArb)
	{
		// We Didn't Find Support For Multisampling, Set Our Flag And Exit Out.
		return 0;
	}

	int pixelFormat;
	BOOL valid;
	UINT numFormats;
	float fAttributes[] = {0,0};

	// These Attributes Are The Bits We Want To Test For In Our Sample
	// Everything Is Pretty Standard, The Only One We Want To 
	// Really Focus On Is The SAMPLE BUFFERS ARB And WGL SAMPLES
	// These Two Are Going To Do The Main Testing For Whether Or Not
	// We Support Multisampling On This Hardware
	int iAttributes[] = { WGL_DRAW_TO_WINDOW_ARB,GL_TRUE,
		WGL_SUPPORT_OPENGL_ARB,GL_TRUE,
		WGL_ACCELERATION_ARB,WGL_FULL_ACCELERATION_ARB,
		WGL_COLOR_BITS_ARB,24,
		WGL_ALPHA_BITS_ARB,8,
		WGL_DEPTH_BITS_ARB,24,
		WGL_STENCIL_BITS_ARB,8,
		WGL_DOUBLE_BUFFER_ARB,GL_TRUE,
		WGL_SAMPLE_BUFFERS_ARB,GL_TRUE,
		WGL_SAMPLES_ARB, multisamples,						// Check For 4x Multisampling
		0,0};

	// First We Check To See If We Can Get A Pixel Format For 4 Samples
	valid = wglChoosePixelFormatArb(hDC,iAttributes,fAttributes,1,&pixelFormat,&numFormats);
 
	// if We Returned True, And Our Format Count Is Greater Than 1
	if (valid && numFormats >= 1)
		return pixelFormat;

	// Our Pixel Format With 4 Samples Failed, Test For 2 Samples
	while(iAttributes[19]/2)
	{
		iAttributes[19]>>=1;
		valid = wglChoosePixelFormatArb(hDC,iAttributes,fAttributes,1,&pixelFormat,&numFormats);
		if (valid && numFormats >= 1)
			return pixelFormat;
	}
return  0;
}
#endif


//---------------------------------------------------------------//
//---------------------------------------------------------------//
#if !defined( WANT_SDL) && defined(WIN32)
namespace __GLBASIC__
{
	LRESULT CALLBACK __WndProc_(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
};
#endif


#ifdef WANT_SDL
	#if SDL_VERSION_ATLEAST(2,0,0) 
	static int GetVideoDisplay()
	{
		const char *variable = SDL_getenv("SDL_VIDEO_FULLSCREEN_DISPLAY");
		if ( !variable ) {
			variable = SDL_getenv("SDL_VIDEO_FULLSCREEN_HEAD");
		}
		if ( variable ) {
			return SDL_atoi(variable);
		} else {
			return 0;
		}
	}
	#endif
#endif






bool OpenGLRainbows::Init(HINSTANCE hInst, int width, int height, int bpp, unsigned int fps, int multisamples, bool fullscreen)
{
static int	been_here=0;
	m_clear_color = RGB(0,0,0);

	SetMaxFPS(fps, false);

#if !defined( WANT_SDL) && defined(WIN32) && !defined(POCKETPC)

	if(!been_here)
	{
		#define STR_GLBASIC	"GLBasic"
		WNDCLASS        wc;
		wc.style                = /*CS_HREDRAW | CS_VREDRAW |*/ CS_BYTEALIGNWINDOW|CS_BYTEALIGNCLIENT|CS_OWNDC;
		wc.lpfnWndProc          = (WNDPROC) __GLBASIC__::__WndProc_;
		wc.cbClsExtra           = 0;
		wc.cbWndExtra           = 0;
		wc.hInstance            = hInst;
		wc.hIcon                = ::LoadIcon(hInst, MAKEINTRESOURCE(101));
		wc.hCursor              = NULL;
		wc.hbrBackground        = NULL;
		wc.lpszMenuName         = NULL;
		wc.lpszClassName        = STR_GLBASIC; // Must be same as below!!!

		if(RegisterClass(&wc) == 0)
			return 0;	//  | WS_CLIPCHILDREN | WS_CLIPSIBLINGS
		been_here = 1;
	}
	m_hinstance = hInst;
#endif
	
	// store for later use
	m_WND.multisample = multisamples;


#ifndef POCKETPC
	#if defined(HAVE_OPENGL_ES) && defined(GPHWIZ) // OpenGL|ES w/o SDL - Wiz, maybe
		g_debug("OGL video GPHWIZ\n");
		__DG_DESKTOPX=320; __DG_DESKTOPY=240;
		m_physical_device_screen_is_landscape = 1;
		// Initialize native OS
		OS_InitFakeOS();
		// Create native window.
		m_WND.hNativeWnd = OS_CreateWindow();	                  
		if(!m_WND.hNativeWnd)
		{
			g_debug("OS_CreateWindow failed\n");
		}

		EGLConfig configs[10];
		EGLint matchingConfigs;	

		/*configAttribs is a integers list that holds the desired format of 
		our framebuffer. We will ask for a framebuffer with 24 bits of 
		color and 16 bits of z-buffer. We also ask for a window buffer, not 
		a pbuffer or pixmap buffer*/	
		const EGLint configAttribs[] =
		{
			EGL_SURFACE_TYPE,   EGL_WINDOW_BIT,
			EGL_NONE,           EGL_NONE
		};

		m_WND.glesDisplay = eglGetDisplay(EGL_DEFAULT_DISPLAY);	 //Ask for an available display

		//Display initialization (we don't care about the OGLES version numbers)
		if(!eglInitialize(m_WND.glesDisplay, 0, 0)) 
		return false;

		/*Ask for the framebuffer confiburation that best fits our 
		parameters. At most, we want 10 configurations*/
		if(!eglChooseConfig(m_WND.glesDisplay, configAttribs, &configs[0], 10,  &matchingConfigs)) 
		return false;

		//If there isn't any configuration enough good
		if (matchingConfigs < 1)  return false;	  

		/*eglCreateWindowSurface creates an onscreen EGLSurface and returns 
		a handle  to it. Any EGL rendering context created with a 
		compatible EGLConfig can be used to render into this surface.*/
		m_WND.glesSurface = eglCreateWindowSurface(m_WND.glesDisplay, configs[0], m_WND.hNativeWnd, configAttribs);	
		if(!m_WND.glesSurface) return false;

		// Let's create our rendering context
		m_WND.glesContext=eglCreateContext(m_WND.glesDisplay,configs[0],0,configAttribs);

		if(!m_WND.glesContext) return false;

		//Now we will activate the context for rendering	
		eglMakeCurrent(m_WND.glesDisplay, m_WND.glesSurface, m_WND.glesSurface, m_WND.glesContext); 

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		eglSwapBuffers(m_WND.glesDisplay, m_WND.glesSurface);

	// Padora-EGL by X11
	#elif defined(PANDORA) && defined(HAVE_OPENGL_ES)
		g_debug("sdl init video PANDORA\n");
		// if( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {  - global call now
		//	g_debug("Couldn't initialize SDL: %s\n", SDL_GetError());
		//	exit( 1 );
		// }
		NativeEGLInitOpenGL();

		const SDL_VideoInfo* info = SDL_GetVideoInfo(); 
		__DG_DESKTOPX = info->current_w; 
		__DG_DESKTOPY = info->current_h;
		g_debug ("SDL_GetVideoInfo = %dx%d @ %d bpp\n", __DG_DESKTOPX, __DG_DESKTOPY, info->vfmt->BitsPerPixel);
		m_physical_device_screen_is_landscape = __DG_DESKTOPX > __DG_DESKTOPY ? 1:0;
		/*
		// also works on OMAP-3 - don't hardcode it
		__DG_DESKTOPX=800; __DG_DESKTOPY=480;
		width=800;
		height=480;
		*/

		SetScreen(width, height, 16, TRUE);
		
//		height = m_WND.height;
//		width = m_WND.width;
		m_WND.height=height;
		m_WND.width=width;
		g_debug("%dx%d sdl\n",width,height);

	#elif defined(RASPBERRY_PI) && defined(HAVE_OPENGL_ES)
		g_debug("sdl init video RaspberryPi\n");


		SDL_Surface* myVideoSurface = SDL_SetVideoMode(0,0, 32,  SDL_SWSURFACE);
		NativeEGLInitParams(width, height);
		NativeEGLInitOpenGL();
		int ix=0, iy=0;
		NativeEGLInfo(&ix, &iy);
		__DG_DESKTOPX = ix; 
		__DG_DESKTOPY = iy;
		g_debug ("NativeEGLInfo = %dx%d\n", __DG_DESKTOPX, __DG_DESKTOPY);
		m_physical_device_screen_is_landscape = __DG_DESKTOPX > __DG_DESKTOPY ? 1:0;


		/*
		if ( SDL_SetVideoMode( ix, iy, 0, SDL_SWSURFACE|SDL_NOFRAME ) == NULL )
		{
			g_debug("SDL - SetVideoMode failed: %s\n", SDL_GetError());
		}
		*/


		OPENGL_CHECK_ERRORS();
		SetScreen(width, height, 32, TRUE);
		OPENGL_CHECK_ERRORS();

//		height = m_WND.height;
//		width = m_WND.width;
		m_WND.height=height;
		m_WND.width=width;
		g_debug("%dx%d sdl\n",width,height);


	// -------- SDL --------
	#elif defined(WANT_SDL)
		g_debug("SDL_init video\n");
		Uint32 video_flags=0;
		int rgb_size[3];
		int done = 0;
		// if( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {  - global call now
		//	g_debug("Couldn't initialize SDL_VIDEO: %s\n", SDL_GetError());
		//	exit( 1 );
		//}

		int desktop_bits = 0;
		#if SDL_VERSION_ATLEAST(2,0,0) 
			SDL_DisplayMode mode;
			SDL_GetDesktopDisplayMode(GetVideoDisplay(), &mode);
			__DG_DESKTOPX = mode.w;
			__DG_DESKTOPY = mode.h;
			SDL_PixelFormat* sdl_vfmt = SDL_AllocFormat(mode.format);
			if(sdl_vfmt)
			{
				desktop_bits = sdl_vfmt->BitsPerPixel;
				SDL_FreeFormat(sdl_vfmt);
			}
		#else
			const SDL_VideoInfo* info = SDL_GetVideoInfo(); 
			__DG_DESKTOPX = info->current_w; 
			__DG_DESKTOPY = info->current_h;
			desktop_bits = info->vfmt->BitsPerPixel;
		#endif

		#if defined(ANDROID)
			// these are totally silly if the main window is not moved or scaled. At least give full drawing area access
			width  = __DG_RESX = __DG_DESKTOPX;
			height = __DG_RESY = __DG_DESKTOPY;
		#endif
		m_physical_device_screen_is_landscape = __DG_DESKTOPX > __DG_DESKTOPY ? 1:0;
		g_debug ("SDL_GetVideoInfo = %dx%d @ %d bpp. Screen %dx%d\n", __DG_DESKTOPX, __DG_DESKTOPY, desktop_bits, __DG_RESX, __DG_RESY);


		// *** SDL 2.0 ***
		// -------------
		#if SDL_VERSION_ATLEAST(2,0,0) 
			g_debug("SDL 2.x mode\n");
			// Sets up OpenGL ES double buffering and visual parameters
			SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 1); // OpenGL|ES 1.1
			SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 1); // OpenGL|ES 1.1
			SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
			SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5); // minimum
			SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 6);
			SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
			SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
			SDL_GL_SetAttribute(SDL_GL_ACCELERATED_VISUAL, 1);

			g_debug("C++ calling SDL_CreateWindow (fs:%d)\n", fullscreen);

			m_WND.sdl_window=SDL_CreateWindow("GLBasic", 0, 0,
				width, height,
				fullscreen
				? (                       SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN | SDL_WINDOW_BORDERLESS)
				: (SDL_WINDOW_RESIZABLE | SDL_WINDOW_OPENGL | SDL_WINDOW_SHOWN)
				);
			if(m_WND.sdl_window == NULL)
			{
				g_debug("Can't create window: %s\n", SDL_GetError());
				exit(-1);
			}


			/*
			m_WND.sdl_renderer = SDL_CreateRenderer(m_WND.sdl_window, -1, SDL_RENDERER_ACCELERATED|SDL_RENDERER_TARGETTEXTURE);
			if (m_WND.sdl_renderer==NULL)
			{
				g_debug("Can't create renderer: %s\n", SDL_GetError());
				return false;
			}
			*/

			// SDL_CreateRenderer(m_WND.window, 0, 0);
			// alternaitive
			m_WND.sdl_glcontext=SDL_GL_CreateContext(m_WND.sdl_window);
			if (m_WND.sdl_glcontext==NULL)
			{
				g_debug("Can't create OpenGL ES context: %s\n", SDL_GetError());
				return false;
			}

			int status=SDL_GL_MakeCurrent(m_WND.sdl_window, m_WND.sdl_glcontext);
			if (status<0)
			{
				g_debug("Can't set current OpenGL ES context: %s\n", SDL_GetError());
				return false;
			}

			// Enable swap on VSYNC
			SDL_GL_SetSwapInterval(1);

			/*
			SDL_SetRenderDrawColor(m_WND.sdl_renderer, 0,0,0,255);
			SDL_RenderClear(m_WND.sdl_renderer);
			SDL_RenderPresent(m_WND.sdl_renderer);
			*/

			g_debug("C++ SDL_CreateWindow stuff finished\n");

		// *** SDL 1.2 ***
		// -------------
		#else // SDL 1.2
			g_debug("SDL 1.2 mode\n");

			// See if we should detect the display depth
			if ( bpp == 0 )
			{
		//		if ( SDL_GetVideoInfo()->vfmt->BitsPerPixel <= 8 )
		//			bpp = 8;
		//		else
					bpp = 24;  // More doesn't seem to work
				bpp = max(bpp, desktop_bits); // desktop might be 32!
			}
			#if defined(GPHWIZ) || defined(GP2X) || defined(PANDORA)
				bpp = 16;
			#endif

			/* Set the flags we want to use for setting the video mode */
			#if defined( MACOSX ) || defined(TARGET_OS_IPHONE) || defined(PANDORA) || defined(GP2X) || defined(GPHWIZ)
			if(fullscreen)
			{
				g_debug("Request fullscreen = %d\n", SDL_FULLSCREEN);
				video_flags|=SDL_FULLSCREEN; // Mac problem
			}
			#endif
		
			#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
				g_debug("Request SDL_OPENGL = %d\n", SDL_OPENGL);
				video_flags |= SDL_OPENGL;

				g_debug("Request double buffer\n");
				SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );;
			#endif



			#if defined( HAVE_OPENGL)
				g_debug("set up SDL for OpenGL mode\n");
				#if defined(EMSCRIPTEN)
					g_debug(" -WebGL mode: 32 bits, SDL_OPENGL=!0x04000000!\n");
					bpp = 32;
					video_flags = /* SDL_compat.h #define SDL_OPENGL        */  0x04000000 | SDL_OPENGL;
				#endif
			#elif defined(HAVE_OPENGL_ES)
				g_debug("set up SDL for OpenGL-ES mode\n");
				SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 1); // OpenGL|ES 1.1

				//** Set depth buffer size to 16 bits 
				SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 ); // minimum - can be 24 or 32
				//** Set stencil buffer size to 16 bits 
				// SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 16 );

				#ifdef PALM_WEBOS
					int want_landscape = width > height ? 1:0;

					g_debug("   want landscape %d,%d: %d\n   device landscape: %d\n", width, height, want_landscape, m_physical_device_screen_is_landscape);

					// 320x480  pre
					// 320x400  pixi
					// 1024x768 touchpad
					m_landscape = 0;
					if(m_physical_device_screen_is_landscape != want_landscape)
					{
						m_landscape=1; // rotate by 90 - palm pre, so "top" is right edge
					}

					width=0;
					height=0;

					bpp=0; // that's what's returned when you put sdl_setvidmode(0,0,0)
					video_flags = SDL_OPENGL; // nothing more
				#endif
				#ifdef ANDROID
					width= __DG_DESKTOPX;
					height=__DG_DESKTOPY;
					bpp = info->vfmt->BitsPerPixel; // SDL_RESIZABLE
					video_flags = SDL_OPENGL|SDL_FULLSCREEN; // SDL_SWSURFACE|SDL_DOUBLEBUF|;
				#endif

			#elif defined(EMSCRIPTEN) // Canvas - SDL + Trenki3D mode
				bpp = 32;
				video_flags = SDL_SWSURFACE;
				g_debug("Emscripten: 32 bits, SDL_SWSURFACE\n");

			#else // TRENKI3D, e.g.
				g_debug("set up SDL for framebuffer mode\n");
				video_flags |= SDL_DOUBLEBUF;
				bpp = 16;
			#endif

#ifdef _DEBUG // SDL
	bpp = 32;
	video_flags = SDL_SWSURFACE;
#endif


			/* Initialize the display */
			switch (bpp)
			{
			case 8:
				rgb_size[0] = 3;
				rgb_size[1] = 3;
				rgb_size[2] = 2;
				break;
			case 15:
				rgb_size[0] = 5;
				rgb_size[1] = 5;
				rgb_size[2] = 5;
				break;
			case 16:
				rgb_size[0] = 5;
				rgb_size[1] = 6;
				rgb_size[2] = 5;
				break;
			default:
				rgb_size[0] = 8;
				rgb_size[1] = 8;
				rgb_size[2] = 8;
				break;
			}

			#if defined(HAVE_OPENGL)
				g_debug("set up SDL for OpenGL mode RGBA=%d,%d,%d,%d; Depth=24, Stencil=8, DoubleBuffer=1\n", rgb_size[0], rgb_size[1], rgb_size[2], (bpp==32)?8:0);
				SDL_GL_SetAttribute( SDL_GL_RED_SIZE,   rgb_size[0]);
				SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, rgb_size[1]);
				SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE,  rgb_size[2]);
				if(bpp==32)
				{
					SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, 8);
				}
				SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24);
				SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 8);
				SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1);

				if(multisamples)
				{
					g_debug("multisamples x%d\n", multisamples);
					SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
					SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, multisamples);
				}
			#endif

			//	video_flags |= SDL_NOFRAME; // Make as fast as possible - ugly flickering
			g_debug("Setting video mode: %dx%dx%d vidflags=0x%x\n", width, height, bpp, video_flags);
			bool ok=true;
			if ( SDL_SetVideoMode( width, height, bpp, video_flags ) == NULL )
			{
				g_debug("SetVideoMode failed: %s\n", SDL_GetError());
			#if !defined(TINYGL) && !defined(TRENKI3D) && !defined(HAVE_OPENGL_ES)
				ok=false;
				if(multisamples)
				{
					g_debug("INIT: Unable to create video surface that supports FSAA; trying without.\n");
					SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 0);
					SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 0);
					if ( SDL_SetVideoMode( width, height, bpp, video_flags ) )
						ok=true;
				}
			#endif
			}
			if(!ok)
			{

				g_debug("INIT: Unable to create video surface that supports requested details; trying SDL_ANYFORMAT.\n");
				if ( SDL_SetVideoMode( width, height, bpp, SDL_ANYFORMAT | video_flags) )
					ok=true;


				g_debug("INIT: Unable to create BPP -> trying 16.\n");
				// retry with 16 bit mode
				if(bpp!=16)
				{
					bpp = 16;
					rgb_size[0] = 5;
					rgb_size[1] = 6;
					rgb_size[2] = 5;
					if ( SDL_SetVideoMode( width, height, bpp, video_flags ) != NULL )
					{
						ok=true;
					}
					else
					{
						g_debug("INIT: Unable to create video surface that supports requested details; trying SDL_ANYFORMAT.\n");
						if ( SDL_SetVideoMode( width, height, bpp, SDL_ANYFORMAT | video_flags) )
							ok=true;
					}
				}
			}
			if(!ok)
			{
				g_debug("Couldn't set GL mode: %s\n", SDL_GetError());
				// SDL_Quit(); - endprogram does?
				return false;
			}
			g_debug("SDL_SetVideoMode seems to be %d\n", ok);

			#if defined(HAVE_OPENGL) && defined(EMSCRIPTEN)
				g_debug("Testing glCommands\n");
				glEnable(GL_TEXTURE_2D);
				glDisable(GL_TEXTURE_2D);
				g_debug("OpenGL is ready for use\n");
			#endif



			// Set the screen size again - I put 0 to get "fullscreen" on Palm
			#ifdef PALM_WEBOS
				// 320x480 pre
				// 320x400 pixi
				bpp=32; // had a peek above ;)

				int ilandscape = m_landscape;

				// __DG_DESKTOPX=800; __DG_DESKTOPY=480; - eaglview did it
				if(m_landscape==1 || m_landscape==3) // wanted to rotate by 90
				{
					int swp = __DG_DESKTOPX; __DG_DESKTOPX=__DG_DESKTOPY; __DG_DESKTOPY=swp;
					g_debug("swap deskx, desky to %d,%d - landscape mode %d\n", __DG_DESKTOPX, __DG_DESKTOPY, m_landscape);
				}
				SetScreen(__DG_DESKTOPX, __DG_DESKTOPY, bpp,1);

				// The HP touchpad would rotate portrait more to 1, which is upside down -> toggle to 3
				if(m_physical_device_screen_is_landscape != want_landscape)
				{
					if(m_physical_device_screen_is_landscape)
						m_landscape=3; // rotate by 270 - touchpad so "top" is at camera
					else
						m_landscape=1; // rotate by 90 - palm pre, so "top" is right edge
				}


				height = m_WND.height;
				width = m_WND.width;

				SDL_Delay(200); // Fix the "hanging" of the touchpad with black screen? (swap waits for vsync)
			#endif

		#endif // SDL 1.2



		#if defined( HAVE_OPENGL)
			g_debug("Screen BPP: %d\n", SDL_GetVideoSurface()->format->BitsPerPixel);
			g_debug("\n");
			g_debug( "Vendor     : %s\n", glGetString( GL_VENDOR ) );
			g_debug( "Renderer   : %s\n", glGetString( GL_RENDERER ) );
			g_debug( "Version    : %s\n", glGetString( GL_VERSION ) );
			g_debug( "Extensions : %s\n", glGetString( GL_EXTENSIONS ) );
			g_debug("\n");

			#if defined(EMSCRIPTEN)
				g_debug("skipping size tests\n");
			#else
				int value;
				SDL_GL_GetAttribute( SDL_GL_RED_SIZE, &value );
				g_debug( "SDL_GL_RED_SIZE: requested %d, got %d\n", rgb_size[0],value);
				SDL_GL_GetAttribute( SDL_GL_GREEN_SIZE, &value );
				g_debug( "SDL_GL_GREEN_SIZE: requested %d, got %d\n", rgb_size[1],value);
				SDL_GL_GetAttribute( SDL_GL_BLUE_SIZE, &value );
				g_debug( "SDL_GL_BLUE_SIZE: requested %d, got %d\n", rgb_size[2],value);
				SDL_GL_GetAttribute( SDL_GL_DEPTH_SIZE, &value );
				g_debug( "SDL_GL_DEPTH_SIZE: requested %d, got %d\n", bpp, value );
				SDL_GL_GetAttribute( SDL_GL_DOUBLEBUFFER, &value );
				g_debug( "SDL_GL_DOUBLEBUFFER: requested 1, got %d\n", value );
			#endif

		#endif
	
		#ifndef TRENKI3D // try to fix a Palm Pre startup issue - didn't work :/
			OPENGL_CHECK_ERRORS();
			ClearScreen(0);

			OPENGL_CHECK_ERRORS();
			m_bDirtyZBuffer=true;
			FlipScreens(NULL);
			OPENGL_CHECK_ERRORS();
			m_bDirtyZBuffer=true;
			FlipScreens(NULL);
		#endif

		this->Make2DMode();

	// -------- WIN32 --------
	#elif defined(WIN32) && (defined(HAVE_OPENGL) || (defined(HAVE_OPENGL_ES) && defined(_DEBUG)) )
		g_debug("WIN32 init video\n");
		__DG_DESKTOPX=::GetSystemMetrics(SM_CXSCREEN);
		__DG_DESKTOPY=::GetSystemMetrics(SM_CYSCREEN);

		HDC hDC;
		RECT	rect;
		m_WND.hwnd = CreateWindow(STR_GLBASIC, STR_GLBASIC, WS_CAPTION,
								0, 0,
								// (::GetSystemMetrics(SM_CXSCREEN)-width)/2,
								// (::GetSystemMetrics(SM_CYSCREEN)-height)/2,
								width, height+::GetSystemMetrics(SM_CYCAPTION)+::GetSystemMetrics(SM_CYBORDER)*6,
								NULL, NULL, hInst, NULL);

		if(m_WND.hwnd == NULL) return false;

		::GetWindowRect(m_WND.hwnd, &rect);
		if (bpp==0) bpp=::GetDeviceCaps(::GetDC(NULL), BITSPIXEL);
		if (width==0) width=abs(rect.right-rect.left);
		if (height==0) height=abs(rect.bottom-rect.top);
		m_WND.bpp=bpp;
		m_WND.height=height;
		m_WND.width=width;
		m_WND.fullscreen=false;
		hDC=::GetDC(m_WND.hwnd);
		// Set Viewport
	//	this->ChangeSize(width, height);
		// Select the pixel format
		int nPixelFormat=0;

		#ifndef PFD_SUPPORT_COMPOSITION
		#define PFD_SUPPORT_COMPOSITION 0x00008000
		#endif

		PIXELFORMATDESCRIPTOR pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),  // Size of this structure
		1,                              // Version of this structure
		PFD_DRAW_TO_WINDOW |// Format Must Support Window
		PFD_SUPPORT_OPENGL |// Format Must Support OpenGL
		PFD_SUPPORT_COMPOSITION| // composition (for accelerated OpenGL on Vista)
		PFD_DOUBLEBUFFER,	// Must Support Double Buffering
		PFD_TYPE_RGBA,		// Request An RGBA Format
		(BYTE)min(24,m_WND.bpp),  // color bits, excluding alpha
		0, 0, 0, 0, 0, 0,	// Color Bits Ignored (?)
		8,					// No Alpha Buffer
		0,					// Shift Bit Ignored (?)
		0,					// No Accumulation Buffer
		0, 0, 0, 0,			// Accumulation Bits Ignored (?)
		24,					// 24 Bit Z-Buffer (Depth Buffer)
		8,					// 8 Bit Stencil Buffer
		0,					// No Auxiliary Buffer (?)
		PFD_MAIN_PLANE,		// Main Drawing Layer
		0,					// Reserved (?)
		0, 0, 0};			// Layer Masks Ignored (?)

		// Choose a pixel format that best matches that described in pfd
		nPixelFormat = ChoosePixelFormat(hDC, &pfd);
		SetPixelFormat(hDC, nPixelFormat, &pfd);

		// Try to share this context
		if(!m_hRC)
			m_hRC = wglCreateContext(hDC);
		// share failed - recreate
		if(!wglMakeCurrent(hDC, m_hRC))
		{
			wglDeleteContext(m_hRC);
			m_hRC = wglCreateContext(hDC);
			wglMakeCurrent(hDC, m_hRC);
		}
		::ReleaseDC(m_WND.hwnd, hDC);

		// We support multisampling!
		m_WND.multisample = 0;
		#if defined(HAVE_OPENGL)
		if(multisamples && (nPixelFormat = ChooseMultisamplePixelFormat(hDC, pfd, m_WND.multisample)))
		{
			Close();

			m_WND.hwnd = CreateWindow(STR_GLBASIC, STR_GLBASIC, WS_CAPTION,
									0, 0,
									// (::GetSystemMetrics(SM_CXSCREEN)-width)/2,
									// (::GetSystemMetrics(SM_CYSCREEN)-height)/2,
									width, height+::GetSystemMetrics(SM_CYCAPTION)+::GetSystemMetrics(SM_CYBORDER)*6,
									NULL, NULL, hInst, NULL);

			if(m_WND.hwnd == NULL) return false;
			hDC=::GetDC(m_WND.hwnd);
			SetPixelFormat(hDC, nPixelFormat, &pfd);
			m_hRC = wglCreateContext(hDC);
			wglMakeCurrent(hDC, m_hRC);
			::ReleaseDC(m_WND.hwnd, hDC);
		}
		#endif
		
		// Check for HW support
		static int did_warn=0;
		if(!did_warn)
		if (DescribePixelFormat(hDC, nPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd))
		{
			static const char* pRenderer = (const char*)glGetString(GL_RENDERER);
			if (pfd.dwFlags & PFD_GENERIC_FORMAT || strcmp(pRenderer, "Mesa Windows GDI Driver")==0)
			{
				did_warn=1;
				g_debug("*** OpenGL Software Emulation - update driver!\n");
				GFFileSystem fs;
				char msg[2048]="";
				memset(msg, 0, 2048);
				if(fs.OpenIn("opengldriver.txt"))
				{
					fs.Read(msg, 2047);
					fs.Close();
				}
				if(strlen(msg)>0)
				{
					MessageBox(NULL, msg, "No OpenGL Hardware Drivers", MB_OK|MB_ICONWARNING);
				}
				else
				{
					MessageBox(NULL, "Please update your GFX adapter's driver\nso it can render OpenGL by HARDWARE\n\n\nBitte aktualisieren Sie Ihren\nGrafkkartentreiber, damit er OpenGL\nper HARDWARE untersttzt!",
					"No OpenGL Hardware Drivers", MB_OK|MB_ICONWARNING);
				}
			}
		}

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // *** SPACE FRACTAL FIXES 2017.3.1 ****
	#elif defined(TARGET_OS_IPHONE)
		// nothing to do, move along
		// all done in iPhoneAppController.m
		// m_physical_device_screen_is_landscape = __DG_DESKTOPX > __DG_DESKTOPY ? 1:0;


        // SPACEFRACTAL:
        // No Orientation swap should been done now from iOS8. Otherwice, you might start
        // the app in a wrong size or orientation instead. You can still use software
        // orientation by setting LANDSCAPE RIGHT only support in the XCODE setting.
		SetScreen(width, height, 16, TRUE);
		height = m_WND.height;
		width = m_WND.width;

		// __DG_DESKTOPX=800; __DG_DESKTOPY=480; - eaglview did it
		// if(m_landscape==1 || m_landscape==3) // landscape mode
		// {
		// 		int swp = __DG_DESKTOPX; __DG_DESKTOPX=__DG_DESKTOPY; __DG_DESKTOPY=swp;
		// }
		g_debug("phy dev is landscape: %d. m_landscape:%d\n", m_physical_device_screen_is_landscape, m_landscape);
		g_debug("desk: %dx%d; screen %dx%d\n", __DG_DESKTOPX, __DG_DESKTOPY, m_WND.width, m_WND.height);

	#elif defined(TRENKI3D) // just trenki3D, no SDL. Use /dev/fb0 or whatever. -> GP2X, e.g.
		g_debug("video init Trenki3D\n");
		// OK, you may pass
		#if defined( GP2X ) || defined(GPHWIZ) || defined(CAANOO)
			__DG_DESKTOPX=320; __DG_DESKTOPY=240;
		#else
			#error screen resolution, please?
		#endif
	#else
		#error no Rainbows Initalized
	#endif


	m_WND.bpp=bpp;
	m_WND.height=height;
	m_WND.width=width;
	m_WND.virtw=0;
	m_WND.virth=0;
	m_WND.fullscreen=false;
	
	// debug landscape on iPad mode
	#if defined(_DEBUG) && defined(WIN32) && defined(HAVE_OPENGL_ES)
		m_WND.height=max(height, width);
		m_WND.width=min(height, width);
		SetScreen(m_WND.width, m_WND.height, 32, false);
	#endif
	
	#if defined(TINYGL) || defined(TRENKI3D)
		Screen.Create(NULL, m_WND.width, m_WND.height, false);
	#endif

#else // POCKETPC
	m_WND.bpp=bpp;
	m_WND.height=height;
	m_WND.width=width;
	m_WND.fullscreen=false;
#endif

	g_debug("get accurate timer - 1st call\n");
	m_next_time=AccurateTime();

	g_debug("flip - 1st call\n");
	this->FlipScreens(); // no glGetString on iOS, otherwise

	// const char* pVendor = (const char*)glGetString( GL_VENDOR );
	// OpenGL extension loading
	#if defined(HAVE_OPENGL)
	{
		char *strExtensions = (char*)glGetString(GL_EXTENSIONS);

		if(strExtensions && strstr(strExtensions, "WGL_EXT_swap_control")!=0)
		{
			gf_wglSwapIntervalEXT = (PFWGL_SWAP_CTRL) glGFGetProcAddressArb("wglSwapIntervalEXT");
		}

		if( strExtensions && strstr(strExtensions, "GL_ARB_multitexture") != 0)
		{
			glActiveTextureArb	=(PFNGLACTIVETEXTUREARBPROC)  glGFGetProcAddressArb("glActiveTextureARB");
			glMultiTexCoord2fArb=(PFNGLMULTITEXCOORD2FARBPROC)glGFGetProcAddressArb("glMultiTexCoord2fARB");
			glClientActiveTextureArb = (PFNGLCLIENTACTIVETEXTUREARBPROC) glGFGetProcAddressArb("glClientActiveTextureARB");

			// VBO Extension Function Pointers
			glGenBuffersArb = (PFNGLGENBUFFERSARBPROC)glGFGetProcAddressArb("glGenBuffersARB");
			glBindBufferArb = (PFNGLBINDBUFFERARBPROC)glGFGetProcAddressArb("glBindBufferARB");
			glBufferDataArb = (PFNGLBUFFERDATAARBPROC)glGFGetProcAddressArb("glBufferDataARB");
			glDeleteBuffersArb = (PFNGLDELETEBUFFERSARBPROC)glGFGetProcAddressArb("glDeleteBuffersARB");
			glMapBufferArb = (PFNGLMAPBUFFERARBPROC)glGFGetProcAddressArb("glMapBufferARB");
			glUnmapBufferArb = (PFNGLUNMAPBUFFERARBPROC)glGFGetProcAddressArb("glUnmapBufferARB");
			glGetBufferParameterivArb = (PFNGLGETBUFFERPARAMETERIVARBPROC)glGFGetProcAddressArb("glGetBufferParameterivARB");
		}
		if( strExtensions && strstr(strExtensions, "GL_ARB_fragment_program") != 0)
		{
			glCreateProgramObjectArb=	(PFNGLCREATEPROGRAMOBJECTARBPROC     ) glGFGetProcAddressArb("glCreateProgramObjectARB");
			glDeleteObjectArb=			(PFNGLDELETEOBJECTARBPROC            ) glGFGetProcAddressArb("glDeleteObjectARB");
			glCreateShaderObjectArb=	(PFNGLCREATESHADEROBJECTARBPROC      ) glGFGetProcAddressArb("glCreateShaderObjectARB");
			glShaderSourceArb=			(PFNGLSHADERSOURCEARBPROC            ) glGFGetProcAddressArb("glShaderSourceARB");
			glCompileShaderArb=			(PFNGLCOMPILESHADERARBPROC           ) glGFGetProcAddressArb("glCompileShaderARB");
			glGetObjectParameterivArb=	(PFNGLGETOBJECTPARAMETERIVARBPROC    ) glGFGetProcAddressArb("glGetObjectParameterivARB");
			glAttachObjectArb=			(PFNGLATTACHOBJECTARBPROC            ) glGFGetProcAddressArb("glAttachObjectARB");
			glGetInfoLogArb=			(PFNGLGETINFOLOGARBPROC              ) glGFGetProcAddressArb("glGetInfoLogARB");
			glLinkProgramArb=			(PFNGLLINKPROGRAMARBPROC             ) glGFGetProcAddressArb("glLinkProgramARB");
			glUseProgramObjectArb=		(PFNGLUSEPROGRAMOBJECTARBPROC        ) glGFGetProcAddressArb("glUseProgramObjectARB");
			glGetUniformLocationArb=	(PFNGLGETUNIFORMLOCATIONARBPROC      ) glGFGetProcAddressArb("glGetUniformLocationARB");
			glUniform1fArb=				(PFNGLUNIFORM1FARBPROC               ) glGFGetProcAddressArb("glUniform1fARB");
			glUniform1iArb=				(PFNGLUNIFORM1IARBPROC               ) glGFGetProcAddressArb("glUniform1iARB");
			glGetAttribLocationArb=		(PFNGLGETATTRIBLOCATIONARBPROC       ) glGFGetProcAddressArb("glGetAttribLocationARB");
			glVertexAttrib1fArb=		(PFNGLVERTEXATTRIB1FARBPROC          ) glGFGetProcAddressArb("glVertexAttrib1fARB");
		}
	}
	#endif


	#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
	{
		char *strExtensions = (char*)glGetString(GL_EXTENSIONS);
		
		m_bSupportBGRA_ext = false;
		if( strExtensions && strstr(strExtensions, /*GL_IMG, GL_APPLE*/"_texture_format_BGRA8888") != 0)
			m_bSupportBGRA_ext = true;
		if( strExtensions && strstr(strExtensions, "GL_EXT_bgra") != 0)
			m_bSupportBGRA_ext = true;

		#ifdef PALM_WEBOS // Pixi crashes with GL_RGBA, but works with GL_BGRA_EXT - no extension signaled, though (bug)
			m_bSupportBGRA_ext = true;
		#endif

		if(m_bSupportBGRA_ext)
		{
			g_debug("BGRA ext supported\n");
		}
		else
		{
			g_debug("BGRA ext NOT supported\n%s\n", strExtensions);
		}

		int				max_tex;
		glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_tex);
		g_debug("Texture size limit: %d\n", max_tex);
	}
	#endif

	g_debug("init fbo\n");
	// Initialize the FrameBuffer extension if available
	FBUFFER::InitFrameBufferObject();

#if !defined( WANT_SDL) && defined(WIN32) && !defined(POCKETPC)
	UpdateWindow(m_WND.hwnd);
	ShowWindow(m_WND.hwnd, SW_HIDE);
#endif

	// Setup OpenGL State
//	this->DoWindowed();
	g_debug("2D VP\n");
	this->Make2DMode();
//pi_test(); exit(0);
	g_debug("OGRB init [OK]\n");
return true;
}

//---------------------------------------------------------------//
//---------------------------------------------------------------//
void OpenGLRainbows::SetMaxFPS(int maxfps, bool bVsync)
{
	#if defined(EMSCRIPTEN)
		// not implemented
	#elif defined(WANT_SDL) && (defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES))
		#if SDL_VERSION_ATLEAST(1,3,0) 
		SDL_GL_SetSwapInterval( bVsync ? 1:0 ); 
		#else /* SDL_VERSION_ATLEAST(1,3,0) */ 
		#ifdef SDL_GL_SWAP_CONTROL 
		SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, bVsync ? 1:0 ); 
		#else /* SDL_GL_SWAP_CONTROL */ 
		g_debug("VSync unsupported on old SDL versions (before 1.2.10).\n");
		#endif /* SDL_GL_SWAP_CONTROL */ 
		#endif /* SDL_VERSION_ATLEAST(1,3,0) */
	#elif defined(WIN32) && !defined(_WIN32_WCE) && !defined(TRENKI3D)
		if(gf_wglSwapIntervalEXT) gf_wglSwapIntervalEXT(bVsync ? 1:0);
	#endif

	if (maxfps<=0.0)
	{
		m_FrameRate = 2.0; // 500 fps = max
	}
	else
	{
		m_FrameRate=DGInt(DGInt(1000.0)/ DGInt(maxfps));
	}
}


//---------------------------------------------------------------//
//---------------------------------------------------------------//
void OpenGLRainbows::SetScreen(int w, int h, int bpp, int fullscreen)
{
#if defined( GP2X ) || defined(GPHWIZ) || defined(PANDORA) || defined(RASPBERRY_PI) || defined(CAANOO)
	return;
#endif

// *** SPACE FRACTAL FIXES 2017.3.1 ****
// Orientation needs ALLWAYS to been started on 0, regaardless of start orientation.
// This is due orientation changes in iOS8.
#if defined(TARGET_OS_IPHONE)
    m_WND.width = __DG_DESKTOPX; // allways start in orientation 0
    m_WND.height= __DG_DESKTOPY;
    m_landscape=0;
    return;
#endif

#if defined(PALM_WEBOS) || defined(ANDROID)
g_debug("Rbow::SetScreen( %d,%d)\n", w,h);
	if(		(w>h &&  m_physical_device_screen_is_landscape) // screen fits orientation, no rotations required
		||	(h>w && !m_physical_device_screen_is_landscape)
	)
	{
		m_landscape=0;
	}
	else
	{
		m_landscape=1;
	}
	// m_WND.width=320; m_WND.height=480; // NOT ON IPAD!!
	
	// m_WND.width = min(w,h);
	// m_WND.height= max(w,h);

	if(m_physical_device_screen_is_landscape)
	{
		m_WND.width = max(__DG_DESKTOPX, __DG_DESKTOPY); // physical size of device
		m_WND.height= min(__DG_DESKTOPX, __DG_DESKTOPY);
	}
	else
	{
		m_WND.width = min(__DG_DESKTOPX, __DG_DESKTOPY);
		m_WND.height= max(__DG_DESKTOPX, __DG_DESKTOPY);
	}

	
	return;
#endif


	if(w) m_WND.width = w;
	if(h) m_WND.height = h;
	if(bpp) m_WND.bpp = bpp;
	m_WND.virth=0; m_WND.virtw=0; // reset - will be set on next viewport call.

#ifndef POCKETPC

	// Go into windowed mode just to be safe
	DoWindowed();


	#ifdef WANT_SDL
		if(w || h || bpp)
		{
	    #if SDL_VERSION_ATLEAST(2,0,0)
			SDL_SetWindowFullscreen(m_WND.sdl_window, SDL_FALSE);
			SDL_SetWindowSize(m_WND.sdl_window, m_WND.width, m_WND.height);
		#else
			Uint32 video_flags=0;
			#if (defined(TINYGL) || defined(TRENKI3D) )
				video_flags |= SDL_DOUBLEBUF;
				bpp = 16;
			#else
				video_flags = SDL_OPENGL;
			#endif
			SDL_SetVideoMode(m_WND.width, m_WND.height, m_WND.bpp, video_flags);
		#endif
		}
	#endif

	if(fullscreen)
		DoFullscreen();
	else
		DoWindowed();

	#if defined(TINYGL) || defined(TRENKI3D)
		HWND hwnd=NULL;
		#if !defined( WANT_SDL) && defined(WIN32)
		hwnd = m_WND.hwnd;
		#endif
		Screen.Create((HWND)hwnd, w,h, w>h ? 1:0);
	#endif

#endif
}



//---------------------------------------------------------------//
//---------------------------------------------------------------//
bool OpenGLRainbows::DoFullscreen()
{
// -------- GP2X --------
#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE)
// -------- Pandora --------
#elif defined(PANDORA) || defined(RASPBERRY_PI)
// -------- SDL --------
#elif defined(WANT_SDL)
	#if !defined( TARGET_OS_IPHONE) && !defined(PALM_WEBOS) && !defined(ANDROID)
	    #if SDL_VERSION_ATLEAST(2,0,0)
			SDL_SetWindowFullscreen(m_WND.sdl_window, SDL_TRUE);
		#else
			SDL_Surface *screen;
			screen = SDL_GetVideoSurface();
			if (!(screen->flags&SDL_FULLSCREEN))
				SDL_WM_ToggleFullScreen(screen);
		#endif
	#endif
#elif defined(MACOSX)
	return true;
#elif defined(POCKETPC)
	return true;
// -------- WIN32 --------
#elif defined(WIN32)
	DEVMODE	dm;
	int		CurrBPP;
	HDC		hdc;
		memset(&dm, 0, sizeof(dm));
		dm.dmSize = sizeof(dm);

		dm.dmPelsWidth  = m_WND.width;
		dm.dmPelsHeight = m_WND.height;
		dm.dmFields     = DM_PELSWIDTH | DM_PELSHEIGHT;

		hdc=::GetDC(NULL);
		CurrBPP=::GetDeviceCaps(hdc, BITSPIXEL);
		ReleaseDC(NULL, hdc);
		if (m_WND.bpp != CurrBPP && m_WND.bpp != 0)
		{
			dm.dmBitsPerPel = m_WND.bpp;
			dm.dmFields |= DM_BITSPERPEL;
		}

		if(ChangeDisplaySettings(&dm, CDS_FULLSCREEN)!=DISP_CHANGE_SUCCESSFUL)
		{
			dm.dmFields=DM_PELSWIDTH | DM_PELSHEIGHT; // No Color Depth Change
			if (ChangeDisplaySettings(&dm, CDS_FULLSCREEN)!=DISP_CHANGE_SUCCESSFUL)
			{
				::MessageBeep(MB_ICONEXCLAMATION);
				g_debug("No Fullscreen Mode available\n");
				this->DoWindowed();
				return false;
			}
		}
		
		OpenGLRainbows::OPENGL_RAINBOWS_WINDOW wnd = m_WND;
//		Close(); // would destroy context

#if 0 // GF: 2015.08.15 this would destroy the old window? We just need to change size and params!
		// (re)store window text
		char wtext[512];
		::GetWindowText(m_WND.hwnd, wtext, 512);
		if(!Init(m_hinstance, wnd.width, wnd.height, wnd.bpp, m_FrameRate>0 ? (int)(1000/m_FrameRate) : 5, wnd.multisample))
		{
			return false;
		}
		else
		{
			::SetWindowText(m_WND.hwnd, wtext);
			::DestroyWindow(wnd.hwnd);
		}
#endif

		this->ChangeSize(m_WND.width, m_WND.height);
		::SetWindowLong(m_WND.hwnd,GWL_STYLE,WS_POPUP|WS_MAXIMIZE);
		::ShowWindow(m_WND.hwnd,SW_RESTORE);
		::UpdateWindow(m_WND.hwnd);
		::SetForegroundWindow(m_WND.hwnd);
		::SetFocus(m_WND.hwnd);
		::SetWindowPos(m_WND.hwnd,HWND_TOPMOST,0,0, m_WND.width, m_WND.height, 0);

		this->Make2DViewport();


#endif

	m_WND.fullscreen=true;
return true;
}


#ifdef WIN32
#if ! (defined( WANT_SDL ) || defined(POCKETPC) )

	#define MAX_WINDOWS 256
	static int num_windows = 0;
	static HWND window_hwnd[MAX_WINDOWS];
	static WINDOWPLACEMENT placement[MAX_WINDOWS];

	BOOL CALLBACK EnumWindowsProc(HWND hwnd, LPARAM lParam)
	{
	  if(num_windows < MAX_WINDOWS)
	  {
		if(IsWindowVisible(hwnd))
		{
		  placement[num_windows].length = sizeof(WINDOWPLACEMENT);
		  if(GetWindowPlacement(hwnd, &placement[num_windows]) != 0)
		  {
			window_hwnd[num_windows] = hwnd;
			num_windows++;
		  }
		}
	  }
	  return TRUE;
	}


	void save_desktop(void)
	{
	  num_windows = 0;
	  EnumWindows(EnumWindowsProc, 0);
	}

	void restore_desktop1(void)
	{
	  int i;
	  for(i = 0; i < num_windows; i++)
	  {
		HWND hwnd = window_hwnd[i];
		if(IsWindow(hwnd))
		{
		  SetWindowPlacement(hwnd, &placement[i]);
		}
	  }
	}

	void restore_desktop(void)
	{
	  /* Okay, this may look strange. I don't like it either.
		 We are always calling this right after ChangeDisplaySettings(),
		 and it seems that the mode change may still be taking place,
		 as sometimes the windows aren't restored. The more delay I
		 put, the more likely it is that it will work. But if there
		 is much delay, the window movement is clearly visible. And
		 delay isn't very robust way to do anything. But I couldn't
		 find a reliable way to check if it worked or not.
	  */

	  restore_desktop1();
	  Sleep(500);
	  restore_desktop1();
	}
#endif
#endif

/*
void CenterWindow(HWND hwnd)
{
    int x, y;
    HWND hwndDeskTop;
    CRect rcWnd, rcDeskTop;
    // Get a handle to the desktop window

    hwndDeskTop = ::GetDesktopWindow();
    // Get dimension of desktop in a rect

    ::GetWindowRect(hwndDeskTop, &rcDeskTop);
    // Get dimension of main window in a rect

    ::GetWindowRect(hwnd, &rcWnd);
    // Find center of desktop

    x = rcDeskTop.Width() / 2;
    y = rcDeskTop.Height() / 2;
    x -= rcWnd.Width() / 2;
    y -= rcWnd.Height() / 2;
    // Set top and left to center main window on desktop

    ::SetWindowPos(hwnd, HWND_TOP, x, y, 0, 0, SWP_NOSIZE);
}
*/


//---------------------------------------------------------------//
//---------------------------------------------------------------//
bool OpenGLRainbows::DoWindowed()
{
	// -------- GP2X --------
	#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
	
	// -------- iPhone --------
	#elif defined(TARGET_OS_IPHONE)
	// -------- Pandora --------
	#elif defined(PANDORA) || defined(RASPBERRY_PI)

	// -------- SDL --------
	#elif defined(WANT_SDL)
		#if !defined( TARGET_OS_IPHONE) && !defined(PALM_WEBOS) && !defined(ANDROID)
			#if SDL_VERSION_ATLEAST(2,0,0)
				SDL_SetWindowFullscreen(m_WND.sdl_window, SDL_FALSE);
			#else
				SDL_Surface *screen;
				screen = SDL_GetVideoSurface();
				if (screen->flags&SDL_FULLSCREEN)
					SDL_WM_ToggleFullScreen(screen);
			#endif
		#endif
	#elif defined(MACOSX)
		return true;
	#elif defined(POCKETPC)
		return true;
	// -------- WIN32 --------
	#elif defined(WIN32)
		if (m_WND.fullscreen)
		{
			::ChangeDisplaySettings(NULL, 0);
			restore_desktop();
		}

		::SetWindowLong(m_WND.hwnd, GWL_EXSTYLE, 0);
		::SetWindowLong(m_WND.hwnd, GWL_STYLE, WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX); // | WS_CLIPCHILDREN | WS_CLIPSIBLINGS);
		::UpdateWindow(m_WND.hwnd);

		WINDOWPLACEMENT	wpl;
		if(0  ) // !::GetWindowPlacement(m_WND.hwnd, &wpl)) // Win98 will fail this here...
		{
			wpl.rcNormalPosition.left=0;
			wpl.rcNormalPosition.top=0;
			wpl.rcNormalPosition.right=m_WND.width;
			wpl.rcNormalPosition.bottom=m_WND.height;
			::OffsetRect(&wpl.rcNormalPosition, (::GetSystemMetrics(SM_CXSCREEN)-m_WND.width)/2, (::GetSystemMetrics(SM_CYSCREEN)-m_WND.width)/2);
		}
		else
		{
			RECT rcCenter, rcArea;
			::GetWindowRect(m_WND.hwnd, &rcCenter);
			HMONITOR hMonitor = MonitorFromRect(&rcCenter, MONITOR_DEFAULTTONEAREST);
			if(hMonitor)
			{
				MONITORINFO MonitorInfo = { sizeof MonitorInfo };
				GetMonitorInfo(hMonitor, &MonitorInfo);
				rcArea = MonitorInfo.rcWork;
			}
			else
				::SystemParametersInfo(SPI_GETWORKAREA, 0, &rcArea, 0);

			wpl.rcNormalPosition.left=0;
			wpl.rcNormalPosition.top=0;
			wpl.rcNormalPosition.right=m_WND.width                                   +::GetSystemMetrics(SM_CXBORDER)*6;
			wpl.rcNormalPosition.bottom=m_WND.height+::GetSystemMetrics(SM_CYCAPTION)+::GetSystemMetrics(SM_CYBORDER)*6;
			::OffsetRect(&wpl.rcNormalPosition, ((rcArea.right-rcArea.left)-m_WND.width)/2, ((rcArea.bottom-rcArea.top)-m_WND.height)/2);
		}

		::SetWindowPos(m_WND.hwnd, HWND_NOTOPMOST,
						wpl.rcNormalPosition.left,
						wpl.rcNormalPosition.top,
						wpl.rcNormalPosition.right - wpl.rcNormalPosition.left,
						wpl.rcNormalPosition.bottom- wpl.rcNormalPosition.top,
						SWP_SHOWWINDOW);
		::ShowWindow(m_WND.hwnd, SW_SHOWNORMAL);
		
		// update desktop
		::InvalidateRect(0, NULL, TRUE);

		::UpdateWindow(m_WND.hwnd);
		::BringWindowToTop(m_WND.hwnd);

		// if (m_WND.fullscreen) ::ShowCursor(TRUE);
		this->Make2DViewport();
	#endif

	m_WND.fullscreen=false;
return true;
}

//---------------------------------------------------------------//
// Hide window
//---------------------------------------------------------------//
bool OpenGLRainbows::HideWindow()
{
// -------- GP2X --------
#if defined(GP2X) || defined(GPHWIZ) || defined(CAANOO)
// -------- iPhone --------
#elif defined(TARGET_OS_IPHONE)
// -------- Pandora --------
#elif defined(PANDORA)||defined(RASPBERRY_PI)
// -------- SDL --------
#elif defined(WANT_SDL)
// #NEED SDL
#elif defined(MACOSX)
	return true;
#elif defined(POCKETPC)
	return true;
// -------- WIN32 --------
#elif defined(WIN32)
	if (m_WND.fullscreen) ::ChangeDisplaySettings(NULL, 0);

	::SetWindowLong(m_WND.hwnd, GWL_EXSTYLE, 0);
	::SetWindowLong(m_WND.hwnd, GWL_STYLE, WS_OVERLAPPED | WS_CLIPCHILDREN | WS_CLIPSIBLINGS);

//	::MoveWindow(m_WND.hwnd, 0, 0, m_WND.width, m_WND.height, TRUE);
	::ShowWindow(m_WND.hwnd, SW_HIDE);
	::UpdateWindow(m_WND.hwnd);

//	if (m_WND.fullscreen) ::ShowCursor(TRUE);
	m_WND.fullscreen=false;
	this->WinMessages();
#endif
return true;
}


#ifndef POCKETPC
//---------------------------------------------------------------//
//---------------------------------------------------------------//
void OpenGLRainbows::ChangeSize(int w, int h)
{
	m_WND.height=h; m_WND.width=w;
	Make2DViewport();
}
#endif



///////////////////////////////////////////////////////////////////
//  POLY WORK
///////////////////////////////////////////////////////////////////


class POLY_POINT2D
{
public:
	float pos[2];
	float tex[2];
	unsigned char col[4]; // must be 4 bytes! TRENKI_COL_PRT (sizeof(float))

	POLY_POINT2D() {}
	POLY_POINT2D(const POLY_POINT2D& pp) {*this=pp;}
	inline POLY_POINT2D& operator=(const POLY_POINT2D& pp)
	{
		// might be faster!?
		if(sizeof(POLY_POINT2D) == 5*sizeof(int))
		{
			unsigned int* pP = (unsigned int*) this;
			const unsigned int* pS = (const unsigned int*) &pp;
			*pP++ = *pS++;
			*pP++ = *pS++;
			*pP++ = *pS++;
			*pP++ = *pS++;
			*pP++ = *pS++;
		}
		else
		{
			pos[0]=pp.pos[0]; pos[1]=pp.pos[1];
			tex[0]=pp.tex[0]; tex[1]=pp.tex[1];
			*(int*)&col = *(int*)&pp.col;
		}
		return *this;
	}
};
Queue<POLY_POINT2D> g_PolyPoints;
int		g_nTexture=0; // the value you pass to X_MAKE3D(here, ...)
int     g_PolyMode=0; // 0:fan, 1:tri, 2:strip
int     g_PolyHasColor=0;
float   g_FtxInv, g_FtyInv;
int		g_bLastPolyStripGroup=0;

//---------------------------------------------------------------//
// Start a poly
//---------------------------------------------------------------//
void OpenGLRainbows::PolyStart(int nTexture, int iMode)
{
//g_nTexture=nTexture;
	SetTexture3D(nTexture, -1, 0);
	g_PolyPoints.Clear(false);
	g_PolyPoints.AllocAtLeast(150); // 480x320 screen with 32x32 tiles ;)
	g_PolyMode=iMode;
	g_PolyHasColor=0;
	if(nTexture>=0)
	{
		g_FtxInv = 1.0f / m_Texture[nTexture].w;
		g_FtyInv = 1.0f / m_Texture[nTexture].h;
	}
	g_bLastPolyStripGroup=0;
}

//---------------------------------------------------------------//
// Add Points to Poly
//---------------------------------------------------------------//
void OpenGLRainbows::PolyPoint(float sx, float sy, float tx, float ty, COLORREF col)
{
	int num = g_PolyPoints.Size();
	g_PolyPoints.SetSize(num + 1 + g_bLastPolyStripGroup );
	POLY_POINT2D& pp = g_PolyPoints[num];
	pp.pos[0]=sx; pp.pos[1]=sy;
	
	// tx/ty come in _PIXELs_, not in texture coords
	// and the user don't fucking care about non power of 2
	// texture sizes!!!
	if(g_nTexture>=0)
	{
		pp.tex[0]=tx* g_FtxInv; pp.tex[1]=((float)m_Texture[g_nTexture].bh - ty)*g_FtyInv;
	}
	if(col==0xffffff)
	{
		pp.col[0]=pp.col[1]=pp.col[2]=255;
	}
	else
	{
		g_PolyHasColor=1;
		pp.col[0] = GetRValue(col);
		pp.col[1] = GetGValue(col);
		pp.col[2] = GetBValue(col);
	}
	if(col == m_CookieColor)
		pp.col[3] =0;
	else if (m_falpha == 1.0f)
		pp.col[3] = 255;
	else
		pp.col[3] = (unsigned char)(m_falpha*255.0f);

	if(g_bLastPolyStripGroup)
	{
		g_PolyPoints[num+1] = pp;
		g_bLastPolyStripGroup = 0;
	}
POLY_POINT2D p2;
	p2 = pp;
}

//---------------------------------------------------------------//
// Add a degenerated poly strip for a new "group" of strips
//---------------------------------------------------------------//
void OpenGLRainbows::PolyNewStrip(void)
{
	int num = g_PolyPoints.Size();
	if(num && !g_bLastPolyStripGroup)
	{
		g_bLastPolyStripGroup = 1;

		g_PolyPoints.SetSize(num+1);
		g_PolyPoints[num] = g_PolyPoints[num-1];
	}
}

/*
Mode = 0 (FAN)
Alle Punkte werden mit dem ersten Punkt erbunden.
1_____0______4
\    / \    /
 \  /   \  /
  \/_____\/
  2       3

Mode = 1 (TRINGLES)
Jeweils 3 Punkte bilden ein eigenstndiges Dreieck.
1_____0  3______5
\    /    \    /
 \  /      \  /
  \/        \/
  2          4

Mode = 1 (STRIP)
// Jeweils 2 Punkte Bilden ein Vieweck.
0_____2______4
\    / \    /
 \  /   \  /
  \/_____\/
  1       3
*/

//---------------------------------------------------------------//
// Finish a Poly
//---------------------------------------------------------------//
void OpenGLRainbows::PolyEnd(void)
{
#if defined(TRENKI3D)

	switch(g_PolyMode)
	{
	case 0:
		tglDrawArraysXYZ_UV_RGBA(GL_TRIANGLE_FAN, g_PolyPoints.Size(), &g_PolyPoints.GetAt(0)->pos[0], 0, 0, g_nTexture>=0 ? 1:0, 1, 1);
		break;
	case 1:
		tglDrawArraysXYZ_UV_RGBA(GL_TRIANGLES, g_PolyPoints.Size(), &g_PolyPoints.GetAt(0)->pos[0], 0, 0, g_nTexture>=0 ? 1:0, 1, 1);
		break;
	case 2:
		if(g_PolyHasColor || m_falpha != 1.0f)
		{
			tglDrawArraysXYZ_UV_RGBA(GL_TRIANGLE_STRIP, g_PolyPoints.Size(), &g_PolyPoints.GetAt(0)->pos[0], 0, 0, g_nTexture>=0 ? 1:0, 1, 1);
		}
		else
		{
			tglDrawArraysXYZ_UV_RGBA(GL_TRIANGLE_STRIP, g_PolyPoints.Size(), &g_PolyPoints.GetAt(0)->pos[0], 0, 0, g_nTexture>=0 ? 1:0, 0, 1);
		}
		break;
	}
#else

	// Enable Pointers
	glEnableClientState( GL_VERTEX_ARRAY );
	glVertexPointer(  2, GL_FLOAT, sizeof(POLY_POINT2D), (char*)&g_PolyPoints.GetAt(0)->pos[0]);
	if(g_nTexture>=0)
	{
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, sizeof(POLY_POINT2D), (char*)&g_PolyPoints.GetAt(0)->tex[0]);
 	}

 	if(g_PolyHasColor || m_falpha != 1.0f)
 	{
 		g_PolyHasColor=1;
		//glbEnable(GL_COLOR_MATERIAL);
		glbEnable(GL_DITHER);
		glShadeModel(GL_SMOOTH);
 		glEnableClientState(GL_COLOR_ARRAY);
 		glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(POLY_POINT2D), (char*)&g_PolyPoints.GetAt(0)->col[0]);
	}

	ApplySmoothShading();


	// Draw All Of The Triangles At Once
	switch(g_PolyMode)
	{
	case 0:
		glDrawArrays(GL_TRIANGLE_FAN, 0, g_PolyPoints.Size());
		break;
	case 1:
		glDrawArrays( GL_TRIANGLES, 0, g_PolyPoints.Size());
		break;
	case 2:
		glDrawArrays( GL_TRIANGLE_STRIP, 0, g_PolyPoints.Size());
		break;
	}
	// Disable Pointers
	glDisableClientState( GL_VERTEX_ARRAY );					// Disable Vertex Arrays
	glDisableClientState( GL_TEXTURE_COORD_ARRAY );				// Disable Texture Coord Arrays


 	if(g_PolyHasColor)
 	{
		glbColor4f(1.0f, 1.0f, 1.0f, m_falpha);
		glDisableClientState( GL_COLOR_ARRAY );						// Disable Color Arrays

		glShadeModel(GL_FLAT);
		glbDisable(GL_DITHER);
		//glbDisable(GL_COLOR_MATERIAL);
 	}
#endif
}


//---------------------------------------------------------------//
// Enable smooth shading mode
//---------------------------------------------------------------//
void OpenGLRainbows::SetSmoothShading(bool smooth)
{
#if !defined(TRENKI3D) && !defined(TINYGL)
	m_SmoothShading = smooth;
#endif
}

//---------------------------------------------------------------//
// Enable smooth shading mode
//---------------------------------------------------------------//
void OpenGLRainbows::ApplySmoothShading()
{
#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
	if(m_SmoothShading)
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	}
	else
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	}
#endif
}


//---------------------------------------------------------------//
// 1.0 -> all solid, no testing
//---------------------------------------------------------------//
void OpenGLRainbows::SetAlphaTesting(float greaterthen)
{
#if defined(TINYGL) || defined(TRENKI3D)
	Screen.AlphaTesting(greaterthen < 1.0f ? true:false);
#else
	if(greaterthen >= 1.0f)
	{
		glbDisable(GL_ALPHA_TEST);
		m_fAlphaTestingGTH = 1.0f;
	}
	else
	{
		// Bug in webgl port, so far
		#if defined(EMSCRIPTEN)
			greaterthen = 1.0f;
			DoAlpha(-1.0);
		#else
			glbEnable(GL_ALPHA_TEST);
			glAlphaFunc(GL_GREATER, greaterthen); // 0.5
		#endif
		m_fAlphaTestingGTH = greaterthen;
	}
#endif
}



#if !defined(TINYGL) && !defined(TRENKI3D)
//---------------------------------------------------------------//
// Draw a 4 noded 2D object (Sprite)
//---------------------------------------------------------------//
void OpenGLRainbows::DrawQuad(const float* pXYUV)
{
	glEnableClientState( GL_VERTEX_ARRAY );
	glVertexPointer( 2, GL_FLOAT, 4*sizeof(float), (char*)pXYUV);
	if(g_nTexture>=0)
	{
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, 4*sizeof(float), (char*)&pXYUV[2]);
 	}

	// Render
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	glDrawArrays(GL_TRIANGLE_FAN, 0, 4);	// Draw All Of The Triangles At Once

	// Disable Pointers
	glDisableClientState( GL_VERTEX_ARRAY );					// Disable Vertex Arrays
	if(g_nTexture>=0)
		glDisableClientState( GL_TEXTURE_COORD_ARRAY );			// Disable Texture Coord Arrays
}
#endif

//---------------------------------------------------------------//
// Draw the vertex array prepared before (3d text / Billboard)
//---------------------------------------------------------------//
void OpenGLRainbows::DrawVertexArray(unsigned int iMode)
{
#ifdef TRENKI3D
//	Trenkti3d();
// #NEED
#else
	if(m_VertexArray.Size())
	{
		glEnableClientState( GL_VERTEX_ARRAY );
		glVertexPointer( 3, GL_FLOAT, sizeof(OpenGLRainbows::VERTEX), (char*)&m_VertexArray[0].x);
		if(g_nTexture>=0)
		{
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, sizeof(OpenGLRainbows::VERTEX), (char*)&m_VertexArray[0].u);
 		}
		// Render GL_TRIANGLES
		glDrawArrays(iMode, 0, m_VertexArray.Size());	// Draw All Of The Triangles At Once
		// Disable Pointers
		glDisableClientState( GL_VERTEX_ARRAY );					// Disable Vertex Arrays
		glDisableClientState( GL_TEXTURE_COORD_ARRAY );				// Disable Texture Coord Arrays
	}
#endif
}




//---------------------------------------------------------------//
// Set Texture for 3D Objects
//---------------------------------------------------------------//
#ifdef TRENKI3D
void OpenGLRainbows::SetTexture3D(int num, int num_multi, int usemipmapping)
{
	g_nTexture = num;
	if(num>=0)
	{
		tglBindTexture(   m_Sprite[g_nTexture].Pixels(), m_Sprite[g_nTexture].m_Width, m_Sprite[g_nTexture].m_Height);
		tglSetCookieColor(m_Sprite[g_nTexture].m_colorkey, m_Sprite[g_nTexture].m_CookieInside);
	}
	else
	{
		tglBindTexture(NULL,0,0);
	}
}
#else
void OpenGLRainbows::SetTexture3D(int num, int num_multi, int usemipmapping, GLuint* pTextures)
{
/* polyvector - all white?
static int o_num=-99, o_num_multi=-99, o_usemipmapping=-99;
static GLuint* po_Textures=NULL;
	if(o_num==num && num_multi==o_num_multi && usemipmapping==o_usemipmapping && pTextures==po_Textures)
		return;
	o_num=num; num_multi=o_num_multi; usemipmapping=o_usemipmapping; pTextures=po_Textures;
*/

	g_nTexture = num;
	usemipmapping = usemipmapping ? 1 : 0;
	#if defined(TINYGL) || defined(TRENKI3D)
		usemipmapping=0;
	#endif
	m_Texture3D[0] = num;
	m_Texture3D[1] = num_multi;
	m_Texture3D[2] = usemipmapping;

	#if defined(TINYGL)
	m_Texture3D[2] = 0;
	if (num >=0 && m_Sprite[num].good() && m_Sprite[num].m_glTextureId==0)
	{
		unsigned int tx = m_Sprite[num].MakeGLTexture();
		m_Texture[num].mipstate[0] = tx;
	}
	#elif defined(TRENKI3D)
		trenki_set_texture
	#endif
	
#if defined(HAVE_OPENGL)
	if (glActiveTextureArb) // Can do MultiTexturing
	{
		glActiveTextureArb(GL_TEXTURE1_ARB);
		if (num_multi>=0 && num>=0) // Set both? Enable 2nd texture
		{
			#ifdef _DEBUG
			m_Texture3D_id_for_chanel[1]=num;
			#endif
			if(pTextures)
			{
				glBindTexture(GL_TEXTURE_2D, pTextures[num]);
			}
			else
			{
				glBindTexture(GL_TEXTURE_2D, this->m_Texture[num].mipstate[usemipmapping]);
			}
			glEnable(GL_TEXTURE_2D);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		}
		else
		{
			#ifdef _DEBUG
			m_Texture3D_id_for_chanel[1]=-1;
			#endif
//			glBindTexture(GL_TEXTURE_2D, 1);
			glDisable(GL_TEXTURE_2D);
		}

		glActiveTextureArb(GL_TEXTURE0_ARB);
		if (num>=0) // Set image texture
		{
			if (num_multi>=0)
			{
				if(pTextures)
				{
					glBindTexture(GL_TEXTURE_2D, pTextures[num_multi]);
				}
				else
				{
					glBindTexture(GL_TEXTURE_2D, m_Texture[num_multi].mipstate[usemipmapping]);
				}
				#ifdef _DEBUG
				m_Texture3D_id_for_chanel[0]=num_multi;
				#endif
			}
			else
			{
				if(pTextures)
				{
					glBindTexture(GL_TEXTURE_2D, pTextures[num]);
				}
				else
				{
					glBindTexture(GL_TEXTURE_2D, m_Texture[num].mipstate[usemipmapping]);
				}
				#ifdef _DEBUG
				m_Texture3D_id_for_chanel[0]=num;
				#endif
			}
			glEnable(GL_TEXTURE_2D);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		}
		else
		{
			#ifdef _DEBUG
			m_Texture3D_id_for_chanel[0]=-1;
			#endif
//			glBindTexture(GL_TEXTURE_2D, 1);
			glDisable(GL_TEXTURE_2D);
		}
		// Active Texture = 0 (Std. Texture)
	}
	else // vanilla opengl
#endif
	{
		if (num>=0)
		{
			if(pTextures)
			{
				glBindTexture(GL_TEXTURE_2D, pTextures[num]);
			}
			else
			{
				glBindTexture(GL_TEXTURE_2D, m_Texture[num].mipstate[usemipmapping]);
			}
			glEnable(GL_TEXTURE_2D);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		}
		else
		{
			glDisable(GL_TEXTURE_2D);
			glbEnable(GL_COLOR_MATERIAL);
		}
	}
}
#endif


///////////////////////////////////////////////////////////////////
//            S P R I T E - O P E R A T I O N S
///////////////////////////////////////////////////////////////////

//---------------------------------------------------------------//
//---------------------------------------------------------------//
void OpenGLRainbows::DrawSprite(int num, int x, int y)
{
#if defined(TINYGL) || defined(TRENKI3D)
	m_Sprite[num].Draw((int)x,(int)y);
#else
	const GLR_TEXTURE& tex = this->m_Texture[num];
	SetTexture3D(num, -1, 0);
	
	ApplySmoothShading();
	const float pos[] = {	
					(float)(x+tex.bw), (float)y, (float)tex.txx, (float)tex.txy,
					(float)x, (float)y, 0.0f, (float)tex.txy,
					(float)x,(float)(y+tex.bh), 0.0f, 0.0f,
					(float)(x+tex.bw), (float)(y+tex.bh), (float)tex.txx, 0.0f
					};
	DrawQuad(pos);
#endif
}


//---------------------------------------------------------------//
//---------------------------------------------------------------//
void OpenGLRainbows::StretchSprite(int num, int x, int y, int wx, int wy)
{
#if defined(TINYGL) || defined(TRENKI3D)
	m_Sprite[num].Zoom((int)x,(int)y,(int)(x+wx), (int)(y+wy));
#else
static int	sx, sy;
	const GLR_TEXTURE& tex = this->m_Texture[num];
	sx=((wx+m_Texture[num].w)/2);
	sy=((wy+m_Texture[num].h)/2);
	SetTexture3D(num, -1, 0);

	const float pos[] = {	(float)(x),    (float)(y+wy), 0.0f,    0.0f,
							(float)(x+wx), (float)(y+wy), tex.txx, 0.0f,
							(float)(x+wx), (float)(y),    tex.txx, tex.txy,
							(float)(x),    (float)(y),    0.0f,    tex.txy};
	ApplySmoothShading();
	DrawQuad(pos);
#endif
}

//---------------------------------------------------------------//
void OpenGLRainbows::RotoSprite(int num, int x, int y, float angle, float scale)
//---------------------------------------------------------------//
{
#if defined(TINYGL) || defined(TRENKI3D)
	static oBitmap bmp;
	m_Sprite[num].CreateRotatation(angle, bmp);
	if(scale==1.0f)
	{
		bmp.Draw((int)x - (bmp.m_Width-m_Sprite[num].m_Width)/2, (int)y - (bmp.m_Height-m_Sprite[num].m_Height)/2);
	}
	else
	{
		int xsize=(int)fabs(scale*bmp.m_Width );
		int ysize=(int)fabs(scale*bmp.m_Height);

		x-=(xsize-m_Sprite[num].m_Width)/2;
		y-=(ysize-m_Sprite[num].m_Height)/2;
		bmp.Zoom((int)x, (int)y, (int)x+(int)xsize, (int)y+(int)ysize);
	}

#else

static float	sx, sy;
float txx, txy;
float t0x, t0y;
	txx=((float)this->m_Texture[num].bw-.5f)/(float)this->m_Texture[num].w;
	txy=((float)this->m_Texture[num].bh-.5f)/(float)this->m_Texture[num].h;
	t0x = 0.5f/(float)m_Texture[num].w;
	t0y = 0.5f/(float)m_Texture[num].h;

	sx=scale*m_Texture[num].bw/2.0f;
	sy=scale*m_Texture[num].bh/2.0f;
	SetTexture3D(num, -1, 0);
	glPushMatrix(); // Due to Rotation save the current matrix
		glTranslatef((GLfloat)(x+(m_Texture[num].bw>>1)), (GLfloat)(y+(m_Texture[num].bh>>1)), 0); // Move to Origin of Sprite
		glRotatef(angle, 0.0f, 0.0f, -1.0f); // Rotate around origin

		float pos[] = {
						-sx, sy, t0x, t0y,
						sx,  sy, txx, t0y, 
						sx, -sy, txx, txy,
						-sx,-sy, t0x, txy
						};
		ApplySmoothShading();
		DrawQuad(pos);
	glPopMatrix(); // Restore the current matrix
#endif
}

//---------------------------------------------------------------//
// Set Tiling sizes for a texture
//---------------------------------------------------------------//
void OpenGLRainbows::TileSprite(int iSprite, int width, int height)
{
	m_Texture[iSprite].tw = (short)width;
	m_Texture[iSprite].th = (short)height;
}


//---------------------------------------------------------------//
// Draw a Tile from a tile texture
//---------------------------------------------------------------//
void OpenGLRainbows::DrawTile(int num, int iTile, int x, int y)
{
	if(m_Texture[num].tw==0 || m_Texture[num].th==0) return;
	int nperline = m_Texture[num].bw / m_Texture[num].tw;
	int tilex, tiley=0;
	if (nperline==0)
	{
		tilex = iTile;
	}
	else
	{
		tilex = iTile % nperline;
		tiley = iTile / nperline;
	}
	

#if defined(TINYGL) || defined(TRENKI3D)
	int cx = tilex*m_Texture[num].tw, cy = tiley*m_Texture[num].th;
	m_Sprite[num].DrawClip((int)x,(int)y, cx,cy, m_Texture[num].tw+cx, m_Texture[num].th+cy);
#else
	// texture coord width/height for one tile
	float onepix_w = 1.0f / (float)m_Texture[num].w;
	float onepix_h = 1.0f / (float)m_Texture[num].h;

	float txx, txy;
	// width of a tile
	txx=(float)m_Texture[num].tw/(float)m_Texture[num].w;
	txy=(float)m_Texture[num].th/(float)m_Texture[num].h;

	// pixel value of upper point
	tiley = m_Texture[num].bh-m_Texture[num].th*(tiley+1);
	// true texture coords
#if 0
	float tx1 = 0.5f*onepix_w + txx*(float)tilex;
	float ty1 = 0.5f*onepix_h + (float)tiley / m_Texture[num].h;
	float tx2 = tx1+txx-onepix_w;
	float ty2 = ty1+txy-onepix_h;
#else
	float tx1 = txx*(float)tilex;
	float ty1 = (float)tiley / m_Texture[num].h;
	float tx2 = tx1+txx;
	float ty2 = ty1+txy;
#endif

	SetTexture3D(num, -1, 0);
	
	ApplySmoothShading();
	#if defined(HAVE_OPENGL)
	glBegin(GL_QUADS);
	glTexCoord2f(tx2, ty2); glVertex2i(x+m_Texture[num].tw, y);
	glTexCoord2f(tx1, ty2); glVertex2i(x, y);
	glTexCoord2f(tx1, ty1); glVertex2i(x,y+m_Texture[num].th);
	glTexCoord2f(tx2, ty1); glVertex2i(x+m_Texture[num].tw, y+m_Texture[num].th);
	glEnd();
	#endif
	
	#if defined(HAVE_OPENGL_ES)
	float xyuv[] ={	(float)(x+m_Texture[num].tw), (float)y,                     (float)tx2, (float)ty2,
					(float)x, (float)y,                                         (float)tx1, (float)ty2,
					(float)x, (float)(y+m_Texture[num].th),                     (float)tx1, (float)ty1,
					(float)(x+m_Texture[num].tw), (float)(y+m_Texture[num].th), (float)tx2, (float)ty1};
	DrawQuad(xyuv);
 
	#endif

#endif
}


//---------------------------------------------------------------//
// Rotozoom a tile
//---------------------------------------------------------------//
void OpenGLRainbows::RotoZoomTile(int num, int iTile, int x, int y, float phi, float scale)
{
	if(m_Texture[num].tw==0 || m_Texture[num].th==0) return;

	int nperline = m_Texture[num].bw / m_Texture[num].tw;
	int tilex, tiley=0;
	if (nperline==0)
	{
		tilex = iTile;
	}
	else
	{
		tilex = iTile % nperline;
		tiley = iTile / nperline;
	}
	// texture coord width/height for one tile
	float onepix_w = 1.0f / (float)m_Texture[num].w;
	float onepix_h = 1.0f / (float)m_Texture[num].h;

	float txx, txy;
	// width of a tile
	txx=(float)m_Texture[num].tw/(float)m_Texture[num].w;
	txy=(float)m_Texture[num].th/(float)m_Texture[num].h;

	// pixel value of upper point
	tiley = m_Texture[num].bh-m_Texture[num].th*(tiley+1);
	// true texture coords

#if 1
	// I had that uncommented, but it seems to work really nice with a 67x18 bmp, 16x16 tiled.
	// Same for 15x15 tiled. Smooth shading true or false.
	// Only if you scale 1x and no rotation, it looks "blurry". (smoothshading FALSE helps, though)
	float tx1 = 0.5f*onepix_w + txx*(float)tilex;
	float ty1 = 0.5f*onepix_h + (float)tiley / m_Texture[num].h;
	float tx2 = tx1+txx-onepix_w;
	float ty2 = ty1+txy-onepix_h;
#else
	float tx1 = txx*(float)tilex;
	float ty1 = (float)tiley / m_Texture[num].h;
	float tx2 = tx1+txx;
	float ty2 = ty1+txy;
#endif



float sphi,cphi;
float dx,dy;

	scale*=0.5f; // half width

	dx=m_Texture[num].tw*scale;
	dy=m_Texture[num].th*scale;
	x+=(m_Texture[num].tw/2);
	y+=(m_Texture[num].th/2);

	sphi = sinf(phi * BOG_INV_F);
	cphi = cosf(phi * BOG_INV_F);

	SetTexture3D(num, -1, 0);
	float xyuv[] ={
					((float)x- cphi*dx-sphi*dy), ((float)y-cphi*dy+sphi*dx), tx1,ty2,
					((float)x- cphi*dx+sphi*dy), ((float)y+cphi*dy+sphi*dx), tx1,ty1,
					((float)x+ cphi*dx+sphi*dy), ((float)y+cphi*dy-sphi*dx), tx2,ty1,
					((float)x+ cphi*dx-sphi*dy), ((float)y-cphi*dy-sphi*dx), tx2,ty2};

#ifdef TRENKI3D
// void tglDrawArraysXYZ_UV_RGBA(const int type, const int count, const float* pData, const int hasZ, const int hasNormal, const int UseUV, const int UseRGBA, const int hasRGBA);
	tglDrawArraysXYZ_UV_RGBA(GL_TRIANGLE_FAN, 4, xyuv,                                0,              0,                   1,               0,                 0);
#elif defined(TINYGL)

#else
	ApplySmoothShading();
	DrawQuad(xyuv);
#endif
}

void OpenGLRainbows::StretchTile(int num, int iTile, int x, int y, int w, int h)
{
	if(m_Texture[num].tw==0 || m_Texture[num].th==0) return;

	int nperline = m_Texture[num].bw / m_Texture[num].tw;
	int tilex, tiley=0;
	if (nperline==0)
	{
		tilex = iTile;
	}
	else
	{
		tilex = iTile % nperline;
		tiley = iTile / nperline;
	}
	// texture coord width/height for one tile
	float txx, txy;
	txx=(float)m_Texture[num].tw/(float)(m_Texture[num].w);
	txy=(float)m_Texture[num].th/(float)(m_Texture[num].h);
	// pixel value of upper point
	tiley = m_Texture[num].bh-m_Texture[num].th*(tiley+1);
	// true texture coords
	float ty1 = (float)tiley / m_Texture[num].h;
	float tx1 = txx*(float)tilex;
	float tx2 = tx1+txx;
	float ty2 = ty1+txy;

	float fx=(float)x;
	float fy=(float)y;
	float fh=(float)h;
	float fw=(float)w;

	float xyuv[] ={
					fx      ,  fy       , tx1,ty2,
					fx      ,  fy+fh+.1f, tx1,ty1,
					fx+fw+.1f, fy+fh+.1f, tx2,ty1,
					fx+fw+.1f, fy       , tx2,ty2
					};
	SetTexture3D(num, -1, 0);
#ifdef TRENKI3D
// void tglDrawArraysXYZ_UV_RGBA(const int type, const int count, const float* pData, const int hasZ, const int hasNormal, const int UseUV, const int UseRGBA, const int hasRGBA);
	tglDrawArraysXYZ_UV_RGBA(GL_TRIANGLE_FAN, 4, xyuv, 0, 0, 1, 0, 0);
#elif defined(TINYGL)

#else
	ApplySmoothShading();
	DrawQuad(xyuv);
#endif
}


//---------------------------------------------------------------//
// return the viewport with respect to rotations
//---------------------------------------------------------------//
const int* OpenGLRainbows::RotatedViewport() const
{
static int vp[4];
	vp[0]=m_WND.viewport.x;
	vp[1]=m_WND.viewport.y;
	vp[2]=m_WND.viewport.width;
	vp[3]=m_WND.viewport.height;
	return vp;
}


//---------------------------------------------------------------//
// Draw 3D Text with constant height and z-buffer
//---------------------------------------------------------------//
void OpenGLRainbows::Text3D(int nFont, const char* szText, float x, float y, float z, float scale, int kerning)
{
	if(m_ttf_fonts[nFont].IsLoaded())
	{
		TextTTF(nFont, szText, x,y,kerning);
		return;
	}

#ifndef TRENKI3D
static float winx, winy, winz;

//static int viewport[4];
static float in[4], out[4];
bool culling = false;
bool lighting=false;

	if(glIsEnabled(GL_CULL_FACE))
	{
		culling = true;
		glbDisable(GL_CULL_FACE);
	}
	if(glIsEnabled(GL_LIGHTING))
	{
		lighting=true;
		glbDisable(GL_LIGHTING);
	}
	// first load the matrices for the project call.
	// glGetFloatv( GL_MODELVIEW_MATRIX,  model);
	const float* model = m_pRbow3D->m_ModelViewCameraMatrix;

	// Font stays constant size - good readability
	if(scale == 0.0f)
	{
		// glGetFloatv  ( GL_PROJECTION_MATRIX, proj );
		const float* proj = m_pRbow3D->m_ProjectionMatrix;
		//glGetIntegerv( GL_VIEWPORT,         viewport );
		const int* viewport = RotatedViewport();

		in[0]=x; in[1]=y; in[2]=z; in[3]=1.0;

		transform_point(out,model,in);
		transform_point(in,proj,out);

		if (in[3]==0.0f) return;

		// Normalize between -1 and 1 :
		in[0]/=in[3];in[1]/=in[3];in[2]/=in[3];

		// In screen coordinates :
		winx = viewport[0]+(1+in[0])*viewport[2]/2;
		winy = viewport[3] - (viewport[1]+(1+in[1])*viewport[3]/2); // GLBasic origin = top, (+) goes down
		winz = -in[2]; // [-1; +1]

//		glViewport(m_WND.viewport.x, m_WND.height-m_WND.viewport.height-m_WND.viewport.y, m_WND.viewport.width, m_WND.viewport.height);
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0.0f, (float)m_WND.viewport.width, (float)m_WND.viewport.height, 0.0f, -1.0f, 1.0f); // Move to upper left corner 0,0

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		// alignment = bottom center
		float len = (float)strlen(szText);
		Text2D(nFont, szText,	winx-(float)(m_Texture[(GLR_MAX_TEXTURES - GLR_MAX_FONT)].bw>>4)*len*.5f-viewport[0],
								winy-(float)(m_Texture[(GLR_MAX_TEXTURES - GLR_MAX_FONT)].bh>>3)+viewport[1],
								winz, kerning, FALSE); //winx, winy, winz);
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();	
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}
	else // font will be scaled - billboard sprite
	{
		float up[3], right[3];
		right[0] = model[0];
		right[1] = model[4];
		right[2] = model[8];

		up[0] = model[1];
		up[1] = model[5];
		up[2] = model[9];
	float		tx, ty, txs, tys, cx, cy;
	unsigned int		asc;

		nFont = nFont + GLR_MAX_TEXTURES - GLR_MAX_FONT;
		SetTexture3D(nFont, -1, 0);

		int			charset=128, yoff=7;

		txs= (m_Texture[nFont].txx) / 16.0f;
		cx = (float)(m_Texture[nFont].bw>>4);
		if(m_Texture[nFont].bw<m_Texture[nFont].bh)
		{
			yoff=15;
			charset=256; // FONT_CHARSET
			tys= (m_Texture[nFont].txy) / 16.0f;
			cy = (float)(m_Texture[nFont].bh>>4);
		}
		else
		{
			tys= (m_Texture[nFont].txy) /  8.0f;
			cy = (float)(m_Texture[nFont].bh>>3);
		}
		gMultVertex(up, cx*scale);
		// gMultVertex(right, scale); // *cx or pixels for kering
		// 4 3
		// 1 2
		float p1[3], p2[3], p3[3], p4[3];

		
		p1[0] = x; p1[1]=y; p1[2]=z;
		// alignment: bottom center
		unsigned int str_l = (unsigned int)strlen(szText);
		
		#if defined(HAVE_OPENGL_ES) // quick workaround - dirty!
			m_VertexArray.SetSize(4);
		#else
			m_VertexArray.SetSize(str_l*4);
		#endif
		OpenGLRainbows::VERTEX* pV = &m_VertexArray[0];

		// 3D
		if(kerning)
		{
			const GLR_FONTKERNING* pKern = m_FontKerning[nFont];
			gMultVertex(right, scale); // * pixels for kering
			gAddVertex(p1, up);
			float len = 0.0f;
			for(unsigned int i=0; i<str_l; ++i)
				len += (float) (pKern[(unsigned int)(szText[i]%charset) ].width);
			len *= 0.5f; // str_l/2.0f;
			p1[0]-=(len*right[0]);
			p1[1]-=(len*right[1]);
			p1[2]-=(len*right[2]);
			
			// calc p2,p3,p4
			gSubVertex(p1, up); gCopyVertex(p2, p1); gAddVertex(p2, right); gCopyVertex(p3, p2);
			gAddVertex(p3, up); gDiffVertex(p4, p3, right);
			// 4 3
			// 1 2
			for (int i=0; szText[i]!='\0'; i++)
			{
				asc=(unsigned)szText[i]%charset;	// Ignore Extended AscII Format - It's Just Shit like 
				const GLR_FONTKERNING& K = pKern[asc];
				tx = (asc%16)*txs;
				ty = tys * (yoff - (asc/16));

				gCopyVertex(p3,p4);
				gCopyVertex(p2,p1);
				// use "up" as a temp for right*width
				gCopyVertex(up, right);
				gMultVertex(up, K.width);
				gAddVertex(p3, up);
				gAddVertex(p2, up);
				// Top left
				(pV++)->assign(p4, K.txa, ty+tys);
				// Bottom left
				(pV++)->assign(p1, K.txa, ty);
				// Top right
				(pV++)->assign(p3, K.txe, ty+tys);
				// Bottom right
				(pV++)->assign(p2, K.txe, ty);

				// Destination
				gCopyVertex(p1, p2);
				gCopyVertex(p4, p3);
			}
		}
		else
		{
			gMultVertex(right, cx*scale); // prepare right vector with fixed font size
			gAddVertex(p1, up);
			float len = (float)str_l/2.0f;
			p1[0]-=(len*right[0]);
			p1[1]-=(len*right[1]);
			p1[2]-=(len*right[2]);
		
			// calc p2,p3,p4
			gSubVertex(p1, up); gCopyVertex(p2, p1); gAddVertex(p2, right); gCopyVertex(p3, p2);
			gAddVertex(p3, up); gDiffVertex(p4, p3, right);
			for (int i=0; szText[i]!='\0'; i++)
			{
				asc=(unsigned)szText[i]%charset;	// Ignore Extended AscII Format - It's Just Shit like 
				tx = (asc%16)*txs;
				ty = tys * (yoff - (asc/16));

				// Top left
				(pV++)->assign(p4, tx, ty+tys);
				// Bottom left
				(pV++)->assign(p1, tx, ty);
				// Top right
				(pV++)->assign(p3, tx+txs, ty+tys);
				// Bottom right
				(pV++)->assign(p2, tx+txs, ty);

				// Destination
				gAddVertex(p1, right);
				gAddVertex(p2, right);
				gAddVertex(p3, right);
				gAddVertex(p4, right);
			}
		}
		DrawVertexArray(GL_TRIANGLE_STRIP);
	}
	if(culling) glbEnable(GL_CULL_FACE);
	if(lighting) glbEnable(GL_LIGHTING);
#endif
}

//---------------------------------------------------------------//
// Draw ttf font
//---------------------------------------------------------------//
void OpenGLRainbows::TextTTF(int nFont, const char * szText, float x, float y, int kerning)
{
	PolyStart(-1, 1);
	float sc = m_FontScaling[nFont];
	const TTF_FONT& font = m_ttf_fonts[nFont];


	// Line overlay with smooth lines
	#if defined(HAVE_OPENGL)
	if(m_SmoothShading)
	{
		glEnable(GL_LINE_SMOOTH);
		glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

		float orig_x=x;
		const char* pOrigTx=szText;
		for(; *szText!='\0'; szText++)
		{
			int c = (int)(unsigned char)(*szText);
			if(c < 33)
			{
				x += font.chars[46].width() * sc; // spaces = 1.1* width of period
				continue;
			}

			const FONTCHAR& fc = font.chars[c];
			for(int it = 0; it < fc.tri.Size(); ++it)
			{
				const FONTCHAR::FONTTRI& tr = fc.tri[it];
				const FONTCHAR::FONTVEC* pV = &fc.vec[tr.i];
				const FONTCHAR::FONTVEC* pV2 = &fc.vec[tr.j];
				DrawLinef(x+pV->x*sc, y+pV->y*sc, x+pV2->x*sc, y+pV2->y*sc, m_TextColor);
				pV = &fc.vec[tr.k];
				DrawLinef(x+pV->x*sc, y+pV->y*sc, x+pV2->x*sc, y+pV2->y*sc, m_TextColor);
				pV2 = &fc.vec[tr.i];
				DrawLinef(x+pV->x*sc, y+pV->y*sc, x+pV2->x*sc, y+pV2->y*sc, m_TextColor);
			}
			if(kerning)
				x += fc.width() * sc;
			else
				x += font.max_char_width * sc;
		}
		glDisable(GL_LINE_SMOOTH);
		x=orig_x;
		szText=pOrigTx;
	}


	#endif

	for(; *szText!='\0'; szText++)
	{
		int c = (int)(unsigned char)(*szText);
		if(c < 33)
		{
			x += font.chars[46].width() * sc; // spaces = 1.1* width of period
			continue;
		}

		const FONTCHAR& fc = font.chars[c];
		for(int it = 0; it < fc.tri.Size(); ++it)
		{
			const FONTCHAR::FONTTRI& tr = fc.tri[it];
			const FONTCHAR::FONTVEC* pV = &fc.vec[tr.i];
			PolyPoint(x+pV->x*sc, y+pV->y*sc, 0,0, m_TextColor);
			pV = &fc.vec[tr.j];
			PolyPoint(x+pV->x*sc, y+pV->y*sc, 0,0, m_TextColor);
			pV = &fc.vec[tr.k];
			PolyPoint(x+pV->x*sc, y+pV->y*sc, 0,0, m_TextColor);
		}
		if(kerning)
			x += fc.width() * sc;
		else
			x += font.max_char_width * sc;
	}
	PolyEnd();
}





//---------------------------------------------------------------//
// Draw a billboard sprite
//---------------------------------------------------------------//
void OpenGLRainbows::DrawBillboard(int num, float x, float y, float z, float scale)
{
#ifndef TRENKI3D
float up[3], right[3], normal[3], modelview[16], a[3], b[3], c[3], d[3], center[3];
// #NEED Get face culling and set to 1 - fixed in glb3d.cpp
	glGetFloatv(GL_MODELVIEW_MATRIX, modelview);

	right[0] = modelview[0];
	right[1] = modelview[4];
	right[2] = modelview[8];

	up[0] = modelview[1];
	up[1] = modelview[5];
	up[2] = modelview[9];

	normal[0] = modelview[ 2];
	normal[1] = modelview[ 6];
	normal[2] = modelview[10];

	glNormal3fv(normal);

	center[0]=x; center[1]=y; center[2]=z;

/*
d-----c
|     |
a--c--b
   e
   nter
*/
	float bx, by;
	bx = scale * (m_Texture[num].bw >> 2);
	by = scale * (m_Texture[num].bh >> 1);
	for (char n=0; n<3; n++)
	{
		a[n] = center[n] - right[n] * bx;
		b[n] = center[n] + right[n] * bx;
		c[n] = center[n] + right[n] * bx + up[n] * by;
		d[n] = center[n] - right[n] * bx + up[n] * by;
	}

	float txx, txy;
	float pxo, pyo;
	pxo = 1.0f/(float)this->m_Texture[num].w;
	pyo = 1.0f/(float)this->m_Texture[num].h;
	txx=((float)this->m_Texture[num].bw-.5f)*pxo;
	txy=((float)this->m_Texture[num].bh-.5f)*pyo;
	pxo=0;//*=0.55f;
	pyo=0;//*=0.55f;

	SetTexture3D(num, -1, 0);

	m_VertexArray.SetSize(4);
	OpenGLRainbows::VERTEX* pV = &m_VertexArray[0];
	(pV++)->assign(a, pxo,pyo);
	(pV++)->assign(b, txx,pyo);
	(pV++)->assign(c, txx,txy);
	(pV++)->assign(d, pxo,txy);
	
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	DrawVertexArray(GL_TRIANGLE_FAN);
#endif
}


//---------------------------------------------------------------//
// width with kerning
//---------------------------------------------------------------//
int  OpenGLRainbows::Text2DWidth(int num_font, const char* text, int kerning)
{
	if(m_ttf_fonts[num_font].IsLoaded())
	{
		float x=0.0;
		float sc = m_FontScaling[num_font];
		const TTF_FONT& font = m_ttf_fonts[num_font];
		for(; *text!='\0'; text++)
		{
			int c = (int)(unsigned char)(*text);
			if(c < 33)
			{
				x += font.chars[46].width() * sc; // spaces = 1.1* width of period
				continue;
			}
			if(kerning)
				x += font.chars[c].width() * sc;
			else
				x += font.max_char_width * sc;
		}
		return int(x);
	}


	int width=0;
	if(kerning)
	{
		const GLR_FONTKERNING* pKern = m_FontKerning[num_font];
		for (int i=0; ; ++i)
		{
			unsigned char asc=((unsigned char)text[i]);
			if (asc==0) break;
			const GLR_FONTKERNING& K = pKern[asc];
			width+=K.width;
		}
		return (int)(width * m_FontScaling[num_font]);
	}
	else
	{
		const GLR_TEXTURE& txtr = m_Texture[num_font + GLR_MAX_TEXTURES - GLR_MAX_FONT];
		return (int)( (strlen(text) * (txtr.bw>>4)) * m_FontScaling[num_font]);
	}
}


//---------------------------------------------------------------//
// Draw text in 2D mode
//---------------------------------------------------------------//
void OpenGLRainbows::Text2D(int num_font, const char* text, float x, float y, float z, int kerning, int apply_scaling)
{
	if(m_ttf_fonts[num_font].IsLoaded())
	{
		TextTTF(num_font, text, x,y,kerning);
		return;
	}

#if defined(TINYGL) || defined(TRENKI3D)
	const GLR_FONTKERNING* pKern = m_FontKerning[num_font];
	m_Font[num_font].Print(text, (int)x,(int)y, kerning ? pKern:NULL);
#else
	unsigned int str_l = (unsigned int) strlen(text);
	if(!str_l) return;
	m_VertexArray.SetSize(str_l*4);
	OpenGLRainbows::VERTEX* pV = &m_VertexArray[0];
	
	float		tx, ty, txs, tys;
	int			i;
	float		cx, cy; // char width + height (no kerning)
	unsigned int asc;
	int			charset=128, yoff=7;
	int num_texture = num_font + GLR_MAX_TEXTURES - GLR_MAX_FONT;
	SetTexture3D(num_texture, -1, 0);
	
	
	const GLR_TEXTURE& txtr = m_Texture[num_texture];

	txs= (txtr.txx) / 16.0f;
	cx = (float)(txtr.bw>>4);
	if(txtr.bw<txtr.bh)
	{
		yoff=15;
		charset=256; // FONT_CHARSET
		tys= (txtr.txy) / 16.0f;
		cy = (float)(txtr.bh>>4);
	}
	else
	{
		tys= (txtr.txy) /  8.0f;
		cy = (float)(txtr.bh>>3);
	}

	x = ceilf(x);
	y = ceilf(y);

	float fscale = 1.0;
	if(apply_scaling)
	{
		fscale = m_FontScaling[num_font];
		cx*=fscale;
		cy*=fscale;
	}


	if(kerning)
	{
		const GLR_FONTKERNING* pKern = m_FontKerning[num_font];
		for (i=0; ; ++i)
		{
			asc=((unsigned)text[i]%charset);
			if (asc==0) break;
			ty = tys * (yoff - (asc/16));
			const GLR_FONTKERNING& K = pKern[asc];

			float kw = (float)K.width * fscale;

			// Top left
			(pV++)->assign(x,    y+cy, z,    K.txa,     ty);
			// Bottom left			      
			(pV++)->assign(x,    y,    z,    K.txa,     ty+tys);
			// Top right					      
			(pV++)->assign(x+kw, y+cy, z,    K.txe, ty);
			// Bottom right					      
			(pV++)->assign(x+kw, y,    z,    K.txe, ty+tys);

			// Destination
			x += kw; // no additional space here!! would need a new poly group
		}
	}
	else
	{
		for (i=0; ; ++i)
		{
			asc=((unsigned)text[i]%charset);
			if (asc==0) break;
			tx = txs * (asc%16);
			ty = tys * (yoff - (asc/16));
			// Top left
			(pV++)->assign(x,    y+cy, z,tx,     ty);
			// Bottom left
			(pV++)->assign(x,    y,    z,tx,     ty+tys);
			// Top right
			(pV++)->assign(x+cx, y+cy, z,tx+txs, ty);
			// Bottom right
			(pV++)->assign(x+cx, y,    z,tx+txs, ty+tys);


			// Destination
			x+=cx;
		}
	}
	DrawVertexArray(GL_TRIANGLE_STRIP);
#endif
}


///////////////////////////////////////////////////////////////////
//             P A S T E   S U R F A C E S
///////////////////////////////////////////////////////////////////


//---------------------------------------------------------------//
//---------------------------------------------------------------//
void OpenGLRainbows::PasteSurface(int num)
{
	#if defined(TINYGL) || defined(TRENKI3D)
		m_Surface[num].Draw(0,0);
	#else
#ifdef GLB_SURFACE_OLDSCHOOL
		if(m_Surface[num].usex == 0) return;

	float	sx=0, sy;
	int		y;


		// GF_TODO the prescaler is not taken care of, yet
		DoAlpha(0.0); // disable blending
		SetAlphaTesting(); // no cookie
		glbDisable(GL_ALPHA_TEST); // No Cookie
		glbDisable(GL_BLEND); // No blend

		// Multi Texturing allowed? Disable Texture 1 and Use Texture 0
		#if defined(HAVE_OPENGL)
		if (glActiveTextureArb)
		{
			glActiveTextureArb(GL_TEXTURE1_ARB);
			glDisable(GL_TEXTURE_2D);

			glActiveTextureArb(GL_TEXTURE0_ARB);
		}
		#endif
		glEnable(GL_TEXTURE_2D);

		g_nTexture = 0; // otherwise DrawQuad fails
		for (int x=0; x<m_Surface[num].usex; x++)
		{
			sy=0; //(float)this->m_Surface[num].sizy[0][0]; // 0-...
			for (y=0; y<m_Surface[num].usey; y++)
			{
				glBindTexture(GL_TEXTURE_2D, m_Surface[num].txt[x][y]);

				// hard interpolation
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

				float txx, txy;
				txx=(float)m_Surface[num].sizbx[x][y]/(float)m_Surface[num].sizx[x][y];
				txy=(float)m_Surface[num].sizby[x][y]/(float)m_Surface[num].sizy[x][y];
				float xyuv[] = {sx, sy+m_Surface[num].sizby[x][y],                           0.0f, 0.0f,
								sx+m_Surface[num].sizbx[x][y], sy+m_Surface[num].sizby[x][y], txx, 0.0f,
								sx+m_Surface[num].sizbx[x][y], sy,                            txx, txy ,
								sx, sy,                                                      0.0f, txy  };
				DrawQuad(xyuv);
				sy+=m_Surface[num].sizy[x][y+1];
			}
			sx+=m_Surface[num].sizx[x][y-1];
		}
		SetTexture3D(-1, -1, 0);
		ApplySmoothShading(); // restore old state

		DoAlpha(0.0f); // might enable blending if no cookie

		// glbEnable(GL_ALPHA_TEST);
#else // just use a sprite
	if(__g_Prescaler.scale> 1.0f)
	{
		// clear, because stuff outside the viewport might flicker.
		// GF_TODO how can you draw outside the viewport!?!?! Must fix that!
		#if defined(TINYGL) || defined(TRENKI3D)
				Screen.Clear();
		#else // GL or EGL
				glClear(GL_COLOR_BUFFER_BIT);
		#endif
	}

	DrawSprite(m_BackSurfaceSpriteID[num], 0,0);
#endif
#endif
}

// --------------------------------------------------------------- //
// Set Cookie Colors
// --------------------------------------------------------------- //
void OpenGLRainbows::SetColorKey(COLORREF col)
{
	m_CookieFont = m_CookieColor = col;
}



// --------------------------------------------------------------- //
// Load Font
// --------------------------------------------------------------- //
bool OpenGLRainbows::LoadFontPIX(int num, const char* szFileName)
{
	m_ttf_fonts[num].Clear();
	m_FontScaling[num] = 1.0f;

	OPENGL_IMAGE image;
	if(!image.LoadImageBMP(szFileName, m_CookieFont) && !image.LoadImagePNG(szFileName, m_CookieFont))
	{
		// not for fonts
		// if(!img.m_bSourceHadAlpha) img.FixNeighbourAlpha();

		#if defined(TINYGL) || defined(TRENKI3D)
			m_Font[num].Delete();
		#else
			FreeSprite(num+GLR_MAX_TEXTURES-GLR_MAX_FONT);
		#endif
		return false;
	}


	// Create 2^n size
	int sx=GLB_MIN_TEX_SIZE/2; int sy=GLB_MIN_TEX_SIZE/2;
	do{sx<<=1;} while(sx < image.m_sizex);
	do{sy<<=1;} while(sy < image.m_sizey);
float txw=(float)sx, txh=(float)sy; // texture dimensions


 	// build kerning table
	int charset=128;
	int cy, cx = image.m_sizex >> 4;
	if(image.m_sizex < image.m_sizey)// FONT_CHARSET
	{
		charset=256; // FONT_CHARSET
		cy = image.m_sizey >> 4;
	}
	else
	{
		cy = image.m_sizey >> 3;
	}
	
	
	for(int ic = 0; ic<charset; ++ic)
	{
		int xoff=0, width=0;
		int x = ic % 16;
		int y = ic / 16;
		GLR_FONTKERNING& K = m_FontKerning[num][ic];
				
		K.width = 0;
		K.xoff  = 0;
		for(int px=0; px<cx; ++px)
		for(int py=0; py<cy; ++py)
		{
			int alpha = image[(x*cx + px + (y*cy+py)*image.m_sizex)*4 + 3];
			if(alpha>0)
			{
				xoff = px;
				goto rightnow;
			}
		}
		rightnow:
		for(int px=cx-1; px>=xoff; --px)
		for(int py=0; py<cy; ++py)
		{
			int alpha = image[(x*cx + px + (y*cy+py)*image.m_sizex)*4 + 3];
			if(alpha>0)
			{
				width = px - xoff + 1;
				goto stopnow;
			}
		}
		stopnow:

		/*
		printf("\n\nCharacter: %d[%c].\nxoff=%d width=%d\n", ic, ic, xoff, width);
		for(int py=0; py<cy; ++py)
		{
			for(int px=0; px<cx; ++px)
			{
						int alpha = image[(x*cx + px + (y*cy+py)*image.m_sizex)*4 + 3];
				printf("%c", alpha ? '#':' ');

			}
		printf("\n");
		}
		*/

		if(width<=0)
		{
			K.xoff =0;
			K.width=cx;
		}
		else
		{
			if(width+xoff < cx) ++width; // nice - we have one pixel for spacing available
			K.xoff =xoff;
			K.width=width;
		}
		K.txa = (float)(xoff       + x*cx) / txw;
		K.txe = (float)(xoff+width + x*cx) / txw;
		
		// fix space - make width of "j"
		if (ic == (int)'j')
		{
			GLR_FONTKERNING& spc = m_FontKerning[num][32];

			spc.width = K.width;
			spc.txa = 0.0f;
			spc.txe = K.txe - K.txa;
		}
		
		if(charset<256)
		{
			m_FontKerning[num][ic+128] = K;
		}
	}

#if defined(TINYGL) || defined(TRENKI3D)
	return m_Font[num].CreateBy(image);
#else
	return LoadSprite(num+GLR_MAX_TEXTURES-GLR_MAX_FONT, image, m_CookieFont, false); // Load a Texture as a Sprite
#endif
}

bool OpenGLRainbows::LoadFontTTF(int num, const short * packed_ttf)
{
	m_ttf_fonts[num].Clear();
	m_FontScaling[num] = 1.0f;

	OPENGL_IMAGE image;
#if defined(TINYGL) || defined(TRENKI3D)
		m_Font[num].Delete();
#else
		FreeSprite(num+GLR_MAX_TEXTURES-GLR_MAX_FONT);
#endif
	BuildFontTTF(packed_ttf, m_ttf_fonts[num]);
	return true;
}



// --------------------------------------------------------------- //
// Load Sprite from file
// --------------------------------------------------------------- //
bool OpenGLRainbows::LoadSprite(int num, const char* szFileName, COLORREF colorkey, bool IsBumpHeightmap)
{
	int inscreen = -1;
	for(int i=0; i<GLR_MAX_FRAMEBUFFER; ++i)
	{
		if(m_FBuffer[i].m_SpriteIndex == num)
		{
			inscreen = i; break;
		}
	}
	
	FreeSprite(num);
	if(szFileName==NULL || szFileName[0]=='\0')
		return false;

	OPENGL_IMAGE image;
	if(!image.LoadImageBMP(szFileName, colorkey) && !image.LoadImagePNG(szFileName, colorkey) && !image.LoadImageJPG(szFileName))
	{
		return false;
	}
	if(!image.m_bSourceHadAlpha)
		image.FixNeighbourAlpha();
	bool ok = LoadSprite(num, image, colorkey, IsBumpHeightmap); // Load a Texture as a Sprite

	if(ok && inscreen>=0)
	{
		if(m_FBuffer[inscreen].Create(m_Texture[num].w, m_Texture[num].h, m_Texture[num].mipstate[0], true))
			m_FBuffer[inscreen].m_SpriteIndex = num;
	}
	return ok;
}

// --------------------------------------------------------------- //
// Load Sprite from memory
// --------------------------------------------------------------- //
bool OpenGLRainbows::LoadSprite(int num, OPENGL_IMAGE& image, COLORREF colorkey, bool IsBumpHeightmap) // Load a Texture as a Sprite
{
	int inscreen = -1;
	for(int i=0; i<GLR_MAX_FRAMEBUFFER; ++i)
	{
		if(m_FBuffer[i].m_SpriteIndex == num)
		{
			inscreen = i; break;
		}
	}

	FreeSprite(num);
	if (image.m_sizex==0 || image.m_sizey==0) return false;

#if defined(TINYGL) || defined(TRENKI3D)
	if(!m_Sprite[num].CreateBy(image, colorkey))
		return false;
	this->m_Texture[num].w =(short)image.m_sizex;
	this->m_Texture[num].h =(short)image.m_sizey;
	this->m_Texture[num].bw=(short)image.m_sizex;
	this->m_Texture[num].bh=(short)image.m_sizey;

	this->m_Texture[num].txx=1.0f;
	this->m_Texture[num].txy=1.0f;
	return true;
#else

	int		sx, sy;
	// Create 2^n size - MUST EQUAL TO MakeTexture!!
	sx=GLB_MIN_TEX_SIZE/2; sy=GLB_MIN_TEX_SIZE/2;
	do{sx<<=1;} while(sx < image.m_sizex);
	do{sy<<=1;} while(sy < image.m_sizey);

	// file, startx, starty, sizex, sizey, &num
	// #NPBUFFER
	bool ok = true;
	if(num>=GLR_MAX_TEXTURES-GLR_MAX_FONT) // is font
		ok = MakeTexture(image, 0, 0, 0, 0, &m_Texture[num].mipstate[0], NULL, false, true);
	else
		ok = MakeTexture(image, 0, 0, 0, 0, &m_Texture[num].mipstate[0], &m_Texture[num].mipstate[1], IsBumpHeightmap, false);

	m_Texture[num].w=sx; // Set Size of Texture
	m_Texture[num].h=sy;
	m_Texture[num].bw=image.m_sizex; // Set Size of BMP File
	m_Texture[num].bh=image.m_sizey;
	m_Texture[num].txx=(float)m_Texture[num].bw/m_Texture[num].w;
	m_Texture[num].txy=(float)m_Texture[num].bh/m_Texture[num].h;
	if(!ok) FreeSprite(num);

	if(ok && inscreen>=0)
	{
		if(m_FBuffer[inscreen].Create(m_Texture[num].w, m_Texture[num].h, m_Texture[num].mipstate[0], true))
			m_FBuffer[inscreen].m_SpriteIndex = num;
	}
	return ok;


return ok;
#endif
}


#if defined(HAVE_OPENGL)
// --------------------------------------------------------------- //
// Create a normal map from a grayscale image
// --------------------------------------------------------------- //
void build_detail_normalmap(OPENGL_IMAGE& image)
{
OPENGL_IMAGE out;
	out.resize(image.m_sizex, image.m_sizey);
	int x, y;
	float dX, dY, nX, nY, nZ, len;

	for(y=0;y<image.m_sizey;y++)
	{
		for(x=0;x<image.m_sizey;x++)
		{
			dY =((float)image[4*(((y+1)%image.m_sizey)*image.m_sizex+((x-1+image.m_sizex)%image.m_sizex))+2])/255.0f*-1.0f;
			dY+=((float)image[4*(((y+1)%image.m_sizey)*image.m_sizex+x%image.m_sizex)+2])/255.0f*-2.0f;
			dY+=((float)image[4*(((y+1)%image.m_sizey)*image.m_sizex+((x+1)%image.m_sizex))+2])/255.0f*-1.0f;
			dY+=((float)image[4*(((y-1+image.m_sizey)%image.m_sizey)*image.m_sizex+((x-1+image.m_sizex)%image.m_sizex))+2])/255.0f*1.0f;
			dY+=((float)image[4*(((y-1+image.m_sizey)%image.m_sizey)*image.m_sizex+(x%image.m_sizex))+2])/255.0f*2.0f;
			dY+=((float)image[4*(((y-1+image.m_sizey)%image.m_sizey)*image.m_sizex+((x+1)%image.m_sizex))+2])/255.0f*1.0f;

			dX =((float)image[4*(((y-1+image.m_sizey)%image.m_sizey)*image.m_sizex+((x-1+image.m_sizex)%image.m_sizex))+2])/255.0f*-1.0f;
			dX+=((float)image[4*((y%image.m_sizey)*image.m_sizex+((x-1+image.m_sizex)%image.m_sizex))+2])/255.0f*-2.0f;
			dX+=((float)image[4*(((y+1)%image.m_sizey)*image.m_sizex+((x-1+image.m_sizex)%image.m_sizex))+2])/255.0f*-1.0f;
			dX+=((float)image[4*(((y-1+image.m_sizey)%image.m_sizey)*image.m_sizex+((x+1)%image.m_sizex))+2])/255.0f*1.0f;
			dX+=((float)image[4*((y%image.m_sizey)*image.m_sizex+((x+1)%image.m_sizex))+2])/255.0f*2.0f;
			dX+=((float)image[4*(((y+1)%image.m_sizey)*image.m_sizex+((x+1)%image.m_sizex))+2])/255.0f*1.0f;

			nX=-dX;
			nY=-dY;
			nZ=1;

			len=1.0f/((float)sqrt((nX*nX)+(nY*nY)+(nZ*nZ)));
			nX*=len;
			nY*=len;
			nZ*=len;

			out[4*(y*image.m_sizex+x)+3]=image[4*(y*image.m_sizex+x)+3];
			out[4*(y*image.m_sizex+x)  ]=(unsigned char)((nX+1.0f)/2.0f*255.0f);
			out[4*(y*image.m_sizex+x)+1]=(unsigned char)((nY+1.0f)/2.0f*255.0f);
			out[4*(y*image.m_sizex+x)+2]=(unsigned char)((nZ+1.0f)/2.0f*255.0f);
		}
	}
//    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, dataout);
	// now swap pointers
	unsigned char* pS = image.m_txdata;
	image.m_txdata = out.m_txdata;
	out.m_txdata = pS;
	// memcpy(datain, dataout, width*height*4);
}
#endif


///////////////////////////////////////////////////////////////////////////////
// OPENGL_IMAGE - functions
///////////////////////////////////////////////////////////////////////////////

// Make a small font (5x6 px) -> 16x8 chars
void OPENGL_IMAGE::MakeDefFont()
{
OPENGL_IMAGE& img = *this;
// 5x6 font ;)
static unsigned int tmap[6][16] = {
	0x00000000, 0x08021080, 0x00002940, 0x15f57d40, 0x08e619c4, 0x14222140, 0x0ed114e0, 0x00001080, 0x08210880, 0x04421040, 0x09575480, 0x08471080, 0x04400000, 0x00070000, 0x08000000, 0x04222100, 
	0x00e5294e, 0x004210c4, 0x00e1390e, 0x00e4390e, 0x0084394a, 0x00e4384e, 0x00e5384e, 0x0042210e, 0x00e5394e, 0x00e4394e, 0x00020080, 0x00220080, 0x00820888, 0x000701c0, 0x00222082, 0x0040310e, 
	0x00e0b526, 0x00a5394e, 0x00e5194e, 0x00e5094e, 0x00e52946, 0x00e1384e, 0x0021384e, 0x00e5284e, 0x00a5394a, 0x00e2108e, 0x00e5210c, 0x00a5194a, 0x00e10842, 0x015ad6bf, 0x00a5294e, 0x00e5294e, 
	0x0021394e, 0x00f294a7, 0x00a5194e, 0x00e4384e, 0x0042108e, 0x00e5294a, 0x0045294a, 0x01fad6b5, 0x00a5114a, 0x0042394a, 0x00e1110e, 0x00610846, 0x01041041, 0x00621086, 0x00000144, 0x00e00000, 
	0x00000082, 0x00e53000, 0x00e53842, 0x00e13800, 0x00e53908, 0x08273800, 0x00423888, 0x08873800, 0x00a51842, 0x00421004, 0x04421004, 0x00a32842, 0x00421084, 0x015abc00, 0x00a51800, 0x00e53800, 
	0x04e51800, 0x10e53000, 0x00253800, 0x00e409c0, 0x00c23884, 0x00c52800, 0x00452800, 0x00ead400, 0x00a22800, 0x1c872800, 0x00e121c0, 0x00821888, 0x00421084, 0x00223082, 0x0002a800, 0x3f18c63f, 
	};

	img.resize(128,64);
	for(unsigned int cx=0; cx<16; ++cx)
	{
		for(unsigned int cy=0; cy<6; ++cy)
		{
			for(unsigned int x=0; x<7; ++x)
			{
				for(unsigned int y=0; y<8; ++y)
				{
					// RGBA
					#define PIXP(px, py) ( ((px)+cx*8 + ((cy+2)*8+(py))*128) *4 )     // ( ((px)+cx*7)*4 + (47+ (cy*8+(py)))*128*4 )
					#define WHITE(px, py) {int pp = PIXP(px,py); img[pp++]=255; img[pp++]=255; img[pp++]=255; img[pp]=255;}
					#define BLACK(px, py) {int pp = PIXP(px,py); img[pp++]=8; img[pp++]=8; img[pp++]=8; img[pp]=255;}
					#define IS_SET(px, py) ( (px<1 || px>5 || py<1 || py>6) ? 0 : (tmap[cy][cx] & (1<<((px)+5*(py) -5-1   ))) )
				
					if(IS_SET(x,y))
					{
						WHITE(x,y);
					}
					else
					{
						if     (x<7 && IS_SET(x+1,y) ) {BLACK(x,y);}
						else if(x>0 && IS_SET(x-1,y) ) {BLACK(x,y);}
						else if(y>0 && IS_SET(x,y-1) ) {BLACK(x,y);}
						else if(y<7 && IS_SET(x,y+1) ) {BLACK(x,y);}
					}
					#undef PIXP
					#undef WHITE
					#undef BLACK
					#undef SET
				}
			}
		}
	}
}

// scale by 2x2
void OPENGL_IMAGE::ScaleX2()
{
OPENGL_IMAGE img;
	img.resize(m_sizex*2, m_sizey*2);
	unsigned int* pSrc = (unsigned int*)m_txdata;
	unsigned int* pDst = (unsigned int*)img.m_txdata;
	for(int line=0; line<m_sizey; ++line)
	{
		for(int src=0; src<m_sizex; ++src)
		{
			pDst[0] = *pSrc;
			pDst[1] = *pSrc;
			pDst[img.m_sizex  ] = *pSrc;
			pDst[img.m_sizex+1] = *pSrc;
			pDst+=2;
			++pSrc;
		}
		pDst+=img.m_sizex;
	}
	*this = img;
}


// #ifndef POCKETPC
	#if defined(HAS_PNG_LIB)
		int gf_png_open(const char* szFileName, int* width, int* height, int* rowbytes, int* channels);
		void gf_png_read(unsigned char* buffer, unsigned char** row_pointers);
		void gf_png_close();
		int gf_write_png(const char *file_name, const unsigned char* pixels32, int width, int height);
	#endif

	int gf_jpeg_read(const char* data, int len, int* pWidth, int* pHeight, int* pChannels, unsigned char** pRGB);
	int gf_write_jpeg(const char *file_name, const unsigned char* pixels32, int width, int height);


	bool OPENGL_IMAGE::LoadImageJPG(const char* szFileName)
	{
		unsigned char* pRGB=NULL;
		int width=0, height=0, channels=0;
		// GF: use GFFileSystem to be able to load from shoebox
		GFFileSystem fs;
		if(!fs.OpenIn(szFileName))
			return false;

		int filelen = (int)fs.GetSize();
		char* buffer = new char[filelen+1]; buffer[0]='\0';
		bool bReadOK =false;
		bReadOK = fs.Read(buffer, filelen);

		if(bReadOK)
			bReadOK = (gf_jpeg_read(buffer, filelen, &width, &height, &channels, &pRGB)==0)?false:true;
		
		if(!bReadOK)
		{
			if(pRGB) delete[] pRGB; // might return 0 after allocation
			delete[] buffer;
			return false;
		}


		switch(channels)
		{
		case 4:
		case 3:
		case 1:
			break;
		default:
			g_debug("JPEG: Channels: %d not supported!\n", channels);
			if(pRGB) delete[] pRGB;
			return false;
		}

		// allocate the memory to hold the image using the fields of png_info.
		OPENGL_IMAGE& image = *this;
		image.resize(width, height);

		if(channels != 4) // no alpha in image
			image.m_bSourceHadAlpha = false;

		unsigned char R=0,G=0,B=0,A=255;
		for(int row = 0; row < height; row++)
		{
			unsigned char *ptr = pRGB + ( width*channels * row);
			unsigned char* pdest = &image[row*image.m_sizex*4];
			for(int x=0; x<width; x++)
			{
				switch(channels)
				{
				case 4:
					R=ptr[0]; G=ptr[1]; B=ptr[2]; A=ptr[3]; break;
				case 3:
					R=ptr[0]; G=ptr[1]; B=ptr[2]; A=255; break;
				case 1:
					R=ptr[0]; G=ptr[0]; B=ptr[0]; A=255; break;
				}

				*pdest++ = R;
				*pdest++ = G;
				*pdest++ = B;
				*pdest++ = A;

				ptr += channels;
			}
		}
		delete[] pRGB;
		delete[] buffer;
		return true;
	}


	/* read a png file.  You may want to return an error code if the read
	fails (depending upon the failure). */
	bool OPENGL_IMAGE::LoadImagePNG(const char* szFileName, COLORREF cookie)
	{
	#if defined(HAS_PNG_LIB)
	int width, height, rowbytes, channels;

		OPENGL_IMAGE& image = *this;

		if(gf_png_open(szFileName, &width, &height, &rowbytes, &channels))
		{
			if(channels != 4) // no alpha in image
				image.m_bSourceHadAlpha = false;

			unsigned char** row_pointers = new unsigned char*[height];
			unsigned char*  buffer = new unsigned char[height*rowbytes+8+64];
			if (buffer!=NULL && row_pointers!=NULL)
			{
				gf_png_read(buffer, row_pointers);

				// allocate the memory to hold the image using the fields of png_info.
				image.resize(width, height);
				// this->resize(,info_ptr->channels);

				unsigned char R=0,G=0,B=0,A=255;
				for(int row = 0; row < height; row++)
				{
					unsigned char *ptr = row_pointers[row];
					unsigned char* pdest = &image[row*image.m_sizex*4];
					for(int x=0; x<width; x++)
					{
						switch(channels)
						{
						case 4:
							R=ptr[0]; G=ptr[1]; B=ptr[2]; A=ptr[3];
							break;
						case 3:
							R=ptr[0]; G=ptr[1]; B=ptr[2]; A=255;
							if(cookie!=(COLORREF)-1 && RGB(R,G,B) == cookie)
							{
								R=0; G=0; B=0; A=0;
							}
							break;
						case 1:
							R=ptr[0]; G=ptr[0]; B=ptr[0]; A=255;
							break;
						}

						*pdest++ = R;
						*pdest++ = G;
						*pdest++ = B;
						*pdest++ = A;

						ptr += channels;
					}
				}
			}
			else
			{
				if(row_pointers)
					delete[] row_pointers;
				if(buffer)
					delete[] buffer;
				// not enough memory
				return false;
			}

			delete[] row_pointers;
			delete[] buffer;
			gf_png_close();
		}
		else
			return false;

		return true;
	#else
		return false;
	#endif // HAS_PNG_LIB
	}


	// -------------------------------- //
	// Load BMP image to RGBA (endianess aware)
	// -------------------------------- //
	bool OPENGL_IMAGE::LoadImageBMP(const char* szFileName, COLORREF cookie)
	{
	OPENGL_IMAGE& image = *this;
	char type_B='x', type_M='x';
	RGBQUAD* pPalette = NULL;
	BITMAPINFO		*pBitmapInfo = NULL;
	unsigned long	lInfoSize = 0;
	unsigned long	lBitSize = 0;
	BYTE			*pBitmapData = NULL;
	BITMAPFILEHEADER bitmapHeader;
	int				BytesPerLine;
	int				BytesPerPixel;
	int i,j, shiftj;
	bool flip_vertical = true;

	int dmy=0;
		// Open the Bitmap file
	GFFileSystem file;
		if(!file.OpenIn(szFileName))
			goto fail;
		#define BITMAPFILEHEADERSIZE 0x0000000e

		if(!file.Read(&type_B, 1) || !file.Read(&type_M, 1))
			goto fail;
		if(type_B !='B' || type_M!='M')
			goto fail;

		if(!file.ReadLong(dmy)) goto fail;
		bitmapHeader.bfSize=dmy; // android crash on direct read - very silly.

		if(!file.ReadWord(bitmapHeader.bfReserved1)) goto fail;
		if(!file.ReadWord(bitmapHeader.bfReserved2)) goto fail;
		if(!file.ReadLong(dmy)) goto fail;
		bitmapHeader.bfOffBits=dmy;


		// Check format of bitmap file
		#ifndef BF_TYPE_BITMAP
		#define BF_TYPE_BITMAP	'M'*256+'B'
		#endif
//		if(bitmapHeader.bfType != BF_TYPE_BITMAP) goto fail;

		// Read in bitmap information structure
		lInfoSize = bitmapHeader.bfOffBits - BITMAPFILEHEADERSIZE;
		pBitmapInfo = (BITMAPINFO *) new BYTE[lInfoSize];
		memset(pBitmapInfo, 0, lInfoSize);

		// if(!file.Read(pBitmapInfo, lInfoSize)) goto fail;
		if(!file.ReadLong(pBitmapInfo->bmiHeader.biSize)) goto fail;

		switch(pBitmapInfo->bmiHeader.biSize)
		{
		case 12: // OS/2 V1 header. Minon brought this....
			if(!file.ReadWord(pBitmapInfo->bmiHeader.biWidth)) goto fail;
			if(!file.ReadWord(pBitmapInfo->bmiHeader.biHeight)) goto fail;
			if(!file.ReadWord(pBitmapInfo->bmiHeader.biPlanes)) goto fail;
			if(!file.ReadWord(pBitmapInfo->bmiHeader.biBitCount)) goto fail;
			break;
		case 40:
			if(!file.ReadLong(pBitmapInfo->bmiHeader.biWidth)) goto fail;
			if(!file.ReadLong(pBitmapInfo->bmiHeader.biHeight)) goto fail;
			if(!file.ReadWord(pBitmapInfo->bmiHeader.biPlanes)) goto fail;
			if(!file.ReadWord(pBitmapInfo->bmiHeader.biBitCount)) goto fail;
			if(!file.ReadLong(pBitmapInfo->bmiHeader.biCompression)) goto fail;
			if(!file.ReadLong(pBitmapInfo->bmiHeader.biSizeImage)) goto fail;
			if(!file.ReadLong(pBitmapInfo->bmiHeader.biXPelsPerMeter)) goto fail;
			if(!file.ReadLong(pBitmapInfo->bmiHeader.biYPelsPerMeter)) goto fail;
			if(!file.ReadLong(pBitmapInfo->bmiHeader.biClrUsed)) goto fail;
			if(!file.ReadLong(pBitmapInfo->bmiHeader.biClrImportant)) goto fail;
			// 40 bytes for that block
			break;
		default:
			goto fail;
		}
		
		
		if(pBitmapInfo->bmiHeader.biHeight < 0)
		{
			flip_vertical = false;
			pBitmapInfo->bmiHeader.biHeight = - pBitmapInfo->bmiHeader.biHeight;
		}
		
		// read palette if any
		if(lInfoSize>40)
		{
			if(pBitmapInfo->bmiHeader.biSize == 12)
			{
				// V1 header - only full size palette
				int cols = 1<<pBitmapInfo->bmiHeader.biBitCount;
				pPalette = new RGBQUAD[cols];
				BYTE* pRGB = new BYTE[3 * cols];
				if(!file.Read(pRGB, cols*3))
				{
					goto fail;
				}
				BYTE* pPal = (BYTE*)pPalette;
				for(int ii=0; ii<cols; ++ii)
				{
					*pPal++ = pRGB[ii*3];
					*pPal++ = pRGB[ii*3+1];
					*pPal++ = pRGB[ii*3+2];
					*pPal++ = 0;
				}
				delete[] pRGB;
			}
			else
			{
				pPalette = new RGBQUAD[ (lInfoSize-40) / sizeof(RGBQUAD) + 1];
				if(!file.Read(pPalette, (int)lInfoSize-40))
				{
					goto fail;
				}
			}
		}

		pBitmapInfo->bmiHeader.biHeight=abs((int)pBitmapInfo->bmiHeader.biHeight);

		// Only supported: 24 Bit & 8 Bit
		if (pBitmapInfo->bmiHeader.biBitCount!=1 &&
			pBitmapInfo->bmiHeader.biBitCount!=4 &&
			pBitmapInfo->bmiHeader.biBitCount!=8 &&
			pBitmapInfo->bmiHeader.biBitCount!=24 &&
			pBitmapInfo->bmiHeader.biBitCount!=16 &&
			pBitmapInfo->bmiHeader.biBitCount!=32)// Num of bytes per pixel
			goto fail;

		// Bits Per Single Line / SinglePixel
		// 4Byte alignment even in 256 Color Image
		BytesPerPixel=pBitmapInfo->bmiHeader.biBitCount/8;
/*
		if (BytesPerPixel==0)
		{
			printf("BytePerP = 0\n");
			goto fail;
		}
*/
		if(pBitmapInfo->bmiHeader.biWidth%4 == 0)
		{
			BytesPerLine=(int)( pBitmapInfo->bmiHeader.biWidth * pBitmapInfo->bmiHeader.biBitCount ) / 8;
		}
		else
		{
			BytesPerLine=(int)( pBitmapInfo->bmiHeader.biWidth * pBitmapInfo->bmiHeader.biBitCount ) / 8;
		//	BytesPerLine=  pBitmapInfo->bmiHeader.biWidth * BytesPerPixel;
			if(BytesPerLine%4)
			{
			BytesPerLine+=(4-(BytesPerLine%4));
		}
		}
		lBitSize=BytesPerLine*pBitmapInfo->bmiHeader.biHeight;

		// Allocate space for the actual bitmap
		pBitmapData = new BYTE[lBitSize];
		if(!pBitmapData) goto fail;

		// Read in the bitmap bits
		if(!file.Read(pBitmapData, (int)lBitSize))
			goto fail;
		file.Close();

		// Convert BMP to GLTexture RGBA
		if(!image.resize((int)pBitmapInfo->bmiHeader.biWidth, (int)pBitmapInfo->bmiHeader.biHeight))
			goto fail;

		shiftj = pBitmapInfo->bmiHeader.biBitCount/8;
		for (int y=0; y<image.m_sizey; ++y) // For each Pixel in BMP
		{
			i=y*image.m_sizex*4;
			// Source Pixel
			j=(int)(pBitmapInfo->bmiHeader.biHeight-y-1)*BytesPerLine;
			for (int x=0; x<image.m_sizex; ++x)
			{
				image[i+3]=255;
				switch(pBitmapInfo->bmiHeader.biBitCount)
				{
				case 24:
				case 32:
					image[i  ]=pBitmapData[j+2];	// Red
					image[i+1]=pBitmapData[j+1];	// Green
					image[i+2]=pBitmapData[j];	// Blue
					break;
					
				case 16: // RGB 5 5 5
					{
					unsigned short col16 = *(unsigned short*)&pBitmapData[j];
					image[i  ]=(((((col16)>>10)&31)*255)/31);	// Red
					image[i+1]=(((((col16)>> 5)&31)*255)/31);	// Green
					image[i+2]=(((((col16)    )&31)*255)/31);	// Blue
					}
					break;
					
					
				case 8:
					{
					RGBQUAD& colquad=pPalette[pBitmapData[j]];
					image[i  ]=colquad.rgbRed;   // Red
					image[i+1]=colquad.rgbGreen; // Green
					image[i+2]=colquad.rgbBlue;  // Blue
					}
					break;
				case 4:
					{
					int index = pBitmapData[j + x/2];
					if(x & 0x01)
					{
						index &=0x0f;
					}
					else
					{
						index>>=4;
					}
					RGBQUAD& colquad=pPalette[ index ];
					image[i  ]=colquad.rgbRed;   // Red
					image[i+1]=colquad.rgbGreen; // Green
					image[i+2]=colquad.rgbBlue;  // Blue
					}
					break;
				case 1:
					{
					int index = (pBitmapData[j + x/8] & ( 1<<(7-(x%8)) ) ) ? 1:0;
					RGBQUAD& colquad=pPalette[ index ];
					image[i  ]=colquad.rgbRed;   // Red
					image[i+1]=colquad.rgbGreen; // Green
					image[i+2]=colquad.rgbBlue;  // Blue
					}
					break;
				}
				// remove cookie color
				if(cookie!=(COLORREF)-1 && RGB(image[i], image[i+1], image[i+2]) == cookie)
				{
					image[i  ]=0;
					image[i+1]=0;
					image[i+2]=0;
					image[i+3]=0;
				}
				j+=shiftj; i+=4;
			}
		}
		
		if(!flip_vertical)
			image.flipy();
		

		if(pPalette)    delete [] pPalette;
		if(pBitmapInfo) delete [] pBitmapInfo;
		if(pBitmapData) delete [] pBitmapData;

		image.m_bSourceHadAlpha = false; // BMP is solid/cookie coloured
		return true;
	fail:
		if(pPalette)    delete [] pPalette;
		if(pBitmapInfo) delete [] pBitmapInfo;
		if(pBitmapData) delete [] pBitmapData;
		image.clear();
		return false;
	}


	// -------------------------------- //
	// for cookie-coloured images
	// -------------------------------- //
	void OPENGL_IMAGE::FixNeighbourAlpha()
	{
		// For transparent pixels:
		// get neighbour pixel color and make mix this into RGB components, if it has no alpha!
		#define ADDFROM(x,y)	if((x)>0 && (y)>0 && (x)<m_sizex-1 && (y)<+m_sizey-1)\
								{\
									unsigned char* pN = &m_txdata[((x) + (y)*m_sizex)*4]; \
									if (pN[3]!=0)\
									{ \
										div=div+1;\
										red+=(float)pN[0]; green+=(float)pN[1]; blue+=(float)pN[2]; \
									} \
								}
		for (int y=0; y<m_sizey; y++) // For each Pixel in BMP
		{
			for (int x=0; x<m_sizex; x++)
			{
				float div=0;
				float red=0; float green=0; float blue=0;
				unsigned char* pPixl;
				pPixl = &m_txdata[ (x + y*m_sizex)*4]; //i=(x+(tex.sizey-y-1-(image.m_sizey-tex.sizey))*sx)*4];
				if(pPixl[3]==0) // transparent pixel
				{
					ADDFROM(x-1,y-1);
					ADDFROM(x  ,y-1);
					ADDFROM(x+1,y-1);
					ADDFROM(x-1,y  );
					ADDFROM(x+1,y  );
					ADDFROM(x-1,y+1);
					ADDFROM(x  ,y+1);
					ADDFROM(x+1,y+1);
					if(div!=0)
					{
						red/=div;
						green/=div;
						blue/=div;
						pPixl[0] = (unsigned char) red;
						pPixl[1] = (unsigned char) green;
						pPixl[1] = (unsigned char) blue;
					}
					else
					{
						pPixl[0] = 0; pPixl[1]=0; pPixl[2]=0; // somewhere in the void
					}
				}
			}
		}
	}

#if 0 // Old PocketPC code - CE 6.x has no AygShell.dll (LoadImage/SHLoadImagegFile)
	// Get 4 byte aligned width
	#define ALIGN_4_BYTES(a) ( ((a) + 3) & ~3 )
	#ifdef _WIN32_WCE
	extern "C" HBITMAP SHLoadImageFile(LPCTSTR);
	#endif

	bool OPENGL_IMAGE::LoadImageJPG(const char* szFileName)
	{
		return LoadImageBMP(szFileName, -1);
	}
	bool OPENGL_IMAGE::LoadImagePNG(const char* szFileName, COLORREF cookie)
	{
		return LoadImageBMP(szFileName, cookie);
	}

	bool OPENGL_IMAGE::LoadImageBMP(const char* szFileName, COLORREF cookie)
	{
	BITMAP bmp;
	HBITMAP hbmp=NULL;
		clear();

		char fname[512];
		GFFileSystem::MakeAbsolutePathCE(szFileName, fname);

		#ifdef UNICODE
			wchar_t szFile[512];
			_mbstowcsz(szFile, fname, strlen(fname)+1);
		#else
			const char* szFile = fname;
		#endif


		#ifdef _WIN32_WCE
			hbmp = SHLoadImageFile(szFile);
		#else
			#ifndef LR_LOADFROMFILE
			#define LR_LOADFROMFILE 0x0010
			#endif
			hbmp = (HBITMAP)::LoadImage(NULL, szFile, IMAGE_BITMAP, 0,0, 0x0010 /*LR_LOADFROMFILE*/);
		#endif

		if(!hbmp) return false;
		if (::GetObject(hbmp, sizeof(BITMAP), &bmp))
		{
			bmp.bmHeight = abs(bmp.bmHeight);

			unsigned char* pixels;

			// Useing CreateDIBSection now
			HDC hDC = GetWindowDC(NULL);

			// Initialize the BITMAPINFOHEADER in m_DIBinfo
			BITMAPINFO bi;
			memset(&bi, 0, sizeof(bi));
			BITMAPINFOHEADER& bih= bi.bmiHeader;
			bih.biSize          = sizeof(BITMAPINFOHEADER);
			bih.biWidth         = bmp.bmWidth;
			bih.biHeight        = bmp.bmHeight;
			bih.biPlanes        = 1;  // Must always be 1 according to docs
			bih.biBitCount      = 24;
			bih.biCompression   = BI_RGB;
			bih.biSizeImage     = ALIGN_4_BYTES(bmp.bmWidth*3) * bmp.bmHeight;

			HBITMAP hTargetBitmap = CreateDIBSection(hDC, (const BITMAPINFO*)&bi, DIB_RGB_COLORS, (void**) &pixels, NULL, 0);
			ReleaseDC(NULL, hDC);		

			// Copy source bitmap into the target bitmap.
			// Create 2 device contexts 
			HDC hMemDc = ::CreateCompatibleDC(NULL);
			HDC hTargetDc = ::CreateCompatibleDC(NULL);

			// Select source bitmap into one DC, target into another
			HBITMAP hOldBitmap1 = (HBITMAP) ::SelectObject(hMemDc, hbmp);
			HBITMAP hOldBitmap2 = (HBITMAP) ::SelectObject(hTargetDc, hTargetBitmap);
			
			// Might only get 16bpp instead of 24??
			if(cookie!=-1)
			{
				::SetPixel(hMemDc, 0,0,cookie);
				cookie = ::GetPixel(hMemDc, 0,0);
			}

			// Copy source bitmap into the target one
			::BitBlt(hTargetDc, 0, 0, bmp.bmWidth, bmp.bmHeight, hMemDc, 0, 0, SRCCOPY);

			// Restore device contexts
			::SelectObject(hMemDc, hOldBitmap1);
			::SelectObject(hTargetDc, hOldBitmap2);

			::DeleteDC(hMemDc);
			::DeleteDC(hTargetDc);

			resize(bmp.bmWidth, bmp.bmHeight);
			unsigned char* pCur;
			unsigned char R,G,B;
			for(int y=0; y<bmp.bmHeight; ++y)
			{
				// HBITMAP is upside down
				unsigned char* pbPixel = pixels + (bmp.bmHeight-y-1) * ALIGN_4_BYTES(bmp.bmWidth*3);
				pCur = this->m_txdata + y*bmp.bmWidth*4;
				for (int x=0; x<bmp.bmWidth; ++x)
				{
					B = *pbPixel++;
					G = *pbPixel++;
					R = *pbPixel++;
					*pCur++ = R;
					*pCur++ = G;
					*pCur++ = B;
					if(cookie == RGB(R,G,B))
					{
						*pCur++ = 0;
					}
					else
					{
						*pCur++ = 255;
					}
				}
			}
			DeleteObject(hTargetBitmap);
		}
		return true;
	}
#endif // POCKETPC


#if defined(HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
// --------------------------------------------------------------- //
// BMP -> Texture
// --------------------------------------------------------------- //
bool OpenGLRainbows::MakeTexture(OPENGL_IMAGE& image, int startx, int starty,
									int sizex, int sizey,
									GLuint* pTextureID, GLuint* pTextureIDMipMap,
									bool IsBumpHeightMap, bool IsFont)
{
int				max_tex;
int				sx, sy;
long			res=FALSE;

	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_tex);

	if(sizex==0) sizex=image.m_sizex-startx;
	if(sizey==0) sizey=image.m_sizey-starty;
	if(sizex<=0 || sizey<=0) return false;
	
	sizex=min(sizex, max_tex);
	sizey=min(sizey, max_tex);

	// Create 2^n size
int yoffset = 0;
	sx=GLB_MIN_TEX_SIZE/2; sy=GLB_MIN_TEX_SIZE/2;
	do{sx<<=1;} while(sx < sizex);
	do{sy<<=1;} while(sy < sizey);
	yoffset = sy - sizey;
	
OPENGL_IMAGE tex;
	if(!tex.resize(sx, sy)) return false;

	for(int x=max(0,startx); x<min(startx+sx, image.m_sizex); ++x) // bmp coords
	for(int y=max(0,starty); y<min(starty+sy, image.m_sizey); ++y)
	{
		if(x-startx<sizex && y-starty<sizey)
		{
			int t, b;
			t = (x-startx + ( sy-(y-starty)-1-yoffset )*sx)*4; // texture
			b = (x + y*image.m_sizex)*4; // bitmap
			tex[t++] = image[b++];
			tex[t++] = image[b++];
			tex[t++] = image[b++];
			tex[t  ] = image[b  ];
		}
	}


	// Now make a GLtexture and free stuff
	// This is specific to the binary format of the data read in.
    glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
	#if defined(HAVE_OPENGL) && !defined(EMSCRIPTEN)
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
	glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
	glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
	#endif

	// get new Id if none given or free mem
	if(*pTextureID==0)
		glGenTextures(1, pTextureID);
	else
		glDeleteTextures(1, pTextureID);

	glBindTexture(GL_TEXTURE_2D, *pTextureID);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	if(!IsFont)
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	}
	else
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	}

	#if defined(HAVE_OPENGL)
		if (IsBumpHeightMap)
			build_detail_normalmap(tex);
		//		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, tex.m_sizex, tex.m_sizey, 0, GL_RGBA, GL_UNSIGNED_BYTE, tex.m_txdata);
		//	#elif defined(HAVE_OPENGL_ES)
		//		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA , tex.m_sizex, tex.m_sizey, 0, GL_RGBA, GL_UNSIGNED_BYTE, tex.m_txdata);
	#endif

	GLint  intformat = GL_RGBA8;
	#if defined(HAVE_OPENGL_ES) || defined(EMSCRIPTEN)
		intformat = GL_RGBA;
	#endif
	glTexImage2D(GL_TEXTURE_2D, 0, intformat , tex.m_sizex, tex.m_sizey, 0, GL_RGBA, GL_UNSIGNED_BYTE, tex.m_txdata);

	// Now build MipMaps
	#if defined(HAVE_OPENGL)
		long components[4];
		if (pTextureIDMipMap)
		{
			glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
			#if defined(HAVE_OPENGL) && !defined(EMSCRIPTEN)
			glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
			glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
			glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
			#endif
			glGenTextures(1, pTextureIDMipMap);
			glBindTexture(GL_TEXTURE_2D, *pTextureIDMipMap);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR); // Trilinear Mipmapping
			int sxn=tex.m_sizex, syn=tex.m_sizey, i=0, x, y, j, compo, pixl;

			// Attention! Mipmap levels must be SQUARE!
			while(sxn> syn) // too wide, half size
			{
				sxn>>=1;
				for (y=0; y<syn; ++y)
				for (x=0; x<sxn; ++x)
				{
					for(j=0; j<4; j++) // RGBA
					{
						tex[x*4 + y*sxn*4 + j] = (tex[x*8 + y*sxn*8 + j]+tex[x*8 +4 + y*sxn*8 + j])>>1;
					}
				}
			}
		
			while(syn> sxn) // too high, half size
			{
				syn>>=1;
				for (y=0; y<syn; ++y)
				for (x=0; x<sxn; ++x)
				{
					for(j=0; j<4; j++) // RGBA
					{
						tex[x*4 + y*sxn*4 + j] = (tex[x*4 + y*sxn*8 + j]+tex[x*4 + y*sxn*8 +sxn*4 + j])>>1;
					}
				}
			}

			while(true)
			{
				glTexImage2D(GL_TEXTURE_2D, i++, intformat, sxn, syn, 0, GL_RGBA, GL_UNSIGNED_BYTE, tex.m_txdata);
				sxn/=2; syn/=2;
				if (sxn<1 && syn<1)  break;
				for (y=0; y<max(syn, 1); y++)
				for (x=0; x<max(sxn, 1); x++)
				{
					for(j=0; j<4; j++) // RGBA
					{
						compo=1;
						components[0] = tex[x*8 + y*sxn*16 + j];
						if (x<sxn-1)
						{
							components[compo]=tex[x*8+4 + y*sxn*16 + j]; compo++;
							if (y<syn-1)
							{
								components[compo]=tex[x*8+4 + (y+1)*sxn*16 + j]; compo++;
							}
						}
						if (y<syn-1)
						{
							components[compo]=tex[x*8 + (y+1)*sxn*16 + j]; compo++;
						}
						// The weight of my pixel is double to each horz or vert.
						// The diagonal pixel is one fourth
						switch(compo)
						{
						case 1: pixl=components[0]; break;
						case 2: pixl=(components[0]*2+components[1])/3; break;
						case 4: pixl=(components[0]*4+components[1]*2+components[2]*2+components[3])/9; break;
						}
						tex[x*4 + y*sxn*4 + j]=(unsigned char)(pixl);
					}
				}
			}
		}
	#endif // mipmap

return true;
}

//---------------------------------------------------------------//
// Get the pixels from the card into SpritePix data
// (for pixel based collisions)
//---------------------------------------------------------------//
void OpenGLRainbows::GetSpritePixels(int n)
{
// OpenGL|ES has no GL_ALPHA for glReadPixels
// must read full RGBA

		// Use a PBuffer an glReadPixels
	#if defined(HAVE_OPENGL) && !defined(EMSCRIPTEN)
		#ifdef SPRITE_ALPHA_PIX_AS_VECTOR
		if(!SpritePix[n].Size() )
		#else
 		if (!SpritePix[n])
		#endif
		{
			GLint width=m_Texture[n].w, height=m_Texture[n].h;
			glBindTexture(GL_TEXTURE_2D, m_Texture[n].mipstate[0]);
			#ifndef EMSCRIPTEN // not for HTML5
			glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH,  &width);
			glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &height);
			#endif

			#ifdef SPRITE_ALPHA_PIX_AS_VECTOR
				SpritePix[n].SetSize(max(8*8, width*height) );
			#else
				SpritePix[n] = new BYTE [ max(8*8, width*height) ];
			#endif
			glGetTexImage(GL_TEXTURE_2D, 0, GL_ALPHA, GL_UNSIGNED_BYTE, &SpritePix[n][0]);
			OPENGL_CHECK_ERRORS();
#if 0
				printf("Alpha pixels: ID=%d\n", n);
				for(int y=0; y<m_Texture[n].h; ++y)
				{
					for(int x=0; x<m_Texture[n].w; ++x)
					{
						printf("%c", SpritePix[n][x + y*m_Texture[n].w]? '#':'_');
					}
					printf("\n");
				}
#endif
		}
	#elif defined(HAVE_OPENGL_ES) || defined(EMSCRIPTEN)
		// Use a PBuffer and glReadPixels
		#ifdef SPRITE_ALPHA_PIX_AS_VECTOR
		if(!SpritePix[n].Size() )
		#else
 		if (!SpritePix[n])
		#endif
		{
			/* if only you had glReadPixels(GL_ALPHA) ...
				FBUFFER pbuffer;
				if(pbuffer.Create(m_Texture[n].w, m_Texture[n].h, false))
				{
					SpritePix[n] = new BYTE [m_Texture[n].w*m_Texture[n].h];
					pbuffer.StartRendering(m_Texture[n].mipstate[0]); // bind mip-mapping texture
					glReadPixels(0,0,m_Texture[n].w, m_Texture[n].h, GL_ALPHA, GL_UNSIGNED_BYTE, SpritePix[n]);
					pbuffer.StopRendering();
				}
			*/

			FBUFFER pbuffer;
			if(pbuffer.Create(m_Texture[n].w, m_Texture[n].h, m_Texture[n].mipstate[0], false))
			{
				#ifdef SPRITE_ALPHA_PIX_AS_VECTOR
					SpritePix[n].SetSize(m_Texture[n].w*m_Texture[n].h);
					Queue<BYTE> pTemp; pTemp.SetSize(m_Texture[n].w*m_Texture[n].h*4);
				#else
					SpritePix[n] = new BYTE [m_Texture[n].w*m_Texture[n].h];
					BYTE* pTemp = new BYTE[m_Texture[n].w*m_Texture[n].h*4];
				#endif
				// memset(pTemp, 0, 4*m_Texture[n].w*m_Texture[n].h);

				if(!pbuffer.Bind()) // bind mip-mapping texture
				{
					g_debug("fail bind FBO on tex %d=%d\n", n, m_Texture[n].mipstate[0]);
				}

				glReadPixels(0,0,m_Texture[n].w, m_Texture[n].h, GL_RGBA, GL_UNSIGNED_BYTE, &pTemp[0]);
				if(!pbuffer.Unbind())
				{
					g_debug("fail Unbind FBO on tex %d=%d\n", n, m_Texture[n].mipstate[0]);
				}

				
				#if defined( SPRITE_ALPHA_PIX_AS_VECTOR)
					int dest=0;
					int src=3;
					for(int y=0; y<m_Texture[n].h; ++y)
					{
						for(int x=0; x<m_Texture[n].w; ++x)
						{
							SpritePix[n][dest] = pTemp[src];
							++dest;
							src+=4;
						}
					}
				#else
					BYTE* pSrc  = pTemp + 3;
					BYTE* pDest = SpritePix[n];
					for(int y=0; y<m_Texture[n].h; ++y)
					{
						for(int x=0; x<m_Texture[n].w; ++x)
						{
							*pDest = *pSrc;
							++pDest;
							pSrc+=4;
						}
					}
					delete[] pTemp;
				#endif
				
#if 0
				printf("RGBA pixels: ID=%d\n", n);
				for(int y=0; y<m_Texture[n].h; ++y)
				{
					for(int x=0; x<m_Texture[n].w; ++x)
					{
						printf("%x", *(int*)(&pTemp[(x + y*m_Texture[n].w)*4]));
					}
					printf("\n");
				}
				printf("Alpha pixels: ID=%d\n", n);
				for(int y=0; y<m_Texture[n].h; ++y)
				{
					for(int x=0; x<m_Texture[n].w; ++x)
					{
						printf("%c", SpritePix[n][x + y*m_Texture[n].w]? '#':'_');
					}
					printf("\n");
				}
#endif
			}
		}
		if(!SpritePix)
			g_debug("can't get alpha pixels\n");
	#endif
}
#endif



// Rect Collision?
#define qmax_(a, b) ((a)>(b) ? (a) : (b))
#define qmin_(a, b) ((a)<(b) ? (a) : (b))
#define __RectRectIntersection(x1, y1, xe1, ye1, x2, y2, xe2, ye2) \
							(qmax_(qmin_(x1, xe1), qmin_(x2, xe2)) <= qmin_(qmax_(x1, xe1), qmax_(x2, xe2))) \
						&&	(qmax_(qmin_(y1, ye1), qmin_(y2, ye2)) <= qmin_(qmax_(y1, ye1), qmax_(y2, ye2)))
// For normalized rects
#define __NormRectRectIntersection(x1, y1, xe1, ye1, x2, y2, xe2, ye2) \
							   ((qmax_(x1, x2) <= qmin_(xe1, xe2)) \
							&& (qmax_(y1, y2) <= qmin_(ye1, ye2)))
//---------------------------------------------------------------//
// Sprite Collision (Pixel based)
//---------------------------------------------------------------//
bool	OpenGLRainbows::SpriteCollision(int n1, int n2, int x1, int y1, int x2, int y2)
{
#if defined(TINYGL) || defined(TRENKI3D)
	return oBitmap::Collision(m_Sprite[n1], x1,y1, m_Sprite[n2], x2,y2);
#else
	if (!__NormRectRectIntersection(x1, y1, x1+m_Texture[n1].bw, y1+m_Texture[n1].bh, x2, y2, x2+m_Texture[n2].bw, y2+m_Texture[n2].bh)) return false;

	// Find Overlapping Rect
	int xs, ys, xe, ye;
	xs=qmax_(x1, x2); ys=qmax_(y1, y2);
	xe=qmin_(x1+m_Texture[n1].bw, x2+m_Texture[n2].bw); ye=qmin_(y1+m_Texture[n1].bh, y2+m_Texture[n2].bh);
	return _PixelCollision(n1,x1,y1, n2,x2,y2, xs,ys,xe,ye);
#endif
}

//---------------------------------------------------------------//
// Sprite-Animation Collision (Pixel based)
//---------------------------------------------------------------//
bool OpenGLRainbows::TileCollision(  int n1, int tile1, int x1, int y1, int n2, int tile2, int x2, int y2)
{
	int tx1, ty1, tx2, ty2; // xy position for tile
	int w1, h1, w2, h2; // width

	if(m_Texture[n1].tw==0)
	{
		tx1 = 0;
		ty1 = 0;
		w1 = m_Texture[n1].bw;
		h1 = m_Texture[n1].bh;
	}
	else
	{
		int nperline = m_Texture[n1].bw / m_Texture[n1].tw;
		if(nperline==0) nperline=1;
		tx1=(   tile1 % nperline)*m_Texture[n1].tw;
		ty1=(0+ tile1 / nperline)*m_Texture[n1].th;
		w1 = m_Texture[n1].tw;
		h1 = m_Texture[n1].th;
	}

	if(m_Texture[n2].tw==0)
	{
		tx2 = 0;
		ty2 = 0;
		w2 = m_Texture[n2].bw;
		h2 = m_Texture[n2].bh;
	}
	else
	{
		// w=336; 
		int nperline = m_Texture[n2].bw / m_Texture[n2].tw;
		if(nperline==0) nperline=1;
		tx2=(   tile2 % nperline)*m_Texture[n2].tw;
		ty2=(0+ tile2 / nperline)*m_Texture[n2].th;
		w2 = m_Texture[n2].tw;
		h2 = m_Texture[n2].th;
	}



#if defined(TINYGL) || defined(TRENKI3D)
	return oBitmap::CollisionClip(	m_Sprite[n1], x1,y1,tx1,ty1,w1,h1,
									m_Sprite[n2], x2,y2,tx2,ty2,w2,h2);
#else
	if (!__NormRectRectIntersection(x1, y1, x1+w1, y1+h1,
									x2, y2, x2+w2, y2+h2)) return false;

	// Find Overlapping Rect
	int xs, ys, xe, ye;
	xs=qmax_(x1, x2); ys=qmax_(y1, y2);
	xe=qmin_(x1+w1, x2+w2); ye=qmin_(y1+h1, y2+h2);
	return _PixelCollision(n1,x1-tx1,y1-ty1, n2,x2-tx2,y2-ty2, xs,ys,xe,ye);

#endif
}



//---------------------------------------------------------------//
// Pixel-Range Collision of Sprites
//---------------------------------------------------------------//
bool OpenGLRainbows::_PixelCollision(int n1, int x1, int y1, int n2, int x2, int y2, int xs, int ys, int xe, int ye)
{
	if(ys==ye || xs==xe) return false; // no pixels - quick shot fails otherwise

#if !defined(TINYGL) && !defined(TRENKI3D)
BYTE	*pBits1, *pBits2;
int		x, y;

	// Get alpha bits, only in SX, SY dimension (2^n x 2^m)
	GetSpritePixels(n1);
	GetSpritePixels(n2);

	pBits1 = &SpritePix[n1][0];
	pBits2 = &SpritePix[n2][0];
	
	// please, don't crash
	if(!pBits1 || !pBits2)
		return false;

	// splitting even/odd rows for even more speed
	BYTE* pRow1, *pRow2;

	#ifdef SPRITE_ALPHA_PIX_AS_VECTOR
		#define VALID_ALPHA(ispr, pos) {BYTE b = SpritePix[(ispr)][(pos)];}
	#else
		#define VALID_ALPHA(ispr, pos)
	#endif
	#ifdef _DEBUG
		#define DBG {if(*pRow1 && *pRow2) this->DrawPixel(x,y,RGB(0,255,0)); else if(*pRow1) this->DrawPixel(x,y,RGB(0,0,255)); else if (*pRow2) this->DrawPixel(x,y,RGB(0,255,255)); else DrawPixel(x,y,RGB(255,255,0)); \
					VALID_ALPHA(n1, (int)(pRow1-pBits1)); VALID_ALPHA(n2, (int)(pRow2-pBits2));}
	#else
		#define DBG {}
	#endif

#if 0
	if(n1==0)
	{
		printf("Collision Range x %d, y %d----\n", xs-x1, ys-y1);
		for (y=ys; y<ye; y++)
		{
			// pBits[x][y] = pBits[x + y*w]
			pRow1 = pBits1 + xs - x1 + (m_Texture[n1].bh -1 - (y-y1) )*m_Texture[n1].w;
			pRow2 = pBits2 + xs - x2 + (m_Texture[n2].bh -1 - (y-y2) )*m_Texture[n2].w;
			for (x=xs; x<xe; x++)
			{
				char out='.';
				if(*pRow1) out='1';
				if(*pRow2) out='2';
				if (*pRow1 && *pRow2)
					out='X';
				printf("%c", out);
				pRow1++; pRow2++;
			}
			printf("\n");
		}
		printf("\n");
	}
#endif

	// very quick shot - center of overlapping
	x = xs+(xe-xs)/2; y=ys+(ye-ys)/2;
	pRow1 = pBits1 - x1 + (m_Texture[n1].bh-1 - (y-y1) )*m_Texture[n1].w + x;
	pRow2 = pBits2 - x2 + (m_Texture[n2].bh-1 - (y-y2) )*m_Texture[n2].w + x;
	DBG
	if (*pRow1 && *pRow2) return true;// Got Collision on both bits
	
	for (y=ys; y<ye; y+=2)
	{
		// pBits[x][y] = pBits[x + y*w]
		pRow1 = pBits1 + xs - x1 + (m_Texture[n1].bh-1 - (y-y1) )*m_Texture[n1].w;
		pRow2 = pBits2 + xs - x2 + (m_Texture[n2].bh-1 - (y-y2) )*m_Texture[n2].w;
		for (x=xs; x<xe; x+=2)
		{
			DBG
			if (*pRow1 && *pRow2) return true;// Got Collision on both bits
			pRow1+=2; pRow2+=2;
		}
	}

	// even rows
		// even pixel cols
		for (y=ys; y<ye; y+=2)
		{
			// pBits[x][y] = pBits[x + y*w]
			pRow1 = pBits1 + xs - x1 + (m_Texture[n1].bh-1 - (y-y1) )*m_Texture[n1].w;
			pRow2 = pBits2 + xs - x2 + (m_Texture[n2].bh-1 - (y-y2) )*m_Texture[n2].w;
			for (x=xs; x<xe; x+=2)
			{
				DBG
				if (*pRow1 && *pRow2) return true;// Got Collision on both bits
				pRow1+=2; pRow2+=2;
			}
		}

		// odd cols
		++xs;
		for (y=ys; y<ye; y+=2)
		{
			// pBits[x][y] = pBits[x + y*w]
			pRow1 = pBits1 + xs - x1 + (m_Texture[n1].bh-1 - (y-y1) )*m_Texture[n1].w;
			pRow2 = pBits2 + xs - x2 + (m_Texture[n2].bh-1 - (y-y2) )*m_Texture[n2].w;
			for (x=xs; x<xe; x+=2)
			{
				DBG
				if (*pRow1 && *pRow2) return true;// Got Collision on both bits
				pRow1+=2; pRow2+=2;
			}
		}
	
	// odd rows
		--xs;
		++ys;
		// even pixel cols
		for (y=ys; y<ye; y+=2)
		{
			// pBits[x][y] = pBits[x + y*w]
			pRow1 = pBits1 + xs - x1 + (m_Texture[n1].bh-1 - (y-y1) )*m_Texture[n1].w;
			pRow2 = pBits2 + xs - x2 + (m_Texture[n2].bh-1 - (y-y2) )*m_Texture[n2].w;
			for (x=xs; x<xe; x+=2)
			{
				DBG
				if (*pRow1 && *pRow2) return true;// Got Collision on both bits
				pRow1+=2; pRow2+=2;
			}
		}
		// odd cols
		++xs;
		for (y=ys; y<ye; y+=2)
		{
			// pBits[x][y] = pBits[x + y*w]
			pRow1 = pBits1 + xs - x1 + (m_Texture[n1].bh-1 - (y-y1) )*m_Texture[n1].w;
			pRow2 = pBits2 + xs - x2 + (m_Texture[n2].bh-1 - (y-y2) )*m_Texture[n2].w;
			for (x=xs; x<xe; x+=2)
			{
				DBG
				if (*pRow1 && *pRow2) return true;// Got Collision on both bits
				pRow1+=2; pRow2+=2;
			}
		}
#undef DBG
#endif
	return false;
}



//---------------------------------------------------------------//
//---------------------------------------------------------------//
bool OpenGLRainbows::LoadSurface(int num, const char* szFileName, bool bGrayScale)
{
	OPENGL_IMAGE image;
		if(!image.LoadImageBMP(szFileName, (COLORREF)-1) && !image.LoadImagePNG(szFileName, (COLORREF)-1) && !image.LoadImageJPG(szFileName))
		{
				return false;
		}
		if (image.m_sizex==0 || image.m_sizey==0) return false;

	#if defined(TINYGL) || defined(TRENKI3D)
		m_Surface[num].CreateBy(image, (COLORREF)-1);
	#else
#ifdef GLB_SURFACE_OLDSCHOOL
		int			tx, ty;
		int			sztex;
		int			bx, by;

		glGetIntegerv(GL_MAX_TEXTURE_SIZE, &sztex);
		sztex=min(512, sztex); // don't make too big, otherwise "grabbing" takes so so long.

		tx=ty=0;
		this->FreeSurface(num);
		for (by=0; by<image.m_sizey; by+=sztex)
		{
			for (bx=0; bx<image.m_sizex; bx+=sztex)
			{
				if (!MakeTexture(image, bx, by, sztex, sztex, &this->m_Surface[num].txt[tx][ty], NULL, false, false))
					return false;
				m_Surface[num].sizx [tx][ty]=sztex;
				m_Surface[num].sizy [tx][ty]=sztex;
				m_Surface[num].sizbx[tx][ty]=sztex;
				m_Surface[num].sizby[tx][ty]=sztex;
				tx++;
			}
			this->m_Surface[num].usex=max(tx, this->m_Surface[num].usex);
			ty++;
			tx=0;
		}
		this->m_Surface[num].usey=ty;
#else // jus use a sprite
	if(m_BackSurfaceSpriteID[num] == -1)
		FreeSprite(m_BackSurfaceSpriteID[num]);

	m_BackSurfaceSpriteID[num] = this->GenSprite();
	if(!this->LoadSprite(m_BackSurfaceSpriteID[num], image, -1, false))
	{
		m_BackSurfaceSpriteID[num] = -1;
		return false;
	}
#endif
#endif
return true;
}

// --------------------------------------------------------------- //
// Get cookie color and fix the 16 bit rounding error
// returned as BIG endian
// --------------------------------------------------------------- //
unsigned int OpenGLRainbows::GetTrueCookieColor()
{
	// pixels = 0x 80 00 ff 00 - BGRA
	// pP     = 0x 00 ff 00 80 - ARGB (Little Endian on x86)
	unsigned long cookie;
	if(m_CookieColor != 0)
	{
		COLORREF old= GetPixel(4,4);
		DrawPixel(4,4,m_CookieColor);
		cookie = GetPixel(4,4);
		cookie = ToBigEndian((unsigned long)cookie);
		DrawPixel(4,4,old);
	}
	else
		cookie = ToBigEndian((unsigned long)m_CookieColor);
return (int)cookie;
}


// --------------------------------------------------------------- //
// Grab a Sprite from BackBuffer
// --------------------------------------------------------------- //
bool OpenGLRainbows::GrabSprite(int num, int x, int y, int wx, int wy)
{
//	this->FreeSprite(num); // might want to reuse the texture

	m_Texture[num].th = m_Texture[num].tw = 0;

#if defined(TINYGL) || defined(TRENKI3D)
	m_Texture[num].w =(short)wx; // Set Size of Texture
	m_Texture[num].h =(short)wy;
	m_Texture[num].bw=(short)wx; // Size of grabbed image
	m_Texture[num].bh=(short)wy;
	m_Texture[num].txx=1.0f;
	m_Texture[num].txy=1.0f;

	m_Sprite[num].GrabFrom(x,y,x+wx, y+wy, m_CookieColor);
	return true;
#else

	// Force clearing of collision pixels!!
	#ifdef SPRITE_ALPHA_PIX_AS_VECTOR
		SpritePix[num].Clear();
	#else
		if (SpritePix[num]) delete[] SpritePix[num];
		SpritePix[num]=NULL;
	#endif

	if(wx==0 || wy==0)
	{
		FreeSprite(num);
		return true;
	}

	int	sx, sy;
	// Create 2^n size
	sx=GLB_MIN_TEX_SIZE/2; sy=GLB_MIN_TEX_SIZE/2;
	do{sx<<=1;} while(sx < wx);
	do{sy<<=1;} while(sy < wy);


	// do not reallocate if not neccessary!
	if (m_Texture[num].w != sx || m_Texture[num].h != sy)
	{
		if(m_Texture[num].mipstate[0])
			glDeleteTextures(1, &m_Texture[num].mipstate[0]);
		if(m_Texture[num].mipstate[1])
			glDeleteTextures(1, &m_Texture[num].mipstate[1]);
		m_Texture[num].mipstate[0]=0;
		m_Texture[num].mipstate[1]=0;

		// This is specific to the binary format of the data
		glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
		#if defined(HAVE_OPENGL) && !defined(EMSCRIPTEN)
		glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
		glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
		glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
		#endif
		// Generate & Bind into Memory
		glGenTextures(1, &m_Texture[num].mipstate[0]);
		glBindTexture(GL_TEXTURE_2D, m_Texture[num].mipstate[0]);
		// Set Texture mapping parameters
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

		GLenum format = GL_RGBA;
		GLint  intformat = GL_RGBA8;
		#if defined(HAVE_OPENGL_ES) || defined(EMSCRIPTEN)
			intformat = GL_RGBA;
		#endif
		// OpenGL|ES does not have BGRA_EXT
		if(m_bSupportBGRA_ext)
		{
			format = GL_BGRA_EXT;
		}
		// must use RGBA, BGRA is not supported
		// glTexImage2D(...      GL_UNSIGNED_SHORT_5_6_5, NULL);
		glTexImage2D(GL_TEXTURE_2D, 0, intformat , sx, sy, 0, format    , GL_UNSIGNED_BYTE, NULL);
	}
	else
	{
		glBindTexture(GL_TEXTURE_2D, m_Texture[num].mipstate[0]);
	}

	m_Texture[num].w =(short)sx; // Set Size of Texture
	m_Texture[num].h =(short)sy;
	m_Texture[num].bw=(short)wx; // Size of grabbed image
	m_Texture[num].bh=(short)wy;
	m_Texture[num].txx=(float)m_Texture[num].bw/m_Texture[num].w;
	m_Texture[num].txy=(float)m_Texture[num].bh/m_Texture[num].h;


#if 0 // crap! glGetTexLevelParameteriv missing on iPhone
	// #iPhone SDL has a Frambuffer instead of a screen
	//         -> also make GRASPRITE work from within USESCREEN
	GLint curr_fbo = FBUFFER::GetActiveFrameBuffer();
	if(curr_fbo)
	{
		FBUFFER fbo;
		fbo.Create(sx,sy, false); // no depth buffer
		if(fbo.StartRendering(m_Texture[num].mipstate[0]))
		{
			Make2DViewport(0,0,sx,sx);
			glViewport(0, 0, sx,sy);

			// curr_fbo is the texture index for the texture to grab from
			// find the size
			GLint org_w=0,  org_h=0;
			glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH,  &org_w);
			glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &org_h);
			// render the quad
			GLfloat   xy[] = {
							wx, 0,
							0 , 0,
							0 , wy,
							wx, wy
							};
			GLfloat   uv[] = {
							(GLfloat)(x+wx)/org_w, (GLfloat)(wy-y)/org_h,
							(GLfloat)     x/org_w, (GLfloat)(wy-y)/org_h,
							(GLfloat)     x/org_w, (GLfloat)(org_h-wy-y   )/org_h,
							(GLfloat)(x+wx)/org_w, (GLfloat)(org_h-wy-y   )/org_h
							};
			glEnableClientState( GL_VERTEX_ARRAY );
			glVertexPointer( 2, GL_FLOAT,    0, (char*)xy);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, 0, (char*)uv);
			glbEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, curr_fbo);

			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
			glDrawArrays(GL_TRIANGLE_FAN, 0, 4);	// Draw All Of The Triangles At Once

			// Disable Pointers
			glDisableClientState( GL_VERTEX_ARRAY );				// Disable Vertex Arrays
			glDisableClientState( GL_TEXTURE_COORD_ARRAY );			// Disable Texture Coord Arrays
		
			fbo.StopRendering(); // restore old frame buffer
			Make2DViewport(0,0,0,0);
		}
		return true;
	}
#endif


	static int has_alpha=-1;
	if(has_alpha == -1)
	{
		glGetIntegerv(GL_ALPHA_BITS, &has_alpha); // 8 bits usually
	}



	int quickmode = 0;

	#ifndef PALM_PIXI // pixi won't do glCopyTexSubImage2D here :(
		if(has_alpha && m_landscape==0)
			quickmode=1;
	#endif


	if(quickmode)
	{
		// Here comes the magic
		// glCopy/glRead had x,y(0,0) = bottom left corner - forced
		//  x,y,w,h = 0,368,32,32
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, x, m_WND.height-y-wy, wx, wy);
	}
	else // no alpha bits -> fix transparancy for compatibility
	{
		static OPENGL_IMAGE img; // saves a lot re-allocations

		// WHAT A MESS! From now on, flip "sx,sy", grab at new screen position and rotate image later
		int xgrab = x;
		int ygrab = m_WND.height-y-wy;

		if(m_landscape)
		{
			if(m_landscape==1)
			{
				// this case is very special because of grab!=pow2
				int nx, ny;
				nx = m_WND.width - y - wy;
				ny = x;
				x=nx;
				y=ny;
				nx = sx; sx=sy; sy=nx;
				nx = wx; wx=wy; wy=nx;
			
				// fix grab positions
				xgrab = x;
				ygrab = m_WND.height-sy-y;
			}
			else
			{
				int xend=xgrab+sx-1;
				int yend=ygrab+sy-1;
				Rot_Display2Screen(xgrab, ygrab, &xgrab, &ygrab);
				Rot_Display2Screen(xend,  yend,  &xend,  &yend);
				if(xend<xgrab)
				{
					int swp = xend; xend=xgrab; xgrab=swp;
				}
				if(yend<ygrab)
				{
					int swp = yend; yend=ygrab; ygrab=swp;
				}
				wx = xend-xgrab+1;
				wy = yend-ygrab+1;
				sx=GLB_MIN_TEX_SIZE/2; sy=GLB_MIN_TEX_SIZE/2;
				do{sx<<=1;} while(sx < wx);
				do{sy<<=1;} while(sy < wy);
			}
		}

		int readx = wx, ready=wy;
		#ifdef GLREAD_POWER_OF_2
			readx = sx;
			ready = sy;
		#endif

		img.resize(readx, ready);

		if(m_bSupportBGRA_ext)
		{
			glReadPixels(xgrab,ygrab,readx,ready,GL_BGRA_EXT, GL_UNSIGNED_BYTE, img.m_txdata);
		}
		else
		{
			glReadPixels(xgrab,ygrab,readx,ready,GL_RGBA    , GL_UNSIGNED_BYTE, img.m_txdata);
		}

		img.expand(sx, sy); // make image power of 2 now (for texture generation)


		if(!has_alpha) // no alpha: fix alpha bits by cookie color -> this is going to be really slow now
		{
			// pixels = 0x 80 00 ff 00 - BGRA
			// pP     = 0x 00 ff 00 80 - ARGB (Little Endian on x86)
			unsigned int cookie=GetTrueCookieColor();

			// DrawPixel might mess this up - uh oh!
			glBindTexture(GL_TEXTURE_2D, m_Texture[num].mipstate[0]);

			unsigned int* pP = (unsigned int*)img.m_txdata, *pEnd = (unsigned int*)img.m_txdata + (sx*sy);
		
			if(m_bSupportBGRA_ext)
			{
				// RGBA instead of BGRA
				// little endian the cookie
				cookie = (int)FromBigEndian((unsigned long)cookie);
				// cookie &= 0x00ffffff;
				do
				{
					if( ((*pP) & 0x00ffffff) == cookie)
					{
						*pP &= 0x00ffffff;
					}
					++pP;
				}while(pP < pEnd);
			}
			else
			{
				if(GF_BIG_ENDIAN)
				{
					// 0xbbggrraa
					do
					{
						if( ((*pP) & 0xffffff00) == cookie)
						{
							*pP &= 0xffffff00;
						}
						++pP;
					}while(pP < pEnd);
				}
				else
				{
					// 0xaarrggbb
					cookie >>= 8;
					do
					{
						if( ((*pP) & 0x00ffffff) == cookie)
						{
							*pP &= 0x00ffffff;
						}
						++pP;
					}while(pP < pEnd);
				}
			}
		} // has_alpha
		

		if(m_landscape)
		{
			for(int i=0; i<m_landscape; ++i)
				img.rotate90right();
			// img.rotate90right();
			
			if(m_landscape==1 || m_landscape==3)
			{
				int swp=sx; sx=sy;sy=swp;
			}
		}
	
		#if 1 // fix formats for gles
			#ifdef PALM_PIXI
				img.clip(wx,wy);
				if(m_bSupportBGRA_ext) // deffo need rgba, not bgra here
					img.swap_red_blue();
				img.flipy();
				MakeTexture(img, 0,0,wx,wy, &m_Texture[num].mipstate[0], NULL, false, false);
			#else
				GLenum format = GL_RGBA;
				GLint  intformat = GL_RGBA8;
				#if defined(HAVE_OPENGL_ES) || defined(EMSCRIPTEN)
					intformat = GL_RGBA;
				#endif
				// OpenGL|ES does not have BGRA_EXT
				if(m_bSupportBGRA_ext)
				{
					format = GL_BGRA_EXT;
				}
 				glTexImage2D(GL_TEXTURE_2D, 0, intformat, sx,sy, 0, format, GL_UNSIGNED_BYTE, img.m_txdata);
			#endif

		#else
			if(m_bSupportBGRA_ext)
			{
				glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, sx,sy, 0, GL_BGRA_EXT, GL_UNSIGNED_BYTE, img.m_txdata);
			}
			else
			{
				// must use RGBA, BGRA is not supported                       //GL_UNSIGNED_SHORT_5_5_5_1
				glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, sx,sy, 0, GL_RGBA     , GL_UNSIGNED_BYTE, img.m_txdata);
			}
		#endif
	}

	OPENGL_CHECK_ERRORS();
return TRUE;
#endif
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void OpenGLRainbows::GrabSurface(int num)
{
	#if defined(TINYGL) || defined(TRENKI3D)
		Screen.SetDrawBuffer();
		m_Surface[num].GrabFrom(0,0,Screen.getWidth(), Screen.getHeight(), m_CookieColor);
	#else
#ifdef GLB_SURFACE_OLDSCHOOL
		int	sx, sy = 0, tx, ty, wx, wy;
		int	bmpx, bmpy;
		int	max_tex;
		// Maximum of one sprite for the surface
		glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_tex);
		bmpx = this->m_WND.width;
		bmpy = this->m_WND.height;
	
		// find minimum required texture size (new cards do 8000x8000, which is _sloooow_ then)
		for(tx=128; tx<max(bmpx, bmpy) && tx<max_tex; tx*=2)
		{}
		max_tex = tx;
	
		int bx=0, by=0;
		tx=ty=0;
	//	this->FreeSurface(num);

		for (by=0; by<bmpy; by+=max_tex) //+64
		{
			for (bx=0; bx<bmpx; bx+=max_tex)
			{
				sx=max_tex;
				sy=max_tex;
				if(m_Surface[num].sizx[tx][ty] != sx || m_Surface[num].sizy[tx][ty]!=sy)
				{
					// This is specific to the binary format of the data read in.
					glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
					#if defined(HAVE_OPENGL) && !defined(EMSCRIPTEN)
					glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
					glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
					glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
					#endif

					// Now Bind the Texture to Memory
					glGenTextures(1, &this->m_Surface[num].txt[tx][ty]);
					glBindTexture(GL_TEXTURE_2D, this->m_Surface[num].txt[tx][ty]);
					// Set Texture mapping parameters
					glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
					glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
					glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
					glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

				#if 1 // fix formats for gles
					GLenum format = GL_RGBA;
					GLint  intformat = GL_RGBA8;
					#if defined(HAVE_OPENGL_ES) || defined(EMSCRIPTEN)
						intformat = GL_RGBA;
					#endif
					// OpenGL|ES does not have BGRA_EXT
					if(m_bSupportBGRA_ext)
					{
						format = GL_BGRA_EXT;
					}
					glTexImage2D(GL_TEXTURE_2D, 0, intformat, sx,sy, 0, format, GL_UNSIGNED_BYTE, NULL);
				#else
					// #if defined(HAVE_OPENGL)
					if(m_bSupportBGRA_ext)
					{
						glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, sx, sy, 0, GL_BGRA_EXT, GL_UNSIGNED_BYTE, NULL);
					}
					else
					{
						//#elif defined(HAVE_OPENGL_ES)
						// must use RGBA, BGRA is not supported
						glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, sx, sy, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
					}
					// #endif
				#endif
				}
				else
					glBindTexture(GL_TEXTURE_2D, this->m_Surface[num].txt[tx][ty]);
			
				wx = sx;
				wy = sy;
				if (bx+sx>bmpx) wx = bmpx-bx;
				if (by+sy>bmpy) wy = bmpy-by;
			

				if(m_bSupportBGRA_ext)
				{
					glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, bx, bmpy-by-wy, wx, wy);
				}
				else
				{
					static OPENGL_IMAGE img; // safes re-allocation a lot
					img.resize(sx,sy);
					// must use RGBA, BGRA is not supported
					glReadPixels(0,0,sx,sy,GL_RGBA    , GL_UNSIGNED_BYTE, img.m_txdata);
					glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, sx,sy, 0, GL_RGBA     , GL_UNSIGNED_BYTE, img.m_txdata);
			
				}
				#ifdef HAVE_OPENGL
				#else
				#endif
				this->m_Surface[num].sizx[tx][ty]=sx;
				this->m_Surface[num].sizy[tx][ty]=sy;
				this->m_Surface[num].sizbx[tx][ty] = wx;
				this->m_Surface[num].sizby[tx][ty] = wy;

				tx++;
			}

			long from = bmpy-by-sy, to = bmpy-by;

			this->m_Surface[num].usex=tx;
			ty++; tx=0;
		}
		this->m_Surface[num].usey=ty;
#else // jus use a sprite
	if(m_BackSurfaceSpriteID[num]==-1)
		FreeSprite(m_BackSurfaceSpriteID[num]);
	m_BackSurfaceSpriteID[num] = GenSprite();
	GrabSprite(m_BackSurfaceSpriteID[num], 0,0,m_WND.virtw, m_WND.virth); // GF_TODO prescaler - I think it's not gonna work...
#endif
	#endif



}

#ifdef EMSCRIPTEN
	#define glReadBuffer(a) {} // bug in Firefox. Not required in HTML5?
#endif

// ------------------------------------------------------------- //
// Get a Pixel's color (approximately)
// ------------------------------------------------------------- //
COLORREF OpenGLRainbows::GetPixel(int x, int y)
{
	Rot_Display2Screen(x, y, &x, &y);

#if defined(TINYGL) || defined(TRENKI3D)
	return UnPackColor(Screen.GetPixel(x,y));
#else
	unsigned char RGBA_pix[4] = {0,0,0,0xff};

	GLenum format = GL_RGBA;
	if(m_bSupportBGRA_ext) format = GL_BGRA_EXT;

	#if defined(HAVE_OPENGL)
		if(m_CurrentFBO>=0)
		{
			glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
			glReadPixels(x, m_Texture[m_FBuffer[m_CurrentFBO].m_SpriteIndex].bh-y-1, 1,1, format, GL_UNSIGNED_BYTE, (void*)RGBA_pix);
		}
		else
		{
			glReadBuffer(GL_BACK); // Select BackBuffer for reading
			glReadPixels(x, m_WND.height-y-1, 1,1, format, GL_UNSIGNED_BYTE, (void*)RGBA_pix);
		}
	#elif defined(HAVE_OPENGL_ES)
		glReadPixels(x, m_WND.height-y-1, 1,1, format, GL_UNSIGNED_BYTE, (void*)RGBA_pix);
	#endif

	if(m_bSupportBGRA_ext)
	{
		return RGB(RGBA_pix[2], RGBA_pix[1], RGBA_pix[0]);
	}
	return RGB(RGBA_pix[0], RGBA_pix[1], RGBA_pix[2]);
#endif
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void OpenGLRainbows::DrawPixel(int sx, int sy, COLORREF col)
{
#if defined(TINYGL) || defined(TRENKI3D)
	Screen.PutPixel((int)sx,(int)sy, PackColor(col));
#else
	#if 1 // #ifdef HAVE_OPENGL_ES // other code is breaking or not really a speedup...
		DrawRect(sx,sy,1,1,col);
		return;
    #else
		if(this->m_WND.virtw != m_WND.virth)
		{
			DrawRect(sx,sy,1,1,col);
			return;
		}


        glbColor4f(	(1.0f/255.0f)*(float)GetRValue(col),
				(1.0f/255.0f)*(float)GetGValue(col),
				(1.0f/255.0f)*(float)GetBValue(col),
				col == m_CookieColor ? 0.0f : m_falpha);

		// static int oldtex[3]; oldtex[0]=m_Texture3D[0]; oldtex[1]=m_Texture3D[1]; oldtex[2]=m_Texture3D[2];
		SetTexture3D(-1, -1, 0);

		#if 1 // always - old code crashes.
			// int pts[] = {static_cast<int>((float)sx), static_cast<int>((float)sy)};
			int pts[] = {sx, sy}; // static_cast<int>((float)sx), static_cast<int>((float)sy)};
			glEnableClientState( GL_VERTEX_ARRAY );
			glVertexPointer( 2, GL_FLOAT, 0, (char*)pts);
			glDrawArrays(GL_POINTS, 0, 1);
			glDisableClientState( GL_VERTEX_ARRAY );
		#else
			// GF: 2015-Aug this crashes on my PC. (gfx card driver update?)
			glMatrixMode(GL_PROJECTION); glPushMatrix(); glTranslatef(0.375, 0.375, 0); glMatrixMode(GL_MODELVIEW);
			glBegin(GL_POINTS);
				glVertex2i(sx, sy);
			glEnd();
			glMatrixMode(GL_PROJECTION); glPopMatrix(); glMatrixMode(GL_MODELVIEW);
		#endif

		glbColor4f(1.0f, 1.0f, 1.0f, m_falpha);
		// Restore state
		// SetTexture3D(oldtex[0], oldtex[1], oldtex[2]);
    #endif

#endif
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void OpenGLRainbows::DrawLine(int sx, int sy, int ex, int ey, COLORREF col)
{
#if defined(TINYGL) || defined(TRENKI3D)
	Screen.DrawLine((int)sx,(int)sy,(int)ex,(int)ey,PackColor(col));
#else
	#if defined(HAVE_OPENGL_ES)
		//glbEnable(GL_COLOR_MATERIAL);
		SetTexture3D(-1, -1, 0);

		glMatrixMode(GL_PROJECTION); glPushMatrix(); glTranslatef(0.375, 0.375, 0); glMatrixMode(GL_MODELVIEW);

		unsigned char cols[8]= {(unsigned char)GetRValue(col),
								(unsigned char)GetGValue(col),
								(unsigned char)GetBValue(col),
								static_cast<unsigned char>(col == m_CookieColor ? 0 : (unsigned char)(255.0f*m_falpha))};
		cols[4]=cols[0];
		cols[5]=cols[1];
		cols[6]=cols[2];
		cols[7]=cols[3];

		float pts[] = {(float)sx,(float)sy,(float)ex,(float)ey};
		glEnableClientState( GL_VERTEX_ARRAY );
		glVertexPointer( 2, GL_FLOAT, 0, (char*)pts);
 
  		glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(4, GL_UNSIGNED_BYTE, 0, (char*)&cols[0]);

		glDrawArrays(GL_LINES, 0, 2);

		glDisableClientState( GL_VERTEX_ARRAY );
		glDisableClientState( GL_COLOR_ARRAY );

		glMatrixMode(GL_PROJECTION); glPopMatrix(); glMatrixMode(GL_MODELVIEW);
	#else
		glbColor4f(	(1.0f/255.0f)*(float)GetRValue(col),
					(1.0f/255.0f)*(float)GetGValue(col),
					(1.0f/255.0f)*(float)GetBValue(col),
					col == m_CookieColor ? 0.0f : m_falpha);
	// static int oldtex[3]; oldtex[0]=m_Texture3D[0]; oldtex[1]=m_Texture3D[1]; oldtex[2]=m_Texture3D[2];
		SetTexture3D(-1, -1, 0);
		glMatrixMode(GL_PROJECTION); glPushMatrix(); glTranslatef(0.375, 0.375, 0); glMatrixMode(GL_MODELVIEW);
		glBegin(GL_LINES);
			glVertex2i(sx, sy);
			glVertex2i(ex, ey);
		glEnd();
		glBegin(GL_POINTS); glVertex2i(ex, ey); glEnd();
		glMatrixMode(GL_PROJECTION); glPopMatrix(); glMatrixMode(GL_MODELVIEW);
		glbColor4f(1.0f, 1.0f, 1.0f, m_falpha);
	#endif

	// Restore state
	// SetTexture3D(oldtex[0], oldtex[1], oldtex[2]);
#endif
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void OpenGLRainbows::DrawRect(int sx, int sy, int wx, int wy, COLORREF col)
{
	// make wx, wy positive
	if(wx<0){sx+=(wx+1); wx=-wx;}
	if(wy<0){sy+=(wy+1); wy=-wy;}

#if defined(TINYGL) || defined(TRENKI3D)
	Screen.PutRect((int)sx,(int)sy,(int)(wx), (int)(wy), PackColor(col));
#else
	// Huh!? Polyvector is using glColor as well...?
	#if 0 // defined(ANDROID) // polyvector workaround. Many drivers don't allow glColor with polys
		this->PolyStart(-1, 0);
		this->PolyPoint(sx   , sy   , 0.0f, 0.0f, col);
		this->PolyPoint(sx   , sy+wy, 0.0f, 0.0f, col);
		this->PolyPoint(sx+wx, sy+wy, 0.0f, 0.0f, col);
		this->PolyPoint(sx+wx, sy   , 0.0f, 0.0f, col);
		this->PolyEnd();
		return;
	#else
		glbColor4f(	(1.0f/255.0f)*(float)GetRValue(col),
					(1.0f/255.0f)*(float)GetGValue(col),
					(1.0f/255.0f)*(float)GetBValue(col),
					col == m_CookieColor ? 0.0f : m_falpha);
		SetTexture3D(-1, -1, 0);

		float fsx=(float)sx;
		float fsy=(float)sy;
		float fex=(float)(sx+wx);
		float fey=(float)(sy+wy);

		// only required for lines and points
//		glMatrixMode(GL_PROJECTION); glPushMatrix(); glTranslatef(0.375f, 0.375f, 0.0f); glMatrixMode(GL_MODELVIEW);
	
		float xyuv[]={	fsx,fey,0.0f,0.0f,
						fex,fey,0.0f,0.0f,
						fex,fsy,0.0f,0.0f,
						fsx,fsy,0.0f,0.0f};
		
		// alpha test/blending is handled by the user now. Default to blend, yes. test, no
		// if(col == m_CookieColor) glbDisable(GL_ALPHA_TEST);
		DrawQuad(xyuv);

//		glMatrixMode(GL_PROJECTION); glPopMatrix(); glMatrixMode(GL_MODELVIEW);
		glbColor4f(1.0f, 1.0f, 1.0f, m_falpha);
		// SetAlphaTesting(m_fAlphaTestingGTH);
	#endif
#endif
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void OpenGLRainbows::FreeSprite(int num)
{
	if(num<0) return;

	// #NPBUFFER
#if defined(TINYGL) || defined(TRENKI3D)
	m_Sprite[num].Delete();
#else
	#ifdef SPRITE_ALPHA_PIX_AS_VECTOR
		SpritePix[num].Clear();
	#else
		if (SpritePix[num]) delete[] SpritePix[num];
		SpritePix[num]=NULL;
	#endif

	this->SetTexture3D(-1, -1, 0); // make sure it's not bound
	if (m_Texture[num].mipstate[0]) {glDeleteTextures(1, &m_Texture[num].mipstate[0]);}
	if (m_Texture[num].mipstate[1]) {glDeleteTextures(1, &m_Texture[num].mipstate[1]);}
#endif
	m_Texture[num].mipstate[0]=0;
	m_Texture[num].mipstate[1]=0;
	m_Texture[num].w=-1;
	m_Texture[num].h=-1;
	m_Texture[num].bw=0;
	m_Texture[num].bh=0;
	m_Texture[num].txx=1.0f;
	m_Texture[num].txy=1.0f;
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void OpenGLRainbows::FreeSurface(int num)
{
	if(num<0) return;
#if defined(TINYGL) || defined(TRENKI3D)
	m_Surface[num].Delete();
#else

#ifdef GLB_SURFACE_OLDSCHOOL
	for (int x=0; x<this->m_Surface[num].usex; x++)
	{
		for (int y=0; y<this->m_Surface[num].usey; y++)
		{
			if(m_Surface[num].txt[x][y])
				glDeleteTextures(1, &m_Surface[num].txt[x][y]);
		}
	}
	memset(&m_Surface[num], 0, sizeof(OPENGL_RAINBOWS_SURFACE));
#else // jus use a sprite
	FreeSprite(m_BackSurfaceSpriteID[num]);
	m_BackSurfaceSpriteID[num]=-1;
#endif


#endif
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void OpenGLRainbows::FreeAll()
{
int	i;
	for (i=0; i<GLR_MAX_TEXTURES; i++)	this->FreeSprite(i);
	for (i=0; i<GLR_MAX_SURFACE; i++)	this->FreeSurface(i);
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
int OpenGLRainbows::GenSprite()
{
	for(DGNat i=GLR_MAX_TEXTURES - GLR_MAX_FONT-1; i>=0; --i)
		if(GetSpriteSize(i).cx == 0)
			return i;
return -1;
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
int OpenGLRainbows::GenFont()
{
	for(DGNat i=GLR_MAX_FONT - 1; i>=0; --i)
		if(GetFontSize(i).cx == 0)
			return i;
return -1;
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
int OpenGLRainbows::GenFBuffer()
{
	for(DGNat i=GLR_MAX_FRAMEBUFFER - 1; i>=0; --i)
		if(m_FBuffer[i].m_Width == 0 || m_FBuffer[i].m_SpriteIndex<0 )
			return i;
return -1;
}


// ------------------------------------------------------------- //
// SAVE BMP FILE
// ------------------------------------------------------------- //
bool OpenGLRainbows::SaveBMPFile (const char* szFilename)
{
	int id = GenSprite();
	if(id==-1) return false;
	if(!this->GrabSprite(id, 0,0,m_WND.width, m_WND.height))
		return false;
	bool ok = this->SaveBMPSprite(id, szFilename);
	this->FreeSprite(id);
	return ok;
}

// ------------------------------------------------------------- //
// get the sprite's texture (visible size)
// ------------------------------------------------------------- //
bool OpenGLRainbows::GetSpriteRGBA(int nSprite, OPENGL_IMAGE& img)
{
	SIZE sz = GetSpriteSize(nSprite);
	SIZE realsz = sz;
	if(sz.cx==0 || sz.cy==0) return false;

#ifdef GLREAD_POWER_OF_2
	int pw2sx=1, pw2sy=1;
	do{pw2sx<<=1;} while(pw2sx < sz.cx);
	do{pw2sy<<=1;} while(pw2sy < sz.cy);

	sz.cx = pw2sx;
	sz.cy = pw2sy;
#endif

	img.resize((int)sz.cx, (int)sz.cy);
#if defined(TINYGL) || defined(TRENKI3D)
	oBitmap& bmp = m_Sprite[nSprite];

	unsigned char* pT = img.m_txdata;
	for(int y=0; y<bmp.m_Height; ++y)
	for(int x=0; x<bmp.m_Width; ++x)
	{
		unsigned short* pC = bmp.Pixels() + x + (sz.cy-1-y)*bmp.m_Width;
		*pT++ = (unsigned char)GETB(*pC);
		*pT++ = (unsigned char)GETG(*pC);
		*pT++ = (unsigned char)GETR(*pC);
		*pT++ = (*pC == bmp.m_colorkey) ? 0 : 255;
	}
	img.flipy();
#else
	// Use a PBuffer and glReadPixels
	if(m_CurrentFBO >=0)
		FBO_end();

	FBUFFER pbuffer;
	if(pbuffer.Create(m_Texture[nSprite].w, m_Texture[nSprite].h, m_Texture[nSprite].mipstate[0], false))
	{
		pbuffer.Bind(); // bind mip-mapping texture
		glReadPixels(0,0, (int)sz.cx, (int)sz.cy, GL_RGBA, GL_UNSIGNED_BYTE, &img.m_txdata[0] ); // #NEED hogs memory with GCC :(
		pbuffer.Unbind();
		img.clip((int)realsz.cx, (int)realsz.cy);
		img.flipy();
	}
#endif
	return true;
}


// ------------------------------------------------------------- //
// Save the sprite's texture (visible size) as a bmp file
// ------------------------------------------------------------- //
bool OpenGLRainbows::SaveBMPSprite( int nSprite, const char* szFilename)
{
/*
bool flipy=false;
#if defined(TINYGL) || defined(TRENKI3D)
	flipy=true;
	oBitmap& bmp = m_Sprite[nSprite];

	unsigned char* pPixl = new unsigned char [bmp.m_Width*bmp.m_Height*3];
	unsigned char* pT = pPixl;
	for(int y=0; y<bmp.m_Height; ++y)
	for(int x=0; x<bmp.m_Width; ++x)
	{
		unsigned short* pC = bmp.Pixels() + x + y*bmp.m_Width;
		*pT++ = (unsigned char)GETB(*pC);
		*pT++ = (unsigned char)GETG(*pC);
		*pT++ = (unsigned char)GETR(*pC);
	}
#else
	glBindTexture(GL_TEXTURE_2D, m_Texture[nSprite].mipstate[0]);
	int sz = m_Texture[nSprite].w * m_Texture[nSprite].h * 3;
	unsigned char* pPixl = new unsigned char [sz];

	#if defined(HAVE_OPENGL)
		glGetTexImage(GL_TEXTURE_2D, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE, pPixl);
	#elif defined(HAVE_OPENGL_ES)
		// #NEED
	#endif
#endif
*/

OPENGL_IMAGE img;
	if(!GetSpriteRGBA(nSprite, img))
		return false;

	bool rv = SaveImageFile(szFilename, img); // SaveMemoryToBMP(szFilename, pPixl, m_Texture[nSprite].bw, m_Texture[nSprite].bh, m_Texture[nSprite].w, m_Texture[nSprite].h, flipy); if (pPixl) delete [] pPixl;
	return rv;
}



// ------------------------------------------------------------- //
// Save a memory buffer to a BMP file
// real_width/height = number of BGR tripples(pixels) width of written image
// width_of_buffer   = number of BGR tripples(pixels) for each line in the buffer
// flip_at_y         = is the first pixel upper left corner ? true : false
// ------------------------------------------------------------- //
bool OpenGLRainbows::SaveImageFile(	const char* szFilename, OPENGL_IMAGE& img)
{
#ifdef HAS_PNG_LIB
	#define LOWR(c) (((c)>='A' && (c)<='Z') ? ((c)-'A'+'a') : ((c)))

	// reverse to BGRA
	int lng = (int)strlen(szFilename);
	if (lng>4 &&
			szFilename[lng-4] == '.' &&
			((
				LOWR(szFilename[lng-3]) == 'p' && 
				LOWR(szFilename[lng-2]) == 'n' &&
				LOWR(szFilename[lng-1]) == 'g'
			)
			||
			(
				LOWR(szFilename[lng-3]) == 'j' && 
				LOWR(szFilename[lng-2]) == 'p' &&
				LOWR(szFilename[lng-1]) == 'g'
			))
		)
	{
		// txdata = BGRA
		int ok;
		if(LOWR(szFilename[lng-3]) == 'j')
		{
			img.swap_red_blue();
			ok = gf_write_jpeg(szFilename, img.m_txdata, img.m_sizex, img.m_sizey);
		}
		else
		{
			ok = gf_write_png(szFilename, img.m_txdata, img.m_sizex, img.m_sizey);
		}
		
		// delete[] txdata;
		return ok?true:false;
	}
#endif
//	img.flipy();

BITMAPFILEHEADER fh;
BITMAPINFOHEADER bi;
	// Make sure its not a NULL filename
	if(szFilename == NULL)
		return false;

GFFileSystem gs;
	if(!gs.OpenOut(szFilename)) return false;

	// Get 4 byte aligned width
	#define BITS_PER_PIXEL 24
	#define DWORD_24BIT_WIDTHBYTES_ALIGN( width )  ((( (width )* (BITS_PER_PIXEL/8)) + 3) & ~3 )

// Buffer sizes
int Height, WidthBytes, line_offset;//, Pitch;
	Height = img.m_sizey;
	WidthBytes = DWORD_24BIT_WIDTHBYTES_ALIGN( img.m_sizex );
	line_offset = WidthBytes - img.m_sizex*3;

	//Write the file header
	((char *)&(fh . bfType))[0] = 'B';
	((char *)&(fh . bfType))[1] = 'M';
	fh . bfSize = (long)(sizeof (BITMAPINFOHEADER)+sizeof (BITMAPFILEHEADER)+WidthBytes*Height); //Size in BYTES
	fh . bfReserved1 = 0;
	fh . bfReserved2 = 0;
	fh . bfOffBits = sizeof (BITMAPINFOHEADER)+sizeof (BITMAPFILEHEADER);
	bi . biSize = sizeof (BITMAPINFOHEADER);
	bi . biWidth =img.m_sizex;
	bi . biHeight =img.m_sizey;
	bi . biPlanes = 1;
	bi . biBitCount = BITS_PER_PIXEL;
	bi . biCompression = BI_RGB;
	bi . biSizeImage = 0;
	bi . biXPelsPerMeter = 10000;
	bi . biYPelsPerMeter = 10000;
	bi . biClrUsed = 0;
	bi . biClrImportant = 0;

	// gs.Write((void *) &fh, sizeof (BITMAPFILEHEADER)); // byte alignment changed to 4 bytes on android!!!
	gs.WriteWord(fh.bfType);
	gs.WriteLong(fh.bfSize);
	gs.WriteWord(fh.bfReserved1);
	gs.WriteWord(fh.bfReserved2);
	gs.WriteLong(fh.bfOffBits);

	// gs.Write((void *) &bi, sizeof (BITMAPINFOHEADER));
	gs.WriteLong(bi.biSize);
	gs.WriteLong(bi.biWidth);
	gs.WriteLong(bi.biHeight);
	gs.WriteWord(bi.biPlanes);
	gs.WriteWord(bi.biBitCount);
	gs.WriteLong(bi.biCompression);
	gs.WriteLong(bi.biSizeImage);
	gs.WriteLong(bi.biXPelsPerMeter);
	gs.WriteLong(bi.biYPelsPerMeter);
	gs.WriteLong(bi.biClrUsed);
	gs.WriteLong(bi.biClrImportant);

	// txdata = BGR
BYTE* txdata = new BYTE[WidthBytes*Height + 4];

	for(int y = 0; y<img.m_sizey; ++y)
	{
		for(int x=0; x<img.m_sizex; ++x)
		{
			int ps, pd;
		//	ps = x*3 + (Height-y-1)*WidthBytes; // inverse Y coord for BMP format
			ps = x*3 + y*WidthBytes; // inverse Y coord for BMP format
			pd = x + (img.m_sizey-y-1)*img.m_sizex;

			// Fix Cookie Color to default GLBasic Cookie Color
			const unsigned char* pPixl = (const unsigned char*)&img.m_txdata[pd*4];
			if(pPixl[3]<255)
			{
				txdata[ps  ] = 128; // BGR
				txdata[ps+1] = 0;
				txdata[ps+2] = 255;
			}
			else
			{
				txdata[ps  ] = pPixl[2];
				txdata[ps+1] = pPixl[1];
				txdata[ps+2] = pPixl[0];
			}
		}
	}
	gs.Write(txdata, WidthBytes*Height);
	gs.Close();
	if (txdata) delete[] txdata;
return true;
}





void OpenGLRainbows::ClearScreen(COLORREF col)
{
	if(col == (COLORREF)-1)
	{
		m_bHasClearColor = false;
	}
	else
	{
		m_bHasClearColor = true;
		#if defined(TINYGL) || defined(TRENKI3D)
			Screen.SetClearColor(col);
			Screen.Clear();
		#else
			float alpha=1.0f;
			if(col == m_CookieColor)
				alpha = 0.0f;

			glClearColor(((BYTE) (col)) / 255.0f,
						 ((BYTE) (((WORD) (col)) >> 8)) / 255.0f,
						 ((BYTE) ((col) >> 16)) / 255.0f,
						 alpha
						);
			glClear(GL_COLOR_BUFFER_BIT);
		#endif
	}
}


void OpenGLRainbows::Rot_Screen2Display(int x, int y, int* pX, int* pY)const
{
	switch(m_landscape)
	{
	default: // case 0:
		*pX=x;               *pY=y; break;
	case 1:
		*pX=m_WND.width-1-y; *pY=x; break;
	case 2:
		*pX=m_WND.width-1-x; *pY=m_WND.height-1 - y; break;
	case 3:
		*pX=y;               *pY=m_WND.height-1 - x; break;
	}
}

void OpenGLRainbows::Rot_Display2Screen(int x, int y, int* pX, int* pY)const
{
	switch(m_landscape)
	{
	default: // case 0:
		*pX=x;               *pY=y; break;
	case 1:
		*pX=y;               *pY=m_WND.width-1-x;break;
	case 2:
		*pX=m_WND.width-1-x; *pY=m_WND.height-1-y; break;
	case 3:
		*pX=m_WND.height-1-y;*pY=x;  break;
	}
}

void OpenGLRainbows::Rot_Gravity2Display(int x, int y, int* pX, int* pY)const
{
	switch(m_landscape)
	{
	default: // case 0:
		*pX=x;  *pY=y;  break;
	case 1:
		*pX=-y;  *pY=x; break;
	case 2:
		*pX=-x; *pY=-y; break;
	case 3:
		*pX=y; *pY=-x;  break;
	}
}



// rect: touch->screen (must be normalized)
void OpenGLRainbows::Rot_Screen2Display(int x, int y, int w, int h, int* pX, int* pY, int* pW, int* pH)const
{
	Rot_Screen2Display(x,y,  pX, pY);
	switch(m_landscape)
	{
		default:
			*pW=w; *pH=h; break;
		case 1:
			*pW=-h; *pH=w; break;
		case 2:
			*pW=-w; *pH=-h; break;
		case 3:
			*pW=h; *pH=-w; break;
	}
}
// rect: screen->touch (must be normalized)
void OpenGLRainbows::Rot_Display2Screen(int x, int y, int w, int h, int* pX, int* pY, int* pW, int* pH)const
{
	Rot_Screen2Display(x,y,  pX, pY);
	switch(m_landscape)
	{
		default:
			*pW=w; *pH=h; break;
		case 1:
			*pW=h; *pH=-w; break;
		case 2:
			*pW=-w; *pH=-h; break;
		case 3:
			*pW=-h; *pH=w; break;
	}
}


// ------------------------------------------------------------- //
// Start drawing in 2D
// x,y,w,h - pixels on the physical screen
// ------------------------------------------------------------- //
void OpenGLRainbows::Make2DViewport(int x, int y, int w, int h, float scale)
{
#if defined(TINYGL) || defined(TRENKI3D)
	if(w<=0 || h<=0)
	{
		if(m_CurrentFBO==-1)
		{
			w=Screen.getWidth(); h=Screen.getHeight();
		}
		else
		{
			const oBitmap& obmp = m_Sprite[m_FBuffer[m_CurrentFBO].m_SpriteIndex];
			w = obmp.m_Width; // width of sprite image
			h = obmp.m_Height; // height of sprite image
		}
	}

	if(m_CurrentFBO==-1)
	{
		if(x+w > Screen.getWidth())  w=Screen.getWidth()-x;
		if(y+h > Screen.getHeight()) h=Screen.getHeight()-y;
		if(w<1) {x=Screen.getWidth()-1;  w=1;}
		if(h<1) {y=Screen.getHeight()-1; h=1;}
	}
	else
	{
		const oBitmap& obmp = m_Sprite[m_FBuffer[m_CurrentFBO].m_SpriteIndex];
		int bw = obmp.m_Width; // width of sprite image
		int bh = obmp.m_Height; // height of sprite image
		if(x+w > bw)  w=bw-x;
		if(y+h > bh)  h=bh-y;
		if(w<1) {x=bw-1;  w=1;}
		if(h<1) {y=bh-1;  h=1;}
	}
	
	
	m_WND.viewport.x=x; m_WND.viewport.y=y; m_WND.viewport.width=w; m_WND.viewport.height=h;
	Screen.Viewport(x,y,w,h);

	#ifdef TRENKI3D
		tglCullFace(GL_FRONT);
		tglLoadIdentity();
		tglEnableDepthTest(0);
		tglViewport(x,y, w,h);
		tglOrtho(0, w, h,0, -1.0f, 1.0f); // Move to upper left corner 0,0
	#endif
#else
	// Set Viewport to window dimensions
	if(m_WND.virth == 0)
	{
		m_WND.virth=m_WND.height;
		m_WND.virtw=m_WND.width;
	}

	if(w<=0 || h<=0)
	{
		x=0; y=0;
		if(m_CurrentFBO==-1)
		{
			if(m_landscape==1 || m_landscape==3)
			{
				w=m_WND.virth;
				h=m_WND.virtw;
			}
			else
			{
				w=m_WND.virtw;
				h=m_WND.virth;
			}
		}
		else
		{
			w = m_FBuffer[m_CurrentFBO].m_Width;
			h = m_FBuffer[m_CurrentFBO].m_Height;
		}
	}
	
	// w,h is device size -> might need to rotate
	if(m_CurrentFBO==-1)
	{
		glMatrixMode(GL_PROJECTION);
		OPENGL_CHECK_ERRORS();
		switch(m_landscape)
		{
		default:	
			{
			glLoadIdentity();
			// float matrix[16] = {1,0,0,0, 0,1,0,0,  0,0,1,0,  0,0,0,1 };
			// glLoadMatrixf(matrix);
			break;
			}
		case 1:
			{
			float matrix[16] = {0,-1,0,0, 1,0,0,0,  0,0,1,0,  0,0,0,1 };
			glLoadMatrixf(matrix);
			// int swp=w; w=h; h=swp;
			break;
			}
		case 2:
			{
			float matrix[16] = {-1,0,0,0, 0,-1,0,0,  0,0,1,0,  0,0,0,1 };
			glLoadMatrixf(matrix);
			break;
			}
		case 3:
			{
			float matrix[16] = {0, 1,0,0, -1,0,0,0,  0,0,1,0,  0,0,0,1 };
			glLoadMatrixf(matrix);
			// int swp=w; w=h; h=swp;
			break;
			}
		}

		int vpx, vpy;
		Rot_Screen2Display(x, y, w, h, &vpx, &vpy, &w, &h);

		// must normalize
		if(w<0)
		{
			vpx=vpx+w+1; w=-w;
		}
		if(h<0)
		{
			vpy=vpy+h+1; h=-h;
		}

		
		glMatrixMode(GL_PROJECTION);
		OPENGL_CHECK_ERRORS();
		// glLoadIdentity();
		// glRotatef(-90  ,0,0,1);
		if(m_landscape == 1 || m_landscape==3)
		{
			glViewport(vpx, m_WND.height-h-vpy, w,h);
			OPENGL_CHECK_ERRORS();
			my_glOrtho(0.0f, (float)h, (float)w, 0,  -1.0f, 1.0f); // Move to upper left corner 0,0
			OPENGL_CHECK_ERRORS();
		}
		else
		{
			glViewport(vpx, m_WND.height-h-vpy, w, h);
			OPENGL_CHECK_ERRORS();
			my_glOrtho(0.0f, (float)w, (float)h, 0,  -1.0f, 1.0f); // Move to upper left corner 0,0
			OPENGL_CHECK_ERRORS();
		}
/*
//		glViewport(m_WND.width-1 -y-h , m_WND.height-w-x, h, w);
		glViewport(m_WND.width -y-h , m_WND.height-w-x, h, w);
		glMatrixMode(GL_PROJECTION);
		// glLoadIdentity();
		// glRotatef(-90  ,0,0,1);
		float matrix[16] = {0,-1,0,0, 1,0,0,0,  0,0,1,0,  0,0,0,1 };
		glLoadMatrixf(matrix);

		glOrtho(0.0f, w, h, 0,  -1.0f, 1.0f); // Move to upper left corner 0,0
*/


		// screen coords now (not device)
		if(m_landscape==1 || m_landscape==3)
		{
			int swp=w; w=h; h=swp;
		}
	}
	else // on FBO
	{
		int height;
		if(m_CurrentFBO==-1)
			height = m_WND.height;
		else
		{
			height = m_Texture[m_FBuffer[m_CurrentFBO].m_SpriteIndex].bh; // height of sprite image
		}
		glViewport(x, height-h-y, w, h);
		OPENGL_CHECK_ERRORS();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho(0.0f, (float)w, (float)h, 0.0f,  -1.0f, 1.0f); // Move to upper left corner 0,0
		OPENGL_CHECK_ERRORS();
	}
	OPENGL_CHECK_ERRORS();

	m_WND.viewport.x=x; m_WND.viewport.y=y; m_WND.viewport.width=w; m_WND.viewport.height=h;
	
//	glTranslatef(0.375, 0.375, 0); // Do this in DRAWLINE and FILLRECT
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	if(scale != 1.0f && m_CurrentFBO==-1)
	{
		glScalef(scale, scale, scale);
	}
#endif
}

// ------------------------------------------------------------- //
// Make Default OpenGL settings for 2D mode. Puts viewport to full screen
// ------------------------------------------------------------- //
void OpenGLRainbows::Make2DMode()
{
	Make2DViewport();

#if defined (HAVE_OPENGL) || defined(HAVE_OPENGL_ES)
	// Options for 2D Mode:
	glbUnEnableAll(); OPENGL_CHECK_ERRORS();

	// Enable textureing
	glbDisable(GL_DEPTH_TEST);	// Drawing Order=Z-Order!
	OPENGL_CHECK_ERRORS();
	glbDisable(GL_CULL_FACE);	// No culling by default (for 2D anyway)
	glFrontFace(GL_CCW);		// Counter clock-wise polygons face out
	OPENGL_CHECK_ERRORS();

	glbColor4f(1.0f, 1.0f, 1.0f, 1.0f);
	OPENGL_CHECK_ERRORS();
	// glClearColor(0.0f, 0.0f, 0.0f, 1.0f); // alpha bits are set(=!cookiecolor) -> use ClearScreen() instead

	glbDisable(GL_FOG);
	glbDisable(GL_LINE_SMOOTH);
	glbDisable(GL_POINT_SMOOTH);
	OPENGL_CHECK_ERRORS();

	glbDisable(GL_STENCIL_TEST);
	glbDisable(GL_NORMALIZE);
	glbDisable(GL_LIGHTING);
	OPENGL_CHECK_ERRORS();
	// Disable Antialaising
	glShadeModel(GL_FLAT);
	glbDisable(GL_COLOR_MATERIAL);
	glbDisable(GL_DITHER);
	OPENGL_CHECK_ERRORS();

	#if defined(HAVE_OPENGL) && !defined(EMSCRIPTEN)
		glbDisable(GL_POLYGON_SMOOTH);
		glbDisable(GL_POLYGON_OFFSET_POINT);
		glbDisable(GL_AUTO_NORMAL);
	#endif

	#if defined(HAVE_OPENGL) && !defined(EMSCRIPTEN)
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
		OPENGL_CHECK_ERRORS();
		glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
	#endif
	OPENGL_CHECK_ERRORS();
#endif
	SetSmoothShading(true);

	// glbEnable(GL_TEXTURE_2D);
	SetTexture3D(-1, -1, 0);					// No Texture

	// Cookie Cut - new defaults
	#ifdef TRENKI3D
		SetAlphaTesting( 0.05f );
		DoAlpha( 0.0f );
	#else
		SetAlphaTesting( 1.0f ); // 0.05f was old default
		DoAlpha(-1.0f);
	#endif
}




// ------------------------------------------------------------- //
// Make Default Font (5x6)
// ------------------------------------------------------------- //
void OpenGLRainbows::DefaultFont(int num)
{
	m_FontScaling[num] = 4.0;
	OPENGL_IMAGE img;
	img.MakeDefFont();

#if 0
	if (this->m_WND.width>=640) {img.ScaleX2();}
#endif
	
#if defined(TINYGL) || defined(TRENKI3D)
	m_Font[num].CreateBy(img);
#else
	int num_tex = num + GLR_MAX_TEXTURES-GLR_MAX_FONT;

	GLR_TEXTURE& txtr = m_Texture[num_tex];
	MakeTexture(img, 0,0,0,0, &txtr.mipstate[0], NULL, false, true);
	txtr.w=img.m_sizex; // Set Size of Texture
	txtr.h=img.m_sizey;
	txtr.bw=img.m_sizex; // Set Size of BMP File
	txtr.bh=img.m_sizey;
	txtr.txx=(float)txtr.bw/txtr.w;
	txtr.txy=(float)txtr.bh/txtr.h;
	
	for(int ic = 0; ic<256; ++ic)
	{
		GLR_FONTKERNING& K = m_FontKerning[num][ic];
		K.xoff  = 0;
		K.width = img.m_sizex/16;
		int cx = ((ic%16) * txtr.bw)/16;
		K.txa   = ((float) cx)            / (float)txtr.w;
		K.txe   = ((float)(cx + K.width)) / (float)txtr.w;
	}
#endif
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
SIZE OpenGLRainbows::GetSpriteSize(int num)
{
#if defined(TINYGL) || defined(TRENKI3D)
	SIZE s; s.cx=m_Sprite[num].m_Width; s.cy=m_Sprite[num].m_Height; return s;
#else
	SIZE s; s.cx=m_Texture[num].bw; s.cy=m_Texture[num].bh; return s;
#endif
}

// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
SIZE OpenGLRainbows::GetFontSize(int num)
{
	if(m_ttf_fonts[num].IsLoaded())
	{
		SIZE sz;
		sz.cx = (LONG)m_ttf_fonts[num].max_char_width;
		sz.cy = (LONG)(fabs(m_ttf_fonts[num].max_ascender) + fabs(m_ttf_fonts[num].max_descender));
		return sz;
	}


#if defined(TINYGL) || defined(TRENKI3D)
	SIZE s; s.cx=m_Font[num].m_cwidth; s.cy=m_Font[num].m_cheight; return s;
#else
	num += (GLR_MAX_TEXTURES-GLR_MAX_FONT);
	SIZE s; s.cx=m_Texture[num].bw/16; s.cy=m_Texture[num].bh/8;
	// 256 char set
	if(m_Texture[num].bw < m_Texture[num].bh)
		s.cy/=2;
	return s;
#endif
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void OpenGLRainbows::ShowDemo(int xpos, float alpha)
{
#if defined(TINYGL) || defined(TRENKI3D)
#define DRWLINE(a,b,c,d,e) Screen.DrawLine((short)(a+xpos), (short)b,(short)(c+xpos),(short)(d),e)
		Screen.Alphamode(255);
		unsigned short c = PackColor(255,255,255), c2 = PackColor(0,255,128);
		DRWLINE(0,0,4,4,c);
		DRWLINE(0,0,0,9,c);
		DRWLINE(0,9,4,4,c);

		DRWLINE(5,0,9,0,c2);
		DRWLINE(5,0,5,9,c2);
		DRWLINE(5,9,9,9,c2);
		DRWLINE(7,3,9,3,c2);
		DRWLINE(7,6,9,6,c2);

		DRWLINE(11,0,15,0,c);
		DRWLINE(11,0,11,9,c);
		DRWLINE(15,0,15,9,c);
		DRWLINE(13,3,13,9,c);
		DRWLINE(15,3,15,9,c);

		DRWLINE(17,0,21,0,c);
		DRWLINE(17,0,17,9,c);
		DRWLINE(21,0,21,9,c);
		DRWLINE(17,9,21,9,c);
#else
    static BYTE tmap[1536] = {
	59,72,209,56,68,196,51,63,182,47,59,169,44,54,155,40,49,141,36,44,128,32,39,114,28,34,100,24,30,87,20,24,73,16,20,59,12,15,46,8,10,32,4,6,18,0,1,5,
	63,77,223,61,74,213,56,69,198,52,64,184,48,59,169,44,53,154,39,48,139,35,43,124,31,38,109,27,33,95,22,27,80,18,22,65,14,17,50,9,12,35,5,7,20,2,3,7,
	63,77,215,68,80,201,97,106,199,95,103,189,96,103,178,88,95,164,85,91,153,82,87,143,79,83,133,76,80,123,73,76,112,73,75,105,70,72,95,65,67,83,44,45,56,13,13,17,
	60,73,209,60,72,201,201,202,219,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,253,253,253,89,89,96,1,1,6,
	57,70,202,57,70,202,102,110,196,253,253,253,253,253,253,254,254,254,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,253,194,194,197,6,7,20,1,1,6,
	55,68,195,55,68,195,55,68,193,190,193,223,188,194,214,21,66,196,14,63,206,12,61,204,12,61,204,12,61,204,17,66,210,67,96,182,242,242,242,56,58,72,5,7,20,1,1,6,
	53,65,187,53,65,187,53,65,187,84,92,184,247,247,248,77,107,197,0,57,224,0,57,224,0,57,224,0,57,224,1,55,212,182,190,212,148,148,156,10,12,36,5,7,20,1,1,6,
	51,63,180,51,63,180,51,63,180,50,62,179,160,164,207,224,227,234,171,182,214,162,173,202,162,173,206,165,176,208,185,192,216,234,235,235,33,36,61,9,12,35,5,7,20,1,1,6,
	49,60,172,49,60,172,49,60,172,49,60,172,59,69,170,225,226,235,165,204,196,59,207,176,55,211,178,73,204,176,240,241,241,134,136,151,14,17,50,9,12,35,5,7,20,1,1,6,
	47,57,165,47,57,165,47,57,165,47,57,165,46,56,164,106,111,164,236,239,238,25,222,181,0,253,200,116,211,191,227,227,230,32,36,74,14,17,50,9,12,35,5,7,20,1,1,6,
	45,55,158,45,55,158,45,55,158,45,55,158,45,55,158,43,52,148,184,185,198,184,223,214,121,206,188,242,242,242,110,112,139,18,22,65,14,17,50,9,12,35,5,7,20,1,1,6,
	42,52,150,42,52,150,42,52,150,42,52,150,42,52,150,42,51,149,76,82,145,249,249,250,254,254,254,201,202,212,24,29,80,18,22,65,14,17,50,9,12,35,5,7,20,1,1,6,
	40,50,143,40,50,143,40,50,143,40,50,143,40,50,143,40,50,143,39,48,138,161,163,188,244,244,244,68,73,116,22,27,80,18,22,65,14,17,50,9,12,35,5,7,20,1,1,6,
	38,47,135,38,47,135,38,47,135,38,47,135,38,47,135,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,14,17,50,9,12,35,5,7,20,1,1,6,
	36,44,128,36,44,128,36,44,128,36,44,128,36,44,128,255,255,255,36,44,128,36,44,124,36,42,110,27,33,95,22,27,80,255,255,255,14,17,50,9,12,35,5,7,20,1,1,6,
	34,41,121,34,41,121,34,41,121,34,41,121,34,41,121,255,255,255,40,50,143,34,41,120,31,38,109,27,33,95,40,50,143,255,255,255,14,17,50,9,12,35,5,7,20,1,1,6,
	32,39,113,32,39,113,32,39,113,32,39,113,32,39,113,40,50,143,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,40,50,143,14,17,50,9,12,35,5,7,20,1,1,6,
	30,37,106,30,37,106,30,37,106,30,37,106,30,37,106,30,37,106,30,37,106,30,37,106,29,36,105,27,33,95,22,27,80,18,22,65,14,17,50,9,12,35,5,7,20,1,1,6,
	28,34,98,28,34,98,28,34,98,28,34,98,28,34,98,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,14,17,50,9,12,35,5,7,20,1,1,6,
	25,31,91,25,31,91,25,31,91,25,31,91,25,31,91,255,255,255,25,31,91,25,31,91,255,255,255,25,30,90,22,27,80,255,255,255,14,17,50,9,12,35,5,7,20,1,1,6,
	23,29,84,23,29,84,23,29,84,23,29,84,23,29,84,255,255,255,23,29,84,23,29,84,23,29,84,23,29,83,22,27,79,255,255,255,14,17,50,9,12,35,5,7,20,1,1,6,
	21,26,76,21,26,76,21,26,76,21,26,76,21,26,76,21,26,76,21,26,76,21,26,76,21,26,76,21,26,76,21,25,75,18,22,65,14,17,50,9,12,35,5,7,20,1,1,6,
	19,24,69,19,24,69,19,24,69,19,24,69,19,24,69,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,14,17,50,9,12,35,5,7,20,1,1,6,
	17,21,61,17,21,61,17,21,61,17,21,61,17,21,61,255,255,255,17,21,61,17,21,61,17,21,61,17,21,61,17,21,61,16,20,60,14,17,50,9,12,35,5,7,20,1,1,6,
	15,18,54,15,18,54,15,18,54,15,18,54,15,18,54,255,255,255,255,255,255,255,255,255,255,255,255,15,18,54,15,18,54,15,18,54,13,16,49,9,12,35,5,7,20,1,1,6,
	13,16,46,13,16,46,13,16,46,13,16,46,13,16,46,255,255,255,13,16,46,13,16,46,13,16,46,13,16,46,13,16,46,13,16,46,12,15,45,9,12,35,5,7,20,1,1,6,
	10,13,39,10,13,39,10,13,39,10,13,39,10,13,39,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,10,13,39,9,11,34,5,7,20,1,1,6,
	8,11,32,8,11,32,8,11,32,8,11,32,8,11,32,8,11,32,8,11,32,8,11,32,8,11,32,8,11,32,8,11,32,8,11,32,8,11,32,8,10,31,5,7,20,1,1,6,
	6,8,24,6,8,24,6,8,24,6,8,24,6,8,24,96,96,96,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,96,96,96,6,8,24,6,8,24,5,6,20,1,1,6,
	4,5,17,4,5,17,4,5,17,4,5,17,4,5,17,255,255,255,96,96,96,4,5,17,4,5,17,4,5,17,96,96,96,255,255,255,4,5,17,4,5,17,3,4,16,1,1,6,
	2,3,9,2,3,9,2,3,9,2,3,9,2,3,9,255,255,255,96,96,96,2,3,9,2,3,9,2,3,9,96,96,96,255,255,255,2,3,9,2,3,9,2,3,9,0,1,5,
	0,0,2,0,0,2,0,0,2,0,0,2,0,0,2,96,96,96,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,96,96,96,0,0,2,0,0,2,0,0,2,0,0,1
	};
	static GLuint num=0;
	if(num==0)
	{
		glGenTextures(1, &num);
		glBindTexture(GL_TEXTURE_2D, num);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA , 16,32, 0, GL_RGB, GL_UNSIGNED_BYTE, tmap);
	}
	SetTexture3D(0, -1, 0);

	glEnable(GL_TEXTURE_2D);
	//glbEnable(GL_COLOR_MATERIAL);
	glBindTexture(GL_TEXTURE_2D, num);
	glbColor4f(1.0f, 1.0f, 1.0f, 1.0f);

	DoAlpha(alpha);

	float w, h;
	if(m_WND.width>500)
	{
		w=160.0f; h=80.0f;
	}
	else
	{
		w=80.0f; h=40.0f;
	}
	float xyuv[] = {
					w+float(xpos-100),     h, 1.0f, 1.0f,
					w+float(xpos-100),  0.0f, 0.0f, 1.0f,
					(float)(xpos-100),  0.0f, 0.0f, 0.0f,
					(float)(xpos-100),     h, 1.0f, 0.0f
					};
	DrawQuad(xyuv);
	// glDeleteTextures(1, &num);
	DoAlpha(0);
#endif
}


inline float ttf_float(short f)
{
	return float(f) /65536.0f;
}


// ------------------------------------------------------------- //
// ------------------------------------------------------------- //
void OpenGLRainbows::BuildFontTTF(const short * ck, TTF_FONT & font)
{
	font.Clear();
	font.chars.SetSize(256);
	for(;;)
	{
		int ic = int(*ck++);
		if(ic < 0) { return; }
		int nv = int(*ck++);
		int nt = int(*ck++);
		if(nv==0 || nt==0) { continue; }
		FONTCHAR& ch = font.chars[ic];
		ch.vec.SetSize(nv);
		ch.tri.SetSize(nt);
		for(int i=0; i<nv; ++i)
		{
			ch.vec[i].x = ttf_float(*ck++);
			ch.vec[i].y =-ttf_float(*ck++);
		}
		ch.minx = ttf_float(*ck++); ch.miny = ttf_float(*ck++);
		ch.maxx = ttf_float(*ck++); ch.maxy = ttf_float(*ck++);

		font.max_char_width = max(font.max_char_width, ch.width());
		font.max_ascender  = max(font.max_ascender,  ch.maxy);
		font.max_descender = max(font.max_descender, ch.miny);

		// maxy = ascender, miny=descender (or other way round ;) )
		for(int i=0; i<nt; ++i)
		{
			FONTCHAR::FONTTRI& t = ch.tri[i];
			t.i = *ck++; t.j = *ck++; t.k = *ck++;
		}
	}
	//assert(*ck == 0xddff); // drue dype font 
}



/// EGL - Wrappers



	
#if defined(HAVE_OPENGL_ES) && !defined(_DEBUG)
	
	// wrapper for EGL - otherwise code is totally unreadable
	struct EGL_VECTOR
	{
		GLfloat x,y,z,w;
		GLfloat u,v;
		unsigned char r,g,b,a;
	};
	EGL_VECTOR	gState= {0.0f,0.0f,0.0f,0.0f,  0.0f,0.0f,  255,255,255,255};
	Queue<EGL_VECTOR>	gEGLvecs;
	int                 gEGLmode=0;
	void eglBegin(int mode)
	{
		gEGLmode = mode;
		gEGLvecs.Clear(false);
	}
	void eglVertex3fv(const float* pv)
	{
		gState.x=*pv++;
		gState.y=*pv++;
		gState.z=*pv;
		gEGLvecs.Push(gState);
	}
	void eglVertex4fv(const float* pv)
	{
		gState.x=*pv++;
		gState.y=*pv++;
		gState.z=*pv++;
		gState.w=*pv;
		gEGLvecs.Push(gState);
	}
	void eglVertex4f(float x, float y, float z, float w)
	{
		gState.x=x;
		gState.y=y;
		gState.z=z;
		gState.w=z;
		gEGLvecs.Push(gState);
	}

	void eglEnd()
	{
		extern int g_nTexture; // the value you pass to X_MAKE3D(here, ...)

		glEnableClientState( GL_VERTEX_ARRAY );
		glVertexPointer( 4, GL_FLOAT, sizeof(EGL_VECTOR), (char*)&gEGLvecs[0].x);
		if(g_nTexture>=0)
		{
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
			glTexCoordPointer(2, GL_FLOAT, sizeof(EGL_VECTOR), (char*)&gEGLvecs[0].u);
 		}

		// Render stuff
		switch(gEGLmode)
		{
			case GL_QUADS:
			{
				for(int i=0; i<(int)gEGLvecs.Size(); i+=4)
				{
					glDrawArrays(GL_TRIANGLE_FAN, i, 4);
				}
				break;
			}
			case GL_LINES:
			{
				glDisableClientState( GL_TEXTURE_COORD_ARRAY );				// Disable Texture Coord Arrays
				glDrawArrays(GL_LINES, 0, (int)(gEGLvecs.Size()));
				break;
			}
			case GL_POINTS:
			{
				glDisableClientState( GL_TEXTURE_COORD_ARRAY );				// Disable Texture Coord Arrays
				glDrawArrays(GL_POINTS, 0, (int)gEGLvecs.Size());
				break;
			}
			default:
				glDrawArrays(gEGLmode, 0, gEGLvecs.Size());	// Draw All Of The Triangles At Once
		}
		// Disable Pointers
		glDisableClientState( GL_VERTEX_ARRAY );					// Disable Vertex Arrays
		glDisableClientState( GL_TEXTURE_COORD_ARRAY );				// Disable Texture Coord Arrays

		gEGLvecs.Clear(false);
	}
	#define glBegin eglBegin
	#define glVertex3fv eglVertex3fv
	#define glVertex4fv eglVertex4fv
	#define glVertex4f eglVertex4f
	#define glEnd eglEnd
#endif

#else // console
	void glb_rainbows_console_dummy() {};
#endif // GLB_CONSOLE

